/* cafe_config_service.c - implements the Cafe Configuration manager service */

/* Copyright (c) 2009-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
14nov14,tle  Fix blocking of TCF thread when emitting events (WB4-4453)
03aug14,sdt  Fixed checked-return warnings (US45196)
17jun14,jmp  Fixed ini_cafe_config_service() to only register cafe@1 service
             in value add (WB4-3814).
15may14,sdt  Fix freeing sysinfo data
02apr14,tle  Add getCapabilities command providing system information (US30548)
11mar14,sdt  Fix 'Checked return' warning
24feb14,sdt  Disable unexpected NO_EFFECT warnings

DESCRIPTION
  Provides access to the Cafe API via a TCF service.

  There are three threads of consequence in this logic.  First is the thread
  which calls ini_cafe_config_service().  There are no restrictions on which
  thread calls this routine.  All the other routines are meant to be called by
  one of the other two threads.

  The second thread of interest is TCF's event queue thread.  TCF invokes all
  command handlers on this thread.  Our service logic is careful to only
  manipulate TCF's InputStream and OutputStream objects on this thread

  The final thread is the Cafe service's own event thread.  In order to avoid
  blocking TCF's event thread, (which is shared between all TCF services) all
  Cafe API calls should issued from this thread.  So in practice, TCF calls a
  command handler on its event thread, this handler unmarshals the command's
  arguments, then dispatches an event to handle the command on Cafe's event
  thread.  This handler resolves the action, then issues another event on TCF's
  event thread to formulate the command's reply.
*/

#include <limits.h>
#include <assert.h>
#include <string.h>  /* for strlen() and friends */
#include <cafe/cafe.h>
#include <cafe/errors.h>
#include <cafe/byte_buffer.h>

#include <tcf/framework/trace.h>
#include <tcf/framework/json.h>
#include <tcf/framework/errors.h>
#include <tcf/framework/events.h>
#include <tcf/framework/event_queue.h>
#include <tcf/framework/myalloc.h>
#include <tcf/framework/proxy.h>
#include <tcf/framework/mdep-threads.h>
#include <tcf/services/streamsservice.h>

#if defined(HAVE_CAFE_CONFIG_H)
#include <cafe/config.h>
#endif
#include <cafe/ep/config_service.h>
#include <cafe/ep/aggregator.h>
#if !defined(__VXWORKS__)
#include <cafe/ep/tcf_upload.h>
#include <cafe/ep/tcf_download.h>
#include <cafe/ep/file_stream_reader.h>
#include <cafe/ep/file_stream_writer.h>
#include <cafe/ep/print.h>
#include <cafe/ep/text_file_provider.h>
#include <cafe/ep/realtime_ep.h>
#include <cafe/ep/timer_ep.h>
#include <cafe/ep/timestamp_normalizer.h>
#if defined(HAVE_PR_SET_NAME)
#include <sys/prctl.h>
#endif
#if defined(__linux__)
#include <cafe/ep/linux_trace.h>
#endif
#endif  /* if !defined(__VXWORKS__) */
#if !defined(VALUE_ADD)
#include <cafe/ep/sysinfo.h>
#endif
#include "tcf_utils.h"


#define LOG_CAFE_DEBUG         (0x80000000) /* bit used to request Cafe service
					       debugging output */

static const char *CONFIG_PREFIX = "AC";
static const char *EP_PREFIX = "EP";
static const char *const SERVICE_NAME = "Cafe";

/* When commands need to take the CAFE_LOCK, they're deferred to our own
   processing thread.  To assist in this, the command_buf variable is used to
   copy the incoming command bytes.  This variable should only be used from the
   TCF event thread. */
static ByteArrayOutputStream command_buf;
static OutputStream *command_buf_os;


/* A byte buffer used by routines called *only* from the cafe service event
   thread. */
static struct cafe_byte_buffer cafe_buf;


static bool json_read_type(InputStream *inp, const struct cafe_value_type *type,
			   cafe_any_value *result, int *err);
static void free_any_value(const struct cafe_value_type *type,
			   cafe_any_value *value);

Channel *cafe_config_channel_up;
Channel *cafe_config_channel_down;

/* Event queue used to dispatch any code which may need to take the CAFE_LOCK()
   for an extended period of time.  Event callbacks on this thread must post
   separate events to the TCF event queue if they need to send replies to a
   peer. */
static struct event_queue *cafe_event_queue;


#define CHANNEL_SESSION_LIST_ENTRY(PTR)					\
	CAFE_LIST_ENTRY((PTR), struct channel_session, channel_session_list)

struct channel_session {
	struct cafe_list_head channel_session_list;
	Channel *c;
	struct cafe_instrument *session;
};

static struct cafe_list_head channel_session_list_head =
	CAFE_LIST_INIT(channel_session_list_head);


static
int read_hex(InputStream *inp, int *err)
{
	int ch = read_stream(inp);
	if (ch >= '0' && ch <= '9') return ch - '0';
	if (ch >= 'A' && ch <= 'F') return ch - 'A' + 10;
	if (ch >= 'a' && ch <= 'f') return ch - 'a' + 10;
	*err = ERR_JSON_SYNTAX;
	return 0;
}


static
int read_hex_char(InputStream *inp, int *err)
{
	int n = read_hex(inp, err) << 12;
	if (*err != 0) return n;
	n |= read_hex(inp, err) << 8;
	if (*err != 0) return n;
	n |= read_hex(inp, err) << 4;
	if (*err != 0) return n;
	n |= read_hex(inp, err);
	return n;
}


static
unsigned int read_esc_char(InputStream *inp, char *utf8, int *err)
{
	int ch = read_stream(inp);

	switch (ch) {
	case '"': break;
	case '\\': break;
	case '/': break;
	case 'b': ch = '\b'; break;
	case 'f': ch = '\f'; break;
	case 'n': ch = '\n'; break;
	case 'r': ch = '\r'; break;
	case 't': ch = '\t'; break;
	case 'u':
		ch = read_hex_char(inp, err);
		if (*err != 0) return 0;
		break;
	default:
		*err = ERR_JSON_SYNTAX;
		return 0;
	}
	/* 'ch' can be wide character - convert it to UTF-8 sequence */
	if (ch < 0x80) {
		utf8[0] = (char)ch;
		return 1;
	}
	if (ch < 0x800) {
		utf8[0] = (char)((ch >> 6) | 0xc0);
		utf8[1] = (char)((ch & 0x3f) | 0x80);
		return 2;
	}
	utf8[0] = (char)((ch >> 12) | 0xe0);
	utf8[1] = (char)(((ch >> 6) & 0x3f) | 0x80);
	utf8[2] = (char)((ch & 0x3f) | 0x80);
	return 3;
}


static
void json_read_string_r(InputStream *inp, char *str, size_t size, int *err)
{
	unsigned int i;
	int ch = read_stream(inp);

	if (ch == 'n') {
		if ((read_stream(inp) != 'u') ||
		    (read_stream(inp) != 'l') ||
		    (read_stream(inp) != 'l'))
		{
			*err = ERR_JSON_SYNTAX;
			return;
		}
		str[0] = 0;
		return;
	}
	if (ch != '"') {
		*err = ERR_JSON_SYNTAX;
		return;
	}
	for (i = 0;;) {
		ch = read_stream(inp);
		if (ch < 0) {
			*err = ERR_JSON_SYNTAX;
			return;
		}
		if (ch == '"') break;
		if (ch == '\\') {
			char utf8[4];
			unsigned int l = read_esc_char(inp, utf8, err);
			unsigned int n;
			for (n = 0; n < l; n++, i++) {
				if (i < size - 1) str[i] = utf8[n];
			}
		} else {
			if (i < size - 1) str[i] = (char)ch;
			i++;
		}
	}
	if (i < size)
		str[i] = 0;
	else
		str[size - 1] = 0;
	return;
}


static
char *json_read_alloc_string_r(InputStream *inp, struct cafe_byte_buffer *buf,
			       int *err)
{
	char * str;
	int ch = read_stream(inp);

	if (ch == 'n') {
		if ((read_stream(inp) != 'u') ||
		    (read_stream(inp) != 'l') ||
		    (read_stream(inp) != 'l'))
		{
			*err = ERR_JSON_SYNTAX;
		}
		return NULL;
	}
	if (ch != '"') {
		*err = ERR_JSON_SYNTAX;
		return NULL;
	}
	for (buf->len = 0;;) {
		ch = read_stream(inp);
		if (ch < 0)  {
			*err = ERR_JSON_SYNTAX;
			return NULL;
		}
		if (ch == '"') break;
		if (ch == '\\') {
			char utf8[4];
			unsigned int l = read_esc_char(inp, utf8, err);
			unsigned int n;
			for (n = 0; n < l; ++n) 
			    (void)cafe_bbuf_write(buf, (utf8[n]));
		}
		else {
			(void)cafe_bbuf_write(buf, ch);
		}
	}
	(void)cafe_bbuf_write(buf, 0);
	str = (char *)loc_alloc(buf->len);
	memcpy(str, buf->data, buf->len);

	return str;
}


typedef void JsonArrayCallBack_r(InputStream *input, void *arg, int *err);

/**
 * json_read_array - generic read array function
 *
 * This function will call the call_back with inp and arg as
 *       arguments for each element of the list.
 * Return 0 if null, 1 otherwise
 */
static
int json_read_array_r(InputStream *inp, JsonArrayCallBack_r *call_back,
		      void *arg, int *err)
{
	int ch = read_stream(inp);

	if (ch == 'n') {
		if ((read_stream(inp) != 'u') ||
		    (read_stream(inp) != 'l') ||
		    (read_stream(inp) != 'l'))
		{
			*err = ERR_JSON_SYNTAX;
		}
		return 0;
	}
	if (ch != '[') {
		*err = ERR_PROTOCOL;
		return 1;
	}
	if (peek_stream(inp) == ']') {
		(void)read_stream(inp);
		return 1;
	}
	for (;;) {
		call_back(inp, arg, err);
		if (*err != 0) break;
		ch = read_stream(inp);
		if (ch == ',') continue;
		if (ch == ']') break;
		*err = ERR_JSON_SYNTAX;
		return 1;
	}
	return 1;
}


static
int json_read_boolean_r(InputStream *inp, int *err)
{
	int ch = read_stream(inp);

	if (ch == 'f') {
		if ((read_stream(inp) != 'a') ||
		    (read_stream(inp) != 'l') ||
		    (read_stream(inp) != 's') ||
		    (read_stream(inp) != 'e'))
		{
			*err = ERR_JSON_SYNTAX;
		}
		return 0;
	}

	if (ch == 't') {
		if ((read_stream(inp) != 'r') ||
		    (read_stream(inp) != 'u') ||
		    (read_stream(inp) != 'e'))
		{
			*err = ERR_JSON_SYNTAX;
		}
		return 1;
	}
	*err = ERR_JSON_SYNTAX;
	return 0;
}


static
long json_read_long_r(InputStream *inp, int *err)
{
	long res;
	int neg = 0;
	int ch = read_stream(inp);

	if (ch == '-') {
		neg = 1;
		ch = read_stream(inp);
	}
	if (ch < '0' || ch > '9') {
		*err = ERR_JSON_SYNTAX;
		return 0;
	}
	res = ch - '0';
	for (;;) {
		ch = peek_stream(inp);
		if (ch < '0' || ch > '9') break;
		read_stream(inp);
		res = res * 10 + (ch - '0');
	}
	if (neg) return -res;
	return res;
}


static
unsigned long json_read_ulong_r(InputStream *inp, int *err)
{
	unsigned long res;
	int neg = 0;
	int ch = read_stream(inp);

	if (ch == '-') {
		neg = 1;
		ch = read_stream(inp);
	}
	if (ch < '0' || ch > '9') {
		*err = ERR_JSON_SYNTAX;
		return 0;
	}
	res = ch - '0';
	for (;;) {
		ch = peek_stream(inp);
		if (ch < '0' || ch > '9') break;
		read_stream(inp);
		res = res * 10 + (ch - '0');
	}
	if (neg) return ~res + 1;
	return res;
}


static
char **json_read_alloc_string_array_r(InputStream *inp, int *cnt,
				      struct cafe_byte_buffer *buf, int *err)
{
	size_t *len_buf;
	unsigned int len_buf_size;
	unsigned int len_pos;
	unsigned int i;
	size_t j;
	char * str;
	char ** arr;
	int ch = read_stream(inp);

	if (ch == 'n') {
		if ((read_stream(inp) != 'u') ||
		    (read_stream(inp) != 'l') ||
		    (read_stream(inp) != 'l'))
		{
			*err = ERR_JSON_SYNTAX;
		}
		if (cnt) *cnt = 0;
		return NULL;
	} else if (ch != '[') {
		*err = ERR_PROTOCOL;
		return NULL;
	}

	buf->len = 0;
	len_buf = NULL;
	len_pos = 0;
	if (peek_stream(inp) == ']') {
	    (void)read_stream(inp);
	} else {
		for (len_buf_size = 0;;) {
			int ch = read_stream(inp);
			size_t len = 0;
			if (len_pos >= len_buf_size) {
				len_buf_size = len_buf_size == 0 ? 0x100 : len_buf_size * 2;
				len_buf = (size_t *)loc_realloc(len_buf, len_buf_size * sizeof(size_t));
			}
			if (ch == 'n') {
				if ((read_stream(inp) != 'u') ||
				    (read_stream(inp) != 'l') ||
				    (read_stream(inp) != 'l'))
				{
					goto fail;
				}
			} else if (ch != '"')
				goto fail;
			else {
				size_t buf_pos0;

				for (buf_pos0 = buf->len;;) {
					ch = read_stream(inp);
					if (ch < 0)
						goto fail;
					if (ch == '"') break;
					if (ch == '\\') {
						char utf8[4];
						unsigned int l = read_esc_char(inp, utf8, err);
						unsigned int n;
						for (n = 0; n < l; ++n)
							cafe_bbuf_write(buf, utf8[n]);
					} else
						(void)cafe_bbuf_write(buf, ch);
				}
				len = buf->len - buf_pos0;
			}
			cafe_bbuf_write(buf, 0);
			len_buf[len_pos++] = len;
			ch = read_stream(inp);
			if (ch == ',') continue;
			if (ch == ']') break;
			goto fail;
		}
	}
	(void)cafe_bbuf_write(buf, 0);
	arr = (char **)loc_alloc((len_pos + 1) * sizeof(char *) + buf->len);
	str = (char *)(arr + len_pos + 1);
	memcpy(str, buf->data, buf->len);
	j = 0;
	for (i = 0; i < len_pos; i++) {
		arr[i] = str + j;
		j += len_buf[i] + 1;
	}
	loc_free(len_buf);
	arr[len_pos] = NULL;
	if (cnt) *cnt = len_pos;
	return arr;

fail:
	loc_free(len_buf);
	*err = ERR_JSON_SYNTAX;
	return NULL;
}


#define service_error(cmd_args, err)		\
	service_error_str((cmd_args), (err), NULL)


#define service_error_str(cmd_args, err, msg)				\
	do {								\
		post_err_reply((cmd_args), (err), (msg));		\
		if (cafe_islocked())					\
			CAFE_UNLOCK();					\
		return;							\
	} while(0)


#define handler_error(err_ptr, err, msg)				\
	do {								\
		*(err_ptr) = err;					\
		(void)printf("%s: %s (%s: %u)\n", errno_to_str((err)), 	\
		             (msg), __FUNCTION__, __LINE__); 		\
		return;							\
	} while(0)


/*******  objects passed to event callbacks  ******/

struct cafe_handler_args {
	struct event_node enode;
	Channel *c;
	const char *token;
	ByteArrayOutputStream os_reply;
	char *command;
	size_t command_sz;
};


static
void free_handler_args(struct cafe_handler_args *args)
{
	loc_free(args->token);
	loc_free(args);
}


/**
 *  Emit a standard error message to the given output stream.
 */
static
void write_err_reply(OutputStream *out, const char *token, unsigned int errnum,
		     const char *err_msg)
{
	write_stringz(out, "R");
	write_stringz(out, token);
	(void)set_errno(errnum, err_msg);
	write_errno(out, errno);
	write_stream(out, MARKER_EOM);
}


/**
 * Defers the command currently queued on the given Channel to the specified
 * handler routine, to be executed on the Cafe event thread.
 */
static
void defer_event(EventQueueCallBack *handler, Channel *c, const char *token)
{
	struct cafe_handler_args *arg =
		loc_alloc_zero(sizeof(struct cafe_handler_args));
	int in;

	assert(is_dispatch_thread());
	arg->enode.handler = handler;
	arg->c = c;
	arg->token = loc_strdup(token);

	in = read_stream(&c->inp);
	while (in != MARKER_EOM) {
		write_stream(command_buf_os, in);
		in = read_stream(&c->inp);
	}

	get_byte_array_output_stream_data(&command_buf,
					  &arg->command, &arg->command_sz);

	channel_lock(c);
	event_queue_post_event(cafe_event_queue, &arg->enode);
}


/*******  TCF event thread callbacks  ******/

struct error_reply_args {
	Channel *c;
	const char *token;
	int err;
	const char *description;
};


static
void error_reply(void *arg)
{
	struct error_reply_args *reply = arg;
	Channel *c = reply->c;
	write_err_reply(&c->out, reply->token,
			reply->err, reply->description);
	loc_free(reply->description);
	loc_free(reply->token);
	loc_free(reply);
	channel_unlock(c);
}


static
void post_err_reply(struct cafe_handler_args *args, unsigned int errnum,
		    const char *err_msg)
{
	struct error_reply_args *reply =
		loc_alloc(sizeof(struct error_reply_args));

	loc_free(args->command);
	reply->c = args->c;
	reply->token = args->token;
	loc_free(args);
	reply->err = errnum;
	if (cafe_islocked() && (cafe_error_cur.rc != CAFE_OK)) {
		struct cafe_byte_buffer buf;
		int rc = cafe_bbuf_init(&buf, cafe_stdlib_allocator, 128);
		if (rc != CAFE_OK) {
			(void)fprintf(stderr, "Error initializing error message buffer: %s\n", cafe_strerror(rc));
			reply->description = NULL;
			goto done;
		}
		rc = cafe_error_bbuf_write(&cafe_error_cur, &buf, err_msg,
					   ", due to\n\t");
		if (rc != CAFE_OK) {
			(void)fprintf(stderr, "Error buffering error message: %s\n", cafe_strerror(rc));
			reply->description = NULL;
			cafe_bbuf_release(&buf);
			goto done;
		}

		cafe_error_clear(NULL);
		reply->description = buf.data;
	} else
		reply->description =
			err_msg == NULL ? NULL : loc_strdup(err_msg);

done:
	post_event(error_reply, reply);
}


/**
 * TCF event posted by Cafe event handlers which have no return value.
 */
static
void simple_reply(void *arg)
{
	struct cafe_handler_args *reply = arg;
	Channel *c = reply->c;
	write_stringz(&c->out, "R");
	write_stringz(&c->out, reply->token);
	write_errno(&c->out, 0);
	write_stream(&c->out, MARKER_EOM);
	free_handler_args(reply);
	channel_unlock(c);
}


/**
 * TCF event posted by Cafe event handlers which want to express return values.
 */
static
void handler_reply(void *arg)
{
	struct cafe_handler_args *reply = arg;
	size_t size;
	char *reply_str;
	OutputStream *out = &reply->c->out;

	get_byte_array_output_stream_data(&reply->os_reply, &reply_str, &size);

	write_stringz(out, "R");
	write_stringz(out, reply->token);
	write_errno(out, 0);
	write_block_stream(out, reply_str, size);
	write_stream(out, MARKER_EOM);
	channel_unlock(reply->c);

	loc_free(reply_str);
	free_handler_args(reply);
}


/**
 *  Write the given context id to an OutputStream.  The prefix argument should
 * indicate the object or context type.
 */
static
void write_id(OutputStream *out, const char *prefix, uintptr_t id)
{
	char buf[256];

	(void)snprintf(buf, sizeof(buf), "%s%" PRIuPTR, prefix, id);
	json_write_string(out, buf);
}


/**
 *  Emit a json-formatted version of the given string array.
 */
static
void write_string_list(OutputStream *out, const char *list[], unsigned int num)
{
	unsigned int i;
	bool is_first = true;

	write_stream(out, '[');
	for (i = 0; i < num; ++i) {
		if (!is_first) {
			write_stream(out, ',');
		} else
			is_first = false;
		json_write_string(out, list[i]);
	}
	write_stream(out, ']');
}


/**
 *  Emit a json-formatted description of the given Event Processor
 * specification.
 */
static
void write_spec(OutputStream *out, const struct cafe_ep_spec *spec)
{
	unsigned int i;
	bool is_first = true;
	assert(spec);

	write_stream(out, '{');

	json_write_string(out, "name");
	write_stream(out, ':');
	json_write_string(out, spec->name);

	write_stream(out, ',');
	json_write_string(out, "workarea");
	write_stream(out, ':');
	json_write_string(out, spec->workarea);

	write_stream(out, ',');
	json_write_string(out, "types");
	write_stream(out, ':');
	json_write_string(out, spec->types);

	write_stream(out, ',');
	json_write_string(out, "input_events");
	write_stream(out, ':');
	write_stream(out, '[');
	for (i = 0; i < spec->n_input_events; ++i) {
		const struct cafe_ep_spec_input_evt *event =
			&spec->input_events[i];
		if (!is_first) {
			write_stream(out, ',');
		} else
			is_first = false;
		write_string_list(out, event->type, event->num_types);
	}
	write_stream(out, ']');

	/* NOTE: order of output_events array must match order of
	 * output_event_od_set_map */
	write_stream(out, ',');
	json_write_string(out, "output_events");
	write_stream(out, ':');
	write_stream(out, '[');
	is_first = true;
	for (i = 0; i < spec->n_output_events; ++i) {
		const struct cafe_ep_spec_output_evt *event =
			&spec->output_events[i];
		if (!is_first) {
			write_stream(out, ',');
		} else
			is_first = false;
		json_write_string(out, event->type);
	}
	write_stream(out, ']');

	write_stream(out, ',');
	json_write_string(out, "orthogonal_types");
	write_stream(out, ':');
	write_string_list(out, spec->output_odata, spec->n_output_odata);

	write_stream(out, ',');
	json_write_string(out, "parameter_types");
	write_stream(out, ':');
	write_string_list(out, NULL, 0);

	write_stream(out, ',');
	json_write_string(out, "is_concentrator");
	write_stream(out, ':');
	if (spec->inject_event == NULL)
		json_write_boolean(out, false);
	else
		json_write_boolean(out, true);

	if (spec->n_output_odata_sets != 0) {
		/* NOTE: order of output_event_od_set_map must match
		 * order of output_events array */
		write_stream(out, ',');
		json_write_string(out, "output_event_od_set_map");
		write_stream(out, ':');
		write_stream(out, '[');
		is_first = true;
		for (i = 0; i < spec->n_output_events; ++i) {
			const struct cafe_ep_spec_output_evt *event =
				&spec->output_events[i];
			if (!is_first) {
				write_stream(out, ',');
			} else
				is_first = false;
			json_write_long(out, event->od_set_idx);
		}
		write_stream(out, ']');

		write_stream(out, ',');
		json_write_string(out, "od_sets");
		write_stream(out, ':');
		write_stream(out, '[');
		is_first = true;
		for (i = 0; i < spec->n_output_odata_sets; ++i) {
			const struct cafe_ep_spec_output_od_set *od_set =
				&spec->output_odata_sets[i];
			int set_size;
			int j;
			bool is_first_od;

			if (!is_first) {
				write_stream(out, ',');
			} else
				is_first = false;

			if (od_set->od_items == NULL)
				set_size = 0;
			else
				set_size = od_set->length;

			write_stream(out, '[');
			is_first_od = true;

			for (j = 0; j < set_size; ++j) {
				if (!is_first_od) {
					write_stream(out, ',');
				} else {
					is_first_od = false;
				}

				json_write_long(out, od_set->od_items[j]);
			}
			write_stream(out, ']');
		}
		write_stream(out, ']');
	}

	write_stream(out, '}');
}


/* argument values necessary for the read_dynarray() callback */
struct read_array_arg {
	const struct cafe_value_type *array_type;
	struct cafe_vm_dynarray *a;
};


/**
 *  Given a read_array_arg structure, populate its dynarray with values read
 * from the given InputStream.  This routine is meant to be used as a callback
 * to json_read_array().
 */
static
void read_dynarray(InputStream *inp, void *args, int *err)
{
	struct read_array_arg *param = args;
	const struct cafe_value_type *element_type =
		cafe_element_type(param->array_type);
	cafe_any_value v;
	void *new_element;
	int rc;

	if (!json_read_type(inp, element_type, &v, err))
		handler_error(err, *err, "unhandled setting type");

	rc = cafe_vm_dynarray_set_size(param->a, param->a->l + 1,
				       param->array_type);
	if (rc != CAFE_OK)
		handler_error(err, rc, "unable to resize dynamic array");

	new_element =
		(char *) param->a->p + ((param->a->l - 1) * element_type->size);
	(void)cafe_value_store_type(element_type, new_element, &v,
			      element_type->ident->ns->allocator);
	free_any_value(element_type, &v);
}


/**
 *  Use the given value type to interpret the next value available on the given
 * InputStream.
 */
static
bool json_read_type(InputStream *inp, const struct cafe_value_type *type,
		    cafe_any_value *result, int *err)
{
	switch(type->kind) {
	case CAFE_TYPE_BOOL:
		result->ull = json_read_boolean_r(inp, err);
		break;

	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_SHORT:
	case CAFE_TYPE_INT:
	case CAFE_TYPE_LONG:
	case CAFE_TYPE_LONGLONG:
		result->ll = json_read_long_r(inp, err);
		break;

	case CAFE_TYPE_USHORT:
	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_UINT:
	case CAFE_TYPE_ULONG:
	case CAFE_TYPE_ULONGLONG:
		result->ull = json_read_ulong_r(inp, err);
		break;

	case CAFE_TYPE_FLOAT:
	case CAFE_TYPE_DOUBLE:
		result->d = json_read_double(inp);
		break;

	case CAFE_TYPE_STRING:
		result->str = json_read_alloc_string_r(inp, &cafe_buf, err);
		break;

	case CAFE_TYPE_DYNARRAY:
	{
		struct read_array_arg arg;

		memset(&result->dary, 0, sizeof(result->dary));
		arg.array_type = type;
		arg.a = &result->dary;
		(void)json_read_array_r(inp, read_dynarray, &arg, err);
		if (*err != 0)
			return false;
		break;
	}

	case CAFE_TYPE_BLOB:
	case CAFE_TYPE_DICT:
	default:
		return false;
	}
	return true;
}


/**
 *  Releases any memory associated with a type read by json_read_type().
 */
static
void free_any_value(const struct cafe_value_type *type,
		    cafe_any_value *value)
{
	switch(type->kind) {
	case CAFE_TYPE_STRING:
		loc_free(value->str);
		break;

	case CAFE_TYPE_DYNARRAY:
	    (void)cafe_vm_dynarray_set_size(&value->dary, 0, type);
		break;

	case CAFE_TYPE_BLOB:
	case CAFE_TYPE_DICT:
		/* placeholders for when we start supporting parameters of
		   these types */

	default:
		break;
	}
}


/**
 *  Format the given value (based on the provided type) as a json value, and
 * write it to an OutputStream.
 */
static
bool json_write_type(OutputStream *out, const struct cafe_value_type *type,
		     const void *value)
{
	switch(type->kind) {
	case CAFE_TYPE_BOOL:
		json_write_boolean(out, *(bool *) value);
		break;

	case CAFE_TYPE_CHAR:
		json_write_long(out, *(char *) value);
		break;
	case CAFE_TYPE_SHORT:
		json_write_long(out, *(short *) value);
		break;
	case CAFE_TYPE_INT:
		json_write_long(out, *(int *) value);
		break;
	case CAFE_TYPE_LONG:
		json_write_long(out, *(long *) value);
		break;
	case CAFE_TYPE_LONGLONG:
		json_write_int64(out, *(long long *) value);
		break;

	case CAFE_TYPE_USHORT:
		json_write_ulong(out, *(unsigned short *) value);
		break;
	case CAFE_TYPE_UCHAR:
		json_write_ulong(out, *(unsigned char *) value);
		break;
	case CAFE_TYPE_UINT:
		json_write_ulong(out, *(unsigned int *) value);
		break;
	case CAFE_TYPE_ULONG:
		json_write_ulong(out, *(unsigned long *) value);
		break;
	case CAFE_TYPE_ULONGLONG:
		json_write_uint64(out, *(unsigned long long *) value);
		break;

	case CAFE_TYPE_FLOAT:
		json_write_double(out, *(float *) value);
		break;
	case CAFE_TYPE_DOUBLE:
		json_write_double(out, *(double *) value);
		break;

	case CAFE_TYPE_STRING:
		json_write_string(out, *(char **) value);
		break;

	case CAFE_TYPE_DYNARRAY:
	{
		const struct cafe_vm_dynarray *a = value;
		const char *element = a->p;
		const struct cafe_value_type *element_type =
			cafe_element_type(type);
		unsigned int i;
		bool is_first = true;

		write_stream(out, '[');
		for (i = 0; i < a->l; ++i) {
			if (!is_first) {
				write_stream(out, ',');
			} else
				is_first = false;
			(void)json_write_type(out, element_type, element);
			element += element_type->size;
		}
		write_stream(out, ']');

		break;
	}

	case CAFE_TYPE_BLOB:
	case CAFE_TYPE_DICT:
	default:
		return false;
	}
	return true;
}


/* argument values necessary for the read_settings() callback */
struct read_settings_arg {
	struct cafe_value_type *workarea_vt;
	struct cafe_vm_dynarray settings_array;
	const char *err;
};


/**
 *  Given a read_settings_arg structure, populate its settings_array with
 * name/value pairs read from the json-formatted input stream.  This routine is
 * meant to be used as a callback to json_read_array().
 */
static
void read_settings(InputStream *inp, void *args, int *err)
{
	struct read_settings_arg *param = args;
	struct cafe_vm_dynarray *array = &param->settings_array;
	struct cafe_cfg_setting *settings;
	struct cafe_vm_attr *attr;

	if (read_stream(inp) != '[')
		handler_error(err, ERR_JSON_SYNTAX, "missing open bracket");
	if (array->l >= array->c) {
		array->c = array->c ? array->c * 2 : 8;
		array->p = loc_realloc(array->p, array->c * sizeof(*settings));
	}

	settings = array->p;
	settings[array->l].name = json_read_alloc_string_r(inp, &cafe_buf, err);

	attr = cafe_type_attr_by_name(param->workarea_vt,
				      settings[array->l].name);
	if (attr == NULL) {
		char buf[256];
		(void)snprintf(buf, sizeof buf,
			 "unrecognized parameter name (%s)",
			 settings[array->l].name);
		handler_error(err, ERR_JSON_SYNTAX, buf);
	}
	if (read_stream(inp) != ',')
		handler_error(err, ERR_JSON_SYNTAX, "missing comma");

	if (!json_read_type(inp, cafe_attr_type(attr),
			    &settings[array->l].value, err))
		handler_error(err, *err, "unhandled setting type");


	if (read_stream(inp) != ']')
		handler_error(err, ERR_JSON_SYNTAX, "missing end bracket");
	++array->l;
}


/**
 * Releases all dynamically allocated memory in an array of settings objects
 * created by read_settings().
 */
static
void free_settings(const struct cafe_value_type *workarea_vt,
		   struct cafe_vm_dynarray *settings_array)
{
	struct cafe_cfg_setting *settings = settings_array->p;
	unsigned int i;

	for (i = 0; i < settings_array->l; ++i) {
		struct cafe_vm_attr *ident;
		ident = cafe_type_attr_by_name(workarea_vt, settings[i].name);
		assert(ident != NULL);
		free_any_value(cafe_attr_type(ident), &settings[i].value);
		loc_free(settings[i].name);
	}
	loc_free(settings);
}


/**
 *  Reads values from a json-object-formatted stream into individual members
 * of a cafe value data structure.  Meant to be used as a callback from
 * json_read_struct().
 */
static
void json_read_struct_member(InputStream *inp, const char *name, void *args)
{
	struct cafe_value_data *value_data = args;
	const struct cafe_vm_attr *attribute;
	cafe_any_value val;
	void *dest;
	int rc = 0;

	attribute = cafe_type_attr_by_name(value_data->type, name);
	if (attribute == NULL) {
		(void)printf("unhandled structure member name (%s: %u)",
		       __FUNCTION__, __LINE__);
		value_data->type = NULL;
		return;
	}

	//td: need to ensure memory allocated here is freed later
	if (!json_read_type(inp, cafe_attr_type(attribute), &val, &rc)) {
		(void)printf("unhandled setting type (%s: %u)", __FUNCTION__, __LINE__);
		value_data->type = NULL;
		return;
	}

	dest = value_data->u.data + attribute->offset;
	rc = cafe_value_store_type(cafe_attr_type(attribute), dest, &val,
				   value_data->type->ident->ns->allocator);
	if (rc != CAFE_OK) {
		(void)printf("unable to populate structure member: %s (%s: %u)",
		       cafe_strerror(rc), __FUNCTION__, __LINE__);
		value_data->type = NULL;
	}
}


/**
 *  Emit a json-formatted description of the given Event Processor
 * specification.
 *
 * TD: need a more useful representation of EP instances.  Also, we need a
 * reasonable way to query for the Configuration state.
 */
static
void write_config(OutputStream *out, const struct cafe_instrument *config)
{
	struct cafe_list_head *list_item;
	bool is_first = true;
	write_stream(out, '{');

	json_write_string(out, "name");
	write_stream(out, ':');
	json_write_string(out, config->name);

	write_stream(out, ',');
	json_write_string(out, "ep_instances");
	write_stream(out, ':');
	write_stream(out, '[');
	cafe_list_foreach(list_item, &config->inst_list_head) {
		struct cafe_ep_inst *inst =
			SESSION_INST_LIST_ENTRY(list_item);
		if (!is_first) {
			write_stream(out, ',');
		} else
			is_first = false;
		write_id(out, EP_PREFIX, (uintptr_t) inst);
	}
	write_stream(out, ']');

	write_stream(out, ',');
	json_write_string(out, "state");
	write_stream(out, ':');
	json_write_string(out, "RUNNING");
	write_stream(out, '}');
	write_stream(out, 0);
}


/**
 * Write all attributes of the given workarea, with the given nature, to a
 * TCF OutputStream.
 */
static
void write_settings(OutputStream *out,
		    const struct cafe_value_type *workarea_vt,
		    const char *workarea,
		    const char *nature)
{
	const struct cafe_vm_attr *attr;
	bool is_first = true;

	if (workarea_vt == NULL) {
		write_stringz(out, "[]");
		return;
	}

	write_stream(out, '[');
	cafe_type_attr_foreach(attr, workarea_vt) {
		if (cafe_find_nature_str(attr->ident, nature) != NULL) {
			if (!is_first) {
				write_stream(out, ',');
			} else
				is_first = false;
			write_stream(out, '[');
			json_write_string(out, attr->ident->name);
			write_stream(out, ',');
			if (!json_write_type(out, cafe_attr_type(attr),
					     workarea + attr->offset)) {
				/* coverity[array_null] */
				assert(!"unhandled setting type");

			}
			write_stream(out, ']');
		}
	}

	write_stream(out, ']');
	write_stream(out, 0);
}


/**
 * Write a list of settings values directly into the given EP instance's
 * workarea.
 */
static
int settings_commit(struct cafe_ep_inst *inst, unsigned int n_settings,
		    const struct cafe_cfg_setting *settings)
{
	unsigned int i;
	int rc;

	for (i = 0; i < n_settings; ++i) {
		rc = cafe_value_store(inst->workarea_type,
				      inst->workarea->u.data,
				      settings[i].name,
				      &settings[i].value,
				      inst->allocator);
		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}


/*******  utility routines  ******/

/**
 *  Given a context name string ("VS0", "AL42", "AC1", etc.) and the expected
 * prefix, parse out the integral context id.  Returns true iff the parsing is
 * successful.
 */
static
bool str2id(const char *prefix, const char *id_str, uintptr_t *id)
{
	const size_t prefix_len = strlen(prefix);

	if (memcmp(id_str, prefix, prefix_len))
		return false;
	*id = strtoul(id_str + prefix_len, 0, 10);

	if ((*id == ULONG_MAX) && (errno == ERANGE))
		return false;

	return true;
}


/**
 *  Locate a Configuration instance by name.  Returns NULL on error.
 */
static
struct cafe_instrument *find_config(uintptr_t handle)
{
	struct cafe_instrument_iter iter;
	struct cafe_instrument *config = cafe_instrument_iter_first(&iter);

	while (config != NULL) {
		if ((uintptr_t) config == handle)
			return config;

		config = cafe_instrument_iter_next(&iter);
	}
	return NULL;
}


/**
 *  Find the EP corresponding to the given handle in the given Configuration.
 */
static
struct cafe_ep_inst *find_config_ep(const struct cafe_instrument *config,
				    uintptr_t ep_handle)
{
	struct cafe_list_head *list_item;
	cafe_list_foreach(list_item, &config->inst_list_head) {
		struct cafe_ep_inst *inst =
			SESSION_INST_LIST_ENTRY(list_item);
		if ((uintptr_t) inst == ep_handle)
			return inst;
	}
	return NULL;
}


/**
 *  Return a handle to the given spec's workarea type.
 */
static
int workarea_type(const struct cafe_ep_spec *spec, struct cafe_namespace *ns,
		  struct cafe_value_type **type)
{
	int rc;

	if (spec->workarea == NULL) {
		*type = NULL;
		return CAFE_OK;
	}
	rc = cafe_type_parse(ns, spec->types, NULL);
	if (rc != CAFE_OK)
		return rc;

	return cafe_type_parse(ns, spec->workarea, type);
}


static
struct cafe_event_descr *get_event_descr(struct cafe_instrument *config,
					 const struct cafe_value_type *vt)
{
	unsigned int i;
	for (i = 0; i < config->events.ind; ++i) {
		if (vt == config->events.list[i]->types[0])
			return config->events.list[i];
	}
	return NULL;
}


/*******  Cafe service event thread callbacks  ******/

static
void list_eps_handler(struct event_node *arg)
{
	struct cafe_handler_args *list_args =
		(struct cafe_handler_args *) arg;
	OutputStream *reply;
	struct cafe_ep_spec *spec;
	bool is_first = true;
	struct cafe_ep_spec_iter iter;

	reply = create_byte_array_output_stream(&list_args->os_reply);

	/* list available Event Processors */
	write_stream(reply, '[');
	CAFE_LOCK();
	spec = cafe_ep_spec_iter_first(&iter);
	while (spec != NULL) {
		if (!is_first) {
			write_stream(reply, ',');
		} else
			is_first = false;
		json_write_string(reply, spec->name);
		spec = cafe_ep_spec_iter_next(&iter);
	}
	CAFE_UNLOCK();
	write_stream(reply, ']');
	write_stream(reply, 0);
	loc_free(list_args->command);
	post_event(handler_reply, list_args);
}


static
void describe_ep_handler(struct event_node *arg)
{
	struct cafe_ep_spec *spec;
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	char id[64];
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);
	OutputStream *reply =
		create_byte_array_output_stream(&cmd_args->os_reply);

	json_read_string_r(in, id, sizeof(id), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	/* find and describe the requested Event Processor */
	CAFE_LOCK();
	spec = cafe_ep_spec_find(id);
	if (spec == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested EP specification");
	loc_free(cmd_args->command);

	write_spec(reply, spec);
	write_stream(reply, 0);
	CAFE_UNLOCK();

	post_event(handler_reply, cmd_args);
}


static
void list_configs_handler(struct event_node *arg)
{
	struct cafe_instrument *config;
	struct cafe_instrument_iter iter;
	bool is_first = true;
	struct cafe_handler_args *list_args =
		(struct cafe_handler_args *) arg;
	OutputStream *reply =
		create_byte_array_output_stream(&list_args->os_reply);

	/* list existing Configurations */
	write_stream(reply, '[');
	CAFE_LOCK();
	config = cafe_instrument_iter_first(&iter);
	while (config != NULL) {
		if (!is_first) {
			write_stream(reply, ',');
		} else
			is_first = false;
		write_id(reply, CONFIG_PREFIX, (uintptr_t) config);
		config = cafe_instrument_iter_next(&iter);
	}
	CAFE_UNLOCK();
	write_stream(reply, ']');
	write_stream(reply, 0);

	post_event(handler_reply, list_args);
}


static
void describe_config_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	uintptr_t config_handle;
	char id[64];
	int rc = 0;
	OutputStream *reply =
		create_byte_array_output_stream(&cmd_args->os_reply);
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, id, sizeof(id), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	/* find and describe the requested Configuration */
	if (!str2id(CONFIG_PREFIX, id, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	CAFE_UNLOCK();

	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Requested instrument not found");
	loc_free(cmd_args->command);

	write_config(reply, config);
	post_event(handler_reply, cmd_args);
}


struct handler_event_args {
	struct event_node enode;
	struct cafe_instrument *session;
	struct cafe_ep_inst *inst;
	const char *name;
};


static
void tcf_handler_event(void *arg)
{
	struct handler_event_args *args = (struct handler_event_args *)arg;
	struct cafe_list_head *list_item;
	Channel *c = NULL;

	CAFE_LOCK();
	cafe_list_foreach(list_item, &channel_session_list_head) {
		struct channel_session *map =
			CHANNEL_SESSION_LIST_ENTRY(list_item);
		if (map->session == args->session) {
			c = map->c;
			break;
		}
	}
	CAFE_UNLOCK();

	if (c != NULL && c->bcg != NULL) {
		OutputStream *out = &c->bcg->out;
		write_stringz(out, "E");
		write_stringz(out, SERVICE_NAME);
		write_stringz(out, args->name);
		write_id(out, CONFIG_PREFIX, (uintptr_t) args->session);
		write_stream(out, 0);
		write_id(out, EP_PREFIX, (uintptr_t) args->inst);
		write_stream(out, 0);
		write_stream(out, MARKER_EOM);
	}
	loc_free(args);
}


static
void cafe_handler_event(struct event_node *arg)
{
	post_event(tcf_handler_event, arg);
}


static
void ep_inst_event(struct cafe_ep_inst *inst, const char *name)
{
	struct handler_event_args *args = loc_alloc_zero(sizeof *args);

	args->enode.handler = cafe_handler_event;
	args->session = inst->session;
	args->inst = inst;
	args->name = name;
	event_queue_post_event(cafe_event_queue, &args->enode);
}


static
void ep_inst_started(struct cafe_ep_inst *inst)
{
	ep_inst_event(inst, "epStarted");
}


static
void ep_inst_stopped(struct cafe_ep_inst *inst)
{
	ep_inst_event(inst, "epStopped");
}


struct status_event_args {
	struct cafe_ep_inst *inst;
	enum cafe_log_level l;
	const char *msg;
};


static
void status_event_handler(void *arg)
{
	const struct status_event_args *a = arg;
	struct cafe_list_head *list_item;
	Channel *c = NULL;
	int local_error;

	CAFE_LOCK();
	cafe_list_foreach(list_item, &channel_session_list_head) {
		struct channel_session *map =
			CHANNEL_SESSION_LIST_ENTRY(list_item);
		if (map->session == a->inst->session) {
			c = map->c;
			break;
		}
	}
	CAFE_UNLOCK();
	if ((c == NULL) || (c->bcg == NULL))
		goto out;

	write_stringz(&c->bcg->out, "E");
	write_stringz(&c->bcg->out, SERVICE_NAME);
	write_stringz(&c->bcg->out, "status");
	json_write_ulong(&c->bcg->out, a->l);
	write_stream(&c->bcg->out, 0);
	write_id(&c->bcg->out, CONFIG_PREFIX, (uintptr_t) a->inst->session);
	write_stream(&c->bcg->out, 0);
	write_id(&c->bcg->out, EP_PREFIX, (uintptr_t) a->inst);
	write_stream(&c->bcg->out, 0);
	local_error = set_errno(ERR_OTHER, a->msg);
	write_error_object(&c->bcg->out, local_error);
	write_stream(&c->bcg->out, 0);
	write_stream(&c->bcg->out, MARKER_EOM);

out:
	loc_free(a->msg);
	loc_free(a);
}


/**
 *  session->status_notifier callback; relays any reported status to clients.
 */
static
void ep_inst_status(struct cafe_ep_inst *inst, enum cafe_log_level l,
		    const char *msg)

{
	struct status_event_args *args = loc_alloc(sizeof(*args));

	args->inst = inst;
	args->l = l;
	args->msg = loc_strdup(msg);
	post_event(status_event_handler, args);
}


static
void config_create_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	char id[64];
	int rc = 0;
	struct channel_session *map;
	struct ByteArrayInputStream input_bytes;
	OutputStream *reply =
		create_byte_array_output_stream(&cmd_args->os_reply);
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, id, sizeof(id), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	CAFE_LOCK();
	rc = cafe_instrument_create(id, &config);
	if (rc == CAFE_OK) {
		config->start_notifier = ep_inst_started;
		config->stop_notifier = ep_inst_stopped;
		config->status_notifier = ep_inst_status;
	} else
		service_error_str(cmd_args, rc, "Unable to create instrument");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);

	/* Return log handle */
	write_id(reply, CONFIG_PREFIX, (uintptr_t) config);
	write_stream(reply, 0);

	/* Register channel/session map so we can delete the session
	 * when the channel is closed */
	map = loc_alloc_zero(sizeof *map);
	map->c = cmd_args->c;
	map->session = config;
	cafe_list_add_first(&map->channel_session_list, &channel_session_list_head);

	post_event(handler_reply, cmd_args);
}


static
void config_delete_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	uintptr_t config_handle;
	char id[64];
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, id, sizeof(id), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	if (!str2id(CONFIG_PREFIX, id, &config_handle)) {
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");
	}

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	rc = cafe_instrument_delete(config);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc, "Unable to delete configuration");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);
	post_event(simple_reply, cmd_args);
}


static
void config_start_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	uintptr_t config_handle;
	char id[64];
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, id, sizeof(id), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	if (!str2id(CONFIG_PREFIX, id, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	rc = cafe_instrument_refresh_output_event_actions(config);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, ERR_OTHER,
				  "Unable to refresh output events");

	rc = cafe_instrument_start(config);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc, "Unable to start configuration");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);
	post_event(simple_reply, cmd_args);
}


static
void config_stop_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	uintptr_t config_handle;
	char id[64];
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, id, sizeof(id), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	if (!str2id(CONFIG_PREFIX, id, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	rc = cafe_instrument_stop(config);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc, "Unable to stop configuration");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);
	post_event(simple_reply, cmd_args);
}


static
void ep_create_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst;
	struct read_settings_arg settings_arg;
	char config_name[64], spec_name[256];
	uintptr_t config_handle;
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);
	OutputStream *reply =
		create_byte_array_output_stream(&cmd_args->os_reply);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, spec_name, sizeof(spec_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	spec = cafe_ep_spec_find(spec_name);
	if (spec == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested EP");

	rc = workarea_type(spec, config->ns, &settings_arg.workarea_vt);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "error parsing spec types");

	memset(&settings_arg.settings_array, 0,
	       sizeof(settings_arg.settings_array));
	settings_arg.err = NULL;
	(void)json_read_array_r(in, read_settings, &settings_arg, &rc);
	if (rc != 0)
		service_error(cmd_args, rc);

	if (settings_arg.err != NULL) {
		free_settings(settings_arg.workarea_vt,
			      &settings_arg.settings_array);
		service_error_str(cmd_args, ERR_JSON_SYNTAX, settings_arg.err);
	}

	/* WRS_TODO: proxy_get_target_channel is not called on the TCF
	 * dispatch thread.  This will work because as long as one
	 * channel is locked both of them will stay around, but it
	 * seems error prone if the proxy implementation change. */
	assert(cafe_config_channel_up == NULL);
	assert(cafe_config_channel_down == NULL);
	cafe_config_channel_up = cmd_args->c;
	cafe_config_channel_down = proxy_get_target_channel(cmd_args->c);
	rc = cafe_ep_inst_create(config, spec,
				 (unsigned int) settings_arg.settings_array.l,
				 settings_arg.settings_array.p, &inst);
	cafe_config_channel_up = NULL;
	cafe_config_channel_down = NULL;
	free_settings(settings_arg.workarea_vt, &settings_arg.settings_array);

	if (rc != CAFE_OK) {
		cafe_error_set(NULL, rc,
			       "Unable to create EP '%s'", spec->name);
		service_error_str(cmd_args, ERR_OTHER, NULL);
	}
	CAFE_UNLOCK();

	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	loc_free(cmd_args->command);

	/* Return log handle */
	write_id(reply, EP_PREFIX, (uintptr_t) inst);
	write_stream(reply, 0);

	write_settings(reply, settings_arg.workarea_vt,
		       inst->workarea->u.data, cafe_setting_result_VT_DEF);
	post_event(handler_reply, cmd_args);
}


static
void ep_delete_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	struct cafe_ep_inst *victim;
	uintptr_t victim_handle, config_handle;
	char config_name[64], victim_name[64];
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, victim_name, sizeof(victim_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	if (!str2id(EP_PREFIX, victim_name, &victim_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted victim EP handle");

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	victim = find_config_ep(config, victim_handle);
	if (victim == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find victim EP");

	rc = cafe_ep_inst_delete(victim);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc,
				  "Unable to delete configuration");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);

	post_event(simple_reply, cmd_args);
}


static
void ep_connect_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	struct cafe_ep_inst *source, *sink;
	uintptr_t source_handle, sink_handle, config_handle;
	char config_name[64], source_name[64], sink_name[64];
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, source_name, sizeof(source_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, sink_name, sizeof(sink_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	/* decode source and sink handles */
	if (!str2id(EP_PREFIX, source_name, &source_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted source handle");

	if (!str2id(EP_PREFIX, sink_name, &sink_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted sink handle");

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	source = find_config_ep(config, source_handle);
	if (source == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find source EP");

	sink = find_config_ep(config, sink_handle);
	if (sink == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find sink EP");

	rc = cafe_ep_inst_connect(source, sink);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc, "Unable to connect EPs");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);

	post_event(simple_reply, cmd_args);
}


static
void ep_disconnect_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	struct cafe_ep_inst *source;
	struct cafe_ep_inst *sink;
	uintptr_t source_handle, sink_handle, config_handle;
	char config_name[64], source_name[64], sink_name[64];
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, source_name, sizeof(source_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, sink_name, sizeof(sink_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	/* decode source and sink handles */
	if (!str2id(EP_PREFIX, source_name, &source_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted source handle");

	if (!str2id(EP_PREFIX, sink_name, &sink_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted sink handle");

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	source = find_config_ep(config, source_handle);
	if (source == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find source EP");

	sink = find_config_ep(config, sink_handle);
	if (sink == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find sink EP");

	rc = cafe_ep_inst_disconnect(source, sink);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc, "Unable to disconnect EPs");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);

	post_event(simple_reply, cmd_args);
}

static
void ep_dynamic_types_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *)arg;
	struct cafe_instrument *config;
	struct cafe_ep_inst *inst;
	char config_name[64], inst_name[64];
	uintptr_t inst_handle, config_handle;
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in;
	OutputStream *reply;

	in = create_byte_array_input_stream(&input_bytes,
					    cmd_args->command,
					    cmd_args->command_sz);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);

	if (rc != 0)
		service_error(cmd_args, rc);

	if (read_stream(in) != 0)
		service_error(cmd_args, ERR_JSON_SYNTAX);

	json_read_string_r(in, inst_name, sizeof(inst_name), &rc);

	if (rc != 0)
		service_error(cmd_args, rc);

	if (read_stream(in) != 0)
		service_error(cmd_args, ERR_JSON_SYNTAX);

	/* decode instance handle */
	if (!str2id(EP_PREFIX, inst_name, &inst_handle)) {
		post_err_reply(cmd_args, ERR_OTHER,
			       "Incorrectly formatted instance handle");
		return;
	}

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle)) {
		post_err_reply(cmd_args, ERR_OTHER,
			       "Incorrectly formatted Config handle");
		return;
	}

	CAFE_LOCK();
	config = find_config(config_handle);

	if (config == NULL) {
		CAFE_UNLOCK();
		post_err_reply(cmd_args, ERR_SYM_NOT_FOUND,
			       "Unable to find requested Configuration");
		return;
	}

	inst = find_config_ep(config, inst_handle);

	if (inst == NULL) {
		CAFE_UNLOCK();
		post_err_reply(cmd_args, ERR_SYM_NOT_FOUND,
			       "Unable to find EP");
		return;
	}

	reply = create_byte_array_output_stream(&cmd_args->os_reply);
	write_stream(reply, '{');
	json_write_string(reply, "types");
	write_stream(reply, ':');
	write_stream(reply, '[');

	if (inst->spec->dynamic_output_events != NULL) {
		char *type_str = NULL;

		rc = inst->spec->dynamic_output_events(inst, &type_str);

		if (rc != CAFE_OK) {
			CAFE_UNLOCK();
			post_err_reply(cmd_args, ERR_OTHER, cafe_strerror(rc));
			return;
		}
		if (type_str != NULL && *type_str != '\0')
			json_write_string(reply, type_str);

		CAFE_FREE(inst->session->allocator, type_str);
	}
	CAFE_UNLOCK();
	write_stream(reply, ']');
	write_stream(reply, '}');
	write_stream(reply, 0);
	loc_free(cmd_args->command);
	post_event(handler_reply, cmd_args);
}


static
void event_enable_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	struct cafe_ep_inst *inst;
	char config_name[64], inst_name[64];
	uintptr_t inst_handle, config_handle;
	int num_types;
	char **type_strs;
	struct cafe_value_type **types;
	struct cafe_event_descr *event;
	unsigned int i;
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, inst_name, sizeof(inst_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	type_strs = json_read_alloc_string_array_r(in, &num_types, &cafe_buf,
						   &rc);
	if (rc != 0) service_error(cmd_args, rc);

	if (read_stream(in) != 0) {
		loc_free(type_strs);
		service_error(cmd_args, ERR_JSON_SYNTAX);
	}

	/* decode instance handle */
	if (!str2id(EP_PREFIX, inst_name, &inst_handle)) {
		post_err_reply(cmd_args, ERR_OTHER,
			       "Incorrectly formatted instance handle");
		goto cleanup_and_exit;
	}

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle)) {
		post_err_reply(cmd_args, ERR_OTHER,
			       "Incorrectly formatted Config handle");
		goto cleanup_and_exit;
	}

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL) {
		CAFE_UNLOCK();
		post_err_reply(cmd_args, ERR_SYM_NOT_FOUND,
			       "Unable to find requested Configuration");
		goto cleanup_and_exit;
	}

	inst = find_config_ep(config, inst_handle);
	if (inst == NULL) {
		CAFE_UNLOCK();
		post_err_reply(cmd_args,  ERR_SYM_NOT_FOUND,
			       "Unable to find EP");
		goto cleanup_and_exit;
	}

	types = loc_alloc(num_types * sizeof(struct cafe_value_type *));
	for (i = 0; i < (unsigned int)num_types; ++i) {
		rc = cafe_type_parse(config->ns, type_strs[i], &types[i]);
		if (rc != CAFE_OK) {
			CAFE_UNLOCK();
			loc_free(types);
			post_err_reply(cmd_args, rc, "Unable to parse type");
			goto cleanup_and_exit;
		}
	}
	loc_free(type_strs);

	rc = cafe_cfg_get_event(config, num_types, types, &event);
	loc_free(types);
	if (rc != CAFE_OK) 
		service_error_str(cmd_args, rc, "Unable to enable event");

	rc = cafe_ep_inst_enable_output_event(inst, event);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc, "Unable to enable event");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);

	post_event(simple_reply, cmd_args);
	return;

cleanup_and_exit:
	loc_free(type_strs);
}


static
void event_disable_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	struct cafe_ep_inst *inst;
	char config_name[64], inst_name[64];
	uintptr_t inst_handle, config_handle;
	int num_types;
	char **type_strs;
	struct cafe_value_type **types;
	struct cafe_event_descr *event;
	unsigned int i;
	int rc = 0;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, inst_name, sizeof(inst_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	type_strs = json_read_alloc_string_array_r(in, &num_types,
						   &cafe_buf, &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) {
		loc_free(type_strs);
		service_error(cmd_args, ERR_JSON_SYNTAX);
	}

	/* decode instance handle */
	if (!str2id(EP_PREFIX, inst_name, &inst_handle)) {
			//TD: fix this??
		post_err_reply(cmd_args, ERR_OTHER,
			       "Incorrectly formatted instance handle");
		goto cleanup_and_exit;
	}

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle)) {
			//TD: fix this??
		post_err_reply(cmd_args, ERR_OTHER,
			       "Incorrectly formatted Config handle");
		goto cleanup_and_exit;
	}

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL) {
		CAFE_UNLOCK();
		post_err_reply(cmd_args, ERR_SYM_NOT_FOUND,
			       "Unable to find requested Configuration");
		goto cleanup_and_exit;
	}

	inst = find_config_ep(config, inst_handle);
	if (inst == NULL) {
		CAFE_UNLOCK();
			//TD: fix this??
		post_err_reply(cmd_args, ERR_SYM_NOT_FOUND,
			       "Unable to find EP");
		goto cleanup_and_exit;
	}

	types = loc_alloc(num_types * sizeof(struct cafe_value_type *));
	for (i = 0; i < (unsigned int)num_types; ++i) {
		rc = cafe_type_parse(config->ns, type_strs[i], &types[i]);
		if (rc != CAFE_OK) {
			CAFE_UNLOCK();
			loc_free(types);
			//TD: fix this??
			post_err_reply(cmd_args, rc, "Unable to parse type");
			goto cleanup_and_exit;
		}
	}
	loc_free(type_strs);

	rc = cafe_cfg_get_event(config, num_types, types, &event);
	loc_free(types);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc, "Unknown event");

	rc = cafe_ep_inst_disable_output_event(inst, event);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, rc, "Unable to disable event");
	CAFE_UNLOCK();

	loc_free(cmd_args->command);

	post_event(simple_reply, cmd_args);
	return;

cleanup_and_exit:
	loc_free(type_strs);
}


static
void command_setting_set_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	uintptr_t config_handle, ep_handle;
	struct cafe_ep_inst *inst;
	struct read_settings_arg settings_arg;
	struct cafe_cfg_setting *settings_arr;
	char config_name[64], ep_name[256];
	unsigned int i;
	int rc = 0;
	bool restart_ep = false;
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);
	OutputStream *reply =
		create_byte_array_output_stream(&cmd_args->os_reply);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, ep_name, sizeof(ep_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	if (!str2id(EP_PREFIX, ep_name, &ep_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted EP handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	inst = find_config_ep(config, ep_handle);
	if (inst == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested EP");

	rc = workarea_type(inst->spec, config->ns, &settings_arg.workarea_vt);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "error parsing spec types");

	memset(&settings_arg.settings_array, 0,
	       sizeof(settings_arg.settings_array));
	settings_arg.err = NULL;
	(void)json_read_array_r(in, read_settings, &settings_arg, &rc);
	if (rc != CAFE_OK)
		service_error(cmd_args, rc);

	if (settings_arg.err != NULL) {
		free_settings(settings_arg.workarea_vt,
			      &settings_arg.settings_array);
		service_error_str(cmd_args, ERR_JSON_SYNTAX, settings_arg.err);
	}

	if (read_stream(in) != 0) {
		free_settings(settings_arg.workarea_vt,
			      &settings_arg.settings_array);
		CAFE_UNLOCK();
		service_error(cmd_args, ERR_JSON_SYNTAX);
	}

	/* make sure all requested values are flagged as settings */
	settings_arr = settings_arg.settings_array.p;
	for (i = 0; i < settings_arg.settings_array.l; ++i) {
		const struct cafe_vm_attr *attr =
			cafe_type_attr_by_name(settings_arg.workarea_vt,
					       settings_arr[i].name);
		if (cafe_find_nature_str(attr->ident, cafe_setting_VT_DEF) == NULL) {
			free_settings(settings_arg.workarea_vt,
				      &settings_arg.settings_array);
			service_error_str(cmd_args, ERR_OTHER,
					  "attempting to set a non-parameter");
		}
	}

	/* If the destination EP is single-threaded, pause its thread. */

	if (((inst->flags & CAFE_EP_INST_FLAG_INPUT_ANY_CTX) == 0) &&
	    (inst->input_ctx != NULL) &&
	    (inst->state == CAFE_EP_INST_STATE_STARTED)) {
		rc = cafe_ep_inst_stop(inst->input_ctx);
		if (rc != CAFE_OK) {
			free_settings(settings_arg.workarea_vt,
				      &settings_arg.settings_array);
			service_error_str(cmd_args, ERR_OTHER,
					  "error pausing thread");
		}
		restart_ep = true;
	}

	assert(cafe_config_channel_up == NULL);
	assert(cafe_config_channel_down == NULL);
	cafe_config_channel_up = cmd_args->c;
	cafe_config_channel_down = proxy_get_target_channel(cmd_args->c);
	if (inst->spec->settings_commit == NULL)
		rc = settings_commit(
			inst,
			(unsigned int) settings_arg.settings_array.l,
			settings_arg.settings_array.p);
	else
		rc = inst->spec->settings_commit(
			inst,
			(unsigned int) settings_arg.settings_array.l,
			settings_arg.settings_array.p);
	cafe_config_channel_up = NULL;
	cafe_config_channel_down = NULL;

	free_settings(settings_arg.workarea_vt, &settings_arg.settings_array);

	if (rc != CAFE_OK)
		service_error_str(cmd_args, ERR_OTHER,
				  "error committing parameters");

	write_settings(reply, inst->workarea_type, inst->workarea->u.data,
		       cafe_setting_result_VT_DEF);

	if (restart_ep) {
		rc = cafe_instrument_refresh_output_event_actions(config);
		if (rc != CAFE_OK)
			service_error_str(cmd_args, ERR_OTHER,
					  "Unable to refresh output events");
		rc = cafe_ep_inst_start(inst->input_ctx);
		if (rc != CAFE_OK)
			service_error_str(cmd_args, ERR_OTHER,
					  "error resuming thread");
	}
	CAFE_UNLOCK();
	loc_free(cmd_args->command);
	post_event(handler_reply, cmd_args);
}


static
void event_inject_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args = (struct cafe_handler_args *) arg;
	struct cafe_instrument *config;
	struct cafe_ep_inst *inst;
	uintptr_t config_handle, ep_handle;
	char config_name[64], ep_name[256], event_name[256];
	struct cafe_value_type *event_vt;
	struct cafe_value_data *value_data;
	int rc = 0;
	struct cafe_event_descr *event_descr;
	void *data[1];
	struct ByteArrayInputStream input_bytes;
	InputStream *in =
		create_byte_array_input_stream(&input_bytes,
					       cmd_args->command,
					       cmd_args->command_sz);

	json_read_string_r(in, config_name, sizeof(config_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, ep_name, sizeof(ep_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);
	json_read_string_r(in, event_name, sizeof(event_name), &rc);
	if (rc != 0) service_error(cmd_args, rc);
	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	if (!str2id(CONFIG_PREFIX, config_name, &config_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted Config handle");

	if (!str2id(EP_PREFIX, ep_name, &ep_handle))
		service_error_str(cmd_args, ERR_OTHER,
				  "Incorrectly formatted EP handle");

	CAFE_LOCK();
	config = find_config(config_handle);
	if (config == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested Configuration");

	inst = find_config_ep(config, ep_handle);
	if (inst == NULL)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "Unable to find requested EP");

	rc = cafe_type_parse(config->ns, event_name, &event_vt);
	if (rc != CAFE_OK)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "error parsing event type");

	if (cafe_value_create(event_vt, &value_data) != CAFE_OK)
		service_error_str(cmd_args, ERR_SYM_NOT_FOUND,
				  "error creating event");

	/* read parameter struct */
	json_read_struct(in, json_read_struct_member, (void *)value_data);
	if (value_data->type == NULL) {
		cafe_value_delete(value_data);
		service_error_str(cmd_args, ERR_OTHER,
				  "error parsing parameters");
	}

	if (read_stream(in) != 0) service_error(cmd_args, ERR_JSON_SYNTAX);

	event_descr = get_event_descr(config, event_vt);
	if (event_descr == NULL || event_descr->num_types != 1) {
		cafe_value_delete(value_data);
		service_error_str(cmd_args, ERR_OTHER,
				  "error dispatching event type");
	}

	/* Send event */
	data[0] = &value_data->u.data;

	if (cafe_ep_inst_inject_event(inst, event_descr, data) != CAFE_OK) {
		cafe_value_delete(value_data);
		service_error_str(cmd_args, ERR_OTHER, "error injecting event");
	}

	CAFE_UNLOCK();

	loc_free(cmd_args->command);
	cafe_value_delete(value_data);
	post_event(simple_reply, cmd_args);
}

#if !defined(VALUE_ADD)
static
int write_sysinfo(OutputStream* out)
{
	struct cafe_sysconfig_info sysinfo;
	struct cafe_namespace *ns;
	struct cafe_value_type *type;
	bool need_comma = false;
	int rc;

	rc = cafe_namespace_create(cafe_stdlib_allocator, &ns);
	if (rc != CAFE_OK) {
		return rc;
	}
	rc = cafe_type_parse(ns,
			 "S[sysinfo_row,k[key,s]v[value,s]]D[t[sysinfo_row]]",
			 &type);
	if (rc != CAFE_OK) {
		cafe_namespace_delete(ns);
		return rc;
	}
	memset(&sysinfo, 0, sizeof (sysinfo));
	cafe_dict_create(&sysinfo.properties, type);
	rc = cafe_sysinfo_get(&sysinfo);
	if (rc != CAFE_OK) {
		cafe_dict_delete(&sysinfo.properties);
		cafe_namespace_delete(ns);
		return rc;
	}

	write_stream(out, '{');
	if (sysinfo.sysname) {
		json_write_string(out, "Name");
		write_stream(out, ':');
		json_write_string(out, sysinfo.sysname);
		need_comma = true;
	}
	if (sysinfo.version) {
		if (need_comma)
			write_stream(out, ',');
		json_write_string(out, "Version");
		write_stream(out, ':');
		json_write_string(out, sysinfo.version);
		need_comma = true;
	}
	if (sysinfo.release) {
		if (need_comma)
			write_stream(out, ',');
		json_write_string(out, "Release");
		write_stream(out, ':');
		json_write_string(out, sysinfo.release);
		need_comma = true;
	}
	if (sysinfo.properties.nentries) {
		struct cafe_dict_iterator iter;
		struct cafe_value_data *value = cafe_dict_first(
				&sysinfo.properties, &iter);
		while (value) {
			struct cafe_sysconfig_property *property =
					(struct cafe_sysconfig_property *) &value->u.data;
			if (need_comma)
				write_stream(out, ',');
			json_write_string(out, property->key);
			write_stream(out, ':');
			json_write_string(out, property->value);
			value = cafe_dict_next(&sysinfo.properties, &iter);
			need_comma = true;
		}
	}
	write_stream(out, '}');
	cafe_sysinfo_free(&sysinfo);
	cafe_dict_delete(&sysinfo.properties);
	cafe_namespace_delete(ns);
	return CAFE_OK;
}
#endif /* !defined(VALUE_ADD) */

static
void get_capabilities_handler(struct event_node *arg)
{
	struct cafe_handler_args *cmd_args =
		(struct cafe_handler_args *) arg;
	OutputStream *reply;

	reply = create_byte_array_output_stream(&cmd_args->os_reply);

	/* write capability properties */
	write_stream(reply, '{');

#if !defined(VALUE_ADD)
	/* <SystemInfo> */
	json_write_string(reply, "SystemInfo");
	write_stream(reply, ':');
	if (write_sysinfo(reply) != CAFE_OK) {
		json_write_string(reply, NULL);
	}
	/* </SystemInfo> */
#endif /* !defined(VALUE_ADD) */

	write_stream(reply, '}');
	write_stream(reply, MARKER_EOA);
	loc_free(cmd_args->command);
	post_event(handler_reply, cmd_args);
}


/**
 *  List registered Event Processor Specifications
 *
 * syntax is : "listEps"
 *
 * return message is <error report> <EP name list>
 */
static
void command_list_eps(char *token, Channel *c, void *arg)
{
	defer_event(list_eps_handler, c, token);
}


/**
 *  Describe an Event Processor Specification
 *
 * syntax is : "describeEp" <EP name>
 *
 * return message is <error report> <EP description object>
 */
static
void command_describe_ep(char *token, Channel *c, void *arg)
{
	defer_event(describe_ep_handler, c, token);
}


/**
 *  List registered Configurations
 *
 * syntax is : "listConfigs"
 *
 * return message is <error report> <Config handle list>
 */
static
void command_list_configs(char *token, Channel *c, void *arg)
{
	defer_event(list_configs_handler, c, token);
}


/**
 *  Describe an existing Configuration
 *
 * syntax is : "describeConfig" <Config handle>
 *
 * return message is <error report> <config description object>
 */
static
void command_describe_config(char *token, Channel *c, void *arg)
{
	defer_event(describe_config_handler, c, token);
}


/**
 *  Create a Collection Configuration
 *
 * syntax is : "configCreate" <Config name>
 *
 * return message is <error report> <Config handle>
 */
static
void command_config_create(char *token, Channel *c, void *arg)
{
	defer_event(config_create_handler, c, token);
}


/**
 *  Delete a Collection Configuration
 *
 * syntax is : "configDelete" <Config handle>
 *
 * return message is <error report>
 */
static
void command_configuration_delete(char *token, Channel *c, void *arg)
{
	defer_event(config_delete_handler, c, token);
}


/**
 *  Start a Collection Configuration
 *
 * syntax is : "configStart" <Config handle>
 *
 * return message is <error report>
 */
static
void command_configuration_start(char *token, Channel *c, void *arg)
{
	defer_event(config_start_handler, c, token);
}


/**
 *  Stop a Collection Configuration
 *
 * syntax is : "configStop" <Config handle>
 *
 * return message is <error report>
 */
static
void command_configuration_stop(char *token, Channel *c, void *arg)
{
	defer_event(config_stop_handler, c, token);
}


/**
 *  Instantiate an Event Processor
 *
 * syntax is : "epCreate" <Config handle> <EP spec name> <EP instance settings>
 *
 * return message is <error report> <EP instance handle>
 *                       <array of name/value return pairs>
 */
static
void command_ep_create(char *token, Channel *c, void *arg)
{
	defer_event(ep_create_handler, c, token);
}


/**
 *  Delete an Event Processor instance
 *
 * syntax is : "epDelete" <Config handle> <victim EP instance handle>
 *
 * return message is <error report>
 */
static
void command_ep_delete(char *token, Channel *c, void *arg)
{
	defer_event(ep_delete_handler, c, token);
}


/**
 *  Set a setting on an Event Processor
 *
 * syntax is : "epSettingSet" <Config handle> <EP instance handle> <setting>
 *
 * return message is <error report>
 */
static
void command_setting_set(char *token, Channel *c, void *arg)
{
	defer_event(command_setting_set_handler, c, token);
}


/**
 *  Send an event to an Event Processor
 *
 * syntax is : "epInjectEvent" <Config handle> <EP instance handle> <eventname> <parameters>
 *
 * return message is <error report>
 */
static
void command_event_inject(char *token, Channel *c, void *arg)
{
	defer_event(event_inject_handler, c, token);
}


/**
 *  Connect two Event Processors
 *
 * syntax is : "epConnect" <Config handle> <source EP> <sink EP>
 *
 * return message is <error report>
 */
static
void command_ep_connect(char *token, Channel *c, void *arg)
{
	defer_event(ep_connect_handler, c, token);
}


/**
 *  Disconnect two Event Processors
 *
 * syntax is : "epDisconnect" <Config handle> <source EP> <sink EP>
 *
 * return message is <error report>
 */
static
void command_ep_disconnect(char *token, Channel *c, void *arg)
{
	defer_event(ep_disconnect_handler, c, token);
}


/**
 *  Enable an output event on the given Event Processor
 *
 * syntax is : "eventEnable" <Config handle> <EP> <event>"
 *
 * return message is <error report>
 */
static
void command_event_enable(char *token, Channel *c, void *arg)
{
	defer_event(event_enable_handler, c, token);
}


/**
 * Disable an output event on the given Event Processor.
 *
 * syntax is : "eventDisable" <Config handle> <EP> <event>"
 *
 * return message is <error report>
 */
static void command_event_disable(char *token, Channel *c, void *arg)
{
	defer_event(event_disable_handler, c, token);
}


/**
 * Get dynamic type information from an Event Processor.
 *
 * syntax is : "epDynamicTypes" <Config handle> <EP>"
 *
 * return message is { "types" : "<value-type-string>" }
 */
static void command_ep_dynamic_types(char *token, Channel *c, void *arg)
{
	defer_event(ep_dynamic_types_handler, c, token);
}


/**
 *  Get capability properties
 *
 * syntax is : "getCapabilities"
 *
 * return message is <error report> <property map>
 */
static
void command_get_capabilities(char *token, Channel *c, void *arg)
{
	defer_event(get_capabilities_handler, c, token);
}


static void register_service_handlers(
	Protocol *proto,
	const char *name)
{
	static struct {
		const char *name;
		ProtocolCommandHandler2 handler;
	} cmd[] = {
		{ "configCreate",	command_config_create },
		{ "configDelete",	command_configuration_delete, },
		{ "configStart",	command_configuration_start },
		{ "configStop",		command_configuration_stop },
		{ "describeConfig",	command_describe_config },
		{ "describeEp",		command_describe_ep },
		{ "epConnect",		command_ep_connect },
		{ "epCreate",		command_ep_create },
		{ "epDelete",		command_ep_delete },
		{ "epDisconnect",	command_ep_disconnect },
		{ "epDynamicTypes",	command_ep_dynamic_types },
		{ "epInjectEvent",	command_event_inject },
		{ "epSettingSet",	command_setting_set },
		{ "eventDisable",	command_event_disable },
		{ "eventEnable",	command_event_enable },
		{ "listConfigs",	command_list_configs },
		{ "listEps",		command_list_eps },
		{ "getCapabilities",	command_get_capabilities }
	};

	size_t i;

	for (i = 0; i < CAFE_NELEMENTS(cmd); i++) {
		add_command_handler2(proto, name,
				     cmd[i].name, cmd[i].handler,
				     NULL);
	}
}


#if defined(VALUE_ADD)
static void channel_redirection_listener(Channel *host, Channel *target)
{
	if (target->state == ChannelStateConnected) {
		unsigned int i;
		long service_cafe = 0;
		const size_t len = strlen(SERVICE_NAME);
		for (i = 0; i < (unsigned int)target->peer_service_cnt; i++) {
			char * nm = target->peer_service_list[i];
			if (strncmp(nm, SERVICE_NAME, len) == 0) {
				if (nm[len] == '\0') {
					if (service_cafe == 0)
						service_cafe = 1;
					continue;
				}
				if (nm[len] == '@') {
					char *endptr;
					long id;
					id = strtol(nm + len + 1, &endptr, 10);
					if (nm + len + 1 < endptr && *endptr == '\0') {
						if (service_cafe <= id)
							service_cafe = id + 1;
						continue;
					}
				}
			}
		}
		if (service_cafe) {
			char name[sizeof SERVICE_NAME + 30];
			snprintf(name, sizeof name, "%s@%ld",
				 SERVICE_NAME, service_cafe);
			register_service_handlers(host->protocol, name);
		}
	}
}
#endif /* VALUE_ADD */


static void channel_close_handler(struct event_node *arg)
{
	struct cafe_handler_args *args = (struct cafe_handler_args *) arg;
	struct cafe_list_head *list_item;

	/* Delete any instruments assocoated with closed channels. */
	CAFE_LOCK();
retry:
	/* Cannot use cafe_list_foreach_safe here because the list can
	 * change in arbitraty ways while we wait in
	 * cafe_instrument_delete(). */
	cafe_list_foreach(list_item, &channel_session_list_head) {
		struct channel_session *map =
			CHANNEL_SESSION_LIST_ENTRY(list_item);
		if (map->c == NULL) {
			(void)cafe_instrument_delete(map->session);
			/* Map entry removed by config_delete_listener */
			goto retry;
		}
	}
	CAFE_UNLOCK();
	loc_free(args);
}


static void channel_close_listener(Channel * c) {
	struct cafe_handler_args *args;
	struct cafe_list_head *list_item;
	bool do_close = false;

	/* Remove any references to this channel. */
	assert(is_dispatch_thread());
	CAFE_LOCK();
	cafe_list_foreach(list_item, &channel_session_list_head) {
		struct channel_session *map =
			CHANNEL_SESSION_LIST_ENTRY(list_item);
		if (map->c == c) {
			map->c = NULL;
			do_close = true;
		}
	}
	CAFE_UNLOCK();

	/* Delete any instruments associated with this channel. */
	if (do_close) {
		args = loc_alloc_zero(sizeof *args);
		args->enode.handler = channel_close_handler;
		event_queue_post_event(cafe_event_queue, &args->enode);
	}
}


static void config_delete_listener(struct cafe_instrument *session)
{
	struct cafe_list_head *list_item;

	cafe_list_foreach(list_item, &channel_session_list_head) {
		struct channel_session *map =
			CHANNEL_SESSION_LIST_ENTRY(list_item);
		if (map->session == session) {
			cafe_list_remove(&map->channel_session_list);
			loc_free(map);
			break;
		}
	}
}


void *cafe_event_thread(void *arg)
{
#if defined(HAVE_PR_SET_NAME)
	if (prctl(PR_SET_NAME, "cafe_tcf_event") != 0)
		perror("Error setting thread name");
#endif
#if (defined(__VXWORKS__) && defined(RUNTIME_ANALYSIS_DEBUG))
	/* Set task as breakable when RUNTIME_ANALYSIS is in debug mode */
	taskOptionsSet(0, VX_UNBREAKABLE, 0);
#endif
	cafe_event_queue = event_queue_create();
	if (cafe_event_queue == NULL) {
		(void)fputs("Error initializing event queue\n", stderr);
		return NULL;
	}

	event_loop_run(cafe_event_queue);
	return NULL;
}


void ini_cafe_config_service(Protocol *proto)
{
#if defined(__linux__) || defined(WIN32) || defined(__APPLE__)
	pthread_t event_thread;
#endif
	int rc;

	rc = cafe_lib_init();
	if (rc != CAFE_OK) {
		(void)printf("Error initializing Cafe lib: %s\n",
		       cafe_strerror(rc));
		return;
	}

	rc = cafe_bbuf_init(&cafe_buf, cafe_stdlib_allocator, 256);
	if (rc != CAFE_OK) {
		(void)printf("Error allocating cafe service buffer: %s\n",
		       cafe_strerror(rc));
		return;
	}

	command_buf_os = create_byte_array_output_stream(&command_buf);

	register_service_handlers(proto, SERVICE_NAME);

	add_channel_close_listener(channel_close_listener);
#if defined(VALUE_ADD)
        /* Only register cafe@1 service in VA */
	add_channel_redirection_listener(channel_redirection_listener);
#endif /* VALUE_ADD */

	CAFE_LOCK();
	cafe_instrument_add_delete_listener(config_delete_listener);
	CAFE_UNLOCK();

#if defined(__linux__) || defined(WIN32) || defined(__APPLE__)
	if (pthread_create(&event_thread, NULL, cafe_event_thread, NULL))
		perror("Error creating Cafe TCF event thread");
#endif
}


#if !defined(__VXWORKS__)
/**
 *  Register all common Cafe EP specifications.  This isn't necessary for
 * VxWorks targets because they toggle availability and initialization of
 * individual specifications using its own component system.
 */
void register_common_EPs(void)
{
	unsigned int i;
	typedef int (*register_fn)(void);
	static struct {
		const char *name;
		register_fn fnp;
	} regfns[] = {
		{ "file stream reader", cafe_fstream_reader_cfg_register },
		{ "file stream writer", cafe_fstream_writer_cfg_register },
		{ "print", cafe_print_configuration_register },
		{ "text file provider", text_file_provider_register },
		{ "realtime", cafe_realtime_ep_register },
#if defined(HAVE_UTSNAME_H) && defined(HAVE_UNAME) && !defined(VALUE_ADD)
		{ "system info", sysinfo_cfg_register },
#endif
#if defined(HAVE_LINUX_PERF) && defined(HAVE_FTW) && !defined(VALUE_ADD)
		{ "linux trace", linux_trace_register },
#endif
		{ "timer", cafe_timer_ep_register },
		{ "timestamp normalizer", cafe_timestamp_normalizer_ep_register },
		{ "TCF upload", cafe_tcf_upload_register },
		{ "TCF download", cafe_tcf_download_register },
		{ "cafe_aggregator", cafe_aggregator_ep_register }
	};

	CAFE_LOCK();

	symresolve_aggregator_ini();

	for (i = 0; i < CAFE_NELEMENTS(regfns); ++i) {
		int rc = regfns[i].fnp();
		if (rc != CAFE_OK) {
			fprintf(stderr, "Cafe EP (%s) registration failed: ",
				regfns[i].name);
			if (cafe_error_cur.rc == CAFE_OK) {
				fprintf(stderr, "%s\n", cafe_strerror(rc));
			} else {
				cafe_error_print(NULL, stderr, ": ");
				cafe_error_clear(NULL);
				fputc('\n', stderr);
			}
			continue;
		}
	}

	CAFE_UNLOCK();
}
#endif  /* if !defined(__VXWORKS__) */
