/* tcf_download.c - copies TCF VirtualStream events to file */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
15may14,sdt  Fix again 'Checked return' warning
11mar14,sdt  Fix 'Checked return' warning
*/

#include <tcf/services/streamsservice.h>
#include <tcf/framework/events.h>
#include <tcf/framework/exceptions.h>
#include <tcf/framework/trace.h>
#include <tcf/framework/json.h>

#include <cafe/stdio.h>
#include <string.h>
#include "../src/cafe_assert.h"
#include <errno.h>
#include <cafe/cafe.h>
#include <cafe/list.h>
#include <cafe/log_stream.h>
#include <cafe/ep/tcf_download.h>
#include <cafe/ep/config_service.h>
#include "tcf_utils.h"
#include "../src/cafe_trace.h"


struct tcf_download_workarea {
	const char *stream_id;
	const char *filename;
};

enum download_state {
	state_connecting,
	state_stopped,
	state_starting,
	state_started,
	state_stopping,
	state_disconnecting,
	state_disconnected
};

struct client_data {
	Channel *channel;
	FILE *out;
	char *stream_id;
	volatile enum download_state state;
	volatile int aborted;
	int pending_read;
	size_t size;
	char buf[4096];
};

struct tcf_download_remote_info {
	cafe_eal_handle_t sink_actions;
};


static void error_struct_callback(
	InputStream *in,
	const char *name,
	void *value)
{
	char buf[1024];
	int size;

	size = snprintf(buf, sizeof(buf), "\t%s: ", name);
	if (strcmp(name, "Time") == 0) {
		(void)snprintf(buf, sizeof(buf) - size,
			 "%"PRId64"\n", json_read_int64(in));
	} else if (strcmp(name, "Code") == 0) {
		(void)snprintf(buf, sizeof(buf) - size,
			 "%"PRId64"\n", json_read_int64(in));
	} else if (strcmp(name, "Format") == 0) {
		(void)json_read_string(in, &buf[size], sizeof(buf) - size);
	} else {
		trace(LOG_ALWAYS, "UNKNOWN field %s when decoding error\n",
		      name);
	}
	trace(LOG_ALWAYS, buf);
}

static void connect_stream_done(
	Channel *c,
	void *client_data,
	int error)
{
	struct cafe_ep_inst *inst = client_data;
	struct client_data *node = inst->client_data;
	int syntax_error = 0;

	if (error) {
		trace(LOG_ALWAYS, "Reply error %d: %s\n",
		      error, errno_to_str(error));
		node->aborted = 1;
	} else {
		if (peek_stream(&c->inp) != 0) {
			trace(LOG_ALWAYS, "ERROR connecting to stream:\n");
			(void)json_read_struct(&c->inp, error_struct_callback,
			                       0);
			node->aborted = 1;
		}
		if (read_stream(&c->inp) != 0 ||
		    read_stream(&c->inp) != MARKER_EOM) {
			node->aborted = 1;
			syntax_error = 1;
		}
	}
	CAFE_LOCK();
	assert(node->state == state_connecting);
	if (!node->aborted) {
		channel_lock(c);
		node->state = state_stopped;
	} else {
		node->state = state_disconnected;
	}
	(void)cafe_notify_signal();
	CAFE_UNLOCK();
	if (syntax_error)
		exception(ERR_JSON_SYNTAX);
}

static void connect_stream(void *arg)
{
	struct cafe_ep_inst *inst = arg;
	struct client_data *node = inst->client_data;
	Channel *c = node->channel;

	/* connect to virtual stream */
	assert(node->state == state_connecting);
	protocol_send_command(c, "Streams", "connect", connect_stream_done,
	                      inst);
	json_write_string(&c->out, node->stream_id);
	write_stream(&c->out, 0);
	write_stream(&c->out, MARKER_EOM);
}

static void disconnect_stream_done(
	Channel *c,
	void *client_data,
	int error)
{
	struct cafe_ep_inst *inst = client_data;
	struct client_data *node = inst->client_data;
	int syntax_error = 0;

	assert(node->channel == c);
	if (error) {
		trace(LOG_ALWAYS, "Reply error %d: %s\n",
		      error, errno_to_str(error));
	} else {
		if (peek_stream(&c->inp) != 0) {
			trace(LOG_ALWAYS, 
			         "ERROR disconnecting from stream:\n");
			(void)json_read_struct(&c->inp, error_struct_callback,
			                       0);
		}
		if (read_stream(&c->inp) != 0 ||
		    read_stream(&c->inp) != MARKER_EOM) {
			syntax_error = 1;
		}
	}
	CAFE_LOCK();
	assert(node->state == state_disconnecting);
	node->state = state_disconnected;
	channel_unlock(c);
	(void)cafe_notify_signal();
	CAFE_UNLOCK();
	if (syntax_error)
		exception(ERR_JSON_SYNTAX);
}


static
void disconnect_stream(void *arg)
{
	struct cafe_ep_inst *inst = arg;
	struct client_data *node = inst->client_data;
	Channel *c = node->channel;

	/* Disconnect from virtual stream */
	assert(node->state == state_disconnecting);
	protocol_send_command(c, "Streams", "disconnect", 
	                      disconnect_stream_done, inst);
	json_write_string(&c->out, node->stream_id);
	write_stream(&c->out, 0);
	write_stream(&c->out, MARKER_EOM);
}

static void read_stream_done(
	Channel *c,
	void *client_data,
	int error)
{
	struct cafe_ep_inst *inst = client_data;
	struct client_data *node = inst->client_data;
	int syntax_error = 0;

	assert(node->size == 0);
	assert(node->pending_read);
	node->pending_read = 0;
	if (error) {
		trace(LOG_ALWAYS, "Reply error %d: %s\n",
		      error, errno_to_str(error));
		node->aborted = 1;
	} else {
		int64_t lost;
		bool end;
		InputStream *inp = &node->channel->inp;
		int ch = peek_stream(inp);
		if (ch == 'n') {
			(void)read_stream(inp);
			if (read_stream(inp) != 'u') goto err_json_syntax;
			if (read_stream(inp) != 'l') goto err_json_syntax;
			if (read_stream(inp) != 'l') goto err_json_syntax;
		} else {
			JsonReadBinaryState state;

			json_read_binary_start(&state, inp);

			for (;;) {
				size_t rd = json_read_binary_data(
					&state, node->buf + node->size,
					sizeof node->buf - node->size);
				if (rd == 0) break;
				node->size += rd;
			}

			assert(state.size_start <= 0 || 
			       node->size == state.size_start);

			json_read_binary_end(&state);
		}

		if (read_stream(inp) != 0) goto err_json_syntax;
		if (peek_stream(inp) != 0) {
			json_read_struct(inp, error_struct_callback, 0);
		}
		if (read_stream(inp) != 0) goto err_json_syntax;
		lost = json_read_int64(inp);
		if (read_stream(inp) != 0) goto err_json_syntax;
		end = json_read_boolean(inp);
		if (read_stream(inp) != 0 ||
		    read_stream(inp) != MARKER_EOM) {
		err_json_syntax:
			syntax_error = 1;
			node->aborted = 1;
		} else if (lost) {
			node->aborted = 1;
		} else if (end) {
			node->aborted = -1;
		}
	}

	if (node->state == state_started) {
		size_t wr = fwrite_unlocked(node->buf, node->size, 1,
					    node->out);
		if (wr != 1) {
			TRACE(("unable to write %"PRIsize_t\
			       " bytes to file, wrote %"PRIsize_t"\n",
			       node->size, wr));
			node->aborted = 1;
		}
		node->size = 0;
		(void)fflush_unlocked(node->out);
	}

	if (node->state == state_disconnected) {
		/* Notify disconnect waiter that pending_read == 0 */
		CAFE_LOCK();
		cafe_ep_inst_state_stopped(inst);
		(void)cafe_notify_signal();
		CAFE_UNLOCK();
	}

	if (syntax_error)
		exception(ERR_JSON_SYNTAX);

	if (node->aborted) {
		CAFE_LOCK();
		if (node->state == state_started) {
			node->state = state_stopped;
			cafe_ep_inst_state_stopped(inst);
			(void)cafe_notify_signal();
		}
		CAFE_UNLOCK();
	} else if (node->state == state_started && !node->pending_read) {
		node->pending_read = 1;
		protocol_send_command(c, "Streams", "read", read_stream_done, 
		                      inst);
		json_write_string(&c->out, node->stream_id);
		write_stream(&c->out, 0);
		json_write_long(&c->out, sizeof node->buf);
		write_stream(&c->out, 0);
		write_stream(&c->out, MARKER_EOM);
	}
}


static
void start_download(void *arg)
{
	struct cafe_ep_inst *inst = arg;
	struct client_data *node = inst->client_data;
	Channel *c = node->channel;

	/* Start reading from virtual stream */
	if (!node->aborted && node->size != 0) {
		size_t wr = fwrite_unlocked(node->buf, node->size, 1,
					    node->out);
		if (wr != 1) {
			TRACE(("unable to write %"PRIsize_t\
			       " bytes to file, wrote %"PRIsize_t"\n",
			       node->size, wr));
			node->aborted = 1;
		}
		node->size = 0;
		(void)fflush_unlocked(node->out);
	}

	if (!node->pending_read && !node->aborted) {
		node->pending_read = 1;
		protocol_send_command(c, "Streams", "read", read_stream_done, 
		                      inst);
		json_write_string(&c->out, node->stream_id);
		write_stream(&c->out, 0);
		json_write_long(&c->out, sizeof node->buf);
		write_stream(&c->out, 0);
		write_stream(&c->out, MARKER_EOM);
	}
	CAFE_LOCK();
	assert(node->state == state_starting);
	if (node->aborted == 0) {
		node->state = state_started;
		cafe_ep_inst_state_started(inst);
	} else {
		node->state = state_stopped;
		cafe_ep_inst_state_stopped(inst);
	}
	CAFE_UNLOCK();
}


static
void stop_download(void *arg)
{
	struct cafe_ep_inst *inst = arg;
	struct client_data *node = inst->client_data;

	/* Stop reading from virtual stream */
	CAFE_LOCK();
	assert(node->state == state_stopping);
	node->state = state_stopped;
	cafe_ep_inst_state_stopped(inst);
	CAFE_UNLOCK();
}

/**
 * "Starts" the configuration by opening the given output file.
 */
static
int tcf_download_create_inst(struct cafe_ep_inst *inst)
{
	struct tcf_download_workarea *w =
		(struct tcf_download_workarea *) inst->workarea->u.data;
	Channel *c = cafe_config_channel_down;
	const char *stream_id = w->stream_id;
	struct client_data *node;
	int rc;

	if (c == NULL || stream_id == NULL || stream_id[0] == '\0')
		return CAFE_ERROR;

	if ((node = CAFE_ALLOC(inst->allocator,
			       sizeof *node + strlen(stream_id) + 1)) == NULL)
		return CAFE_NOMEM;
	inst->client_data = node;
	memset(node, 0, sizeof *node);
	node->channel = c;
	node->stream_id = (char *)(node + 1);
	strncpy(node->stream_id, stream_id, strlen(stream_id)+1);

	/* Connect to virtual stream */
	node->state = state_connecting;
	post_event(connect_stream, inst);
	while (node->state == state_connecting) {
		rc = cafe_notify_wait();
		assert(rc == CAFE_OK);
	}
	if (node->state != state_stopped) {
		rc = CAFE_ERROR;
		goto error;
	}

	node->out = fopen(w->filename, "wb");
	if (node->out == 0) {
		rc = CAFE_ERROR;
		goto error;
	}

	return CAFE_OK;

error:
	if (node->state != state_disconnected) {
		/* Disconnect from virtual stream */
		assert(node->state == state_stopped);
		node->state = state_disconnecting;
		post_event(disconnect_stream, inst);
		while (node->state == state_disconnecting) {
			rc = cafe_notify_wait();
			assert(rc == CAFE_OK);
		}
		assert(node->state == state_disconnected);
	}
	CAFE_FREE(inst->allocator, node);
	return rc;
}

static
void tcf_download_delete_inst(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	int rc;

	if (node->state != state_disconnected) {
		/* Disconnect from virtual stream */
		assert(node->state == state_stopped);
		node->state = state_disconnecting;
		post_event(disconnect_stream, inst);
		while (node->state == state_disconnecting || 
		       node->pending_read) {
			rc = cafe_notify_wait();
			assert(rc == CAFE_OK);
			if (rc != CAFE_OK)
				(void)fprintf(stderr,
				              "Error waiting for wakeup signal: %s\n",
					cafe_strerror(rc));
		}
		assert(node->state == state_disconnected);
		assert(!node->pending_read);
	}
	(void)fclose(node->out);
	CAFE_FREE(inst->allocator, node);
}


/**
 * "Starts" the configuration by opening the given output file.
 */
static
int tcf_download_start(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;

	assert(node->state == state_stopped);
	node->state = state_starting;
	cafe_post_event_sync(start_download, inst);
	if (node->state != state_started) {
		TRACE(("unable to start stream %s\n", node->stream_id));
		return CAFE_ERROR;
	}
	return CAFE_OK;
}


static
int tcf_download_stop(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;

	assert(node->state == state_started);
	node->state = state_stopping;
	cafe_post_event_sync(stop_download, inst);
	assert(node->state == state_stopped);
	return CAFE_OK;
}

static
int tcf_download_update_output_event_actions(
	struct cafe_ep_inst *inst)
{
	return CAFE_OK;
}

static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};

#define WORKAREA_DEF					\
	"S[tcf_download.workarea,"			\
	"v[stream_id,s," cafe_setting_VT_NAT "]"	\
	"v[filename,s," cafe_setting_VT_NAT "],"	\
	cafe_output_inst_thread_VT_NAT			\
	"]"

#define WORKAREA_REF "t[tcf_download.workarea]"

#define ALL_TYPES				\
	cafe_output_inst_thread_VT_DEF		\
	cafe_wildcard_VT_DEF			\
	cafe_setting_VT_DEF			\
	WORKAREA_DEF

static struct cafe_ep_spec spec = {
	"tcf_download",
	tcf_download_create_inst,
	tcf_download_delete_inst,
	tcf_download_start,
	tcf_download_stop,
	NULL,
	tcf_download_update_output_event_actions,
	NULL,			/* inject event */
	ALL_TYPES,
	WORKAREA_REF,
	0,			/* n_input_events */
	NULL,			/* input event list */
	1,
	output_events
};


static bool cfg_registered = false;


void cafe_tcf_download_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}


int cafe_tcf_download_register(void)
{
	int rc;

	if (cfg_registered)
		return CAFE_OK;

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		cfg_registered = true;

	return rc;
}
