/* tcf_upload.c - pipes event streams via TCF VirtualStream */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
11mar14,sdt  Fix 'Checked return' warning
*/

#include <stdio.h>
#include <string.h>  /* for strlen() and friends */
#include <cafe/value_print.h>
#include <cafe/log_stream.h>
#include <cafe/log_buffer.h>
#include <cafe/cafe.h>
#include <cafe/atomic.h>

#include <tcf/services/streamsservice.h>
#include <tcf/framework/events.h>
#include <tcf/framework/exceptions.h>
#include <assert.h>
#include <cafe/ep/tcf_upload.h>
#include "../src/cafe_trace.h"
#include "tcf_utils.h"


/* Maximum number of useconds data will stay buffered. */
static const unsigned int BATCH_DELAY = 100 * 1000;

static int tcf_upload_create(struct cafe_ep_inst *inst);
static void tcf_upload_destroy(struct cafe_ep_inst *inst);
static int tcf_upload_start(struct cafe_ep_inst *inst);
static int tcf_upload_stop(struct cafe_ep_inst *inst);
static int tcf_upload_get_input_event_actions(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list *actions);


static const char *const STREAM_TYPE = "Cafe event stream";

static const char *upload_event_TYPE[] = {cafe_wildcard_VT_REF};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT(upload_event)
};


static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};


/* types used to parameterize this configuration */
#define PARAM_TYPES   "v[buffer_size,L," cafe_setting_VT_NAT ",16000]"


static struct cafe_ep_spec spec = {
	"tcf_upload",
	tcf_upload_create,			/* create_hook */
	tcf_upload_destroy,			/* delete_hook */
	tcf_upload_start,
	tcf_upload_stop,
	tcf_upload_get_input_event_actions,
	NULL,
	NULL,
	cafe_setting_result_VT_DEF
	"S[tcf_upload.workarea," PARAM_TYPES
	"v[stream_version,I]"
	"v[vs_id,s," cafe_setting_result_VT_NAT "]]"
	cafe_metadata_VT_DEF
	cafe_wildcard_VT_DEF, /* types */
	"t[tcf_upload.workarea]",
	CAFE_NELEMENTS(input_events),
	input_events,
	CAFE_NELEMENTS(output_events),
	output_events
};


struct upload_workspace {
	unsigned long buffer_size;
	unsigned int stream_version;
	char *vs_id;
};


struct client_data {
	struct cafe_slog_encoded_inst *slog;
	struct cafe_output_stream *ostream;
	size_t upload_threshold;
	cafe_atomic_val bytes_logged;  /* number of bytes currently in the log,
					  used to calculate when TCF should be
					  notified to start uploading */
	VirtualStream *vstream;
	/* private local buffer for constructing encoded events before copying
	   them into the virtual stream's output buffer */
	struct cafe_log_buf *buffer;

	/* true when a delayed batch buffer drain event has been posted to the
	   TCF event queue */
	cafe_atomic_val is_batch_queued;
	bool pending_forward_tcf_stream_event;
};


/**
 *  Writes as much of the given log buffer to the given virtual stream as
 * possible, until either there's no buffered data left, or the virtual
 * stream's own buffer is full.
 */
static
void log_event_write(VirtualStream *vstream, struct cafe_log_buf *buf,
		     cafe_atomic_val *bytes_logged)
{
	size_t size;
	size_t wrote;
	cafe_atomic_val bsize;
	char *msg = cafe_log_read_reserve(buf, &size);

	while (size != 0) {
		if (virtual_stream_add_data(vstream, msg, size, &wrote, 0)
		    == -1)
		{
			TRACE(("stream add data failed\n"));
			return;
		}
		if (wrote != size)
			TRACE(("\tonly wrote %" PRIsize_t " bytes\n", wrote));
		if (wrote == 0)
			/* vstream is full */
			return;

		/* this is a workaround for slog_encoded_write() returning less
		   bytes than actually buffered */
		bsize = cafe_atomic_get(bytes_logged);
		if (bsize > (cafe_atomic_val)wrote) {
			(void)cafe_atomic_val_sub(bytes_logged, wrote);
		} else {
			(void)cafe_atomic_val_sub(bytes_logged, bsize);
		}

		cafe_log_read_commit(buf, wrote);
		msg = cafe_log_read_reserve(buf, &size);
	}
}


/**
 *  When the virtual stream notifies us it has space available for writing,
 * fill it with any available buffered event data.
 */
static
void stream_forward_callback(VirtualStream *stream, int event, void *arg)
{
	struct client_data *client_data = arg;

	assert(client_data);
	assert(client_data->vstream);
	assert(client_data->buffer);
	switch (event) {
	case VS_EVENT_SPACE_AVAILABLE:
		log_event_write(client_data->vstream, client_data->buffer,
				&client_data->bytes_logged);

		break;
	case VS_EVENT_DATA_AVAILABLE:
	default:
		str_exception(ERR_OTHER, "Unhandled stream callback\n");
	}
}

/**
 *  A post_event() callback used to drain the output event log into a
 * VirtualStream.
 */
static
void forward_tcf_stream_event(void *arg)
{
	struct client_data *client_data = arg;

	assert(is_dispatch_thread());
	assert(client_data);
	assert(client_data->vstream);
	assert(client_data->buffer);

	client_data->pending_forward_tcf_stream_event = false;
	log_event_write(client_data->vstream, client_data->buffer,
			&client_data->bytes_logged);
}

static void trigger_forward_tcf_stream(void *arg)
{
	struct client_data *client_data = arg;

	if (client_data->pending_forward_tcf_stream_event)
		return;
	client_data->pending_forward_tcf_stream_event = true;
	post_event(forward_tcf_stream_event, client_data);
}

/**
 *  A post_event() callback used to drain the output event log into a
 * VirtualStream.
 */
static
void batch_tcf_stream_event(void *arg)
{
	struct client_data *client_data = arg;

	assert(is_dispatch_thread());
	assert(client_data);
	assert(client_data->vstream);
	assert(client_data->buffer);

	(void)cafe_atomic_set(&client_data->is_batch_queued, false);
	log_event_write(client_data->vstream, client_data->buffer,
			&client_data->bytes_logged);
}

/**
 * Arguments to the tcf_virtual_stream_create callback routine
 */
struct tcf_virtual_stream_create_arg {
	size_t buf_len;
	struct client_data *client_data;
};

/**
 * TCF event callback for creating the virtual stream.
 */
static
void tcf_virtual_stream_create(void *arg)
{
	struct tcf_virtual_stream_create_arg *create_args = arg;

	virtual_stream_create(STREAM_TYPE,
			      NULL, /* context ID */
			      create_args->buf_len,	/* buf_len */
			      VS_ENABLE_REMOTE_READ, 	/* access flags */
			      stream_forward_callback, 	/* callback */
			      create_args->client_data, /* callback args */
			      &create_args->client_data->vstream);  /* result */
}


static
int tcf_upload_create(struct cafe_ep_inst *inst)
{
	int rc;
	struct client_data *client_data;
	char vs_id[256];
	struct upload_workspace *w =
		(struct upload_workspace *) inst->workarea->u.data;
	struct tcf_virtual_stream_create_arg create_args;

	client_data = CAFE_ALLOC(inst->allocator, sizeof(struct client_data));
	if (!client_data)
		return CAFE_NOMEM;
	inst->client_data = client_data;
	client_data->pending_forward_tcf_stream_event = false;
	client_data->buffer = cafe_log_create(w->buffer_size, inst->allocator);
	if (client_data->buffer == NULL) {
		CAFE_FREE(inst->allocator, client_data);
		return CAFE_NOMEM;
	}

	create_args.buf_len = w->buffer_size;
	create_args.client_data = client_data;
	cafe_post_event_sync(tcf_virtual_stream_create, &create_args);

	rc = cafe_logstream_open_output(client_data->buffer, inst->allocator,
					&client_data->ostream,
					trigger_forward_tcf_stream,
					client_data);

	if (rc != CAFE_OK)
		return rc;

	rc = cafe_slog_encoded_create(inst->session, client_data->ostream,
				      &client_data->slog);
	if (rc != CAFE_OK)
		return rc;
	virtual_stream_get_id(client_data->vstream, vs_id, sizeof(vs_id));
	w->vs_id = CAFE_STRDUP(inst->allocator, vs_id);
	if (w->vs_id == NULL)
		return CAFE_NOMEM;

	cafe_atomic_set(&client_data->bytes_logged, 0);
	client_data->upload_threshold = client_data->buffer->size >> 1;
	w->stream_version = cafe_slog_version;
	return CAFE_OK;
}


static
void close_vstream(void *arg)
{
	VirtualStream *vstream = arg;
	size_t wrote;

	/* flag end-of-stream, then delete our virtual output stream */
	if (virtual_stream_add_data(vstream, NULL, 0, &wrote, 1))
		(void)printf("Failed to add EOS marker\n");
	virtual_stream_delete(vstream);
}


static
void tcf_upload_destroy(struct cafe_ep_inst *inst)
{
	struct client_data *client_data = inst->client_data;

	CAFE_FREE(inst->allocator, client_data->ostream);
	cafe_slog_encoded_delete(client_data->slog);

	cafe_post_event_sync(close_vstream, client_data->vstream);
	CAFE_FREE(inst->allocator, client_data->buffer);
	CAFE_FREE(inst->allocator, client_data);
}


static
int tcf_upload_start(struct cafe_ep_inst *inst)
{
	struct client_data *client_data = inst->client_data;
	cafe_atomic_val bytes = cafe_atomic_get(&client_data->bytes_logged);

	(void)cafe_atomic_set(&client_data->is_batch_queued, false);
	cafe_ep_inst_state_started(inst);

	if (bytes >= (cafe_atomic_val)client_data->upload_threshold)
		trigger_forward_tcf_stream(inst->client_data);
	else if (bytes > 0) {
		(void)cafe_atomic_set(&client_data->is_batch_queued, true);
		post_event_with_delay(batch_tcf_stream_event, inst->client_data,
				      BATCH_DELAY);
	}

	return CAFE_OK;
}


static
void cancel_events(void *arg)
{
	struct client_data *c = arg;

	(void)cancel_event(forward_tcf_stream_event, c, 0);
	(void)cancel_event(batch_tcf_stream_event, c, 0);
}


static
int tcf_upload_stop(struct cafe_ep_inst *inst)
{
	cafe_post_event_sync(cancel_events, inst->client_data);
	cafe_ep_inst_state_stopped(inst);
	return CAFE_OK;
}


static
void upload_event(struct cafe_event_action_info *event, void *event_data[])
{
	struct client_data *client_data = event->inst->client_data;
	size_t wrote, original;
#if defined(TCF_UPLOAD_DEBUG)
	struct cafe_value_print_options options = CAFE_VALUE_PRINT_OPTIONS_INIT;
	unsigned int i;

	options.field_names = true;
	for (i = 0; i < event->event->num_types; ++i)
		cafe_value_print(stdout, NULL, event->event->types[i],
				 event_data[i], &options);
#endif

	wrote = cafe_slog_encoded_write(client_data->slog, event->event, event_data);

	original = cafe_atomic_val_add(&client_data->bytes_logged, wrote);
	if ((original < client_data->upload_threshold)
	    && ((original + wrote) >= client_data->upload_threshold))
		trigger_forward_tcf_stream(event->inst->client_data);
	else if (cafe_atomic_cas(&client_data->is_batch_queued, false, true))
		post_event_with_delay(batch_tcf_stream_event,
				      event->inst->client_data,
				      BATCH_DELAY);
}


static
int tcf_upload_get_input_event_actions(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *event_info;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);
	if (rc != CAFE_OK || event_info == NULL)
		return rc;

	return cafe_add_event_action(inst, actions, upload_event, event_info);
}


int cafe_tcf_upload_register(void)
{
	static bool cfg_registered = false;
	int rc;

	if (cfg_registered)
		return CAFE_OK;

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		cfg_registered = true;

	return rc;
}
