/* text_file_provider.c - read information from a text file and inject */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
13may14,sdt  Rename cafe tasks to tAnalysis* - US37854
*/

#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/stdio.h>
#include <cafe/list.h>
#include <cafe/value_type.h>
#include <cafe/dict.h>
#include <cafe/expr.h>
#include <cafe/ep_worker.h>
#include <cafe/value_print.h>
#include <cafe/ep/text_file_provider.h>
#include "../src/time_utils.h"


struct signal_event_type {
	cafe_eal_handle_t sink_actions;
};

struct client_data {
	FILE *f;
	struct cafe_ep_worker *worker;
	unsigned long time_ms;
	unsigned long sequence_count;
};

struct text_file_provider_workarea
{
	char *filename;
	struct cafe_dict_struct events;
};

struct text_file_provider_event {
	struct cafe_vm_blob label;
	struct cafe_event_descr *event;
	cafe_eal_handle_t sink_actions;
	unsigned long output_event_action_change_count;
};

static const char *optwsp(const char *s)
{
	while (*s != '\0' && isspace(*s & 0xFF))
		s++;
	return s;
}

static const char *skipwsp(const char *s)
{
	if (*s == '\0' || !isspace(*s & 0xFF))
		return NULL;
	return optwsp(s + 1);
}

static const char *skipident(const char *s)
{
	if (*s == '\0' ||
	    !(isalpha(*s & 0xFF) || *s == '_'))
		return NULL;
	do {
		s++;
	} while (*s != '\0' && (isalnum(*s & 0xFF) || *s == '_'));
	return s;
}

static void event_list_read(void *arg)
{
	struct text_file_provider_workarea *w;
	struct cafe_ep_inst *inst = (struct cafe_ep_inst *)arg;
	struct client_data *node = inst->client_data;
	struct timespec t;
	char linebuf[1000];
	const char *s;
	const char *cmd;
	const char *label;
	ptrdiff_t labelsz;
	size_t linesz;
	cafe_any_value keys[1];
	struct cafe_value_data *data;
	int typeno;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;

	assert(inst->workarea != NULL);
	w = (struct text_file_provider_workarea *)inst->workarea->u.data;
	while (cafe_ep_worker_isrunning(node->worker)) {
		if (fgets_unlocked(linebuf, sizeof linebuf, node->f) == NULL) {
			/* EOF or error - stop this configuration */
			CAFE_LOCK();
			(void)cafe_ep_worker_stop(node->worker);
			CAFE_UNLOCK();
			break;
		}
		linesz = strlen(linebuf);
		if (linesz > 0 && linebuf[linesz-1] == '\n') {
			linebuf[--linesz] = '\0';
		}
		s = optwsp(linebuf);
		/* Skip lines that are either blank, a comment or have
		 * no content. */
		if (linesz == 0 || linebuf[0] == '#' || *s == '\0')
			continue;
		cmd = s;
		s = skipident(s);
		if (s - cmd == 4 && strncmp(cmd, "type", 4) == 0) {
			/* type <ident> <valueType>[,<valueType>]* */
			struct cafe_event_action_list *sink_actions;
			cafe_eal_handle_t sink_actions_handle;
			struct text_file_provider_event *e;
			int rc;

			assert(s != NULL);
			s = skipwsp(s);
			if (s == NULL)
				goto parseerr;
			label = s;
			s = skipident(s);
			labelsz = s - label;
			if (labelsz == 0) {
				(void)fprintf(stderr, "invalid label: %s\n", label);
				goto lineerr;
			}
			assert(s != NULL);
			s = skipwsp(s);
			if (s == NULL)
				goto parseerr;
			typeno = 0;
			for (;;) {
				if (typeno >= CAFE_MAX_EVENT_TYPES) {
					(void)fprintf(stderr, "too many types: %s\n", linebuf);
					goto lineerr;
				}
				rc = cafe_type_parse2(inst->session->ns, &s,
						      ",", &types[typeno]);
				if (rc != CAFE_OK || s == NULL || types[typeno] == NULL) {
					goto parseerr;
				}
				typeno++;
				if (*s != ',')
					break;
				s++;
			}
			if (*s != '\0') {
				goto parseerr;
			}
			CAFE_LOCK();
			rc = cafe_cfg_get_event(inst->session, typeno, types, &event);
			if (rc != CAFE_OK) {
				CAFE_UNLOCK();
				(void)fprintf(stderr, "cannot get event description: %s\n", linebuf);
				goto lineerr;
			}
			keys[0].blob.p = (void *)label;
			keys[0].blob.l = labelsz;
			data = cafe_dict_get(&w->events, keys);
			if (data == NULL) {
				CAFE_UNLOCK();
				(void)fprintf(stderr, "cannot get events row: %s\n", linebuf);
				goto lineerr;
			}
			e = (struct text_file_provider_event *)data->u.data;
			if (e->event != NULL) {
				(void)fprintf(stderr, " row: %s\n", linebuf);
				goto lineerr;
			}
			e->event = event;
			rc = cafe_event_action_list_alloc(inst, event,
							  &sink_actions,
							  &sink_actions_handle);
			if (rc != CAFE_OK) {
				CAFE_UNLOCK();
				(void)fprintf(stderr, "cannot allocate sink action list: %s: %s\n",
					linebuf, strerror(rc));
				goto lineerr;
			}
			rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
			if (rc != CAFE_OK) {
				cafe_event_action_list_release(&sink_actions_handle);
				CAFE_UNLOCK();
				(void)fprintf(stderr, "cannot get sink actions: %s: %s\n",
					linebuf, strerror(rc));
				goto lineerr;
			}
			cafe_event_action_list_swap(&e->sink_actions,
						    sink_actions_handle);
			e->output_event_action_change_count =
				inst->session->output_event_action_change_count;
			CAFE_UNLOCK();
		} else if (s - cmd == 5 && strncmp(cmd, "event", 5) == 0) {
			/* event <ident> <valueData>[,<valueData>]* */
			struct text_file_provider_event *e;
			void *event_data[CAFE_MAX_EVENT_TYPES];
			struct cafe_value_data *objects[CAFE_MAX_EVENT_TYPES];
			unsigned int i;
			int rc;

			assert(s != NULL);
			s = skipwsp(s);
			if (s == NULL)
				goto parseerr;
			label = s;
			s = skipident(s);
			labelsz = s - label;
			if (labelsz == 0) {
				fprintf(stderr, "invalid label: %s\n", label);
				goto lineerr;
			}
			assert(s != NULL);
			s = skipwsp(s);
			if (s == NULL)
				goto parseerr;
			keys[0].blob.p = (void *)label;
			keys[0].blob.l = labelsz;
			data = cafe_dict_lookup(&w->events, keys);
			if (data == NULL) {
				(void)fprintf(stderr, "no event named: %.*s\n", (int)labelsz, label);
				goto lineerr;
			}
			e = (struct text_file_provider_event *)data->u.data;
			event = e->event;
			for (i = 0; i < event->num_types; i++) {
				struct cafe_vm_expr *expr;
				struct cafe_value_context *context;
				if (i != 0 && *s++ != ',')
					goto parseerr;
				context = cafe_value_context_create(inst->allocator);
				assert(context != NULL);
				context->ns = inst->session->ns;
				context->type = event->types[i]->ident;
				rc = cafe_expr_parse(context, &s, &expr);
				cafe_value_context_delete(context);
				if (rc != CAFE_OK)
					goto parseerr;
				rc = cafe_value_create2(event->types[i], expr, &data);
				cafe_expr_free(inst->allocator, expr);
				if (rc != CAFE_OK)
					goto parseerr;
				event_data[i] = data->u.data;
				objects[i] = data;
			}
			cafe_dispatch_event(&e->sink_actions, event_data);
			for (i = 0; i < event->num_types; i++) {
				cafe_value_delete(objects[i]);
			}
			node->sequence_count++;
		} else if (s - cmd == 5 && strncmp(cmd, "delay", 5) == 0) {
			/* delay <mseconds> */
			unsigned long delay;
			char *nd;

			assert(s != NULL);
			s = skipwsp(s);
			if (s == NULL)
				goto parseerr;
			errno = 0;
			delay = strtoul(s, &nd, 0);
			if (errno != 0 || s >= nd)
				goto parseerr;
			s = optwsp(nd);
			if (*s != '\0')
				goto parseerr;
			node->time_ms += delay;
			timespec_from_ticks(delay, MILLIS_PER_SEC, &t);
			(void)nanosleep(&t, NULL);
		} else {
			fprintf(stderr, "unknown command: %s\n", cmd);
			goto lineerr;
		}
	}
	return;

parseerr:
	fprintf(stderr, "cannot parse types: %s\n", linebuf);
lineerr:
	CAFE_LOCK();
	cafe_ep_worker_abort(node->worker);
	CAFE_UNLOCK();
	return;
}

static int text_file_provider_inst_create(struct cafe_ep_inst *inst)
{
	struct text_file_provider_workarea *w;
	struct client_data *node;
	int rc;

	assert(inst->workarea != NULL);
	w = (struct text_file_provider_workarea *)inst->workarea->u.data;
	if(w->filename == NULL) {
		(void)fprintf(stderr, "No filename specified\n");
		return CAFE_ERROR;
	}


	if ((node = CAFE_ALLOC(inst->allocator, sizeof *node)) == NULL)
		return CAFE_NOMEM;
	memset(node, 0, sizeof *node);
	node->f = fopen(w->filename, "r");
	if (node->f == NULL) {
		(void)fprintf(stderr, "Unable to open specified file `%s': %s\n",
			w->filename, strerror(errno));
		CAFE_FREE(inst->allocator, node);
		return CAFE_ERROR;
	}
	inst->client_data = node;

	rc = cafe_ep_worker_create("tAnalysisTextReader", inst, event_list_read,
	                           inst, &node->worker);
	if (rc != 0) {
		(void)fclose(node->f);
		CAFE_FREE(inst->allocator, node);
		return CAFE_ERROR;
	}

	return CAFE_OK;
}

static void text_file_provider_inst_delete(struct cafe_ep_inst *inst)
{
	struct text_file_provider_workarea *w =
		(struct text_file_provider_workarea *)inst->workarea->u.data;
	struct client_data *node = inst->client_data;
	struct cafe_value_data *data;
	struct cafe_dict_iterator iter;

	cafe_ep_worker_delete(node->worker);

	(void)fclose(node->f);

	data = cafe_dict_first(&w->events, &iter);
	while (data != NULL) {
		struct text_file_provider_event *e =
			(struct text_file_provider_event *)data->u.data;
		cafe_event_action_list_release(&e->sink_actions);
		data = cafe_dict_next(&w->events, &iter);
	}
	CAFE_FREE(inst->allocator, inst->client_data);
}

static int text_file_provider_start(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	int rc;

	rc = cafe_ep_worker_start(node->worker);
	if (rc != CAFE_OK)
		return rc;
	return cafe_ep_worker_wait(node->worker);
}

static int text_file_provider_stop(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	int rc;

	rc = cafe_ep_worker_stop(node->worker);
	if (rc != CAFE_OK)
		return rc;
	return cafe_ep_worker_wait(node->worker);
}

static int text_file_provider_update_output_event_actions(
	struct cafe_ep_inst *inst)
{
	struct text_file_provider_workarea *w =
		(struct text_file_provider_workarea *)inst->workarea->u.data;
	struct cafe_dict_iterator iter;
	struct cafe_value_data *data;
	int rc;

	for (data = cafe_dict_first(&w->events, &iter);
	     data != NULL; data = cafe_dict_next(&w->events, &iter)) {
		struct text_file_provider_event *e =
			(struct text_file_provider_event *)data->u.data;
		struct cafe_event_action_list *sink_actions;
		cafe_eal_handle_t sink_actions_handle;

		if (e->output_event_action_change_count ==
		    inst->session->output_event_action_change_count)
			continue;

		rc = cafe_event_action_list_alloc(inst, e->event,
						  &sink_actions,
						  &sink_actions_handle);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sink_actions_handle);
			return rc;
		}

		cafe_event_action_list_swap(&e->sink_actions,
					    sink_actions_handle);
		e->output_event_action_change_count =
			inst->session->output_event_action_change_count;
	}
	return CAFE_OK;
}

static struct cafe_ep_spec_output_evt signal_output_events[] = {
	CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};

static struct cafe_ep_spec spec = {
	"text_file_provider",
	text_file_provider_inst_create,
	text_file_provider_inst_delete,
	text_file_provider_start,
	text_file_provider_stop,
	NULL,
	text_file_provider_update_output_event_actions,
	NULL,
	cafe_output_inst_thread_VT_DEF
	cafe_wildcard_VT_DEF,
	"S[text_file_provider_event,k[label,b]v[event,L]v[sink_actions,L]"
	"v[output_event_action_change_count,L]]"
	"S[text_file_provider_workarea,v[filename,s," cafe_setting_VT_NAT "]"
	"v[events,D[t[text_file_provider_event]]],"
	cafe_output_inst_thread_VT_NAT "]",
	0,			/* 0 input events */
	NULL,			/* input events */
	CAFE_NELEMENTS(signal_output_events),
	signal_output_events
};

int text_file_provider_unregister(void)
{
	return cafe_ep_spec_unregister(&spec);
}

int text_file_provider_register(void)
{
	return cafe_ep_spec_register(&spec);
}
