/* Copyright (c) 2010-2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/* Implementations of Cafe OS primitives for UNIX. */

#if CAFE_OS_UNIX

/* The only things expected here are the differences between different
 * flavours of UNIX and/or Linux. */
#include "cafe_os.h"
#include <cafe/cafe.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "cafe_trace.h"
#include "cafe_assert.h"
#include "time_utils.h"
#include "cafe_dbg.h"

static pthread_mutex_t cafe_mutex = PTHREAD_MUTEX_INITIALIZER;
static volatile pthread_t cafe_lock_thread_owner = 0;

static pthread_cond_t cafe_cond = PTHREAD_COND_INITIALIZER;

int cafe_lock_init(void)
{
	return CAFE_OK;
}

bool inline cafe_islocked(void)
{
	return (cafe_lock_thread_owner == pthread_self());
}

void __cafe_lock(const char *file, int line)
{
	int rc;
	pthread_t curr_thr = pthread_self();

	TRACE(("cafe_lock(%s:%d): thr_id=%llx\n", file, line, (long long)curr_thr));

	__cafe_assert(!cafe_islocked(), file, line, NULL);

	if ((rc = pthread_mutex_lock(&cafe_mutex)) != 0) {
		char buf[128];
		snprintf(buf, sizeof(buf),
			 "pthread_mutex_lock() != 0 (%s)", strerror(rc));
		__cafe_assert_rtn(buf, file, line, NULL);
	}

	cafe_lock_thread_owner = curr_thr;
}

void __cafe_unlock(const char *file, int line)
{
	int rc;

	TRACE(("cafe_unlock(%s:%d): thr_id=%llx\n", file, line, (long long)pthread_self()));

	__cafe_assert(cafe_islocked(), file, line, NULL);
	__cafe_assert(pthread_self() == cafe_lock_thread_owner, file, line, "cafe_lock/unlock on same thread");

	cafe_lock_thread_owner = 0;
	if ((rc = pthread_mutex_unlock(&cafe_mutex)) != 0) {
		char buf[128];
		snprintf(buf, sizeof(buf),
			 "pthread_mutex_unlock() != 0 (%s)", strerror(rc));
		__cafe_assert_rtn(buf, file, line, NULL);
	}
}

int cafe_notify_signal(void)
{
	int cond_broadcast_result;

	CAFE_ASSERT_LOCKED();
	cond_broadcast_result = pthread_cond_broadcast(&cafe_cond);
	assert(cond_broadcast_result == 0);
	if (cond_broadcast_result != 0)
		fprintf(stderr, "error sending notify signal: %s\n",
			strerror(cond_broadcast_result));
	return CAFE_OK;
}

int cafe_notify_wait(void)
{
	int cond_wait_result;

	CAFE_ASSERT_LOCKED();
	cafe_lock_thread_owner = 0;
	cond_wait_result = pthread_cond_wait(&cafe_cond, &cafe_mutex);
	assert(cond_wait_result == 0);
	assert(cafe_lock_thread_owner == 0);
	if (cond_wait_result != 0)
		fprintf(stderr, "error waiting for notify signal: %s\n",
			strerror(cond_wait_result));
	cafe_lock_thread_owner = pthread_self();
	return CAFE_OK;
}

void __cafe_assert_rtn(const char *expr, const char *file,
		       int line, const char *preferred_msg)
{
	fprintf(stderr, "%s:%d: **** Assertion `%s' failed ****\n",
		file, line, preferred_msg ? preferred_msg : expr);
#ifdef HAVE_CAFE_PRINTSTACK
	cafe_print_stacktrace(stderr);
#endif
	abort();
}

#ifdef __APPLE__

int cafe_clock_gettime(clockid_t clk_id, struct timespec *tp)
{
	struct timeval tv;

	switch (clk_id) {
	case CLOCK_REALTIME:
		if (gettimeofday(&tv, NULL) != 0)
			return -1;
		tp->tv_sec = tv.tv_sec;
		tp->tv_nsec = tv.tv_usec * 1000;
		return 0;
	default:
		errno = EINVAL;
		return -1;
	}
}

#endif	/* __APPLE__ */

#endif	/* CAFE_OS_UNIX */
