/* errors.c - cafe error handling */

/*
 * Copyright (c) 2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cafe/errors.h>
#include <cafe/cafe.h>
#include "cafe_assert.h"


#define  ADJUST_ERR_HANDLE(handle)			\
	do {						\
		if ((handle) == NULL) {			\
			if (!cafe_islocked())		\
				return;			\
			(handle) = &cafe_error_cur;	\
		}					\
	} while(0)


/**
 *  Set the given error report object to the given Cafe error status, and
 * optionally store any associated error message.
 */
void cafe_error_set(struct cafe_error_report *e, int cafe_status,
		    const char *fmt, ...)
{
	size_t len = 0;
	char *msg = NULL;
	void *tmp;

	ADJUST_ERR_HANDLE(e);

	if (e->rc == CAFE_OK)
		e->rc = cafe_status;

	if (fmt == NULL)
		return;

	do {
		va_list ap;
		int n;

		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf(msg, len, fmt, ap);
		va_end(ap);

		if (n < 0)
			return;

		/* If that worked, we're finished. */
		if (n < (int)len)
			break;

		/* Else try again with more space. */
		len = n + 1;

		tmp = CAFE_REALLOC(cafe_stdlib_allocator, msg, len);
		if (tmp == NULL) {
			CAFE_FREE(cafe_stdlib_allocator, msg);
			return;
		}
		msg = tmp;
	} while(1);

	tmp = CAFE_REALLOC(cafe_stdlib_allocator, e->messages,
			   (e->n_messages + 1) * sizeof(*e->messages));
	if (tmp == NULL) {
		CAFE_FREE(cafe_stdlib_allocator, msg);
		return;
	}

	e->messages = tmp;
	e->messages[e->n_messages++] = msg;
}


/**
 *  Provide an optional OS-specific error value, e.g., "errno".
 */
void cafe_error_set_errno(struct cafe_error_report *e, int err)
{
	ADJUST_ERR_HANDLE(e);

	e->os_errno = err;
}


/**
 *  Emit a representation of the given error object to the given output stream.
 * Seperates different message sections with the given delimiter string.
 */
void cafe_error_print(const struct cafe_error_report *e,
		      FILE *out, const char *delim)
{
	unsigned int i;

	assert(out != NULL);
	ADJUST_ERR_HANDLE(e);

	for (i = e->n_messages; i > 0; --i) {
		(void)fputs(e->messages[i - 1], out);
		(void)fputs(delim, out);
	}
	(void)fputs(cafe_strerror(e->rc), out);

	if (e->os_errno != 0) {
		(void)fputs(delim, out);
		(void)fputs(strerror(e->os_errno), out);
	}
}


static
int bbuf_append_str(struct cafe_byte_buffer *buf, const char *str)
{
	size_t len;

	if (str == NULL)
		return CAFE_OK;
	len = strlen(str);
	return cafe_bbuf_append(buf, str, len);
}


/**
 *  Emits an error report to the given byte-buffer, separating each entry with
 * the provided delimiter string.
 */
int cafe_error_bbuf_write(const struct cafe_error_report *e,
			  struct cafe_byte_buffer *buf,
			  const char *err_msg, const char *delim)
{
	size_t len = delim == NULL ? 0 : strlen(delim);
	unsigned int i;
	int rc;

	assert(e != NULL);
	assert(buf != NULL);

	if (err_msg != NULL) {
		rc = bbuf_append_str(buf, err_msg);
		if (rc != CAFE_OK)
			goto fail;
		rc = cafe_bbuf_append(buf, delim, len);
		if (rc != CAFE_OK)
			goto fail;
	}

	for (i = e->n_messages; i > 0; --i) {
		rc = bbuf_append_str(buf, e->messages[i - 1]);
		if (rc != CAFE_OK)
			goto fail;
		rc = cafe_bbuf_append(buf, delim, len);
		if (rc != CAFE_OK)
			goto fail;
	}

	rc = bbuf_append_str(buf, cafe_strerror(e->rc));
	if (rc != CAFE_OK)
		goto fail;

	if (e->os_errno != 0) {
		rc = cafe_bbuf_append(buf, delim, len);
		if (rc != CAFE_OK)
			goto fail;
		rc = bbuf_append_str(buf, strerror(e->os_errno));
		if (rc != CAFE_OK)
		    goto fail;
	}

	rc = cafe_bbuf_write(buf, 0);
	if (rc != CAFE_OK)
		goto fail;

	return CAFE_OK;

fail:
	buf->data[buf->len] = 0;
	return rc;
}


/**
 *  Clear the given error report object, resetting status to CAFE_OK, and
 * releasing any messages and other allocated memory.
 */
void cafe_error_clear(struct cafe_error_report *e)
{
	unsigned int i;

	ADJUST_ERR_HANDLE(e);
	e->rc = CAFE_OK;
	e->os_errno = 0;
	for (i = 0; i < e->n_messages; ++i)
		CAFE_FREE(cafe_stdlib_allocator, e->messages[i]);
	if (e->n_messages > 0) {
		e->n_messages = 0;
		CAFE_FREE(cafe_stdlib_allocator, e->messages);
		e->messages = NULL;
	}
}
