/* event_map.c - Maps local events to events from external address spaces. */

/* Copyright (c) 2010-2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/* An Event Table allows clients to interpret events from external address
   spaces.
*/

#include <cafe/cafe.h>
#include <cafe/event_map.h>
#include <cafe/value_type.h>
#include <cafe/dict.h>
#include <string.h>
#include "cafe_assert.h"
#include "cafe_trace.h"

#define UNUSED_ARG(X) ((X) = *(&(X)))

#define CAFE_EVENT_MAP_DICT						\
	"S[cafe_remote_event,k[remote_id,Q]v[event,L]v[client_data,L]]"	\
	"D[t[cafe_remote_event]]"

struct cafe_event_map {
	struct cafe_instrument *session;
	struct cafe_dict_struct dict;
};


/**
 *  Returns a newly created empty namespace, or NULL on failure.
 */
struct cafe_event_map *cafe_event_map_create(struct cafe_instrument *session)
{
	struct cafe_event_map *event_map;
	struct cafe_value_type *dict_type;
	int rc;

	event_map = CAFE_ALLOC(session->allocator, sizeof *event_map);
	if (event_map == 0)
		return 0;
	memset(event_map, 0, sizeof *event_map);

	rc = cafe_type_parse(session->ns, CAFE_EVENT_MAP_DICT, &dict_type);
	if (rc != CAFE_OK || dict_type == NULL) {
		CAFE_FREE(session->allocator, event_map);
		return NULL;
	}

	cafe_dict_create(&event_map->dict, dict_type);
	event_map->session = session;
	return event_map;
}


/**
 *  Free all resources associated with a previously created namespace.
 */
void cafe_event_map_destroy(struct cafe_event_map *event_map)
{
	if (event_map == NULL)
		return;

	cafe_dict_delete(&event_map->dict);
	CAFE_FREE(event_map->session->allocator, event_map);
}


/**
 *  Iterate over all remote events associated with the event map.
 */
void cafe_event_map_iter(struct cafe_event_map *event_map,
			 bool (*cb)(void *client_data,
				    struct cafe_remote_event *),
			 void *client_data)
{
	struct cafe_dict_iterator iter;
	struct cafe_value_data *data;

	data = cafe_dict_first(&event_map->dict, &iter);
	while (data != NULL) {
		if (!cb(client_data, (struct cafe_remote_event *)data->u.data))
			return;
		data = cafe_dict_next(&event_map->dict, &iter);
	}
}


/**
 *  Create an event within the given namespace, as described by the given
 * metadata.  Associates this new event with the remote event id contained
 * in the metadata.  Returns a pointer to the resulting output event spec on
 * success, and NULL on failure.
 */
int cafe_event_map_add_event(
	struct cafe_event_map *event_map,
	const struct cafe_metadata *description,
	void *client_data,
	struct cafe_remote_event **result)
{
	int rc;
	unsigned int i;
	cafe_any_value keys[1];
	struct cafe_value_data *data;
	struct cafe_remote_event *remote;
	struct cafe_event_descr *event;
	struct cafe_value_type **types;

	*result = NULL;

	types = CAFE_ALLOC(event_map->session->allocator,
			   description->event_data.l * sizeof *types);
	if (types == NULL)
		return CAFE_NOMEM;
	for (i = 0; i < description->event_data.l; ++i) {
		char *vt_str = (char *)((char **)description->event_data.p)[i];
		UNUSED_ARG(vt_str); /* only used when TRACEing is enabled. */
		rc = cafe_type_parse(event_map->session->ns,
				     ((char **)description->event_data.p)[i],
				     &types[i]);
		if (rc != CAFE_OK || types[i] == NULL) {
			if (rc == CAFE_OK)
				rc = CAFE_UNRESOLVED_VALUE_TYPE;
			TRACE(("value type parse error: %s\n", vt_str));
			CAFE_FREE(event_map->session->allocator, types);
			return rc;
		}
	}
	rc = cafe_cfg_get_event(event_map->session,
				(unsigned int) description->event_data.l,
				types, &event);
	CAFE_FREE(event_map->session->allocator, types);
	if (rc != CAFE_OK)
		return rc;

	keys[0].ull = description->eid;
	data = cafe_dict_get(&event_map->dict, keys);
	if (data == NULL)
		return CAFE_NOMEM;

	remote = (struct cafe_remote_event *)data->u.data;
	if (remote->event)
		return CAFE_ERROR;

	remote->event = event;
	remote->client_data = client_data;
	*result = remote;
	return CAFE_OK;
}


/**
 *  Return the local event ID corresponding to the given remote event id, or
 * NULL if no mapping was found.
 */
struct cafe_remote_event *cafe_event_map_lookup(
	const struct cafe_event_map *event_map,
	unsigned long long event_id)
{
	cafe_any_value keys[1];
	struct cafe_value_data *data;

	keys[0].ull = event_id;
	data = cafe_dict_lookup(&event_map->dict, keys);
	if (data == NULL)
		return NULL;

	return (struct cafe_remote_event *)data->u.data;
}
