/*
 * Copyright (c) 2010-2011, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

#include "cafe_os.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <cafe/hashtbl.h>
#include <cafe/cafe.h>

/* Max size of hash table must not be greater than 2e29
 * IOW, (HASHTBL_MAX_TABLE_SIZE * sizeof (struct cafe_hashtbl_entry*)) < 2e32
 * to avoid variable overflow on 32bit arch.
 */
#ifndef HASHTBL_MAX_TABLE_SIZE
#define HASHTBL_MAX_TABLE_SIZE	(1 << 29)
#endif

/*
 * A hash table implementation based on chaining.
 *
 * Each slot in the bucket array is a pointer to a linked list that
 * contains the key-value pairs that are hashed to the same
 * location. Lookup requires scanning the list for an entry with the
 * given key. Insertion requires appending a new entry record to
 * either end of the list in the hashed slot. Deletion requires
 * searching the list and removing the element.
 */

struct cafe_hashtbl_entry {
	void *key;
	void *val;
	unsigned long hash;
	struct cafe_hashtbl_entry *next;
};

struct cafe_hashtbl {
	struct cafe_hashtbl_entry **table;
	const struct cafe_allocator *allocator;
	CAFE_HASHTBL_HASH_FUNC hashfun;
	CAFE_HASHTBL_EQUALS_FUNC hasheql;
	unsigned int count;
	unsigned int table_size;
	unsigned int resize_threshold;
	int autoresize;
	unsigned int load_factor; /* threshold to trigger resize */
};

/* djb2
 *
 * This algorithm was first reported by Dan Bernstein many years ago
 * in comp.lang.c.
 */
static inline unsigned long djb2_hash(const unsigned char *str)
{
	unsigned long hash = 5381;

	if (str != NULL) {
		int c;
		while ((c = *str++) != '\0')
			hash = ((hash << 5) + hash) + c; /* hash*33 + c */
	}
	return hash;
}

/* Robert Jenkins' 32 bit integer hash function. */

static inline unsigned int int32hash(unsigned int a)
{
	a = (a + 0x7ed55d16) + (a << 12);
	a = (a ^ 0xc761c23c) ^ (a >> 19);
	a = (a + 0x165667b1) + (a << 5);
	a = (a + 0xd3a2646c) ^ (a << 9);
	a = (a + 0xfd7046c5) + (a << 3);
	a = (a ^ 0xb55a4f09) ^ (a >> 16);
	return a;
}

static inline struct cafe_hashtbl_entry **slot(struct cafe_hashtbl *h,
					       unsigned long hval)
{
	return &h->table[hval % h->table_size];
}

static inline void unlink_entry(struct cafe_hashtbl *h,
				struct cafe_hashtbl_entry **head,
				struct cafe_hashtbl_entry *entry)
{
	*head = entry->next;
	h->count--;
}

static inline void link_entry(struct cafe_hashtbl *h, struct cafe_hashtbl_entry *entry)
{
	struct cafe_hashtbl_entry **head = slot(h, entry->hash);
	entry->next = *head;
	*head = entry;
	h->count++;
}

static inline struct cafe_hashtbl_entry *hashtbl_entry_new(
	const struct cafe_allocator *allocator,
	void *k, void *v,
	unsigned long hval)
{
	struct cafe_hashtbl_entry *e;

	if ((e = CAFE_ALLOC(allocator, sizeof *e)) != NULL) {
		e->key = k;
		e->val = v;
		e->hash = hval;
		e->next = NULL;
	}
	return e;
}

static inline struct cafe_hashtbl_entry *lookup_internal(struct cafe_hashtbl *h,
							 const void *k,
							 unsigned long hv)
{
	struct cafe_hashtbl_entry **head = slot(h, hv);
	struct cafe_hashtbl_entry *entry = *head;

	while (entry != NULL) {
		if (entry->hash == hv && h->hasheql(entry->key, k)) {
			return entry;
		}
		entry = entry->next;
	}
	return NULL;
}

int cafe_hashtbl_resize(struct cafe_hashtbl *h, unsigned int capacity)
{
	size_t i;
	struct cafe_hashtbl_entry **new_table;
	size_t nbytes;
	struct cafe_hashtbl tmp_h;

	if (capacity < 1)
		capacity = 1;

	if (capacity > HASHTBL_MAX_TABLE_SIZE)
		capacity = HASHTBL_MAX_TABLE_SIZE;

	nbytes = capacity * sizeof *new_table;

	if ((tmp_h.table = CAFE_ALLOC(h->allocator, nbytes)) == NULL)
		return CAFE_NOMEM;

	memset(tmp_h.table, 0, nbytes);
	tmp_h.count = 0;
	tmp_h.table_size = capacity;

	/* Transfer all entries from old table to new table. */

	for (i = 0; i < h->table_size; i++) {
		struct cafe_hashtbl_entry **head = &h->table[i];
		struct cafe_hashtbl_entry *entry;
		while ((entry = *head) != NULL) {
			unlink_entry(h, head, entry);
			link_entry(&tmp_h, entry);
			/* Look for other chained keys in this slot */
			head = &h->table[i];
		}
	}

	if (h->table != NULL)
		CAFE_FREE(h->allocator, h->table);

	h->table = tmp_h.table;
	h->table_size = tmp_h.table_size;
	h->count = tmp_h.count;
	h->resize_threshold = capacity * h->load_factor / 100;

	return CAFE_OK;
}

int cafe_hashtbl_insert(struct cafe_hashtbl *h, void *k, void *v)
{
	struct cafe_hashtbl_entry *new_entry;

	if (h->autoresize && h->count > h->resize_threshold) {
		/* Autoresize failures are benign only when the table
		 * has some size.  The hashtbl can continue to
		 * function in this mode albeit less optimally than
		 * requested. */
		(void)cafe_hashtbl_resize(h, 2 * h->table_size);
		assert(h->table_size > 0);
	}

	new_entry = hashtbl_entry_new(h->allocator, k, v, h->hashfun(k));

	if (new_entry == NULL)
		return -1;	/* WRS_XXX should be CAFE_NOMEM */

	link_entry(h, new_entry);

	return 0;		/* WRS_XXX should be CAFE_OK */
}

void *cafe_hashtbl_lookup(struct cafe_hashtbl *h, const void *k)
{
	struct cafe_hashtbl_entry *entry;
	entry = lookup_internal(h, k, h->hashfun(k));
	return (entry != NULL) ? entry->val : NULL;
}

void *cafe_hashtbl_replace(struct cafe_hashtbl *h, void *k, void *v)
{
	void *old_val = NULL;
	unsigned long hv = h->hashfun(k);
	struct cafe_hashtbl_entry *entry = lookup_internal(h, k, hv);
	if (entry != NULL) {
		old_val = entry->val;
		entry->val = v;
	}
	return old_val;
}

void *cafe_hashtbl_remove(struct cafe_hashtbl *h, const void *k,
			  CAFE_HASHTBL_KEY_FREE_FUNC kfreefunc, void *kfreearg)
{
	unsigned long hv = h->hashfun(k);
	struct cafe_hashtbl_entry **head = slot(h, hv);
	struct cafe_hashtbl_entry *entry = *head;

	while (entry != NULL) {
		if (h->hasheql(entry->key, k)) {
			void *value = entry->val;
			/* cull the end of the previous entry */
			*head = entry->next;
			h->count--;
			if (kfreefunc)
				kfreefunc(entry->key, kfreearg);
			CAFE_FREE(h->allocator, entry);
			return value;
		}
		head = &entry->next;
		entry = entry->next;
	}
	return NULL;
}

void cafe_hashtbl_clear(struct cafe_hashtbl *h,
			CAFE_HASHTBL_KEY_FREE_FUNC kfreefunc, void *kfreearg,
			CAFE_HASHTBL_VAL_FREE_FUNC vfreefunc, void *vfreearg)
{
	unsigned int i;
	struct cafe_hashtbl_entry *entry;
	struct cafe_hashtbl_entry *next;

	for (i = 0; i < h->table_size; i++) {
		next = h->table[i];
		while ((entry = next) != NULL) {
			if (kfreefunc != NULL)
				kfreefunc(entry->key, kfreearg);
			if (vfreefunc != NULL)
				vfreefunc(entry->val, vfreearg);
			next = entry->next;
			entry->next = NULL;
			CAFE_FREE(h->allocator, entry);
			--h->count;
		}
		assert(next == NULL);
		h->table[i] = next;
	}
	assert(h->count == 0);
}

void cafe_hashtbl_delete(struct cafe_hashtbl *h,
			 CAFE_HASHTBL_KEY_FREE_FUNC kfreefunc, void *kfreearg,
			 CAFE_HASHTBL_VAL_FREE_FUNC vfreefunc, void *vfreearg)
{
	if (h != NULL) {
		cafe_hashtbl_clear(h, kfreefunc, kfreearg, vfreefunc, vfreearg);
		CAFE_FREE(h->allocator, h->table);
		CAFE_FREE(h->allocator, h);
	}
}

unsigned int cafe_hashtbl_count(const struct cafe_hashtbl *h)
{
	return (h != NULL) ? h->count : 0;
}

struct cafe_hashtbl *cafe_hashtbl_new(unsigned int initial_size,
				      const struct cafe_allocator *allocator,
				      CAFE_HASHTBL_HASH_FUNC hashfun,
				      CAFE_HASHTBL_EQUALS_FUNC hasheql)
{
	int rc;
	struct cafe_hashtbl *h;

	assert(hashfun != NULL);
	assert(hasheql != NULL);

	if ((h = CAFE_ALLOC(allocator, sizeof *h)) == NULL)
		return NULL;

	h->table = NULL;
	h->allocator = allocator;
	h->hashfun = hashfun;
	h->hasheql = hasheql;
	h->count = 0;
	h->table_size = 0;	/* must be 0 for resize to work */
	h->resize_threshold = 0;
	h->autoresize = 0;
	h->load_factor = 75;	/* percent */
	rc = cafe_hashtbl_resize(h, initial_size);

	if (rc != CAFE_OK) {
		CAFE_FREE(allocator, h);
		return NULL;
	}

	return h;
}

unsigned long cafe_hashtbl_string_hash(const void *k)
{
	return djb2_hash((const unsigned char *)k);
}

int cafe_hashtbl_string_equals(const void *a, const void *b)
{
	return strcmp(((const char *)a), (const char *)b) == 0;
}

unsigned long cafe_hashtbl_int_hash(const void *k)
{
	return int32hash(*((const unsigned int *)k));
}

int cafe_hashtbl_int_equals(const void *a, const void *b)
{
	return *((const int *)a) == *((const int *)b);
}

unsigned long cafe_hashtbl_direct_hash(const void *k)
{
	return (unsigned long)k;
}

int cafe_hashtbl_direct_equals(const void *a, const void *b)
{
	return a == b;
}

unsigned int cafe_hashtbl_capacity(const struct cafe_hashtbl *h)
{
	return h->table_size;
}

void cafe_hashtbl_iter_init(struct cafe_hashtbl *h,
			    struct cafe_hashtbl_iter *iter)
{
	iter->key = iter->val = NULL;
	iter->pos = 0;
	iter->entry = NULL;
}

int cafe_hashtbl_iter_next(struct cafe_hashtbl *h,
			   struct cafe_hashtbl_iter *iter)
{
	unsigned int i;

	if (iter->entry != NULL) {
		iter->entry = iter->entry->next;
		if (iter->entry != NULL) {
			iter->key = iter->entry->key;
			iter->val = iter->entry->val;
			return 1;
		} else {
			iter->pos += 1;
		}
	}

	for (i = iter->pos; i < h->table_size; i++) {
		iter->entry = h->table[i];
		iter->pos = i;
		if (iter->entry != NULL) {
			iter->key = iter->entry->key;
			iter->val = iter->entry->val;
			return 1;
		}
	}

	return 0;
}

void cafe_hashtbl_autoresize_enable(struct cafe_hashtbl *h, unsigned int load_factor)
{
	h->load_factor = load_factor;
	h->autoresize = 1;
}

void cafe_hashtbl_autoresize_disable(struct cafe_hashtbl *h)
{
	h->autoresize = 0;
}
