/* io_demo.c - sample code to generate io events for Cafe */

/* Copyright (c) 2011-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */


#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>

#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

#define DEFAULT_LOOPCOUNT 10
#define DEFAULT_WRITE_CNT 64
#define DEFAULT_SIZE      1024

static bool dbg = false;

union iodemo_file_handle {
	int fd;
	FILE *file;
};

static int (*readint_ptr)(union iodemo_file_handle *f, int n, int *result);
static int (*writeint_ptr)(union iodemo_file_handle *f, int n, int *value);
static int (*open_ptr)(union iodemo_file_handle *f, const char *name,
		       const char *mode);
static int (*close_ptr)(union iodemo_file_handle *f);

static
void usage(void)
{
	printf("Usage: cafe-io-demo [-t <type>] [-n loopcount] [-c <writecount>] [-s num_ints] [-f] [-d distribution]\n"
	       "       type 0 - file access using read() and write() (default)\n"
	       "       type 1 - file access using fread() and fwrite()\n"
	       "\n"
	       "       loopcount - number of iterations around the outer loop. A negative value will\n"
	       "	           cause the test to run indefinitely (default %d)\n"
	       "\n"
	       "       writecount - number of writes in each iteration (default %d)\n"
	       "\n"
	       "       num_ints - number of integers to read and write (controls the maximum write size)\n"
	       "                  (default %d)\n"
	       "\n"
	       "       The -f option causes the demo to run as fast as possible,\n"
	       "       with no delay between iterations. Otherwise, there is a one\n"
	       "       second delay between iterations.\n"
	       "\n"
	       "       distribution - control the distribution of write sizes\n"
	       "                      0 - triangular random distribution from 0 to <num_ints>, with the\n"
	       "                          most popular write size being <num_ints>/2\n"
	       "                      1 - flat distribution, write sizes from 0 to <num_ints> are\n"
	       "                          covered evenly\n"
	       "                      2 - fixed size. All writes are of size <num_ints>\n"
	       "\n\n",
	       DEFAULT_LOOPCOUNT, DEFAULT_WRITE_CNT, DEFAULT_SIZE);
	exit(EXIT_SUCCESS);
}

static int readint_read(union iodemo_file_handle *f, int n, int *result)
{
	if (read(f->fd, result, n * sizeof(*result)) == n * sizeof(*result))
		return 1;
	else
		return -1;
}

static int readint_fread(union iodemo_file_handle *f, int n, int *result)
{
	size_t size = fread(result, sizeof(*result), n, f->file);
	if (size == n)
		return 1;
	else
		return -1;
}

static int writeint_write(union iodemo_file_handle *f, int n, int *value)
{
	ssize_t size = write(f->fd, value, n * sizeof(*value));
	if (size == n * sizeof(*value))
		return 1;
	else
		return -1;
}

static int writeint_fwrite(union iodemo_file_handle *f, int n, int *value)
{
	size_t size = fwrite(value, sizeof(*value), n, f->file);
	if (size == n)
		return 1;
	else
		return -1;
}

static int open_open(union iodemo_file_handle *f, const char *name,
		     const char *mode)
{
	int flags = 0;

	if (strcmp(mode, "r") == 0)
		flags = O_RDONLY;
	else if (strcmp(mode, "w") == 0)
		flags = O_CREAT | O_WRONLY | O_TRUNC;
	else
		return -1;

	f->fd = open(name, flags, 0660);
	if (f->fd == -1)
		return -1;
	return 0;
}

static int open_fopen(union iodemo_file_handle *f, const char *name,
		      const char *mode)
{
	f->file = fopen(name, mode);
	if (f->file == NULL)
		return -1;
	else {
		return 0;
	}
}

static int close_close(union iodemo_file_handle *f)
{
	return close(f->fd);
}


static int close_fclose(union iodemo_file_handle *f)
{
	return fclose(f->file);
}

static void setup_test(int (*readint_func)(union iodemo_file_handle *f, int n,
					   int *v),
		       int (*writeint_func)(union iodemo_file_handle *f, int n,
					    int *v),
		       int (*open_func)(union iodemo_file_handle *f,
					const char *name, const char *mode),
		       int (*close_func)(union iodemo_file_handle *f))
{
	readint_ptr = readint_func;
	writeint_ptr = writeint_func;
	open_ptr = open_func;
	close_ptr = close_func;
}

/* given a value x, with a uniform distribution in the range 0 .. input_range, return an
 * integer value y with a trianglular distribution over the range
 * 0 .. output_range

   http://en.wikipedia.org/wiki/Triangular_distribution#Generating_Triangular-distributed_random_variates

 */

static float uniform_to_triangle(int x, int input_range, int output_range)
{
	float a = 0.0;
	float b = 1.0;
	float c = 0.5;
	float u = (float)x / input_range;
	int y;

	if(u < 0.5)
		y = (int)(output_range * (sqrt(u*(b-a)*(c-a))) + 0.5);
	else
		y = (int)(output_range * (b - sqrt((1-u)*(b-a)*(b-c))) + 0.5);
	return y;
}

/*
  size : biggest size chunk to write
  num_writes: number of iterations round the loop
  mode: 0 = triangular distribution
  	1 = uniform distribution
	2 = fixed size
*/
static int write_test(int size, int write_cnt, size_t *ints_written, int mode)
{
	union iodemo_file_handle f;
	int rc;
	int i;
	int write_size; /* number of ints */
	int *p;

	rc = (*open_ptr)(&f, "/tmp/cafe-test", "w");
	if (rc != 0)
		return -1;
	if (size <= 0)
		size = 32;

	p = malloc(size * sizeof(*p));
	if (p == NULL)
		return -1;

	for (i = 0; i < write_cnt; i++) {
		if (mode == 0)
			write_size = uniform_to_triangle(rand(), RAND_MAX, size);
		else if (mode == 1)
			write_size = i * size / write_cnt;
		else if (mode == 2)
			write_size = size;
		else {
			free(p);
			return -1;
		}
		if (dbg)
			printf("write_size: %d\n", write_size);

		rc = (*writeint_ptr)(&f, write_size, p);
		if (rc != 1) {
			printf("write failed\n");
			close_ptr(&f);
			free(p);
			return -1;
		}
		*ints_written += write_size;
	}
	close_ptr(&f);
	free(p);
	return 0;
}

static int read_test(int size, size_t ints_written)
{
	union iodemo_file_handle f;
	int rc;
	int ints_remaining = ints_written;
	int ints_to_read;
	int values[10];
	rc = (*open_ptr)(&f, "/tmp/cafe-test", "r");
	if (rc != 0)
		return -1;

	while (ints_remaining > 0) {
		if (ints_remaining > sizeof(values)/sizeof(values[0]))
			ints_to_read = sizeof(values)/sizeof(values[0]);
		else
			ints_to_read = ints_remaining;

		rc = (*readint_ptr)(&f, ints_to_read, &values[0]);
		if (rc != 1) {
			printf("read failed\n");
			close_ptr(&f);
			return -1;
		}
		ints_remaining -= ints_to_read;
	}
	close_ptr(&f);
	return 0;
}

int main(int argc, char *argv[])
{
	int c;
	char *endptr;
	long loopcount = DEFAULT_LOOPCOUNT;
	long size = DEFAULT_SIZE;
	int type = 0;
	int mode = 0;
	bool fast = false;
	size_t ints_written = 0;
	int write_cnt = DEFAULT_WRITE_CNT;

	dbg = false;
	while ((c = getopt(argc, argv, "Xfn:t:s:d:c:")) != -1) {
		switch (c) {
		case 'n':
			loopcount = strtol(optarg, &endptr, 0);
			if (endptr == optarg)
				usage();
			break;
		case 'c':
			write_cnt = strtol(optarg, &endptr, 0);
			if (endptr == optarg)
				usage();
			break;
		case 's':
			size = strtol(optarg, &endptr, 0);
			if (endptr == optarg)
				usage();
			break;
		case 't':
			type = strtol(optarg, &endptr, 0);
			if (endptr == optarg)
				usage();
			if (type != 0 && type != 1)
				usage();
			break;
		case 'd':
			mode = strtol(optarg, &endptr, 0);
			if (endptr == optarg)
				usage();
			if ( mode < 0 || mode > 2)
				usage();
			break;
		case 'f':
			fast = true;
			break;
		case 'X':
			dbg = true;
			break;
		default:
			usage();
		}
	}

	if (type == 0) {
		setup_test(readint_read, writeint_write, open_open, close_close);
	} else if (type == 1) {
		setup_test(readint_fread, writeint_fwrite, open_fopen, close_fclose);
	} else
		return EXIT_FAILURE;

	while (loopcount != 0) {
		ints_written = 0;
		if (write_test(size, write_cnt, &ints_written, mode) != 0)
			break;

		if (read_test(size, ints_written) != 0)
			break;
		if (!fast)
			sleep(1);
		if (loopcount > 0)
			loopcount--;
	}

	return EXIT_SUCCESS;
}
