/* vt-tool.c - quick validator for valuetype format strings */

/* Copyright (c) 2010-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <cafe/cafe.h>
#include <cafe/errors.h>
#include <cafe/value_type.h>
#include "../src/cafe_assert.h"
#include "../lang/strbuf.h"

static int getln(struct cafe_byte_buffer *sb, FILE *stream)
{
	int c;
	int rc;

	if (feof(stream))
		return EOF;

	while ((c = fgetc(stream)) != EOF) {
		if (c == '\n')
			break;

		rc = strbuf_addc(sb, c);

		if (rc != CAFE_OK)
			return rc;
	}

	if (sb->len == 0)
		return EOF;

	sb->data[sb->len] = '\0';

	return 0;
}


static
void error(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (cafe_error_cur.rc == CAFE_OK) {
		vfprintf(stderr, fmt, ap);
		va_end(ap);
		fputs("(untracked)\n", stderr);
		return;
	}

	vfprintf(stderr, fmt, ap);
	va_end(ap);
	cafe_error_print(NULL, stderr, ": ");
	cafe_error_clear(NULL);
	fputc('\n', stderr);
}


int main(int argc, char * argv[])
{
	int rc;
	struct cafe_namespace *ns;
	struct cafe_value_type *type;
	FILE *stream = stdin;
	struct cafe_byte_buffer sb = STRBUF_INIT(cafe_stdlib_allocator);
	int ln = 1;
	const char *filename = "<stdin>";

	/* Read from stdin if no string specified on the command
	   line. */

	if (argc > 1) {
		filename = argv[1];
		stream = fopen(filename, "rb");
		if (stream == NULL) {
			fprintf(stderr, "cannot open: '%s': %s\n",
				filename, strerror(errno));
			exit(EXIT_FAILURE);
		}
	}

	CAFE_LOCK();
	rc = cafe_namespace_create(cafe_stdlib_allocator, &ns);
	if (rc != CAFE_OK) {
		error("Error building test namespace: %s\n", cafe_strerror(rc));
		return EXIT_FAILURE;
	}

	while (getln(&sb, stream) != EOF) {
		rc = cafe_type_parse(ns, sb.data, &type);

		if (rc != CAFE_OK) {
			error("%s:%d: error parsing value type string: %s\n",
			      filename, ln, cafe_strerror(rc));
		} else {
			printf("%s:%d: %s ### size %zu, align %u.\n",
			       filename, ln, sb.data, type->size, type->align);
		}
		ln++;
		strbuf_reset(&sb);
	}
	CAFE_UNLOCK();
	if (stream != stdin)
		fclose(stream);

	strbuf_release(&sb);
	cafe_namespace_delete(ns);

	return rc == CAFE_OK ? EXIT_SUCCESS : EXIT_FAILURE;
}
