/* ostimer.c - posix EP timer */

/* Copyright (c) 2009-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
14may14,sdt  Fix 'checked_return' warning
13may14,sdt  Rename cafe tasks to tAnalysis* - US37854
24feb14,sdt  Disable SLEEP warnings

DESCRIPTION
 This cafe configuration emit timer event

 */

#include "../src/cafe_assert.h"
#include "../src/cafe_os.h"
#include "../src/time_utils.h"
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/value_type.h>
#include <cafe/ep_worker.h>
#include <cafe/ep/ostimer.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#define WORKAREA_DEF "S[cafe_ostimer.workarea,"				\
	"v[interval,Q," cafe_setting_VT_NAT "],"			\
	cafe_output_inst_thread_VT_NAT "]"
#define WORKAREA_REF "t[cafe_ostimer.workarea]"

struct client_data {
	struct cafe_ep_worker *worker;
	pthread_cond_t cond;
	pthread_mutex_t mutex;

	long change_count;
	cafe_eal_handle_t timer_actions;
	void *timer_data[2];
};

struct ostimer_workarea {
	unsigned long long interval;	// in microseconds
};

struct ostimer_event {
	unsigned long long tick;
};

static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(OSTIMER_REF)
};


static
void timer_thread(void *arg)
{
	struct ostimer_workarea *w;
	struct cafe_ep_inst *inst = (struct cafe_ep_inst *)arg;
	struct client_data *node = inst->client_data;
	int ret;
	struct timespec timeout;

	assert(inst->workarea != NULL);
	w = (struct ostimer_workarea *)inst->workarea->u.data;

	while (cafe_ep_worker_isrunning(node->worker)) {
		/* lock removed by a different thread */
		/* coverity[sleep] */
		(void)pthread_mutex_lock(&node->mutex);
		if (clock_gettime(CLOCK_REALTIME, &timeout)) {
			(void)pthread_mutex_unlock(&node->mutex);
			perror("ostimer thread: error getting current time");
			cafe_ep_worker_abort(node->worker);
			return;
		}

		timespec_ticks_add((unsigned long)w->interval, USECS_PER_SEC,
		                   &timeout);
		
		/* lock removed by a different thread */
		/* coverity[sleep] */
		ret = pthread_cond_timedwait(&node->cond, &node->mutex,
					     &timeout);
		(void)pthread_mutex_unlock(&node->mutex);
		if (ret != ETIMEDOUT)
			break;

		cafe_dispatch_event(&node->timer_actions, node->timer_data);
	}
}

static
int ostimer_inst_create(struct cafe_ep_inst *inst)
{
	struct ostimer_workarea *w;
	struct client_data *node;
	int ret;

	assert(inst->workarea != NULL);
	w = (struct ostimer_workarea *) inst->workarea->u.data;
	if (w->interval == 0) {
		(void)fprintf(stderr, "No interval specified, use 1 second as default.\n");
		w->interval = USECS_PER_SEC;
	}

	if ((node = CAFE_ALLOC(inst->allocator, sizeof *node)) == NULL)
		return CAFE_NOMEM;

	memset(node, 0, sizeof *node);
	ret = pthread_cond_init(&node->cond, NULL);
	if (ret != 0) {
		CAFE_FREE(inst->allocator, node);
		return CAFE_ERROR;
	}

	ret = pthread_mutex_init(&node->mutex, NULL);
	if (ret != 0) {
		(void)pthread_cond_destroy(&node->cond);
		CAFE_FREE(inst->allocator, node);
		return CAFE_ERROR;
	}

	ret = cafe_ep_worker_create("tAnalysisOstimer", inst, timer_thread, inst,
				    &node->worker);
	if (ret != CAFE_OK) {
		(void)pthread_mutex_destroy(&node->mutex);
		(void)pthread_cond_destroy(&node->cond);
		CAFE_FREE(inst->allocator, node);
		return ret;
	}

	node->timer_actions = CAFE_EAL_HANDLE_NULL;

	inst->client_data = node;

	return CAFE_OK;
}

static
void ostimer_inst_delete(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;

	cafe_ep_worker_abort(node->worker);
	(void)pthread_mutex_lock(&node->mutex);
	/* previous call of cafe_ep_worker_abort should ensure that
	 * cond_broadcast will be not blocking */
	/* coverity[sleep] */
	(void)pthread_cond_broadcast(&node->cond);
	(void)pthread_mutex_unlock(&node->mutex);
	cafe_ep_worker_delete(node->worker);

	(void)pthread_mutex_destroy(&node->mutex);
	(void)pthread_cond_destroy(&node->cond);
	cafe_event_action_list_release(&node->timer_actions);
	CAFE_FREE(inst->allocator, node);
}

static
int ostimer_inst_start(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	int rc;

	rc = cafe_ep_worker_start(node->worker);
	if (rc != CAFE_OK)
		return rc;
	return cafe_ep_worker_wait(node->worker);
}

static
int ostimer_inst_stop(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	int rc;

	rc = cafe_ep_worker_stop(node->worker);
	if (rc != CAFE_OK)
		return rc;

	(void)pthread_mutex_lock(&node->mutex);
	/* previous call of cafe_ep_worker_stop should ensure that 
	 * cond_broadcast will be not blocking */
	/* coverity[sleep] */
	(void)pthread_cond_broadcast(&node->cond);
	(void)pthread_mutex_unlock(&node->mutex);

	return cafe_ep_worker_wait(node->worker);
}

static
int ostimer_cfg_update_output_event_actions(struct cafe_ep_inst *inst)
{
	int rc;
	struct client_data *inst_data = (struct client_data *)inst->client_data;
	unsigned int i;

	if (inst_data->change_count == inst->session->output_event_action_change_count)
		return CAFE_OK;

	inst_data->change_count = inst->session->output_event_action_change_count;

	for (i = 0; i < inst->enabled_output_events.ind; ++i) {
		struct cafe_event_descr *event =
			inst->enabled_output_events.list[i].event;
		struct cafe_event_action_list *sink_actions;
		cafe_eal_handle_t sink_actions_handle = CAFE_EAL_HANDLE_NULL;

		rc = cafe_event_action_list_alloc(inst, event,
				&sink_actions,
				&sink_actions_handle);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sink_actions_handle);
			return rc;
		}

		cafe_event_action_list_swap(&inst_data->timer_actions,
				sink_actions_handle);
	}

	return CAFE_OK;
}

static struct cafe_ep_spec spec = {
	"ostimer",
	ostimer_inst_create,
	ostimer_inst_delete,
	ostimer_inst_start,
	ostimer_inst_stop,
	NULL,
	ostimer_cfg_update_output_event_actions,
	NULL,
	OSTIMER_DEF
	cafe_output_inst_thread_VT_DEF
	cafe_setting_VT_DEF
	WORKAREA_DEF, /* types */
	WORKAREA_REF,
	0,
	NULL,
	CAFE_NELEMENTS(output_events),
	output_events
};

static bool cfg_registered = false;

int cafe_ostimer_cfg_register(void)
{
	int rc;

	if (cfg_registered)
		return CAFE_OK;

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		cfg_registered = true;

	return rc;
}

void cafe_ostimer_cfg_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}
