/* cpu_hog_pthread.c - exercise system with a known load pattern */

/* Copyright (c) 2011-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#if defined(__linux__)
#include <sys/syscall.h>
#elif defined(__sun__)
#include <sys/processor.h>
#include <sys/procset.h>
#endif

#define NELEMENTS(A)		((sizeof((A)) / sizeof((A))[0]))
#define CLOCK_ID		CLOCK_REALTIME
#define ONE_EXP9		1000000000L

/* 8192 should be enough for this year! */
#define MAX_THREADS		8192

static unsigned long long tickget(void)
{
	int rc;
	struct timespec ts;
	rc = clock_gettime(CLOCK_ID, &ts);
	if (rc != 0)
		fprintf(stderr, "Error getting current time: %s\n",
			strerror(errno));
	assert(rc == 0);
	return ts.tv_sec * ONE_EXP9 + ts.tv_nsec;
}

static void busyloop(unsigned long long duration)
{
	unsigned long long tick = tickget();
	while (tickget() < tick + duration);
}

static void goidle(long nanoseconds)
{
	struct timespec ts;

	ts.tv_sec = 0;
	ts.tv_nsec = nanoseconds;
	if (ts.tv_nsec == ONE_EXP9)
		ts.tv_nsec = nanoseconds -1;
	assert(ts.tv_nsec < ONE_EXP9);
	while (nanosleep(&ts, &ts))
		perror("nanosleep");
}

#if defined(__linux__)
static pid_t gettid(void)
{
	return syscall(SYS_gettid);
}
#endif

static void *thread_worker(void *arg)
{
	float *load = (float *)arg;

#define GRANULARITY 5

	while (1) {
		assert(*load <= 1.0);
		busyloop((*load * ONE_EXP9) / GRANULARITY);
		goidle((1.0 - *load) * ONE_EXP9 / GRANULARITY);
	}

	return arg;
}

static void usage(const char *progname)
{
	fprintf(stderr,
		"usage: %s [-n <seconds>] [-t <seconds>] [-N <nthreads>]\n"
		"\n"
		"This program will generate a known load pattern over all the online\n"
		"cores.  If the machine has only one core then the control loop and\n"
		"the busy loop will run on the same core.\n"
		"\n"
		"The known pattern looks like a sawtooth:\n"
		"\n"
		"  100     /|    /|    /|\n"
		"         / |   / |   / |\n"
		"        /  |  /  |  /  |  /\n"
		"       /   | /   | /   | /\n"
		"    0 /    |/    |/    |/\n"
		"\n"
		"The time it takes for the load to hit 100%% can be controlled\n"
		"with the -n option; the larger the value the longer it takes.\n"
		"\n"
		"OPTIONS\n"
		"  -n <seconds>  - number of seconds to reach 100%% usage (default 10)\n"
		"  -t <seconds>  - execution time (default is forever)\n"
		"  -N <nthreads> - the number of threads\n\n",
		progname);
}

int main(int argc, char *argv[])
{
	int opt;
	int num_steps = 10;
	int timeout_in_secs = -1;
	size_t i;
	pthread_t tid[MAX_THREADS];
	float load = 0.0;
	long num_threads = 0;
	long num_cores = 1;

	/* WRS_XXX use getopt_long for legibility. */

	while ((opt = getopt(argc, argv, "N:n:t:")) != -1) {
		switch (opt) {
		case 'n':
			num_steps = atoi(optarg);
			break;
		case 'N':
			num_threads = atoi(optarg);
			break;
		case 't':
			timeout_in_secs = atoi(optarg);
			break;
		default:
			usage(argv[0]);
			exit(EXIT_FAILURE);
		}
	}

	if (num_steps <= 0) {
		fprintf (stderr, "%s: invalid number of steps; must be > 0\n",
			 argv[0]);
		return EXIT_FAILURE;
	}

	num_cores = sysconf(_SC_NPROCESSORS_ONLN);
	if (num_threads == 0)
		num_threads = num_cores;
	assert(num_threads >= 1);
	assert(num_threads < NELEMENTS(tid));

	for (i = num_threads > 1 ? 1 : 0; i < num_threads; i++) {
#if defined(__linux__)
		cpu_set_t cpuset;
#endif
		int rc;

		rc = pthread_create(&tid[i], NULL, thread_worker, &load);
		if (rc != 0) {
			perror("error: pthread_create");
			exit(EXIT_FAILURE);
		}

#if defined(__linux__)
		CPU_ZERO(&cpuset);
		CPU_SET(i % num_cores, &cpuset);

		rc = sched_setaffinity(gettid(), sizeof(cpu_set_t), &cpuset);
#elif defined(__sun__)
		rc = processor_bind(P_LWPID, tid[i], i % num_cores, NULL);
#else
#  warning "Don't know how to set thread affinity for this operating system"
#endif
		/* Affinity errors are benign, for now. */
		if (rc != 0) {
			fprintf(stderr,
				"%s: ignoring failure to set cpu affinity "
				"(%s)\n", argv[0], strerror(rc));
		}
	}

	while (1) {
		for (i = 0; i <= num_steps; i++) {
			load = i / (float)num_steps;
			sleep(1);
			if (timeout_in_secs > 0) {
				if (--timeout_in_secs == 0)
					exit(EXIT_SUCCESS);
			}
		}
	}

	return EXIT_SUCCESS;
}
