/* cafe_print.c - print routines for cafe objects */

/* Copyright (c) 2010, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

#include <stdio.h>
#include <string.h>
#include <cafe/cafe.h>
#include <cafe/cafe_print.h>
#include <cafe/value_print.h>

/* Don't check return code of printf and fputs (fix Prevent warning) */

#define CAFE_PRINTF	(void)printf
#define CAFE_PUTS(s)	(void)fputs(s, stdout)

int cafe_ep_spec_print(const struct cafe_ep_spec *spec, int flags)
{
	const char *out_ev_spec;
	const struct cafe_ep_spec_input_evt *in_ev_spec;
	unsigned int i;
	int j;

	if (spec == NULL)
		return CAFE_ERROR;

	CAFE_PRINTF("namespace:			%s\n", spec->name);

	if (flags & CAFE_CFG_PRINT_FUNCS) {
		CAFE_PRINTF("inst_create hook:		%p\n",
		       spec->create_inst);
		CAFE_PRINTF("inst_delete hook:		%p\n",
		       spec->delete_inst);

		CAFE_PRINTF("start hook:			%p\n",
		       spec->start);
		CAFE_PRINTF("stop hook:			%p\n",
		       spec->stop);
	}

	CAFE_PRINTF("types string:                   %s\n", spec->types);
	CAFE_PRINTF("workarea vt string:		%s\n\n", spec->workarea);

	CAFE_PRINTF("%u input events\n", spec->n_input_events);

	if (flags & CAFE_CFG_PRINT_INPUT)
		for (i = 0; i < spec->n_input_events; i++) {
			in_ev_spec = spec->input_events+i;
			CAFE_PRINTF("%-3u vt string:    %s\n",
			       i, in_ev_spec->type[0]);
		}

	CAFE_PRINTF("\n%u output orthogonal data types\n", spec->n_output_odata);
	if (flags & CAFE_CFG_PRINT_OUTPUT) {
		for (i = 0; i < spec->n_output_odata; i++) {
			out_ev_spec = spec->output_odata[i];
			CAFE_PRINTF("%-3u vt string:    %s\n",
			       i, out_ev_spec);
		}
	}

	CAFE_PRINTF("\n%u output orthogonal data sets\n", spec->n_output_odata_sets);
	if (flags & CAFE_CFG_PRINT_OUTPUT) {
		for (i = 0; i < spec->n_output_odata_sets; i++) {
			CAFE_PRINTF("%d: (", i);
			for (j = 0; j < spec->output_odata_sets[i].length; j++) {
				if (j != 0)
					CAFE_PRINTF(" ");
				CAFE_PRINTF("%d", *(spec->output_odata_sets[i].od_items + j));
			}
			CAFE_PRINTF(")\n");
		}
	}

	CAFE_PRINTF("\n%u output events\n", spec->n_output_events);

	if (flags & CAFE_CFG_PRINT_OUTPUT)
		for (i = 0; i < spec->n_output_events; i++) {
			out_ev_spec = spec->output_events[i].type;
			if (spec->n_output_odata_sets == 0)
				CAFE_PRINTF("vt string:    %s\n",
				       out_ev_spec);
			else
				CAFE_PRINTF("od_set: %d      vt string:    %s\n",
				       spec->output_events[i].od_set_idx,
				       out_ev_spec);
		}

	CAFE_PUTS("\n\n");
	return CAFE_OK;
}


int cafe_ep_spec_print_all(int flags)
{
	struct cafe_ep_spec_iter iter;
	const struct cafe_ep_spec *cfg_spec;

	CAFE_PUTS("\n\n"
	      "Available instruments\n"
	      "------------------------\n");

	CAFE_LOCK();
	cfg_spec = cafe_ep_spec_iter_first(&iter);

	while (cfg_spec != NULL) {
		(void)cafe_ep_spec_print(cfg_spec, flags);
		CAFE_PUTS("=======================================\n\n");
		cfg_spec = cafe_ep_spec_iter_next(&iter);
	}
	CAFE_UNLOCK();
	return CAFE_OK;
}


int cafe_instrument_print(const struct cafe_instrument *session, int flags,
			  bool lock_taken)
{
	const struct cafe_list_head *ptr;
	unsigned int num_instances = 0;

	if (session == NULL)
		return CAFE_ERROR;

	CAFE_PRINTF("==== Session ==================\n\n"
	       "Session:                        %p\n"
	       "Session name:			%s\n", session, session->name);

	if (!lock_taken)
		CAFE_LOCK();

	cafe_list_foreach(ptr, &session->inst_list_head)
		++num_instances;

	if (!lock_taken)
		CAFE_UNLOCK();

	CAFE_PRINTF("%u instance%s\n\n",
	       num_instances, num_instances == 1 ? "" : "s");
	return CAFE_OK;
}


int cafe_instrument_print_all(int flags)
{
	struct cafe_instrument_iter iter;
	const struct cafe_instrument *session;

	CAFE_PUTS("\n\n"
	      "Available sessions\n"
	      "-------------------\n");

	CAFE_LOCK();
	session = cafe_instrument_iter_first(&iter);
	while (session != NULL) {
		(void)cafe_instrument_print(session, flags, 1);
		session = cafe_instrument_iter_next(&iter);
	}
	CAFE_UNLOCK();
	return CAFE_OK;
}


int cafe_ep_inst_print(const struct cafe_ep_inst *inst, int flags,
		       bool lock_taken)
{
	const struct cafe_list_head *list_item;
	unsigned int i;

	if (inst == NULL)
		return CAFE_ERROR;

	CAFE_PRINTF("==== Instance ==================\n\n"
	       "Instance:			%p\n", inst);
	CAFE_PRINTF("Specification name:		%s\n", inst->spec->name);

	CAFE_PRINTF("Workarea data:			%p\n", inst->workarea);

	CAFE_PRINTF("Sources:			");

	if (!lock_taken)
		CAFE_LOCK();

	cafe_list_foreach(list_item, &inst->source_list_head) {
		const struct cafe_ep_inst *source =
			SOURCE_LIST_NODE_ENTRY(list_item)->source;
		CAFE_PRINTF("%p(%s) ", source, source != NULL ?
		       source->spec->name : "None");
	}

	CAFE_PRINTF("\nSinks:				");
	cafe_list_foreach(list_item, &inst->sink_list_head) {
		const struct cafe_ep_inst *sink =
			SINK_LIST_NODE_ENTRY(list_item)->sink;
		CAFE_PRINTF("%p(%s) ", sink, sink != NULL ?
		       sink->spec->name : "None");
	}

	if (inst->event_action_data_list.ind == 1)
		CAFE_PUTS("\n1 input event");
	else
		CAFE_PRINTF("\n%u input events\n", inst->event_action_data_list.ind);

	for (i = 0; i < inst->event_action_data_list.ind; ++i) {
		struct cafe_event_action_info *action =
			inst->event_action_data_list.list[i];
		struct cafe_event_action_list *sink_actions;
		cafe_eal_handle_t unlock_handle;
		unsigned int j;
		int rc;
		CAFE_PRINTF("event:				%s\n",
		       action->event->types[0]->type_str);
		for (j = 1; j < action->event->num_types; ++j)
			CAFE_PRINTF("\t\t%s\n",
			       action->event->types[j]->type_str);
		if (action->sink_actions == CAFE_EAL_HANDLE_NULL)
			continue;

		rc = cafe_event_action_list_read_lock(
			&action->sink_actions, &unlock_handle, &sink_actions);
		if (rc != CAFE_OK)
			return rc;
		if (sink_actions == NULL)
			continue;

		CAFE_PUTS("\t\t(");
		for (j = 0; j < sink_actions->ind; ++j) {
			if (j > 0)
				CAFE_PUTS(",");
			CAFE_PRINTF("%p", sink_actions->list[j].handler);
		}
		CAFE_PUTS(")\n");
		cafe_event_action_list_read_unlock(unlock_handle);
	}

	if (!(flags & CAFE_CFG_PRINT_OUTPUT))
		return CAFE_OK;

	if (inst->enabled_output_events.ind == 1)
		CAFE_PUTS("\n1 enabled output event");
	else
		CAFE_PRINTF("\n%u enabled output events\n",
		       inst->enabled_output_events.ind);

	for (i = 0; i < inst->enabled_output_events.ind; ++i) {
		struct cafe_event_descr *event =
			inst->enabled_output_events.list[i].event;
		unsigned int j;

		CAFE_PRINTF("event:				%s\n",
		       event->types[0]->type_str);
		for (j = 1; j < event->num_types; ++j)
			CAFE_PRINTF("\t\t%s\n", event->types[j]->type_str);
	}

	if (!lock_taken)
		CAFE_UNLOCK();

	CAFE_PUTS("\n");
	return CAFE_OK;
}

int cafe_ep_inst_print_all(struct cafe_instrument *session,
			   int flags)
{
	struct cafe_ep_inst_iter iter = { NULL, NULL };
	const struct cafe_ep_inst *cfg_inst;

	if (session == NULL)
		return CAFE_ERROR;

	CAFE_PUTS("\n\n"
	          "Available instances\n"
	          "-------------------\n");

	CAFE_LOCK();
	cfg_inst = cafe_ep_inst_iter_first(session, &iter);
	while (cfg_inst != NULL) {
		(void)cafe_ep_inst_print(cfg_inst, flags, true);
		CAFE_PUTS("=======================================\n\n");
		cfg_inst = cafe_ep_inst_iter_next(&iter);
	}
	CAFE_UNLOCK();
	return CAFE_OK;
}


/* Note: This function is recursive */
int cafe_ep_inst_print_chain(const struct cafe_ep_inst *inst, int flags,
			     bool lock_taken)
{
	const struct cafe_ep_inst *source;
	const struct cafe_list_head *list_item;

	CAFE_PRINTF("\n\n"
	            "Instance chain, starting at %p(%s)\n\n", inst, 
	            inst->spec->name);

	if (!lock_taken)
		CAFE_LOCK();

	(void)cafe_ep_inst_print(inst, flags, 1);

	cafe_list_foreach(list_item, &inst->source_list_head) {
		source = SOURCE_LIST_NODE_ENTRY(list_item)->source;
		(void)cafe_ep_inst_print_chain(source, flags, true);
	}
	if (!lock_taken)
		CAFE_UNLOCK();
	return CAFE_OK;
}


int cafe_event_data_print(struct cafe_value_type *vt, const void *data)
{
	struct cafe_value_print_options options = CAFE_VALUE_PRINT_OPTIONS_INIT;

	if (vt == NULL) {
		CAFE_PUTS("(...)\n");
		return CAFE_OK;
	}

	(void)fputc('(', stdout);
	options.field_names = 1;
	cafe_value_print(stdout, NULL, vt, data, &options);
	CAFE_PUTS(")\n");
	return CAFE_OK;
}
