/* log_buffer.c - memory-based byte-oriented ring buffer */

/* Copyright (c) 2009-2010 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
  This buffer internally handles reader/writer interlocks, but the client must
  handle any possibility of multiple readers and/or multiple writers.

  Differentiates itself from rngLib.c by:
  - providing a non-malloc-based creation usage
  - when needed, mallocing everything in one chunk
  - "zero copy" read and writes

  Differentiates itself from rBuffLib.c by:
  - being simpler
  - not providing ability to grow dynamically

  To support the zero-copy-based interface, read and write pointers returned to
  callers always represent contiguous memory.
  Does not attempt to support the flight-recorder usecase, since that doesn't
  require read/write synchronization.
*/


#include <cafe/cafe.h>  /* for CAFE_[ALLOC,FREE]() */
#include <cafe/log_buffer.h>


/**
 *  Initialize the given address as a new log buffer.  This can be useful for
 * creating a log buffer based on pre-allocated memory.  It can also be useful
 * if the client wants to pre-fill the area with a pattern (to check high-water
 * marks, etc.).
 */
bool cafe_log_init(struct cafe_log_buf *log, size_t size)
{
	if (size <= sizeof(*log))
		return false;

	log->size = size - sizeof(struct cafe_log_buf);
	log->read_head = log->write_head = log->data;

	/* set trailing_top doesn't need to be set here, since it will only be
	   read after write_head has wrapped, but if we don't set it,
	   validation may result in false failures */
	log->trailing_top = log->data + log->size;

	return true;
}


/**
 *  Allocates space for a buffer of the given size and initializes the buffer.
 * When using this routine, it is the caller's responsibility to free() the
 * returned pointer when the buffer is no longer in use.
 */
struct cafe_log_buf *cafe_log_create(size_t size,
				     const struct cafe_allocator *allocator)
{
	struct cafe_log_buf *log =
		CAFE_ALLOC(allocator, sizeof(struct cafe_log_buf) + size);
	if (0 == log)
		return 0;

	if (!cafe_log_init(log, size + sizeof(struct cafe_log_buf))) {
		CAFE_FREE(allocator, log);
		return 0;
	}

	return log;
}


/**
 *  Reserve as much log space as possible for writing data. Returns 0
 * if the minimum number of bytes are not available, and a pointer to the
 * writeable area on success.  The bytes written to the area are not available
 * to be read until commit() is called.  It is the caller's responsibility to
 * not write more than "available" bytes to the returned pointer.
 * Note that the write head will always point at an empty byte.  This acts as
 * a flag so we can tell whether the buffer is completely full or completely
 * empty.  Otherwise, when read == write, then we have no way of knowing.
 */
char *cafe_log_write_reserve_min(struct cafe_log_buf *buf, size_t min,
				 size_t *bytes_available)
{
	/* save a function-local copy here, so concurrent reads don't disrupt
	   our calculations */
	char *const read_head = buf->read_head;
	size_t available;

	if (buf->write_head >= read_head) {
		available = buf->data + buf->size - buf->write_head;

		if (min <= available) {
			*bytes_available = available;
			return buf->write_head;
		}

		/* provide access to bytes potentially available wraparound
		   bytes */
		if (read_head == buf->data)
			/* we can only write to the end of the buf, which we've
			   already calculated to be too small. */
			return 0;

		available = read_head - buf->data - 1;
		if (min <= available) {
			buf->trailing_top = buf->write_head;
			buf->write_head = buf->data;
			*bytes_available = available;
			return buf->write_head;
		}

		return 0;
	}

	/* write head has wrapped around, but the read head hasn't yet */
	available = read_head - buf->write_head - 1;
	if (min <= available) {
		*bytes_available = available;
		return buf->write_head;
	}

	return 0;
}


/**
 *  Reserve enough space in the log for the given byte size.  Returns 0 on
 * failure and a pointer to the writeable area on success.  The bytes written
 * to the area are not available to be read until commit() is called.  It is
 * the caller's responsibility to not write more than "size" bytes to the
 * returned pointer.
 * Note that the write head will always point at an empty byte.  This acts as
 * a flag so we can tell whether the buffer is completely full or completely
 * empty.  Otherwise, when read == write, then we have no way of knowing.
 */
char *cafe_log_write_reserve(struct cafe_log_buf *buf, size_t size)
{
	size_t available;

	return cafe_log_write_reserve_min(buf, size, &available);
}


/**
 *  Flags the currently available write-area as being finished and available
 * for reading.  Callers may pass a smallersize here than what they actually
 * reserved, but they must never pass a greater size.  It is currently the
 * client's responsibility to ensure the reserve() and commit() routines are
 * always called in matching pairs.
 */
void cafe_log_write_commit(struct cafe_log_buf *buf, size_t size)
{
	buf->write_head += size;
	//td: if inlined, this probably needs a write memory barrier, etc.
}


/**
 *  Get a pointer to the next readable section of the buffer.  The size of the
 * readable area is returned in the "size" parameter, will be 0 if there's no
 * bytes currently written to the buffer.  The readable bytes will remain so
 * on successive calls to this function, until commit() is called.
 * It is the caller's responsibility to not read more than *size bytes from the
 * returned pointer.
 */
char *cafe_log_read_reserve(struct cafe_log_buf *buf, size_t *size)
{
	/* save a function-local copy here, so concurrent writes don't disrupt
	   our calculations */
	char *const write_head = buf->write_head;

	if (write_head >= buf->read_head) {
		*size = write_head - buf->read_head;
	} else {
		/* write head has wrapped around, but the read head hasn't
		   yet */
		*size = buf->trailing_top - buf->read_head;

		if (0 == *size) {
			/* need to wrap the read head */
			buf->read_head = buf->data;
			*size = write_head - buf->read_head;
		}
	}

	return buf->read_head;
}


/**
 *  Advances the buffer's read head by "size" bytes.  The given size must
 * never exceed the value returned by a preceeding call to read_reserve(), but
 * may be less than that value if the client consumed less than the currently
 * available buffer.
 */
void cafe_log_read_commit(struct cafe_log_buf *buf, size_t size)
{
	buf->read_head += size;
}
