/* log-dump.c - parse event stream files and emits the contents to stdout */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

/*
 * The log-dump application parses a phoenix event stream (file) and
 * prints the individual events and metadata to stdout.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <cafe/cafe.h>
#include <cafe/stdio.h>
#include <cafe/log_stream.h>
#include <cafe/cafe_print.h>
#include <cafe/autogenerated/cafe.type.h>	\

static
void usage(void)
{
	(void)printf("log-dump <log file>\n");
}


int main(int argc, char *argv[])
{
	unsigned long version;
	int rc;
	unsigned int sequence;
	struct cafe_event_map *event_map;
	FILE *input;
	struct cafe_byte_buffer data;
	struct cafe_instrument *session;
	struct cafe_value_type *metadata_event;

	if (argc != 2) {
		usage();
		return EXIT_FAILURE;
	}

	if ((rc = cafe_lib_init()) != CAFE_OK) {
		(void)fprintf(stderr, "error: cannot initialize cafe library: %s\n",
			cafe_strerror(rc));
		return EXIT_FAILURE;
	}

	CAFE_LOCK();
	
	rc = cafe_instrument_create("log-dump", &session);
	if (rc != CAFE_OK) {
		(void)fprintf(stderr, "error: cannot create session");
		return EXIT_FAILURE;
	}

	input = fopen(argv[1], "rb");
	if (input == 0) {
		fprintf(stderr, "Error opening %s: %s\n",
			argv[1], strerror(errno));
		return EXIT_FAILURE;
	}

	rc = cafe_fdecode_stream_header(input, &version);
	if (rc != CAFE_OK) {
		fprintf(stderr, "Error reading stream header from %s: %s\n",
			argv[1], cafe_strerror(rc));
		rewind(input);
	} else
		printf("Dumping log format %lu\n", version);

	cafe_bbuf_init(&data, cafe_stdlib_allocator, 0);

	event_map = cafe_event_map_create(session);
	assert(event_map != NULL);

	rc = cafe_type_parse(session->ns,
			     cafe_metadata_VT_DEF, &metadata_event);
	if (rc != CAFE_OK) {
		fprintf(stderr, "error: cannot initialize metadata event: %s\n",
			cafe_strerror(rc));
		return EXIT_FAILURE;
	}

	for(sequence = 0;;++sequence) {
		unsigned int i;
		unsigned long long eid, size;
		struct cafe_remote_event *remote;
		ssize_t header_bytes;
		void **output_event_data;

		header_bytes = cafe_fdecode_event_header(input, &size, &eid);
		if (header_bytes < 0) {
			if (feof_unlocked(input))
				break;
			printf("Error decoding event header @ 0x%lld\n",
			       (long long int)ftello(input));
			return EXIT_FAILURE;
		}

		if (eid == 0) {
			/* metadata event. */
			struct cafe_metadata md;
			struct cafe_remote_event *remote2;
			
			/* argument values */
			cafe_fdecode_val(metadata_event, &md,
					 cafe_stdlib_allocator, event_map,
					 input);
			rc = cafe_event_map_add_event(event_map, &md, NULL,
						      &remote2);
			if (rc != CAFE_OK)
				printf("Failed to add event description "
				       "\"%s\": %s\n",
				       ((char **)md.event_data.p)[0],
				       cafe_strerror(rc));
			printf("%u (%llx, %llx): New EID 0x%llx", sequence,
			       (long long int)ftello(input), size, md.eid);
			for (i = 0; i < md.event_data.l; ++i)
				printf(" [%s]", ((char **)md.event_data.p)[i]);
			printf("\n");
			cafe_release_values(metadata_event, &md,
					    cafe_stdlib_allocator);
			continue;
		}
		remote = cafe_event_map_lookup(event_map, eid);
		if (remote == 0) {
			printf("No metadata for event %llx (%llu), skipping\n",
			       eid, size);
			printf("seeking %lld -> %llx\n",
			       (long long int)ftello(input),
			       (long long)ftello(input) + size);
			fseeko(input, (off_t)size, SEEK_CUR);
			continue;
		}

		printf("%u (%llx, %llx): 0x%llx \"%s\"\n", sequence,
		       (long long int)ftello(input),
		       size, eid, remote->event->types[0]->ident->name);

		/* read event data orthogonal data payload */
		if (cafe_fread_event_data(input, remote->event,
					  cafe_stdlib_allocator, event_map,
					  &data) != CAFE_OK)
		{
			printf("unable to allocate %"PRIsize_t" bytes, skipping event\n",
			       remote->event->types[0]->size);
			fseeko(input, (off_t)size, SEEK_CUR);
			continue;
		}

		/* print event data and any orthogonal data */
		output_event_data = (void **)data.data;
		for (i = 0; i < remote->event->num_types; ++i) {
			printf("\t");
			cafe_event_data_print(remote->event->types[i],
					      output_event_data[i]);
			cafe_release_values(remote->event->types[i],
					    output_event_data[i],
					    cafe_stdlib_allocator);
		}
	}

	fclose(input);
	cafe_bbuf_release(&data);
	cafe_event_map_destroy(event_map);
	rc = cafe_instrument_delete(session);
	CAFE_UNLOCK();

	if (rc != CAFE_OK)
		fprintf(stderr, "Error deleting instrument: %s\n",
			cafe_strerror(rc));
	return EXIT_SUCCESS;
}
