/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
11mar14,sdt  Fix 'Checked Return' warnings
24feb14,sdt  Disable unexpected NO_EFFECT warnings
21feb14,sdt  Fix compiler warnings

DESCRIPTION

*/

#include "cafe_os.h"
#include "cafe_assert.h"
#include <string.h>
#include <errno.h>
#include <ctype.h>		/* for isascii() */
#include <cafe/cafe.h>
#include <cafe/errors.h>
#include <cafe/value_type.h>
#include <cafe/expr.h>
#include <cafe/byte_buffer.h>
#include "strtoll.h"
#include <cafe/array.h>
#include "cafe_trace.h"


enum {
	RESOLVED = 1,
	UNRESOLVED = 2
};

struct parse_context {
	struct cafe_namespace *ns;
	struct cafe_array new_types;
};

static void free_nature_ref(struct cafe_namespace *ns,
			    struct cafe_value_nature_ref *nature);

#if defined(_MSC_VER)
#pragma warning(disable:4244)	/* conversion error: possible loss of data */
#endif

static int build_single_type(struct parse_context *pc,
			     struct cafe_vm_ident **namep,
			     const char **fmt,
			     bool refonly);
static void free_ident(struct cafe_vm_ident *ident);
static void free_attributes(struct cafe_namespace *ns,
			    struct cafe_vm_attr *attr,
			    unsigned int elements);


static void parse_context_init(struct parse_context *pc,
			       struct cafe_namespace *ns)
{
	pc->ns = ns;
	cafe_array_init(&pc->new_types, ns->allocator,
			sizeof(struct cafe_vm_ident *));
}


static struct cafe_value_type *cafe_type_alloc(struct cafe_vm_ident *ident,
					       cafe_value_type_tag kind)
{
	struct cafe_value_type *type;

	assert(ident->type == NULL);
	type = CAFE_ALLOC(ident->ns->allocator, sizeof *type);
	if (type == NULL) {
		return NULL;
	}
	memset(type, 0, sizeof *type);
	type->kind = kind;
	type->ident = ident;
	ident->type = type;
	return type;
}

static void cafe_type_free(struct cafe_value_type *type)
{
	if (type == NULL) {
		return;
	}
	switch (type->kind) {
	case CAFE_TYPE_STRUCT:
	case CAFE_TYPE_NATURE_DEF:
		free_attributes(type->ident->ns, type->attributes,
				type->elements);
		free_ident(type->init_attr);
		break;
	}
	if (type->keymap) {
		CAFE_FREE(type->ident->ns->allocator, type->keymap);
	}
	CAFE_FREE(type->ident->ns->allocator, type->type_str);
	CAFE_FREE(type->ident->ns->allocator, type);
}

static const char *parseULL(const char *fmt, unsigned long long *result)
{
	char *str;

	errno = 0;
	*result = strtoull(fmt, &str, 0);
	if (errno != 0 || fmt >= str)
		return NULL;
	return str;
}

/*
 * Check if character is valid for ident
 */
static bool is_valid_ident(int c)
{
	return (c >= 'A' && c <= 'Z') ||
		(c >= 'a' && c <= 'z') ||
		(c == '%' || c == '_');
}

/*
 * Skip over an identifier
 */
static const char *skip_ident(const char *fmt)
{
	int hexmode = 0;
	int c;

	c = *fmt++;
	if (!is_valid_ident(c)) {
		return NULL;
	}
	for (;;) {
		c = *fmt;
		if (hexmode) {
			if (!((c >= 'A' && c <= 'F') ||
			      (c >= 'a' && c <= 'f') ||
			      (c >= '0' && c <= '9'))) {
				return NULL;
			}
			hexmode--;
		} else {
			if (!(is_valid_ident(c) || (c >= '0' && c <= '9'))) {
				break;
			}
			if (c == '%') {
				hexmode = 2;
			}
		}
		fmt++;
	}
	return fmt;
}


static size_t write_string(char *buf, const char *str, size_t len)
{
	char c;
	size_t l = 0;

	if (buf != NULL) {
		*buf++ = '"';
	}
	l++;
	while (len-- > 0) {
		c = *str++;
		if (c == '\0') {
			if (buf != NULL) {
				*buf++ = '\\';
				*buf++ = '0';
				*buf++ = '0';
			}
			l += 3;
		} else {
			if (buf != NULL)
				*buf++ = c;
			l++;
		}
	}
	if (buf != NULL) {
		*buf++ = '"';
		*buf = '\0';
	}
	l++;
	return l;
}

/*
 * Allocate identifier object
 */
static struct cafe_vm_ident *alloc_ident(
	struct cafe_namespace *ns,
	cafe_ident_type_tag kind,
	size_t size)
{
	struct cafe_vm_ident *ident;

	ident = CAFE_ALLOC(ns->allocator, size + sizeof(*ident));
	if (ident == NULL) {
		cafe_error_set(NULL, CAFE_NOMEM, "Building new identifier");
		return NULL;
	}
	memset(ident, 0, sizeof *ident);
	ident->kind = kind;
	ident->ns = ns;
	ident->len = size;
	return ident;
}

/*
 * build identifier object
 */
static struct cafe_vm_ident *build_ident(
	struct cafe_namespace *ns,
	cafe_ident_type_tag kind,
	const char *name,
	size_t size)
{
	struct cafe_vm_ident *ident;

	ident = alloc_ident(ns, kind, size);
	if (ident == NULL)
		return NULL;
	memcpy(ident->name, name, size);
	ident->name[size] = '\0';
	return ident;
}

/*
 * Free identifier object
 */
static void free_ident(struct cafe_vm_ident *ident)
{
	struct cafe_vm_ident *next;

	while (ident != NULL) {
		next = ident->peer;
		cafe_type_free(ident->type);
		if (ident->kind == CAFE_IDENT_TYPE) {
			free_ident(ident->children);
		}
		free_nature_ref(ident->ns, ident->natures);
		cafe_expr_free(ident->ns->allocator, ident->expr);
		CAFE_FREE(ident->ns->allocator, ident);
		ident = next;
	}
}

/*
 * Parse qualified name
 */
static const char *get_qualified_ident(struct cafe_namespace *ns,
				       const char *fmt,
				       struct cafe_vm_ident **identp)
{
	struct cafe_vm_ident **peers;
	struct cafe_vm_ident **tail;
	struct cafe_vm_ident *parent;
	struct cafe_vm_ident *ident;
	const char *name;
	size_t namesz;

	parent = NULL;
	peers = &ns->children;
	for (;;) {
		name = fmt;
		fmt = skip_ident(fmt);
		if (fmt == NULL) {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "Badly formed identifier \"%s\"", name);
			return NULL;
		}
		tail = peers;
		namesz = fmt - name;
		while ((ident = *tail) != NULL) {
			if (ident->len == namesz &&
			    memcmp(ident->name, name, namesz) == 0) {
				break;
			}
			tail = &ident->peer;
		}
		if (ident == NULL) {
			*tail = ident = build_ident(ns, CAFE_IDENT_TYPE, name,
						    namesz);
			if (ident == NULL) {
				return NULL;
			}
			ident->parent = parent;
		}
		if (*fmt != '.') {
			*identp = ident;
			return fmt;
		}
		fmt++;
		parent = ident;
		peers = &ident->children;
	}
}

static int get_basic_type_ident(
	struct cafe_namespace *ns,
	int kind,
	struct cafe_vm_ident **result)
{
	struct cafe_vm_ident **peers;
	struct cafe_vm_ident *ident;
	struct cafe_value_type *type;
	const char *name;
	size_t namelen;
	unsigned int align;
	size_t size;
	char type_str[32];

#define BUILTIN_TYPE_INIT(NAME, TYPE)					\
	do {								\
		name = "<" #NAME ">";					\
		namelen = (sizeof(#NAME)) + 2 - 1;			\
		align = alignof(TYPE);					\
		size = sizeof(TYPE);					\
		(void)snprintf(type_str, sizeof(type_str),		\
			"B[%s,%"PRIsize_t",%u]",			\
			 #NAME, size, align);				\
	} while(0)

	switch (kind) {
	case CAFE_TYPE_CHAR:
		BUILTIN_TYPE_INIT(c, char);
		break;

	case CAFE_TYPE_UCHAR:
		BUILTIN_TYPE_INIT(C, unsigned char);
		break;

	case CAFE_TYPE_SHORT:
		BUILTIN_TYPE_INIT(h, short);
		break;

	case CAFE_TYPE_USHORT:
		BUILTIN_TYPE_INIT(H, unsigned short);
		break;

	case CAFE_TYPE_INT:
		BUILTIN_TYPE_INIT(i, int);
		break;

	case CAFE_TYPE_UINT:
		BUILTIN_TYPE_INIT(I, unsigned int);
		break;

	case CAFE_TYPE_LONG:
		BUILTIN_TYPE_INIT(l, long);
		break;

	case CAFE_TYPE_ULONG:
		BUILTIN_TYPE_INIT(L, unsigned long);
		break;

	case CAFE_TYPE_LONGLONG:
		BUILTIN_TYPE_INIT(q, long long);
		break;

	case CAFE_TYPE_ULONGLONG:
		BUILTIN_TYPE_INIT(Q, unsigned long long);
		break;

	case CAFE_TYPE_STRING:
		BUILTIN_TYPE_INIT(s, char *);
		break;

	case CAFE_TYPE_BLOB:
		BUILTIN_TYPE_INIT(b, struct cafe_vm_blob);
		break;

	case CAFE_TYPE_BOOL:
		BUILTIN_TYPE_INIT(?, int);
		break;

	case CAFE_TYPE_TYPE:
		BUILTIN_TYPE_INIT(Y, struct cafe_value_type *);
		break;

	case CAFE_TYPE_FLOAT:
		BUILTIN_TYPE_INIT(f, float);
		break;

	case CAFE_TYPE_DOUBLE:
		BUILTIN_TYPE_INIT(d, double);
		break;

	default:
		cafe_error_set(NULL, CAFE_ERROR, "Expected builtin type but received %u", kind);
		return CAFE_ERROR;
	}
	peers = &ns->children;
	while ((ident = *peers) != NULL) {
		if (ident->len == namelen &&
		    memcmp(ident->name, name, namelen) == 0) {
			*result = ident;
			return CAFE_OK;
		}
		peers = &ident->peer;
	}
	ident = build_ident(ns, CAFE_IDENT_TYPE, name, namelen);
	if (ident == NULL) {
		return CAFE_NOMEM;
	}
	type = cafe_type_alloc(ident, kind);
	if (type == NULL) {
		CAFE_FREE(ns->allocator, ident);
		return CAFE_NOMEM;
	}
	type->align = align;
	type->size = size;
	type->type_str = CAFE_STRDUP(ns->allocator, type_str);
	if (type->type_str == NULL) {
		CAFE_FREE(ns->allocator, ident);
		return CAFE_NOMEM;
	}
	*peers = ident;
	*result = ident;
	return CAFE_OK;
}

static struct cafe_vm_ident *get_composit_ident(struct cafe_namespace *ns,
						int kind,
						unsigned int elements,
						const struct cafe_vm_ident *element_name,
						struct cafe_vm_expr *exprs)
{
	struct cafe_vm_ident **peers;
	struct cafe_vm_ident *parent;
	struct cafe_vm_ident *ident;
	struct cafe_vm_expr *elem;
	const char *name;
	const char *prefix = NULL;
	const char *suffix = NULL;
	struct cafe_byte_buffer buf = {NULL};
	size_t namelen;
	size_t prefixlen;
	size_t suffixlen = 0;
	char suffixbuf[30];
	int rc;

	assert(element_name != NULL);
	assert(element_name->kind == CAFE_IDENT_TYPE);
	assert(element_name->ns == ns);
	name = element_name->name;
	namelen = element_name->len;
	parent = element_name->parent;
	if (parent != NULL) {
		peers = &parent->children;
	} else {
		peers = &ns->children;
	}
	switch (kind) {
	case CAFE_TYPE_DICT:
		prefix = "dict<";
		suffix = ">";
		break;
	case CAFE_TYPE_ARRAY:
		(void)snprintf(suffixbuf, sizeof suffixbuf, ",%u>", elements);
		prefix = "array<";
		suffix = suffixbuf;
		break;
	case CAFE_TYPE_DYNARRAY:
		prefix = "array<";
		suffix = ">";
		break;
	case CAFE_TYPE_INSTANCE:
		rc = cafe_bbuf_init(&buf, ns->allocator, 0);
		if (rc != CAFE_OK) {
			cafe_error_set(NULL, rc,
			               "Failed to init buf of ident string");
			return NULL;
		}
		for (elem = exprs ? exprs->u.e : NULL;
		     elem != NULL; elem = elem->u.elem.next) {
			struct cafe_vm_expr *expr = elem->u.elem.e;

			if (expr->kind == CAFE_EXPR_UVALUE ||
			    expr->kind == CAFE_EXPR_SVALUE) {
				suffixlen = 30;
			} else if (expr->kind == CAFE_EXPR_STRING) {
				suffixlen = write_string(NULL, expr->u.str.p,
							 expr->u.str.l);
			} else {
				/* coverity[array_null] */
				assert(!"unexpected expression");
			}
			rc = cafe_bbuf_capacity_set(&buf,
						    buf.len + suffixlen + 3);
			if (rc != CAFE_OK) {
				cafe_error_set(NULL, rc,
					       "Failed to build ident string");
				return NULL;
			}

			if(expr != exprs->u.e)
				buf.data[buf.len++] = ',';

			if (expr->kind == CAFE_EXPR_UVALUE) {
				buf.len += snprintf(buf.data + buf.len,
						    buf.capacity - buf.len,
						    "%llu", expr->u.ull);
			} else if (expr->kind == CAFE_EXPR_SVALUE) {
				buf.len += snprintf(buf.data + buf.len,
						    buf.capacity - buf.len,
						    "%lld", expr->u.ll);
			} else if (expr->kind == CAFE_EXPR_STRING) {
				buf.len += write_string(buf.data + buf.len,
							expr->u.str.p,
							expr->u.str.l);
			} else {
				/* coverity[array_null] */
				assert(!"unexpected expression");
			}
		}
		prefix = "instance<";
		suffix = buf.data;
		break;

	default:
		/* coverity[array_null] */
		assert(!"unexpected kind");
	}
	prefixlen = strlen(prefix);
	suffixlen = strlen(suffix);
	while ((ident = *peers) != NULL) {
		if (ident->len == prefixlen + namelen + suffixlen &&
		    memcmp(ident->name, prefix, prefixlen) == 0 &&
		    memcmp(ident->name + prefixlen, name, namelen) == 0 &&
		    memcmp(ident->name + prefixlen + namelen, suffix, suffixlen) == 0) {
			cafe_bbuf_release(&buf);
			return ident;
		}
		peers = &ident->peer;
	}
	ident = alloc_ident(ns, CAFE_IDENT_TYPE,
			    prefixlen + namelen + suffixlen);
	if (ident == NULL) {
		cafe_bbuf_release(&buf);
		cafe_error_set(NULL, CAFE_NOMEM,
			       "allocating identifier object");
		return NULL;
	}
	memcpy(ident->name, prefix, prefixlen);
	memcpy(ident->name + prefixlen, name, namelen);
	memcpy(ident->name + prefixlen + namelen, suffix, suffixlen);
	ident->name[prefixlen + namelen + suffixlen] = '\0';
	*peers = ident;
	ident->parent = parent;
	cafe_bbuf_release(&buf);
	return ident;
}


/*
 * Free attribute list
 */
static void free_attributes(
	struct cafe_namespace *ns,
	struct cafe_vm_attr *attr,
	unsigned int elements)
{
	unsigned int i;

	for (i = 0; i < elements; i++) {
		free_ident(attr[i].ident);
	}
	if (attr)
		CAFE_FREE(ns->allocator, attr);
}


static int resolve_type(
	struct cafe_value_context *context,
	struct cafe_value_type *type)
{
	struct cafe_vm_attr *attr;
	struct cafe_value_type *attrtype;
	struct cafe_value_type *elemtype;
	struct cafe_vm_expr *e;
	int rval = 0;
	int keys;
	unsigned int align;
	size_t offset;

	if (type == NULL || type->align != 0) {
		return 0;
	}
	switch (type->kind) {
	case CAFE_TYPE_STRUCT:
	case CAFE_TYPE_NATURE_DEF:
		assert(type->ident->expr == NULL);
		context->type = type->ident;
		align = 1;
		offset = 0;
		cafe_type_attr_foreach(attr, type) {
			assert(attr->kind == CAFE_ATTR_VAR ||
			       attr->kind == CAFE_ATTR_KEY ||
			       attr->kind == CAFE_ATTR_CONST ||
			       attr->kind == CAFE_ATTR_PARAM);
			assert(type->attributes + attr->index == attr);
			attr->type = type;
			if (attr->ident->children == NULL) {
				rval |= UNRESOLVED;
				break;
			}
			attrtype = cafe_attr_type(attr);
			if (attrtype == NULL || attrtype->align == 0) {
				rval |= UNRESOLVED;
				break;
			}
			e = attr->ident->expr;
			if (e != NULL && e->type == NULL) {
				context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
				attr->ident->expr = e = cafe_expr_resolve_for_type(
					context, attr->ident->expr, cafe_base_type(attrtype));
				assert(e != NULL);
				if (e->type == NULL) {
					rval |= UNRESOLVED;
					break;
				}
			}
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			offset = (offset + attrtype->align - 1) & (0 - attrtype->align);
			attr->offset = offset;
			offset += attrtype->size;
			if (align < attrtype->align) {
				align = attrtype->align;
			}
		}
		if (rval & UNRESOLVED) {
			break;
		}
		type->align = align;
		type->size = (offset + align - 1) & (0 - align);
		rval |= RESOLVED;
		break;

	case CAFE_TYPE_DICT:
		assert(type->ident->expr == NULL);
		elemtype = cafe_element_type(type);
		if (elemtype == NULL || elemtype->align == 0) {
			rval |= UNRESOLVED;
			break;
		}
		elemtype = cafe_base_type(elemtype);
		assert(elemtype != NULL);
		if (elemtype->kind != CAFE_TYPE_STRUCT) {
			rval |= UNRESOLVED;
			break;
		}

		/* Count how many keys we have so we can allocate keymap */
		keys = 0;
		cafe_type_attr_foreach(attr, elemtype) {
			if (attr->kind == CAFE_ATTR_KEY) {
				keys++;
			}
		}
		if (keys == 0) {
			rval |= UNRESOLVED;
			break;
		}
		type->keymap = CAFE_ALLOC(type->ident->ns->allocator,
					  keys * sizeof *type->keymap);
		if (type->keymap == NULL) {
			rval |= UNRESOLVED;
			break;
		}
		type->elements = keys;

		/* Build keymap */
		keys = 0;
		cafe_type_attr_foreach(attr, elemtype) {
			if (attr->kind == CAFE_ATTR_KEY) {
				type->keymap[keys++] = attr;
			}
		}
		assert(keys == type->elements);
		type->size = sizeof(struct cafe_dict_struct);
		type->align = alignof(struct cafe_dict_struct);
		rval |= RESOLVED;
		break;

	case CAFE_TYPE_ARRAY:
		assert(type->ident->expr == NULL);
		elemtype = cafe_element_type(type);
		if (elemtype == NULL || elemtype->align == 0) {
			rval |= UNRESOLVED;
			break;
		}
		type->size = elemtype->size * type->elements;
		type->align = elemtype->align;
		rval |= RESOLVED;
		break;

	case CAFE_TYPE_DYNARRAY:
		assert(type->ident->expr == NULL);
		elemtype = cafe_element_type(type);
		if (elemtype == NULL || elemtype->align == 0) {
			rval |= UNRESOLVED;
			break;
		}
		type->size = sizeof(struct cafe_vm_dynarray);
		type->align = alignof(struct cafe_vm_dynarray);
		rval |= RESOLVED;
		break;

	case CAFE_TYPE_TYPEDEF:
		assert(type->ident->expr == NULL);
		elemtype = cafe_element_type(type);
		if (elemtype == NULL || elemtype->align == 0) {
			rval |= UNRESOLVED;
			break;
		}
		type->size = elemtype->size;
		type->align = elemtype->align;
		rval |= RESOLVED;
		break;

	case CAFE_TYPE_INSTANCE:
		elemtype = cafe_element_type(type);
		if (elemtype == NULL || elemtype->align == 0) {
			rval |= UNRESOLVED;
			break;
		}
		assert(elemtype->kind == CAFE_TYPE_STRUCT);
		assert(type->ident->expr != NULL);
		context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
		type->ident->expr = e = cafe_expr_resolve_for_type(
			context, type->ident->expr, type);
		if (e->type == NULL) {
			rval |= UNRESOLVED;
			break;
		}
		type->size = elemtype->size;
		type->align = elemtype->align;
		rval |= RESOLVED;
		break;
	}
	return rval;
}

static int resolve_types(
	struct cafe_namespace *ns,
	struct cafe_vm_ident **types,
	size_t num_types)
{
	size_t i;
	int rval;
	int retries;
	struct cafe_value_context *context;

	context = cafe_value_context_create(ns->allocator);
	if (context == NULL)
		return CAFE_NOMEM;

	/* Resolve all types */
	retries = 10;
	do {
		rval = 0;
		for (i = 0; i < num_types; i++) {
			rval |= resolve_type(context, types[i]->type);
		}
	} while (rval & RESOLVED && rval & UNRESOLVED && retries-- > 0);
	cafe_value_context_delete(context);
	return (rval & UNRESOLVED) != 0 ? CAFE_UNRESOLVED_VALUE_TYPE : CAFE_OK;
}

/*
 * Parse an expression.
 *
 * Use *exprp as the stack root to avoid explicit cleanup handling on
 * error.  The caller will perform the cleanup both on okay and error.
 */
static const char *get_expression(struct cafe_value_context *context,
				  struct cafe_vm_expr **exprp,
				  const char *fmt)
{
	int rc = cafe_expr_parse(context, &fmt, exprp);
	return rc == CAFE_OK ? fmt : NULL;
}

static struct cafe_value_nature_ref *nature_create(
	struct cafe_vm_ident *ident,
	struct cafe_vm_expr *args)
{
	struct cafe_value_nature_ref *nature = CAFE_ALLOC(ident->ns->allocator,
							  sizeof(*nature));
	if (nature == NULL)
		return NULL;

	nature->def = ident;
	nature->args = args;
	nature->next = NULL;

	return nature;
}

static const char *build_nature_ref_list(struct parse_context *pc,
					 struct cafe_value_nature_ref **natures,
					 const char *fmt)
{
	struct cafe_vm_ident *ident;
	struct cafe_vm_expr *args;

	while(*fmt == 'n') {
		fmt++;
		if (*fmt++ != '[') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "Expected '[' in nature list");
			return NULL;
		}
		fmt = get_qualified_ident(pc->ns, fmt, &ident);
		if (fmt == NULL) {
			return NULL;
		}
		if (*fmt == ',') {
			struct cafe_value_context *context =
				cafe_value_context_create(pc->ns->allocator);
			int rc;
			assert(context != NULL);
			context->ns = pc->ns;
			context->type = ident;
			fmt++;
			rc = cafe_expr_list_parse(context, &fmt, &args);
			cafe_value_context_delete(context);
			if (rc != CAFE_OK)
				return NULL;
		} else
			args = NULL;

		if (*fmt++ != ']') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "Expected ']' in nature list");
			return NULL;
		}

		*natures = nature_create(ident, args);
		if (*natures == NULL) {
			cafe_error_set(NULL, CAFE_NOMEM,
				       "creating nature list");
			return NULL;
		}
		natures = &(*natures)->next;
	}
	return fmt;
}

static bool same_nature_ref_list(const struct cafe_value_nature_ref *nature1,
				 const struct cafe_value_nature_ref *nature2)
{
	while(nature1 != NULL && nature2 != NULL) {
		if (nature1->def != nature2->def ||
		    !cafe_expr_same(nature1->args, nature2->args)) {
			return 0;
		}
		nature1 = nature1->next;
		nature2 = nature2->next;
	}
	return nature1 == NULL && nature2 == NULL;
}

static void free_nature_ref(struct cafe_namespace *ns,
			    struct cafe_value_nature_ref *nature)
{
	struct cafe_value_nature_ref *next;

	while (nature != NULL) {
		cafe_expr_free(ns->allocator, nature->args);
		next = nature->next;
		CAFE_FREE(ns->allocator, nature);
		nature = next;
	}
}

static const char *get_settings(struct cafe_namespace *ns,
				const char *fmt,
				struct cafe_value_type *type,
				struct cafe_vm_expr **result)
{
	struct cafe_vm_expr *list;
	struct cafe_vm_expr **exprs;
	struct cafe_vm_attr *attr;
	struct cafe_value_type *attrtype;
	struct cafe_vm_expr *expr;

	list = cafe_expr_new_list(ns->allocator);
	if (list == NULL)
		return NULL;
	exprs = &list->u.e;
	cafe_type_attr_foreach(attr, type) {
		if (attr->kind != CAFE_ATTR_PARAM) {
			assert(attr->kind == CAFE_ATTR_VAR ||
			       attr->kind == CAFE_ATTR_KEY ||
			       attr->kind == CAFE_ATTR_CONST);
			continue;
		}
		if (*fmt != ',' && *fmt != ']') {
			struct cafe_value_context *context;
			context = cafe_value_context_create(ns->allocator);
			assert(context != NULL);
			context->ns = ns;
			context->type = cafe_attr_type(attr)->ident;
			fmt = get_expression(context, &expr, fmt);
			cafe_value_context_delete(context);
			if (fmt == NULL)
				goto error;
		} else if (attr->ident->expr != NULL) {
			expr = cafe_expr_copy(ns->allocator, attr->ident->expr);
			if (expr == NULL)
				goto error;
		} else {
			goto error;
		}
		*exprs = expr = cafe_expr_new_element(ns->allocator, expr);
		if (expr == NULL)
			goto error;
		exprs = &expr->u.elem.next;
		attrtype = cafe_base_type(cafe_attr_type(attr));
		assert(attrtype != NULL);
		switch (attrtype->kind) {
		case CAFE_TYPE_CHAR:
		case CAFE_TYPE_UCHAR:
		case CAFE_TYPE_SHORT:
		case CAFE_TYPE_USHORT:
		case CAFE_TYPE_INT:
		case CAFE_TYPE_UINT:
		case CAFE_TYPE_LONG:
		case CAFE_TYPE_ULONG:
		case CAFE_TYPE_LONGLONG:
		case CAFE_TYPE_ULONGLONG:
		case CAFE_TYPE_BOOL:
			if (!cafe_expr_is_iconst(expr->u.elem.e))
				goto error;
			break;

		case CAFE_TYPE_STRING:
		case CAFE_TYPE_BLOB:
			if (expr->u.elem.e->kind != CAFE_EXPR_STRING)
				goto error;
			break;

		/* WRS_TODO: support more types */
		default:
			goto error;
		}
		if (*fmt == ',')
			fmt++;
	}
	*result = list;
	return fmt;
error:
	cafe_expr_free(ns->allocator, list);
	return NULL;
}

static const char *build_instance_ref(struct parse_context *pc,
				      struct cafe_vm_ident **identp,
				      const char *fmt)
{
	struct cafe_vm_ident *element_ident;
	struct cafe_vm_expr *init_args;
	struct cafe_vm_ident *ident;
	struct cafe_value_type *type;
	int rc;

	if (*fmt++ != '[') {
		cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
			       "Expected '[' in instance reference");
		return NULL;
	}
	fmt = get_qualified_ident(pc->ns, fmt, &element_ident);
	if (fmt == NULL || element_ident->type == NULL ||
	    element_ident->type->kind != CAFE_TYPE_STRUCT) {
		return NULL;
	}
	if (*fmt == ',') {
		fmt++;
	} else if (*fmt != ']') {
		cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
			       "Expected ']' in instance reference");
		return NULL;
	}
	fmt = get_settings(pc->ns, fmt, element_ident->type, &init_args);
	if (fmt == NULL) {
		cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
			       "error building settings for instance reference");
		return NULL;
	}
	if (*fmt++ != ']') {
		cafe_expr_free(pc->ns->allocator, init_args);
		cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
			       "Expected ']' in instance reference");
		return NULL;
	}
	*identp = ident = get_composit_ident(pc->ns, CAFE_TYPE_INSTANCE,
					     0, element_ident, init_args);
	if (ident == NULL) {
		cafe_expr_free(pc->ns->allocator, init_args);
		return NULL;
	}
	if (ident->type != NULL) {
		cafe_expr_free(pc->ns->allocator, init_args);
		return fmt;
	}
	rc = cafe_array_append(&pc->new_types, &ident);
	if (rc != CAFE_OK) {
		cafe_expr_free(pc->ns->allocator, init_args);
		return NULL;
	}
	type = cafe_type_alloc(ident, CAFE_TYPE_INSTANCE);
	if (type == NULL) {
		cafe_expr_free(pc->ns->allocator, init_args);
		return NULL;
	}
	type->element = element_ident;
	ident->expr = init_args;
	return fmt;
}

static const char *build_filter(struct cafe_namespace *ns,
				struct cafe_vm_ident **identp, const char *fmt)
{
	static const char *NAME_PREFIX = "filter<";
	struct cafe_vm_ident **tail;
	struct cafe_vm_ident *ident;
	struct cafe_vm_expr *e;
	struct cafe_value_type *type;
	struct cafe_value_context *context;
	struct cafe_byte_buffer namebuf;
	const char *expr_start;
	ptrdiff_t expr_len;
	int rc;

	if (*fmt++ != '[') {
		cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
			       "Expected '[' in filter definition");
		return NULL;
	}

	context = cafe_value_context_create(ns->allocator);
	assert(context != NULL);
	context->ns = ns;

	expr_start = fmt;
	rc = cafe_expr_parse(context, &fmt, &e);
	expr_len = fmt - expr_start;
	cafe_value_context_delete(context);

	if (rc != CAFE_OK) {
		cafe_error_set(NULL, rc,
			       "error parsing filter definition expression");
		return NULL;
	}
	if (*fmt++ != ']') {
		cafe_expr_free(ns->allocator, e);
		cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
			       "Expected ']' in filter definition");
		return NULL;
	}

	rc = cafe_bbuf_init(&namebuf, ns->allocator, 128);
	if (rc != CAFE_OK) {
		cafe_expr_free(ns->allocator, e);
		return NULL;
	}
	/* Above allocation (128) is enough */
	/* coverity[secure_coding] */
	strcpy(namebuf.data, NAME_PREFIX);
	namebuf.len = strlen(NAME_PREFIX);

	rc = cafe_bbuf_append(&namebuf, expr_start, expr_len);
	if (rc != CAFE_OK) {
		cafe_bbuf_release(&namebuf);
		cafe_expr_free(ns->allocator, e);
		return NULL;
	}
	rc = cafe_bbuf_append(&namebuf, ">", 2);
	if (rc != CAFE_OK) {
		cafe_bbuf_release(&namebuf);
		cafe_expr_free(ns->allocator, e);
		return NULL;
	}

	ident = build_ident(ns, CAFE_IDENT_TYPE, namebuf.data, namebuf.len);
	if (ident == NULL) {
		cafe_bbuf_release(&namebuf);
		cafe_expr_free(ns->allocator, e);
		return NULL;
	}

	cafe_bbuf_release(&namebuf);
	type = cafe_type_alloc(ident, CAFE_TYPE_FILTER);
	if (type == NULL) {
		CAFE_FREE(ns->allocator, ident);
		cafe_expr_free(ns->allocator, e);
		return NULL;
	}
	type->align = 1;
	ident->expr = e;

	tail = &ns->children;
	while (*tail != NULL)
		tail = &(*tail)->peer;
	*tail = *identp = ident;

	return fmt;
}

static const char *build_builtin(struct parse_context *pc,
				 struct cafe_vm_ident **identp,
				 const char *fmt)
{
	struct cafe_vm_ident *ident;
	struct cafe_vm_expr *exprs;
	struct cafe_vm_expr *e1;
	struct cafe_vm_expr *e2;
	struct cafe_value_context *context;
	struct cafe_value_type *type;
	int rc;

	*identp = NULL;
	if (*fmt++ != '[') {
		return NULL;
	}
	rc = build_single_type(pc, &ident, &fmt, true);
	if (rc != CAFE_OK) {
		return NULL;
	}
	if (*fmt++ != ',') {
		return NULL;
	}
	context = cafe_value_context_create(pc->ns->allocator);
	assert(context != NULL);
	rc = cafe_builtin_type(pc->ns, CAFE_TYPE_INT, &type);
	if (rc != CAFE_OK) {
		cafe_value_context_delete(context);
		return NULL;
	}
	context->ns = pc->ns;
	context->type = type->ident;
	rc = cafe_expr_list_parse(context, &fmt, &exprs);
	cafe_value_context_delete(context);
	if (rc != CAFE_OK ||
	    (e1 = exprs->u.e) == NULL ||
	    (e2 = e1->u.elem.next) == NULL ||
	    e2->u.elem.next != NULL ||
	    !cafe_expr_is_iconst(e1->u.elem.e) ||
	    !cafe_expr_is_iconst(e2->u.elem.e))
	{
		cafe_expr_free(pc->ns->allocator, exprs);
		return NULL;
	}

	/* ident->type->size = cafe_expr_to_ull(e1->u.elem.e); */
	/* ident->type->align = cafe_expr_to_ull(e2->u.elem.e); */

	cafe_expr_free(pc->ns->allocator, exprs);
	if (*fmt++ != ']') {
		return NULL;
	}
	*identp = ident;
	return fmt;
}

static const char *build_attribute_list(struct parse_context *pc,
					struct cafe_vm_attr **attrs,
					unsigned int *elements,
					const char *fmt)
{
	struct cafe_vm_ident *ident;
	struct cafe_vm_attr *attr = NULL;
	unsigned int max = 0;
	unsigned int ind = 0;
	const char *name;
	cafe_ident_type_tag kind = CAFE_IDENT_TYPE;
	int rc;
	int c;

	for (;;) {
		c = *fmt++;
		switch (c) {
		case 'v':
		case 'k':
		case 'K':
		case 'p':
			switch (c) {
			case 'v':
				kind = CAFE_ATTR_VAR;
				break;

			case 'k':
				kind = CAFE_ATTR_KEY;
				break;

			case 'K':
				kind = CAFE_ATTR_CONST;
				break;

			case 'p':
				kind = CAFE_ATTR_PARAM;
				break;

			default:
				/* coverity[array_null] */
				assert(!"unexpected type");
			}
			if (*fmt++ != '[') {
				goto free_attr_and_error;
			}
			name = fmt;
			fmt = skip_ident(fmt);
			if (fmt == NULL) {
				goto free_attr_and_error;
			}
			ident = build_ident(pc->ns, CAFE_IDENT_ATTR, name,
					    fmt - name);
			if (ident == NULL) {
				goto free_attr_and_error;
			}
			if (*fmt == ',') {
				fmt++;
				rc = build_single_type(pc, &ident->children,
						       &fmt, true);
				if (rc != CAFE_OK) {
					cafe_error_set(NULL, rc, "Error building type at '%s'", fmt);
					goto free_ident_and_error;
				}
			}
			if (*fmt == ',') {
				fmt = build_nature_ref_list(pc, &ident->natures, fmt + 1);
				if (fmt == NULL) {
					goto free_ident_and_error;
				}
			}
			if (*fmt == ',') {
				struct cafe_value_context *context =
					cafe_value_context_create(pc->ns->allocator);
				assert(context != NULL);
				context->ns = pc->ns;
				context->type = ident->children;
				fmt = get_expression(context, &ident->expr,
						     fmt + 1);
				cafe_value_context_delete(context);
				if (fmt == NULL) {
					goto free_ident_and_error;
				}
			}
			if (*fmt++ != ']') {
				cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
					       "expected ']' before '%s'",
					       fmt - 1);
				goto free_ident_and_error;
			}
			break;

		default:
			*attrs = attr;
			*elements = ind;
			return fmt - 1;
		}
		if (ind == max) {
			struct cafe_vm_attr *newattr;
			max = max ? max * 2 : 1;
			newattr = CAFE_REALLOC(pc->ns->allocator, attr,
					       max * sizeof *attr);
			if (newattr == NULL)
				goto free_ident_and_error;
			attr = newattr;
			memset(attr + ind, 0, (max - ind) * sizeof *attr);
		}
		attr[ind].kind = kind;
		attr[ind].ident = ident;
		attr[ind].index = ind;
		ind++;
	}
free_ident_and_error:
	free_ident(ident);
free_attr_and_error:
	free_attributes(pc->ns, attr, ind);
	return NULL;
}

static bool same_attribute_list(const struct cafe_vm_attr *attr1,
				const struct cafe_vm_attr *attr2,
				unsigned int elements)
{
	while (elements > 0) {
		assert(attr1->kind == CAFE_ATTR_VAR ||
		       attr1->kind == CAFE_ATTR_KEY ||
		       attr1->kind == CAFE_ATTR_CONST ||
		       attr1->kind == CAFE_ATTR_PARAM);
		assert(attr2->kind == CAFE_ATTR_VAR ||
		       attr2->kind == CAFE_ATTR_KEY ||
		       attr2->kind == CAFE_ATTR_CONST ||
		       attr2->kind == CAFE_ATTR_PARAM);
		assert(attr1->ident->ns == attr2->ident->ns);
		if (attr1->kind != attr2->kind ||
		    attr1->ident->len != attr2->ident->len ||
		    strcmp(attr1->ident->name, attr2->ident->name) != 0 ||
		    attr1->ident->children != attr2->ident->children ||
		    !same_nature_ref_list(attr1->ident->natures, attr2->ident->natures) ||
		    !cafe_expr_same(attr1->ident->expr, attr2->ident->expr)) {
			return false;
		}
		attr1++;
		attr2++;
		elements--;
	}
	return true;
}

int cafe_type_dynarray(
	const struct cafe_value_type *element_type,
	struct cafe_value_type **result)
{
	struct cafe_value_type *type;
	struct cafe_vm_ident *element_ident = element_type->ident;
	struct cafe_vm_ident *ident = get_composit_ident(
		element_ident->ns, CAFE_TYPE_DYNARRAY, 0, element_ident, NULL);
	if (ident == NULL)
		return CAFE_ERROR;
	if ((type = ident->type) == NULL) {
		type = cafe_type_alloc(ident, CAFE_TYPE_DYNARRAY);
		if (type == NULL)
			return CAFE_NOMEM;
		type->element = element_ident;
	}

	type->size = sizeof(struct cafe_vm_dynarray);
	type->align = alignof(struct cafe_vm_dynarray);

	*result = type;
	return CAFE_OK;
}

/*
 * Build single value type
 */
static int build_single_type(struct parse_context *pc,
			     struct cafe_vm_ident **identp,
			     const char **fmtp,
			     bool refonly)
{
	const char *fmt = *fmtp;
	int c;
	int i = 0;
	int rc;
	cafe_value_type_tag kind;
	struct cafe_vm_ident *ident = NULL;
	struct cafe_value_type *type;
	struct cafe_vm_ident *element_ident;

	c = *fmt++;
	switch (c) {
	case 'c':
	case 'C':
	case 'h':
	case 'H':
	case 'i':
	case 'I':
	case 'l':
	case 'L':
	case 'q':
	case 'Q':
	case 's':
	case 'b':
	case '?':
	case 'Y':
	case 'f':
	case 'd':
		switch (c) {
		case 'c':
			kind = CAFE_TYPE_CHAR;
			break;

		case 'C':
			kind = CAFE_TYPE_UCHAR;
			break;

		case 'h':
			kind = CAFE_TYPE_SHORT;
			break;

		case 'H':
			kind = CAFE_TYPE_USHORT;
			break;

		case 'i':
			kind = CAFE_TYPE_INT;
			break;

		case 'I':
			kind = CAFE_TYPE_UINT;
			break;

		case 'l':
			kind = CAFE_TYPE_LONG;
			break;

		case 'L':
			kind = CAFE_TYPE_ULONG;
			break;

		case 'q':
			kind = CAFE_TYPE_LONGLONG;
			break;

		case 'Q':
			kind = CAFE_TYPE_ULONGLONG;
			break;

		case 's':
			kind = CAFE_TYPE_STRING;
			break;

		case 'b':
			kind = CAFE_TYPE_BLOB;
			break;

		case '?':
			kind = CAFE_TYPE_BOOL;
			break;

		case 'Y':
			kind = CAFE_TYPE_TYPE;
			break;

		case 'f':
			kind = CAFE_TYPE_FLOAT;
			break;

		case 'd':
			kind = CAFE_TYPE_DOUBLE;
			break;

		default:
			/* coverity[array_null] */
			assert(!"unexpected type");
			type = NULL;
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		rc = get_basic_type_ident(pc->ns, kind, identp);
		if (rc != CAFE_OK)
			return rc;
		break;

	case 'T': {
		struct cafe_value_nature_ref *natures = NULL;

		if (refonly) {
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		if (*fmt++ != '[') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected '[' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		fmt = get_qualified_ident(pc->ns, fmt, &ident);
		if (fmt == NULL) 
			return CAFE_INVALID_VALUE_TYPE_STR;
		if (*fmt++ != ',') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected ',' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}

		*identp = ident;
		rc = build_single_type(pc, &element_ident, &fmt, true);
		if (rc != CAFE_OK)
			return rc;
		if (*fmt == ',') {
			fmt = build_nature_ref_list(pc, &natures, fmt + 1);
			if (fmt == NULL) {
				free_nature_ref(pc->ns, natures);
				return CAFE_INVALID_VALUE_TYPE_STR;
			}
		}
		if (*fmt++ != ']') {
			free_nature_ref(pc->ns, natures);
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected ']' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		if ((type = ident->type) != NULL) {
			if (type->element != element_ident ||
			    !same_nature_ref_list(ident->natures, natures)) {
				free_nature_ref(pc->ns, natures);
				return CAFE_REDECL_VALUE_TYPE;
			}
			free_nature_ref(pc->ns, natures);
			break;
		}
		rc = cafe_array_append(&pc->new_types, &ident);
		if (rc != CAFE_OK) {
			free_nature_ref(pc->ns, natures);
			return rc;
		}
		type = cafe_type_alloc(ident, CAFE_TYPE_TYPEDEF);
		if (type == NULL) {
			free_nature_ref(pc->ns, natures);
			return CAFE_NOMEM;
		}
		type->element = element_ident;
		ident->natures = natures;
		break;
	}

	case 't':
		if (*fmt++ != '[') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected '[' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		fmt = get_qualified_ident(pc->ns, fmt, identp);
		if (fmt != NULL && *fmt++ != ']') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected ']' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		break;

	case 'S':
	case 'N': {
		unsigned int elements = 0;
		struct cafe_vm_attr *attributes = NULL;
		struct cafe_value_nature_ref *natures = NULL;

		if (refonly) {
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		switch (c) {
		case 'S':
			i = CAFE_TYPE_STRUCT;
			break;

		case 'N':
			i = CAFE_TYPE_NATURE_DEF;
			break;

		default:
			/* coverity[array_null] */
			assert(!"unexpected type");
		}
		if (*fmt++ != '[') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected '[' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		fmt = get_qualified_ident(pc->ns, fmt, &ident);
		if (fmt == NULL) {
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		*identp = ident;
		if (*fmt == ',') {
			fmt = build_attribute_list(pc, &attributes, &elements,
						   fmt + 1);
			if (fmt == NULL) {
				return CAFE_INVALID_VALUE_TYPE_STR;
			}
		}
		if (*fmt == ',') {
			fmt = build_nature_ref_list(pc, &natures, fmt + 1);
			if (fmt == NULL) {
				free_attributes(pc->ns, attributes, elements);
				free_nature_ref(pc->ns, natures);
				return CAFE_INVALID_VALUE_TYPE_STR;
			}
		}
		if (*fmt++ != ']') {
			free_attributes(pc->ns, attributes, elements);
			free_nature_ref(pc->ns, natures);
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected ']' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		if ((type = ident->type) != NULL) {
			if (type->elements != elements ||
			    !same_attribute_list(type->attributes, attributes, elements) ||
			    !same_nature_ref_list(ident->natures, natures)) {
				free_attributes(pc->ns, attributes, elements);
				free_nature_ref(pc->ns, natures);
				return CAFE_REDECL_VALUE_TYPE;
			}
			free_attributes(pc->ns, attributes, elements);
			free_nature_ref(pc->ns, natures);
			break;
		}
		rc = cafe_array_append(&pc->new_types, &ident);
		if (rc != CAFE_OK) {
			free_attributes(pc->ns, attributes, elements);
			free_nature_ref(pc->ns, natures);
			return rc;
		}
		type = cafe_type_alloc(ident, i);
		if (type == NULL) {
			free_attributes(pc->ns, attributes, elements);
			free_nature_ref(pc->ns, natures);
			return CAFE_NOMEM;
		}
		type->elements = elements;
		type->attributes = attributes;
		ident->natures = natures;
		break;
	}

	case 'D':
	case 'a': {
		unsigned long long elements;

		if (*fmt++ != '[') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected '[' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		elements = 0;
		element_ident = NULL;
		rc = build_single_type(pc, &element_ident, &fmt, true);
		if (rc != CAFE_OK)
			return rc;
		assert(element_ident != NULL);
		if (c == 'D') {
			kind = CAFE_TYPE_DICT;
		} else if (*fmt == ',') {
			kind = CAFE_TYPE_ARRAY;
			fmt = parseULL(fmt + 1, &elements);
			if (fmt == NULL) {
				return CAFE_INVALID_VALUE_TYPE_STR;
			}
		} else {
			kind = CAFE_TYPE_DYNARRAY;
		}
		if (*fmt++ != ']') {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "expected ']' before '%s'",
				       fmt - 1);
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		*identp = ident = get_composit_ident(pc->ns, kind,
						     (unsigned int) elements,
						     element_ident, NULL);
		if (ident == NULL) {
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		if (ident->type != NULL) {
			break;
		}
		rc = cafe_array_append(&pc->new_types, &ident);
		if (rc != CAFE_OK) {
			return rc;
		}
		type = cafe_type_alloc(ident, kind);
		if (type == NULL) {
			return CAFE_NOMEM;
		}
		type->elements = (unsigned int) elements;
		type->element = element_ident;
		break;
	}

	case 'P':
		fmt = build_instance_ref(pc, identp, fmt);
		if (fmt == NULL)
			return CAFE_INVALID_VALUE_TYPE_STR;
		break;

	case 'B':
		fmt = build_builtin(pc, identp, fmt);
		if (fmt == NULL) {
			cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
				       "Failed parsing builtin type string");
			return CAFE_INVALID_VALUE_TYPE_STR;
		}
		break;

	case 'F':
		fmt = build_filter(pc->ns, identp, fmt);
		if (fmt == NULL)
			return CAFE_INVALID_VALUE_TYPE_STR;
		break;
	default:
		cafe_error_set(NULL, CAFE_INVALID_VALUE_TYPE_STR,
			       "Unknown type identifier '%c'", c);
		return CAFE_INVALID_VALUE_TYPE_STR;
	}
	*fmtp = fmt;
	return CAFE_OK;
}

/*
 * Construct a value.  If the type is complex then recurse to
 * construct each element.
 */
static int construct_value(struct cafe_value_type *type_orig,
			   void *base,
			   const struct cafe_vm_expr *expr)
{
	int rc;
	unsigned int i;
	cafe_any_value v;
	struct cafe_vm_attr *attr;
	struct cafe_value_type *type;

	type = cafe_base_type(type_orig);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_SHORT:
	case CAFE_TYPE_USHORT:
	case CAFE_TYPE_INT:
	case CAFE_TYPE_UINT:
	case CAFE_TYPE_LONG:
	case CAFE_TYPE_ULONG:
	case CAFE_TYPE_LONGLONG:
	case CAFE_TYPE_ULONGLONG:
	case CAFE_TYPE_BOOL:
		if (expr == NULL)
			return CAFE_OK;
		if (!cafe_expr_is_iconst(expr))
			return CAFE_INVALID_INITIALIZER;
		v.ull = cafe_expr_to_ull(expr);
		return cafe_value_store_type(type, base, &v,
					     type->ident->ns->allocator);

	case CAFE_TYPE_FLOAT:
	case CAFE_TYPE_DOUBLE:
		if (expr == NULL)
			return CAFE_OK;
		return CAFE_INVALID_INITIALIZER;

	case CAFE_TYPE_STRUCT:
		if (expr != NULL) {
			if (expr->kind != CAFE_EXPR_LIST)
				return CAFE_INVALID_INITIALIZER;
			expr = expr->u.e;
		}
		cafe_type_attr_foreach(attr, type) {
			struct cafe_value_type *attrtype =
				cafe_attr_type(attr);
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			rc = construct_value(attrtype,
					     ((char *)base) + attr->offset,
					     expr ? expr->u.elem.e :
					     attr->ident->expr);
			if (rc != CAFE_OK)
				return rc;
			if (expr)
				expr = expr->u.elem.next;
		}
		break;

	case CAFE_TYPE_ARRAY: {
		struct cafe_value_type *element_type =
			cafe_element_type(type);

		if (expr != NULL) {
			if (expr->kind != CAFE_EXPR_LIST)
				return CAFE_INVALID_INITIALIZER;
			expr = expr->u.e;
		}
		for (i = 0; i < type->elements; i++) {
			rc = construct_value(element_type, base,
					     expr ? expr->u.elem.e : NULL);
			if (rc != CAFE_OK)
				return rc;
			if (expr)
				expr = expr->u.elem.next;
			base = ((char *)base) + element_type->size;
		}
		break;
	}

	case CAFE_TYPE_DYNARRAY: {
		struct cafe_vm_dynarray *a = base;
		struct cafe_value_type *element_type =
			cafe_element_type(type);
		struct cafe_vm_expr *e;

		if (expr == NULL)
			return CAFE_OK;
		if (expr->kind != CAFE_EXPR_LIST)
			return CAFE_INVALID_INITIALIZER;
		for (i = 0, e = expr->u.e; e != NULL; e = e->u.elem.next)
			i++;
		if (i == 0)
			return CAFE_OK;
		rc = cafe_vm_dynarray_set_size(a, i, type);
		if (rc != CAFE_OK)
			return rc;

		base = a->p;
		for (e = expr->u.e; e != NULL; e = e->u.elem.next) {
			rc = construct_value(element_type, base, e->u.elem.e);
			if (rc != CAFE_OK)
				return rc;
			base = ((char *)base) + element_type->size;
		}
		break;
	}

	case CAFE_TYPE_STRING:
		if (expr == NULL)
			return CAFE_OK;
		if (expr->kind != CAFE_EXPR_STRING)
			return CAFE_INVALID_INITIALIZER;
		v.str = expr->u.str.p;
		return cafe_value_store_type(type, base, &v,
					     type->ident->ns->allocator);

	case CAFE_TYPE_BLOB:
		if (expr == NULL)
			return CAFE_OK;
		if (expr->kind != CAFE_EXPR_STRING)
			return CAFE_INVALID_INITIALIZER;
		v.blob.l = expr->u.str.l;
		v.blob.p = expr->u.str.p;
		return cafe_value_store_type(type, base, &v,
					     type->ident->ns->allocator);

	case CAFE_TYPE_DICT: {
		struct cafe_value_type *element_type =
			cafe_element_type(type);
		struct cafe_vm_expr *e;

		cafe_dict_create((struct cafe_dict_struct *)base, type_orig);
		if (expr == NULL)
			break;
		if (expr->kind != CAFE_EXPR_LIST)
			return CAFE_INVALID_INITIALIZER;
		for (e = expr->u.e; e != NULL; e = e->u.elem.next) {
			struct cafe_value_data *data;

			rc = cafe_value_create2(element_type, e->u.elem.e,
						&data);
			if (rc != CAFE_OK)
				return rc;
			if (cafe_dict_replace((struct cafe_dict_struct *)base,
					      data) == NULL)
				return CAFE_NOMEM;
		}
		break;
	}

	default:
		cafe_error_set(NULL, CAFE_ERROR, "%s: unexpected type %c\n",
			       __FUNCTION__, type->kind);
		return CAFE_ERROR;
	}
	return CAFE_OK;
}

/*
 * Destruct a value.  If the type is complex then recurse to destruct
 * each element.
 */
static void destruct_value(const struct cafe_value_type *type, void *base)
{
	unsigned int i;
	void *p;
	struct cafe_vm_attr *attr;

	type = cafe_base_type(type);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_STRUCT:
		cafe_type_attr_foreach(attr, type) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			destruct_value(cafe_attr_type(attr),
				       ((char *)base) + attr->offset);
		}
		break;

	case CAFE_TYPE_ARRAY:
		for (i = 0; i < type->elements; i++) {
			destruct_value(cafe_element_type(type), base);
			base = ((char *)base) + cafe_element_type(type)->size;
		}
		break;

	case CAFE_TYPE_DYNARRAY:
		(void)cafe_vm_dynarray_set_size(base, 0, type);
		break;

	case CAFE_TYPE_STRING:
		p = *(char **)base;
		if (p) {
			CAFE_FREE(type->ident->ns->allocator, p);
		}
		break;

	case CAFE_TYPE_BLOB:
		p = ((struct cafe_vm_blob *)base)->p;
		if (p) {
			CAFE_FREE(type->ident->ns->allocator, p);
		}
		break;

	case CAFE_TYPE_DICT:
		cafe_dict_delete((struct cafe_dict_struct *)base);
		break;
	}
}

/*
 * Clear a value.  If the type is complex then recurse to destruct
 * each element.
 */
static void clearValue(const struct cafe_value_type *type, void *base)
{
	int rc;
	unsigned int i;
	void *p;
	cafe_any_value v;
	struct cafe_vm_attr *attr;

	type = cafe_base_type(type);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_SHORT:
	case CAFE_TYPE_USHORT:
	case CAFE_TYPE_INT:
	case CAFE_TYPE_UINT:
	case CAFE_TYPE_LONG:
	case CAFE_TYPE_ULONG:
	case CAFE_TYPE_LONGLONG:
	case CAFE_TYPE_ULONGLONG:
	case CAFE_TYPE_BOOL:
		v.ull = 0;
		rc = cafe_value_store_type(type, base, &v,
					   type->ident->ns->allocator);
		assert(rc == CAFE_OK);
		if (rc != CAFE_OK)
			TRACE(("Error clearing bool type (%s)\n",
			       cafe_strerror(rc)));
		break;

	case CAFE_TYPE_FLOAT:
	case CAFE_TYPE_DOUBLE:
		v.d = 0;
		rc = cafe_value_store_type(type, base, &v,
					   type->ident->ns->allocator);
		assert(rc == CAFE_OK);
		if (rc != CAFE_OK)
			TRACE(("Error clearing double type (%s)\n",
			       cafe_strerror(rc)));
		break;

	case CAFE_TYPE_STRUCT:
		cafe_type_attr_foreach(attr, type) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			clearValue(cafe_attr_type(attr),
				   ((char *)base) + attr->offset);
		}
		break;

	case CAFE_TYPE_ARRAY:
		for (i = 0; i < type->elements; i++) {
			clearValue(cafe_element_type(type), base);
			base = ((char *)base) + cafe_element_type(type)->size;
		}
		break;

	case CAFE_TYPE_DYNARRAY:
		(void)cafe_vm_dynarray_set_size(base, 0, type);
		break;

	case CAFE_TYPE_STRING:
		p = *(char **)base;
		if (p) {
			CAFE_FREE(type->ident->ns->allocator, p);
			*(char **)base = NULL;
		}
		break;

	case CAFE_TYPE_BLOB:
		p = ((struct cafe_vm_blob *)base)->p;
		if (p) {
			CAFE_FREE(type->ident->ns->allocator, p);
			((struct cafe_vm_blob *)base)->p = NULL;
			((struct cafe_vm_blob *)base)->l = 0;
		}
		break;

	case CAFE_TYPE_DICT:
		cafe_dict_reset((struct cafe_dict_struct *)base);
		break;

	default:
		/* coverity[array_null] */
		assert(!"unexpected type");
	}
}

/*
 * valueTypeBuild - Build type data structure from type description string
 */
int cafe_type_parse2(struct cafe_namespace *ns, const char **fmtp,
		     const char *endchars, struct cafe_value_type **typep)
{
	const char *fmt = *fmtp;
	struct cafe_vm_ident *ident = NULL;
	const char *type_str;
	int rc;
	struct parse_context pc;

	assert(ns != NULL);
	if (typep != NULL) {
		*typep = NULL;
	}
	if (fmt == NULL) {
		return CAFE_INVALID_VALUE_TYPE_STR;
	}

	parse_context_init(&pc, ns);
	while (*fmt != '\0' &&
	       (endchars == NULL || strchr(endchars, *fmt) == NULL)) {
		ident = NULL;
		type_str = fmt;
		rc = build_single_type(&pc, &ident, &fmt, false);
		if (rc != CAFE_OK)
			goto error;
		assert(fmt != NULL);
		if (ident && ident->type && ident->type->type_str == NULL) {
			ident->type->type_str = CAFE_ALLOC(ns->allocator,
							   fmt - type_str + 1);
			if (ident->type->type_str == NULL) {
				rc = CAFE_NOMEM;
				goto error;
			}
			memcpy(ident->type->type_str, type_str, fmt - type_str);
			ident->type->type_str[fmt - type_str] = '\0';
		}
	}
	rc = resolve_types(ns, pc.new_types.data, pc.new_types.len);
	if (rc != CAFE_OK)
		goto error;
	cafe_array_release(&pc.new_types);
	*fmtp = fmt;
	if (ident == NULL) {
		return CAFE_OK;
	}
	if (ident->type == NULL)
		return CAFE_UNRESOLVED_VALUE_TYPE;
	//assert(ident->type->align > 0);
	if (typep != NULL) {
		*typep = ident->type;
	}
	return CAFE_OK;
error:
	cafe_array_release(&pc.new_types);
	return rc;
}

int cafe_type_ref_parse(
	struct cafe_namespace *ns,
	const char **fmtp,
	struct cafe_vm_ident **identp)
{
	struct parse_context pc;
	int rc;

	assert(ns != NULL);
	assert(*fmtp != NULL);
	parse_context_init(&pc, ns);
	rc = build_single_type(&pc, identp, fmtp, true);
	assert(pc.new_types.len == 0);
	cafe_array_release(&pc.new_types);
	return rc;
}

int cafe_builtin_type(struct cafe_namespace *ns,
		      cafe_value_type_tag kind,
		      struct cafe_value_type **result)
{
	struct cafe_vm_ident *ident;
	int rc;

	rc = get_basic_type_ident(ns, kind, &ident);
	if (rc != CAFE_OK)
		return rc;
	*result = ident->type;
	return CAFE_OK;
}

bool cafe_type_based_on(const struct cafe_value_type *type,
			const struct cafe_value_type *base)
{
	if (type == NULL || base == NULL)
		return false;
	while (base->kind == CAFE_TYPE_TYPEDEF)
		base = cafe_element_type(base);
	while (type != base &&
	       (type->kind == CAFE_TYPE_TYPEDEF ||
		type->kind == CAFE_TYPE_INSTANCE))
		type = cafe_element_type(type);
	if (type == base)
		return true;
	if (type->kind != base->kind)
		return false;
	if (type->kind == CAFE_TYPE_DICT ||
	    (type->kind == CAFE_TYPE_ARRAY &&
	     type->elements == base->elements) ||
	    type->kind == CAFE_TYPE_DYNARRAY) {
		return cafe_type_based_on(
			cafe_element_type(type),
			cafe_element_type(base));
	}
	return false;
}

static int cafe_ident_name_str(struct cafe_byte_buffer *db,
			       const struct cafe_vm_ident *ident)
{
	int rc;

	assert(ident);
	assert(db);
	if (ident->parent) {
		(void)cafe_ident_name_str(db, ident->parent);
	}
	rc = cafe_bbuf_capacity_set(db, db->len + ident->len + 1);
	if (rc != CAFE_OK)
		return rc;
	if (ident->parent)
		db->data[db->len++] = '.';
	memcpy(db->data + db->len, ident->name, ident->len);
	db->len += ident->len;
	return CAFE_OK;
}

static int bbuf_type_ref_append(struct cafe_byte_buffer *b,
				const struct cafe_vm_ident *ident)
{
	int rc;

	if ((strlen(ident->name) == 3)
	    && (ident->name[0] == '<')
	    && (ident->name[2] == '>'))
	{
		rc = cafe_bbuf_write(b, ident->name[1]);
		return rc;
	}

	if (ident->type->kind == CAFE_TYPE_FILTER) {
		char *start, *end;
		rc = cafe_bbuf_append(b, "F[", 2);
		if (rc != CAFE_OK)
			return rc;

		start = strchr(ident->name, '<') + 1;
		end = strrchr(ident->name, '>') ;
		rc = cafe_bbuf_append(b, start, end - start);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_bbuf_write(b, ']');
		if (rc != CAFE_OK)
			return rc;

		return CAFE_OK;
	}
	rc = cafe_bbuf_append(b, "t[", 2);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_ident_name_str(b, ident);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_bbuf_write(b, ']');
	if (rc != CAFE_OK)
		return rc;

	return CAFE_OK;
}

static int bbuf_ull_append(struct cafe_byte_buffer *b, unsigned long long value)
{
	int rc = 0;
	int digit = (int)(value % 10);
	unsigned long long remainder = value / 10;

	if (remainder > 0)
		rc = bbuf_ull_append(b, remainder);

	if (rc == 0)
		rc = cafe_bbuf_write(b, '0' + digit);
	return rc;
}

static int bbuf_nature_ref_append(struct cafe_byte_buffer *b,
				  const struct cafe_value_nature_ref *nature)
{
	int rc;

	rc = cafe_bbuf_append(b, "n[", 2);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_ident_name_str(b, nature->def);
	if (rc != CAFE_OK)
		return rc;

	if (nature->args != NULL) {
		const struct cafe_vm_expr *trav;

		for (trav = nature->args->u.e; trav != NULL; trav = trav->u.elem.next) {
			const struct cafe_vm_expr *e;
			if (trav->kind != CAFE_EXPR_ELEMENT) {
				(void)fprintf(stderr, "%s: Non-element found in list (%u)\n",
				              __FUNCTION__, trav->kind);
				continue;
			}

			rc = cafe_bbuf_write(b, ',');
			if (rc != CAFE_OK)
				return rc;

			e = trav->u.elem.e;
			if (e->kind == CAFE_EXPR_STRING) {
				rc = cafe_bbuf_write(b, '"');
				if (rc != CAFE_OK)
					return rc;

				rc = cafe_bbuf_append(b,
						      e->u.str.p, e->u.str.l);
				if (rc != CAFE_OK)
					return rc;

				rc = cafe_bbuf_write(b, '"');
				if (rc != CAFE_OK)
					return rc;
			} else if (e->kind == CAFE_EXPR_UVALUE) {
				rc = bbuf_ull_append(b, e->u.ull);
				if (rc != CAFE_OK)
					return rc;
			} else if (e->kind == CAFE_EXPR_SVALUE) {
				long long value = e->u.ll;
				if (value < 0) {
					rc = cafe_bbuf_write(b, '-');
					if (rc != CAFE_OK)
						return rc;
					value = -value;
				}
				rc = bbuf_ull_append(b, (unsigned long long) value);
				if (rc != CAFE_OK)
					return rc;
			} else {
				(void)fprintf(stderr, "%s: unhandled nature expression type: %d\n", __FUNCTION__, e->kind);
			}
		}
	}

	rc = cafe_bbuf_write(b, ']');
	if (rc != CAFE_OK)
		return rc;

	return CAFE_OK;
}

static int bbuf_type_append(struct cafe_byte_buffer *b,
			    const struct cafe_value_type *type)
{
	int rc;
	switch(type->kind) {
	case CAFE_TYPE_DYNARRAY:
		rc = cafe_bbuf_append(b, "a[", 2);
		if (rc != CAFE_OK)
			return rc;
		rc = bbuf_type_ref_append(b, cafe_element_type(type)->ident);
		if (rc != CAFE_OK)
			return rc;
		rc = cafe_bbuf_write(b, ']');
		if (rc != CAFE_OK)
			return rc;
		break;
	default:
		rc = bbuf_type_ref_append(b, type->ident);
		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}

static int bbuf_nature_list_append(struct cafe_byte_buffer *b,
				   const struct cafe_value_nature_ref *natures)
{
	const struct cafe_value_nature_ref *trav;
	int rc;

	if (natures == NULL)
		return CAFE_OK;

	rc = cafe_bbuf_write(b, ',');
	if (rc != CAFE_OK)
		return rc;

	for (trav = natures; trav != NULL; trav = trav->next) {
		rc = bbuf_nature_ref_append(b, trav);
		if (rc != CAFE_OK)
			return rc;
	}
	return CAFE_OK;
}

static int bbuf_element_append(struct cafe_byte_buffer *b,
			       const struct cafe_vm_attr *attr)
{
	int rc;

	switch(attr->kind) {
	case CAFE_ATTR_VAR:
		rc = cafe_bbuf_append(b, "v[", 2);
		if (rc != CAFE_OK)
			return rc;
		break;

	case CAFE_ATTR_KEY:
		rc = cafe_bbuf_append(b, "k[", 2);
		if (rc != CAFE_OK)
			return rc;
		break;

	case CAFE_ATTR_CONST:
	case CAFE_ATTR_PARAM:
	default:
		cafe_error_set(NULL, CAFE_ERROR,
			       "%s: Unexpected element type: %u\n",
			       __FUNCTION__, attr->kind);
		return CAFE_ERROR;
	}

	rc = cafe_ident_name_str(b, attr->ident);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_bbuf_write(b, ',');
	if (rc != CAFE_OK)
		return rc;

	rc = bbuf_type_append(b, cafe_attr_type(attr));
	if (rc != CAFE_OK)
		return rc;

	rc = bbuf_nature_list_append(b, attr->ident->natures);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_bbuf_write(b, ']');
	if (rc != CAFE_OK)
		return rc;

	return CAFE_OK;
}

char *cafe_type_ref_str(const struct cafe_value_type *type)
{
	struct cafe_byte_buffer dynstr;
	struct cafe_byte_buffer *db = &dynstr;
	int rc;

	assert(type->kind == CAFE_TYPE_STRUCT
	       || type->kind == CAFE_TYPE_TYPEDEF
	       || type->kind == CAFE_TYPE_FILTER);
	rc = cafe_bbuf_init(db, type->ident->ns->allocator, 4);
	if (rc != CAFE_OK)
		return NULL;

	rc = bbuf_type_ref_append(db, type->ident);
	if (rc != CAFE_OK) {
		cafe_bbuf_release(db);
		return NULL;
	}

	rc = cafe_bbuf_write(db, 0);
	if (rc != CAFE_OK) {
		cafe_bbuf_release(db);
		return NULL;
	}

	return db->data;
}

char *cafe_type_name_str(const struct cafe_value_type *type)
{
	struct cafe_byte_buffer db;
	int rc;

	rc = cafe_bbuf_init(&db, type->ident->ns->allocator, 0);
	if (rc != CAFE_OK) {
		return NULL;
	}
	rc = cafe_ident_name_str(&db, type->ident);
	if (rc != CAFE_OK) {
		cafe_bbuf_release(&db);
		return NULL;
	}
	rc = cafe_bbuf_write(&db, '\0');
	if (rc != CAFE_OK) {
		cafe_bbuf_release(&db);
		return NULL;
	}

	return db.data;
}

void cafe_type_iter_dep_types(struct cafe_value_type *type,
			      bool (*pre_cb)(struct cafe_value_type *, void *),
			      void (*post_cb)(struct cafe_value_type *, void *),
			      void *arg)
{
	struct cafe_vm_attr *attr;
	struct cafe_value_nature_ref *nature;

	if (type == NULL) {
		return;
	}
	if (pre_cb(type, arg)) {
		return;
	}
	switch (type->kind) {
	case CAFE_TYPE_DICT:
	case CAFE_TYPE_ARRAY:
	case CAFE_TYPE_DYNARRAY:
	case CAFE_TYPE_TYPEDEF:
	case CAFE_TYPE_INSTANCE:
		cafe_type_iter_dep_types(cafe_element_type(type), pre_cb,
					 post_cb, arg);
		break;

	case CAFE_TYPE_STRUCT:
	case CAFE_TYPE_NATURE_DEF:
		cafe_type_attr_foreach(attr, type) {
			cafe_type_iter_dep_types(cafe_attr_type(attr), pre_cb,
						 post_cb, arg);
			for (nature = attr->ident->natures; nature != NULL;
			     nature = nature->next)
			{
				cafe_type_iter_dep_types(nature->def->type,
							 pre_cb, post_cb, arg);
			}
		}
		break;
	}
	for (nature = type->ident->natures; nature != NULL;
	     nature = nature->next)
	{
		cafe_type_iter_dep_types(nature->def->type, pre_cb, post_cb,
					 arg);
	}
	post_cb(type, arg);
}

int cafe_namespace_create(
	const struct cafe_allocator *allocator,
	struct cafe_namespace **nsp)
{
	struct cafe_namespace *ns;

	*nsp = ns = CAFE_ALLOC(allocator, sizeof *ns);
	if (ns == NULL) {
		return CAFE_NOMEM;
	}
	memset(ns, 0, sizeof *ns);
	ns->allocator = allocator;
	return CAFE_OK;
}

void cafe_namespace_delete(struct cafe_namespace *ns)
{
	free_ident(ns->children);
	CAFE_FREE(ns->allocator, ns);
}

/*
 * cafe_value_create - Create object based on value type
 *
 * <type>: pointer to type of object
 */
int cafe_value_create2(struct cafe_value_type *type,
		       const struct cafe_vm_expr *expr,
		       struct cafe_value_data **data)
{
	struct cafe_value_data *p;

	assert(type != NULL);
	*data = p = CAFE_ALLOC(type->ident->ns->allocator,
			       offsetof(struct cafe_value_data, u) + type->size);
	if (p == NULL) {
		return CAFE_NOMEM;
	}
	memset(p, 0, offsetof(struct cafe_value_data, u)+type->size);
	p->type = type;
	return construct_value(type, p->u.data, expr);
}

/*
 * cafe_value_delete - Delete object
 *
 * <p>: pointer to object
 */
void cafe_value_delete(struct cafe_value_data *p)
{
	struct cafe_value_type *type = p->type;

	p->type = NULL;
	destruct_value(type, p->u.data);
	CAFE_FREE(type->ident->ns->allocator, p);
}


/*
 * cafe_value_clear - Clear object
 *
 * <p>: pointer to object
 */
void cafe_value_clear(struct cafe_value_data *p)
{
	clearValue(p->type, p->u.data);
}

/*
 * cafe_value_load_type - Load type from memory to value
 *
 * <type>: type to load
 *
 * <base>: pointer to data in memory
 *
 * <value>: value to load into
 */
void cafe_value_load_type(const struct cafe_value_type *type,
			  const void *base,
			  cafe_any_value *value)
{
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
		value->ll = *(char *)base;
		break;

	case CAFE_TYPE_UCHAR:
		value->ull = *(unsigned char *)base;
		break;

	case CAFE_TYPE_SHORT:
		value->ll = *(short *)base;
		break;

	case CAFE_TYPE_USHORT:
		value->ull = *(unsigned short *)base;
		break;

	case CAFE_TYPE_INT:
		value->ll = *(int *)base;
		break;

	case CAFE_TYPE_UINT:
		value->ull = *(unsigned int *)base;
		break;

	case CAFE_TYPE_LONG:
		value->ll = *(long *)base;
		break;

	case CAFE_TYPE_ULONG:
		value->ull = *(unsigned long *)base;
		break;

	case CAFE_TYPE_LONGLONG:
		value->ll = *(long long *)base;
		break;

	case CAFE_TYPE_ULONGLONG:
		value->ull = *(unsigned long long *)base;
		break;

	case CAFE_TYPE_BOOL:
		value->ll = *(bool *)base;
		break;

	case CAFE_TYPE_FLOAT:
		value->d = *(float *)base;
		break;

	case CAFE_TYPE_DOUBLE:
		value->d = *(double *)base;
		break;

	case CAFE_TYPE_STRUCT:
	case CAFE_TYPE_DICT:
		value->p = (void *)base;
		break;

	case CAFE_TYPE_ARRAY:
		value->ary.l = type->elements;
		value->ary.p = (void *)base;
		break;

	case CAFE_TYPE_DYNARRAY:
		value->dary.l = ((struct cafe_vm_dynarray *)base)->l;
		value->dary.p = ((struct cafe_vm_dynarray *)base)->p;
		break;

	case CAFE_TYPE_STRING:
		value->str = *(char **)base;
		break;

	case CAFE_TYPE_BLOB:
		value->blob = ((cafe_any_value *) base)->blob;
		break;

	default:
		/* coverity[array_null] */
		assert(!"unexpected cafe_value_type in cafe_value_load_type()");
		break;
	}
}

/*
 * cafe_value_store_string - store string to memory
 *
 * <dest>: pointer to memory to store value
 *
 * <src>: value to store
 *
 * <allocator>: memory allocator
 */
int cafe_value_store_string(char **dest,
			    const char *src,
			    const struct cafe_allocator *allocator)
{
	size_t l = strlen(src);
	void *p = CAFE_ALLOC(allocator, l + 1);
	if (p == NULL)
		return CAFE_NOMEM;
	memcpy(p, src, l + 1);
	if (*dest) {
		CAFE_FREE(allocator, *dest);
	}
	*dest = p;
	return CAFE_OK;
}

/*
 * cafe_value_store_blob - store blob to memory
 *
 * <dest>: pointer to destination
 *
 * <src>: value to store
 *
 * <allocator>: memory allocator
 */
int cafe_value_store_blob(struct cafe_vm_blob *dest,
			  const struct cafe_vm_blob *src,
			  const struct cafe_allocator *allocator)
{
	size_t l = src->l;
	void *p = CAFE_ALLOC(allocator, l);
	if (p == NULL)
		return CAFE_NOMEM;
	CAFE_FREE(allocator, dest->p);
	dest->l = l;
	dest->p = p;
	memcpy(p, src->p, l);
	return CAFE_OK;
}

/*
 * cafe_value_store_type - store type from value to memory
 *
 * <type>: type to load
 *
 * <base>: pointer to memory to store value
 *
 * <value>: value to store
 *
 * <allocator>: memory allocator
 */
int cafe_value_store_type(const struct cafe_value_type *type,
			  void *base,
			  const cafe_any_value *value,
			  const struct cafe_allocator *allocator)
{
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
		*(char *)base = (char)value->ll;
		break;

	case CAFE_TYPE_UCHAR:
		*(unsigned char *)base = (unsigned char)value->ull;
		break;

	case CAFE_TYPE_SHORT:
		*(short *)base = (short)value->ll;
		break;

	case CAFE_TYPE_USHORT:
		*(unsigned short *)base = (unsigned short)value->ull;
		break;

	case CAFE_TYPE_INT:
		*(int *)base = (int)value->ll;
		break;

	case CAFE_TYPE_UINT:
		*(unsigned int *)base = (unsigned int)value->ull;
		break;

	case CAFE_TYPE_LONG:
		*(long *)base = (long)value->ll;
		break;

	case CAFE_TYPE_ULONG:
		*(unsigned long *)base = (unsigned long)value->ull;
		break;

	case CAFE_TYPE_LONGLONG:
		*(long long *)base = (long long)value->ll;
		break;

	case CAFE_TYPE_ULONGLONG:
		*(unsigned long long *)base = (unsigned long long)value->ull;
		break;

	case CAFE_TYPE_BOOL:
		*(bool *)base = (bool)value->ll;
		break;

	case CAFE_TYPE_FLOAT:
		*(float *)base = (float) value->d;
		break;

	case CAFE_TYPE_DOUBLE:
		*(double *)base = (double) value->d;
		break;

	case CAFE_TYPE_STRUCT: {
		struct cafe_vm_attr *attr;
		void *src_base = value->p;

		assert(src_base);
		cafe_type_attr_foreach(attr, type) {
			void *dest;
			void *src;
			struct cafe_value_type *attrtype;
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;

			attrtype = cafe_attr_type(attr);
			dest = ((char *)base) + attr->offset;
			src = ((char *)src_base) + attr->offset;
			clearValue(attrtype, dest);
			(void)cafe_value_copy(attrtype, dest, src, allocator);
		}
		break;
	}

	case CAFE_TYPE_ARRAY: {
		size_t i;
		size_t l = type->elements < value->ary.l ? type->elements : value->ary.l;
		void *dest;
		void *src;
		struct cafe_value_type *element_type;

		dest = base;
		src = value->ary.p;
		element_type = cafe_element_type(type);
		for (i = 0; i < l; i++) {
			clearValue(element_type, dest);
			(void)cafe_value_copy(element_type, dest, src, allocator);
			dest = ((char *)dest) + element_type->size;
			src = ((char *)src) + element_type->size;
		}
		l = type->elements;
		for (; i < l; i++) {
			clearValue(element_type, dest);
			dest = ((char *)dest) + element_type->size;
		}
		break;
	}

	case CAFE_TYPE_DYNARRAY: {
		int rc;
		size_t i;
		size_t l = value->dary.l;
		void *dest;
		void *src;
		struct cafe_value_type *element_type;

		rc = cafe_vm_dynarray_set_size(base, l, type);
		if (rc != CAFE_OK)
			return rc;

		dest = ((struct cafe_vm_dynarray *)base)->p;
		src = value->dary.p;
		element_type = cafe_element_type(type);
		for (i = 0; i < l; i++) {
			clearValue(element_type, dest);
			(void)cafe_value_copy(element_type, dest, src, allocator);
			dest = ((char *)dest) + element_type->size;
			src = ((char *)src) + element_type->size;
		}
		break;
	}

	case CAFE_TYPE_STRING:{
		size_t l = strlen(value->str);
		void *p = CAFE_ALLOC(allocator, l + 1);
		if (p == NULL)
			return CAFE_NOMEM;
		clearValue(type, base);
		*(char **)base = p;
		memcpy(p, value->str, l + 1);
		break;
	}

	case CAFE_TYPE_BLOB:{
		size_t l = value->blob.l;
		void *p = CAFE_ALLOC(allocator, l);
		if (p == NULL)
			return CAFE_NOMEM;
		clearValue(type, base);
		((cafe_any_value *) base)->blob.l = l;
		((cafe_any_value *) base)->blob.p = p;
		memcpy(p, value->blob.p, l);
		break;
	}

	case CAFE_TYPE_TYPEDEF:
		return cafe_value_store_type(cafe_element_type(type), base,
					     value, allocator);

	default:
		/* coverity[array_null] */
		assert(!"unexpected cafe_value_type in cafe_value_store_type()");
		break;
	}
	return CAFE_OK;
}

/*
 * cafe_value_store - Store a value into named field
 *
 * <type>: pointer to type of data
 *
 * <base>: pointer to data
 *
 * <name>: pointer to name of field
 *
 * <value>: pointer to value to store
 *
 * For now this function only supports storing into a a direct field
 * in a struct.  In the future we can enhance this to allow name to be
 * a '.' separated list of fields perhaps also have a syntax for array
 * and dict looks up, if needed.
 */
int cafe_value_store(const struct cafe_value_type *type, void *base,
		     const char *name, const cafe_any_value *value,
		     const struct cafe_allocator *allocator)
{
	struct cafe_vm_attr *attr;

	if (type->kind != CAFE_TYPE_STRUCT)
		return CAFE_ERROR;
	attr = cafe_type_attr_by_name(type, name);
	if (attr == NULL ||
	    (attr->kind != CAFE_ATTR_VAR && attr->kind != CAFE_ATTR_KEY))
		return CAFE_ERROR;
	return cafe_value_store_type(cafe_attr_type(attr),
				     (char *)base + attr->offset,
				     value,
				     allocator);
}


/*
 * cafe_find_nature - find nature associated with identifier
 *
 * <ident>: identifier to search
 *
 * <nature_name>: name of nature to look for
 *
 * returns pointer to nature reference if found or NULL
 */
static const struct cafe_value_nature_ref *cafe_find_nature_type(
	const struct cafe_vm_ident *ident,
	const struct cafe_value_type *nature)
{
	struct cafe_value_nature_ref *natures = ident->natures;

	while (natures != 0) {
		if (cafe_type_based_on(natures->def->type, nature))
			return natures;
		natures = natures->next;
	}
	switch (ident->kind) {
	case CAFE_IDENT_TYPE:
		if (ident->type->kind == CAFE_TYPE_DICT ||
		    ident->type->kind == CAFE_TYPE_ARRAY ||
		    ident->type->kind == CAFE_TYPE_DYNARRAY ||
		    ident->type->kind == CAFE_TYPE_TYPEDEF ||
		    ident->type->kind == CAFE_TYPE_INSTANCE) {
			return cafe_find_nature_type(
				cafe_element_type(ident->type)->ident, nature);
		}
		break;

	case CAFE_IDENT_ATTR:
		return cafe_find_nature_type(ident->children, nature);
	}
	return NULL;
}

const struct cafe_value_nature_ref *cafe_find_nature_str(
	const struct cafe_vm_ident *ident,
	const char *nature_name)
{
	struct cafe_value_type *nature;

	assert(ident != NULL);
	if (cafe_type_parse(ident->ns, nature_name, &nature) != CAFE_OK)
		return NULL;
	return cafe_find_nature_type(ident, nature);
}

/*
 * cafe_dynarray_set_size - set size of dynamic array
 *
 * Return CAFE_OK or CAFE_NOMEM if the size could not be accommodated.
 */
int cafe_vm_dynarray_set_size(struct cafe_vm_dynarray *a, size_t new_l,
			      const struct cafe_value_type *type)
{
	const size_t old_l = a->l;
	const size_t elemsize = cafe_element_type(type)->size;

	assert(type->kind == CAFE_TYPE_DYNARRAY);
	if (old_l < new_l) {
		/* Grow */
		if (a->c < new_l) {
			size_t new_c = a->c * 2 < new_l ? new_l : a->c * 2;
			void *new_p = CAFE_REALLOC(type->ident->ns->allocator,
						   a->p, new_c * elemsize);
			if (new_p == NULL)
				return CAFE_NOMEM;
			a->c = new_c;
			a->p = new_p;
		}
		memset((char *)a->p + old_l * elemsize, 0,
		       (new_l - old_l) * elemsize);
	} else {
		void *base = (char *)a->p + new_l * elemsize;
		size_t i;
		for (i = new_l; i < old_l; i++) {
			clearValue(cafe_element_type(type), base);
			base = ((char *)base) + elemsize;
		}
		if (new_l == 0) {
			CAFE_FREE(type->ident->ns->allocator, a->p);
			a->p = NULL;
			a->c = 0;
		} else if (new_l * 2 < a->c) {
			a->p = CAFE_REALLOC(type->ident->ns->allocator,
					    a->p, new_l * elemsize);
			assert(a->p != NULL);
			a->c = new_l;
		}
	}
	a->l = new_l;
	return CAFE_OK;
}

int cafe_vm_dynarray_elem_get(struct cafe_vm_dynarray *array,
			      const struct cafe_value_type *type,
			      size_t index,
			      void **result)
{
	struct cafe_value_type *element_type;

	assert(array);
	assert(type);
	assert(result);
	assert(type->kind == CAFE_TYPE_DYNARRAY);
	*result = NULL;
	element_type = cafe_element_type(type);
	assert(element_type);

	if (index + 1 > array->l) {
		size_t i;
		const size_t old_l = array->l;
		void *base;
		int rc;

		rc = cafe_vm_dynarray_set_size(array, index + 1, type);

		if (rc != CAFE_OK)
			return rc;

		base = ((char *)array->p) + old_l * element_type->size;

		for (i = old_l; i < array->l; i++) {
			rc = construct_value(element_type, base, NULL);
			if (rc != CAFE_OK)
				return rc;
			base = ((char *) base) + element_type->size;
		}
	}

	*result = &((char *)array->p)[index * element_type->size];

	return CAFE_OK;
}

struct cafe_vm_attr *cafe_type_first_attr(
	const struct cafe_value_type *type,
	struct cafe_type_attr_iter *iter)
{
	struct cafe_vm_attr *attr;

	type = cafe_base_type(type);
	assert(type != NULL);
	assert(type->kind == CAFE_TYPE_NATURE_DEF ||
	       type->kind == CAFE_TYPE_STRUCT);
	iter->attrcur = attr = type->attributes;
	iter->attrend = attr + type->elements;
	assert(attr == NULL ||
	       attr->kind == CAFE_ATTR_VAR ||
	       attr->kind == CAFE_ATTR_KEY ||
	       attr->kind == CAFE_ATTR_CONST ||
	       attr->kind == CAFE_ATTR_PARAM);
	return attr;
}


struct cafe_vm_attr *cafe_type_next_attr(
	struct cafe_type_attr_iter *iter)
{
	struct cafe_vm_attr *attr = iter->attrcur + 1;

	if (attr >= iter->attrend)
		return NULL;
	iter->attrcur = attr;
	assert(attr->kind == CAFE_ATTR_VAR ||
	       attr->kind == CAFE_ATTR_KEY ||
	       attr->kind == CAFE_ATTR_CONST ||
	       attr->kind == CAFE_ATTR_PARAM);
	return attr;
}


struct cafe_vm_attr *cafe_type_attr_by_index(
	const struct cafe_value_type *type,
	size_t idx)
{
	type = cafe_base_type(type);
	assert(type != NULL);
	assert(type->kind == CAFE_TYPE_NATURE_DEF ||
	       type->kind == CAFE_TYPE_STRUCT);
	return (idx < type->elements) ? type->attributes + idx : NULL;
}


struct cafe_vm_attr *cafe_type_attr_by_name(
	const struct cafe_value_type *type,
	const char *name)
{
	struct cafe_vm_attr *attr;

	type = cafe_base_type(type);
	assert(type != NULL);
	assert(type->kind == CAFE_TYPE_NATURE_DEF ||
	       type->kind == CAFE_TYPE_STRUCT);
	cafe_type_attr_foreach(attr, type) {
		if (strcmp(attr->ident->name, name) == 0) {
			return attr;
		}
	}
	return NULL;
}


struct cafe_value_context *cafe_value_context_create(
	const struct cafe_allocator *allocator)
{
	struct cafe_value_context *newctx;

	newctx = CAFE_ALLOC(allocator, sizeof *newctx);
	if (newctx == NULL)
		return NULL;
	memset(newctx, 0, sizeof *newctx);
	newctx->allocator = allocator;
	return newctx;
}

struct cafe_value_context *cafe_value_context_add(
	struct cafe_value_context *curctx)
{
	struct cafe_value_context *newctx;

	assert(curctx);
	newctx = CAFE_ALLOC(curctx->allocator, sizeof *newctx);
	if (newctx == NULL)
		return NULL;
	*newctx = *curctx;
	newctx->parent = curctx;
	return newctx;
}

struct cafe_value_context *cafe_value_context_delete(
	struct cafe_value_context *curctx)
{
	struct cafe_value_context *parent = curctx->parent;

	CAFE_FREE(curctx->allocator, curctx);
	return parent;
}


struct cafe_vm_expr *cafe_param_value_get(
	const struct cafe_value_type *type,
	const char *param_name)
{
	struct cafe_vm_attr *attr = cafe_type_attr_by_name(type, param_name);
	if (attr == NULL)
		return NULL;

	switch (type->kind) {
	case CAFE_TYPE_INSTANCE:
		return type->ident->expr->u.e->u.elem.e;
	case CAFE_TYPE_STRUCT:
		return attr->ident->expr;
	}

	return NULL;
}


static
int nature_ref_list_copy(const struct cafe_value_nature_ref *input,
			 struct cafe_value_nature_ref **result)
{
	const struct cafe_value_nature_ref *trav;
	struct cafe_vm_expr *args;

	if (input == NULL) {
		*result = NULL;
		return CAFE_OK;
	}

	for (trav = input; trav != NULL; trav = trav->next) {
		args = (trav->args == NULL) ?
			NULL : cafe_expr_copy(trav->def->ns->allocator,
					      trav->args);
		*result = nature_create(trav->def, args);
		if (*result == NULL) {
			cafe_error_set(NULL, CAFE_NOMEM, "copying nature list");
			return CAFE_NOMEM;
		}
		result = &(*result)->next;
	}

	return CAFE_OK;
}


/*
 * Create new value type representing a structure with the given name,
 * attributes, and natures.
 */
int cafe_value_struct_create(const char *name, struct cafe_namespace *ns,
			     struct cafe_array *attr_list,
			     struct cafe_value_nature_ref *natures,
			     struct cafe_value_type **type)
{
	struct cafe_vm_ident *ident;
	struct cafe_byte_buffer namebuf;
	unsigned int i;
	int rc;

	assert(name != NULL);
	assert(ns != NULL);
	assert(attr_list != NULL);
	assert(type != NULL);

	name = get_qualified_ident(ns, name, &ident);
	if ((name == NULL) || (*name != 0))
		return CAFE_INVALID_VALUE_TYPE_STR;
	if (ident->type != NULL) {
		if (same_attribute_list(ident->type->attributes,
					attr_list->data, 
					(unsigned int)attr_list->len) &&
		    same_nature_ref_list(ident->natures, natures))
		{
			free_attributes(ns, attr_list->data, 
			                (unsigned int)attr_list->len);
			*type = ident->type;
			return CAFE_OK;
		}
		return CAFE_REDECL_VALUE_TYPE;
	}

	*type = cafe_type_alloc(ident, CAFE_TYPE_STRUCT);
	if (*type == NULL)
		return CAFE_NOMEM;

	(*type)->attributes = attr_list->data;
	(*type)->elements = (unsigned int) attr_list->len;
	(*type)->align = 0;
	rc = nature_ref_list_copy(natures, &ident->natures);
	if (rc != CAFE_OK) {
		cafe_type_free(*type);
		return rc;
	}

	rc = cafe_bbuf_init(&namebuf, ns->allocator, 64);
	if (rc != CAFE_OK) {
		cafe_type_free(*type);
		return rc;
	}
	namebuf.data[namebuf.len++] = 'S';
	namebuf.data[namebuf.len++] = '[';
	rc = cafe_ident_name_str(&namebuf, ident);
	if (rc != CAFE_OK)
		goto fail;

	if (attr_list->len > 0) {
		rc = cafe_bbuf_write(&namebuf, ',');
		if (rc != CAFE_OK)
			goto fail;
	}

	for (i = 0; i < attr_list->len; ++i) {
		struct cafe_vm_attr *a =
			cafe_array_val_ptr(attr_list, struct cafe_vm_attr, i);
		struct cafe_value_type *attr_type = cafe_attr_type(a);

		a->type = *type;
		if ((*type)->align < attr_type->align)
			(*type)->align = attr_type->align;

		rc = bbuf_element_append(&namebuf, a);
		if (rc != CAFE_OK)
			goto fail;
	}

	if (attr_list->len != 0) {
		const struct cafe_vm_attr *last =
			cafe_array_val_ptr(attr_list, struct cafe_vm_attr,
					   attr_list->len - 1);
		(*type)->size = (last->offset + cafe_attr_type(last)->size + 
				(*type)->align - 1) & (0 - (*type)->align);
	} else
		(*type)->size = 0;

	rc = bbuf_nature_list_append(&namebuf, natures);
	if (rc != CAFE_OK)
		goto fail;

	rc = cafe_bbuf_append(&namebuf, "]", 2);
	if (rc != CAFE_OK)
		goto fail;

	(*type)->type_str = namebuf.data;

	return CAFE_OK;

fail:
	cafe_type_free(*type);
	cafe_bbuf_release(&namebuf);
	return rc;
}

/*
 * Add a new entry to the given attribute list.  The new entry will have the
 * given kind, ident, natures, and name.
 */
int cafe_value_attr_list_add_ident(cafe_attr_type_tag kind,
				   struct cafe_vm_ident *ident,
				   const char *name,
				   struct cafe_value_nature_ref *natures,
				   struct cafe_array *list)
{
	struct cafe_vm_attr *new;
	struct cafe_value_type *attr_type;
	unsigned int index;
	int rc;

	assert(ident != NULL);
	assert(list != NULL);
	assert(name != NULL);
	rc = cafe_array_capacity_set(list, list->len + 1);
	if (rc != CAFE_OK)
		return rc;

	index = (unsigned int) list->len++;
	new = cafe_array_val_ptr(list, struct cafe_vm_attr, index);
	new->kind = kind;
	new->ident = build_ident(ident->ns, CAFE_IDENT_ATTR, name,
				 strlen(name));
	assert(new->ident != NULL);
	new->ident->children = ident;
	rc = nature_ref_list_copy(natures, &new->ident->natures);
	if (rc != CAFE_OK)
		return rc;

	new->index = index;
	attr_type = ident->type;
	if (index != 0) {
		const struct cafe_vm_attr *prev =
			cafe_array_val_ptr(list, struct cafe_vm_attr,
					   index - 1);
		new->offset = (prev->offset + cafe_attr_type(prev)->size + 
				attr_type->align - 1) & (0 - attr_type->align);
	} else
		new->offset = 0;

	return CAFE_OK;
}

/*
 * Given an attribute describing an existing struct-type's member, add a new
 * entry to the given list representing an attribute with the same name and
 * type.
 */
int cafe_value_attr_list_add(struct cafe_vm_attr *attr,
			     struct cafe_array *list)
{
	assert(attr != NULL);
	assert(list != NULL);
	return cafe_value_attr_list_add_ident(attr->kind, attr->ident->children,
					      attr->ident->name,
					      attr->ident->natures, list);
}
