/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
21may14,tle  Improve formatting of arrays in tables and skip empty tables

DESCRIPTION

*/

#include "cafe_assert.h"
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>		/* for isascii() */
#include <cafe/cafe.h>
#include <cafe/value_print.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/autogenerated/cafe_aggregator.type.h>
#include <cafe/dict.h>
#include "strtoll.h"


struct column_desc {
	const struct cafe_value_type *member_type;
	unsigned int size;
	bool variable_length;
	bool hex;
};


static void cafe_dict_table_print(FILE *out, const struct cafe_dict_struct *dict,
				  const struct cafe_value_print_options *options);

/*
 * value_type_display_width - calculate the optimum character width for
 * displaying the given type
 *
 * <vt>: value type to calculate for
 */
static
unsigned int value_type_display_width(const struct cafe_value_type *vt)
{
	switch (vt->kind) {
	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_SHORT:
	case CAFE_TYPE_USHORT:
	case CAFE_TYPE_INT:
	case CAFE_TYPE_UINT:
	case CAFE_TYPE_LONG:
	case CAFE_TYPE_ULONG:
	case CAFE_TYPE_BOOL:
	case CAFE_TYPE_FLOAT:
	case CAFE_TYPE_DOUBLE:
		return 10;

	case CAFE_TYPE_STRING:
	case CAFE_TYPE_BLOB:
		return 30;

	case CAFE_TYPE_LONGLONG:
	case CAFE_TYPE_ULONGLONG:
		return 15;

	case CAFE_TYPE_ARRAY:
		return (value_type_display_width(cafe_base_type(cafe_element_type(vt))) + 1) * vt->elements + 1;

	case CAFE_TYPE_DYNARRAY:
		return 0;

	default:
		return 52;
	}
}

/*
 * value_is_private - Return true if the given value is not meant for display
 *
 * <ident>: identifier to test for private-ness
 */
static
bool value_is_private(const struct cafe_vm_ident *ident,
		      const struct cafe_value_print_options *options)
{
	if (options && options->include_private)
		return false;

	//td: PRIVATE and LABEL natures need to be defined publicly somewhere and not be "magic" values
	if (0 != cafe_find_nature_str(ident, cafe_PRIVATE_VT_DEF))
		return true;

	return false;
}


/*
 * value_label_get - Determine the given value's most appropriate label
 *
 * <ident>: identifier to get label for
 */
static
const char *value_label_get(const struct cafe_vm_ident *ident)
{
	const struct cafe_value_nature_ref *label;
	const char *name;

	label = cafe_find_nature_str(ident, cafe_LABEL_VT_DEF);
	if (0 != label && 0 != label->args &&
	    label->args->kind == CAFE_EXPR_STRING)
		name = label->args->u.str.p;
	else
		name = ident->name;

	return name;
}


static
void string_print(FILE *out, const char *str, int field_width)
{
	if (field_width > 0) {
		/* truncate/pad strings to "field_width" length */
		const size_t len = strlen(str);
		if (len < (size_t)field_width)
			(void)fprintf(out, "%*s", field_width, str);
		else
			(void)fwrite_unlocked(str, field_width, 1, out);
	} else
		(void)fputs_unlocked(str, out);
}


/*
 * settable_column_init - initialize a table column entry representing a value
 * annotated with a cafe.SET nature, and prints this column's header.
 *
 * <out>: file stream to receive the column header string
 *
 * <column>: entry where calculated values will be recorded 
 *
 * <attr>: description of the settable value 
 */
static
void settable_column_init(FILE *out, struct column_desc *column,
			  const struct cafe_vm_attr *attr)
{
	const char *name = value_label_get(attr->ident);
	struct cafe_vm_attr *subattr;
	const struct cafe_value_type *value_type = NULL;
	struct cafe_value_type * vt = cafe_base_type(cafe_attr_type(attr));
	assert(vt != NULL);

	cafe_type_attr_foreach(subattr, vt) {
		if (strcmp(subattr->ident->name, "value") == 0) {
			value_type = cafe_base_type(cafe_attr_type(subattr));
			break;
		}
	}
	assert(value_type != NULL);

	column->member_type = value_type;
	column->size = value_type_display_width(value_type);
	if (column->size < strlen(name))
		column->size = (unsigned int) strlen(name);

	(void)fprintf(out, "%*s ", (int)column->size, name);
	column->hex = false;
}


/*
 * settable_print - print a value annotated with the cafe.SET nature to the
 * given stream; collapses the settable structure to a single value.
 *
 * <out>: file stream to receive emitted value
 *
 * <type>: description of the settable value 
 *
 * <p>: pointer to value's base address
 *
 * <options>: options used to transform any output
 */
static
void settable_print(FILE *out, const struct cafe_value_type *type,
		    const char *p,
		    const struct cafe_value_print_options *options)
{
	const void *value = NULL;
	const struct cafe_value_type *value_type = NULL;
	bool *set = NULL;
	bool is_set = false;
	struct cafe_vm_attr *subattr;

	cafe_type_attr_foreach(subattr, type) {
		if (strcmp(subattr->ident->name, "is_set") == 0) {
			set = (bool *) (p + subattr->offset);
			is_set = *set;
			if (!is_set)
				goto not_set;
		} else {
			value = p + subattr->offset;
			value_type = cafe_base_type(cafe_attr_type(subattr));
		}
	}

	if ((set != NULL) && is_set) {
		assert(value != NULL);
		assert(value_type != NULL);
		cafe_value_print(out, NULL, value_type, value, options);
	} else
	not_set:
		string_print(out, "Not Set", options->field_width);
}


/*
 * cafe_value_print - Print the given value using a value-specific format
 *
 * <out>: output file stream
 *
 * <attr>: optional attribute to get natures from
 *
 * <type>: pointer to type of data
 *
 * <p>: pointer to data
 *
 * <options>: optional print options
 */
void cafe_value_print(FILE *out, const struct cafe_vm_ident *attr,
		      const struct cafe_value_type *type,
		      const char *p,
		      const struct cafe_value_print_options *options)
{
	int field_width = options ? options->field_width : -1;
	int radix = options ? options->radix : 0;

	if (attr == NULL)
		attr = type->ident;

	if (options && options->type_name) {
		struct cafe_value_print_options attropts = *options;
		char *name = cafe_type_name_str(type);
		(void)fprintf(out, "%s:[", name);
		CAFE_FREE(type->ident->ns->allocator, name);
		attropts.type_name = false;
		cafe_value_print(out, attr, type, p, &attropts);
		(void)fputc(']', out);
		return;
	}

	if (p == NULL) {
		(void)fputs("NULL", out);
		return;
	}

	type = cafe_base_type(type);
	assert(type != NULL);

	if (field_width == 0)
		field_width = value_type_display_width(type);

	switch (type->kind) {
	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_SHORT:
	case CAFE_TYPE_INT:
	case CAFE_TYPE_LONG:
	case CAFE_TYPE_LONGLONG: {
		cafe_any_value v;

		/* prefer hexadecimal format for address types */
		if (radix == 0 && 0 != cafe_find_nature_str(attr, cafe_ADDRESS_VT_DEF))
			radix = 16;

		cafe_value_load_type(type, p, &v);

		if (field_width == -1)
			field_width = 0;
		if (radix == 16) {
			(void)fprintf(out, "%#*llx", field_width, v.ll);
		} else {
			(void)fprintf(out, "%*lld", field_width, v.ll);
		}
		break;
	}

	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_USHORT:
	case CAFE_TYPE_UINT:
	case CAFE_TYPE_ULONG:
	case CAFE_TYPE_ULONGLONG: {
		cafe_any_value v;

		/* prefer hexadecimal format for address types */
		if (radix == 0 && 0 != cafe_find_nature_str(attr, cafe_ADDRESS_VT_DEF))
			radix = 16;

		cafe_value_load_type(type, p, &v);

		if (field_width == -1)
			field_width = 0;
		if (radix == 16) {
			(void)fprintf(out, "%#*llx", field_width, v.ull);
		} else {
			(void)fprintf(out, "%*llu", field_width, v.ull);
		}
		break;
	}

	case CAFE_TYPE_BOOL:
		(void)fprintf(out, "%*s", field_width, *(bool *)p ? "true" : "false");
		break;

	case CAFE_TYPE_FLOAT:
	case CAFE_TYPE_DOUBLE: {
		cafe_any_value v;

		cafe_value_load_type(type, p, &v);

		if (field_width == -1)
			field_width = 0;
		(void)fprintf(out, "%*g", field_width, v.d);
		break;
	}

	case CAFE_TYPE_STRING: {
		const char *str = *(char **)p == NULL ? "NULL" : *(char **)p;
		assert(options != NULL);
		string_print(out, str, options->field_width);
		break;
	}

	case CAFE_TYPE_BLOB: {
		size_t i;
		const struct cafe_vm_blob * blob = (struct cafe_vm_blob*)p;

		(void)fprintf (out, "blob[%lu]: ", (unsigned long)blob->l);
		for (i = 0; i < blob->l; ++i)
			if (!isascii(*((char *)blob->p + i)))
				break;
		if (i < blob->l) {
			/* not everything was interpretible as ascii, so just
			   emit the bytes */
			for (i = 0; i < blob->l; ++i) {
				(void)fprintf(out, "%02x", *((char *) blob->p + i));
				if (i < blob->l - 1)
					(void)fputs(", ", out);
			}
		} else
			for (i = 0; i < blob->l; ++i)
				(void)fputc(*((char *) blob->p + i), out);
		break;
	}

	case CAFE_TYPE_ARRAY: {
		unsigned int i;
		struct cafe_value_type *subtype = cafe_element_type(type);
		struct cafe_value_print_options attropts;
		int field_width = 0;
		int remaining_width = 0;

		if (options) {
			attropts = *options;
			field_width = options->field_width;
			remaining_width = field_width-1;
		}
		attropts.field_width = (field_width > 0 && subtype->kind != CAFE_TYPE_STRUCT) ? 10 : -1;
		attropts.field_names = false;
		fputc('[', out);
		for (i = 0; i < type->elements; ++i) {
			if (i > 0) {
				(void)fputc(',', out);
				--remaining_width;
			}
			if (attropts.field_width > 0 && remaining_width < 20) {
				if (remaining_width > 3) {
					remaining_width -= 3;
					(void)fprintf(out, "...");
				}
				break;
			}
			cafe_value_print(out, attr, subtype,
					 p + (i * subtype->size), &attropts);
			if (attropts.field_width > 0)
				remaining_width -= attropts.field_width;
		}
		(void)fputc(']', out);
		while (--remaining_width > 0)
			fputc(' ', out);
		break;
	}

	case CAFE_TYPE_DYNARRAY: {
		const struct cafe_vm_dynarray *array =
			(struct cafe_vm_dynarray *) p;
		unsigned int i;
		struct cafe_value_type *subtype = cafe_element_type(type);
		struct cafe_value_print_options attropts;
		int field_width = 0;
		int remaining_width = 0;

		if (options) {
			attropts = *options;
			field_width = options->field_width;
			remaining_width = field_width-1;
		}
		attropts.field_width = (field_width > 0 && subtype->kind != CAFE_TYPE_STRUCT) ? 10 : -1;
		attropts.field_names = false;
		fputc('[', out);
		for (i = 0; i < array->l; ++i) {
			if (i > 0) {
				(void)fputc(',', out);
				--remaining_width;
			}
			if (attropts.field_width > 0 && remaining_width < 20) {
				if (remaining_width > 3) {
					remaining_width -= 3;
					(void)fprintf(out, "...");
				}
				break;
			}
			cafe_value_print(out, attr, subtype,
					 (char *)array->p + (i * subtype->size),
					 &attropts);
			if (attropts.field_width > 0)
				remaining_width -= attropts.field_width;
		}
		(void)fputc(']', out);
		while (--remaining_width > 0)
			fputc(' ', out);
		break;
	}

	case CAFE_TYPE_DICT:
		cafe_dict_table_print(out, (struct cafe_dict_struct *)p,
				      options);
		break;

	case CAFE_TYPE_STRUCT: {
		struct cafe_vm_attr *subattr;
		bool at_least_one_written = false;

		if (cafe_find_nature_str(type->ident, "N[AGGREGATOR_TABLE]")
				&& type->elements == 1) {
			struct cafe_dict_struct *dict;
			subattr = type->attributes;
			dict = (struct cafe_dict_struct *) (p + subattr->offset);
			if (dict->nentries == 0) {
				fprintf(out, "empty table: %s", value_label_get(type->ident));
				return;
			}
		}
		cafe_type_attr_foreach(subattr, type) {
			struct cafe_value_type *attrtype =
				cafe_base_type(cafe_attr_type(subattr));
			assert(attrtype != NULL);
			if (subattr->kind != CAFE_ATTR_VAR &&
			    subattr->kind != CAFE_ATTR_KEY)
				continue;
			if (value_is_private(subattr->ident, options))
				continue;
			if (field_width == -1) {
				/* separate new value from the last one we
				   emitted */
				if (at_least_one_written) {
					if (attrtype->kind != CAFE_TYPE_DICT)
						(void)fputs(", ", out);
					else
						(void)fputc('\n', out);
				}
				at_least_one_written = true;
			}
			if (options && options->field_names) {
				/* emit the most appropriate name for the value */
				const char *name = value_label_get(subattr->ident);
				if (name) {
					(void)fprintf(out, "%s:", name);
					if (attrtype->kind == CAFE_TYPE_DICT) {
						struct cafe_dict_struct *dict = (struct cafe_dict_struct *) (p + subattr->offset);
						if (dict->nentries) {
							/* emit table name */
							(void)fprintf(out, " %s\n", value_label_get(cafe_element_type(dict->type)->ident));
						}
					}
				}
			}
			cafe_value_print(out, subattr->ident, attrtype,
					 p + subattr->offset, options);
		}
		break;
	}

	case CAFE_TYPE_FILTER:
		/* no actual data to relay */
		break;

	default:
		(void)fprintf(out, "???unknown type %d???", type->kind);
	}
}


/*
 * columns_initialize - Initialize the given column descriptor list and emit
 * column header names by recursively traversing the given value_type.
 * Structures are omitted; only their members are included, essentially
 * flattening any hierarchy.
 *
 * <out>: file stream to emit to
 *
 * <vt>: value_type to build column headers from
 *
 * <column>: current offset within cols array
 *
 * <cols>: column header array to be populated.  Must be NULL or
 * preallocated to an appropriate size.
 */
static
unsigned int columns_initialize(FILE *out, const struct cafe_value_type *vt,
				unsigned int column,
				struct column_desc *cols,
				const struct cafe_value_print_options *options)
{
	struct cafe_vm_attr *attr;
	struct cafe_vm_attr *variable_length_attr = NULL;

	assert(cols == NULL || out);
	assert(vt);
	assert(vt->kind == CAFE_TYPE_STRUCT);

	cafe_type_attr_foreach(attr, vt) {
		struct cafe_value_type *attr_type;

		if (attr->kind != CAFE_ATTR_VAR &&
		    attr->kind != CAFE_ATTR_KEY)
			continue;

		attr_type = cafe_base_type(cafe_attr_type(attr));
		assert(attr_type != NULL);
		if (attr_type->kind == CAFE_TYPE_STRUCT) {
			if (cafe_find_nature_str(attr->ident,
						 cafe_SET_VT_DEF))
			{
				if (cols)
					settable_column_init(out, &cols[column], attr);
				++column;
			} else
				column += columns_initialize(out, attr_type,
							     column, cols,
							     options);
			continue;
		}

		if (value_is_private(attr->ident, options))
			continue;

		if (cols) {
			const char *name = value_label_get(attr->ident);
			if (name == NULL)
				name = "";

			cols[column].member_type = vt;
			cols[column].size = value_type_display_width(attr_type);
			cols[column].variable_length = cols[column].size == 0;
			if (cols[column].size < strlen(name)) {
				cols[column].size = (unsigned int) strlen(name);
			}
			cols[column].hex = false;

			if (cols[column].variable_length && !variable_length_attr)
				variable_length_attr = attr;
			else
				(void)fprintf(out, "%*s ", (int)cols[column].size, name);
		}
		++column;
	}

	if (variable_length_attr)
		(void)fprintf(out, "%s", value_label_get(variable_length_attr->ident));

	return column;
}


/*
 * cafe_dict_header_print - Print a useful header for visualizing the
 * dictionary as a table to the given file stream.  Returns a column
 * descriptor to be used when emitting column data, or 0 on error.
 * Clients must free this descriptor when finished.
 *
 * <out>: file stream to emit to
 *
 * <dict>: handle to dictionary to be used
 */
static
struct column_desc *cafe_dict_header_print(
	FILE *out,
	const struct cafe_dict_struct *dict,
	const struct cafe_value_print_options *options)
{
	const struct cafe_value_type *vt =
		cafe_base_type(cafe_element_type(dict->type));
	struct column_desc *cols;
	unsigned int i;
	unsigned int ncols;
	unsigned int variable_length_column_size = 0;

	assert(vt != NULL);
	assert(vt->kind == CAFE_TYPE_STRUCT);

	ncols = columns_initialize(NULL, vt, 0, NULL, options);
	cols = CAFE_ALLOC(vt->ident->ns->allocator, (ncols + 1) * sizeof *cols);
	if (cols == NULL)
		return NULL;
	memset(cols, 0, (ncols + 1) * sizeof *cols);
	(void)columns_initialize(out, vt, 0, cols, options);
	(void)fputc('\n', out);

	for (i = 0; i < ncols; ++i) {
		unsigned int j;

		if (cols[i].variable_length && !variable_length_column_size) {
			variable_length_column_size = cols[i].size;
			continue;
		}
		if (cols[i].size > 0) {
			for (j = 0; j < cols[i].size; ++j)
				fputc('-', out);
			fputc(' ', out);
		}
	}
	for (i = 0; i < variable_length_column_size; ++i)
		fputc('-', out);
	(void)fputc('\n', out);

	return cols;
}


/*
 * dict_table_row_print - Print all data found at a specific "row" in the
 * dictionary (where a "row" is all data with the same lookup key).
 *
 * <out>: file stream to emit to
 *
 * <vt>: handle to row's value type
 *
 * <cols>: handle to column descriptor
 *
 * <data>: pointer to the begining of the row's data

 * <nested_offset>: base offset if recursively printing a nested structure
 */
static
int dict_table_row_print(FILE *out, const struct cafe_value_type *vt,
			 unsigned int column,
			 const struct column_desc *cols,
			 const struct cafe_value_data *data,
			 const struct cafe_value_print_options *options,
			 int nested_offset)
{
	struct cafe_vm_attr *attr;
	struct cafe_value_print_options attropts = CAFE_VALUE_PRINT_OPTIONS_INIT;
	unsigned int variable_length_column = 0;
	struct cafe_vm_attr *variable_length_attr = NULL;

	assert(vt->kind == CAFE_TYPE_STRUCT);

	if (options)
		attropts = *options;

	cafe_type_attr_foreach(attr, vt) {
		const void *arg = data->u.data + attr->offset + nested_offset;
		const struct cafe_value_type *attrtype =
			cafe_base_type(cafe_attr_type(attr));

		if (attr->kind != CAFE_ATTR_VAR &&
		    attr->kind != CAFE_ATTR_KEY)
			continue;

		attropts.field_width = cols[column].size;
		attropts.radix = cols[column].hex ? 16 : 0;

		if (value_is_private(attr->ident, &attropts))
			continue;

		assert(attrtype != NULL);
		if (attrtype->kind == CAFE_TYPE_STRUCT) {
			if (cafe_find_nature_str(attr->ident, cafe_SET_VT_DEF)) {
				settable_print(out, attrtype, arg, &attropts);
				fputc_unlocked(' ', out);
				++column;
			} else
				column += dict_table_row_print(out, attrtype,
							       column, cols,
							       data, &attropts,
							       (unsigned int) attr->offset);
			continue;
		}

		if (cols[column].variable_length && !variable_length_attr) {
			variable_length_attr = attr;
			variable_length_column = column;
		} else {
			cafe_value_print(out, attr->ident, attrtype, arg, &attropts);
			(void)fputc(' ', out);
		}
		++column;
	}

	if (variable_length_attr) {
		const void *arg = data->u.data + variable_length_attr->offset + nested_offset;
		const struct cafe_value_type *attrtype =
			cafe_base_type(cafe_attr_type(variable_length_attr));

		attropts.field_width = -1;
		attropts.radix = cols[variable_length_column].hex ? 16 : 0;

		cafe_value_print(out, variable_length_attr->ident, attrtype, arg, &attropts);
		(void)fputc(' ', out);
	}
	return column;
}


/*
 * cafe_dict_table_data_print - Print all data found in the given dictionary, in
 * tabular format.
 *
 * <out>: file stream to emit to
 *
 * <dict>: handle to dictionary to be used
 *
 * <cols>: handle to column descriptor
 */
static
void cafe_dict_table_data_print(FILE *out, const struct cafe_dict_struct *dict,
				const struct column_desc *cols,
				const struct cafe_value_print_options *options)
{
	struct cafe_dict_iterator iter;
	struct cafe_value_data *val;
	struct cafe_value_type *element_type = cafe_base_type(cafe_element_type(dict->type));

	assert(element_type != NULL);

	/* emit each row in turn */
	val = cafe_dict_first(dict, &iter);
	while (val != 0) {
		(void)dict_table_row_print(out, element_type, 0, cols, val, options, 0);
		(void)fputc('\n', out);
		val = cafe_dict_next(dict, &iter);
	}
}


/*
 * calc_inverted_columns - When printing an "inverted" table, there are only
 * two widths we need to discover, the width of the label column (width of the
 * longest label) and the width of the value column (the width of the widest
 * value in the given value type).
 *
 * <vt>: value type to use for this calculation
 *
 * <label_width>: Returned value; width (in chars) of the label column
 *
 * <value_width>: Returned value; width (in chars) of the value column
 */
static
void calc_inverted_columns(const struct cafe_value_type *vt,
			   unsigned int *label_width,
			   unsigned int *value_width)
{
	struct cafe_vm_attr *attr;

	assert(vt->kind == CAFE_TYPE_STRUCT);
	cafe_type_attr_foreach(attr, vt) {
		const char *name;
		size_t val_len;
		struct cafe_value_type *attrtype =
			cafe_base_type(cafe_attr_type(attr));

		if (attr->kind != CAFE_ATTR_VAR &&
		    attr->kind != CAFE_ATTR_KEY)
			continue;

		assert(attrtype != NULL);
		if (attrtype->kind == CAFE_TYPE_STRUCT) {
			calc_inverted_columns(attrtype, label_width,
					      value_width);
			continue;
		}

		name = value_label_get(attr->ident);
		if (name != 0) {
			size_t len = strlen(name);
			if (len > *label_width)
				*label_width = (unsigned int) len;
		}

		val_len = value_type_display_width(attrtype);
		if (val_len > *value_width)
			*value_width = (unsigned int) val_len;
	}
}


/*
 * print_inverted_row - Print a single row of an inverted table
 *
 * <out>: stream to emit output to
 *
 * <label_width>: width (in chars) of the label column
 *
 * <value_width>: width (in chars) of the value column
 *
 * <dict>: dictionary representing this table
 *
 * <vt>: value type of the member to print in this row
 *
 * <offset>: offset (from a dictionary row entry base) of the member we're
 *           printing in this row
 */
static
void print_inverted_row(FILE *out,
			unsigned int label_width, unsigned int value_width,
			const struct cafe_dict_struct *dict,
			const struct cafe_vm_attr *attr,
			unsigned int offset,
			const struct cafe_value_print_options *options)
{
	struct cafe_value_type *vt;
	struct cafe_value_data *val;
	struct cafe_dict_iterator iter;
	unsigned int len;
	struct cafe_value_print_options attropts = CAFE_VALUE_PRINT_OPTIONS_INIT;

	assert(attr != NULL);
	vt = cafe_base_type(cafe_attr_type(attr));
	assert(vt != NULL);
	assert(attr->kind == CAFE_ATTR_VAR || attr->kind == CAFE_ATTR_KEY);
	if (value_is_private(attr->ident, options))
		return;
	if (vt->kind == CAFE_TYPE_STRUCT) {
		struct cafe_vm_attr *subattr;
		const struct cafe_value_nature_ref *label =
			cafe_find_nature_str(attr->ident, cafe_LABEL_VT_DEF);
		if (0 != label && 0 != label->args &&
		    label->args->kind == CAFE_EXPR_STRING) {
			const char *name = label->args->u.str.p;
			fprintf(out, "%s\n", name);
		}
		cafe_type_attr_foreach(subattr, vt) {
			if (subattr->kind != CAFE_ATTR_VAR &&
			    subattr->kind != CAFE_ATTR_KEY)
				continue;
			print_inverted_row(out, label_width, value_width, dict,
					   subattr, (unsigned int) attr->offset + offset, options);
		}
		return;
	}

	/* emit label column */
	len = fprintf(out, "%s", value_label_get(attr->ident));
	for (; len < label_width; ++len)
		(void)fputc(' ', out);
	(void)fputc(' ', out);

	if (options)
		attropts = *options;

	/* emit data columns */
	val = cafe_dict_first(dict, &iter);
	while (val != 0) {
		attropts.field_width = value_width;
		cafe_value_print(out, attr->ident, vt,
				 val->u.data + attr->offset + offset, &attropts);
		val = cafe_dict_next(dict, &iter);
		(void)fputc(' ', out);
	}
	(void)fputc('\n', out);
}

/*
 * dict_inverted_print - Print the given dictionary as an "inverted" table; use
 * the dictionary rows as table columns, and the dictionary row members as
 * rows.
 *
 * <out>: file stream to emit to
 *
 * <dict>: handle to dictionary to print
 */
static
void dict_inverted_print(FILE *out, const struct cafe_dict_struct *dict,
			 const struct cafe_value_print_options *options)
{
	const struct cafe_value_type *vt = cafe_base_type(cafe_element_type(dict->type));
	struct cafe_vm_attr *attr;
	unsigned int i;
	unsigned int ncols = (unsigned int) cafe_dict_count_rows(dict);
	unsigned int label_width = 0;
	unsigned int value_width = 0;

	assert(vt != NULL);
	assert(vt->kind == CAFE_TYPE_STRUCT);

	calc_inverted_columns(vt, &label_width, &value_width);

	/* print keys */
	for (i = 0; i < dict->type->elements; i++) {
		attr = dict->type->keymap[i];
		print_inverted_row(out, label_width, value_width,
				   dict, attr, 0, options);
	}

	/* header separater */
	for (i = 0; i < label_width; ++i)
		(void)fputc('-', out);
	(void)fputc(' ', out);

	for (i = 0; i < ncols; ++i) {
		unsigned int j;
		for (j = 0; j < value_width; ++j)
			(void)fputc('-', out);
		(void)fputc(' ', out);
	}
	fputc('\n', out);

	/* print values */
	cafe_type_attr_foreach(attr, vt) {
		if (attr->kind != CAFE_ATTR_VAR)
			continue;
		print_inverted_row(out, label_width, value_width,
				   dict, attr, 0, options);
	}
}


/*
 * cafe_dict_table_print - Print the given dictionary as a table; chooses the most
 * appropriate formate (normal or inverted) based on the table's dimensions.
 *
 * <out>: file stream to emit to
 *
 * <dict>: handle to dictionary to print
 */
static
void cafe_dict_table_print(FILE *out, const struct cafe_dict_struct *dict,
			   const struct cafe_value_print_options *options)
{
	int invert = cafe_find_nature_str(dict->type->ident, cafe_INVERT_VT_DEF) != NULL;
	int invertable = cafe_find_nature_str(dict->type->ident, cafe_INVERTABLE_VT_DEF) != NULL;
	size_t rows = cafe_dict_count_rows(dict);

	if (rows == 0) {
		(void)fprintf(out, "<empty>");
		return;
	}

	if (!invert && invertable) {
		/* count dictionary rows */
		const unsigned int ncols = columns_initialize(
			NULL, cafe_base_type(cafe_element_type(dict->type)),
			0, NULL, options);
		invert = ncols > rows;
	}
	if (!invert) {
		struct column_desc *cols = cafe_dict_header_print(out, dict, options);
		assert(cols != NULL);
		cafe_dict_table_data_print(out, dict, cols, options);
		CAFE_FREE(dict->type->ident->ns->allocator, cols);
	} else
		dict_inverted_print(out, dict, options);

}


#if !defined(NDEBUG)
static
void indent(const unsigned int level)
{
	unsigned int i;
	for (i = 0; i < level; ++i)
		printf("\t");
}


void value_type_print(const char *name, const struct cafe_value_type *vt,
		      unsigned int indent_lvl)
{
	char *type_name;

	assert(vt);
	indent(indent_lvl);
	if (name != NULL)
		printf("%s : ", name);

	type_name = cafe_type_name_str(vt);
	printf("%s (%p): ", type_name, vt);
	CAFE_FREE(vt->ident->ns->allocator, type_name);
	++indent_lvl;

	switch (vt->kind) {
	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_SHORT:
	case CAFE_TYPE_USHORT:
		printf("CAFE_TYPE_SHORT\n");
		break;
	case CAFE_TYPE_INT:
	case CAFE_TYPE_UINT:
		printf("CAFE_TYPE_INT\n");
		break;
	case CAFE_TYPE_LONG:
	case CAFE_TYPE_ULONG:
		printf("CAFE_TYPE_LONG\n");
		break;
	case CAFE_TYPE_LONGLONG:
	case CAFE_TYPE_ULONGLONG:
		printf("CAFE_TYPE_LONGLONG\n");
		break;
	case CAFE_TYPE_BOOL:
		printf("CAFE_TYPE_BOOL\n");
		break;
	case CAFE_TYPE_FLOAT:
		printf("CAFE_TYPE_FLOAT\n");
		break;
	case CAFE_TYPE_DOUBLE:
		printf("CAFE_TYPE_DOUBLE\n");
		break;

	case CAFE_TYPE_STRING:
		printf("CAFE_TYPE_STRING\n");
		break;

	case CAFE_TYPE_BLOB:
		printf("CAFE_TYPE_BLOB\n");
		break;

	case CAFE_TYPE_ARRAY:
		printf("CAFE_TYPE_ARRAY (%u elements)\n", vt->elements);
		break;

	case CAFE_TYPE_DICT:
	{
		printf("CAFE_TYPE_DICT\n");
		value_type_print(NULL, cafe_element_type(vt), indent_lvl);
		break;
	}
	case CAFE_TYPE_STRUCT:
	{
		struct cafe_vm_attr *subattr;

		printf("CAFE_TYPE_STRUCT\n");
		cafe_type_attr_foreach(subattr, vt) {
			value_type_print(subattr->ident->name,
					 cafe_attr_type(subattr), indent_lvl);
		}
		break;
	}
	case CAFE_TYPE_DYNARRAY:
		printf("CAFE_TYPE_{DYNARRAY}-not-implemented\n");
		break;

	case CAFE_TYPE_TYPEDEF:
		printf("CAFE_TYPE_TYPEDEF\n");
		break;

	case CAFE_TYPE_INSTANCE:
		printf("CAFE_TYPE_INSTANCE\n");
		break;

	default:
		printf("???unknown type %d???\n", vt->kind);
	}

	indent(indent_lvl);
	printf("size: %" PRIsize_t "\n", vt->size);
}
#endif  /* if !defined(NDEBUG) */
