/* Copyright (c) 2003-2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#if !defined(__kernel_map_parser_h)
#define __kernel_map_parser_h

#if defined(__cplusplus)
    extern "C" {
#endif


#include <stdint.h>
#include <stdbool.h>


/* length of longest symbol name.  Taken out of linux/kallsyms.h.  If this
   value changes, change the field widths in the scanf() calls below as well. */
#define KSYM_NAME_LEN     (128)

/* length of longest symbol name.  Taken out of linux/module.h.  If this value
   changes, change the field widths in the scanf() calls below as well. */
#define MODULE_NAME_LEN   (60)

typedef struct {
	uintptr_t address;
#if !defined(CAFE_LINUX24)
	char type;
#endif
	char name[KSYM_NAME_LEN];
	char module[MODULE_NAME_LEN];
} ksym_entry_t;


/**
 *  Locate the address of a kernel symbol.
 * symbolname - name of the symbol to look for.  Must be a valid null-
 *              terminated string.
 * mapFilename - pointer to a string which holds a suggested location of
 *               the system map.  If the mapFilename is 0, or the specified
 *               file isn't found, default locations are searched. mapFilename
 *               is set to point at the full filename of the actual map
 *               used, or zero if no maps were found.
 * returns true if address now corresponds to the given symbol
 * within the kernel, and false otherwise.
 */
extern bool kernel_symbol_find(const char *const symbol_name,
			       uintptr_t *address,
			       size_t *size,
			       const char **const map_filename);


/**
 *  Remove the checksum wart from any kernel symbol string.  The pointer passed
 * must point at a valid null-terminated string.
 */
extern void kernel_symbol_strip_checksum(char *sym);


/**
 *  This function reads a single entry in the public symbol map exported
 * via the procfs.
 *
 *   entry_num - the line number corresponding to the symbol requested.  Note
 *               that this function currently skips entries associated with
 *               modules.
 *   symbol - a pointer to space allocated for a ksym_entry_t.  This structure
 *            will be populated with the symbol data found, or left untouched
 *            if parsing fails.
 *
 * Returns - true iff symbol has been populated with the requested info.
 */
extern bool kernel_symbol_read_exported(const unsigned int entry_num,
					ksym_entry_t *symbol);


/**
 *  Reads kernel symbols exported via the proc filesystem, and returns them to
 * the caller.  This function allocates an array of symbol entries and returns
 * them in the "symbols" parameter.  This parameter must point to a valid
 * "ksym_entry_t *" when this function is called.  Further, the caller is
 * responsible for free()ing the returned pointer.
 *
 * Returns the number of symbols read, or -1 on error.
 */
extern int kernel_symbols_read(ksym_entry_t **symbols);


#if defined(__cplusplus)
    }   /* extern "C" */
#endif

#endif  /* if !defined(__kernel_map_parser_h) */
