/* file_stream_reader.c - EP used to fire events from file */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
15may14,sdt  Fix 'Checked return' warning
13may14,sdt  Rename cafe tasks to tAnalysis* - US37854
11mar14,sdt  Fix 'Checked return' warning


DESCRIPTION
 * This event processor supports a number of settings for controlling
 * for "file replay".  These settings control how many events should
 * be dispatched from a starting record-number to an ending
 * record-number where the first record in the log is 0.
 *
 * The "range_start" setting specifies which record to start from,
 * where 0 means the beginning of the log.

 * The "range_end" setting specifies which record to stop at, where 0
 * means the end of the log.
 *
 * Whenever a range is specified that range is half-open (i.e., the
 * range 37146-37150 would yield 4 records. Or, put another way, it is
 * START..END-1 (inclusive).
 *
 * If "range_end < range_start" the enumeration order is undefined.
 *
 * There are corresponding settings for dispatching a custom
 * (value)type whenever the range conditions are met.
 *
 * The "range_start_type" is dispatched when "range_start" is met.
 *
 * The "range_end_type" is dispatched when "range_end" is met.
 *
 * The "start_type" is always dispatched when replaying the log.
 *
 * Examples:
 *
 * If you have a log with 100 events in it and you configure the range
 * settings with:
 *
 *   range_start = 10
 *   range_end   = 15
 *
 * then you will get 5 records dispatched (10, 11, 12, 13, 14).
 *
 * If in this example you also specified associated types for the
 * range conditions you would get the following sequence of events:
 *
 *   <your-start-type>
 *   all-events-up-until-range-start
 *   <your-range-start-type>
 *   all-events-up-until-range-end
 *   <your-range-end-type>
 *   cafe.stop
 *
 * Any time the settings are changed the file position is set back to
 * the beginning (i.e., it is rewound).
 *
 * The default values for "range_start" and "range_end" are 0 which
 * means the whole log would be replayed.  There are no default types
 * specified for "start_type", "range_start_type" or "range_end_type".
 */

#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <cafe/stdio.h>
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/autogenerated/file_stream_reader.type.h>
#include <cafe/log_stream.h>
#include <cafe/event_map.h>
#include <cafe/ep_worker.h>
#include <cafe/errors.h>
#include <cafe/ep/file_stream_reader.h>
#include "../src/cafe_trace.h"
#include "../src/leb128.h"
#include "../src/cafe_assert.h"
#include "../src/cafe_os.h"  /* for nanosleep() */

typedef unsigned long long fsr_seqno_t;

enum {
	PLAYBACK_START = 0,
	PLAYBACK_RANGE_START,
	PLAYBACK_RANGE_END,
	PLAYBACK_START_DISPATCH,
	PLAYBACK_END_OF_FILE,
	PLAYBACK_NEVENTS
};

enum {
	ODATA_SEQUENCE = 0,
	ODATA_CLIENT_DATA
};

struct playback_event {
	struct cafe_value_type *types[2];
	struct cafe_event_descr *event_descr;
	cafe_eal_handle_t sink_actions;
};

struct client_data {
	FILE *in;
	struct cafe_value_type *metadata_event;
	struct cafe_value_type *sequence_event;
	struct cafe_value_type *client_data_event;
	struct cafe_event_map *event_map;
	struct cafe_ep_worker *worker;
	struct playback_event playback_events[PLAYBACK_NEVENTS];
	struct cafe_byte_buffer bytebuf;
	fsr_seqno_t seqno;
	bool dispatch_start_event_pending;
};

struct fstream_remote_info {
	cafe_eal_handle_t sink_actions;
};

static void start_parsing(void *arg);


static inline
bool have_playback_start_type(const struct client_data *node)
{
	return node->playback_events[PLAYBACK_START].types[0] != NULL;
}


static inline
bool have_playback_range_start_type(const struct client_data *node)
{
	return node->playback_events[PLAYBACK_RANGE_START].types[0] != NULL;
}


static inline
bool have_playback_range_end_type(const struct client_data *node)
{
	return node->playback_events[PLAYBACK_RANGE_END].types[0] != NULL;
}


static inline
bool have_playback_start_dispatch_type(const struct client_data *node)
{
	return node->playback_events[PLAYBACK_START_DISPATCH].types[0] != NULL;
}


static inline
bool have_playback_end_of_file_type(const struct client_data *node)
{
	return node->playback_events[PLAYBACK_END_OF_FILE].types[0] != NULL;
}


static
int update_output_playback_event(struct cafe_ep_inst *inst,
				 struct playback_event *pbe)
{
	struct cafe_event_action_list *sink_actions;
	cafe_eal_handle_t sink_actions_handle = CAFE_EAL_HANDLE_NULL;
	int rc;

	assert(pbe->types[0] != NULL);
	rc = cafe_event_action_list_alloc(inst, pbe->event_descr,
					  &sink_actions, &sink_actions_handle);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);

	if (rc == CAFE_OK)
		cafe_event_action_list_swap(&pbe->sink_actions,
					    sink_actions_handle);
	else
		cafe_event_action_list_release(&sink_actions_handle);

	return rc;
}


static
int update_output_playback_events(struct cafe_ep_inst *inst)
{
	size_t i;
	struct client_data *node;

	node = inst->client_data;

	for (i = 0; i < CAFE_NELEMENTS(node->playback_events); i++) {
		int rc;
		struct playback_event *pbe;
		pbe = &node->playback_events[i];
		if (pbe->types[0] == NULL)
			continue;
		rc = update_output_playback_event(inst, pbe);
		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}


static
void dispatch_playback_start(struct cafe_ep_inst *inst, fsr_seqno_t seqno)
{
	void *data[2];
	struct client_data *node = inst->client_data;

	assert(node->playback_events[PLAYBACK_START].types[0]);
	data[0] = NULL;
	data[1] = &seqno;
	cafe_dispatch_event(&node->playback_events[PLAYBACK_START].sink_actions, data);
}


static
void dispatch_playback_range_start(struct cafe_ep_inst *inst, fsr_seqno_t seqno)
{
	void *data[2];
	struct client_data *node = inst->client_data;

	assert(node->playback_events[PLAYBACK_RANGE_START].types[0]);
	data[0] = NULL;
	data[1] = &seqno;
	cafe_dispatch_event(&node->playback_events[PLAYBACK_RANGE_START].sink_actions, data);
}


static
void dispatch_playback_range_end(struct cafe_ep_inst *inst, fsr_seqno_t seqno)
{
	void *data[2];
	struct client_data *node = inst->client_data;

	assert(node->playback_events[PLAYBACK_RANGE_END].types[0]);
	data[0] = NULL;
	data[1] = &seqno;
	cafe_dispatch_event(&node->playback_events[PLAYBACK_RANGE_END].sink_actions, data);
}


static
void dispatch_playback_start_dispatch(struct cafe_ep_inst *inst,
				      unsigned long client_data)
{
	void *data[2];
	struct client_data *node = inst->client_data;

	assert(node->playback_events[PLAYBACK_START_DISPATCH].types[0]);
	data[0] = NULL;
	data[1] = &client_data;
	cafe_dispatch_event(&node->playback_events[PLAYBACK_START_DISPATCH].sink_actions, data);
}


static
void dispatch_playback_end_of_file(struct cafe_ep_inst *inst)
{
	void *data[2];
	struct client_data *node = inst->client_data;

	assert(node->playback_events[PLAYBACK_END_OF_FILE].types[0]);
	data[0] = NULL;
	data[1] = NULL;
	cafe_dispatch_event(&node->playback_events[PLAYBACK_END_OF_FILE].sink_actions, data);
}


static
int process_settings(struct cafe_ep_inst *inst)
{
	size_t i;
	int rc;
	struct file_stream_reader_workarea *w;
	struct client_data *node;
	const char *type_str[PLAYBACK_NEVENTS];
	unsigned int num_types;

	w = (struct file_stream_reader_workarea *)inst->workarea->u.data;
	node = inst->client_data;
	type_str[0] = w->start_type;
	type_str[1] = w->range_start_type;
	type_str[2] = w->range_end_type;
	type_str[3] = w->start_dispatch_type;
	type_str[4] = w->end_of_file_type;

	for (i = 0; i < CAFE_NELEMENTS(node->playback_events); i++) {
		struct playback_event *pbe;

		pbe = &node->playback_events[i];
		cafe_event_action_list_release(&pbe->sink_actions);

		if (type_str[i] == NULL || *type_str[i] == '\0') {
			pbe->types[0] = NULL;
			pbe->types[1] = NULL;
			pbe->event_descr = NULL;
			assert(pbe->sink_actions == CAFE_EAL_HANDLE_NULL);
			continue;
		}

		rc = cafe_type_parse(inst->session->ns, type_str[i], pbe->types);

		if (rc != CAFE_OK)
			return rc;

		switch (i) {
		case PLAYBACK_START_DISPATCH:
			pbe->types[1] = node->client_data_event;
			num_types = 2;
			break;
		case PLAYBACK_END_OF_FILE:
			num_types = 1;
			break;
		default:
			pbe->types[1] = node->sequence_event;
			num_types = 2;
			break;
		}

		rc = cafe_cfg_get_event(inst->session,
					num_types,
					pbe->types, &pbe->event_descr);

		if (rc != CAFE_OK)
			return rc;
	}

	rc = update_output_playback_events(inst);

	if (rc != CAFE_OK)
		return rc;

	clearerr_unlocked(node->in);	/* clear feof() */

	if (w->rewind) {
		node->seqno = 0;
		node->dispatch_start_event_pending = true;
		rewind(node->in);
	} else {
		node->dispatch_start_event_pending = false;
	}

	return CAFE_OK;
}


/**
 * "Starts" the configuration by opening the given output file.
 */
static
int fstream_reader_cfg_create_inst(struct cafe_ep_inst *inst)
{
	size_t i;
	struct file_stream_reader_workarea *w =
		(struct file_stream_reader_workarea *)inst->workarea->u.data;
	struct client_data *node;
	int rc;

	if ((node = CAFE_ALLOC(inst->allocator, sizeof *node)) == NULL)
		return CAFE_NOMEM;
	memset(node, 0, sizeof *node);
	node->in = fopen(w->filename, "rb");
	if (node->in == 0) {
		rc = CAFE_ERROR;
		cafe_error_set(NULL, rc, "Unable to open file \"%s\"", w->filename);
		cafe_error_set_errno(NULL, errno);
		goto error;
	}
#if defined(HAVE_POSIX_FADVISE)
	rc = fileno_unlocked(node->in);
	if (rc != -1) {
		posix_fadvise(rc, 0, 0, POSIX_FADV_SEQUENTIAL);
	}
#endif

	node->event_map = cafe_event_map_create(inst->session);
	if (node->event_map == 0) {
		rc = CAFE_ERROR;
		goto error;
	}

	rc = cafe_type_parse(inst->session->ns,
			     cafe_metadata_VT_DEF, &node->metadata_event);
	if (rc != CAFE_OK)
		goto error;

	rc = cafe_type_parse(inst->session->ns,
			     file_stream_reader_sequence_VT_DEF,
			     &node->sequence_event);

	if (rc != CAFE_OK)
		goto error;

	rc = cafe_type_parse(inst->session->ns,
			     file_stream_reader_client_data_VT_DEF,
			     &node->client_data_event);

	if (rc != CAFE_OK)
		goto error;

	inst->client_data = node;
	rc = cafe_ep_worker_create("tAnalysisFsReader", inst, start_parsing, inst,
				   &node->worker);
	if (rc != CAFE_OK) {
		rc = CAFE_ERROR;
		goto error;
	}

	for (i = 0; i < CAFE_NELEMENTS(node->playback_events); i++)
		node->playback_events[i].sink_actions = CAFE_EAL_HANDLE_NULL;

	rc = process_settings(inst);
	if (rc != CAFE_OK)
		goto error;

	rc = cafe_bbuf_init(&node->bytebuf, inst->allocator, 0);
	if (rc != CAFE_OK)
		goto error;

	return CAFE_OK;

error:
	if (node->event_map)
		cafe_event_map_destroy(node->event_map);
	if (node->in)
		(void)fclose(node->in);
	CAFE_FREE(inst->allocator, node);
	return rc;
}


static
bool fstream_event_map_iter_handler(void *client_data,
				    struct cafe_remote_event *remote)
{
	struct cafe_ep_inst *inst = client_data;
	struct fstream_remote_info *remote_info = remote->client_data;

	assert(remote_info != NULL);
	cafe_event_action_list_release(&remote_info->sink_actions);
	remote->client_data = NULL;
	CAFE_FREE(inst->allocator, remote_info);
	return true;
}


static
void fstream_reader_cfg_delete_inst(struct cafe_ep_inst *inst)
{
	size_t i;
	struct client_data *node = inst->client_data;

	cafe_ep_worker_delete(node->worker);

	cafe_event_map_iter(node->event_map,
			    fstream_event_map_iter_handler,
			    inst);

	cafe_event_map_destroy(node->event_map);
	(void)fclose(node->in);
	cafe_bbuf_release(&node->bytebuf);

	for (i = 0; i < CAFE_NELEMENTS(node->playback_events); i++)
		cafe_event_action_list_release(&node->playback_events[i].sink_actions);

	CAFE_FREE(inst->allocator, node);
}


/**
 *  Consume metadata event parameters from the input stream.
 */
static
int parse_metadata_event(struct cafe_ep_inst *inst)
{
	int rc;
	struct cafe_metadata md;
	struct cafe_remote_event *remote;
	struct client_data *node = inst->client_data;
	struct cafe_event_action_list *sink_actions;
	struct fstream_remote_info *remote_info;
	cafe_eal_handle_t sink_actions_handle;

	/* argument values */
	if (cafe_fdecode_val(node->metadata_event, &md, inst->allocator,
			     node->event_map, node->in) < 0)
		return CAFE_ERROR;
	remote = cafe_event_map_lookup(node->event_map, md.eid);
	if (remote != NULL) {
		cafe_release_values(node->metadata_event, &md, inst->allocator);
		return CAFE_OK;
	}
	TRACE(("Metadata for event %llu\n", md.eid));
	remote_info =
		CAFE_ALLOC(inst->allocator, sizeof *remote_info);
	if (remote_info == NULL) {
		rc = CAFE_NOMEM;
		goto error;
	}
	memset(remote_info, 0, sizeof *remote_info);
	remote_info->sink_actions = CAFE_EAL_HANDLE_NULL;
	rc = cafe_event_map_add_event(node->event_map, &md, remote_info, &remote);
	if (rc != CAFE_OK) {
		CAFE_FREE(inst->allocator, remote_info);
		goto error;
	}
	TRACE(("%s(0x%llx)\n", ((char **)md.event_data.p)[0], md.eid));

	rc = cafe_event_action_list_alloc(inst, remote->event,
					  &sink_actions,
					  &sink_actions_handle);
	if (rc != CAFE_OK)
		goto error;
	rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
	if (rc != CAFE_OK) {
		cafe_event_action_list_release(&sink_actions_handle);
		goto error;
	}
	cafe_event_action_list_swap(&remote_info->sink_actions,
				    sink_actions_handle);
	cafe_release_values(node->metadata_event, &md, inst->allocator);
	return CAFE_OK;
error:
	cafe_release_values(node->metadata_event, &md, inst->allocator);
	return rc;
}


static
void parse_header(struct cafe_ep_inst *inst)
{
	struct file_stream_reader_workarea *w =
		(struct file_stream_reader_workarea *)inst->workarea->u.data;
	struct client_data *node = inst->client_data;

	if (ftello(node->in) != 0)
		return;

	while (cafe_ep_worker_isrunning(node->worker)) {
		int rc;

		rc = cafe_fdecode_stream_header(node->in, NULL);
		if (rc != CAFE_OK) {
			if (feof_unlocked(node->in) && !ferror_unlocked(node->in)) {
				if (w->eof_poll) {
					struct timespec delay;
					delay.tv_sec = w->eof_poll / 1000;
					delay.tv_nsec = (w->eof_poll % 1000) * 1000 * 1000;
					(void)nanosleep(&delay, NULL);

					clearerr(node->in); /* clear feof() */
					rewind(node->in);
					continue;
				}
			}
			rewind(node->in);
		}
		break;
	}
}


static inline
bool valid_playback_range(struct cafe_ep_inst *inst)
{
	struct file_stream_reader_workarea *w;
	bool wide_open;

	w = (struct file_stream_reader_workarea *)inst->workarea->u.data;
	wide_open = w->range_start == 0 && w->range_end == 0;

	return wide_open || w->range_end - w->range_start > 0;
}


static
void parse_body(struct cafe_ep_inst *inst)
{
	struct file_stream_reader_workarea *w =
		(struct file_stream_reader_workarea *)inst->workarea->u.data;
	struct client_data *node = inst->client_data;
	off_t curpos = ftello(node->in);
	off_t cursize = 0;		/* Updated lazily */
	bool aborted = false;

	if (curpos == -1) {
		perror("Error accessing file stream");
		goto stop;
	}

	if (valid_playback_range(inst)) {
		// DE323: bail out if already at or beyond range_end
		if (w->range_end > 0 && node->seqno >= w->range_end) {
			goto stop;
		}
	}

	if (have_playback_start_type(node) && node->dispatch_start_event_pending) {
		dispatch_playback_start(inst, node->seqno);
		node->dispatch_start_event_pending = false;
	}

	if (have_playback_start_dispatch_type(node))
		dispatch_playback_start_dispatch(inst, w->client_data);

	while (cafe_ep_worker_isrunning(node->worker)) {
		unsigned long long eid, size;
		struct cafe_remote_event *remote;
		struct fstream_remote_info *remote_info;
		ssize_t header_bytes;
		unsigned int i;
		void **output_event_data;
		int rc;

		header_bytes = cafe_fdecode_event_header(node->in, &size, &eid);
		if (header_bytes < 0) {
			if (feof_unlocked(node->in) && !ferror_unlocked(node->in)) {
			handle_eof:
				if (w->eof_poll) {
					struct timespec delay;
					delay.tv_sec = w->eof_poll / 1000;
					delay.tv_nsec = (w->eof_poll % 1000) * 1000 * 1000;
					(void)nanosleep(&delay, NULL);

					clearerr(node->in); /* clear feof() */
					(void)fseeko(node->in, 0, SEEK_END);
					cursize = ftello(node->in);
					(void)fseeko(node->in, curpos, SEEK_SET);
					continue;
				}
				if (header_bytes == LEB128ERR_NONUM) {
					CAFE_LOCK();
					(void)cafe_ep_worker_stop(node->worker);
					CAFE_UNLOCK();
					break;
				}
			}

			/* Not EOF or partial event header */
			CAFE_LOCK();
			cafe_ep_worker_abort(node->worker);
			CAFE_UNLOCK();
			(void)printf("Error decoding event header @ 0x%llx\n",
			             (long long int)ftello(node->in));
			aborted = true;
			break;
		}
		if (cursize < curpos + header_bytes + size) {
			assert(curpos + header_bytes == ftello(node->in));
			(void)fseeko(node->in, 0, SEEK_END);
			cursize = ftello(node->in);
			(void)fseeko(node->in, curpos + header_bytes, SEEK_SET);
			if (cursize < curpos + header_bytes + size)
				goto handle_eof;
		}

		if (eid == 0) {
			/* metadata event. */
			CAFE_LOCK();
			rc = parse_metadata_event(inst);
			if (rc == CAFE_OK) {
				CAFE_UNLOCK();
				curpos += (off_t)(header_bytes + size);
				continue;
			}
			cafe_ep_worker_abort(node->worker);
			CAFE_UNLOCK();
			(void)printf("error binding new event: seqno=%llu: %s\n", node->seqno, cafe_strerror(rc));
			aborted = true;
			break;
		}

		if (have_playback_range_start_type(node) &&
		    valid_playback_range(inst)) {
			if (node->seqno == w->range_start)
				dispatch_playback_range_start(inst, node->seqno);
		}

		remote = cafe_event_map_lookup(node->event_map, eid);

		if (remote == NULL) {
			TRACE(("No metadata for event %llu\n", eid));
			goto stop;
		}

		TRACE(("%llu (%lld, %llx): %llx \"%s\"\n", node->seqno,
		       (long long int)ftello(node->in), size, eid,
		       remote->event->types[0]->type_str));

		/* read event data orthogonal data payload */
		rc = cafe_fread_event_data(node->in, remote->event,
					   inst->allocator, node->event_map,
					   &node->bytebuf);

		if (rc != CAFE_OK) {
			TRACE(("unable to decode event: %s\n", cafe_strerror(rc)));
			goto stop;
		}

		output_event_data = (void **)node->bytebuf.data;
		remote_info = remote->client_data;

		if (valid_playback_range(inst))
			cafe_dispatch_event(&remote_info->sink_actions, output_event_data);

		curpos += (off_t)(header_bytes + size);
		++node->seqno;

		for (i = 0; i < remote->event->num_types; ++i) {
			cafe_release_values(remote->event->types[i],
					    output_event_data[i],
					    inst->allocator);
		}

		if (valid_playback_range(inst)) {
			if (w->range_end > 0 && node->seqno == w->range_end) {
				if (have_playback_range_end_type(node)) {
					dispatch_playback_range_end(inst, node->seqno - 1);
				}
				break;
			}
		}
	}

	assert(aborted || curpos == ftello(node->in));

	if (!aborted && (curpos != ftello(node->in)))
		TRACE(("Error: exiting without being aborted\n"));
stop:
	if (!aborted && w->eof_poll == 0 &&
	    (feof_unlocked(node->in) && !ferror_unlocked(node->in))) {
		if (have_playback_end_of_file_type(node))
			dispatch_playback_end_of_file(inst);
	}

	if (cafe_ep_worker_isrunning(node->worker)) {
		CAFE_LOCK();
		(void)cafe_ep_worker_stop(node->worker);
		CAFE_UNLOCK();
	}
}


/**
 * Begin parsing this configuration instance's input file.
 */
static
void start_parsing(void *arg)
{
	struct cafe_ep_inst *inst = arg;
	parse_header(inst);
	parse_body(inst);
}


/**
 * "Starts" the configuration by opening the given output file.
 */
static
int fstream_reader_cfg_start(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	int rc;

	rc = cafe_ep_worker_start(node->worker);
	if (rc != CAFE_OK)
		return rc;
	return cafe_ep_worker_wait(node->worker);
}


static
int fstream_reader_cfg_stop(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	int rc;

	rc = cafe_ep_worker_stop(node->worker);
	if (rc != CAFE_OK)
		return rc;
	return cafe_ep_worker_wait(node->worker);
}

struct fstream_update_info {
	struct cafe_ep_inst *inst;
	int error;
};


static
bool fstream_event_map_iter_update_event_actions(
	void *client_data,
	struct cafe_remote_event *remote)
{
	struct fstream_update_info *info = client_data;
	struct cafe_ep_inst *inst = info->inst;
	struct fstream_remote_info *remote_info = remote->client_data;
	struct cafe_event_action_list *sink_actions;
	cafe_eal_handle_t sink_actions_handle;

	info->error = cafe_event_action_list_alloc(inst, remote->event,
						   &sink_actions,
						   &sink_actions_handle);
	if (info->error != CAFE_OK)
		return false;

	info->error = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
	if (info->error != CAFE_OK) {
		cafe_event_action_list_release(&sink_actions_handle);
		return false;
	}
	cafe_event_action_list_swap(&remote_info->sink_actions,
				    sink_actions_handle);

	return true;
}


static
int fstream_reader_cfg_update_output_event_actions(
	struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	struct fstream_update_info info;

	info.inst = inst;
	info.error = CAFE_OK;
	cafe_event_map_iter(node->event_map,
			    fstream_event_map_iter_update_event_actions,
			    &info);
	if (info.error == CAFE_OK)
		info.error = update_output_playback_events(inst);
	return info.error;
}


static
int fstream_reader_cfg_discover_events(struct cafe_ep_inst *inst, char **typep)
{
	int rc;
	struct file_stream_reader_workarea *w;
	FILE *fp;
	struct client_data *node;
	struct cafe_byte_buffer bbuf;

	*typep = NULL;
	w = (struct file_stream_reader_workarea *)inst->workarea->u.data;
	fp = fopen(w->filename, "rb");

	if (fp == NULL) {
		cafe_error_set(NULL, CAFE_ERROR,
			       "cannot open \"%s\"", w->filename);
		cafe_error_set_errno(NULL, errno);
		return CAFE_ERROR;
	}
#if defined(HAVE_POSIX_FADVISE)
	rc = fileno_unlocked(fp);
	if (rc != -1) {
		posix_fadvise(rc, 0, 0,
			      POSIX_FADV_SEQUENTIAL | POSIX_FADV_NOREUSE);
	}
#endif

	node = (struct client_data *)inst->client_data;
	rc = cafe_bbuf_init(&bbuf, inst->session->allocator, 128);

	if (rc != CAFE_OK) {
		(void)fclose(fp);
		return rc;
	}

	rc = cafe_fdecode_stream_header(fp, NULL);

	if (rc != CAFE_OK) {
		/* Warning: cannot read stream header (no version) */
		rewind(fp);
	}

	while (!feof_unlocked(fp)) {
		unsigned long long eid;
		unsigned long long size;
		ssize_t header_bytes;

		header_bytes = cafe_fdecode_event_header(fp, &size, &eid);

		if (header_bytes < 0) {
			if (feof_unlocked(fp))
				break;
			cafe_error_set(NULL, CAFE_ERROR,
				       "Error decoding event header @ 0x%llu\n",
				       (unsigned long long)ftello(fp));
			rc = CAFE_ERROR;
			goto error;
		}

		if (eid == 0) {
			unsigned int i;
			ssize_t nbytes;
			struct cafe_metadata md;

			nbytes = cafe_fdecode_val(node->metadata_event, &md,
						  inst->session->allocator,
						  NULL, fp);

			if (nbytes < 0) {
				cafe_error_set(NULL, CAFE_ERROR,
					"Error decoding metadata event @ 0x%llu\n",
					(unsigned long long)ftello(fp));
				rc = CAFE_ERROR;
				goto error;
			}

			for (i = 0; i < md.event_data.l; i++) {
				char *p = (((char **)md.event_data.p)[i]);
				size_t l = strlen(p);
				rc = cafe_bbuf_append(&bbuf, p, l);
				if (rc != CAFE_OK) {
					cafe_release_values(node->metadata_event,
							    &md, inst->session->allocator);
					goto error;
				}
			}

			cafe_release_values(node->metadata_event,
					    &md, inst->session->allocator);
		} else {
			if (fseeko(fp, (off_t)size, SEEK_CUR) == -1) {
				rc = CAFE_ERROR;
				goto error;
			}
		}
	}

	rc = cafe_bbuf_write(&bbuf, '\0');
	if (rc == CAFE_OK)
		*typep = bbuf.data;
error:
	if (fp != NULL)
		(void)fclose(fp);

	if (*typep == NULL)
		cafe_bbuf_release(&bbuf);

	return rc;
}


static
int fstream_reader_cfg_settings_commit(struct cafe_ep_inst *inst,
				       unsigned int num_settings,
				       const struct cafe_cfg_setting *settings)
{
	size_t i;

	for (i = 0; i < num_settings; i++) {
		int rc;

		rc = cafe_value_store(inst->workarea_type,
				      inst->workarea->u.data,
				      settings[i].name,
				      &settings[i].value,
				      inst->allocator);
		if (rc != CAFE_OK)
			return rc;
	}

	return process_settings(inst);
}


static
int fstream_reader_cfg_stop_dispatch(struct cafe_ep_inst *inst)
{
	struct client_data *node = inst->client_data;
	int rc;

	CAFE_LOCK();
	rc = cafe_ep_worker_stop(node->worker);
	CAFE_UNLOCK();
	return rc;
}


static int od_set0[] = {
	ODATA_SEQUENCE
};


static int od_set1[] = {
	ODATA_CLIENT_DATA
};


static struct cafe_ep_spec_output_od_set od_sets[] = {
	{ 0, NULL },
	{ CAFE_NELEMENTS(od_set0), od_set0 },
	{ CAFE_NELEMENTS(od_set1), od_set1 }
};


static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT3(file_stream_reader_range_end_VT_REF, NULL, 1),
	CAFE_OUTEV_INIT3(file_stream_reader_range_start_VT_REF, NULL, 1),
	CAFE_OUTEV_INIT3(file_stream_reader_start_VT_REF, NULL, 1),
	CAFE_OUTEV_INIT3(file_stream_reader_start_dispatch_VT_REF, NULL, 2),
	CAFE_OUTEV_INIT3(file_stream_reader_end_of_file_VT_REF, NULL, 0),
	CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};


static const char *orthogonal_data[] = {
	file_stream_reader_sequence_VT_REF,
	file_stream_reader_client_data_VT_REF
};


#define ALL_TYPES				\
	cafe_output_inst_thread_VT_DEF		\
	cafe_wildcard_VT_DEF			\
	cafe_setting_VT_DEF			\
	FILE_STREAM_READER_TYPE_ALL_VT_DEFS


static struct cafe_ep_spec spec = {
	"file_stream_reader",
	fstream_reader_cfg_create_inst,
	fstream_reader_cfg_delete_inst,
	fstream_reader_cfg_start,
	fstream_reader_cfg_stop,
	NULL,
	fstream_reader_cfg_update_output_event_actions,
	NULL,			/* inject event */
	ALL_TYPES,
	file_stream_reader_workarea_VT_REF,
	0,			/* n_input_events */
	NULL,			/* input event list */
	CAFE_NELEMENTS(output_events),
	output_events,
	CAFE_NELEMENTS(orthogonal_data),
	orthogonal_data,
	fstream_reader_cfg_settings_commit,
	NULL,                    /* input_state_change */
	CAFE_NELEMENTS(od_sets), /* n_output_odata_sets */
	od_sets,                 /* output_odata_sets */
	{ NULL, NULL },          /* spec_list */
	fstream_reader_cfg_discover_events,
	fstream_reader_cfg_stop_dispatch
};


static bool cfg_registered = false;


void cafe_fstream_reader_cfg_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}


int cafe_fstream_reader_cfg_register(void)
{
	int rc;

	if (cfg_registered)
		return CAFE_OK;

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		cfg_registered = true;

	return rc;
}
