/* file_stream_writer_cfg.c - Config used to tee events to a file */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

#include <cafe/stdio.h>
#include <string.h>
#include "../src/cafe_assert.h"
#include <cafe/cafe.h>
#include <cafe/list.h>
#include <cafe/value_type.h>
#include <cafe/log_stream.h>
#include <cafe/ep/file_stream_writer.h>
#include <cafe/dict.h>


struct fstream_workspace {
	struct cafe_event_descr *metadata_event;
	const char *filename;
	bool unbuffered;
	FILE *out;
	struct cafe_output_stream *ostream;
	struct cafe_slog_encoded_inst *flog_inst;
	unsigned int stream_version;
};


static
int cafe_fstream_writer_cfg_inst_create(struct cafe_ep_inst *inst)
{
	int rc;
	struct fstream_workspace *w =
		(struct fstream_workspace *) inst->workarea->u.data;
	struct cafe_value_type *metadata_type;

	rc = cafe_type_parse(inst->session->ns,
			     cafe_metadata_VT_REF, &metadata_type);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_cfg_get_event(inst->session, 1, &metadata_type,
				&w->metadata_event);
	if (rc != CAFE_OK)
		return rc;

	w->out = fopen(w->filename, "wb");
	if (w->out == 0)
		return CAFE_ERROR;

	rc = cafe_filestream_open_output(w->out, inst->allocator, &w->ostream);
	if (rc != CAFE_OK) {
		fclose(w->out);
		return rc;
	}

	rc = cafe_slog_encoded_create(inst->session, w->ostream, &w->flog_inst);
	if (rc != CAFE_OK) {
		fclose(w->out);
		CAFE_FREE(inst->allocator, w->ostream);
		return rc;
	}

	w->stream_version = cafe_slog_version;
	return CAFE_OK;
}


static
void cafe_fstream_writer_cfg_inst_delete(struct cafe_ep_inst *inst)
{
	struct fstream_workspace *w =
		(struct fstream_workspace *) inst->workarea->u.data;
	cafe_slog_encoded_delete(w->flog_inst);
	CAFE_FREE(inst->allocator, w->ostream);
	(void)fclose(w->out);
}


static
int cafe_fstream_writer_cfg_stop(struct cafe_ep_inst *inst)
{
	struct fstream_workspace *w =
		(struct fstream_workspace *) inst->workarea->u.data;

	(void)fflush_unlocked(w->out);
	cafe_ep_inst_state_stopped(inst);
	return CAFE_OK;
}


static
void fstream_log_event(struct cafe_event_action_info *event_info,
		      void **event_data)
{
	struct cafe_ep_inst *inst = event_info->inst;
	struct fstream_workspace *w =
		(struct fstream_workspace *) inst->workarea->u.data;

	/* log the event to our file */
	if (!cafe_slog_encoded_write(w->flog_inst, event_info->event,
				     event_data)) {
		cafe_dispatch_error(event_info->inst, CAFE_ERROR);
		return;
	}
	if (w->unbuffered) {
		(void)fflush_unlocked(w->out);
	}
}


static int cafe_fstream_writer_cfg_get_input_event_actions(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *event_info;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);
	if (rc != CAFE_OK || event_info == NULL)
		return rc;
	rc = cafe_add_event_action(inst, actions, fstream_log_event,
				   event_info);
	if (rc != CAFE_OK)
		return rc;
	return cafe_ep_inst_get_sink_event_actions(inst, actions);
}


const char *fstream_log_event_TYPE[] = {cafe_wildcard_VT_REF};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT(fstream_log_event)
};


static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};


#define PARAM_TYPES   "v[filename,s," cafe_setting_VT_NAT "]"\
					  "v[unbuffered,?," cafe_setting_VT_NAT ",0]"

static struct cafe_ep_spec spec = {
	"file_stream_writer",
	cafe_fstream_writer_cfg_inst_create,
	cafe_fstream_writer_cfg_inst_delete,
	NULL,
	cafe_fstream_writer_cfg_stop,
	cafe_fstream_writer_cfg_get_input_event_actions,
	NULL,
	NULL,
	cafe_setting_VT_DEF cafe_wildcard_VT_DEF cafe_metadata_VT_DEF,
	"S[file_stream_writer.workarea,v[metadata_event,L]"
	PARAM_TYPES
	"v[out,L]v[ostream,L]v[flog_inst,L]"
	"v[stream_version,I]]",
	CAFE_NELEMENTS(input_events),
	input_events,
	CAFE_NELEMENTS(output_events),
	output_events
};


static bool cfg_registered = false;


void cafe_fstream_writer_cfg_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}


int cafe_fstream_writer_cfg_register(void)
{
	int rc;

	if (cfg_registered)
		return CAFE_OK;

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		cfg_registered = true;

	return rc;
}
