/* Copyright (c) 2009-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <cafe/cafe.h>
#include <cafe/dict.h>
#include <cafe/ep/ostimer.h>
#include <cafe/ep/procfs_cfg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <stdlib.h>


/* macros */
#define BUF_SIZE        (64*1024)
#define STAT_FILE       "/proc/stat"
#define UPTIME_FILE     "/proc/uptime"
#define LOADAVG_FILE    "/proc/loadavg"
#define MEMINFO_FILE    "/proc/meminfo"
#define INTERRUPT_FILE  "/proc/interrupts"
#define NETDEV_FILE     "/proc/net/dev"

/* events */
#define PROCFS_LOADAVG_EVT_DEF "S[loadavg.info,v[avg1,d]v[avg5,d]v[avg15,d]]"
#define PROCFS_LOADAVG_EVT_REF "t[loadavg.info]"

#define PROCFS_UPTIME_EVT_DEF "S[uptime.info,v[up,d]v[idle,d]]"
#define PROCFS_UPTIME_EVT_REF "t[uptime.info]"

#define PROCFS_LPFR_DEF "S[lpfr,v[fd_stat,i]v[fd_uptime,i]v[fd_loadavg,i]" \
                        "v[fd_meminfo,i]v[fd_intr,i]v[fd_netdev,i]" \
                        "v[cpu_avg,t[stat_info]]" \
                        "v[avg,t[loadavg.info]]v[uptime,t[uptime.info]]" \
                        "v[net_stats,t[net_stats]]v[mem,t[mem_info]]]"
#define PROCFS_LPFR_REF "t[lpfr]"

#define PROCFS_CPU_STAT_DEF "S[cpu_stat,v[time_user,Q]v[time_low,Q]v[time_sys,Q]v[time_idle,Q]v[time_iowait,Q]v[time_irq,Q]v[time_softirq,Q]v[time_steal,Q]]"
#define PROCFS_CPU_STAT_REF "t[cpu_stat]"
#define PROCFS_CPU_DICT_ROW_DEF "S[cpu_dict_row,k[key,s]v[value,t[cpu_stat]]]"

#define PROCFS_STAT_INFO_EVT_DEF "S[stat_info,v[cpu_avg,t[cpu_stat]]v[cpus,D[t[cpu_dict_row]]]v[ctxt_switch,Q]v[interrupt,Q]v[page_in,L]v[page_out,L]v[swap_in,L]v[swap_out,L]v[boot_time,I]v[boot_procs,I]v[procs_running,I]v[procs_blocked,I]]"
#define PROCFS_STAT_INFO_EVT_REF "t[stat_info]"

#define PROCFS_NET_DEV_STATS_DEF "S[net_device_stats,v[rx_bytes,L]v[rx_packets,L]v[rx_errs,L]v[rx_drops,L]v[rx_fifo,L]v[rx_frame,L]v[rx_compressed,L]v[rx_multi,L]v[tx_bytes,L]v[tx_packets,L]v[tx_errs,L]v[tx_drops,L]v[tx_fifo,L]v[tx_colls,L]v[tx_carrier,L]v[tx_compressed,L]]"
#define PROCFS_NET_STATS_ROW_DEF "S[net_stats_row,k[key,s]v[value,t[net_device_stats]]]"
#define PROCFS_NET_STATS_EVT_DEF "S[net_stats,v[stats,D[t[net_stats_row]]]]"
#define PROCFS_NET_STATS_EVT_REF "t[net_stats]"

#define PROCFS_MEM_INFO_DEF "S[mem_info,v[kb_active,L]v[kb_anon_pages,L]v[kb_bounce,L]v[kb_main_buffers,L]v[kb_main_cached,L]v[kb_commit_limit,L]v[kb_committed_as,L]v[kb_dirty,L]v[kb_high_free,L]v[kb_high_total,L]v[kb_inactive,L]v[kb_low_free,L]v[kb_low_total,L]v[kb_mapped,L]v[kb_main_free,L]v[kb_main_total,L]v[kb_pagetables,L]v[kb_swap_reclaimable,L]v[kb_swap_unreclaimable,L]v[kb_slab,L]v[kb_swap_cached,L]v[kb_swap_free,L]v[kb_swap_total,L]v[kb_vmalloc_chunk,L]v[kb_vmalloc_total,L]v[kb_vmalloc_used,L]v[kb_writeback,L]]"

#define PROCFS_MEM_INFO_REF "t[mem_info]"

/* /proc/stat */
struct cpu_stat {
	unsigned long long time_user;
	unsigned long long time_low;
	unsigned long long time_sys;
	unsigned long long time_idle;

	unsigned long long time_iowait;
	unsigned long long time_irq;
	unsigned long long time_softirq;
	unsigned long long time_steal;
};

struct cpu_dict_row {
	char *key;
	struct cpu_stat value;
};

struct stat_info {
	/* in units of USER_HZ */
	struct cpu_stat cpu_avg;
	struct cafe_dict_struct cpus;

	unsigned long long ctxt_switch;
	unsigned long long interrupt;

	unsigned long page_in;
	unsigned long page_out;
	unsigned long swap_in;
	unsigned long swap_out;

	unsigned int boot_time;    /* in seconds */
	unsigned int boot_procs;

	unsigned int procs_running;
	unsigned int procs_blocked;
};

/* /proc/loadavg */
struct loadavg_info{
	double avg1;
	double avg5;
	double avg15;
};

/* /proc/uptime */
struct uptime_info {
	double up;
	double idle;
};

/* /proc/net/dev, refer to: kernel:net/core/dev.c */
struct net_device_stats {
	unsigned long rx_bytes;
	unsigned long rx_packets;
	unsigned long rx_errs;
	unsigned long rx_drops;
	unsigned long rx_fifo;
	unsigned long rx_frame;
	unsigned long rx_compressed;
	unsigned long rx_multi;

	unsigned long tx_bytes;
	unsigned long tx_packets;
	unsigned long tx_errs;
	unsigned long tx_drops;
	unsigned long tx_fifo;
	unsigned long tx_colls;
	unsigned long tx_carrier;
	unsigned long tx_compressed;
};

struct net_stats_row {
	char *key;
	struct net_device_stats value;
};

struct net_stats {
	struct cafe_dict_struct stats;
};

/* /proc/meminfo */
struct mem_tbl_item {
	const char *name;
	unsigned long offset;
};


struct mem_info {
	unsigned long kb_active;
	unsigned long kb_anon_pages;
	unsigned long kb_bounce;
	unsigned long kb_main_buffers;
	unsigned long kb_main_cached;
	unsigned long kb_commit_limit;
	unsigned long kb_committed_as;
	unsigned long kb_dirty;
	unsigned long kb_high_free;
	unsigned long kb_high_total;
	unsigned long kb_inactive;
	unsigned long kb_low_free;
	unsigned long kb_low_total;
	unsigned long kb_mapped;
	unsigned long kb_main_free;
	unsigned long kb_main_total;
	unsigned long kb_pagetables;
	unsigned long kb_swap_reclaimable;
	unsigned long kb_swap_unreclaimable;
	unsigned long kb_slab;
	unsigned long kb_swap_cached;
	unsigned long kb_swap_free;
	unsigned long kb_swap_total;
	unsigned long kb_vmalloc_chunk;
	unsigned long kb_vmalloc_total;
	unsigned long kb_vmalloc_used;
	unsigned long kb_writeback;
};

#define offset_entry(_type, _member) \
	(unsigned long)&(((_type* )(0))->_member)

#define offset_mem_info(_member) offset_entry(struct mem_info, _member)

/* The list of strings must be ordered for binary search */
static const struct mem_tbl_item mem_tbl[] = {
	{"Active",          offset_mem_info(kb_active)},
	{"AnonPages",       offset_mem_info(kb_anon_pages)},
	{"Bounce",          offset_mem_info(kb_bounce)},
	{"Buffers",         offset_mem_info(kb_main_buffers)},
	{"Cached",          offset_mem_info(kb_main_cached)},
	{"CommitLimit",     offset_mem_info(kb_commit_limit)},
	{"Committed_AS",    offset_mem_info(kb_committed_as)},
	{"Dirty",           offset_mem_info(kb_dirty)},
	{"HighFree",        offset_mem_info(kb_high_free)},
	{"HighTotal",       offset_mem_info(kb_high_total)},
	{"Inactive",        offset_mem_info(kb_inactive)},
	{"LowFree",         offset_mem_info(kb_low_free)},
	{"LowTotal",        offset_mem_info(kb_low_total)},
	{"Mapped",          offset_mem_info(kb_mapped)},
	{"MemFree",         offset_mem_info(kb_main_free)},
	{"MemTotal",        offset_mem_info(kb_main_total)},
	{"PageTables",      offset_mem_info(kb_pagetables)},
	{"SReclaimable",    offset_mem_info(kb_swap_reclaimable)},
	{"SUnreclaim",      offset_mem_info(kb_swap_unreclaimable)},
	{"Slab",            offset_mem_info(kb_slab)},
	{"SwapCached",      offset_mem_info(kb_swap_cached)},
	{"SwapFree",        offset_mem_info(kb_swap_free)},
	{"SwapTotal",       offset_mem_info(kb_swap_total)},
	{"VmallocChunk",    offset_mem_info(kb_vmalloc_chunk)},
	{"VmallocTotal",    offset_mem_info(kb_vmalloc_total)},
	{"VmallocUsed",     offset_mem_info(kb_vmalloc_used)},
	{"Writeback",       offset_mem_info(kb_writeback)},
};

struct lpfr {
	/* fd caching */
	int fd_stat;
	int fd_uptime;
	int fd_loadavg;
	int fd_meminfo;
	int fd_intr;
	int fd_netdev;

	/* status */
	struct stat_info stats;
	struct loadavg_info avg;
	struct uptime_info uptime;
	struct net_stats net_stats;
	struct mem_info mem;
};

static
bool procfs_cpu_dict_add(struct cafe_dict_struct *dict, const char *key,
		struct cpu_stat *value);
static
bool procfs_net_dict_add(struct cafe_dict_struct *dict, const char *key,
		struct net_device_stats *value);
static
void lpfr_deinit(struct lpfr *t)
{
	if (t->fd_stat)
		close(t->fd_stat);

	if (t->fd_uptime)
		close(t->fd_uptime);

	if (t->fd_loadavg)
		close(t->fd_loadavg);

	if (t->fd_meminfo)
		close(t->fd_meminfo);

	if (t->fd_intr)
		close(t->fd_intr);

	if (t->fd_netdev)
		close(t->fd_netdev);
}

static
int lpfr_get_stat(struct lpfr *t)
{
	char buf[BUF_SIZE];
	const char *b;
	struct stat_info *stats = &t->stats;
	ssize_t ret;

	if (t->fd_stat > 0)
		lseek(t->fd_stat, 0L, SEEK_SET);
	else {
		t->fd_stat = open(STAT_FILE, O_RDONLY, 0);
		if (t->fd_stat == -1)
			return CAFE_ERROR;
	}

	ret = read(t->fd_stat, buf, BUF_SIZE - 1);
	if (ret == -1) {
		perror("Error reading stat");
		return CAFE_ERROR;
	}
	buf[ret] = 0;

	b = strstr(buf, "cpu ");
	if (b) {
		struct cpu_stat *cpu = &stats->cpu_avg;
		sscanf(b, "cpu  %Lu %Lu %Lu %Lu %Lu %Lu %Lu %Lu",
				&cpu->time_user, &cpu->time_low,
				&cpu->time_sys, &cpu->time_idle,
				&cpu->time_iowait, &cpu->time_irq,
				&cpu->time_softirq, &cpu->time_steal);

		/* read more info about each cpu */
		for(;;) {
			char n[3];
			struct cpu_stat s;
			b += 4;
			b = strstr(b, "cpu");
			if (b) {
				sscanf(b, "cpu%2s %Lu %Lu %Lu %Lu %Lu %Lu %Lu %Lu",
						n,
						&s.time_user, &s.time_low,
						&s.time_sys, &s.time_idle,
						&s.time_iowait, &s.time_irq,
						&s.time_softirq, &s.time_steal);
				procfs_cpu_dict_add(&stats->cpus, n, &s);
			} else
				break;
		}
	}

	b = strstr(buf, "page ");
	if (b)
		sscanf(b, "page %lu %lu", &stats->page_in, &stats->page_out);

	b = strstr(buf, "swap ");
	if (b)
		sscanf(b, "swap %lu %lu", &stats->swap_in, &stats->swap_out);

	b = strstr(buf, "intr ");
	if (b)
		sscanf(b, "intr %Lu", &stats->interrupt);

	b = strstr(buf, "ctxt ");
	if (b)
		sscanf(b, "ctxt %Lu", &stats->ctxt_switch);

	b = strstr(buf, "btime ");
	if (b)
		sscanf(b, "btime %u", &stats->boot_time);

	b = strstr(buf, "processes ");
	if (b)
		sscanf(b, "processes %u", &stats->boot_procs);

	b = strstr(buf, "procs_running ");
	if (b)
		sscanf(b, "procs_running %u", &stats->procs_running);

	b = strstr(buf, "procs_blocked ");
	if (b)
		sscanf(b, "procs_blocked %u", &stats->procs_blocked);

	return CAFE_OK;
}

static
int compare_mem_tbl_item(const void *key, const void *item)
{
	const char *s1 = key;
	const char *s2 = ((struct mem_tbl_item *) item)->name;

	return strcmp(s1, s2);
}

static
int lpfr_get_meminfo(struct lpfr *t)
{
	char buf[BUF_SIZE];
	const int mem_tbl_count = sizeof(mem_tbl)/sizeof(struct mem_tbl_item);
	ssize_t ret;
	char *head;
	char *tail;
	struct mem_tbl_item *found;

	if (t->fd_meminfo > 0)
		lseek(t->fd_meminfo, 0L, SEEK_SET);
	else {
		t->fd_meminfo = open(MEMINFO_FILE, O_RDONLY, 0);
		if (t->fd_meminfo == -1)
			return CAFE_ERROR;
	}

	ret = read(t->fd_meminfo, buf, BUF_SIZE - 1);
	if (ret == -1) {
		perror("Error reading meminfo");
		return CAFE_ERROR;
	}
	buf[ret] = 0;

	head = buf;
	for (;;) {
		tail = strchr(head, ':');
		if (!tail)
			break;

		*tail = '\0';
		found = bsearch(head, mem_tbl, mem_tbl_count,
				sizeof(struct mem_tbl_item), compare_mem_tbl_item);
		head = tail + 1;
		if (!found)
			goto nextline;

		//set value by offset
		*(unsigned long *)((char *)(&t->mem) + found->offset) =
			strtoul(head, &tail, 10);
nextline:
		tail = strchr(head, '\n');
		if (!tail)
			break;

		head = tail + 1;
	}

	return CAFE_OK;
}

static
int lpfr_get_loadavg(struct lpfr *t)
{
	char buf[BUF_SIZE];
	struct loadavg_info *avg = &t->avg;
	ssize_t ret;

	if (t->fd_loadavg > 0)
		lseek(t->fd_loadavg, 0L, SEEK_SET);
	else {
		t->fd_loadavg = open(LOADAVG_FILE, O_RDONLY, 0);
		if (t->fd_loadavg == -1)
			return CAFE_ERROR;
	}

	ret = read(t->fd_loadavg, buf, BUF_SIZE - 1);
	if (ret == -1) {
		perror("Error reading loadavg");
		return CAFE_ERROR;
	}
	buf[ret] = 0;

	if (sscanf(buf, "%lf %lf %lf", &avg->avg1, &avg->avg5, &avg->avg15) < 3)
		return CAFE_ERROR;

	return CAFE_OK;
}

static
int lpfr_get_uptime(struct lpfr *t)
{
	char buf[BUF_SIZE];
	struct uptime_info *up = &t->uptime;
	char *savelocale;
	ssize_t ret;

	if (t->fd_uptime > 0)
		lseek(t->fd_uptime, 0L, SEEK_SET);
	else {
		t->fd_uptime = open(UPTIME_FILE, O_RDONLY, 0);
		if (t->fd_uptime == -1)
			return CAFE_ERROR;
	}

	ret = read(t->fd_uptime, buf, BUF_SIZE - 1);
	if (ret == -1) {
		perror("Error reading uptime");
		return CAFE_ERROR;
	}
	buf[ret] = 0;

	savelocale = setlocale(LC_NUMERIC, NULL);
	setlocale(LC_NUMERIC, "C");
	if (sscanf(buf, "%lf %lf", &up->up, &up->idle) < 2)
		return CAFE_ERROR;
	setlocale(LC_NUMERIC, savelocale);

	return CAFE_OK;
}

static
inline char *skipline(char *p)
{
	while (*p && *p != '\n') p++;

	return ++p;
}

static
int lpfr_get_net_stats(struct lpfr *t)
{
	struct net_device_stats s;
	char name[16];
	char *begin;
	char buf[BUF_SIZE];
	ssize_t ret;
	char *p;

	if (t->fd_netdev > 0)
		lseek(t->fd_netdev, 0L, SEEK_SET);
	else {
		t->fd_netdev = open(NETDEV_FILE, O_RDONLY, 0);
		if (t->fd_netdev == -1)
			return CAFE_ERROR;
	}

	ret = read(t->fd_netdev, buf, BUF_SIZE - 1);
	if (ret == -1) {
		perror("Error reading net stats");
		return CAFE_ERROR;
	}
	buf[ret] = 0;

	/* parsing */
	p = buf;
	p = skipline(p);
	p = skipline(p);

	do {
		begin  = strchr(p, ':');
		if (begin) {
			*begin++ = 0;
			snprintf(name, 15, "%s", p);
			p = begin;
			if(sscanf(p, "%lu %lu %lu %lu %lu %lu %lu %lu"
						" %lu %lu %lu %lu %lu %lu %lu %lu",
						&s.rx_bytes, &s.rx_packets, &s.rx_errs, &s.rx_drops,
						&s.rx_fifo, &s.rx_frame, &s.rx_compressed, &s.rx_multi,
						&s.tx_bytes, &s.tx_packets, &s.tx_errs, &s.tx_drops,
						&s.tx_fifo, &s.tx_colls, &s.tx_carrier, &s.tx_compressed
					 ) == 16) {
				procfs_net_dict_add(&t->net_stats.stats, name, &s);
				p = skipline(p);
			}
		}
	} while(begin);

	return CAFE_OK;
}

static const char *timer_event_TYPE[] = {OSTIMER_REF};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT(timer_event)
};

static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(PROCFS_LOADAVG_EVT_REF),
	CAFE_OUTEV_INIT(PROCFS_UPTIME_EVT_REF),
	CAFE_OUTEV_INIT(PROCFS_STAT_INFO_EVT_REF),
	CAFE_OUTEV_INIT(PROCFS_NET_STATS_EVT_REF),
	CAFE_OUTEV_INIT(PROCFS_MEM_INFO_REF),
};

struct procfs_inst_data {
	unsigned long output_event_action_change_count;
	cafe_eal_handle_t sink_actions[CAFE_NELEMENTS(output_events)];
};

static
bool procfs_cpu_dict_add(struct cafe_dict_struct *dict, const char *key,
		struct cpu_stat *value)
{
	cafe_any_value keys[1];
	struct cafe_value_data *val;
	struct cpu_dict_row *row;

	keys[0].str = (char *) key;

	val = cafe_dict_get(dict, keys);
	if (val == NULL)
		return false;

	row = (struct cpu_dict_row *)val->u.data;
	memcpy(&row->value, value, sizeof(struct cpu_stat));

	return true;
}

static
bool procfs_net_dict_add(struct cafe_dict_struct *dict, const char *key,
		struct net_device_stats *value)
{
	cafe_any_value keys[1];
	struct cafe_value_data *val;
	struct net_stats_row *row;

	keys[0].str = (char *) key;

	val = cafe_dict_get(dict, keys);
	if (val == NULL)
		return false;

	row = (struct net_stats_row *)val->u.data;
	memcpy(&row->value, value, sizeof(struct net_device_stats));

	return true;
}

static
int send_event(struct cafe_ep_inst *inst, unsigned int type_idx, void *data)
{
	int rc;
	struct procfs_inst_data *inst_data = inst->client_data;
	void *output_event_data[2];
	struct cafe_event_action_list *action_list;
	cafe_eal_handle_t unlock_handle;

	rc = cafe_event_action_list_read_lock(&inst_data->sink_actions[type_idx],
			&unlock_handle, &action_list);
	if (rc != CAFE_OK)
		return rc;
	output_event_data[0] = data;

	cafe_dispatch_event_unlocked(action_list, output_event_data);
	cafe_event_action_list_read_unlock(unlock_handle);
	return CAFE_OK;
}

static
void procfs_send_loadavg_event(struct cafe_ep_inst *inst)
{
	struct lpfr *t = (struct lpfr *)inst->workarea->u.data;
	void *args[1];

	if (lpfr_get_loadavg(t) == CAFE_OK) {
		args[0] = &t->avg;
		send_event(inst, 0, args[0]);
	}
}

static
void procfs_send_uptime_event(struct cafe_ep_inst *inst)
{
	struct lpfr *t = (struct lpfr *)inst->workarea->u.data;
	void *args[1];

	if (lpfr_get_uptime(t) == CAFE_OK) {
		args[0] = &t->uptime;
		send_event(inst, 1, args[0]);
	}
}

static
void procfs_send_statinfo_event(struct cafe_ep_inst *inst)
{
	struct lpfr *t = (struct lpfr *)inst->workarea->u.data;
	void *args[1];

	if (lpfr_get_stat(t) == CAFE_OK) {
		args[0] = &t->stats;
		send_event(inst, 2, args[0]);
	}
}

static
void procfs_send_netstat_event(struct cafe_ep_inst *inst)
{
	struct lpfr *t = (struct lpfr *)inst->workarea->u.data;
	void *args[1];

	if (lpfr_get_net_stats(t) == CAFE_OK) {
		args[0] = &t->net_stats;
		send_event(inst, 3, args[0]);
	}
}

static
void procfs_send_meminfo_event(struct cafe_ep_inst *inst)
{
	struct lpfr *t = (struct lpfr *)inst->workarea->u.data;
	void *args[1];

	if (lpfr_get_meminfo(t) == CAFE_OK) {
		args[0] = &t->mem;
		send_event(inst, 4, args[0]);
	}
}

static
void procfs_request_handler(struct cafe_event_action_info *event_info,
		void *event_data[])
{
	procfs_send_loadavg_event(event_info->inst);
	procfs_send_uptime_event(event_info->inst);
	procfs_send_statinfo_event(event_info->inst);
	procfs_send_netstat_event(event_info->inst);
	procfs_send_meminfo_event(event_info->inst);
}

static
int procfs_inst_create(struct cafe_ep_inst *inst)
{
	struct procfs_inst_data *inst_data;

	if ((inst_data = CAFE_ALLOC(inst->allocator, sizeof *inst_data)) == NULL)
		return CAFE_NOMEM;

	memset(inst_data, 0, sizeof *inst_data);
	inst_data->output_event_action_change_count =
		inst->session->output_event_action_change_count - 1;
	inst->client_data = inst_data;

	return CAFE_OK;
}

static
void procfs_inst_delete(struct cafe_ep_inst *inst)
{
	struct procfs_inst_data *inst_data = inst->client_data;
	struct lpfr *t = (struct lpfr *)inst->workarea->u.data;
	int i;

	for (i = 0; i < CAFE_NELEMENTS(output_events); ++i)
		cafe_event_action_list_release(&inst_data->sink_actions[i]);

	CAFE_FREE(inst->allocator, inst_data);
	lpfr_deinit(t);
}

static
int procfs_cfg_get_input_event_actions(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *event_info;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);
	if (rc != CAFE_OK || event_info == NULL)
		return rc;

	return cafe_add_event_action(
			inst, actions, procfs_request_handler, event_info);
}

static
int procfs_cfg_update_output_event_actions(struct cafe_ep_inst *inst)
{
	struct procfs_inst_data *inst_data = inst->client_data;
	int rc;
	unsigned int i;

	if (inst_data->output_event_action_change_count ==
			inst->session->output_event_action_change_count)
		return CAFE_OK;
	inst_data->output_event_action_change_count =
		inst->session->output_event_action_change_count;

	for (i = 0; i < inst->enabled_output_events.ind; ++i) {
		struct cafe_event_descr *event =
			inst->enabled_output_events.list[i].event;
		unsigned long idx = (unsigned long)
			inst->enabled_output_events.list[i].impl_data;
		struct cafe_event_action_list *sink_actions;
		cafe_eal_handle_t sink_actions_handle;

		rc = cafe_event_action_list_alloc(inst, event,
				&sink_actions,
				&sink_actions_handle);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sink_actions_handle);
			return rc;
		}
		cafe_event_action_list_swap(&inst_data->sink_actions[idx],
				sink_actions_handle);
	}
	return CAFE_OK;
}

static struct cafe_ep_spec spec = {
	"procfs",
	procfs_inst_create,
	procfs_inst_delete,
	NULL,
	NULL,
	procfs_cfg_get_input_event_actions,
	procfs_cfg_update_output_event_actions,
	NULL,
	PROCFS_MEM_INFO_DEF
	PROCFS_NET_DEV_STATS_DEF
	PROCFS_NET_STATS_ROW_DEF
	PROCFS_NET_STATS_EVT_DEF
	PROCFS_CPU_DICT_ROW_DEF
	PROCFS_CPU_STAT_DEF
	PROCFS_STAT_INFO_EVT_DEF
	OSTIMER_DEF
	PROCFS_UPTIME_EVT_DEF
	PROCFS_LOADAVG_EVT_DEF
	PROCFS_LPFR_DEF,
	PROCFS_LPFR_REF,
	CAFE_NELEMENTS(input_events),
	input_events,
	CAFE_NELEMENTS(output_events),
	output_events
};

static bool cfg_registered = false;

int procfs_cfg_register(void)
{
	unsigned int i;
	int rc;

	if (cfg_registered)
		return CAFE_OK;

	for (i = 0; i < CAFE_NELEMENTS(output_events); ++i)
		output_events[i].impl_data = (void *)(unsigned long)i;

	rc = cafe_ep_spec_register(&spec);

	if (rc == CAFE_OK)
		cfg_registered = true;

	return rc;
}

void procfs_cfg_unregister(void)
{
	if (cfg_registered) {
		cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}
