/* symresolve_actions.c - symbol resolution engine */

/*
 * Copyright (c) 2011-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
18nov14,tle  Emit entry for unnamed object files (WB4-4663)
10nov14,tle  Fix resolution in files with multiple program segments (WB4-4590)
07nov14,tle  Emit entries for unresolved symbols and object files (US48740)
29oct14,tle  Fix symbol resolution for RTP events before rtpLoad (WB4-4383)
08oct14,tle  Fix multiple symbols for same location (WB4-4180)
29sep14,bpn  Fix call to elf_map_to_link_time_address() for new version of
             TCF opensource code (tcf-c-core commit 6052f2c).
09sep14,tle  Fix bogus symbol lookup if size does not match (WB4-4158)
03aug14,sdt  Fixed checked-return warnings (US45196)
06aug14,tle  Fix crash if symbol cannot be resolved (WB4-3977)
11jun14,tle  Fix lookup of 0-length symbols (WB4-3683)
             Fix symbol resolution for HV images (WB4-3716)
16may14,tle  Fix line info lookup for RTPs (WB4-3499)
15may14,sdt  Fix 'checked return' warning
24feb14,sdt  Disable unexpected NO_EFFECT warnings
21feb14,sdt  Fix compiler warnings

DESCRIPTION

*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#if CAFE_OS_UNIX || CAFE_OS_WINDOWS
#  include <sys/types.h>
#  include <sys/stat.h>
#endif
#if defined(HAVE_CAFE_CONFIG_H)
#  include <cafe/config.h>
#endif
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/array.h>
#include <cafe/ptr_array.h>
#include <cafe/byte_buffer.h>
#include <cafe/hashtbl.h>
#include <cafe/inttypes.h>
#include <cafe/utils.h>
#include <cafe/ep/config_service.h>
#if defined(HAVE_TCF)
#  include <tcf/framework/channel.h>
#  include <tcf/framework/myalloc.h>
#  if defined(ENABLE_ContextMux) && ENABLE_ContextMux
#    include <tcf/framework/context-dispatcher.h>
#    include <tcf/services/linenumbers.h>
#  endif
#  include <tcf/services/memorymap.h>
#  include <tcf/services/tcf_elf.h>
#  include <tcf/services/pathmap.h>
#  include "tcf_utils.h"
#endif
#include "symresolve_actions.h"
#if defined(HAVE_DEMANGLER)
#   include "demangler.h"
#endif

extern long cafe_aggregator_debug;

static const unsigned int INITIAL_ADDRESS_SPACE_MAP_SIZE = 128;
static const unsigned int INITIAL_OBJECT_CACHE_SIZE = 128;
static const char *UNMAPPED_STR = "[unmapped]";
static const char *UNNAMED_STR = "[unknown object]";

/* Types of data possible to calculate for a given callstack entry. */
enum {
	RESOLVED_ADDRESS_FIELD_OBJECT_NAME = 0,
	RESOLVED_ADDRESS_FIELD_SYMBOL_INDEX,
	RESOLVED_ADDRESS_FIELD_LINEINFO_INDEX,
	RESOLVED_ADDRESS_FIELD_OBJECT_ADDRESS,
	RESOLVED_ADDRESS_FIELD_SYMBOL_NAME,
	RESOLVED_ADDRESS_FIELD_SYMBOL_OFFSET,
	RESOLVED_ADDRESS_FIELD_STACK_INDEX,
	RESOLVED_ADDRESS_FIELD_ORIGINAL_ADDRESS,
	RESOLVED_LINEINFO_FIELD_FILE_NAME,
	RESOLVED_LINEINFO_FIELD_LINE_NUMBER,
	RESOLVED_LINEINFO_FIELD_COLUMN_NUMBER
};


struct symresolve_actions {
	const struct cafe_allocator *allocator;
	struct cafe_value_type *delta_vt;
	struct cafe_value_type *no_delta_vt;
	struct cafe_value_type *delta_list_vt;

	struct cafe_dict_struct *symbol_index;
	struct cafe_dict_struct *lineinfo_index;
	unsigned long next_sym_index;
	unsigned long next_lineinfo_index;

	/* a hash of: (address space UID) -> (address space record) */
	struct cafe_hashtbl *address_space_map;

	/* a hash of: (object file name) -> (symbol cache) */
	struct cafe_hashtbl *object_cache;

	void (*error_handler) (const char *msg, void *arg);
	void *error_handler_arg;
#if defined(SYMRESOLVE_PROFILE_CACHE)
	unsigned int hits, misses, lhits, lmisses;
#endif
	struct symbol_cache_entry *unmapped_entry;
	struct symbol_cache_entry *unnamed_entry;
#if defined(HAVE_TCF)
	Channel *c;
	/* Handle to a context object used to provide the correct function
	   pointer interface for source line lookup. */
	Context *ctx;
	/* Listens to TCF path map changes to clear cache of unresolved object files */
	PathMapEventListener path_map_listener;
	/* Indicates whether path map changed since last object cache lookup */
	bool path_map_changed;
#endif
};


/* Represents an address space, and its memory map over time.  The address_uid
   should be unique and is used as a hash key.  The mm_generations array is an
   an array of memory maps, representing how the address space has changed over
   time; changes to the memory map are generally folded into the most recent
   generation, unless that generation has been referenced externally (by a 
   client querying for the current generation number). */
struct address_space_record {
	unsigned long long address_uid;
	struct cafe_array mm_generations;  /* array of memory_map_record */
};


/* Captures a single generation of an address space's memory map. */
struct memory_map_record {
	unsigned long long generation;
	unsigned long long parent_uid;
	unsigned long long parent_generation;
	bool committed;
	bool generated;
	struct cafe_array regions;    /* array of memory_map_region */
};


/* A cache of persistent objects */
struct pcache {
	struct cafe_ptr_array list;
	struct cafe_byte_buffer data;
};


/* Captures an individual memory region within a memory map */
struct memory_map_region {
	unsigned long long start_addr;
	unsigned long long len;
	unsigned long long file_offset;
	const char *file_name;
	struct symbol_cache *cache;
};


/* Represents a single object file cache entry; a function symbol */
struct symbol_cache_entry {
	uint64_t address;
	uint64_t size;
	unsigned long index;
	char name[0];
};


/* Represents a single lineinfo cache entry */
struct lineinfo_cache_entry {
	uint64_t address;
	uint64_t size;
	unsigned long line;
	unsigned long column;
	unsigned long index;
	char file[0];
};


/* Caches any information previously looked up calculated for a given object
   file */
struct symbol_cache {
	struct pcache syms;
	struct pcache lineinfo;
	unsigned int refs;
	bool is_mapped_text_section;
	char *remote_filename;
	int64_t mtime;
	/* Symbol entry for the symbol-not-resolved case */
	struct symbol_cache_entry *unresolved_entry;
	char local_filename[0];
};


/* Marker records are memory map entries where the len field is 0.
 * They are used to communicate special events in the delta table.
 * The start_address field is used to hold the MARKER_EVENT code.  The
 * file_offset and file_name fields are marker event specific
 * arguments and should be set to 0 and NULL, respectively, when not
 * used. */
enum {
	MARKER_len = 0
};
enum {
	MARKER_EVENT_ADDRESS_SPACE = 0,
	MARKER_EVENT_GENERATION = 1,
	MARKER_EVENT_PARENT_GENERATION = 2,
	MARKER_EVENT_CLEAR = 3,
	MARKER_EVENT_DELETE = 4
};


static void symbol_cache_dereference(struct memory_map_region *region,
				     struct symresolve_actions *actions);


#if defined(ENABLE_ContextMux) && ENABLE_ContextMux

static int symresolve_context_get_memory_map(Context *ctx, MemoryMap *map);
static Context *symresolve_context_get_group(Context *ctx, int group);

static
const struct lineinfo_cache_entry *lineinfo_cache_lookup(
	struct symresolve_actions *actions, struct symbol_cache *cache,
	uint64_t offs, unsigned long long address,
	struct memory_map_record *map, bool object_cache_updated);

static ContextIf SYMRESOLVE_CTX_IF = {0};


/* Should only be referenced from the TCF event thread */
static Channel *ctx_channel;
static LINK tcf_ctx_list = TCF_LIST_INIT(tcf_ctx_list);


/* Should only be referenced from the TCF event thread */
static struct memory_map_record *ctx_map;
#endif


#if defined(HAVE_TCF) && ENABLE_ELF
/**
 *  Initialize a new persistent cache object.
 */
static
int pcache_new(struct pcache *cache, unsigned int initial_entries,
	       unsigned initial_size, const struct cafe_allocator *allocator)
{
	int rc;

	assert(cache != NULL);
	cafe_ptr_array_init(&cache->list, allocator);
	rc = cafe_ptr_array_capacity_set(&cache->list, initial_entries);
	if (rc != CAFE_OK) {
		cafe_ptr_array_release(&cache->list);
		return rc;
	}
	rc = cafe_bbuf_init(&cache->data, allocator, initial_size);
	if (rc != CAFE_OK) {
		cafe_ptr_array_release(&cache->list);
		return rc;
	}

	return CAFE_OK;
}


/**
 *  Insert an object into the given persistent cache, at the given index. On
 * success, a pointer to a new region of "size" bytes is returned.
 */
static
void *pcache_insert(struct pcache *cache, unsigned int index, size_t size)
{
	char *slot;
	const char *base;
	int rc;

	assert(cache != NULL);
	assert(index <= cache->list.len);

	base = cache->data.data;
	rc = cafe_bbuf_reserve(&cache->data, size, &slot);
	if (rc != CAFE_OK)
		return NULL;

	if (base != cache->data.data) {
		unsigned int i;
		ptrdiff_t off = cache->data.data - base;
		for (i = 0; i < cache->list.len; ++i)
			cafe_ptr_array_val(&cache->list, i) =
				off + (char *) cafe_ptr_array_val(&cache->list,
								  i);
	}

	rc = cafe_ptr_array_capacity_set(&cache->list, ++cache->list.len);
	if (rc != CAFE_OK)
		return NULL;

	memmove(&cafe_ptr_array_val(&cache->list, index + 1),
		&cafe_ptr_array_val(&cache->list, index),
		(cache->list.len - index - 1) * sizeof(void *));
	cafe_ptr_array_val(&cache->list, index) = slot;
	return slot;
}

/**
 * TCF post_event() callback for adding a path map event listener.
 */
static
void tcf_add_path_map_event_listener(void *arg)
{
	struct symresolve_actions *actions = arg;
	add_path_map_event_listener(&actions->path_map_listener, actions);
}


/**
 * TCF post_event() callback for removing a path map event listener.
 */
static
void tcf_rem_path_map_event_listener(void *arg)
{
	PathMapEventListener *listener = arg;
	rem_path_map_event_listener(listener);
}


/**
 * Path map change listener. On path map changes, set a flag to delete all cache entries
 * of object files which were not found before.
 */
static
void tcf_path_mapping_changed(Channel *c, void *arg)
{
	struct symresolve_actions *actions = arg;

	/* Just set the flag, stale entries will be flushed
	 * on next invocation of object_cache_lookup
	 */
	actions->path_map_changed = true;
}
#endif  /* if defined(HAVE_TCF) && ENABLE_ELF */


/**
 *  Returns the current number of cache entries.
 */
static inline
unsigned int pcache_size(const struct pcache *cache)
{
	assert(cache != NULL);
	return (unsigned int) cache->list.len;
}


/**
 *  Release resouces claimed by the given pcache.
 */
static
void pcache_free(struct pcache *cache)
{
	assert(cache != NULL);

	cafe_ptr_array_release(&cache->list);
	cafe_bbuf_release(&cache->data);
}


/**
 *  Return cache entry at the given index.
 */
static inline
void *pcache_val(const struct pcache *cache, unsigned int index)
{
	assert(cache != NULL);
	assert(index < cache->list.len);
	return cafe_ptr_array_val(&cache->list, index);
}


#if !defined(NDEBUG)
void dump_delta(const struct cafe_vm_dynarray *delta);
void dump_delta(const struct cafe_vm_dynarray *delta)
{
	size_t i;
	struct cafe_mm_delta_entry *e;

	for (i = 0, e = (struct cafe_mm_delta_entry *)delta->p;
	     i < delta->l; i++, e++) {
		if (e->len == MARKER_len && e->start_addr == MARKER_EVENT_ADDRESS_SPACE) {
			printf("address_space %llu\n", e->file_offset);
		} else if (e->len == MARKER_len && e->start_addr == MARKER_EVENT_GENERATION) {
			printf("generation %llu\n", e->file_offset);
		} else if (e->len == MARKER_len && e->start_addr == MARKER_EVENT_PARENT_GENERATION) {
			printf("parent generation %llu\n", e->file_offset);
		} else if (e->len == MARKER_len && e->start_addr == MARKER_EVENT_CLEAR) {
			printf("clear %llu\n", e->file_offset);
		} else if (e->len == MARKER_len && e->start_addr == MARKER_EVENT_DELETE) {
			printf("delete\n");
		} else {
			printf("0x%08llx %llu %llu %s\n",
			       e->start_addr, e->len, e->file_offset, e->file_name);
		}
	}
}


void dump_symbol_cache(const struct symbol_cache *sc);
void dump_symbol_cache(const struct symbol_cache *sc)
{
	unsigned int i;

	printf("Symbol cache \"%s\" (%u)\n", sc->local_filename, sc->refs);
	for (i = 0; i < pcache_size(&sc->syms); ++i) {
		struct symbol_cache_entry *symentry = pcache_val(&sc->syms, i);
		printf("\t%u, %p: 0x%"PRIx64" - 0x%"PRIx64"\t%s\n",
		       i, symentry, symentry->address,
		       symentry->address + symentry->size, symentry->name);
		if (i > 0) {
			struct symbol_cache_entry *prev =
				pcache_val(&sc->syms, i - 1);
			assert(symentry->address > prev->address);
		}
	}
}


void dump_lineinfo_cache(const struct symbol_cache *sc);
void dump_lineinfo_cache(const struct symbol_cache *sc)
{
	unsigned int i;

	printf("Symbol cache \"%s\" (%u)\n", sc->local_filename, sc->refs);
	for (i = 0; i < pcache_size(&sc->lineinfo); ++i) {
		const struct lineinfo_cache_entry *lineinfo =
			pcache_val(&sc->lineinfo, i);
		printf("\t%u, %p: 0x%"PRIx64" - 0x%"PRIx64"\t%s  %lu\n",
		       i, lineinfo, lineinfo->address,
		       lineinfo->address + lineinfo->size, lineinfo->file,
		       lineinfo->line);
		if (i > 0) {
			struct lineinfo_cache_entry *prev =
				pcache_val(&sc->lineinfo, i - 1);
			assert(lineinfo->address > prev->address);
		}
	}
}
#endif  /* if !defined(NDEBUG) */


/**
 *  Construct a new empty memory map and append it to the given generation
 * array.  Note that this may move the location of array entries, so clients
 * should be sure to recalculate pointers to such as necessary.  Returns NULL
 * on error.
 */
static
struct memory_map_record *mm_new(struct cafe_array *generations,
				 unsigned long long address_uid)
{
	struct memory_map_record *mm;
	int rc;

	rc = cafe_array_capacity_set(generations, ++generations->len);
	if (rc != CAFE_OK)
		return NULL;

	mm = &cafe_array_val(generations, struct memory_map_record,
			     generations->len - 1);
	cafe_array_init(&mm->regions, generations->allocator,
			sizeof(struct memory_map_region));
	if (generations->len > 1) {
		struct memory_map_record *prev = &cafe_array_val(
			generations, struct memory_map_record,
			generations->len - 2);
		mm->generation = prev->generation + 1;
		mm->parent_uid = prev->parent_uid;

		/* if an address space has no parent (e.g., kernel address
		   space) then the parent generation should always be the
		   current generation */
		if (address_uid != mm->parent_uid)
			mm->parent_generation = prev->parent_generation;
		else
			mm->parent_generation = mm->generation;
	} else {
		mm->generation = 0;
		mm->parent_uid = 0;
		mm->parent_generation = 0;
	}
	mm->committed = false;
	mm->generated = false;
	return mm;
}


/**
 *  Remove all region mappings from this memory map.
 */
static
void mm_clear(struct memory_map_record *mm, struct symresolve_actions *actions)
{
	unsigned int i;

	assert(!mm->committed);
	for (i = 0; i < mm->regions.len; ++i) {
		struct memory_map_region *r =
			&cafe_array_val(&mm->regions, struct memory_map_region,
					i);
		symbol_cache_dereference(r, actions);
		CAFE_FREE(mm->regions.allocator, (char *)r->file_name);
	}
	mm->regions.len = 0;
}


/**
 *  Release all memory associated with the given memory map record.
 */
static
void mm_free(struct memory_map_record *mm, struct symresolve_actions *actions)
{
#if !defined(NDEBUG)
	mm->committed = false;
#endif
	mm_clear(mm, actions);
	cafe_array_release(&mm->regions);
}


static
void generations_clear(struct cafe_array *generations,
		       struct symresolve_actions *actions)
{
	unsigned int i;

	for (i = 0; i < generations->len; ++i) {
		struct memory_map_record *mm =
			&cafe_array_val(generations,
					struct memory_map_record, i);
		mm_free(mm, actions);
	}
	generations->len = 0;
}


static
void clear_uncommitted(struct cafe_array *generations,
		       struct symresolve_actions *actions)
{
	struct memory_map_record *mm;

	/* Clear uncommitted memmory map entries */
	while (generations->len > 0) {
		mm = &cafe_array_val(
			generations, struct memory_map_record,
			generations->len - 1);
		if (mm->committed)
			break;
		mm_free(mm, actions);
		generations->len--;
	}
}


/**
 *  Create an address space object, and initialize its initial (empty) memory
 * map.  Returns NULL on error.
 */
static
struct address_space_record *address_space_new(struct symresolve_actions *actions,
					       unsigned long long address_space_uid)
{
	struct address_space_record *asr =
		CAFE_ALLOC(actions->allocator, sizeof(*asr));
	if (asr == NULL)
		return NULL;

	cafe_array_init(&asr->mm_generations, actions->allocator,
			sizeof(struct memory_map_record));

	asr->address_uid = address_space_uid;

	if (cafe_hashtbl_insert(actions->address_space_map,
				&asr->address_uid, asr) == -1)
	{
		cafe_array_release(&asr->mm_generations);
		CAFE_FREE(actions->allocator, asr);
		return NULL;
	}
	return asr;
}


/**
 *  Release all memory associated with the given address space record.  Meant to
 * be used as a callback from hashtbl_delete().
 */
static
void address_space_free(void *v, void *arg)
{
	struct address_space_record *asr = v;
	struct symresolve_actions *actions = arg;

	generations_clear(&asr->mm_generations, actions);
	cafe_array_release(&asr->mm_generations);
	CAFE_FREE(actions->allocator, v);
}


static
void address_space_remove(
	struct symresolve_actions *actions,
	unsigned long long address_space_uid)
{
	struct address_space_record *asr;

	asr = cafe_hashtbl_remove(actions->address_space_map,
				  &address_space_uid,
				  NULL, NULL);
	if (asr != NULL)
		address_space_free(asr, actions);
}


static
unsigned long hashtbl_ull_hash(const void *k)
{
	//td: this could be improved.
	const unsigned long long *ull = k;
	return (unsigned long) (-1UL & *ull);
//	return int32hash(*((const unsigned int *)k));
}


static
int hashtbl_ull_equals(const void *a, const void *b)
{
	return *((const unsigned long long *)a) == *((const unsigned long long *)b);
}


/**
 *  Format the given printf-style message string and emit it using the given
 * symresolve action object's error message handler.
 */
#define error(actions, fmt, ...)					\
	do {								\
		char *msg = cafe_mprintf((actions)->allocator, (fmt),	\
					 ## __VA_ARGS__);		\
		if (msg == NULL) {					\
			break;						\
		}							\
		(actions)->error_handler(msg, (actions)->error_handler_arg); \
		CAFE_FREE((actions)->allocator, msg);			\
	} while(0)


#if defined(HAVE_TCF) && ENABLE_ELF
/**
 *  Allocate a symbol cache object pairing the given local and remote
 * filenames.  Should be deallocated by symbol_cache_free() when no longer in
 * use.
 */
static
bool symbol_cache_new(struct symbol_cache **sc, const char *local_filename,
		      const char *remote_filename,
		      struct symresolve_actions *actions)
{
	size_t len_local = strlen(local_filename) + 1;
	size_t len_remote = strlen(remote_filename) + 1;
	int fd;
	int rc;

	*sc = CAFE_ALLOC(actions->allocator,
			sizeof(**sc) + len_local + len_remote);
	if (*sc == NULL)
		return false;

	memcpy((*sc)->local_filename, local_filename, len_local);
	(*sc)->remote_filename = &(*sc)->local_filename[len_local];
	memcpy((*sc)->remote_filename, remote_filename, len_remote);
	(*sc)->refs = 0;
	(*sc)->is_mapped_text_section = false;
	(*sc)->unresolved_entry = NULL;

	fd = open(local_filename, O_RDONLY, 0);
	if (fd == -1) {
		error(actions, "Error locating %s locally: %s\n",
		      local_filename, strerror(errno));
		(*sc)->local_filename[0] = 0;
	} else
		close(fd);

	rc = pcache_new(&(*sc)->syms, 128,
			128 * sizeof(struct symbol_cache_entry),
			actions->allocator);
	if (rc != CAFE_OK) {
		CAFE_FREE(actions->allocator, (*sc));
		return false;
	}

	rc = pcache_new(&(*sc)->lineinfo, 128,
			128 * sizeof(struct lineinfo_cache_entry),
			actions->allocator);
	if (rc != CAFE_OK) {
		pcache_free(&(*sc)->syms);
		CAFE_FREE(actions->allocator, (*sc));
		return false;
	}

	if (cafe_hashtbl_insert(actions->object_cache, (*sc)->remote_filename,
				(*sc)) == -1)
	{
		pcache_free(&(*sc)->lineinfo);
		pcache_free(&(*sc)->syms);
		CAFE_FREE(actions->allocator, (*sc));
		return false;
	}

	return true;
}


/**
 *  Reference the given symbol cache entry from the given map region.
 */
static
void symbol_cache_reference(struct symbol_cache *sc,
			    struct memory_map_region *region)
{
	region->cache = sc;
	++sc->refs;
	if (sc->is_mapped_text_section &&
	    (region->file_offset == SYMRESOLVE_OFFSET_UNKNOWN))
	{
		region->file_offset = 0;
	}
}
#endif  /* if defined(HAVE_TCF) && ENABLE_ELF */


/**
 *  Release all memory associated with the given address space record.  Meant to
 * be used as a callback from hashtbl_delete().
 */
static
void symbol_cache_free(void *v, void *arg)
{
	struct symbol_cache *sc = v;
	const struct cafe_allocator *a = arg;

	assert(sc->refs == 0);
	pcache_free(&sc->lineinfo);
	pcache_free(&sc->syms);
	CAFE_FREE(a, sc->unresolved_entry);
	CAFE_FREE(a, v);
}


/**
 *  De-reference the given map region's cache entry.  Free the cache entry if
 * this was the last handle to it.
 */
static
void symbol_cache_dereference(struct memory_map_region *region,
			      struct symresolve_actions *actions)
{
	if (region->cache == NULL)
		return;

	assert(region->cache->refs > 0);
	if (--region->cache->refs == 0) {
		/* Don't free the cache immediately, it might get reused */
	}
	region->cache = NULL;
}


#if defined(HAVE_TCF)
/**
 *  TCF post_event() callback routine for locking a given channel object.
 */
static
void channel_lock_callback(void *arg)
{
	Channel *c = arg;
	channel_lock(c);
}


/**
 *  TCF post_event() callback routine for unlocking a given channel object.
 */
static
void channel_unlock_callback(void *arg)
{
	Channel *c = arg;
	channel_unlock(c);
}


#  if defined(ENABLE_ContextMux) && ENABLE_ContextMux
/**
 *  A TCF Context structure callback routine.  Copies the memory map for the
 * current address space into the MemoryMap structure provided by the caller.
 */
static
int symresolve_context_get_memory_map(Context *ctx, MemoryMap *map)
{
	MemoryRegion *r1;
	const struct memory_map_region *r2;
	unsigned int i, j;

	assert(ctx_map != NULL);
	assert(map != NULL);

	if (ctx_map->regions.len == 0) {
		map->region_cnt = ctx_map->regions.len;
		return 0;
	}

	if (map->region_max < ctx_map->regions.len) {
		map->region_max = ctx_map->regions.len;
		map->regions = (MemoryRegion *) realloc(map->regions, sizeof(MemoryRegion) * map->region_max);
	}

	r1 = map->regions;
	r2 = ctx_map->regions.data;
	for (i = 0, j = 0; i < ctx_map->regions.len; ++i) {
		const char *file_name;
		char *tmp;

		/* skip non-file-backed memory regions */
		if (r2[i].file_name == NULL)
			continue;

		if (r2[j].cache == NULL || r2[j].cache->local_filename[0] == '\0') {
			tmp = apply_path_map(ctx_channel, NULL,
						   (char *) r2[i].file_name,
						   PATH_MAP_TO_LOCAL);
		} else {
			/* reuse mapped path from cache */
			tmp = (char *) &r2[j].cache->local_filename;
		}
		if (tmp != NULL)
			file_name = tmp;
		else
			file_name = r2[i].file_name;

		memset(&r1[j], 0, sizeof(MemoryRegion));

		if ((r2[i].cache != NULL) &&
				r2[i].cache->is_mapped_text_section) {
			r1[j].addr = r2[i].start_addr;
			r1[j].size = r2[i].len;
			r1[j].sect_name = loc_strdup(".text");
		} else {
			if (j > 0 && strcmp(r1[j-1].file_name, file_name) == 0) {
				/* region in same object file - use lowest load addr */
				if (r1[j-1].addr > r2[i].start_addr)
					r1[j-1].addr = r2[i].start_addr;
				continue;
			}
			/* This triggers the "Linux module" case of search_regions. */
			r1[j].addr = r2[i].start_addr;
		}
		r1[j].file_name = loc_strdup(file_name);

		++j;
	}

	map->region_cnt = j;
	return 0;
}


/**
 *  TCF Context get_group() callback.
 */
static
Context *symresolve_context_get_group(Context *ctx, int group)
{
	return ctx;
}


struct lineinfo_arg {
	const char *file_name;
	const char *file_dir;
	unsigned int line;
	unsigned int column;
	ContextAddress address;
	uint64_t size;
	bool found;
};


/**
 *
 */
static
void TCF_line_number_callback(CodeArea *ca, void *arg)
{
	struct lineinfo_arg *a = arg;

	a->file_name = ca->file;
	a->file_dir = ca->directory;
	a->line = ca->start_line;
	a->column = ca->start_column;
	a->address = ca->start_address;
	a->size = ca->end_address - ca->start_address;
	a->found = true;
}
#  endif  /* if ENABLE_ContextMux */

#  if defined(ENABLE_ELF) && ENABLE_ELF
struct elf_parse_arg {
	unsigned long long *start_addr;
	unsigned long long *file_len;
	unsigned long long *file_offset;
	const char *file_name;
	Channel *c;
};


/**
 *  Attempts to parse the object file given by the delta entry record, and
 * updates the record with the program header information found in the first
 * load entry.
 */
static
void elf_parse_tcf_callback(void *arg)
{
	struct elf_parse_arg *a = arg;
	char *file_name;
	const ELF_File *file;
	const ELF_PHeader *p;
	unsigned int i;
	char *tmp;

	assert(a != NULL);
	file_name = (char *)a->file_name;
	tmp = apply_path_map(a->c, NULL, file_name,
				   PATH_MAP_TO_LOCAL);
	if (tmp != NULL)
		file_name = tmp;

	file = elf_open(file_name);
	if (file == NULL)
		return;

	/* locate the corresponding program header */
	for (i = 0; i < file->pheader_cnt; ++i) {
		p = &file->pheaders[i];
		if (p->type != PT_LOAD)
			continue;

		*a->file_offset = p->offset;
		if (*a->start_addr == 0) {
			*a->start_addr = p->address;
			*a->file_len = p->mem_size;
		}
		return;
	}
}


/**
 *
 */
static
void get_elf_load_address(Channel *c, unsigned long long *start_addr,
			  unsigned long long *file_len,
			  unsigned long long *file_offset,
			  const char *file_name)
{
	struct elf_parse_arg a;

	a.start_addr = start_addr;
	a.file_len = file_len;
	a.file_offset = file_offset;
	a.file_name = file_name;
	a.c = c;
	if (cafe_islocked())
		cafe_post_event_sync(elf_parse_tcf_callback, &a);
	else
		post_event_sync(elf_parse_tcf_callback, &a);
}
#  endif  /* if ENABLE_ELF */

#endif  /* if defined(HAVE_TCF) */


/**
 *  When clients choose not to specify a custom error message handler, simply
 * post the messages to stderr.
 */
static
void default_errhandler(const char *msg, void *arg)
{
	(void)fputs(msg, stderr);
}


/**
 *  Create a symbol resolver actions engine using a TCF-based backend.
 */
int symresolve_actions_create(struct cafe_value_type *delta_vt,
			      const struct cafe_allocator *allocator,
			      struct symresolve_actions **actions)
{
	int rc;
	char *debug;
#if defined(HAVE_TCF) && ENABLE_ELF
#  if defined(ENABLE_ContextMux) && ENABLE_ContextMux
        static int initialized = 0;
	if (!initialized) {
	    SYMRESOLVE_CTX_IF.context_get_group = symresolve_context_get_group;
	    SYMRESOLVE_CTX_IF.context_get_memory_map = symresolve_context_get_memory_map;
	    initialized = 1;
	};
#  endif
#endif

	assert(delta_vt != NULL);
	assert(allocator != NULL);
	assert(actions != NULL);

	if ((debug = getenv("CAFE_AGGREGATOR_DEBUG")) != NULL)
		cafe_aggregator_debug = atol(debug);

	*actions = CAFE_ALLOC(allocator, sizeof(**actions));
	if (*actions == NULL)
		return CAFE_NOMEM;

#if defined(SYMRESOLVE_PROFILE_CACHE)
	(*actions)->hits = 0;
	(*actions)->misses = 0;
	(*actions)->lhits = 0;
	(*actions)->lmisses = 0;
#endif
	rc = cafe_type_parse(delta_vt->ident->ns,
			     CAFE_MM_DELTA_ENTRY_DEF CAFE_MM_DELTA_DEF,
			     &(*actions)->delta_vt);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_type_parse(delta_vt->ident->ns,
			     CAFE_MM_NO_DELTA_DEF,
			     &(*actions)->no_delta_vt);
	if (rc != CAFE_OK)
		return rc;

	if (delta_vt == (*actions)->delta_vt) {
		struct cafe_vm_attr *attr;
		assert((*actions)->delta_vt->elements == 1);
		attr = cafe_type_attr_by_index((*actions)->delta_vt, 0);
		assert(attr->kind == CAFE_ATTR_VAR);
		assert(attr->offset == 0);
		(*actions)->delta_list_vt = cafe_base_type(cafe_attr_type(attr));
	} else if (delta_vt == (*actions)->no_delta_vt) {
		(*actions)->delta_list_vt = NULL;
	} else
		return CAFE_ERROR;

	(*actions)->address_space_map = cafe_hashtbl_new(INITIAL_ADDRESS_SPACE_MAP_SIZE,
							 allocator,
							 hashtbl_ull_hash,
							 hashtbl_ull_equals);
	if ((*actions)->address_space_map == NULL) {
		CAFE_FREE(allocator, *actions);
		return CAFE_NOMEM;
	}

	(*actions)->object_cache = cafe_hashtbl_new(INITIAL_OBJECT_CACHE_SIZE,
						    allocator,
						    cafe_hashtbl_string_hash,
						    cafe_hashtbl_string_equals);
	if ((*actions)->object_cache == NULL) {
		cafe_hashtbl_delete((*actions)->address_space_map, NULL, NULL,
				    address_space_free, actions);
		CAFE_FREE(allocator, *actions);
		return CAFE_NOMEM;
	}

	(*actions)->allocator = allocator;
#if defined(HAVE_TCF) && ENABLE_ELF
#  if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	(*actions)->ctx = create_context("cafe_scratch");
	if (context_set_interface((*actions)->ctx, &SYMRESOLVE_CTX_IF)) {
		cafe_hashtbl_delete((*actions)->object_cache, NULL, NULL,
				    symbol_cache_free,
				    (void *)(*actions)->allocator);
		cafe_hashtbl_delete((*actions)->address_space_map, NULL, NULL,
				    address_space_free, *actions);
		CAFE_FREE(allocator, *actions);
		(void)fputs("Error initializing symbol resolution context dispatcher "
		      "structure", stderr);
		return CAFE_ERROR;
	}
	(*actions)->ctx->ref_count = 1;
	list_add_first(&(*actions)->ctx->ctxl, &tcf_ctx_list);

	(*actions)->lineinfo_index = NULL;
	(*actions)->next_lineinfo_index = 0;
#  endif

	(*actions)->c = cafe_config_channel_up;
	if ((*actions)->c != NULL)
		post_event(channel_lock_callback, (*actions)->c);

	(*actions)->path_map_listener.mapping_changed = tcf_path_mapping_changed;
	post_event(tcf_add_path_map_event_listener, *actions);
	(*actions)->path_map_changed = false;
#endif  /* if HAVE_TCF && ENABLE_ELF */

	(*actions)->unmapped_entry = NULL;
	(*actions)->unnamed_entry = NULL;
	(*actions)->symbol_index = NULL;
	(*actions)->next_sym_index = 0;

	(*actions)->error_handler = default_errhandler;
	(*actions)->error_handler_arg = NULL;
	return CAFE_OK;
}


/**
 *  Releases resources acquired by create().
 */
void symresolve_actions_delete(struct symresolve_actions *actions)
{
	const struct cafe_allocator *a = actions->allocator;

#if defined(SYMRESOLVE_PROFILE_CACHE)
	printf("Hits/misses:  %u / %u\t %u / %u\n",
	       actions->hits, actions->misses,
	       actions->lhits, actions->lmisses);
#endif
#if defined(HAVE_TCF) && ENABLE_ELF
	post_event(tcf_rem_path_map_event_listener, &actions->path_map_listener);
	if (actions->c != NULL)
		post_event(channel_unlock_callback, actions->c);
#  if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	send_context_exited_event(actions->ctx);
#  endif
#endif

	CAFE_FREE(a, actions->unmapped_entry);
	CAFE_FREE(a, actions->unnamed_entry);
	cafe_hashtbl_delete(actions->address_space_map, NULL, NULL,
			    address_space_free, actions);
	cafe_hashtbl_delete(actions->object_cache, NULL, NULL,
			    symbol_cache_free, (void *)a);
	CAFE_FREE(a, actions);
}


/**
 *  Specify the dictionary to be populated with new symbol record entries.
 */
void symresolve_set_symindex(struct symresolve_actions *actions,
			     struct cafe_dict_struct *symbol_table)
{
	assert(actions != NULL);
	actions->symbol_index = symbol_table;
}


/**
 *  Specify the dictionary to be populated with new source info record entries.
 */
void symresolve_set_lineinfoindex(struct symresolve_actions *actions,
				  struct cafe_dict_struct *lineinfo_table)
{
	assert(actions != NULL);
	actions->lineinfo_index = lineinfo_table;
}


/**
 *  Specify the callback to be invoked when symbol resolution needs to
 * communicate an error condition.
 */
void symresolve_set_errhandler(struct symresolve_actions *actions,
			       void (*handler_callback) (const char *msg, void *arg),
			       void *arg)
{
	assert(actions != NULL);

	if (handler_callback == NULL) {
		actions->error_handler = default_errhandler;
		actions->error_handler_arg = NULL;
	} else {
		actions->error_handler = handler_callback;
		actions->error_handler_arg = arg;
	}
}


/**
 *   Reset symbol resolver actions engine.
 */
void symresolve_actions_reset(struct symresolve_actions *actions)
{
	cafe_hashtbl_clear(actions->address_space_map, NULL, NULL,
			   address_space_free, actions);
	cafe_hashtbl_clear(actions->object_cache, NULL, NULL,
			   symbol_cache_free, (void *)actions->allocator);
}


/**
 *  Create a memory map region and append it to the given memory map.
 */
static
int append_mm_region(struct memory_map_record *mm,
		     unsigned long long start_addr,
		     unsigned long long len,
		     unsigned long long file_offset,
		     const char *file_name)
{
	struct memory_map_region *region;
	int rc = cafe_array_capacity_set(&mm->regions, ++mm->regions.len);
	if (rc != CAFE_OK)
		return rc;

	region = &cafe_array_val(&mm->regions, struct memory_map_region,
				 mm->regions.len - 1);
	if (file_name != NULL) {
		region->file_name = CAFE_STRDUP(mm->regions.allocator, file_name);
		if (region->file_name == NULL)
			return CAFE_NOMEM;
	} else {
		region->file_name = NULL;
	}

	region->start_addr = start_addr;
	region->len = len;
	region->file_offset = file_offset;
	region->cache = NULL;
	return CAFE_OK;
}


/**
 *
 */
static
int insert_mm_region(struct cafe_array *regions, unsigned int slot,
		     unsigned long long start_addr,
		     unsigned long long len,
		     unsigned long long file_offset,
		     const char *file_name)
{
	struct memory_map_region *trav;
	int rc = cafe_array_capacity_set(regions, ++regions->len);
	if (rc != CAFE_OK)
		return rc;

	trav = (struct memory_map_region *) regions->data + slot;
	memmove((struct memory_map_region *) regions->data + (slot + 1),
		trav,
		(regions->len - slot) * sizeof(struct memory_map_region));

	trav->start_addr = start_addr;
	trav->len = len;
	trav->file_offset = file_offset;
	trav->cache = NULL;
	if (file_name != NULL) {
		trav->file_name = CAFE_STRDUP(regions->allocator, file_name);
		if (trav->file_name == NULL)
			return CAFE_NOMEM;
	} else {
		trav->file_name = NULL;
	}
	return CAFE_OK;
}


static
bool same_file_name(const char *file_name1, const char *file_name2)
{
	if (file_name1 == file_name2)
		return true;
	if (file_name1 == NULL || file_name2 == NULL)
		return false;
	return strcmp(file_name1, file_name2) == 0;
}


/**
 *
 */
static
int update_mm(struct memory_map_record *m,
	      unsigned long long start_addr,
	      unsigned long long len,
	      unsigned long long file_offset,
	      const char *file_name,
	      struct symresolve_actions *actions)
{
	struct memory_map_region *trav, *next = NULL;
	unsigned int i;
	unsigned int j;

	assert(len != MARKER_len);

#if defined(HAVE_TCF) && defined(ENABLE_ELF) && ENABLE_ELF
	if (file_offset == SYMRESOLVE_OFFSET_UNKNOWN)
		get_elf_load_address(actions->c, &start_addr, &len,
				     &file_offset, file_name);
#endif

	/* check whether we can simply append this new region */
	if (m->regions.len == 0)
		return append_mm_region(m, start_addr, len, file_offset,
					file_name);

	trav = &cafe_array_val(&m->regions, struct memory_map_region,
			       m->regions.len - 1);
	assert(trav->len != MARKER_len);
	if (start_addr >= trav->start_addr + trav->len) {
		if (start_addr == trav->start_addr + trav->len &&
		    file_offset == trav->file_offset + trav->len &&
		    same_file_name(file_name, trav->file_name)) {
			trav->len += len;
			return CAFE_OK;
		}
		return append_mm_region(m, start_addr, len, file_offset,
					file_name);
	}

	/* locate the closest existing region */
	for (i = 0; i < m->regions.len; ++i) {
		trav = &cafe_array_val(&m->regions, struct memory_map_region, i);
		if (start_addr <= trav->start_addr + trav->len)
			break;
	}

check_next_region:
	assert(i < m->regions.len);
	if (start_addr + len < trav->start_addr)
		/* Not adjacent or overlapping */
		return insert_mm_region(&m->regions, i, start_addr,
					len, file_offset, file_name);

	if (start_addr <= trav->start_addr) {
		/* Adjacent or overlapping with beginning of existing region */
		if (file_offset + (trav->start_addr - start_addr) == trav->file_offset &&
		    same_file_name(file_name, trav->file_name)) {
			/* Join with existing region */
			trav->len += trav->start_addr - start_addr;
			trav->start_addr = start_addr;
			trav->file_offset = file_offset;
			if (len <= trav->len)
				return CAFE_OK;
			trav->len = len;
		} else if (start_addr + len >= trav->start_addr + trav->len) {
			/* Replace whole region */
			trav->start_addr = start_addr;
			trav->file_offset = file_offset;
			trav->len = len;
			CAFE_FREE(m->regions.allocator, (char *)trav->file_name);
			if (file_name != NULL) {
				trav->file_name = CAFE_STRDUP(m->regions.allocator, file_name);
				if (trav->file_name == NULL)
					return CAFE_NOMEM;
			} else {
				trav->file_name = NULL;
			}

			symbol_cache_dereference(trav, actions);
		} else {
			/* New region not matching with existing region */
			if (start_addr + len >= trav->start_addr) {
				/* Move start of existing region forward if overlapping. */
				trav->len -= start_addr + len - trav->start_addr;
				trav->file_offset += start_addr + len - trav->start_addr;
				trav->start_addr = start_addr + len;
			}
			return insert_mm_region(&m->regions, i, start_addr,
						len, file_offset, file_name);
		}
	} else if (file_offset - (start_addr - trav->start_addr) == trav->file_offset &&
		   same_file_name(file_name, trav->file_name)) {
		/* Join with existing region */
		trav->len = start_addr + len - trav->start_addr;
	} else {
		/* Could not join with this region, shrink this region
		 * if needed and try to join with next. */
		if (start_addr < trav->start_addr + trav->len) {
			trav->len = start_addr - trav->start_addr;
		}
		if (++i >= m->regions.len) {
			return append_mm_region(m, start_addr, len, file_offset,
						file_name);
		}
		trav = &cafe_array_val(&m->regions, struct memory_map_region, i);
		goto check_next_region;
	}

	for (j = i + 1; j < m->regions.len; ++j) {
		next = &cafe_array_val(&m->regions, struct memory_map_region, j);
		if (trav->start_addr + trav->len < next->start_addr + next->len) {
			if (trav->start_addr + trav->len >= next->start_addr) {
				/* Adjacent or partially overlapping */
				if (next->file_offset - (next->start_addr - trav->start_addr) == trav->file_offset &&
				    same_file_name(file_name, trav->file_name)) {
					/* Join regions of same type */
					trav->len = next->start_addr + next->len - trav->start_addr;
					continue;
				}
				if (trav->start_addr + trav->len > next->start_addr) {
					next->start_addr += trav->start_addr + trav->len - next->start_addr;
					next->file_offset += trav->start_addr + trav->len - next->start_addr;
				}
			}
			break;
		}
	}
	if (++i < j) {
		trav = &cafe_array_val(&m->regions, struct memory_map_region, i);
		if (j < m->regions.len)
			memmove(trav, next, (m->regions.len - j) * sizeof(struct memory_map_region));
		m->regions.len -= j - i;
	}
	return CAFE_OK;
}


/**
 *  Create a new generation memory map for the given address space, by copying
 * the current latest generation map.
 */
static
int clone_latest_generation(const struct memory_map_record *m1,
			    struct memory_map_record *m2,
			    const struct cafe_allocator *a)
{
	unsigned int i;
	int rc;

	assert(!m2->committed);
	if (m1->regions.len > 0 &&
	    cafe_array_val(&m1->regions, struct memory_map_region, 0).len == MARKER_len)
		return CAFE_OK;
	rc = cafe_array_capacity_set(&m2->regions, m1->regions.len);
	if (rc != CAFE_OK)
		return rc;
	memcpy(m2->regions.data, m1->regions.data,
	       m1->regions.len * m1->regions.element_size);
	m2->regions.len = m1->regions.len;
	for (i = 0; i < m2->regions.len; ++i) {
		struct memory_map_region *r;

		r = &cafe_array_val(&m2->regions, struct memory_map_region, i);
		if (r->file_name != NULL) {
			r->file_name = CAFE_STRDUP(a, r->file_name);
			if (r->file_name == NULL)
				return CAFE_NOMEM;
		} else {
			r->file_name = NULL;
		}

		if (r->cache != NULL)
			++r->cache->refs;
	}
	return CAFE_OK;
}


/**
 *
 */
static
int mm_table_insert(struct cafe_vm_dynarray *delta,
		    struct cafe_value_type *delta_vt,
		    unsigned long long start_addr,
		    unsigned long long len,
		    unsigned long long file_offset,
		    const char *file_name)
{
	struct cafe_mm_delta_entry *e;
	size_t l = delta->l;
	int rc;

	rc = cafe_vm_dynarray_set_size(delta, l + 1, delta_vt);
	if (rc != CAFE_OK)
		return rc;
	e = ((struct cafe_mm_delta_entry *)delta->p) + l;
	e->start_addr = start_addr;
	e->len = len;
	e->file_offset = file_offset;
	e->file_name = file_name;
	return CAFE_OK;
}


static
int generate_delta(struct symresolve_actions *actions,
		   struct cafe_vm_dynarray *delta,
		   struct address_space_record *asr)
{
#if !defined(NDEBUG)
	struct memory_map_record *old_m;
#endif
	struct memory_map_record *m = NULL;
	unsigned int i;
	unsigned int last_committed;
	int first_time = 1;
	int rc;

	i = (unsigned int) asr->mm_generations.len;
	while (i-- > 0) {
		struct memory_map_record *m = &cafe_array_val(
			&asr->mm_generations, struct memory_map_record, i);
		if (m->committed) {
			if (m->generated)
				return CAFE_OK;
			break;
		}
	}
	last_committed = i + 1;
	for (i = 0; i < last_committed; ++i) {
		unsigned int j;

#if !defined(NDEBUG)
		old_m = m;
#endif
		m = &cafe_array_val(&asr->mm_generations, struct memory_map_record, i);
		if (m->generated)
			continue;
		if (first_time) {
			first_time = 0;
			rc = mm_table_insert(delta,
					     actions->delta_list_vt,
					     MARKER_EVENT_ADDRESS_SPACE,
					     MARKER_len,
					     asr->address_uid,
					     NULL);
			if (rc != CAFE_OK)
				return rc;
		}
		rc = mm_table_insert(delta,
				     actions->delta_list_vt,
				     MARKER_EVENT_GENERATION,
				     MARKER_len,
				     m->generation,
				     NULL);
		if (rc != CAFE_OK)
			return rc;

		rc = mm_table_insert(delta,
				     actions->delta_list_vt,
				     MARKER_EVENT_PARENT_GENERATION,
				     MARKER_len,
				     m->parent_generation,
				     NULL);
		if (rc != CAFE_OK)
			return rc;

		/* WRS_TODO: optimize delta and handle unmap */
#if !defined(NDEBUG)
		assert(old_m == NULL || m->generation == old_m->generation + 1);
#endif
		for (j = 0; j < m->regions.len; ++j) {
			int rc;
			struct memory_map_region *r = &cafe_array_val(
				&m->regions, struct memory_map_region, j);
			rc = mm_table_insert(delta,
					     actions->delta_list_vt,
					     r->start_addr,
					     r->len,
					     r->file_offset,
					     r->file_name);
			if (rc != CAFE_OK)
				return rc;
		}
		m->generated = true;
	}
	return CAFE_OK;
}


/**
 *  Generate memory map delta table from current memory maps.  Non
 *  committed generations, i.e. generation for which
 *  symresolve_mm_current_gen() has not been called, are not included
 *  in the delta.
 */
int symresolve_mm_gen_delta(struct symresolve_actions *actions,
			    void *delta_arg)
{
	struct cafe_vm_dynarray *delta = (struct cafe_vm_dynarray *)delta_arg;
	struct cafe_hashtbl_iter iter;
	int first_error = CAFE_OK;

	assert(actions != NULL);
	assert(delta_arg != NULL);
	if (actions->delta_list_vt == NULL) {
		actions->error_handler("Warning: Attempting to generate delta when no output table type provided", actions->error_handler_arg);
		return CAFE_OK;
	}
	cafe_hashtbl_iter_init(actions->address_space_map, &iter);
	while (cafe_hashtbl_iter_next(actions->address_space_map, &iter) != 0) {
		struct address_space_record *asr = iter.val;
		int rc;

		if (asr == NULL)
			continue;

		rc = generate_delta(actions, delta, asr);
		if (rc != CAFE_OK && first_error == CAFE_OK)
			first_error = rc;
	}
	return first_error;
}


/**
 *  Clear memory map delta table
 */
void symresolve_mm_clear_delta(struct symresolve_actions *actions,
			       void *delta_arg)
{
	struct cafe_vm_dynarray *delta = (struct cafe_vm_dynarray *)delta_arg;

	assert(actions != NULL);
	assert(delta_arg != NULL);
	delta->l = 0;
}


/**
 *  Update the current memory map table with a single new record.  Used to
 * build delta tables.  A newly unmapped region is expressed as a range with
 * no access permissions.
 */
int symresolve_mm_update(struct symresolve_actions *actions,
			 unsigned long long address_space_uid,
			 unsigned long long start_addr,
			 unsigned long long len,
			 unsigned long long file_offset,
			 const char *file_name,
			 bool map)
{
	struct memory_map_record *m;
	struct address_space_record *asr;
	int rc;

	assert(actions != NULL);
	if (len == MARKER_len)
		return CAFE_ERROR;

	asr = cafe_hashtbl_lookup(actions->address_space_map,
				  &address_space_uid);
	if (asr == NULL) {
		rc = symresolve_address_space_clear(actions, address_space_uid,
						    address_space_uid);
		if (rc != CAFE_OK)
			return rc;
		asr = cafe_hashtbl_lookup(actions->address_space_map,
					  &address_space_uid);
		if (asr == NULL)
			return CAFE_ERROR;
	}

	if (asr->mm_generations.len == 0)
		return CAFE_ERROR;
	m = &cafe_array_val(&asr->mm_generations, struct memory_map_record,
			    asr->mm_generations.len - 1);
	if (m->committed ||
	    (m->regions.len > 0 &&
	     cafe_array_val(&m->regions, struct memory_map_region, 0).len == MARKER_len)) {
		struct memory_map_record *new_map;

		new_map = mm_new(&asr->mm_generations, address_space_uid);
		if (new_map == NULL)
			return CAFE_NOMEM;
		m = &cafe_array_val(&asr->mm_generations, struct memory_map_record,
				    asr->mm_generations.len - 2);
		rc = clone_latest_generation(m, new_map, actions->allocator);
		if (rc != CAFE_OK)
			return rc;
		m = new_map;
	}
	return update_mm(m, start_addr, len, file_offset, file_name, actions);
}


static
bool is_clear_marker(const struct memory_map_region *r)
{
	return (r->len == MARKER_len) && (r->start_addr == MARKER_EVENT_CLEAR);
}


static
bool is_delete_marker(const struct memory_map_region *r)
{
	return (r->len == MARKER_len) && (r->start_addr == MARKER_EVENT_DELETE);
}


/**
 *  Update the current memory map table with a delta table.
 */
int symresolve_mm_delta(struct symresolve_actions *actions, void *delta_arg)
{
	const struct cafe_vm_dynarray *delta =
		(struct cafe_vm_dynarray *)delta_arg;
	struct address_space_record *asr = NULL;
	struct memory_map_record *map = NULL;
	struct cafe_mm_delta_entry *e;
	unsigned int i;
	int rc;

	assert(actions != NULL);
	assert(delta != NULL);

	for (i = 0, e = (struct cafe_mm_delta_entry *)delta->p;
	     i < delta->l; ++i, ++e) {
		if (e->len == MARKER_len) {
			if (e->start_addr == MARKER_EVENT_ADDRESS_SPACE) {
				if (map != NULL)
					map->committed = true;
				map = NULL;
				asr = cafe_hashtbl_lookup(actions->address_space_map,
							  &e->file_offset);
				if (asr == NULL) {
					asr = address_space_new(actions, e->file_offset);
					if (asr == NULL)
						return CAFE_NOMEM;
				}
				continue;
			}
			if (e->start_addr == MARKER_EVENT_GENERATION) {
				assert(asr != NULL);
				if (map != NULL)
					map->committed = true;
				map = mm_new(&asr->mm_generations,
					     asr->address_uid);
				if (map == NULL)
					return CAFE_NOMEM;
				if (map->generation != e->file_offset)
					return CAFE_ERROR;
				continue;
			}
			if (map == NULL)
				return CAFE_ERROR;
			if (e->start_addr == MARKER_EVENT_PARENT_GENERATION) {
				map->parent_generation = e->file_offset;
				continue;
			} else if (e->start_addr == MARKER_EVENT_CLEAR)
				map->parent_uid = e->file_offset;
			else if (e->start_addr != MARKER_EVENT_DELETE)
				return CAFE_ERROR;
			rc = append_mm_region(map, e->start_addr, e->len,
					      e->file_offset, e->file_name);
			if (rc != CAFE_OK)
				return rc;
			continue;
		}

		if (map == NULL)
			return CAFE_ERROR;

		if (asr->mm_generations.len < 2 ||
		    (map->regions.len > 0 &&
		     cafe_array_val(&map->regions, struct memory_map_region, 0).len == MARKER_len))
			return CAFE_ERROR;

		if (map->regions.len == 0) {
			struct memory_map_record *old_map = &cafe_array_val(
				&asr->mm_generations, struct memory_map_record,
				asr->mm_generations.len - 2);
			rc = clone_latest_generation(old_map, map, actions->allocator);
			if (rc != CAFE_OK)
				return rc;
		}
		rc = update_mm(map, e->start_addr, e->len, e->file_offset,
			       e->file_name, actions);
		if (rc != CAFE_OK)
			return rc;
	}
	if (map != NULL)
		map->committed = true;
	return CAFE_OK;
}


/**
 *  Drop all but the most recent generation of existing memory maps.
 */
int symresolve_mm_flush(struct symresolve_actions *actions)
{
	struct cafe_hashtbl_iter iter;

	cafe_hashtbl_iter_init(actions->address_space_map, &iter);
	while (cafe_hashtbl_iter_next(actions->address_space_map, &iter) != 0) {
		struct memory_map_record *map, *dest;
		unsigned int i;
		struct address_space_record *asr = iter.val;

		if (asr == NULL)
			continue;

		if (asr->mm_generations.len > 0) {
			map = &cafe_array_val(&asr->mm_generations,
					      struct memory_map_record,
					      asr->mm_generations.len - 1);
		} else {
			map = NULL;
		}
		if (map == NULL ||
		    map->regions.len == 0 ||
		    is_delete_marker(&cafe_array_val(&map->regions, struct memory_map_region, 0))) {
			address_space_remove(actions, asr->address_uid);
			cafe_hashtbl_iter_init(actions->address_space_map, &iter);
			continue;
		}

		if (asr->mm_generations.len == 1)
			continue;

		dest = &cafe_array_val(&asr->mm_generations,
				       struct memory_map_record, 0);
		for (i = 0; i < asr->mm_generations.len - 1; ++i) {
			map = &cafe_array_val(&asr->mm_generations,
					      struct memory_map_record, i);
			mm_free(map, actions);
		}

		++map;
		memcpy(dest, map, sizeof(*map));
		asr->mm_generations.len = 1;
	}
	return CAFE_OK;
}


/**
 *  Clients should invoke this routine when a new address space is created
 * using an existing address space's memory map.
 */
int symresolve_address_space_clone(struct symresolve_actions *actions,
				   unsigned long long address_space_uid,
				   unsigned long long new_uid)
{
	struct address_space_record *asr, *new_asr;
	struct memory_map_record *m_orig, *m_new;
	unsigned int i;
	int rc;

	assert(actions != NULL);
	if (address_space_uid == new_uid)
		return CAFE_ERROR;

	asr = cafe_hashtbl_lookup(actions->address_space_map,
				  &address_space_uid);
	if (asr == NULL)
		return CAFE_OK;

	if (asr->mm_generations.len == 0)
		return CAFE_ERROR;

	m_orig = &cafe_array_val(&asr->mm_generations, struct memory_map_record,
				 asr->mm_generations.len - 1);
	if (m_orig->regions.len == 0 ||
	    is_delete_marker(&cafe_array_val(&m_orig->regions, struct memory_map_region, 0))) {
		return CAFE_ERROR;
	}

	rc = symresolve_address_space_clear(actions, new_uid, m_orig->parent_uid);
	if (rc != CAFE_OK)
		return rc;

	if (cafe_array_val(&m_orig->regions, struct memory_map_region, 0).len == MARKER_len)
		return CAFE_OK;

	new_asr = cafe_hashtbl_lookup(actions->address_space_map, &new_uid);
	assert(new_asr != NULL);
	assert(new_asr->mm_generations.len > 0);
	m_new = mm_new(&new_asr->mm_generations, address_space_uid);
	if (m_new == NULL)
		return CAFE_NOMEM;

	for (i = 0; i < cafe_array_len(&m_orig->regions); ++i) {
		struct memory_map_region *r = &cafe_array_val(&m_orig->regions,
							      struct memory_map_region, i);

		rc = append_mm_region(m_new, r->start_addr, r->len,
				      r->file_offset, r->file_name);
		if (rc != CAFE_OK)
			return rc;
	}
	return CAFE_OK;
}


/**
 *  Signals that the given address space has cleared its memory mappings (e.g.,
 * on a unix exec() call).
 */
int symresolve_address_space_clear(struct symresolve_actions *actions,
				   unsigned long long address_space_uid,
				   unsigned long long parent_uid)
{
	struct address_space_record *asr;
	struct memory_map_record *mm;

	assert(actions != NULL);
	asr = cafe_hashtbl_lookup(actions->address_space_map,
				  &address_space_uid);
	if (asr == NULL) {
		asr = address_space_new(actions, address_space_uid);
		if (asr == NULL)
			return CAFE_NOMEM;
	}

	clear_uncommitted(&asr->mm_generations, actions);

	if (asr->mm_generations.len > 0) {
		mm = &cafe_array_val(&asr->mm_generations, struct memory_map_record,
				     asr->mm_generations.len - 1);
		if (mm->regions.len > 0 &&
		    is_clear_marker(&cafe_array_val(&mm->regions, struct memory_map_region, 0)))
			return CAFE_OK;
	}

	mm = mm_new(&asr->mm_generations, address_space_uid);
	if (mm == NULL)
		return CAFE_NOMEM;
	mm->parent_uid = parent_uid;
	asr = cafe_hashtbl_lookup(actions->address_space_map, &parent_uid);
	if (asr != NULL) {
		const struct memory_map_record *parent_mm =
			&cafe_array_val(&asr->mm_generations,
					struct memory_map_record,
					asr->mm_generations.len - 1);
		mm->parent_generation = parent_mm->generation;
	}

	return append_mm_region(mm, MARKER_EVENT_CLEAR, MARKER_len, parent_uid, NULL);
}


/**
 *  Should be called when an address space has been entirely deleted.  Clients
 * can still resolve symbols on past generations of this address space until
 * mm_flush() is called.
 */
int symresolve_address_space_delete(struct symresolve_actions *actions,
				    unsigned long long address_space_uid)
{
	struct address_space_record *asr;
	struct memory_map_record *mm;
	int rc;

	assert(actions != NULL);
	asr = cafe_hashtbl_lookup(actions->address_space_map,
				  &address_space_uid);
	if (asr == NULL)
		return CAFE_OK;

	clear_uncommitted(&asr->mm_generations, actions);
	if (asr->mm_generations.len == 0) {
		address_space_remove(actions, address_space_uid);
		return CAFE_OK;
	}

	mm = &cafe_array_val(&asr->mm_generations, struct memory_map_record,
			     asr->mm_generations.len - 1);
	if (mm->regions.len > 0 &&
	    is_delete_marker(&cafe_array_val(&mm->regions, struct memory_map_region, 0)))
		return CAFE_OK;

	mm = mm_new(&asr->mm_generations, address_space_uid);
	if (mm == NULL)
		return CAFE_NOMEM;

	rc = append_mm_region(mm, MARKER_EVENT_DELETE, MARKER_len, 0, NULL);
	if (rc != CAFE_OK)
		return rc;

	mm->committed = true;
	return CAFE_OK;
}


/**
 *  Returns the current memory map generation number for the given address
 * space.
 */
int symresolve_mm_current_gen(struct symresolve_actions *actions,
			      unsigned long long address_space_uid,
			      unsigned long long *current_gen)
{
	struct address_space_record *asr, *parent_asr;
	struct memory_map_record *map;

	assert(actions != NULL);
	asr = cafe_hashtbl_lookup(actions->address_space_map,
				  &address_space_uid);
	if (asr == NULL || asr->mm_generations.len == 0) {
		*current_gen = 0;
		return CAFE_OK;
	}

	map = &cafe_array_val(&asr->mm_generations, struct memory_map_record,
			      asr->mm_generations.len - 1);
	if (address_space_uid == map->parent_uid)
		goto done;

	parent_asr = cafe_hashtbl_lookup(actions->address_space_map,
					 &map->parent_uid);
	if ((parent_asr != NULL) && (parent_asr->mm_generations.len != 0)) {
		struct memory_map_record *parent_map =
			&cafe_array_val(&parent_asr->mm_generations,
					struct memory_map_record,
					parent_asr->mm_generations.len - 1);
		unsigned long long parent_gen = parent_map->generation;
		parent_map->committed = true;
		/* if the parent address space has changed since last time it
		   was referenced, update its generation and create a new child
		   generation to reference it if the current child gen is
		   in use. */
		if (map->parent_generation != parent_gen) {
			if (map->committed) {
				const struct memory_map_record *old_map;
				int rc;

				map = mm_new(&asr->mm_generations,
					     address_space_uid);
				if (map == NULL)
					return CAFE_NOMEM;
				old_map = &cafe_array_val(
					&asr->mm_generations,
					struct memory_map_record,
					asr->mm_generations.len - 2);
				rc = clone_latest_generation(old_map, map,
							     actions->allocator);
				if (rc != CAFE_OK)
					return rc;
			}
			map->parent_generation = parent_gen;
		}
	}

done:
	map->committed = true;
	*current_gen = map->generation;
	return CAFE_OK;
}


/**
 *  Create a symbol info table entry with given symbol name.
 */
static
struct symbol_cache_entry *create_symentry(struct symresolve_actions *actions, const char *symname)
{
	struct symbol_cache_entry *symentry;
	const size_t len = strlen(symname) + 1;

	symentry = CAFE_ALLOC(actions->allocator, sizeof(*symentry) + len);
	if (symentry == NULL)
		return NULL;

	symentry->address = 0;
	symentry->size = 0;
	symentry->index = actions->next_sym_index++;
	memcpy(&symentry->name, symname, len);

	if (actions->symbol_index != NULL) {
		struct cafe_value_data *new_row;
		struct cafe_normalized_symbol_entry *entry;
		cafe_any_value key[1];

		key[0].ull = symentry->index;
		if (cafe_dict_lookup(actions->symbol_index, key) != NULL)
			return symentry;
		new_row = cafe_dict_get(actions->symbol_index, key);
		if (new_row == NULL)
			return NULL;

		entry = (struct cafe_normalized_symbol_entry *) new_row->u.data;
		entry->index = symentry->index;
		entry->name = CAFE_STRDUP(actions->allocator, symname);
		entry->object_name = CAFE_STRDUP(actions->allocator, symname);
		entry->object_address = symentry->address;
	}

	return symentry;
}


/**
 *  Retrieve handle to the symbol info table entry representing the error
 * condition where the given address occurs outside of any known memory region.
 */
static
struct symbol_cache_entry *get_unmapped_entry(struct symresolve_actions *actions)
{
	struct symbol_cache_entry *symentry = actions->unmapped_entry;

	if (symentry == NULL) {
		symentry = actions->unmapped_entry = create_symentry(actions, UNMAPPED_STR);
	}

	return symentry;
}

/**
 *  Retrieve handle to the symbol info table entry representing the error
 * condition where the given address occurs inside an unnamed memory region.
 */
static
struct symbol_cache_entry *get_unnamed_entry(struct symresolve_actions *actions)
{
	struct symbol_cache_entry *symentry = actions->unnamed_entry;

	if (symentry == NULL) {
		symentry = actions->unnamed_entry = create_symentry(actions, UNNAMED_STR);
	}

	return symentry;
}


#if defined(HAVE_TCF) && ENABLE_ELF
/**
 *  A bsearch() callback used to find the memory_map_region corresponding to a
 * given address.
 */
static
int region_bsearch(const void *arg1, const void *arg2)
{
	const unsigned long long *key = arg1;
	const struct memory_map_region *region = arg2;

	if (*key < region->start_addr)
		return -1;

	if (*key >= (region->start_addr + region->len))
		return 1;

	return 0;
}


static int is_valid_elf_symbol(ELF_SymbolInfo * info) {
    switch (info->type) {
    case STT_NOTYPE:
        if (info->name == NULL || info->name[0] == '$') break;
        return 1;
    case STT_FUNC:
    case STT_OBJECT:
        return info->name != NULL;
    }
    return 0;
}


/**
 *  Given a file and a link time address, locate the closest corresponding ELF
 * symbol.  Returns true on success, populating the "sym" and "function_offset"
 * parameters with the results, and false on failure.
 */
static
bool locate_sym_in_file(struct symresolve_actions *actions,
			ContextAddress lt_addr, const ELF_File *file,
			ELF_Section *section,
			ELF_SymbolInfo *sym, uint64_t *function_offset)
{
	unsigned int i;

	/* locate the corresponding section, and resolve the virtualized
	   address */
	if (section == NULL) {
		for (i = 1; i < file->section_cnt; ++i) {
			ELF_Section *s = &file->sections[i];
			if ((s->flags & SHF_ALLOC) == 0)
				continue;
			if (s->addr <= lt_addr && s->addr + s->size > lt_addr) {
				section = s;
				break;
			}
		}
	}
	if (section != NULL) {
		elf_find_symbol_by_address(section, lt_addr, sym);
		if (!is_valid_elf_symbol(sym) || sym->size == 0) {
			ELF_SymbolInfo good_sym = *sym;
			while (sym->addr_index > 0) {
				elf_prev_symbol_by_address(sym);
				if (!is_valid_elf_symbol(sym))
					continue;
				if (lt_addr < sym->value)
					break;
				if (lt_addr < sym->value + sym->size)
					goto success;
				if (sym->size == 0) {
					/* remember the symbol but keep searching for better match */
					if ((sym->type == STT_FUNC && good_sym.type == STT_NOTYPE)
							|| !is_valid_elf_symbol(&good_sym))
						good_sym = *sym;
					continue;
				}
				break;
			}
			if (is_valid_elf_symbol(&good_sym)) {
				*sym = good_sym;
				goto success;
			}
			return false;
		}
	success:
		if (sym->size > 0 && lt_addr >= sym->value + sym->size) {
			return false;
		}
		/* ensure the returned address is relative to the map's
		   base address */
		*function_offset = lt_addr - sym->value;

		return true;
	}

	error(actions, "Unable to locate section containing address "
	      "0x%"PRIx64" in file %s\n", (uint64_t)lt_addr, file->name);
	return false;
}


struct tcf_locate_sym_arg {
	struct symresolve_actions *actions;
	unsigned long long rt_address;
	bool object_cache_updated;
	struct memory_map_record *map;
	const char *file_name;
	uint64_t function_offset;
	uint64_t sym_size;
	char name_buf[512];
	bool found;
};


/**
 *  Given a file and a file offset, locate the closest corresponding ELF
 * symbol.  Sets arg.found to true on success, its sym_size, function_offset,
 * and name parameters with the results.  Meant to be used as a TCF event
 * callback.
 */
static
void tcf_locate_sym(void *arg)
{
	struct tcf_locate_sym_arg *a = arg;
	ELF_File *file;
	ELF_Section *section = NULL;
	ContextAddress lt_addr;
	ELF_SymbolInfo sym;

	assert(a != NULL);
#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	if ((ctx_map != a->map) || a->object_cache_updated) {
		ctx_channel = a->actions->c;
		ctx_map = a->map;
		memory_map_event_mapping_changed(a->actions->ctx);
	}

	lt_addr = elf_map_to_link_time_address(a->actions->ctx, a->rt_address, 1, &file, &section);
	if (section == NULL)
		goto fail;
#else
	/* WRS_TODO this is just to satisfy the legacy linux platform build */
	file = elf_open(a->file_name);
	if (file == NULL) {
		error(a->actions, "Unable to parse ELF information from \"%s\": %s\n",
		      a->file_name, errno_to_str(errno));
		goto fail;
	}
	lt_addr = a->rt_address;
#endif /* defined(ENABLE_ContextMux) && ENABLE_ContextMux */

	if (locate_sym_in_file(a->actions, lt_addr, file, section, &sym, &a->function_offset))
		goto succeed;

#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	goto fail;
#else
	/* WRS_TODO this is just to satisfy the legacy linux platform build */
	/* try using any available debug info files */
	if (file->debug_info_file_name == NULL)
		goto fail;

	file = elf_open(file->debug_info_file_name);
	if (file == NULL) {
		error(a->actions, "Unable to parse ELF information from \"%s\": %s\n",
		      file->debug_info_file_name, errno_to_str(errno));
		goto fail;
	}

	if (!locate_sym_in_file(a->actions, lt_addr, file, section, &sym, &a->function_offset))
		goto fail;
#endif /* defined(ENABLE_ContextMux) && ENABLE_ContextMux */
succeed:
	strncpy(a->name_buf, sym.name, sizeof(a->name_buf));
	a->name_buf[sizeof(a->name_buf) - 1] = 0;

	/* some assembly routines do not record their size in the symbol table,
	   e.g., vmlinux */
	if (sym.size == 0) {
		ELF_SymbolInfo next_sym = sym;
		do {
			elf_next_symbol_by_address(&next_sym);
		} while (next_sym.value <= sym.value && next_sym.sym_section != NULL);
		if (next_sym.value > sym.value)
			a->sym_size = next_sym.value - sym.value;
		else
			if ((strcmp(sym.section->name, ".fini") == 0) ||
			    (strcmp(sym.section->name, ".init") == 0))
				a->sym_size = sym.section->size;
			else
				a->sym_size = a->function_offset + 1;
	} else
		a->sym_size = sym.size;

	a->found = true;
	return;

fail:
	a->found = false;
}


struct tcf_cache_object_arg {
	struct symresolve_actions *actions;
	struct memory_map_region *region;
};


/**
 *  Build an object cache for the given memory map region.
 */
static
void tcf_cache_object(void *arg)
{
	struct tcf_cache_object_arg *a = arg;
	const char *local_file_name;
	struct symbol_cache *sc;
	char *tmp;
	const ELF_File *file;

	tmp = apply_path_map(a->actions->c, NULL,
				   (char *) a->region->file_name,
				   PATH_MAP_TO_LOCAL);
	if (tmp != NULL)
		local_file_name = tmp;
	else
		local_file_name = a->region->file_name;

	if (!symbol_cache_new(&sc, local_file_name,
			      a->region->file_name, a->actions))
	{
		a->region->cache = NULL;
		return;
	}

	if (sc->local_filename[0] == '\0')
		goto out;

	file = elf_open(local_file_name);
	if (file == NULL)
		goto out;

	sc->mtime = file->mtime;

	/* VxWorks DKMs lack program headers.  In this case we just need to
	   name the mapped section so TCF knows how to interpret the file. */
	if (a->region->file_offset == SYMRESOLVE_OFFSET_UNKNOWN) {
		if (file->pheader_cnt != 0) {
			error(a->actions, "Region marked with unknown file"
			      "offset, but contains program headers: %s\n",
			      local_file_name);
			goto out;
		}

		sc->is_mapped_text_section = true;
	}

out:
	symbol_cache_reference(sc, a->region);
}


#  if defined(ENABLE_ContextMux) && ENABLE_ContextMux
/**
 *  Returns true iff the provided path is absolute.  Makes an attempt to
 * provide the correct answer for paths coming from any build host type.
 */
static
bool cafe_is_absolute_path(const char *path)
{
	size_t len;
	unsigned int i;

	/* check for absolute unix paths */
	if (path[0] == '/')
		return true;

	/* check for absolute windows paths */
	if (path[0] == '\\')
		return true;

	len = strlen(path);
	/* check for drive-rooted windows paths */
	for (i = 1; i < 3; ++i) {
		static const char *WIN_DRIVE_IDENT = ":\\";
		static const char *WIN_DRIVE_IDENT2 = ":/";

		if (i + 2 >= len)
			break;

		if ( (memcmp(&path[i], WIN_DRIVE_IDENT, 2) == 0) ||
		     (memcmp(&path[i], WIN_DRIVE_IDENT2, 2) == 0) )
			return true;
	}

	return false;
}


struct tcf_locate_lineinfo_arg {
	struct symresolve_actions *actions;
	struct memory_map_record *map;
	bool object_cache_updated;
	unsigned long long address;
	uint64_t size;
	unsigned long line;
	unsigned long column;
	char source_file[4096];
	bool found;
};


/**
 *  Given a memory map and an address, locate the corresponding line
 * information.  Sets arg.found to true on success.
 */
static
void tcf_locate_lineinfo(void *arg)
{
	struct lineinfo_arg callback_arg;
	struct tcf_locate_lineinfo_arg *a = arg;

	assert(a != NULL);

	if ((ctx_map != a->map) || a->object_cache_updated) {
		ctx_channel = a->actions->c;
		ctx_map = a->map;
		memory_map_event_mapping_changed(a->actions->ctx);
	}
	callback_arg.found = false;
	if (address_to_line(a->actions->ctx,
			    a->address, a->address + 1,
			    TCF_line_number_callback, &callback_arg))
	{
		struct memory_map_region *region =
			bsearch(&a->address, a->map->regions.data,
				a->map->regions.len,
				a->map->regions.element_size, region_bsearch);

		if (region == NULL)
			error(a->actions, "Error parsing dwarf info: no memory "
			      "region found for 0x%llx\n", a->address);
		else
			error(a->actions, "Error parsing dwarf info from "
			      "%s: %s\n",
			      region->file_name, errno_to_str(errno));
		a->line = 0;
		a->column = 0;
		snprintf(a->source_file, sizeof(a->source_file), "[%s]",
			 errno_to_str(errno));
		goto fail;
	}
	if (!callback_arg.found) {
		a->line = 0;
		a->column = 0;
		strncpy(a->source_file, "[SOURCE INFO NOT FOUND]",
			sizeof(a->source_file));
		goto fail;
	}

	if (cafe_is_absolute_path(callback_arg.file_name)) {
		strncpy(a->source_file, callback_arg.file_name,
			sizeof(a->source_file));
	} else {
		size_t len1 = strlen(callback_arg.file_dir);
		size_t len2 = strlen(callback_arg.file_name) + 1;
		if (len1 + len2 + 1 < sizeof(a->source_file)) {
			memcpy(a->source_file, callback_arg.file_dir, len1);
			a->source_file[len1] = '/';
			memcpy(a->source_file + len1 + 1,
			       callback_arg.file_name, len2);
		} else {
			strncpy(a->source_file, callback_arg.file_name, sizeof(a->source_file));
		}
	}

	a->line = callback_arg.line;
	a->column = callback_arg.column;
	a->address = callback_arg.address;
	a->size = callback_arg.size;
	a->found = true;
	return;

fail:
	a->found = false;
}


/**
 *  Look up source information for the given (address, mapped region) using any
 * available cached data.  Populates cache with the results before returning.
 * Returns the cached information on success and NULL on failure.
 */
static
const struct lineinfo_cache_entry *lineinfo_cache_lookup(
	struct symresolve_actions *actions, struct symbol_cache *cache,
	uint64_t offs, unsigned long long address,
	struct memory_map_record *map, bool object_cache_updated)
{
	struct tcf_locate_lineinfo_arg a;
	struct lineinfo_cache_entry *lineinfo;
	struct lineinfo_cache_entry *lineinfo_low;
	unsigned long index;
	size_t len;

	if (pcache_size(&cache->lineinfo) != 0) {
		unsigned long low, hi, mid;
		bool found = false;

		low = 0;
		hi = pcache_size(&cache->lineinfo) - 1;

		lineinfo = pcache_val(&cache->lineinfo, hi);
		lineinfo_low = pcache_val(&cache->lineinfo, low);

		if ((offs >= lineinfo_low->address) && (offs < lineinfo->address)) {
			while((low+1) < hi) {
				mid = (low+hi)/2;
				lineinfo = pcache_val(&cache->lineinfo, mid);
				if (offs >= lineinfo->address) {
					low = mid;
				} else {
					hi = mid;
				}
			}
			lineinfo = pcache_val(&cache->lineinfo, low);
			if ((offs >= lineinfo->address) && (offs < (lineinfo->address + lineinfo->size))) {
				found = true;
			}

		} else if (offs >= lineinfo->address) {
			if (offs < (lineinfo->address + lineinfo->size)) {
				/* The last cache element matches. */
				found = true;
			} else {
				/* Not found and insert point is at the end of the cache. */
				low = hi;
			}
		}

		if (found) {
#if defined(SYMRESOLVE_PROFILE_CACHE)
			++actions->lhits;
#endif
			return lineinfo;
		}
		index = low;

		lineinfo = pcache_val(&cache->lineinfo, index);
		if (offs > lineinfo->address) {
			/* due to the way the binary search works, we may end
			   up with either the entry before or the entry after
			   the desired insertion point.  After this adjustment,
			   "index" points at the insertion slot, any current
			   contents should be shifted down. */
			++index;
		}
	} else
		index = 0;

#if defined(SYMRESOLVE_PROFILE_CACHE)
	++actions->lmisses;
#endif
	a.address = address;
	a.actions = actions;
	a.map = map;
	a.object_cache_updated = object_cache_updated;
	post_event_sync(tcf_locate_lineinfo, &a);

	if (!a.found)
		return NULL;

	/* assert(address >= a.address && address < a.address + a.size); */
	if (!(address >= a.address && address < a.address + a.size)) {
#if !defined(NDEBUG)
		(void)fputs("Line info lookup error: address not within code range",
			     stderr);
#endif
		return NULL;
	}

	len = strlen(a.source_file) + 1;
	lineinfo = pcache_insert(&cache->lineinfo, index,
				 sizeof(*lineinfo) + len);
	if (lineinfo == NULL)
		return NULL;

	/* Correct offset to the start of the code range */
	lineinfo->address = offs - (address - a.address);
	lineinfo->size = a.size;
	lineinfo->line = a.line;
	lineinfo->column = a.column;
	lineinfo->index = actions->next_lineinfo_index++;
	memcpy(&lineinfo->file, a.source_file, len);

	if (actions->lineinfo_index != NULL) {
		struct cafe_value_data *new_row;
		struct cafe_normalized_lineinfo_entry *entry;
		cafe_any_value key[1];

		key[0].ull = lineinfo->index;
		assert(cafe_dict_lookup(actions->lineinfo_index, key) == NULL);
		new_row = cafe_dict_get(actions->lineinfo_index, key);
		if (new_row == NULL) {
			(void)fputs("Error adding normalized symbol entry", stderr);
			return lineinfo;
		}

		entry = (struct cafe_normalized_lineinfo_entry *) new_row->u.data;
		entry->index = lineinfo->index;
		entry->line = lineinfo->line;
		entry->column = lineinfo->column;
		entry->file_name = CAFE_STRDUP(actions->allocator, lineinfo->file);
	}

	return lineinfo;
}
#endif


/**
 * Clear all references to unresolved cache entries
 * We don't touch the cache of resolved object files, because this would cause
 * inconsistency with symbol data already sent out to the client.
 */
static
void object_cache_flush_unresolved(struct symresolve_actions *actions) {
	struct cafe_hashtbl_iter iter;
	cafe_hashtbl_iter_init(actions->address_space_map, &iter);
	while (cafe_hashtbl_iter_next(actions->address_space_map, &iter) != 0) {
		struct address_space_record *asr = iter.val;
		size_t i;

		if (asr == NULL)
			continue;

		i = asr->mm_generations.len;
		while (i-- > 0) {
			struct memory_map_record *m = &cafe_array_val(
				&asr->mm_generations, struct memory_map_record, i);
			struct cafe_array regions = m->regions;
			size_t j;

			for (j=0; j<regions.len; ++j) {
				struct memory_map_region *region = &cafe_array_val(
						&regions, struct memory_map_region, j);
				struct symbol_cache *sc = region->cache;
				/* Unresolved object files have an empty local file name */
				if (sc != NULL && sc->local_filename[0] == 0) {
					symbol_cache_dereference(region, actions);
				}
			}
		}
	}
}

/**
 * Look up file mapping for the given mapped region using any
 * available cached data.  Returns the cached information on success or NULL on
 * failure.
 */
static
struct symbol_cache *object_cache_lookup(struct symresolve_actions *actions,
					 struct memory_map_region *region,
					 bool *cache_updated)
{
	if (actions->path_map_changed) {
		/* Flush caches of unresolved object files */
		actions->path_map_changed = false;
		object_cache_flush_unresolved(actions);
	}

	if (region->cache == NULL) {
		struct symbol_cache *sc;
		sc = cafe_hashtbl_lookup(actions->object_cache,
					 region->file_name);
		if (sc != NULL && sc->refs == 0) {
			/* Flush cache if object file has changed */
			struct stat st;
			if (sc->local_filename[0] == '\0'
					|| stat(sc->local_filename, &st) !=0
					|| sc->mtime != st.st_mtime) {
				sc = cafe_hashtbl_remove(actions->object_cache,
				                         region->file_name,
				                         NULL, NULL);
				assert(sc != NULL);
				symbol_cache_free(sc, (void *) actions->allocator);
				sc = NULL;
			}
		}
		if (sc == NULL) {
			struct tcf_cache_object_arg arg;

			arg.actions = actions;
			arg.region = region;
			post_event_sync(tcf_cache_object, &arg);
			if (region->cache == NULL)
				return NULL;
		} else {
			symbol_cache_reference(sc, region);
		}
		*cache_updated = true;
	} else
		*cache_updated = false;

	return region->cache;
}


static int symbol_cache_entry_bsearch_comparator(const void *a, const void *b)
{
	const struct symbol_cache_entry *x = a;
	const struct symbol_cache_entry *y = *(const struct symbol_cache_entry **)b;

	if (x->address < y->address)
		return -1;

	if (x->address >= y->address + y->size)
		return 1;

	return 0;
}

static struct symbol_cache_entry *symbol_cache_lookup2(
	struct symresolve_actions *actions,
	struct symbol_cache *cache,
	uint64_t offs)
{
	struct symbol_cache_entry key;
	struct symbol_cache_entry **symentry_ptr;

	key.address = offs;
	key.size = 0;
	symentry_ptr = bsearch(&key, &cache->syms.list.data[0],
			       cache->syms.list.len,
			       sizeof *symentry_ptr,
			       symbol_cache_entry_bsearch_comparator);

	return symentry_ptr ? *symentry_ptr : NULL;
}

/*
 * Create an entry for symbols unresolved in an object file.
 */
static
struct symbol_cache_entry* get_unresolved_entry(
		struct symbol_cache* cache, struct symresolve_actions* actions) {

	struct symbol_cache_entry* symentry;
	size_t len;
	char *basename;

	symentry = cache->unresolved_entry;
	if (symentry != NULL)
		return symentry;

	basename = cache->local_filename + strlen(cache->local_filename) - 1;
	while (basename > cache->local_filename && *basename != '/' && *basename != '\\')
		--basename;
	if (basename > cache->local_filename)
		++basename;
	len = strlen(basename) + 16 + 1; /* [unresolved in objfile]\0 */
	symentry = CAFE_ALLOC(actions->allocator, sizeof(*symentry) + len);
	if (symentry == NULL)
		return NULL;

	symentry->address = 0;
	symentry->size = 0;
	symentry->index = actions->next_sym_index++;
	snprintf(symentry->name, len, "[unresolved in %s]", basename);
	cache->unresolved_entry = symentry;

	if (actions->symbol_index != NULL) {
		struct cafe_value_data *new_row;
		struct cafe_normalized_symbol_entry *entry;
		cafe_any_value key[1];

		key[0].ull = symentry->index;
		if (cafe_dict_lookup(actions->symbol_index, key) != NULL)
			return symentry;
		new_row = cafe_dict_get(actions->symbol_index, key);
		if (new_row == NULL)
			return NULL;
		entry = (struct cafe_normalized_symbol_entry *) new_row->u.data;
		entry->index = symentry->index;
		entry->name = CAFE_STRDUP(actions->allocator, symentry->name);
		entry->object_name = CAFE_STRDUP(actions->allocator, cache->local_filename);
		entry->object_address = symentry->address;
		if (actions->symbol_index->type->element->type->elements > 4) {
			entry->symbol_lineinfo_index = -1ULL;
		}
	}
	return symentry;
}

/**
 *  Look up symbol information for the given (address, mapped region) using any
 * available cached data.  Populates cache with the results before returning.
 * Returns the cached information on success and NULL on failure.
 */
static
const struct symbol_cache_entry *symbol_cache_lookup(
	struct symresolve_actions *actions,
	struct symbol_cache *cache,
	uint64_t offs, unsigned long long address,
	struct memory_map_record *map, bool *object_cache_updated)
{
	struct symbol_cache_entry *symentry;
	size_t len;
	unsigned long index;
	struct tcf_locate_sym_arg arg;

	if (pcache_size(&cache->syms) != 0) {
		unsigned long low, hi, mid;
		bool found = false;

		low = 0;
		hi = pcache_size(&cache->syms) - 1;
		do {
			mid = (low + hi) / 2;
			symentry = pcache_val(&cache->syms, mid);
			if (symentry->address > offs) {
				if (mid == low)
					break;
				hi = mid - 1;
			} else if (symentry->address + symentry->size < offs) {
				if (mid == hi)
					break;
				low = mid + 1;
			} else {
				found = true;
				break;
			}

			if (low == hi) {
				symentry = pcache_val(&cache->syms, low);
				if ((symentry->address <= offs) && (symentry->address + symentry->size > offs)) {
					found = true;
					break;
				}
			}
				
		} while (1);

		if (found) {
#if defined(SYMRESOLVE_PROFILE_CACHE)
			++actions->hits;
#endif
			return symentry;
		}
		index = low;

		symentry = pcache_val(&cache->syms, index);
		if (offs > symentry->address)
			/* due to the way the binary search works, we may end
			   up with either the entry before or the entry after
			   the desired insertion point.  After this adjustment,
			   "index" points at the insertion slot, any current
			   contents should be shifted down. */
			++index;
	} else
		index = 0;

#if defined(SYMRESOLVE_PROFILE_CACHE)
	++actions->misses;
#endif

	if (cache->local_filename[0] != 0) {
		char* symname;
		arg.actions = actions;
		arg.rt_address = address;
		arg.map = map;
		arg.object_cache_updated = *object_cache_updated;
		arg.file_name = cache->local_filename;
		post_event_sync(tcf_locate_sym, &arg);
		*object_cache_updated = false;
		if (!arg.found) {
			return get_unresolved_entry(cache, actions);
		}

		if ((symentry=symbol_cache_lookup2(actions, cache, offs - arg.function_offset)) != NULL) {
			/* Found symbol in cache although file offset is not within symbol bounds */
			if (offs - arg.function_offset == symentry->address && arg.sym_size == symentry->size)
				return symentry;
			return get_unresolved_entry(cache, actions);
		}
		symname = arg.name_buf;
#if defined(HAVE_DEMANGLER)
		/* Demangle C++ symbols */
		if ((symname[0] == '_') && ((symname[1] == 'Z') || (symname[1] == '_'))) {
			char * signature = demangle(symname, DMGL_STYLE_AUTO, DMGL_MODE_COMPLETE);
			if (signature != NULL) {
				if (signature[0] != '\0') {
					symname = signature;
				} else {
					demangleResultFree(signature);
				}
			}
		}
#endif /* defined(HAVE_DEMANGLER) */
		len = strlen(symname) + 1;
		symentry = pcache_insert(&cache->syms, index,
					 sizeof(*symentry) + len);
		if (symentry == NULL)
			return NULL;
		memcpy(&symentry->name, symname, len);
#if defined(HAVE_DEMANGLER)
		if (symname != arg.name_buf) {
			demangleResultFree(symname);
		}
#endif /* defined(HAVE_DEMANGLER) */
	} else {
		// insert a single symbol entry spanning the whole address range
		char *basename = cache->remote_filename + strlen(cache->remote_filename) - 1;
		while (basename > cache->remote_filename && *basename != '/' && *basename != '\\')
			--basename;
		if (basename > cache->remote_filename)
			++basename;
		len = strlen(basename) + 12 + 1; /* [objfile not found]\0 */
		symentry = pcache_insert(&cache->syms, index,
					 sizeof(*symentry) + len);
		if (symentry == NULL)
			return NULL;
		arg.function_offset = offs;
		arg.sym_size = ULLONG_MAX;
		snprintf(symentry->name, len, "[%s not found]", basename);
	}

	symentry->address = offs - arg.function_offset;
	symentry->size = arg.sym_size;
	symentry->index = actions->next_sym_index++;

	if (actions->symbol_index != NULL) {
		struct cafe_value_data *new_row;
		struct cafe_normalized_symbol_entry *entry;
		cafe_any_value key[1];

		key[0].ull = symentry->index;
		assert(cafe_dict_lookup(actions->symbol_index, key) == NULL);
		new_row = cafe_dict_get(actions->symbol_index, key);
		if (new_row == NULL) {
			(void)fputs("Error adding normalized symbol entry", stderr);
			return symentry;
		}

		entry = (struct cafe_normalized_symbol_entry *) new_row->u.data;
		entry->index = symentry->index;
		entry->name = CAFE_STRDUP(actions->allocator, symentry->name);
		entry->object_name = CAFE_STRDUP(actions->allocator, cache->remote_filename);
		entry->object_address = symentry->address;
		if (actions->symbol_index->type->element->type->elements > 4) {
			/* Extended symbol entry with lineinfo for the symbol */
			const struct lineinfo_cache_entry *lineinfo = NULL;
#  if defined(ENABLE_ContextMux) && ENABLE_ContextMux
			if (cache->local_filename[0] != 0 && actions->lineinfo_index != NULL) {
				lineinfo = lineinfo_cache_lookup(actions, cache,
								 offs - arg.function_offset,
								 address - arg.function_offset,
								 map, *object_cache_updated);
				*object_cache_updated = false;
			}
#  endif
			if (lineinfo != NULL) {
				entry->symbol_lineinfo_index = lineinfo->index;
			} else {
				entry->symbol_lineinfo_index = -1ULL;
			}
		}
	}
	return symentry;
}
#endif  /* if defined(HAVE_TCF) && ENABLE_ELF */


/**
 *  Given the expected result types, populate the result structure with the
 * calculated values.
 */
static
void populate_result(unsigned int type_map[10],
		     const struct cafe_value_type *result_type,
		     const char *object_filename,
		     const struct symbol_cache_entry *symentry,
		     const struct lineinfo_cache_entry *lineinfo,
		     unsigned int index, unsigned long long original_address,
		     uint64_t object_address,
		     const struct cafe_allocator *allocator,
		     void *result)
{
	unsigned int j;
	assert(original_address != CAFE_STACK_MARKER_KERNEL);
	for (j = 0; j < result_type->elements; ++j) {
		struct cafe_vm_attr *result_attr =  result_type->attributes + j;
		void *result_field = (char *)result + result_attr->offset;

		switch (type_map[j]) {
		case RESOLVED_ADDRESS_FIELD_OBJECT_NAME:
			*(char **)result_field =
				CAFE_STRDUP(allocator, object_filename);
			break;
		case RESOLVED_ADDRESS_FIELD_SYMBOL_INDEX:
			if (symentry != NULL)
				*(unsigned long long *)result_field = symentry->index;
			else
				*(unsigned long long *)result_field = -1ULL;
			break;
		case RESOLVED_ADDRESS_FIELD_LINEINFO_INDEX:
			if (lineinfo != NULL)
				*(unsigned long long *)result_field = lineinfo->index;
			else
				*(unsigned long long *)result_field = -1ULL;
			break;
		case RESOLVED_ADDRESS_FIELD_SYMBOL_NAME:
			if (symentry != NULL)
				*(char **)result_field =
					CAFE_STRDUP(allocator, symentry->name);
			else {
				char namebuf[32];
				(void)snprintf(namebuf, sizeof(namebuf), "0x%llx",
					 original_address);
				*(char **)result_field =
					CAFE_STRDUP(allocator, namebuf);
			}
			break;

		case RESOLVED_ADDRESS_FIELD_OBJECT_ADDRESS:
			if (symentry != NULL)
				*(unsigned long long *)result_field =
					symentry->address;
			else
				*(unsigned long long *)result_field =
					original_address;
			break;
		case RESOLVED_ADDRESS_FIELD_SYMBOL_OFFSET:
			if (symentry != NULL)
				*(unsigned long long *)result_field =
					object_address - symentry->address;
			else
				*(unsigned long long *)result_field = 0;
			break;
		case RESOLVED_ADDRESS_FIELD_STACK_INDEX:
			*(unsigned long long *)result_field = index;
			break;
		case RESOLVED_ADDRESS_FIELD_ORIGINAL_ADDRESS:
			*(unsigned long long *)result_field = original_address;
			break;

		case RESOLVED_LINEINFO_FIELD_FILE_NAME:
		{
			const char *file;
			if (lineinfo != NULL)
				file = lineinfo->file;
			else
				file = "[SOURCE INFO NOT FOUND]";
			*(char **)result_field = CAFE_STRDUP(allocator, file);
			break;
		}
		case RESOLVED_LINEINFO_FIELD_LINE_NUMBER:
			if (lineinfo != NULL)
				*(unsigned long long *)result_field =
					lineinfo->line;
			else
				*(unsigned long long *)result_field = 0;

			break;
		case RESOLVED_LINEINFO_FIELD_COLUMN_NUMBER:
			if (lineinfo != NULL)
				*(unsigned long long *)result_field =
					lineinfo->column;
			else
				*(unsigned long long *)result_field = 0;
			break;

		default:
			/* coverity[array_null] */
			assert(!"unexpected address resolve field");
		}
	}
}


/**
 *  Resolve the given callstack in the context of the given address space's
 * memory map, returning the result in the resolved_stack parameter.
 */
int symresolve_callstack(struct symresolve_actions *actions,
			 unsigned long long address_space_uid,
			 unsigned long long mm_generation,
			 const struct cafe_vm_dynarray *raw_stack,
			 const struct cafe_value_type *result_type,
			 struct cafe_vm_dynarray *result)
{
#if defined(HAVE_TCF) && ENABLE_ELF
	struct address_space_record *asr;
	struct memory_map_record *cur_map, *parent_map, *map;
#endif
	unsigned int type_map[10];
	bool need_syminfo, need_lineinfo;
	const unsigned long long *stack = raw_stack->p;
	const struct cafe_value_type *result_element_type;
	void *result_row;
	unsigned int n_resolved;
	unsigned int i;
	int rc;
#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	int leaf_pc;
#endif

	assert(actions != NULL);
	assert(result_type != NULL);
	assert(result != NULL);
	assert(result_type->kind == CAFE_TYPE_DYNARRAY);
	result_element_type = cafe_base_type(cafe_element_type(result_type));
	assert(result_element_type != NULL);
	assert(result_element_type->kind == CAFE_TYPE_STRUCT);

	/* prepare our output array */
	rc = cafe_vm_dynarray_set_size(result, raw_stack->l, result_type);
	if (rc != CAFE_OK)
		return rc;

	if (raw_stack->l == 0)
		return CAFE_OK;

	/* determine which types we need to calculate */
	need_syminfo = false;
	need_lineinfo = false;
	for (i = 0; i < result_element_type->elements; ++i) {
		const struct cafe_vm_attr *attr =
			result_element_type->attributes + i;
		assert(attr->kind == CAFE_ATTR_VAR);
		assert(i < sizeof(type_map)/sizeof(unsigned int));
		if (strcmp(attr->ident->name, "object_name") == 0) {
			type_map[i] = RESOLVED_ADDRESS_FIELD_OBJECT_NAME;
		} else if (strcmp(attr->ident->name, "symbol_index") == 0) {
			type_map[i] = RESOLVED_ADDRESS_FIELD_SYMBOL_INDEX;
			need_syminfo = true;
		} else if (strcmp(attr->ident->name, "lineinfo_index") == 0) {
			type_map[i] = RESOLVED_ADDRESS_FIELD_LINEINFO_INDEX;
			need_lineinfo = true;
		} else if (strcmp(attr->ident->name, "object_address") == 0) {
			type_map[i] = RESOLVED_ADDRESS_FIELD_OBJECT_ADDRESS;
			need_syminfo = true;
		} else if (strcmp(attr->ident->name, "symbol_name") == 0) {
			type_map[i] = RESOLVED_ADDRESS_FIELD_SYMBOL_NAME;
			need_syminfo = true;
		} else if (strcmp(attr->ident->name, "symbol_offset") == 0) {
			type_map[i] = RESOLVED_ADDRESS_FIELD_SYMBOL_OFFSET;
			need_syminfo = true;
		} else if (strcmp(attr->ident->name, "stack_index") == 0) {
			type_map[i] = RESOLVED_ADDRESS_FIELD_STACK_INDEX;
			need_syminfo = true;
		} else if (strcmp(attr->ident->name, "original_address") == 0) {
			type_map[i] = RESOLVED_ADDRESS_FIELD_ORIGINAL_ADDRESS;
			need_syminfo = true;
		} else if (strcmp(attr->ident->name, "source_name") == 0) {
			type_map[i] = RESOLVED_LINEINFO_FIELD_FILE_NAME;
			need_lineinfo = true;
		} else if (strcmp(attr->ident->name, "line_number") == 0) {
			type_map[i] = RESOLVED_LINEINFO_FIELD_LINE_NUMBER;
			need_lineinfo = true;
		} else if (strcmp(attr->ident->name, "column_number") == 0) {
			type_map[i] = RESOLVED_LINEINFO_FIELD_COLUMN_NUMBER;
			need_lineinfo = true;
		} else {
			/* coverity[array_null] */
			assert(!"unexpected symbol resolution field name");
			return CAFE_ERROR;
		}
	}

#if defined(HAVE_TCF) && ENABLE_ELF
	asr = cafe_hashtbl_lookup(actions->address_space_map,
				  &address_space_uid);
	if (asr == NULL) {
		/* WB4-4383 Simple heuristic: Try address_space_uid=1 assuming
		 * this is the id of the kernel map (which is true for SB).
		 */
		address_space_uid = 1;
		asr = cafe_hashtbl_lookup(actions->address_space_map,
					  &address_space_uid);
		if (asr == NULL)
			goto nomap;
		mm_generation = asr->mm_generations.len - 1;
	}

	cur_map = NULL;
	for (i = 0; i < asr->mm_generations.len; ++i) {
		struct memory_map_record *trav =
			&cafe_array_val(&asr->mm_generations,
					struct memory_map_record, i);
		if (trav->generation == mm_generation) {
			cur_map = trav;
			break;
		}
	}
	if (cur_map == NULL)
		goto nomap;

	asr = cafe_hashtbl_lookup(actions->address_space_map,
				  &cur_map->parent_uid);
	if (asr != NULL) {
		parent_map = NULL;
		for (i = 0; i < asr->mm_generations.len; ++i) {
			struct memory_map_record *trav =
				&cafe_array_val(&asr->mm_generations,
						struct memory_map_record, i);
			if (trav->generation == cur_map->parent_generation) {
				parent_map = trav;
				break;
			}
		}
	} else
		parent_map = cur_map;

	map = cur_map;
	result_row = result->p;
	n_resolved = 0;
#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	leaf_pc = 1;
#endif

	for (i = 0; i < raw_stack->l; ++i) {
		static uint64_t last_unmapped_address = -1ULL;
		struct symbol_cache *object_cache;
		const struct symbol_cache_entry *symentry = NULL;
		const struct lineinfo_cache_entry *lineinfo = NULL;
		const char *object_name;
		struct memory_map_region *region;
		bool object_cache_updated;
		/* offset of address from beginning of mapped region */
		uint64_t offs;

		if (stack[i] == CAFE_STACK_MARKER_KERNEL) {
			map = parent_map;
			continue;
		} else if (stack[i] == CAFE_STACK_MARKER_USER) {
			map = cur_map;
			continue;
		}

		region = bsearch(&stack[i], map->regions.data,
				 map->regions.len,
				 map->regions.element_size, region_bsearch);

		if (region == NULL) {
			char *msg;

			symentry = get_unmapped_entry(actions);
			object_name = UNMAPPED_STR;
			offs = stack[i];

			if (last_unmapped_address == offs) {
				/* Skip same unmapped address */
				goto fail;
			}
			last_unmapped_address = offs;

			msg = cafe_mprintf(actions->allocator,
					   "Address found in unmapped region (%d 0x%llx)\n",
					   map == cur_map, stack[i]);

			if (cafe_aggregator_debug & 4) {
				/* Detailed logging: print also memory map */
				struct cafe_byte_buffer buf;
				unsigned int j;

				if (cafe_bbuf_init(&buf, actions->allocator, 512) != CAFE_OK)
					goto fail;
				if (msg != NULL)
					cafe_bbuf_append(&buf, msg, strlen(msg));
				CAFE_FREE(actions->allocator, msg);

				for (j = 0; j < map->regions.len; ++j) {
					struct memory_map_region *r =
						&cafe_array_val(&map->regions,
								struct memory_map_region,
								j);
					msg = cafe_mprintf(actions->allocator,
							   "\t%llx - %llx  %s\n",
							   r->start_addr,
							   r->start_addr + r->len,
							   r->file_name);
					if (msg != NULL)
						cafe_bbuf_append(&buf, msg,
								 strlen(msg));
					CAFE_FREE(actions->allocator, msg);
				}
				cafe_bbuf_write(&buf, 0);
				actions->error_handler(buf.data,
						       actions->error_handler_arg);
				cafe_bbuf_release(&buf);
			} else if (msg != NULL) {
				actions->error_handler(msg,
						       actions->error_handler_arg);
				CAFE_FREE(actions->allocator, msg);
			}
			goto fail;
		} else if (region->file_name == NULL) {
			symentry = get_unnamed_entry(actions);
			offs = stack[i];
			goto fail;
		}

		object_cache = object_cache_lookup(actions, region,
						   &object_cache_updated);
		object_name = region->file_name;
		offs = stack[i] - region->start_addr + region->file_offset;
		if (object_cache == NULL) {
			error(actions, "No object cache for %s", object_name);
			goto fail;
		}

		if (need_syminfo)
			symentry = symbol_cache_lookup(actions, object_cache,
						       offs, stack[i], map,
						       &object_cache_updated);

#  if defined(ENABLE_ContextMux) && ENABLE_ContextMux
		if (need_lineinfo && (object_cache->local_filename[0] != 0)) {
			unsigned long long call_address;
			uint64_t call_offs;

			if ((symentry != NULL) && (symentry->name != NULL) &&
				((strcmp(symentry->name,"ST_trace_error")==0)
				||(strcmp(symentry->name,"ST_trace_truncated")==0))) {
				call_address = stack[i];
				call_offs = offs;
			} else if(leaf_pc == 1) {
				call_address = stack[i];
				call_offs = offs;
				leaf_pc = 0;
			} else {
				call_address = stack[i] - 1;
				call_offs = offs - 1;
			}

			lineinfo = lineinfo_cache_lookup(actions, object_cache,
							 call_offs, call_address, map,
							 object_cache_updated);
		}
#  endif
	fail:
		assert(stack[i] != CAFE_STACK_MARKER_KERNEL);
		populate_result(type_map, result_element_type,
				object_name, symentry, lineinfo,
				n_resolved, stack[i], offs,
				actions->allocator, result_row);

		++n_resolved;
		result_row = (char *)result_row + result_element_type->size;
	}

	result->l = n_resolved;

	return CAFE_OK;
nomap:
#endif  /* if defined(HAVE_TCF) && ENABLE_ELF */

	result_row = result->p;
	n_resolved = 0;
	for (i = 0; i < raw_stack->l; ++i) {
		if (stack[i] == CAFE_STACK_MARKER_KERNEL)
			continue;
		else if (stack[i] == CAFE_STACK_MARKER_USER)
			continue;

		populate_result(type_map, result_element_type,
				UNMAPPED_STR, get_unmapped_entry(actions), NULL,
				n_resolved, stack[i], stack[i],
				actions->allocator, result_row);
		result_row = (char *)result_row + result_element_type->size;
		++n_resolved;
	}
	result->l = n_resolved;
	if (need_lineinfo)
		need_lineinfo = need_syminfo;
	return CAFE_OK;
}
