/* symresolve_actions.h - symbol resolution engine */

/*
 * Copyright (c) 2011-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

#if !defined(__symresolve_actions_h)
#define __symresolve_actions_h


#include <cafe/cafe.h>


#if defined(__cplusplus)
extern "C" {
#endif


#define CAFE_MM_DELTA_ENTRY_DEF						\
	"S[cafe.mm_delta_entry,v[start_addr,Q," cafe_ADDRESS_VT_NAT "]"	\
	"v[len,Q]v[file_offset,Q]v[file_name,s]]"
#define CAFE_MM_DELTA_ENTRY_REF "t[cafe.mm_delta_entry]"

#define CAFE_MM_DELTA_DEF "S[cafe.mm_delta,v[delta_list,a[" CAFE_MM_DELTA_ENTRY_REF "]]]"
#define CAFE_MM_DELTA_REF "t[cafe.mm_delta]"

#define CAFE_MM_NO_DELTA_DEF "S[cafe.mm_no_delta,," cafe_PRIVATE_VT_NAT "]"
#define CAFE_MM_NO_DELTA_REF "t[cafe.mm_no_delta]"

struct cafe_mm_delta_entry {
	unsigned long long start_addr;
	unsigned long long len;
	unsigned long long file_offset;
	const char *file_name;
};


#define CAFE_NORMALIZED_SYMBOL_DEF					\
	"S[normalized_symbol_entry,"					\
	"k[index,Q]v[name,s]v[object_name,s]"				\
	"v[object_address,Q," cafe_ADDRESS_VT_NAT "]]"
#define CAFE_NORMALIZED_SYMBOL_REF  "t[normalized_symbol_entry]"

/* extension to the normalized symbol entry:
 * additional lineinfo_index of the symbol
 */
#define CAFE_NORMALIZED_SYMBOL_EXT_DEF					\
	"S[normalized_symbol_entry_ext,"				\
	"k[index,Q]v[name,s]v[object_name,s]"				\
	"v[object_address,Q," cafe_ADDRESS_VT_NAT "]"			\
	"v[symbol_lineinfo_index,Q]]"
#define CAFE_NORMALIZED_SYMBOL_EXT_REF  "t[normalized_symbol_entry_ext]"

struct cafe_normalized_symbol_entry {
	unsigned long long index;
	char *name;
	char *object_name;
	unsigned long long object_address;
	unsigned long long symbol_lineinfo_index;
};


#define CAFE_NORMALIZED_LINEINFO_DEF					\
	"S[normalized_lineinfo_entry,"					\
	"k[index,Q]v[line,Q]v[column,Q]"				\
	"v[file_name,s]]"
#define CAFE_NORMALIZED_LINEINFO_REF  "t[normalized_lineinfo_entry]"

struct cafe_normalized_lineinfo_entry {
	unsigned long long index;
	unsigned long long line;
	unsigned long long column;
	char *file_name;
};


struct symresolve_actions;


/* Clients should use this value for file_offset when calling mm_update for a
   memory region about which the base load address and file offset aren't known.
*/
#if defined(_MSC_VER)
  /* Turn off message: warning C4146: unary minus operator applied to unsigned type, result still unsigned */
  #pragma warning (disable: 4146)
#endif
static const unsigned long long SYMRESOLVE_OFFSET_UNKNOWN = -1ULL;


/**
 *  Create a symbol resolver actions engine using a TCF-based backend.  The
 * given mm_table dictionary should have been created prior to this call, and
 * should not be modified by the client directly, until after calling
 * symresolve_actions_delete().  This routine locks the given Channel until a
 * subsequent call to symresolve_actions_delete().
 */
extern int symresolve_actions_create(struct cafe_value_type *delta_vt,
				     const struct cafe_allocator *allocator,
				     struct symresolve_actions **actions);


/**
 *  Releases resources acquired by create().
 */
extern void symresolve_actions_delete(struct symresolve_actions *actions);


/**
 *  Specify the dictionary to be populated with new symbol record entries.
 */
extern void symresolve_set_symindex(struct symresolve_actions *actions,
				    struct cafe_dict_struct *symbol_table);


/**
 *  Specify the dictionary to be populated with new source info record entries.
 */
extern void symresolve_set_lineinfoindex(struct symresolve_actions *actions,
					 struct cafe_dict_struct *lineinfo_table);


/**
 *  Specify the callback to be invoked when symbol resolution needs to
 * communicate an error condition.
 */
extern void symresolve_set_errhandler(struct symresolve_actions *actions,
				      void (*handler_callback) (const char *msg, void *arg),
				      void *arg);


/**
 *   Reset symbol resolver actions engine.
 */
extern void symresolve_actions_reset(struct symresolve_actions *actions);


/**
 *  Update the current memory map table with a single new record.  Used to
 * build delta tables.  Specify the "map" parameter as false when a region
 * should be unmapped.
 */
extern int symresolve_mm_update(struct symresolve_actions *actions,
				unsigned long long address_space_uid,
				/* common memory map fields */
				unsigned long long start_addr,
				unsigned long long size,
				unsigned long long file_offset,
				const char *file_name,
				bool map);


/**
 *  Update the current memory map table with a delta table.
 */
extern int symresolve_mm_delta(struct symresolve_actions *actions,
			       void *delta);


/**
 *  Drop all but the most recent generation of existing memory maps.
 */
extern int symresolve_mm_flush(struct symresolve_actions *actions);


/**
 *  Clients should invoke this routine when a new address space is created
 * using an existing address space's memory map.
 */
extern int symresolve_address_space_clone(struct symresolve_actions *actions,
					  unsigned long long address_space_uid,
					  unsigned long long new_uid);


/**
 *  Signals that the given address space has cleared its memory mappings (e.g.,
 * on a unix exec() call).  This call also sets the "parent" address space for
 * the given address space; the space faulted into whenever a trap is taken
 * (i.e., not the parent process).
 */
extern int symresolve_address_space_clear(struct symresolve_actions *actions,
					  unsigned long long address_space_uid,
					  unsigned long long parent_uid);


/**
 *  Should be called when an address space has been entirely deleted.  Clients
 * can still resolve symbols on past generations of this address space until
 * mm_flush() is called.
 */
extern int symresolve_address_space_delete(struct symresolve_actions *actions,
					   unsigned long long address_space_uid);


/**
 *  Returns the current memory map generation number for the given address
 * space.
 */
extern int symresolve_mm_current_gen(struct symresolve_actions *actions,
				     unsigned long long address_space_uid,
				     unsigned long long *current_gen);


/**
 *  Resolve the given callstack in the context of the given address space's
 * memory map, returning the result in the resolved_stack parameter.  The
 * input parameter "stack" should have a value type of "a[Q]".
 */
extern int symresolve_callstack(struct symresolve_actions *actions,
				unsigned long long address_space_uid,
				unsigned long long mm_generation,
				const struct cafe_vm_dynarray *stack,
				const struct cafe_value_type *result_type,
				struct cafe_vm_dynarray *result);


/**
 *  Generate memory map delta table from current memory maps.  Non
 *  committed generations, i.e. generation for which
 *  symresolve_mm_current_gen() has not been called, are not included
 *  in the delta.
 */
extern int symresolve_mm_gen_delta(struct symresolve_actions *actions,
				   void *delta);

/**
 *  Clear memory map delta table
 */
extern void symresolve_mm_clear_delta(struct symresolve_actions *actions,
				      void *delta_arg);

#if defined(__cplusplus)
}
#endif	/* __cplusplus */

#endif /* __symresolve_actions_h */
