/* validates file_stream_reader playback logic. */

/* Copyright (c) 2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
08jul14,sdt  Used CUT_PRINTF/CUT_PRINT_ERR macro to make clear output
*/

#include "../../src/cafe_os.h"
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <errno.h>
#include <test/cafe_CUnitTest.h>
#include <cafe/ep/file_stream_reader.h>
#include <cafe/ep/text_file_provider.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/autogenerated/file_stream_reader.type.h>
#include <cafe/utils.h>
#include <cafe/ep_worker.h>

#ifdef _WIN32
#include <io.h>
#endif

#define TEST_FSR_A_VT_DEF "S[test_file_stream_playback.A]"
#define TEST_FSR_B_VT_DEF "S[test_file_stream_playback.B]"
#define TEST_FSR_C_VT_DEF "S[test_file_stream_playback.C]"

static unsigned int debug_action_handlers = 0;

static const char *test_content = "\
type t_0 " TEST_FSR_A_VT_DEF "\n\
type t_1 " TEST_FSR_B_VT_DEF "\n\
type t_2 " TEST_FSR_C_VT_DEF "\n\
event t_0 []\n\
event t_0 []\n\
event t_0 []\n\
event t_0 []\n\
event t_1 []\n\
event t_1 []\n\
event t_1 []\n\
event t_0 []\n\
event t_1 []\n\
event t_0 []\n\
event t_0 []\n\
event t_1 []\n";

static char *text_filename;
static char *instlog_filename;
static bool setup_complete;
static volatile unsigned int instrument_stopped;

struct results {
	unsigned int n_end_of_file_events;
	unsigned int n_abc_events;
	unsigned int n_abc_events_in_range;
	unsigned int n_range_end_events;
	unsigned int n_range_start_events;
	unsigned int n_start_dispatch_events;
	unsigned int n_start_events;
	unsigned long client_data;
	unsigned long long sequence;
	unsigned long range_end;
	unsigned long range_start;
};

/* WRS_XXX make this instance specific. */
static struct results results;

#ifdef _WIN32
static const char *tempdir(void)
{
	if (getenv("TEMP") != NULL)
		return getenv("TEMP");
	else if (getenv("TMP") != NULL)
		return getenv("TMP");
	else
		return NULL;
}
#else
static const char *tempdir(void)
{
	if (getenv("TMPDIR") != NULL)
		return getenv("TMPDIR");
	else
		return "/tmp";
}
#endif

static char *temporary_filename(const struct cafe_allocator *allocator,
				const char *template)
{
	struct timespec ts;

	if (clock_gettime(CLOCK_REALTIME, &ts) != 0) {
		perror("Error getting time for use in temporary filename");
		ts.tv_sec = 0;
		ts.tv_nsec = 0;
	}

	return cafe_mprintf(allocator,
			    "%s_%ld_%lu_%ld", template, ts.tv_sec, ts.tv_nsec,
#ifdef _WIN32
			    (long)GetCurrentProcessId());
#else
			    (long)getpid());
#endif
}

static char *make_pathname(const char *dir, const char *template, const char *suffix)
{
	char *path;
	char *filename = temporary_filename(cafe_stdlib_allocator, template);

	path = cafe_mprintf(cafe_stdlib_allocator, "%s%s%s.%s",
			    (dir != NULL) ? dir : "",
			    (dir != NULL) ? "/" : "",
			    filename, suffix);

	CAFE_FREE(cafe_stdlib_allocator, filename);

	return path;
}

static void stop_notifier(struct cafe_ep_inst *inst)
{
	instrument_stopped = true;
}

static int test_prep_content(void)
{
	struct cafe_instrument *instrument;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst[2];
	struct cafe_cfg_setting settings[1];
	struct timespec timespec = { 0, 100000000 }; /* 0.1s */
	size_t test_content_sz = strlen(test_content);
	FILE *fp;

	text_filename = make_pathname(tempdir(), "test_file_stream_playback_text", "txt");
	instlog_filename = make_pathname(tempdir(), "test_file_stream_playback_instlog", "instlog");
	CUT_ASSERT_NOT_NULL(text_filename);
	CUT_ASSERT_NOT_NULL(instlog_filename);

	CUT_ASSERT_NOT_NULL((fp = fopen(text_filename, "wb")));
	CUT_ASSERT_EQUAL(1, fwrite(test_content, test_content_sz, 1, fp));
	CUT_ASSERT_EQUAL(0, fclose(fp));

	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(cafe_instrument_create("test_file_stream_playback", &instrument));

	CUT_ASSERT_NOT_NULL((spec = cafe_ep_spec_find("text_file_provider")));
	settings[0].name = "filename";
	settings[0].value.str = text_filename;
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument, spec, 1, settings, &inst[0]));

	CUT_ASSERT_NOT_NULL((spec = cafe_ep_spec_find("file_stream_writer")));
	settings[0].name = "filename";
	settings[0].value.str = instlog_filename;
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument, spec, 1, settings, &inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));
	CAFE_UNLOCK();

	instrument_stopped = false;
	instrument->stop_notifier = stop_notifier;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(instrument));
	CAFE_UNLOCK();

	return 0;
}

static int test_playback(void)
{
	size_t i;
	struct cafe_instrument *instrument;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst[3];
	struct cafe_cfg_setting settings[10];
	struct timespec timespec = { 0, 1000 };
	char *fsw_filename;

	CUT_ASSERT_NOT_NULL(text_filename);
	CUT_ASSERT_NOT_NULL(instlog_filename);
	fsw_filename = make_pathname(tempdir(), "test_file_stream_playback_fsw", "instlog");
	CUT_ASSERT_NOT_NULL(fsw_filename);

	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(cafe_instrument_create("test_file_stream_playback", &instrument));

	CUT_ASSERT_NOT_NULL((spec = cafe_ep_spec_find("file_stream_reader")));
	settings[0].name = "filename";
	settings[0].value.str = instlog_filename;
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument, spec, 1, settings, &inst[0]));

	CUT_ASSERT_NOT_NULL((spec = cafe_ep_spec_find("file_stream_writer")));
	settings[0].name = "filename";
	settings[0].value.str = fsw_filename;
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument, spec, 1, settings, &inst[1]));

	CUT_ASSERT_NOT_NULL((spec = cafe_ep_spec_find("test_file_stream_playback")));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument, spec, 0, NULL, &inst[2]));

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[2]));

	CAFE_UNLOCK();

	instrument->stop_notifier = stop_notifier;

	/*
	 * Validate total number of events with no playback settings
	 * specified.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(0, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(0, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);
	CUT_ASSERT_EQUAL(0, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_start_dispatch_events);
	CUT_ASSERT_EQUAL(0, results.n_start_events);
	CUT_ASSERT_EQUAL(0, results.client_data);

	settings[0].name = "start_type";
	settings[0].value.str = (char *)file_stream_reader_start_VT_REF;
	settings[1].name = "range_start_type";
	settings[1].value.str = (char *)file_stream_reader_range_start_VT_REF;
	settings[2].name = "range_end_type";
	settings[2].value.str = (char *)file_stream_reader_range_end_VT_REF;
	settings[3].name = "start_dispatch_type";
	settings[3].value.str = (char *)file_stream_reader_start_dispatch_VT_REF;
	settings[4].name = "end_of_file_type";
	settings[4].value.str = (char *)file_stream_reader_end_of_file_VT_REF;

	/*
	 * Validate total number of events with range_start = 0.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 0;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 6, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(12, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);

	/*
	 * Validate total number of events with range_end = 0.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_end";
	settings[5].value.ll = results.range_end = 0;
	settings[6].name = "client_data";
	settings[6].value.ll = 10;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 7, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(10, results.client_data);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(12, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);

	/*
	 * Validate total number of events with range_start = 0 and range_end == 0.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 0;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 0;
	settings[7].name = "client_data";
	settings[7].value.ll = 20;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(20, results.client_data);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(12, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);

	/*
	 * Validate total number of events with range_start = 1 and range_end == 1.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 1;
	settings[7].name = "client_data";
	settings[7].value.ll = 30;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(30, results.client_data);
	CUT_ASSERT_EQUAL(0, results.n_abc_events);
	CUT_ASSERT_EQUAL(0, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);

	/*
	 * Validate total number of events with range_start = 0 and range_end == 4.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 0;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 4;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 7, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(0, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(30, results.client_data); /* should be unchanged */
	CUT_ASSERT_EQUAL(4, results.n_abc_events);
	CUT_ASSERT_EQUAL(4, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_end_events);

	/*
	 * Validate total number of events with range_start = 1 and range_end == 4.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 4;
	settings[7].name = "client_data";
	settings[7].value.ll = 40;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(0, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(40, results.client_data);
	CUT_ASSERT_EQUAL(4, results.n_abc_events);
	CUT_ASSERT_EQUAL(3, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_end_events);

	/*
	 * Validate total number of events with range_start = 1 and range_end == 0.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 0;
	settings[7].name = "client_data";
	settings[7].value.ll = 0;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(0, results.client_data);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(11, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);

	/*
	 * Validate total number of events with range_start = 100 and range_end == 0.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 100;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 0;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 7, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(0, results.client_data);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(0, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);

	/*
	 * Validate total number of events with range_start = 100 and range_end == 200.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 100;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 200;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 7, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(0, results.client_data);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(0, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);

	/*
	 * Validate total number of events with range_start = 1 and range_end == 2.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 2;
	settings[7].name = "rewind";
	settings[7].value.ull = 1;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(0, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(0, results.client_data);
	CUT_ASSERT_EQUAL(2, results.n_abc_events);
	CUT_ASSERT_EQUAL(1, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_end_events);

	/*
	 * Validate +1 more event because of rewind == false.
	 */
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 3;
	settings[7].name = "rewind";
	settings[7].value.ll = 0;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(0, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(0, results.client_data);
	CUT_ASSERT_EQUAL(3, results.n_abc_events);
	CUT_ASSERT_EQUAL(2, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(2, results.n_range_end_events);

	/*
	 * Validate +N more events because of rewind == false.
	 */
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 0;
	settings[7].name = "rewind";
	settings[7].value.ll = 0;
	settings[8].name = "client_data";
	settings[8].value.ll = 100;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 9, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(100, results.client_data);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(11, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(2, results.n_range_end_events);

	/*
	 * Validate +N more events because of rewind == false and
	 * already at end of file.
	 */
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 0;
	settings[7].name = "rewind";
	settings[7].value.ll = 0;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(2, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(100, results.client_data);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(11, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(2, results.n_range_end_events);

	/*
	 * Validate clear of rewind.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 0;
	settings[7].name = "rewind";
	settings[7].value.ll = 1;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(100, results.client_data);
	CUT_ASSERT_EQUAL(12, results.n_abc_events);
	CUT_ASSERT_EQUAL(11, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);

	/*
	 * Validate DE323: no more events if resumed and already at or beyond range_end.
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 1;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 4;
	settings[7].name = "rewind";
	settings[7].value.ll = 1;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(0, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(100, results.client_data);
	CUT_ASSERT_EQUAL(4, results.n_abc_events);
	CUT_ASSERT_EQUAL(3, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_end_events);

	// restart with same settings - no more events expected
	instrument_stopped = false;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(0, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(100, results.client_data);
	CUT_ASSERT_EQUAL(4, results.n_abc_events);
	CUT_ASSERT_EQUAL(3, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_start_events);
	CUT_ASSERT_EQUAL(1, results.n_range_end_events);

#if 0
	/*
	 * Validate invalid range.  Behaviour as of 1/May/2012 is undefined. */
	 */
	memset(&results, 0, sizeof results);
	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = 5;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = 3;
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 7, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(100, results.client_data);
	CUT_ASSERT_EQUAL(0, results.n_abc_events);
	CUT_ASSERT_EQUAL(0, results.n_abc_events_in_range);
	CUT_ASSERT_EQUAL(1, results.n_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_start_events);
	CUT_ASSERT_EQUAL(0, results.n_range_end_events);
#endif

	/* Validate orthogonal data, both sequence and client_data. */

	memset(&results, 0, sizeof results);

	for (i = 0; i < 12; i++) {
		instrument_stopped = false;
		settings[5].name = "range_start";
		settings[5].value.ll = results.range_start = i;
		settings[6].name = "range_end";
		settings[6].value.ll = results.range_end = i + 1;
		settings[7].name = "rewind";
		settings[7].value.ll = i == 0;
		settings[8].name = "client_data";
		settings[8].value.ll = 1024 * 1024;
		results.n_abc_events_in_range = 0;

		CAFE_LOCK();
		CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 9, settings));
		CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
		CAFE_UNLOCK();

		while (!instrument_stopped)
			nanosleep(&timespec, NULL);

		CUT_ASSERT_EQUAL(1, results.n_abc_events_in_range);
		CUT_ASSERT_EQUAL(i + 1, results.n_abc_events);
		CUT_ASSERT_EQUAL(i, results.sequence);
		CUT_ASSERT_EQUAL(i + 1, results.n_start_dispatch_events);
		CUT_ASSERT_EQUAL(i + 1, results.n_range_start_events);
		CUT_ASSERT_EQUAL(i + 1, results.n_range_end_events);
		CUT_ASSERT_EQUAL((1024 * 1024), results.client_data);
	}
	CUT_ASSERT_EQUAL(i, results.n_abc_events);
	CUT_ASSERT_EQUAL(0, results.n_end_of_file_events);

	instrument_stopped = false;
	settings[5].name = "range_start";
	settings[5].value.ll = results.range_start = i;
	settings[6].name = "range_end";
	settings[6].value.ll = results.range_end = i + 1;
	settings[7].name = "rewind";
	settings[7].value.ll = 0;

	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(inst[0]->spec->settings_commit(inst[0], 8, settings));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));
	CAFE_UNLOCK();

	while (!instrument_stopped)
		nanosleep(&timespec, NULL);

	CUT_ASSERT_EQUAL(1, results.n_end_of_file_events);
	CUT_ASSERT_EQUAL(11, results.sequence);

	/* Done! */
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(instrument));
	CAFE_UNLOCK();

	remove(fsw_filename);
	remove(text_filename);
	remove(instlog_filename);
	CAFE_FREE(cafe_stdlib_allocator, text_filename);
	CAFE_FREE(cafe_stdlib_allocator, instlog_filename);
	CAFE_FREE(cafe_stdlib_allocator, fsw_filename);

	return 0;
}

static void range_start_action_handler(struct cafe_event_action_info *info, void **data)
{
	struct file_stream_reader_sequence *seq = data[1];

	if (debug_action_handlers)
		CUT_PRINTF("%s\n", info->event->types[0]->type_str);

	results.n_range_start_events++;
	results.sequence = seq->number;
}

static void range_end_action_handler(struct cafe_event_action_info *info, void **data)
{
	struct file_stream_reader_sequence *seq = data[1];

	if (debug_action_handlers)
		CUT_PRINTF("%s\n", info->event->types[0]->type_str);

	results.n_range_end_events++;
	results.sequence = seq->number;
}

static void start_action_handler(struct cafe_event_action_info *info, void **data)
{
	struct file_stream_reader_sequence *seq = data[1];

	if (debug_action_handlers)
		CUT_PRINTF("%s\n", info->event->types[0]->type_str);

	results.n_start_events++;
	results.sequence = seq->number;
}

static void start_dispatch_action_handler(struct cafe_event_action_info *info, void **data)
{
	struct file_stream_reader_client_data *client_data = data[1];

	if (debug_action_handlers)
		CUT_PRINTF("%s\n", info->event->types[0]->type_str);

	results.n_start_dispatch_events++;
	results.client_data = client_data->value;
}

static void end_of_file_action_handler(struct cafe_event_action_info *info, void **data)
{
	if (debug_action_handlers)
		CUT_PRINTF("%s\n", info->event->types[0]->type_str);

	results.n_end_of_file_events++;
}

static void test_fsr_ABC_action_handler(struct cafe_event_action_info *info, void **data)
{
	if (debug_action_handlers)
		CUT_PRINTF("%s\n", info->event->types[0]->type_str);

	results.n_abc_events++;

	if (results.n_range_start_events)
		results.n_abc_events_in_range++;
}

static const char *range_start_TYPE[] = {
	file_stream_reader_range_start_VT_REF
};

static const char *range_end_TYPE[] = {
	file_stream_reader_range_end_VT_REF
};

static const char *start_TYPE[] = {
	file_stream_reader_start_VT_REF
};

static const char *start_dispatch_TYPE[] = {
	file_stream_reader_start_dispatch_VT_REF
};

static const char *end_of_file_TYPE[] = {
	file_stream_reader_end_of_file_VT_REF
};

static const char *test_fsr_A_TYPE[] = {
	TEST_FSR_A_VT_DEF
};

static const char *test_fsr_B_TYPE[] = {
	TEST_FSR_B_VT_DEF
};

static const char *test_fsr_C_TYPE[] = {
	TEST_FSR_C_VT_DEF
};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT2(range_start, range_start_action_handler),
	CAFE_INPEV_INIT2(range_end, range_end_action_handler),
	CAFE_INPEV_INIT2(start, start_action_handler),
	CAFE_INPEV_INIT2(start_dispatch, start_dispatch_action_handler),
	CAFE_INPEV_INIT2(end_of_file, end_of_file_action_handler),
	CAFE_INPEV_INIT2(test_fsr_A, test_fsr_ABC_action_handler),
	CAFE_INPEV_INIT2(test_fsr_B, test_fsr_ABC_action_handler),
	CAFE_INPEV_INIT2(test_fsr_C, test_fsr_ABC_action_handler)
};

static int get_input_event_actions(struct cafe_ep_inst *inst,
				   struct cafe_event_descr *event,
				   struct cafe_event_action_list *actions)
{
	int rc;
	CAFE_ACTION_HANDLER handler;
	struct cafe_event_action_info *event_info;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);

	if (rc != CAFE_OK || event_info == NULL)
		return rc;

	assert(event_info->impl_data);
	handler = (CAFE_ACTION_HANDLER)event_info->impl_data;

	return cafe_add_event_action(inst, actions, handler, event_info);
}

#define ALL_TYPES				\
	FILE_STREAM_READER_TYPE_ALL_VT_DEFS	\
	TEST_FSR_A_VT_DEF			\
	TEST_FSR_B_VT_DEF			\
	TEST_FSR_C_VT_DEF

static struct cafe_ep_spec spec = {
	"test_file_stream_playback",
	NULL,			/* create */
	NULL,			/* delete */
	NULL,			/* start */
	NULL,			/* stop */
	get_input_event_actions,
	NULL,			/* update_output_event_actions */
	NULL,			/* inject event */
	ALL_TYPES,
	"S[test_file_stream_reader_workarea,v[dummy,L]]",
	CAFE_NELEMENTS(input_events),
	input_events
};

static void setup(void)
{
	int rc;

	assert(!setup_complete);
	CAFE_LOCK();

	if ((rc = cafe_ep_spec_register(&spec)) != CAFE_OK)
		goto error;

	if ((rc = text_file_provider_register()) != CAFE_OK)
		goto error;

error:
	setup_complete = (rc == CAFE_OK);
	CAFE_UNLOCK();
}

static void teardown(void)
{
	if (setup_complete) {
		CAFE_LOCK();
		cafe_ep_spec_unregister(&spec);
		text_file_provider_unregister();
		CAFE_UNLOCK();
	}
	setup_complete = false;
}

CUT_BEGIN_SUITE_F(fsr_playback_test_suite, &setup, &teardown)
CUT_RUN_TEST(test_prep_content);
CUT_RUN_TEST(test_playback);
CUT_END_SUITE_F
