/* test_instance_api.c - unit tests for configuration instance routines */

/* Copyright (c) 2010-2011, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
15may14,sdt  Fix ressource leak issue

DESCRIPTION
*/

#include <string.h>
#include <cafe/types.h>
#include <cafe/utils.h>
#include <cafe/cafe_print.h>
#include <cafe/value_type.h>
#include <cafe/ep/print.h>
#include <test/cafe_CUnitTest.h>
#include "dummy_cfg.h"
#include "verify_cfg.h"


#define TEST_ITERS (43)

static bool setup_complete = false;	/* setup routine completed OK */
static 	struct cafe_event_descr *eventTstEna = NULL;


static
void setup(void)
{
	int rc;

	CAFE_LOCK();
	rc = dummy_cfg_register();
	if (rc != CAFE_OK)
		goto error;

	rc = verify_cfg_register();
	if (rc != CAFE_OK)
		goto error;

error:
	setup_complete = (rc == CAFE_OK);
}


static
void teardown(void)
{
	if (setup_complete) {
		verify_cfg_unregister();
		dummy_cfg_unregister();
	}
	CAFE_UNLOCK();
	setup_complete = false;
}


/**
 *  Unit test for the event_enable() api.
 */

static void test_free(void)
{
	if (eventTstEna != NULL) {
	    free(eventTstEna);
	    eventTstEna = NULL;
	}
}

static
int test_event_enable(void)
{
	static const char *UNKNOWN_TYPE = "S[unknown_type]";
	struct cafe_instrument *session;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst;
	struct cafe_event_descr *event = eventTstEna;

	CUT_ASSERT_CAFE_OK(dummy_cfg_register());

	CUT_ASSERT_CAFE_OK(cafe_instrument_create(
				__func__, &session));

	spec = cafe_ep_spec_find("dummy_source");
	CUT_ASSERT_NOT_NULL(spec);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec, 0, NULL, &inst));

	/* set up an event description handle for us to reuse */
	event = eventTstEna = malloc(sizeof(struct cafe_event_descr)
	                             + sizeof(struct cafe_value_type *));
	CUT_ASSERT_NOT_NULL(event);

	/* first try some expected success cases */
	event->num_types = 1;
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, DUMMY_EVT_VT_REF,
					   &event->types[0]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst, event));
	/* re-enabling the same output event */
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst, event));

	/* unhandled event type */
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, UNKNOWN_TYPE,
					   &event->types[0]));
	CUT_ASSERT(cafe_ep_inst_enable_output_event(inst, event) != CAFE_OK);

	/* unhandled orthogonal type */
	event->num_types = 2;
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, DUMMY_EVT_VT_REF,
					   &event->types[0]));
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, UNKNOWN_TYPE,
					   &event->types[1]));
	CUT_ASSERT(cafe_ep_inst_enable_output_event(inst, event) != CAFE_OK);

	/* clean up, disconnect, and deallocate configurations... */
	test_free();
	cafe_ep_inst_delete(inst);
	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(session));
	return 0;
}


static
int test_instance_order(void)
{
	struct cafe_instrument *session;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst[2];

	CUT_ASSERT_TRUE(setup_complete);

	CUT_ASSERT_CAFE_OK(cafe_instrument_create(
				"test_instance_order", &session));

	spec = cafe_ep_spec_find("dummy_source");
	CUT_ASSERT_NOT_NULL(spec);

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       0, NULL, &inst[0]));
	CUT_ASSERT_CAFE_OK(cafe_inst_enable_all_output_events(inst[0]));

	spec = cafe_ep_spec_find("verify_cfg");
	CUT_ASSERT_NOT_NULL(spec);

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       0, NULL, &inst[1]));

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));

	CUT_ASSERT_TRUE(inst[0]->order < inst[1]->order);

	/* Forms a circle */
	CUT_ASSERT_EQUAL(cafe_ep_inst_connect(inst[1], inst[0]), 
	                 CAFE_INVALID_CONNECTION);

	CUT_ASSERT_TRUE(inst[0]->order < inst[1]->order);

	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(session));

	return 0;
}


CUT_BEGIN_SUITE_F(instance_api_suite, &setup, &teardown)
	CUT_RUN_TEST_F(test_event_enable, NULL, test_free);
	CUT_RUN_TEST(test_instance_order);
CUT_END_SUITE_F
