/* Copyright (c) 2011, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
24feb14,sdt  Disable unexpected NO_EFFECT warnings

DESCRIPTION

*/

#include <unistd.h>
#include <sched.h>
#include <sys/swap.h>
#include <string.h>
#include <assert.h>
#include <test/cafe_CUnitTest.h>
#include <cafe/ep/linux_trace.h>

#define NUM_ROOT_INSTANCES  (3)

#define EP_NAME "linux_perf_enable_disable_test"


/* use the swapon and swapoff events because we can be reasonably sure no-one
   else in the system is generating these while we're running this test.
   There may be a more ideal method to produce events in a deterministic way.
*/
enum {
	SWAPON = 0,
	SWAPOFF
};


#define EVENT_COUNT(index, event)					\
	(((struct inst_data *)instance_list[(index)]->client_data)->event_data[(event)].event_count)


static struct cafe_instrument *instrument[NUM_ROOT_INSTANCES];
static struct cafe_ep_inst *instance_list[NUM_ROOT_INSTANCES];

struct event_data {
	unsigned long event_count;
	bool enabled;
};

static const char *SWAPON_TYPE[] = {
	"t[syscalls.sys_enter_swapon]"
};

static const char *SWAPOFF_TYPE[] = {
	"t[syscalls.sys_enter_swapoff]"
};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT2(SWAPON, (void *) SWAPON),
	CAFE_INPEV_INIT2(SWAPOFF, (void *) SWAPOFF)
};

struct inst_data {
	unsigned int index;
	struct event_data event_data[CAFE_NELEMENTS(input_events)];
	struct cafe_ep_inst *source;
};


static
void count_handler(struct cafe_event_action_info *info, void **data)
{
	struct inst_data *inst_data;
	unsigned long event_index;

	inst_data = (struct inst_data *)info->inst->client_data;
	event_index = (uintptr_t) info->impl_data;
	inst_data->event_data[event_index].event_count++;
	assert(inst_data->event_data[event_index].enabled);
}


static
int inst_create(struct cafe_ep_inst *inst)
{
	unsigned int i;
	struct inst_data *inst_data;

 	for (i = 0; i < CAFE_NELEMENTS(instance_list); ++i)
		if (instance_list[i] == NULL)
			break;

	if (i == CAFE_NELEMENTS(instance_list))
		return CAFE_NOMEM;

	inst_data = CAFE_ALLOC(inst->allocator, sizeof(*inst_data));

	if (inst_data == NULL)
		return CAFE_NOMEM;

	memset(&inst_data->event_data, 0, sizeof(inst_data->event_data));
	inst_data->index = i;
	instance_list[i] = inst;
	inst->client_data = inst_data;

	return CAFE_OK;
}


static
void inst_delete(struct cafe_ep_inst *inst)
{
	struct inst_data *inst_data =
		(struct inst_data *) inst->client_data;

	instance_list[inst_data->index] = NULL;
	CAFE_FREE(inst->allocator, inst_data);
}


static
int get_input_event_actions(struct cafe_ep_inst *inst,
			    struct cafe_event_descr *event,
			    struct cafe_event_action_list *actions)
{
	int rc;
	struct inst_data *inst_data;
	struct cafe_event_action_info *action;

	inst_data = (struct inst_data *) inst->client_data;
	rc = cafe_ep_inst_event_action_info_get(inst, event, &action);
	assert(inst_data->event_data[(uintptr_t) action->impl_data].enabled);
	if (!inst_data->event_data[(uintptr_t) action->impl_data].enabled)
		return CAFE_ERROR;

	if (rc != CAFE_OK || action == NULL)
		return rc;

	return cafe_add_event_action(inst, actions, count_handler, action);
}


static struct cafe_ep_spec spec = {
	EP_NAME,
	inst_create,
	inst_delete,
	NULL,			/* start */
	NULL,			/* stop */
	get_input_event_actions,
	NULL,			/* update outputs */
	NULL,			/* inject event */
	NULL,			/* types */
	NULL,			/* workarea type */
	CAFE_NELEMENTS(input_events),
	input_events
};

static bool test_ep_registered = false;


static
void test_ep_unregister(void)
{
	if (test_ep_registered) {
		cafe_ep_spec_unregister(&spec);
		test_ep_registered = false;
	}
}


static
int test_ep_register(void)
{
	int rc;

	if (test_ep_registered)
		return CAFE_OK;

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		test_ep_registered = true;

	return rc;
}


static
int build_instruments(void)
{
	unsigned int i;
	struct cafe_ep_spec *spec[2];

	CAFE_LOCK();

	CUT_ASSERT_CAFE_OK(linux_trace_register());
	CUT_ASSERT_CAFE_OK(test_ep_register());

	spec[0] = cafe_ep_spec_find("linux_trace");
	spec[1] = cafe_ep_spec_find(EP_NAME);
	CUT_ASSERT_NOT_NULL(spec[0]);
	CUT_ASSERT_NOT_NULL(spec[1]);

	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		static struct cafe_ep_inst *inst[2];
		struct inst_data *inst_data;

		CUT_ASSERT_CAFE_OK(cafe_instrument_create(
					EP_NAME, &instrument[i]));

		CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument[i], spec[0],
						       0, NULL, &inst[0]));

		CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument[i], spec[1],
						       0, NULL, &inst[1]));

		/* Bypass threading compatibility checks. */
		inst[0]->flags |= CAFE_EP_INST_FLAG_OUTPUT_SESSION_THREAD;
		inst_data = (struct inst_data *const)inst[1]->client_data;
		inst_data->source = inst[0];

		CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));

		CUT_ASSERT_CAFE_OK(cafe_instrument_refresh_output_event_actions(instrument[i]));

		CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument[i]));
	}
	CAFE_UNLOCK();
	return 0;
}


static
int delete_instruments(void)
{
	unsigned int i;

	CAFE_LOCK();
	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		if (instrument[i] != NULL) {
			CUT_ASSERT_CAFE_OK(cafe_instrument_delete(instrument[i]));
			instrument[i] = NULL;
		}
	}
	linux_trace_unregister();
	test_ep_unregister();
	CAFE_UNLOCK();
	return 0;
}


static
void dispatch_event(int event_id, unsigned int nevents)
{
	unsigned int i;

	for (i = 0; i < nevents; ++i) {
		switch (event_id) {
		case SWAPON: {
			swapon("/non/existant/path", 0);
			break;
		}
		case SWAPOFF:
			swapoff("/non/existant/path");
			break;
		default:
			/* coverity[array_null] */
			assert(!"unexpected event_id");
		}
	}
}


static
int enable_event(struct cafe_ep_inst *inst, unsigned int event_index)
{
	int rc;
	struct cafe_event_descr *const type = inst->input_events[event_index];
	struct inst_data *const inst_data =
		(struct inst_data *) inst->client_data;

	inst_data->event_data[event_index].enabled = true;
	inst_data->event_data[event_index].event_count = 0;

	CAFE_LOCK();
	rc = cafe_ep_inst_enable_output_event(inst_data->source, type);
	CAFE_UNLOCK();

	return rc;
}


static
int disable_event(struct cafe_ep_inst *inst, size_t event_index)
{
	int rc;
	struct cafe_event_descr *const type = inst->input_events[event_index];
	struct inst_data *const inst_data =
		(struct inst_data *) inst->client_data;

	inst_data->event_data[event_index].enabled = false;
	inst_data->event_data[event_index].event_count = 0;

	CAFE_LOCK();
	rc = cafe_ep_inst_disable_output_event(inst_data->source, type);
	CAFE_UNLOCK();

	return rc;
}


static
int refresh_instance(size_t inst_index)
{
	int rc;

	CAFE_LOCK();
	rc = cafe_instrument_refresh_output_event_actions(instrument[inst_index]);
	CAFE_UNLOCK();

	return rc;
}


static
int test(void)
{
	unsigned int i;

	/* enable all. */
	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		CUT_ASSERT_CAFE_OK(enable_event(instance_list[i], SWAPON));
		CUT_ASSERT_CAFE_OK(enable_event(instance_list[i], SWAPOFF));
		CUT_ASSERT_CAFE_OK(refresh_instance(i));
	}

	dispatch_event(SWAPON, 5);
	dispatch_event(SWAPOFF, 5);
	sleep(3);

	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		CUT_ASSERT_EQUAL(5, EVENT_COUNT(i, SWAPON));
		CUT_ASSERT_EQUAL(5, EVENT_COUNT(i, SWAPOFF));
	}

	/* disable all. */
	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		CUT_ASSERT_CAFE_OK(disable_event(instance_list[i], SWAPON));
		CUT_ASSERT_CAFE_OK(disable_event(instance_list[i], SWAPOFF));
		CUT_ASSERT_CAFE_OK(refresh_instance(i));
	}

	dispatch_event(SWAPON, 5);
	dispatch_event(SWAPOFF, 5);
	sleep(3);

	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		CUT_ASSERT_EQUAL(0, EVENT_COUNT(i, SWAPON));
		CUT_ASSERT_EQUAL(0, EVENT_COUNT(i, SWAPOFF));
	}

	/* enable SWAPOFF for even numbered instances. */
	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		bool is_even = (i % 2) == 0;

		CUT_ASSERT_CAFE_OK(disable_event(instance_list[i], SWAPON));
		CUT_ASSERT_CAFE_OK(disable_event(instance_list[i], SWAPOFF));
		CUT_ASSERT_CAFE_OK(refresh_instance(i));

		if (is_even)
			CUT_ASSERT_CAFE_OK(enable_event(instance_list[i], SWAPOFF));

		CUT_ASSERT_CAFE_OK(refresh_instance(i));
	}

	dispatch_event(SWAPON, 5);
	dispatch_event(SWAPOFF, 5);
	sleep(3);

	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		int expected = (i % 2) == 0 ? 5 : 0;
		CUT_ASSERT_EQUAL(0, EVENT_COUNT(i, SWAPON));
		CUT_ASSERT_EQUAL(expected, EVENT_COUNT(i, SWAPOFF));
	}

	for (i = NUM_ROOT_INSTANCES / 2; i < NUM_ROOT_INSTANCES; ++i) {
		CUT_ASSERT_CAFE_OK(disable_event(instance_list[i], SWAPON));
		CUT_ASSERT_CAFE_OK(disable_event(instance_list[i], SWAPOFF));
		CUT_ASSERT_CAFE_OK(refresh_instance(i));
	}

	for (i = 0; i < NUM_ROOT_INSTANCES / 2; ++i) {
		CUT_ASSERT_CAFE_OK(disable_event(instance_list[i], SWAPON));
		CUT_ASSERT_CAFE_OK(enable_event(instance_list[i], SWAPOFF));
		CUT_ASSERT_CAFE_OK(refresh_instance(i));
	}

	dispatch_event(SWAPON, 5);
	dispatch_event(SWAPOFF, 5);
	sleep(3);

	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		if (i >= NUM_ROOT_INSTANCES / 2) {
			CUT_ASSERT_EQUAL(0, EVENT_COUNT(i, SWAPON));
			CUT_ASSERT_EQUAL(0, EVENT_COUNT(i, SWAPOFF));
		} else {
			CUT_ASSERT_EQUAL(0, EVENT_COUNT(i, SWAPON));
			CUT_ASSERT_EQUAL(5, EVENT_COUNT(i, SWAPOFF));
		}
	}

	for (i = 0; i < NUM_ROOT_INSTANCES / 2; ++i) {
		CUT_ASSERT_CAFE_OK(enable_event(instance_list[i], SWAPON));
		CUT_ASSERT_CAFE_OK(enable_event(instance_list[i], SWAPOFF));
		CUT_ASSERT_CAFE_OK(refresh_instance(i));
	}

	dispatch_event(SWAPON, 5);
	dispatch_event(SWAPOFF, 5);
	sleep(3);

	for (i = 0; i < NUM_ROOT_INSTANCES; ++i) {
		if (i >= NUM_ROOT_INSTANCES / 2) {
			CUT_ASSERT_EQUAL(0, EVENT_COUNT(i, SWAPON));
			CUT_ASSERT_EQUAL(0, EVENT_COUNT(i, SWAPOFF));
		} else {
			CUT_ASSERT_EQUAL(5, EVENT_COUNT(i, SWAPON));
			CUT_ASSERT_EQUAL(5, EVENT_COUNT(i, SWAPOFF));
		}
	}

	return 0;
}


CUT_BEGIN_SUITE(linux_perf_disable_event_suite)
	CUT_RUN_TEST(build_instruments);
	CUT_RUN_TEST(test);
	CUT_RUN_TEST(delete_instruments);
CUT_END_SUITE
