/* test_streams.c - test streaming events to and from files */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
16sep14,sdt  Fixed build warning
08aug14,sdt  Fixed secure coding warning
03aug14,sdt  Fixed checked-return warnings (US45196)
25jun14,sdt  Fix file path when cafe_cunit_logRootDirGet return empty string
20jun14,sdt  Make test runnable on simics target

DESCRIPTION
*/

#include "../../src/cafe_os.h"
#include <assert.h>
#include <string.h>  /* for strcmp() */
#include <sys/stat.h>
#include <cafe/utils.h>
#include <cafe/value_type.h>
#include <cafe/cafe_print.h>
#include <cafe/log_stream.h>
#include <cafe/ep/file_stream_writer.h>
#include <cafe/ep/file_stream_reader.h>
#include <cafe/ep/text_file_provider.h>
#include <cafe/ep/print.h>
#include <test/cafe_CUnitTest.h>
#include "dummy_cfg.h"
#include "verify_cfg.h"
#include "../src/time_utils.h"

#if defined(_MSC_VER)
    /* Turn off message: warning C4090: '=' : different 'const' qualifiers*/
#   pragma warning (disable: 4090)
#endif

#define TEST_ITERS (43)

static bool setup_complete = false;	/* setup routine completed OK */
static bool test_fstream_write_complete = false;
static char * filenamePath = NULL;
static char * textfilenamePath = NULL;
static int stop_count;
static FILE *fTestStream = NULL;


static char * fullFilenameGet(char ** pFile, char * file) {
    char * pRes = file;

#if defined(__VXWORKS__)
    if (*pFile != NULL) {
	pRes = *pFile;
    } else if ((*pFile == NULL) &&
		(cafe_cunit_logRootDirGet() != NULL)) {
	size_t sz = strlen(cafe_cunit_logRootDirGet()) + strlen(file) + 2;
	*pFile = malloc(sz);
	if (*pFile != NULL) {
	    memset(*pFile, 0, sz);
	    if (strlen(cafe_cunit_logRootDirGet()) > 0) {
		(void)sprintf(*pFile, "%s/%s", cafe_cunit_logRootDirGet(), file);
	    } else {
		strncpy(*pFile, file, sz);
		*pFile[sz - 1] = 0;
	    }
	    pRes = *pFile;
	}
    }
#endif

    return pRes;
}

static char * filenameGet(void) {
    return fullFilenameGet(&filenamePath, (char *)"test_stream.instlog");
}

static char * textfilenameGet(void) {
    return fullFilenameGet(&textfilenamePath, (char *)"test_stream.tinstlog");
}

static void test_stop_notifier(
	struct cafe_ep_inst *inst)
{
	stop_count++;
	(void)cafe_notify_signal();
}

static
void setup(void)
{
	int rc;
	
	fTestStream = NULL;
	assert(!setup_complete);
	CAFE_LOCK();

	if ((rc = dummy_cfg_register()) != CAFE_OK)
		goto error;

	if ((rc = verify_cfg_register()) != CAFE_OK)
		goto error;

	if ((rc = text_file_provider_register()) != CAFE_OK)
		goto error;

error:
	setup_complete = (rc == CAFE_OK);
}


static
void teardown(void)
{
	if (fTestStream != NULL)
		(void)fclose(fTestStream);

	if (setup_complete) {
		(void)text_file_provider_unregister();
		verify_cfg_unregister();
		dummy_cfg_unregister();
	}
	
	if (filenamePath != NULL) {
	    free(filenamePath);
	    filenamePath = NULL;
	}
	if (textfilenamePath != NULL) {
	    free(textfilenamePath);
	    textfilenamePath = NULL;
	}
	
	CAFE_UNLOCK();
	setup_complete = false;
}


static
int test_fstream_write(void)
{
	struct cafe_instrument *session;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst[2];
	unsigned int i;
	struct cafe_cfg_setting settings[1];

	CUT_ASSERT_TRUE(setup_complete);

	CUT_ASSERT_CAFE_OK(cafe_instrument_create("test_fstream_write", &session));

	spec = cafe_ep_spec_find("dummy_source");
	CUT_ASSERT_NOT_NULL(spec);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       0, NULL, &inst[0]));
	CUT_ASSERT_CAFE_OK(cafe_inst_enable_all_output_events(inst[0]));

	spec = cafe_ep_spec_find("file_stream_writer");
	CUT_ASSERT_NOT_NULL(spec);

	settings[0].name = "filename";
	settings[0].value.str = (char *)filenameGet();
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       CAFE_NELEMENTS(settings),
					       settings, &inst[1]));

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));

	CUT_ASSERT_CAFE_OK(cafe_instrument_refresh_output_event_actions(session));

	CUT_ASSERT_CAFE_OK(cafe_instrument_start(session));

	for (i = 0; i < TEST_ITERS; ++i)
		CUT_ASSERT_CAFE_OK(send_dummy_event(inst[0], i));

	/* Stop additional events from being generated */
	CUT_ASSERT_CAFE_OK(cafe_instrument_stop(session));

	/* clean up, disconnect, and deallocate configurations... */
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_disconnect(inst[0], inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_delete(inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_delete(inst[0]));

	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(session));

	test_fstream_write_complete = true;

	return 0;
}


#if 0
/**
 * make sure fstream forwards events to sink configs.
 */
static
int test_fstream_forward(void)
{
	CUT_FAIL("Test not written.");
	return 0;
}
#endif

/**
 *  Test reading an event stream from file entirely in one shot (that is,
 * as fast as we can process them).
 */
static
int test_fstream_read_oneshot(void)
{
	struct event_verify_spec events[TEST_ITERS];
	struct dummy_args args[TEST_ITERS];
	static const char *types[] = {DUMMY_EVT_VT_DEF};
	struct cafe_instrument *session;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst[2];

	struct cafe_cfg_setting settings[2];
	struct verify_cfg_workarea *vcw;
	unsigned int i;

	CUT_ASSERT_TRUE(setup_complete);
	CUT_ASSERT_TRUE(test_fstream_write_complete);

	CUT_ASSERT_CAFE_OK(cafe_instrument_create(
				"test_fstream_read_oneshot", &session));
	session->stop_notifier = test_stop_notifier;

	spec = cafe_ep_spec_find("file_stream_reader");
	CUT_ASSERT_NOT_NULL(spec);

	settings[0].name = "filename";
	settings[0].value.str = (char *)filenameGet();
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       1, settings, &inst[0]));

	spec = cafe_ep_spec_find("verify_cfg");
	CUT_ASSERT_NOT_NULL(spec);

	/* create verification input */
	for (i = 0; i < TEST_ITERS; ++i) {
		events[i].types.l = 1;
		events[i].types.p = types;
		events[i].data.l = sizeof(struct dummy_args);
		events[i].data.p = &args[i];
		args[i].data = i;
		args[i].magic = DUMMY_MAGIC;
	}

	settings[0].name = "events";
	settings[0].value.dary.l = TEST_ITERS;
	settings[0].value.dary.p = events;

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       1, settings, &inst[1]));

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));

	stop_count = 0;
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(session));

	while (stop_count == 0) {
		CUT_ASSERT_CAFE_OK(cafe_notify_wait());
	}

	/* clean up, disconnect, and deallocate configurations... */
	CUT_ASSERT_CAFE_OK(cafe_instrument_stop(session));
	vcw = (struct verify_cfg_workarea *)inst[1]->workarea->u.data;
	CUT_ASSERT_EQUAL(vcw->failed, 0);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_disconnect(inst[0], inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_delete(inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_delete(inst[0]));

	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(session));

	return 0;
}


#if 0
/**
 *  Test reading a events from file using delays suggested by the events
 * themselves.
 */
static
int test_fstream_read_timed(void)
{
	CUT_FAIL("Test not written yet.");
	return 0;
}
#endif


static
int test_text_file_provider(void)
{
	static const char *SIMPLE_TYPE[] = {"S[text_test.simple]"};
	static const char *TEST_TYPE[] = {"S[text_test.a]"};
	struct cafe_instrument *session;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst[2];

	struct cafe_cfg_setting settings[2];
	struct verify_cfg_workarea *vcw;
	struct event_verify_spec *events;
	unsigned int i;
	static const char text_file_content[] =
		"type simple S[text_test.simple]\n"
		"event simple []\n"
		"delay 100\n"
		"event simple []\n"
		"type a_b S[text_test.a],S[text_test.b]\n"
		"event a_b [],[]\n"
		"";

	CUT_ASSERT_TRUE(setup_complete);

	CUT_ASSERT_NOT_NULL((fTestStream = fopen(textfilenameGet(), "w")));
	CUT_ASSERT_EQUAL(fwrite(text_file_content, strlen(text_file_content), 1,
	                        fTestStream), 1);
	CUT_ASSERT_EQUAL(fclose(fTestStream), 0);
	fTestStream = NULL;

	CUT_ASSERT_CAFE_OK(cafe_instrument_create(
				"test_text_file_provider", &session));
	session->stop_notifier = test_stop_notifier;

	spec = cafe_ep_spec_find("text_file_provider");
	CUT_ASSERT_NOT_NULL(spec);

	settings[0].name = "filename";
	settings[0].value.str = (char *)textfilenameGet();
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       1, settings, &inst[0]));

	spec = cafe_ep_spec_find("verify_cfg");
	CUT_ASSERT_NOT_NULL(spec);

	/* create verification input */
	events = malloc(TEST_ITERS * sizeof *events);
	CUT_ASSERT_NOT_NULL(events);

	i = 0;
	events[i].types.l = 1;
	events[i].types.p = SIMPLE_TYPE;
	events[i].data.l = 0;
	events[i].data.p = NULL;
	i++;

	events[i].types.l = 1;
	events[i].types.p = SIMPLE_TYPE;
	events[i].data.l = 0;
	events[i].data.p = NULL;
	i++;

	events[i].types.l = 1;
	events[i].types.p = TEST_TYPE;
	events[i].data.l = 0;
	events[i].data.p = NULL;
	i++;

	settings[0].name = "events";
	settings[0].value.dary.l = i;
	settings[0].value.dary.p = events;

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       1, settings, &inst[1]));

	free(events);

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));

	stop_count = 0;
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(session));

	while (stop_count == 0) {
		CUT_ASSERT_CAFE_OK(cafe_notify_wait());
	}

	/* clean up, disconnect, and deallocate configurations... */
	CUT_ASSERT_CAFE_OK(cafe_instrument_stop(session));
	vcw = (struct verify_cfg_workarea *)inst[1]->workarea->u.data;
	CUT_ASSERT_EQUAL(vcw->failed, 0);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_disconnect(inst[0], inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_delete(inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_delete(inst[0]));

	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(session));

	return 0;
}


static
int test_filter_marshalling(void)
{
	const char *vt[] = {
		"S[main,v[arg,i]]", "F[1]", "S[orth,v[arg,i]]"
	};
	struct cafe_instrument *session;
	struct cafe_event_descr *event;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	int main_data = 42;
	int orth_data = 42 * 42;
	void *data[CAFE_MAX_EVENT_TYPES];
	struct cafe_log_buf *buf;
	struct cafe_slog_encoded_inst *slog;
	struct cafe_output_stream *ostream;
	unsigned int i;
	int8_t *read_head;
	unsigned long long eid, msg_size;
	struct cafe_event_map *event_map;
	struct cafe_value_type *metadata_event;
	struct cafe_remote_event *remote;
	int **read_data;
	struct cafe_byte_buffer bbuf;

	data[0] = &main_data;
	data[1] = NULL;
	data[2] = &orth_data;

	CUT_ASSERT_CAFE_OK(cafe_bbuf_init(&bbuf, cafe_stdlib_allocator, 128));

	buf = cafe_log_create(128, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(buf);

	CUT_ASSERT_CAFE_OK(cafe_logstream_open_output(
				   buf, cafe_stdlib_allocator,
				   &ostream, NULL, NULL));

	CUT_ASSERT_CAFE_OK(cafe_instrument_create(__FUNCTION__, &session));
	CUT_ASSERT_CAFE_OK(cafe_slog_encoded_create(session, ostream, &slog));

	for (i = 0; i < CAFE_NELEMENTS(vt); ++i) {
		CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, vt[i],
						   &types[i]));
	}
	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(session, CAFE_NELEMENTS(vt),
					      types, &event));

	cafe_slog_encoded_write(slog, event, data);
	cafe_slog_encoded_delete(slog);
	CAFE_FREE(cafe_stdlib_allocator, ostream);

	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, cafe_metadata_VT_DEF,
					   &metadata_event));
	event_map = cafe_event_map_create(session);
	CUT_ASSERT_NOT_NULL(event_map);
	read_head = cafe_mdecode_stream_header((int8_t *)buf->data, buf->size,
					       NULL);
	CUT_ASSERT_NOT_NULL(read_head);
	do {
		struct cafe_metadata md;
		read_head = cafe_mdecode_event_header(read_head, buf->size,
						      &msg_size, &eid);
		CUT_ASSERT_NOT_NULL(read_head);
		if (eid == 0) {
			read_head = cafe_mdecode_val(metadata_event, &md,
						     cafe_stdlib_allocator,
						     event_map, read_head);
			CUT_ASSERT_NOT_NULL(read_head);

			CUT_ASSERT_CAFE_OK(cafe_event_map_add_event(event_map, &md, NULL, &remote));
			cafe_release_values(metadata_event, &md,
					    cafe_stdlib_allocator);
		}
	} while (eid == 0);
	remote = cafe_event_map_lookup(event_map, eid);
	CUT_ASSERT_NOT_NULL(remote);
	CUT_ASSERT_CAFE_OK(cafe_mread_event_data(read_head, remote->event,
						 cafe_stdlib_allocator,
						 event_map, &bbuf));
	CUT_ASSERT(bbuf.len > 0);
	read_data = (int **) bbuf.data;
	CUT_ASSERT_EQUAL(*read_data[0], main_data);
	CUT_ASSERT_EQUAL(read_data[1], NULL);
	CUT_ASSERT_EQUAL(*read_data[2], orth_data);
	cafe_bbuf_release(&bbuf);

	cafe_event_map_destroy(event_map);
	CAFE_FREE(cafe_stdlib_allocator, buf);
	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(session));
	return 0;
}


CUT_BEGIN_SUITE_F(stream_suite, &setup, &teardown)
	CUT_RUN_TEST(test_fstream_write);
	CUT_RUN_TEST(test_fstream_read_oneshot);
	if (remove(filenameGet()))
		perror("unlinking test stream output file");
	CUT_RUN_TEST(test_text_file_provider);
	if (remove(textfilenameGet()))
		perror("unlinking test stream text file");
	CUT_RUN_TEST(test_filter_marshalling);
CUT_END_SUITE_F
