/* test_symresolve.c -	*/

/* Copyright (c) 2011-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
07nov14,tle  Emit entries for unresolved symbols and object files (US48740)
08jul14,sdt  Used CUT_PRINTF/CUT_PRINT_ERR macro to make clear output
13jun14,tle  Fix symresolve test failure due to fix for WB4-3716

DESCRIPTION

*/

#include <assert.h>
#include <dlfcn.h>
#include <pthread.h>
#include <sys/stat.h>
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/array.h>
#include <cafe/atomic.h>
#include <cafe/ep/linux_trace.h>
#include <cafe/ep/config_service.h>
#include <test/cafe_CUnitTest.h>
#include <tcf/framework/channel.h>
#include <tcf/framework/events.h>
#include <tcf/services/memorymap.h>
#include <tcf/services/pathmap.h>
#include "../tcf_utils.h"
#include "../symresolve_actions.h"
#include "../../../platforms/linux/include/memory_map.h"


#define RESOLVED_ADDRESS_DEF						\
	"S[cafe.resolved_address,"					\
	"v[object_name,s]v[object_address,Q]"				\
	"v[symbol_name,s]v[symbol_offset,Q]]"
#define RESOLVED_ADDRESS_REF  "t[cafe.resolved_address]"


struct resolved_address {
	char *object_name;
	unsigned long long address;
	char *name;
	unsigned long long offset;
};


#define RESOLVED_LINEINFO_DEF					   \
	"S[cafe.resolved_lineinfo,"				   \
	"v[source_name,s]v[line_number,Q]v[column_number,Q]"	   \
	"v[object_address,Q]v[symbol_offset,Q]v[symbol_name,s]]"
#define RESOLVED_LINEINFO_REF  "t[cafe.resolved_lineinfo]"

/*  */
struct resolved_lineinfo {
	char *file_name;
	unsigned long long line_number;
	unsigned long long column_number;
	unsigned long long address;
	unsigned long long offset;
	char *name;
};


static const char *const MAPPED_LIB = "libdl";
static const char *const MAPPED_FUNC = "dlopen";
static const char *const UNMAPPED = "[unmapped]";
static uintptr_t mapped_address;
static uintptr_t kernel_address;


static const struct memory_region ADDRESS_SPACE_START = {
	.start = 0xcafe,
	.end = 0xcafe
};


static bool setup_complete = false;	/* setup routine completed OK */
static cafe_atomic_val tcf_init_complete = false;


static struct cafe_namespace *ns;
static struct cafe_value_type *mm_delta_vt, *stacktrace_vt, *resolved_entry_vt,
	*resolved_indexed_vt, *lineinfo_indexed_vt, *lineinfo_vt;
static pthread_t event_thread;
static struct memory_map self_map;
static struct memory_map mapped_regions;
static struct memory_map kernel_regions;
static uintptr_t test_base_addr = 0;


static
void *tcf_event_thread(void *arg)
{
	Protocol *protocol;
	TCFBroadcastGroup *bcg;

	ini_events_queue();
	protocol = protocol_alloc();
	bcg = broadcast_group_alloc();
	ini_path_map_service(protocol, bcg);
#if defined(SERVICE_MemoryMap) && SERVICE_MemoryMap
	ini_memory_map_service(protocol, NULL);
#endif
#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	ini_contexts();
#endif
	cafe_atomic_set(&tcf_init_complete, true);
	run_event_loop();
	broadcast_group_free(bcg);
	protocol_release(protocol);
	return NULL;
}


static
void setup(void)
{
	int rc;

	assert(!setup_complete);
	rc = cafe_namespace_create(cafe_stdlib_allocator, &ns);
	if (rc != CAFE_OK)
		CUT_PRINTF_ERR("%s (%u): Error creating namespace: %s\n",
			__FILE__, __LINE__, cafe_strerror(rc));

	rc = cafe_type_parse(ns, cafe_ADDRESS_VT_DEF , NULL);
	if (rc != CAFE_OK)
		CUT_PRINTF_ERR("%s (%u): Error parsing type: %s\n",
			__FILE__, __LINE__, cafe_strerror(rc));

	rc = cafe_type_parse(ns, CAFE_MM_DELTA_ENTRY_DEF CAFE_MM_DELTA_DEF,
			     &mm_delta_vt);
	if (rc != CAFE_OK)
		CUT_PRINTF_ERR("%s (%u): Error parsing type: %s\n",
			__FILE__, __LINE__, cafe_strerror(rc));
	assert(mm_delta_vt != NULL);

	cafe_type_parse(ns, "a[Q]", &stacktrace_vt);
	assert(stacktrace_vt != NULL);

	cafe_type_parse(ns, RESOLVED_ADDRESS_DEF "a["RESOLVED_ADDRESS_REF"]",
			&resolved_entry_vt);
	assert(resolved_entry_vt != NULL);

	cafe_type_parse(ns, RESOLVED_LINEINFO_DEF "a["RESOLVED_LINEINFO_REF"]",
			&lineinfo_vt);
	assert(lineinfo_vt != NULL);

	cafe_type_parse(ns, "S[normalized_stack,v[symbol_index,Q]v[symbol_offset,Q]]"
			"a[t[normalized_stack]]", &resolved_indexed_vt);
	assert(resolved_indexed_vt != NULL);

	cafe_type_parse(ns, "S[normalized_lineinfo_stack,v[lineinfo_index,Q]]"
			"a[t[normalized_lineinfo_stack]]",
			&lineinfo_indexed_vt);
	assert(lineinfo_indexed_vt != NULL);

	setup_complete = true;
}


static
void teardown(void)
{
	if (!setup_complete)
		return;
	cafe_namespace_delete(ns);
	setup_complete = false;
}


#if 0 /* DEBUG only */
void dump_resolved_address(const struct resolved_address *a);
void dump_resolved_address(const struct resolved_address *a)
{
	printf("Resolved address:\n"
	       "\tname: %s\n"
	       "\tobject: %s\n"
	       "\taddress: 0x%llx\n"
	       "\toffset: 0x%llx\n",
	       a->name, a->object_name, a->address, a->offset);
}
#endif


static
int test_symresolve_lifecycle(void)
{
	struct symresolve_actions *actions;

	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						     cafe_stdlib_allocator,
						     &actions));
	symresolve_actions_delete(actions);
	return 0;
}


/**
 *
 */
static
bool populate_map(pid_t pid, struct symresolve_actions *actions,
		  const struct memory_map *map)
{
	unsigned int i;

	for (i = 0; i < map->region_count; ++i)
		if (symresolve_mm_update(actions, pid,
					 map->regions[i].start,
					 map->regions[i].end - map->regions[i].start,
					 map->regions[i].file_offset,
					 map->regions[i].name, true) != CAFE_OK)
			return false;

	return true;
}


static
int test_symresolve(void)
{
	struct cafe_vm_dynarray stack, resolved_stack;
	struct symresolve_actions *actions;
	unsigned long long *stack_arr;
	struct resolved_address *results;
	pid_t pid = getpid();
	unsigned long long gen1, gen2, gen3;

	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						     cafe_stdlib_allocator,
						     &actions));
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(actions, pid, 0));
	CUT_ASSERT(populate_map(pid, actions, &self_map));

	memset(&resolved_stack, 0, sizeof(resolved_stack));
	memset(&stack, 0, sizeof(stack));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 4, stacktrace_vt));
	stack_arr = stack.p;

	/* test basic success case */
	stack.l = 1;
	stack_arr[0] = (uintptr_t) test_symresolve;
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(actions, pid, &gen1));
	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid, gen1, &stack,
						resolved_entry_vt, &resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_EQUAL((uintptr_t)test_symresolve,
			 results[0].address + test_base_addr);
	CUT_ASSERT_EQUAL(0, strcmp("test_symresolve", results[0].name));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	/* test basic failure case */
	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid + 1, 0, &stack,
						resolved_entry_vt, &resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL((uintptr_t)test_symresolve, results[0].offset);
	CUT_ASSERT_EQUAL(0, results[0].address);
	CUT_ASSERT_EQUAL(0, strncmp(UNMAPPED, results[0].name, 2));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	/* test parent address space resolution */
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(actions, pid + 1, pid));
	CUT_ASSERT(populate_map(pid + 1, actions, &mapped_regions));
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(actions, pid + 1, &gen2));

	stack.l = 4;
	stack_arr[0] = CAFE_STACK_MARKER_KERNEL;
	stack_arr[1] = (uintptr_t) test_symresolve;
	stack_arr[2] = CAFE_STACK_MARKER_USER;
	stack_arr[3] = mapped_address;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid + 1, gen2, &stack,
					     resolved_entry_vt, &resolved_stack));
	CUT_ASSERT_EQUAL(2, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_EQUAL((uintptr_t)test_symresolve,
			 results[0].address + test_base_addr);
	CUT_ASSERT_EQUAL(0, strcmp("test_symresolve", results[0].name));
	CUT_ASSERT_EQUAL(0, results[1].offset);
	CUT_ASSERT_NOT_NULL(strstr(results[1].object_name, MAPPED_LIB));
	CUT_ASSERT_EQUAL(0, strncmp(results[1].name, MAPPED_FUNC,
				    strlen(MAPPED_FUNC)));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	/* test updating parent address space independently of the child */
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(actions, pid, pid));
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(actions, pid + 1, &gen3));
	CUT_ASSERT_NOT_EQUAL(gen2, gen3);
	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid + 1, gen3, &stack,
						resolved_entry_vt, &resolved_stack));
	CUT_ASSERT_EQUAL(2, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL((uintptr_t)test_symresolve, results[0].offset);
	CUT_ASSERT_EQUAL(0, results[0].address);
	CUT_ASSERT_EQUAL(0, strcmp(UNMAPPED, results[0].name));
	CUT_ASSERT_EQUAL(0, results[1].offset);
	CUT_ASSERT_NOT_NULL(strstr(results[1].object_name, MAPPED_LIB));
	CUT_ASSERT_EQUAL(0, strncmp(results[1].name, MAPPED_FUNC,
				    strlen(MAPPED_FUNC)));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	free(stack.p);
	symresolve_actions_delete(actions);
	return 0;
}


/**
 *  Verify the symbol resolution engine can normalize symbol entries
 * effectively.
 */
static
int test_indexed_symresolve(void)
{
	struct cafe_vm_dynarray stack, resolved_stack;
	struct symresolve_actions *actions;
	unsigned long long *stack_arr;
	unsigned long long *results;
	pid_t pid = getpid();
	unsigned long long gen1;
	struct cafe_value_type *index_table_vt;
	struct cafe_dict_struct index_dict;
	cafe_any_value key[1];
	struct cafe_value_data *dict_row;
	struct cafe_normalized_symbol_entry *entry;

	CUT_ASSERT_CAFE_OK(cafe_type_parse(ns, CAFE_NORMALIZED_SYMBOL_DEF
					   "D["CAFE_NORMALIZED_SYMBOL_REF"]",
					   &index_table_vt));

	memset(&index_dict, 0, sizeof(index_dict));
	cafe_dict_create(&index_dict, index_table_vt);

	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						     cafe_stdlib_allocator,
						     &actions));
	symresolve_set_symindex(actions, &index_dict);

	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(actions, pid, 0));
	CUT_ASSERT(populate_map(pid, actions, &self_map));

	memset(&resolved_stack, 0, sizeof(resolved_stack));
	memset(&stack, 0, sizeof(stack));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 4, stacktrace_vt));
	stack_arr = stack.p;

	/* test basic success case */
	stack.l = 1;
	stack_arr[0] = (uintptr_t) test_symresolve;
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(actions, pid, &gen1));
	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid, gen1, &stack,
						resolved_indexed_vt,
						&resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0]);

	CUT_ASSERT_EQUAL(1, index_dict.nentries);
	key[0].ull = 0;
	dict_row = cafe_dict_get(&index_dict, key);
	CUT_ASSERT_NOT_NULL(dict_row);
	entry = (struct cafe_normalized_symbol_entry *) dict_row->u.data;
	CUT_ASSERT_EQUAL(0, entry->index);
	CUT_ASSERT_EQUAL((uintptr_t)test_symresolve,
			 entry->object_address + test_base_addr);
	CUT_ASSERT_EQUAL(0, strcmp("test_symresolve", entry->name));

	free(stack.p);
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_indexed_vt);
	symresolve_actions_delete(actions);
	cafe_dict_delete(&index_dict);

	return 0;
}

#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
static
int test_indexed_symresolve_with_lineinfo(void)
{
	struct cafe_vm_dynarray stack, resolved_stack;
	struct symresolve_actions *actions;
	unsigned long long *stack_arr;
	unsigned long long *results;
	pid_t pid = getpid();
	unsigned long long gen1;
	struct cafe_value_type *sym_index_table_vt;
	struct cafe_dict_struct sym_index_dict;
	struct cafe_value_type *lineinfo_index_table_vt;
	struct cafe_dict_struct lineinfo_index_dict;
	cafe_any_value key[1];
	struct cafe_value_data *dict_row;
	struct cafe_normalized_symbol_entry *sym_entry;
	struct cafe_normalized_lineinfo_entry *lineinfo_entry;

	CUT_ASSERT_CAFE_OK(cafe_type_parse(ns, CAFE_NORMALIZED_SYMBOL_EXT_DEF
					   "D["CAFE_NORMALIZED_SYMBOL_EXT_REF"]",
					   &sym_index_table_vt));
	CUT_ASSERT_CAFE_OK(cafe_type_parse(ns, CAFE_NORMALIZED_LINEINFO_DEF
					   "D["CAFE_NORMALIZED_LINEINFO_REF"]",
					   &lineinfo_index_table_vt));

	memset(&sym_index_dict, 0, sizeof(sym_index_dict));
	cafe_dict_create(&sym_index_dict, sym_index_table_vt);
	memset(&lineinfo_index_dict, 0, sizeof(lineinfo_index_dict));
	cafe_dict_create(&lineinfo_index_dict, lineinfo_index_table_vt);

	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						     cafe_stdlib_allocator,
						     &actions));
	symresolve_set_symindex(actions, &sym_index_dict);
	symresolve_set_lineinfoindex(actions, &lineinfo_index_dict);

	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(actions, pid, 0));
	CUT_ASSERT(populate_map(pid, actions, &self_map));

	memset(&resolved_stack, 0, sizeof(resolved_stack));
	memset(&stack, 0, sizeof(stack));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 4, stacktrace_vt));
	stack_arr = stack.p;

	/* test basic success case */
	stack.l = 1;
	stack_arr[0] = (uintptr_t) test_symresolve;
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(actions, pid, &gen1));
	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid, gen1, &stack,
						resolved_indexed_vt,
						&resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0]);

	CUT_ASSERT_EQUAL(1, sym_index_dict.nentries);
	key[0].ull = 0;
	dict_row = cafe_dict_get(&sym_index_dict, key);
	CUT_ASSERT_NOT_NULL(dict_row);
	sym_entry = (struct cafe_normalized_symbol_entry *) dict_row->u.data;
	CUT_ASSERT_EQUAL(0, sym_entry->index);
	CUT_ASSERT_EQUAL(0, sym_entry->symbol_lineinfo_index);
	CUT_ASSERT_EQUAL((uintptr_t)test_symresolve,
			 sym_entry->object_address + test_base_addr);
	CUT_ASSERT_EQUAL(0, strcmp("test_symresolve", sym_entry->name));

	CUT_ASSERT_EQUAL(1, lineinfo_index_dict.nentries);
	key[0].ull = 0;
	dict_row = cafe_dict_get(&lineinfo_index_dict, key);
	CUT_ASSERT_NOT_NULL(dict_row);
	lineinfo_entry = (struct cafe_normalized_lineinfo_entry *) dict_row->u.data;
	CUT_ASSERT_EQUAL(0, lineinfo_entry->index);
	CUT_ASSERT_NOT_NULL(lineinfo_entry->file_name);
	CUT_ASSERT_EQUAL(0, strcmp(__FILE__, lineinfo_entry->file_name));
	CUT_ASSERT_NOT_EQUAL(0, lineinfo_entry->line);

	free(stack.p);
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_indexed_vt);
	symresolve_actions_delete(actions);
	cafe_dict_delete(&sym_index_dict);
	cafe_dict_delete(&lineinfo_index_dict);

	return 0;
}
#endif /* if defined(ENABLE_ContextMux) && ENABLE_ContextMux */

/**
 *  Make sure that addresses which only become valid after a memory map update
 * are resolved with the expected generation of the process' map.
 */
static
int test_dlopen(void)
{
	struct cafe_vm_dynarray stack, resolved_stack;
	struct symresolve_actions *actions;
	unsigned long long *stack_arr;
	struct resolved_address *results;
	unsigned long long gen1, gen2;
	pid_t pid = getpid();

	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						     cafe_stdlib_allocator,
						     &actions));
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(actions, pid, 0));
	CUT_ASSERT(populate_map(pid, actions, &self_map));
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(actions, pid, &gen1));

	memset(&resolved_stack, 0, sizeof(resolved_stack));
	memset(&stack, 0, sizeof(stack));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 1, stacktrace_vt));
	stack_arr = stack.p;

	CUT_ASSERT(populate_map(pid, actions, &mapped_regions));
	stack_arr[0] = mapped_address;

	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(actions, pid, &gen2));
	CUT_ASSERT_NOT_EQUAL(gen1, gen2);

	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid, gen1, &stack,
						resolved_entry_vt, &resolved_stack));
	results = resolved_stack.p;
	CUT_ASSERT_NOT_NULL(results[0].object_name);
	CUT_ASSERT_EQUAL(0, strcmp(UNMAPPED, results[0].object_name));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid, gen2, &stack,
						resolved_entry_vt, &resolved_stack));
	results = resolved_stack.p;
	CUT_ASSERT_NOT_NULL(strstr(results[0].object_name, MAPPED_LIB));
	CUT_ASSERT_NOT_NULL(results[0].name);
	CUT_ASSERT_EQUAL(0, strncmp(results[0].name, MAPPED_FUNC,
				    strlen(MAPPED_FUNC)));
	CUT_ASSERT_EQUAL(0, results[0].offset);
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	free(stack.p);
	symresolve_actions_delete(actions);
	return 0;
}


/**
 *  Make sure that addresses which only become valid after a memory map update
 * are resolved with the expected generation of the process' map.
 */
static
int test_remote_symresolve(void)
{
	struct cafe_vm_dynarray stack, resolved_stack;
	unsigned long long *stack_arr;
	unsigned long long gen1, gen2, kernel_gen, forked_gen, execed_gen;
	struct resolved_address *results;
	struct symresolve_actions *local_actions;
	struct symresolve_actions *remote_actions;
	const pid_t pid = getpid();
	struct cafe_value_data *mm_delta;

	/* build the mmap delta using the local symbol engine */
	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						     cafe_stdlib_allocator,
						     &local_actions));
	/* add a "kernel" address space */
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(local_actions, 0, 0));
	CUT_ASSERT(populate_map(0, local_actions, &kernel_regions));
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(local_actions, 0, &kernel_gen));

	/* add a non-exec()ed initial process */
	/* WRS_TODO: should we support non-exec address spaces? */
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(local_actions, pid, 0));
	CUT_ASSERT(populate_map(pid, local_actions, &self_map));
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(local_actions, pid, &gen1));

	/* add a second generation to the delta */
	CUT_ASSERT(populate_map(pid, local_actions, &mapped_regions));
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(local_actions, pid, &gen2));

	/* add a forked process */
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clone(local_actions,
							  pid, pid + 1));
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(local_actions, pid + 1,
						     &forked_gen));
	/* add an exec()ed generation */
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(local_actions, pid + 1,
							  0));
	CUT_ASSERT(populate_map(pid + 1, local_actions, &self_map));
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(local_actions, pid + 1,
						     &execed_gen));
	/* retire the exec()ed address space */
	CUT_ASSERT_CAFE_OK(symresolve_address_space_delete(local_actions,
							   pid + 1));

	/* populate "remote" symbol engine using the local's delta */
	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						     cafe_stdlib_allocator,
						     &remote_actions));

	CUT_ASSERT_CAFE_OK(cafe_value_create(mm_delta_vt, &mm_delta));
	CUT_ASSERT_CAFE_OK(symresolve_mm_gen_delta(local_actions, mm_delta->u.data));
	CUT_ASSERT_CAFE_OK(symresolve_mm_delta(remote_actions, mm_delta->u.data));
	symresolve_mm_clear_delta(local_actions, mm_delta->u.data);
	cafe_value_delete(mm_delta);
	symresolve_actions_delete(local_actions);

	memset(&resolved_stack, 0, sizeof(resolved_stack));
	memset(&stack, 0, sizeof(stack));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 4, stacktrace_vt));
	stack_arr = stack.p;

	/* test resolution of generation 1 address */
	stack.l = 1;
	stack_arr[0] = (uintptr_t) test_remote_symresolve;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(remote_actions, pid, gen1,
						&stack, resolved_entry_vt,
						&resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_EQUAL((uintptr_t)test_remote_symresolve,
			 test_base_addr + results[0].address);
	CUT_ASSERT_EQUAL(0, strcmp("test_remote_symresolve", results[0].name));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	/* test resolution of generation 2 address */
	stack_arr[0] = mapped_address;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(remote_actions, pid, gen2,
						&stack, resolved_entry_vt,
						&resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_NOT_NULL(strstr(results[0].object_name, MAPPED_LIB));
	CUT_ASSERT_EQUAL(0, strncmp(results[0].name, MAPPED_FUNC,
				    strlen(MAPPED_FUNC)));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	/* test resolution of forked process */
	stack_arr[0] = mapped_address;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(remote_actions, pid + 1,
						forked_gen, &stack,
						resolved_entry_vt, &resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_NOT_NULL(strstr(results[0].object_name, MAPPED_LIB));
	CUT_ASSERT_EQUAL(0, strncmp(results[0].name, MAPPED_FUNC,
				    strlen(MAPPED_FUNC)));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	/* test resolution of exec()ed generation */
	stack_arr[0] = (uintptr_t) test_remote_symresolve;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(remote_actions,
						pid + 1, execed_gen,
						&stack, resolved_entry_vt, &resolved_stack));
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_EQUAL((uintptr_t)test_remote_symresolve,
			 test_base_addr + results[0].address);
	CUT_ASSERT_EQUAL(0, strcmp("test_remote_symresolve", results[0].name));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	/* test parent address space resolution */
	stack.l = 4;
	stack_arr[0] = CAFE_STACK_MARKER_KERNEL;
	stack_arr[1] = kernel_address;
	stack_arr[2] = CAFE_STACK_MARKER_USER;
	stack_arr[3] = (uintptr_t) test_remote_symresolve;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(remote_actions, pid + 1,
						execed_gen, &stack,
						resolved_entry_vt, &resolved_stack));
	CUT_ASSERT_EQUAL(2, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_NOT_NULL(strstr(results[0].object_name, MAPPED_LIB));
	CUT_ASSERT_EQUAL(0, strncmp(results[0].name, MAPPED_FUNC,
				    strlen(MAPPED_FUNC)));
	CUT_ASSERT_EQUAL(0, results[1].offset);
	CUT_ASSERT_EQUAL((uintptr_t)test_remote_symresolve,
			 results[1].address + test_base_addr);
	CUT_ASSERT_EQUAL(0, strcmp("test_remote_symresolve", results[1].name));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	/* make sure flushing removes the first generation but not the second */
	stack.l = 1;
	CUT_ASSERT_CAFE_OK(symresolve_mm_flush(remote_actions));
	stack_arr[0] = (uintptr_t) test_remote_symresolve;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(remote_actions, pid, gen1,
						&stack, resolved_entry_vt,
						&resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, strncmp(UNMAPPED, results[0].name, 2));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&resolved_stack, 0,
						     resolved_entry_vt));
	CUT_ASSERT_CAFE_OK(symresolve_callstack(remote_actions, pid, gen2,
						&stack, resolved_entry_vt, &resolved_stack));
	cafe_vm_dynarray_set_size(&resolved_stack, 0, resolved_entry_vt);

	free(stack.p);
	symresolve_actions_delete(remote_actions);
	return 0;
}


static
int get_exe_path(char *buf, size_t bufsz)
{
	ssize_t len = readlink("/proc/self/exe", buf, bufsz);
	if (len != -1) {
		buf[len] = 0;
		return CAFE_OK;
	}
	return CAFE_ERROR;
}


/**
 *  
 */
static
int test_pathmap(void)
{
	static const char *FAUX_EXEC_NAME = "faux";
	struct cafe_vm_dynarray stack, resolved_stack;
	unsigned long long *stack_arr;
	struct resolved_address *results;
	struct symresolve_actions *actions;
	pid_t pid = getpid();
	unsigned int i;
	unsigned long long gen;
	char this_exe_path[PATH_MAX];
	
	CUT_ASSERT_CAFE_OK(get_exe_path(this_exe_path, sizeof this_exe_path));
	tcf_add_pathmap("faux", this_exe_path);

	/* build the mmap delta using the local symbol engine */
	cafe_config_channel_up = NULL;
	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						     cafe_stdlib_allocator,
						     &actions));
	CUT_ASSERT_CAFE_OK(symresolve_address_space_clear(actions, pid, 0));
	for (i = 0; i < self_map.region_count; ++i) {
		if (strstr(self_map.regions[i].name, this_exe_path) != NULL) {
			CUT_ASSERT_CAFE_OK(symresolve_mm_update(actions, pid,
								self_map.regions[i].start,
								self_map.regions[i].end - self_map.regions[i].start,
								self_map.regions[i].file_offset,
								FAUX_EXEC_NAME,
								true));
			break;
		}
	}
	CUT_ASSERT(i < self_map.region_count);

	memset(&resolved_stack, 0, sizeof(resolved_stack));
	memset(&stack, 0, sizeof(stack));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 1, stacktrace_vt));
	stack_arr = stack.p;
	stack_arr[0] = (uintptr_t) test_pathmap;

	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(actions, pid, &gen));
	CUT_ASSERT_CAFE_OK(symresolve_callstack(actions, pid, gen, &stack,
						resolved_entry_vt, &resolved_stack));

	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_EQUAL((uintptr_t)test_pathmap,
			 test_base_addr + results[0].address);
	CUT_ASSERT_EQUAL(0, strcmp("test_pathmap", results[0].name));

	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&resolved_stack, 0,
						     resolved_entry_vt));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 0, stacktrace_vt));
	symresolve_actions_delete(actions);
	return 0;
}


#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
/**
 *  Test basic source line resolution.
 */
static
int test_lineinfo(void)
{
	struct cafe_vm_dynarray stack, resolved_stack;
	unsigned long long *stack_arr;
	unsigned long long gen1;
	struct resolved_lineinfo *results;
	struct symresolve_actions *local_actions;
	pid_t pid = getpid();

	/* build the mmap delta using the local symbol engine */
	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						  cafe_stdlib_allocator,
						  &local_actions));
	populate_map(pid, local_actions, &self_map);
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(local_actions, pid, &gen1));


	memset(&resolved_stack, 0, sizeof(resolved_stack));
	memset(&stack, 0, sizeof(stack));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 1, stacktrace_vt));
	stack_arr = stack.p;

	/* test resolution of generation 1 address */
	stack_arr[0] = (uintptr_t) test_lineinfo;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(local_actions, pid, gen1,
						&stack,
						lineinfo_vt, &resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0].offset);
	CUT_ASSERT_EQUAL((uintptr_t)test_lineinfo,
			 test_base_addr + results[0].address);
	CUT_ASSERT_EQUAL(0, strcmp("test_lineinfo", results[0].name));
	CUT_ASSERT_NOT_NULL(results[0].file_name);
	CUT_ASSERT_EQUAL(0, strcmp(__FILE__, results[0].file_name));
	CUT_ASSERT_NOT_EQUAL(0, results[0].line_number);

	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&resolved_stack, 0,
						     lineinfo_vt));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 0, stacktrace_vt));
	symresolve_actions_delete(local_actions);
	return 0;
}


/**
 *  Verify the symbol resolution engine can normalize debug info entries
 * effectively.
 */
static
int test_indexed_lineinfo(void)
{
	struct cafe_vm_dynarray stack, resolved_stack;
	unsigned long long *stack_arr;
	unsigned long long gen1;
	unsigned long long *results;
	struct symresolve_actions *local_actions;
	pid_t pid = getpid();
	struct cafe_value_type *index_table_vt;
	struct cafe_dict_struct lineinfo_index;
	struct cafe_value_data *dict_row;
	cafe_any_value key[1];
	struct cafe_normalized_lineinfo_entry *entry;

	CUT_ASSERT_CAFE_OK(cafe_type_parse(ns, CAFE_NORMALIZED_LINEINFO_DEF
					   "D["CAFE_NORMALIZED_LINEINFO_REF"]",
					   &index_table_vt));

	memset(&lineinfo_index, 0, sizeof(lineinfo_index));
	cafe_dict_create(&lineinfo_index, index_table_vt);

	/* build the mmap delta using the local symbol engine */
	CUT_ASSERT_CAFE_OK(symresolve_actions_create(mm_delta_vt,
						  cafe_stdlib_allocator,
						  &local_actions));
	symresolve_set_lineinfoindex(local_actions, &lineinfo_index);
	populate_map(pid, local_actions, &self_map);
	CUT_ASSERT_CAFE_OK(symresolve_mm_current_gen(local_actions, pid, &gen1));


	memset(&resolved_stack, 0, sizeof(resolved_stack));
	memset(&stack, 0, sizeof(stack));
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 1, stacktrace_vt));
	stack_arr = stack.p;

	/* test resolution of generation 1 address */
	stack_arr[0] = (uintptr_t) test_lineinfo;
	CUT_ASSERT_CAFE_OK(symresolve_callstack(local_actions, pid, gen1,
						&stack,
						lineinfo_indexed_vt,
						&resolved_stack));
	CUT_ASSERT_EQUAL(1, resolved_stack.l);
	results = resolved_stack.p;
	CUT_ASSERT_EQUAL(0, results[0]);

	CUT_ASSERT_EQUAL(1, lineinfo_index.nentries);
	key[0].ull = 0;
	dict_row = cafe_dict_get(&lineinfo_index, key);
	CUT_ASSERT_NOT_NULL(dict_row);
	entry = (struct cafe_normalized_lineinfo_entry *) dict_row->u.data;
	CUT_ASSERT_EQUAL(0, entry->index);
	CUT_ASSERT_NOT_NULL(entry->file_name);
	CUT_ASSERT_EQUAL(0, strcmp(__FILE__, entry->file_name));
	CUT_ASSERT_NOT_EQUAL(0, entry->line);

	free(results);
	CUT_ASSERT_CAFE_OK(cafe_vm_dynarray_set_size(&stack, 0, stacktrace_vt));
	symresolve_actions_delete(local_actions);
	cafe_dict_delete(&lineinfo_index);
	return 0;
}
#endif  /* if ENABLE_ContextMux */


/**
 *
 */
static
void relocate_regions(const struct memory_region *r1,
		      const struct memory_region *r2,
		      struct memory_map *in, unsigned int *index)
{
	for (; *index < in->region_count; ++(*index)) {
		size_t len, avail;

		len = in->regions[*index].end - in->regions[*index].start;
		avail = r2->start - r1->end;
		if (avail < len)
			return;

		in->regions[*index].start = r1->end;
		in->regions[*index].end = in->regions[*index].start + len;
		r1 = &in->regions[*index];
	}
}


static
bool suite_setup(void)
{
	void *lib_handle, *mapped_func;
	uintptr_t mapped_base = 0;
	struct memory_region r;
	unsigned int i, j;
	int rc;

	if (!memory_map_get_self(&self_map)) {
		CUT_PRINTF_ERR("Error reading memory map, skipping symresolve"
			" API test suite\n");
		return false;
	}

	mapped_regions.regions = malloc(10 * sizeof(mapped_regions.regions[0]));
	if (mapped_regions.regions == NULL)
		return false;
	kernel_regions.regions = malloc(10 * sizeof(kernel_regions.regions[0]));
	if (kernel_regions.regions == NULL)
		return false;
	mapped_regions.region_count = 0;
	kernel_regions.region_count = 0;
	for (i = 0; i < self_map.region_count; ++i) {
		r = self_map.regions[i];
		if (strstr(r.name, MAPPED_LIB) != NULL) {
			mapped_regions.regions[mapped_regions.region_count++] = r;
			if (mapped_base == 0)
				mapped_base = r.start;
			r.start += 0x1fff;
			r.end += 0x1fff;
			kernel_regions.regions[kernel_regions.region_count++] = r;
		}
		if ((test_base_addr == 0) && (strstr(r.name, "cafe-cfg-unit-test-suite") != NULL))
			test_base_addr = r.start;

		if (mapped_regions.region_count >= 10)
			return false;
	}
	if (mapped_regions.region_count == 0)
		return false;

	j = 0;
	r = ADDRESS_SPACE_START;
	for (i = 0; i < self_map.region_count; ++i) {
		relocate_regions(&r, &self_map.regions[i], &mapped_regions, &j);
		r = self_map.regions[i];
	}
	if (j != mapped_regions.region_count)
		return false;

	lib_handle = dlopen(mapped_regions.regions[0].name, RTLD_LAZY);
	if (lib_handle == NULL) {
		CUT_PRINTF_ERR("Error mapping %s: %s\n",
			mapped_regions.regions[0].name, dlerror());
		return false;
	}
	mapped_func = dlsym(lib_handle, MAPPED_FUNC);
	dlclose(lib_handle);
	if (mapped_func == NULL) {
		CUT_PRINTF_ERR("Error locating symbol \"%s\": %s\n",
			MAPPED_FUNC, dlerror());
		return false;		
	}

	mapped_address = mapped_regions.regions[0].start +
		(uintptr_t) mapped_func - mapped_base;

	kernel_address = kernel_regions.regions[0].start +
		(uintptr_t) mapped_func - mapped_base;

	rc = pthread_create(&event_thread, NULL, tcf_event_thread, NULL);
	if (rc != 0) {
		CUT_PRINTF_ERR("Error starting TCF event thread: %s\n",
			strerror(rc));
		return false;
	}

	while (!cafe_atomic_get(&tcf_init_complete))
		sched_yield();
	return true;
}


CUT_BEGIN_SUITE_F(symresolve_api_suite, &setup, &teardown)
	if (!suite_setup()) {
		CUT_PRINTF_ERR("Error initializing, skipping symresolve"
			" API test suite\n");
		return;
	}
	CUT_RUN_TEST(test_symresolve_lifecycle);
	CUT_RUN_TEST(test_symresolve);
	CUT_RUN_TEST(test_dlopen);
	CUT_RUN_TEST(test_remote_symresolve);
	CUT_RUN_TEST(test_pathmap);
#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	CUT_RUN_TEST(test_lineinfo);
	CUT_RUN_TEST(test_indexed_lineinfo);
	CUT_RUN_TEST(test_indexed_symresolve_with_lineinfo);
#endif  /* if ENABLE_ContextMux */
	CUT_RUN_TEST(test_indexed_symresolve);
	cancel_event_loop();
	memory_map_destroy(&self_map);
	memory_map_destroy(&mapped_regions);
	memory_map_destroy(&kernel_regions);
	pthread_join(event_thread, NULL);
CUT_END_SUITE_F
