/* test_user_event.c - test user events */

/* Copyright (c) 2010-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <assert.h>
#include <cafe/ep/user_event.h>
#include <test/cafe_CUnitTest.h>
#include "verify_cfg.h"

#if defined(_MSC_VER)
    /* Turn off message: warning C4090: '=' : different 'const' qualifiers*/
#   pragma warning (disable: 4090)
#endif


#define TEST_ITERS (43)

static bool setup_complete = false;	/* setup routine completed OK */


static
void setup(void)
{
	int rc;

	assert(!setup_complete);
	CAFE_LOCK();

	rc = verify_cfg_register();
	if (rc == CAFE_OK)
		setup_complete = true;
}


static
void teardown(void)
{
	if (setup_complete)
		verify_cfg_unregister();

	CAFE_UNLOCK();
	setup_complete = false;
}

#define USER_EVENT_1 "S[user_event_1,v[arg1,L]v[arg2,L]v[arg3,s]]"
struct user_event_1_struct {
	unsigned long arg1;
	unsigned long arg2;
	const char *arg3;
};

/**
 *  Enable the output event with the given type string on the given EP
 * instance.
 */
static
int enable_event(struct cafe_ep_inst *inst, const char *vt)
{
	struct cafe_event_descr *event;
	struct cafe_value_type *types[1];
	int rc;

	rc = cafe_type_parse(inst->session->ns, vt, &types[0]);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_cfg_get_event(inst->session, 1, types, &event);
	if (rc != CAFE_OK)
		return rc;

	return cafe_ep_inst_enable_output_event(inst, event);
}

static
int test_user_event(void)
{
	struct cafe_instrument *session;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst[2];

	struct cafe_cfg_setting settings[2];
	struct verify_cfg_workarea *vcw;
	struct event_verify_spec *events;
	unsigned int i;
	struct cafe_user_event_handle *l6_handle;

	static const char *L1_TYPE[] = {user_event_signed_VT_DEF};
	static const char *L2_TYPE[] = {user_event_signed_VT_DEF};
	static const char *L3_TYPE[] = {user_event_float_VT_DEF};
	static const char *L4_TYPE[] = {user_event_string_VT_DEF};
	static const char *L5_TYPE[] = {user_event_unsigned_VT_DEF};
	static const char *L6_TYPE[] = {USER_EVENT_1};

	struct user_event_signed l1 = { (char *) "l1", 100 };
	struct user_event_signed l2 = { (char *) "l2", -200 };
	struct user_event_float l3 = { (char *) "l3", 3.1415926 };
	struct user_event_string l4 = { (char *) "l4",
					(char *) "last was a float" };
	struct user_event_unsigned l5 = { (char *) "l5", 345 };
	struct user_event_1_struct l6 = { 123, 234, (char *) "l6" };

	CUT_ASSERT_TRUE(setup_complete);

	CUT_ASSERT_CAFE_OK(cafe_user_event_add(
				cafe_stdlib_allocator,
				USER_EVENT_1,
				&l6_handle));

	CUT_ASSERT_CAFE_OK(cafe_instrument_create("test_user_event", &session));

	spec = cafe_ep_spec_find("vxworks_kernel");
	CUT_ASSERT_NOT_NULL(spec);

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       0, NULL, &inst[0]));

	spec = cafe_ep_spec_find("verify_cfg");
	CUT_ASSERT_NOT_NULL(spec);

	/* create verification input */
	events = malloc(TEST_ITERS * sizeof *events);
	CUT_ASSERT_NOT_NULL(events);

	i = 0;
	events[i].types.l = 1;
	events[i].types.p = L1_TYPE;
	events[i].data.l = sizeof l1;
	events[i].data.p = &l1;
	i++;

	events[i].types.l = 1;
	events[i].types.p = L2_TYPE;
	events[i].data.l = sizeof l2;
	events[i].data.p = &l2;
	i++;

	events[i].types.l = 1;
	events[i].types.p = L3_TYPE;
	events[i].data.l = sizeof l3;
	events[i].data.p = &l3;
	i++;

	events[i].types.l = 1;
	events[i].types.p = L4_TYPE;
	events[i].data.l = sizeof l4;
	events[i].data.p = &l4;
	i++;

	events[i].types.l = 1;
	events[i].types.p = L5_TYPE;
	events[i].data.l = sizeof l5;
	events[i].data.p = &l5;
	i++;

	events[i].types.l = 1;
	events[i].types.p = L6_TYPE;
	events[i].data.l = sizeof l6;
	events[i].data.p = &l6;
	i++;

	settings[0].name = "events";
	settings[0].value.dary.l = i;
	settings[0].value.dary.p = events;

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(session, spec,
					       1, settings, &inst[1]));

	free(events);

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));
	CUT_ASSERT_CAFE_OK(enable_event(inst[0], user_event_signed_VT_REF));
	CUT_ASSERT_CAFE_OK(enable_event(inst[0], user_event_unsigned_VT_REF));
	CUT_ASSERT_CAFE_OK(enable_event(inst[0], user_event_float_VT_REF));
	CUT_ASSERT_CAFE_OK(enable_event(inst[0], user_event_string_VT_REF));
	CUT_ASSERT_CAFE_OK(cafe_instrument_start(session));

	CAFE_UNLOCK();

	cafe_user_event_signed("l1", 100);
	cafe_user_event_signed("l2", -200);
	cafe_user_event_float("l3", 3.1415926);
	cafe_user_event_string("l4", "last was a float");
	cafe_user_event_unsigned("l5", 345);
	cafe_user_event_generate(l6_handle, &l6);

	CAFE_LOCK();

	cafe_user_event_release(l6_handle);

	/* clean up, disconnect, and deallocate configurations... */
	CUT_ASSERT_CAFE_OK(cafe_instrument_stop(session));
	vcw = (struct verify_cfg_workarea *)inst[1]->workarea->u.data;
	CUT_ASSERT_EQUAL(vcw->failed, 0);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_disconnect(inst[0], inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_delete(inst[1]));
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_delete(inst[0]));

	CUT_ASSERT_CAFE_OK(cafe_instrument_delete(session));

	return 0;
}


CUT_BEGIN_SUITE_F(user_event_suite, &setup, &teardown)
	CUT_RUN_TEST(test_user_event);
CUT_END_SUITE_F
