/* Copyright (c) 2011-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
08jul14,sdt  Used CUT_PRINTF/CUT_PRINT_ERR macro to make clear output
*/

#include "../../src/cafe_os.h"
#include <string.h>
#include <test/cafe_CUnitTest.h>
#include <cafe/atomic.h>
#include <cafe/autogenerated/vxworks.type.h>
#include <cafe/ep/vxworks_kernel.h>
#include <assert.h>
#include "../vxworks_events.h"

#ifdef __VXWORKS__
#include <memPartLib.h>
#endif

#define EP_NAME "vxworks_skipcount_test"

#define WORKAREA_DEF "S[" EP_NAME ".workarea,v[dummy,q]]"
#define WORKAREA_REF "t[" EP_NAME ".workarea]"

#define NTHREADS 10
#define NEVENTS  1000

#define TEST_EVENT_ID EVENT_MEMALLOC

#define TEST_EVENT_DEF				\
	vxWorks_memAlloc_VT_DEF

#define TEST_EVENT_REF				\
	vxWorks_memAlloc_VT_REF

#define ALL_TYPES				\
	WORKAREA_DEF				\
	TEST_EVENT_REF

#if defined(__VXWORKS__)
static PART_ID mem_part_id;
static char *pool;
static unsigned int poolsize = NTHREADS * NEVENTS + 1000;
#endif

static struct cafe_instrument *instrument = NULL;
static struct cafe_ep_inst *inst[2];
static bool input_event_enabled = false;

static const char *test_event_TYPE[] = {
	TEST_EVENT_REF
};

static const char *skipcount_default[] = {
	TEST_EVENT_REF,
	cafe_timestamp_VT_REF,
	cafe_skipcount_VT_REF
};

static const char *skipcount_as_2[] = {
	TEST_EVENT_REF,
	cafe_timestamp_VT_REF,
	"P[cafe.skipcount,2]"
};

static const char *skipcount_as_20[] = {
	TEST_EVENT_REF,
	cafe_timestamp_VT_REF,
	"P[cafe.skipcount,20]"
};

struct {
	cafe_atomic_val action_handler_count;
	unsigned long long skipcount_value;
	cafe_atomic_val event_count;
} event_data;

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT(test_event)
};

static void action_handler(struct cafe_event_action_info *info, void **data)
{
#if defined(__VXWORKS__)
	struct vxWorks_memAlloc *e =
		(struct vxWorks_memAlloc *)data[0];
#endif
	assert(input_event_enabled);
#if defined(__VXWORKS__)
	if (e->partId != (unsigned long)mem_part_id)
		return;
	assert(e->nBytesRequested == 1);
#endif
	(void)cafe_atomic_val_inc(&event_data.action_handler_count);

	if (info->event->num_types == 3) {
		struct cafe_skipcount *skipcount;

		skipcount = (struct cafe_skipcount *)data[2];
		event_data.skipcount_value = skipcount->value;
		(void)cafe_atomic_val_add(&event_data.event_count, 
		                    (cafe_atomic_val)skipcount->value);
	}
}

static int get_input_event_actions(struct cafe_ep_inst *inst,
				   struct cafe_event_descr *event,
				   struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *info;

	if (!input_event_enabled)
		return CAFE_OK;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &info);

	if (rc != CAFE_OK || info == NULL)
		return rc;

	return cafe_add_event_action(inst, actions, action_handler, info);
}

static struct cafe_ep_spec spec = {
	EP_NAME,
	NULL,			/* create */
	NULL,			/* delete */
	NULL,			/* start */
	NULL,			/* stop */
	get_input_event_actions,
	NULL,			/* update outputs */
	NULL,			/* inject event */
	ALL_TYPES,
	WORKAREA_REF,
	CAFE_NELEMENTS(input_events),
	input_events
};

static bool skipcount_ep_registered = false;

static void skipcount_ep_unregister(void)
{
	if (skipcount_ep_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		skipcount_ep_registered = false;
	}
}

static int skipcount_ep_register(void)
{
	int rc = CAFE_OK;

	if (!skipcount_ep_registered) {
		rc = cafe_ep_spec_register(&spec);
		if (rc == CAFE_OK)
			skipcount_ep_registered = true;
	}

	return rc;
}

/*
 * I'm deliberately not asserting much in here: we have other tests
 * that verify instrument instantiation.
 */
static int build_instrument(void)
{
	int rc = CAFE_OK;
	size_t i;
	struct cafe_ep_spec *spec[2];

	CAFE_LOCK();

	if ((rc = skipcount_ep_register()) != CAFE_OK)
		goto cafe_error;

	if ((rc = cafe_instrument_create(
		     EP_NAME, &instrument)) != CAFE_OK)
		goto cafe_error;

	spec[0] = cafe_ep_spec_find("vxworks_kernel");
	spec[1] = cafe_ep_spec_find(EP_NAME);
	CUT_ASSERT_NOT_NULL(spec[0]);
	CUT_ASSERT_NOT_NULL(spec[1]);

	rc = cafe_ep_inst_create(instrument, spec[0], 0, NULL, &inst[0]);
	if (rc != CAFE_OK)
		goto cafe_error;

	rc = cafe_ep_inst_create(instrument, spec[1], 0, NULL, &inst[1]);
	if (rc != CAFE_OK)
		goto cafe_error;

	for (i = 0; i < spec[1]->n_input_events; i++) {
		struct cafe_event_descr *const t = inst[1]->input_events[i];

		rc = cafe_ep_inst_enable_output_event(inst[0], t);
		if (rc != CAFE_OK)
			goto cafe_error;
	}

	/* Bypass threading compatibility checks. */
	inst[0]->flags |= CAFE_EP_INST_FLAG_OUTPUT_SESSION_THREAD;

	rc = cafe_ep_inst_connect(inst[0], inst[1]);
	if (rc != CAFE_OK)
		goto cafe_error;

	input_event_enabled = false;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	if (rc != CAFE_OK)
		goto cafe_error;

	rc = cafe_instrument_start(instrument);
	if (rc != CAFE_OK)
		goto cafe_error;

	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();
	return 0;

cafe_error:
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();
	return 1;
}

static int delete_instrument(void)
{
	if (instrument != NULL) {
		CAFE_LOCK();
		CUT_ASSERT_CAFE_OK(cafe_instrument_delete(instrument));
		CAFE_UNLOCK();
		instrument = NULL;
	}
	CAFE_LOCK();
	skipcount_ep_unregister();
	CAFE_UNLOCK();
	return 0;
}

static void dispatch_event(size_t count)
{
	size_t i;

	for (i = 0; i < count; i++) {
#if defined(__VXWORKS__)
		void *p = memPartAlloc(mem_part_id, 1);
		(void)memPartFree(mem_part_id, p);
#else
		cafe_cfg_vxworks_ev_process_id(TEST_EVENT_ID, NULL, NULL);
#endif
	}
}

static int test_with_invalid_limit(void)
{
	int rc;
	size_t i;
	unsigned int j;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;
	static const char *type1_str[] = {
		TEST_EVENT_DEF,
		cafe_timestamp_VT_REF,
		"P[cafe.skipcount,4294967296]",
	};
	static const char *type2_str[] = {
		TEST_EVENT_DEF,
		cafe_timestamp_VT_REF,
		"P[cafe.skipcount,0]"
	};
	const char **args[2] = { type1_str, type2_str };

	CUT_ASSERT_NOT_NULL(inst[0]);
	CUT_ASSERT_NOT_NULL(inst[1]);

	CAFE_LOCK();

	for (i = 0; i < 2; i++) {
		for (j = 0; j < 3; j++) {
			rc = cafe_type_parse(instrument->ns,
					     args[i][j],
					     &types[j]);
			CUT_ASSERT_CAFE_OK(rc);
		}

		CUT_ASSERT_EQUAL(3, j);
		rc = cafe_cfg_get_event(instrument, j, types, &event);
		CUT_ASSERT_CAFE_OK(rc);
		CUT_ASSERT_EQUAL(3, event->num_types);
		rc = cafe_ep_inst_enable_output_event(inst[0], event);
		CUT_ASSERT_CAFE_OK(rc);
		input_event_enabled = true;
		rc = cafe_instrument_refresh_output_event_actions(instrument);
		CUT_ASSERT_EQUAL(CAFE_ERROR, rc);
		rc = cafe_ep_inst_disable_output_event(inst[0], event);
		CUT_ASSERT_CAFE_OK(rc);
		input_event_enabled = false;
		rc = cafe_instrument_refresh_output_event_actions(instrument);
		CUT_ASSERT_CAFE_OK(rc);
	}

	CAFE_UNLOCK();
	return 0;
}

static int test_skipcount(void)
{
	int rc;
	unsigned int i;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;

	CAFE_LOCK();

	/* Test with default skipcount value. */

	for (i = 0; i < 3; i++) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_default[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}

	CUT_ASSERT_EQUAL(3, i);
	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(3, event->num_types);
	rc = cafe_ep_inst_enable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	memset(&event_data, 0, sizeof event_data);
	dispatch_event(5);
	CUT_ASSERT_EQUAL(5, event_data.action_handler_count);
	CUT_ASSERT_EQUAL(1, event_data.skipcount_value);
	rc = cafe_ep_inst_disable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = false;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);

	/* Test with skipcount = 2. */

	for (i = 0; i < 3; i++) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_as_2[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}

	CUT_ASSERT_EQUAL(3, i);
	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(3, event->num_types);
	rc = cafe_ep_inst_enable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	memset(&event_data, 0, sizeof event_data);
	dispatch_event(5);
	CUT_ASSERT_EQUAL(2, event_data.action_handler_count);
	dispatch_event(2);
	CUT_ASSERT_EQUAL(2, event_data.skipcount_value);
	CUT_ASSERT_EQUAL(3, event_data.action_handler_count);
	rc = cafe_ep_inst_disable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = false;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);

	/* Test with skipcount = 20. */

	for (i = 0; i < 3; i++) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_as_20[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}

	CUT_ASSERT_EQUAL(3, i);
	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(3, event->num_types);
	rc = cafe_ep_inst_enable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	memset(&event_data, 0, sizeof event_data);
	dispatch_event(19);
	CUT_ASSERT_EQUAL(0, event_data.skipcount_value);
	CUT_ASSERT_EQUAL(0, event_data.action_handler_count);
	dispatch_event(1);
	CUT_ASSERT_EQUAL(20, event_data.skipcount_value);
	CUT_ASSERT_EQUAL(1, event_data.action_handler_count);
	memset(&event_data, 0, sizeof event_data);
	for (i = 0; i < 2000 / 20; i++) {
		dispatch_event(20);
		CUT_ASSERT_EQUAL(i + 1 % 20, event_data.action_handler_count);
		CUT_ASSERT_EQUAL(20, event_data.skipcount_value);
	}
	CUT_ASSERT_EQUAL(2000 / 20, event_data.action_handler_count);
	rc = cafe_ep_inst_disable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);

	CAFE_UNLOCK();

	return 0;
}

struct multithreaded_context {
	unsigned int n_events;
	cafe_atomic_val setup_complete;
};

static void *thread_worker(void *arg)
{
	struct multithreaded_context *ctx;

	ctx = (struct multithreaded_context *)arg;
	while (!cafe_atomic_get(&ctx->setup_complete)) {
#if defined(__unix__) || defined(__VXWORKS__)
		(void)sched_yield();
#elif defined(_WIN32)
		Sleep(0);
#endif
	}
	dispatch_event(ctx->n_events);
	return NULL;
}

static int test_multithreaded_with_skipcount_using_defaults(void)
{
	int rc;
	unsigned int i;
	pthread_t tid[NTHREADS];
	struct multithreaded_context ctx;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;

	/* Test with skipcount default limit. */

	CAFE_LOCK();

	for (i = 0; i < 3; i++) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_default[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}
	CUT_ASSERT_EQUAL(3, i);
	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(3, event->num_types);
	rc = cafe_ep_inst_enable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();

	ctx.n_events = NEVENTS;
	cafe_atomic_set(&ctx.setup_complete, 0);

	for (i = 0; i < CAFE_NELEMENTS(tid); i++) {
		rc = pthread_create(&tid[i], NULL, thread_worker, &ctx);
		CUT_ASSERT_EQUAL(0, rc);
	}

	cafe_atomic_set(&ctx.setup_complete, 1);

	for (i = 0; i < CAFE_NELEMENTS(tid); i++) {
		rc = pthread_join(tid[i], NULL);
		CUT_ASSERT_EQUAL(0, rc);
	}

	CUT_ASSERT_TRUE(event_data.event_count >= 1);
	CUT_ASSERT_TRUE(event_data.event_count <= (cafe_atomic_val)((ctx.n_events * NTHREADS) + NTHREADS * NUM_ROOT_INSTANCES));
	CUT_ASSERT_EQUAL(1, event_data.event_count / event_data.action_handler_count);

	CAFE_LOCK();
	rc = cafe_ep_inst_disable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = false;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();

	return 0;
}

static int test_multithreaded_with_skipcount_set_as_twenty(void)
{
	int rc;
	unsigned int i;
	pthread_t tid[NTHREADS];
	struct multithreaded_context ctx;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;

	CAFE_LOCK();

	for (i = 0; i < 3; i++) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_as_20[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}
	CUT_ASSERT_EQUAL(3, i);
	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(3, event->num_types);
	rc = cafe_ep_inst_enable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();

	ctx.n_events = NEVENTS;
	cafe_atomic_set(&ctx.setup_complete, 0);

	for (i = 0; i < CAFE_NELEMENTS(tid); i++) {
		rc = pthread_create(&tid[i], NULL, thread_worker, &ctx);
		CUT_ASSERT_EQUAL(0, rc);
	}

	cafe_atomic_set(&ctx.setup_complete, 1);

	for (i = 0; i < CAFE_NELEMENTS(tid); i++) {
		rc = pthread_join(tid[i], NULL);
	}

#if defined(__VXWORKS__)
	CUT_PRINTF("%s:%d: event_count: %d\n", __FILE__, __LINE__,
		event_data.event_count);
	CUT_PRINTF("%s:%d: max_events : %d\n", __FILE__, __LINE__,
	       ctx.n_events * (NTHREADS));
#endif
	CUT_ASSERT_TRUE(event_data.event_count >= 20);
	CUT_ASSERT_TRUE(event_data.event_count >= (cafe_atomic_val)(ctx.n_events * (NTHREADS -1)));
	CUT_ASSERT_TRUE(event_data.event_count <= (cafe_atomic_val)(ctx.n_events * NTHREADS) + NTHREADS * NUM_ROOT_INSTANCES);
	CUT_ASSERT_EQUAL(20, event_data.event_count / event_data.action_handler_count);
	CAFE_LOCK();
	rc = cafe_ep_inst_disable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();

	return 0;
}

static void setup(void)
{
#if defined(__VXWORKS__)
	pool = malloc(poolsize);
	assert(pool != NULL);
	mem_part_id = memPartCreate(pool, poolsize);
	assert(mem_part_id != NULL);
#endif
	event_data.action_handler_count = 0;
	event_data.skipcount_value = 0;
	event_data.event_count = 0;
}

static void teardown(void)
{
#if defined(__VXWORKS__)
	(void)memPartDelete(mem_part_id);
	free(pool);
#endif
}

CUT_BEGIN_SUITE_F(vxworks_skipcount_suite, setup, teardown)
CUT_RUN_TEST(build_instrument);
CUT_RUN_TEST(test_with_invalid_limit);
CUT_RUN_TEST(test_skipcount);
#if !defined(_WIN32)
/* There's a number of issues with multithreaded event dispatch on windows
   hosts.  We're disabling these tests to unblock automated testing until
   they're fixed. */
CUT_RUN_TEST(test_multithreaded_with_skipcount_using_defaults);
CUT_RUN_TEST(test_multithreaded_with_skipcount_set_as_twenty);
#endif
CUT_RUN_TEST(delete_instrument);
CUT_END_SUITE_F
