/* vxworks_kernel.c - Cafe configuration for vxWorks */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
30jun14,sdt  Add support of PPC decrementer int enter - WB4-3797
18apr14,wzc  ignore 2 frame for x86 32bit also.(WB4-3276)
12mar14,tle  Disable use of alloca() on PPC and ARM (WB4-2840, WB4-3122)


DESCRIPTION

  This Event Processor routes events from the vxWorks OS into the Cafe
analysis system. In the Cafe system, it appears as an EP which can be
instantiated multiple times. It has no input events, a minimal private
workarea, and output events.

  It is also a root provider of user defined events.  For the simplest and most
common use cases, it has APIs to generate predefined user events with a string
"label" and a "value" of signed, unsigned, float or string.  For more complex
types of data, this EP has APIs to create new event type(s), generate events of
the created type(s) and finally release the handle to the created type(s).
*/


#include <string.h>
#include <limits.h>
#include <stdio.h>
#include <cafe/cafe.h>
#include <cafe/expr.h>
#include <cafe/utils.h>
#include "../src/cafe_assert.h"
#include "vxworks_events.h"
#include <cafe/autogenerated/vxworks.type.h>
#include <cafe/ep/vxworks_kernel.h>
#include <cafe/ep/user_event.h>

#if defined(__VXWORKS__)
#  include <taskLib.h>
#  include <private/eventP.h>
#  include <dataCollectorLib.h>
#  if defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
#    include <vxworks/cafe_vx_stacktrace.h>
#  endif
#endif

#ifndef ALLOW_MULTIPLE_STATE_EVENTS
#define ALLOW_MULTIPLE_STATE_EVENTS
#endif

/* These events are not yet implemented in the OS */
#ifndef CAFE_VXWORKS_POWER_MANAGEMENT_EVENTS
/* #define CAFE_VXWORKS_POWER_MANAGEMENT_EVENTS */
#endif

#if defined(__VXWORKS__) && defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
#  define MAX_STACK_PARAM  "v[max_stack_depth,I," cafe_setting_VT_NAT "]"
#  define MAX_STACK_DEPTH  (128)
#  if (_VX_CPU_FAMILY != _VX_PPC) && (_VX_CPU_FAMILY != _VX_ARM)
#    ifdef  __GNUC__
#      define CAFE_ALLOCA(size)   __builtin_alloca(size)
#    elif defined(__DCC__)
#      define CAFE_ALLOCA(size)   __alloca(size)
#    elif defined(__ICC)
#      define CAFE_ALLOCA(size)   _alloca(size)
#    endif
#  endif  /* if (_VX_CPU_FAMILY != _VX_PPC) */
#  if (_VX_CPU_FAMILY == _VX_I80X86) || (_VX_CPU_FAMILY == _VX_SIMNT)	\
	|| (_VX_CPU_FAMILY == _VX_SIMLINUX)
	/* x86 targets need to skip:
	     - cafe_cfg_vxworks_ev_process_handle
	     - eventPointSwExec */
        static const unsigned int FRAMES_TO_SKIP = 2;

#  elif (_VX_CPU_FAMILY == _VX_PPC)
    /* PPC targets need to skip:
             - cafe_cfg_vxworks_ev_process_handle(),
             - eventPointSwExec(), and
             - eventPointSwHandle() */
        static const unsigned int FRAMES_TO_SKIP = 3;

#  elif (_VX_CPU_FAMILY == _VX_ARM)
    /* ARM targets need to skip:
             - cafe_cfg_vxworks_ev_process_handle(),
             - eventPointSwExec(), and
             - eventPointSwHandle() */
        static const unsigned int FRAMES_TO_SKIP = 3;

#  else
#    warning FRAMES_TO_SKIP is not set for this architecture
        static const unsigned int FRAMES_TO_SKIP = 0;
#  endif  /* if (_VX_CPU_FAMILY == XXX) */

#else
#  define MAX_STACK_PARAM  ""
#endif

#define EVENT_TYPEDEFS					\
	cafe_setting_VT_DEF				\
	cafe_timestamp_config_VT_DEF			\
	cafe_num_cores_VT_DEF				\
	cafe_timestamp_VT_DEF				\
	cafe_core_id_VT_DEF				\
	cafe_CURRENT_CORE_ID_VT_DEF			\
	cafe_stacktrace_VT_DEF				\
	cafe_skipcount_VT_DEF				\
	VXWORKS_TYPE_ALL_VT_DEFS			\
	USER_EVENT_TYPE_ALL_VT_DEFS			\
	"S[vxWorks.workarea," MAX_STACK_PARAM "]"

/* statedump_request states */
enum {
	STATEDUMP_NONE,
	STATEDUMP_REQUEST,
	STATEDUMP_DONE
};

static int dump_state_for_inst(struct cafe_ep_inst *inst, bool from_start);

#if defined(__VXWORKS__) && defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
/* Per-vxWorks instance configuration data */
struct workarea {
	unsigned int max_stack_depth;
};
#endif

#if (((_WRS_VXWORKS_MAJOR > 6) || \
	  (_WRS_VXWORKS_MAJOR == 6) && (_WRS_VXWORKS_MINOR > 8)) && \
      defined(_WRS_CONFIG_RTP))
#    define RTP_INSTRUMENTATION 1
#else
#    define RTP_INSTRUMENTATION 0
#endif

/* globals */

bool cafe_vx_ignore_allocations = true;

/* locals */

static bool vxWorksInitialized = false;
static struct cafe_ep_inst *instance_list[NUM_ROOT_INSTANCES];

struct vxworks_inst_data {
	unsigned int index;
	bool started;
	long change_count;
};

#define EV_HANDLER_INFO_INIT(EVENT_ID)					\
static struct vxworks_event_handle ev_handler_info_##EVENT_ID = {	\
	{0}, {0}, {0}, {0}, {0},					\
	EVENT_##EVENT_ID						\
};

EV_HANDLER_INFO_INIT(TASKSPAWN)
EV_HANDLER_INFO_INIT(TASKDELAY)
EV_HANDLER_INFO_INIT(MEMALLOC)
EV_HANDLER_INFO_INIT(MEMFREE)
EV_HANDLER_INFO_INIT(MEMREALLOC)
EV_HANDLER_INFO_INIT(MEMPARTCREATE)
EV_HANDLER_INFO_INIT(MEMADDTOPOOL)
EV_HANDLER_INFO_INIT(INT_EXIT_K)
EV_HANDLER_INFO_INIT(INT_EXIT)
EV_HANDLER_INFO_INIT(INT_ENTER)
#if defined(__VXWORKS__) && (_VX_CPU_FAMILY == _VX_PPC)
    /* Add specific PPC events
     * Note: these events don't follow event id naming convention so redefine
     * locally event ID to be able to use EV_HANDLER_INFO_INIT macro
     */
    #define EVENT_PPC_DECR_INT_ID	PPC_DECR_INT_ID
    #define EVENT_PPC_WD_INT_ID		PPC_WD_INT_ID

    EV_HANDLER_INFO_INIT(PPC_DECR_INT_ID)
    EV_HANDLER_INFO_INIT(PPC_WD_INT_ID)
#endif
EV_HANDLER_INFO_INIT(WIND_EXIT_DISPATCH)
EV_HANDLER_INFO_INIT(WIND_EXIT_DISPATCH_PI)
EV_HANDLER_INFO_INIT(WIND_EXIT_IDLE)
EV_HANDLER_INFO_INIT(TASKLOCK)

#if defined(__VXWORKS__) && ((_WRS_VXWORKS_MAJOR > 6) || \
	((_WRS_VXWORKS_MAJOR == 6) && (_WRS_VXWORKS_MINOR > 8)) || \
    ((_WRS_VXWORKS_MAJOR == 6) && (_WRS_VXWORKS_MINOR == 8) && (_WRS_VXWORKS_MAINT >= 2)))

EV_HANDLER_INFO_INIT(TASKUNLOCK_ZERO)
EV_HANDLER_INFO_INIT(TASKUNLOCK_NONZERO)

#else

EV_HANDLER_INFO_INIT(TASKUNLOCK)

#endif

EV_HANDLER_INFO_INIT(TASKNAME)
EV_HANDLER_INFO_INIT(TICKANNOUNCE)
EV_HANDLER_INFO_INIT(WINDDELETE)
EV_HANDLER_INFO_INIT(TIMESTAMP_CONFIG)
EV_HANDLER_INFO_INIT(NUM_CPU)
EV_HANDLER_INFO_INIT(IOCREAT)
EV_HANDLER_INFO_INIT(IOOPEN)
EV_HANDLER_INFO_INIT(IODELETE)
EV_HANDLER_INFO_INIT(IOCLOSE)
EV_HANDLER_INFO_INIT(IOREAD)
EV_HANDLER_INFO_INIT(IOWRITE)
EV_HANDLER_INFO_INIT(IOIOCTL)
EV_HANDLER_INFO_INIT(WINDPRIORITYSETRAISE)
EV_HANDLER_INFO_INIT(WINDPRIORITYSETLOWER)
EV_HANDLER_INFO_INIT(SEMMCREATE)
EV_HANDLER_INFO_INIT(SEMCCREATE)
EV_HANDLER_INFO_INIT(SEMBCREATE)
EV_HANDLER_INFO_INIT(SEMDELETE)
EV_HANDLER_INFO_INIT(SEMGIVE)
EV_HANDLER_INFO_INIT(SEMTAKE)
EV_HANDLER_INFO_INIT(SEMFLUSH)
EV_HANDLER_INFO_INIT(MSGQSEND)
EV_HANDLER_INFO_INIT(MSGQRECEIVE)
EV_HANDLER_INFO_INIT(MSGQCREATE)
EV_HANDLER_INFO_INIT(MSGQDELETE)
EV_HANDLER_INFO_INIT(WDCREATE)
EV_HANDLER_INFO_INIT(WDDELETE)
EV_HANDLER_INFO_INIT(WDCANCEL)
EV_HANDLER_INFO_INIT(WDSTART)
EV_HANDLER_INFO_INIT(OBJ_SEMGIVE)
EV_HANDLER_INFO_INIT(OBJ_SEMTAKE)
EV_HANDLER_INFO_INIT(OBJ_MSGSEND)
EV_HANDLER_INFO_INIT(OBJ_MSGRECEIVE)
EV_HANDLER_INFO_INIT(OBJ_SEMRTAKE)
EV_HANDLER_INFO_INIT(OBJ_SEMWTAKE)
EV_HANDLER_INFO_INIT(WINDPENDQPUT)
EV_HANDLER_INFO_INIT(WINDPENDQGET)

#if RTP_INSTRUMENTATION
EV_HANDLER_INFO_INIT(RTPMEMPARTCREATE)
EV_HANDLER_INFO_INIT(RTPMEMALLOC)
EV_HANDLER_INFO_INIT(RTPMEMFREE)
EV_HANDLER_INFO_INIT(RTPMEMREALLOC)
#endif

#ifdef CAFE_VXWORKS_POWER_MANAGEMENT_EVENTS
EV_HANDLER_INFO_INIT(POWERDOWN)
EV_HANDLER_INFO_INIT(POWERUP)
EV_HANDLER_INFO_INIT(POWER_PSTATE_SET)
EV_HANDLER_INFO_INIT(POWER_CSTATE_SET)
EV_HANDLER_INFO_INIT(POWER_DSTATE_SET)
EV_HANDLER_INFO_INIT(POWER_DEV_ENABLE_SET)
EV_HANDLER_INFO_INIT(POWER_DEV_CAPABILITY_SET)
EV_HANDLER_INFO_INIT(POWER_DEV_TIMEOUT_SET)

EV_HANDLER_INFO_INIT(PPM_POLICY_SET_SYS)
EV_HANDLER_INFO_INIT(PPM_POLICY_SET_CPU)
EV_HANDLER_INFO_INIT(PPM_POLICY_SET_DEV)
EV_HANDLER_INFO_INIT(PPM_POLICY_DISABLE)
EV_HANDLER_INFO_INIT(PPM_POLICY_ENABLE)
#endif

EV_HANDLER_INFO_INIT(DPRINTF_LOG)
EV_HANDLER_INFO_INIT(EDR_ERROR_INJECT)
EV_HANDLER_INFO_INIT(EVENTRECEIVE)
EV_HANDLER_INFO_INIT(EVENTSEND)
EV_HANDLER_INFO_INIT(EXCEPTION)
EV_HANDLER_INFO_INIT(ISR_CREATE)
EV_HANDLER_INFO_INIT(ISR_DESTROY)
EV_HANDLER_INFO_INIT(ISR_DISPATCHER_DONE)
EV_HANDLER_INFO_INIT(ISR_DISPATCHER_INVOKE)
EV_HANDLER_INFO_INIT(KILL)
EV_HANDLER_INFO_INIT(OBJ_EVENTRECEIVE)
EV_HANDLER_INFO_INIT(OBJ_EVENTSEND)
EV_HANDLER_INFO_INIT(OBJ_MSGDELETE)
EV_HANDLER_INFO_INIT(OBJ_SEMFLUSH)
EV_HANDLER_INFO_INIT(OBJ_SIGKILL)
EV_HANDLER_INFO_INIT(OBJ_SIGPAUSE)
EV_HANDLER_INFO_INIT(OBJ_SIGSUSPEND)
EV_HANDLER_INFO_INIT(OBJ_SIGWAIT)
EV_HANDLER_INFO_INIT(OBJ_WINDROTATE)
EV_HANDLER_INFO_INIT(OWNERSET)
EV_HANDLER_INFO_INIT(PAUSE)
EV_HANDLER_INFO_INIT(RTPCHILDWAIT)
EV_HANDLER_INFO_INIT(RTPKILL)
EV_HANDLER_INFO_INIT(RTPPAUSE)
EV_HANDLER_INFO_INIT(RTPSIGNAL)
EV_HANDLER_INFO_INIT(RTPSIGPENDING)
EV_HANDLER_INFO_INIT(RTPSIGPROCMASK)
EV_HANDLER_INFO_INIT(RTPSIGQUEUE)
EV_HANDLER_INFO_INIT(RTPSIGTIMEDWAIT)
EV_HANDLER_INFO_INIT(RTPSIGWRAPPER)
EV_HANDLER_INFO_INIT(RTPSIGWRAPPER_DONE)
EV_HANDLER_INFO_INIT(RTPSUSPEND)
EV_HANDLER_INFO_INIT(RTPTASKKILL)
EV_HANDLER_INFO_INIT(RTPTASKSIGQUEUE)
EV_HANDLER_INFO_INIT(RTP_DELETE)
EV_HANDLER_INFO_INIT(RTP_INIT_TASK)
EV_HANDLER_INFO_INIT(RTP_LOAD)
EV_HANDLER_INFO_INIT(RTP_SPAWN)
EV_HANDLER_INFO_INIT(SD_CREATE)
EV_HANDLER_INFO_INIT(SD_DELETE)
EV_HANDLER_INFO_INIT(SD_MAP)
EV_HANDLER_INFO_INIT(SD_UNMAP)
EV_HANDLER_INFO_INIT(SEMMGIVEFORCE)
EV_HANDLER_INFO_INIT(SEMRTAKE)
EV_HANDLER_INFO_INIT(SEMWTAKE)
EV_HANDLER_INFO_INIT(SEMRWCREATE)
EV_HANDLER_INFO_INIT(SIGNAL)
EV_HANDLER_INFO_INIT(SIGSUSPEND)
EV_HANDLER_INFO_INIT(SIGWRAPPER)
EV_HANDLER_INFO_INIT(SYSCALL_BAD_ID)
EV_HANDLER_INFO_INIT(SYSCALL_RETURN_ERR)
EV_HANDLER_INFO_INIT(SYSCALL_RETURN_OK)
EV_HANDLER_INFO_INIT(TASKCONT)
EV_HANDLER_INFO_INIT(TASKDESTROY)
EV_HANDLER_INFO_INIT(TASKPRIORITYSET)
EV_HANDLER_INFO_INIT(TASKRESUME)
EV_HANDLER_INFO_INIT(TASKROTATE)
EV_HANDLER_INFO_INIT(TASKSAFE)
EV_HANDLER_INFO_INIT(TASKSTOP)
EV_HANDLER_INFO_INIT(TASKSUSPEND)
EV_HANDLER_INFO_INIT(TASKUNSAFE)
EV_HANDLER_INFO_INIT(TASK_CPU_AFFINITY_SET)
EV_HANDLER_INFO_INIT(TIMER_ROLLOVER)
EV_HANDLER_INFO_INIT(WINDCONT)
EV_HANDLER_INFO_INIT(WINDDELAY)
EV_HANDLER_INFO_INIT(WINDPENDQFLUSH)
EV_HANDLER_INFO_INIT(WINDPENDQTERMINATE)
EV_HANDLER_INFO_INIT(WINDRESUME)
EV_HANDLER_INFO_INIT(WINDSEMDELETE)
EV_HANDLER_INFO_INIT(WINDTICKANNOUNCETMRSLC)
EV_HANDLER_INFO_INIT(WINDTICKANNOUNCETMRWD)
EV_HANDLER_INFO_INIT(WINDSPAWN)
EV_HANDLER_INFO_INIT(WINDSUSPEND)
EV_HANDLER_INFO_INIT(WINDSTOP)
EV_HANDLER_INFO_INIT(WINDTICKTIMEOUT)
EV_HANDLER_INFO_INIT(WINDTICKUNDELAY)
EV_HANDLER_INFO_INIT(WINDWDCANCEL)
EV_HANDLER_INFO_INIT(WINDWDSTART)
EV_HANDLER_INFO_INIT(WINDUNDELAY)
EV_HANDLER_INFO_INIT(ISR_INFO)
EV_HANDLER_INFO_INIT(RTPNAME)
EV_HANDLER_INFO_INIT(CAFE_MODULE_LOAD)
EV_HANDLER_INFO_INIT(CAFE_MODULE_UNLOAD)
EV_HANDLER_INFO_INIT(CAFE_RTP_UNLOAD)


#define USEREV_HANDLER_INFO_INIT(EVENT_NAME)				\
static struct vxworks_event_handle ev_handler_info_##EVENT_NAME = {	\
	{0}, {0}, {0}, {0}, {0},					\
	0								\
};


USEREV_HANDLER_INFO_INIT(user_event_signed)
USEREV_HANDLER_INFO_INIT(user_event_unsigned)
USEREV_HANDLER_INFO_INIT(user_event_float)
USEREV_HANDLER_INFO_INIT(user_event_string)


#define USER_EVENT_HANDLE_INIT(TYPE)					\
	do {								\
	ev_handler_info_user_event_##TYPE.event_type.user.refcnt = 1;	\
	ev_handler_info_user_event_##TYPE.event_type.user.vt2str =	\
		user_event_##TYPE##_VT_REF;				\
	} while(0)


static struct cafe_list_head user_defined_events =
	CAFE_LIST_INIT(user_defined_events);


/* indices into orthogonal data items */

enum {
	ODATA_INDEX_TS = 0,
	ODATA_INDEX_CORE,
	ODATA_INDEX_SKIPCOUNT,
#if defined(__VXWORKS__)
	ODATA_INDEX_CONTEXT,
#endif
#if defined(__VXWORKS__) && defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
	ODATA_INDEX_STACK,
#endif
	NUM_ODATA_ITEMS
};

static int set0[] = {
	ODATA_INDEX_TS, ODATA_INDEX_CORE, ODATA_INDEX_SKIPCOUNT,
#if defined(__VXWORKS__)
	ODATA_INDEX_CONTEXT,
#endif
#if defined(__VXWORKS__) && defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
	ODATA_INDEX_STACK,
#endif
};

static struct cafe_ep_spec_output_od_set od_sets[] = {
	{CAFE_NELEMENTS(set0), set0}
};

static struct cafe_ep_spec_output_evt outputs[] = {
#if defined(__VXWORKS__) && ((_WRS_VXWORKS_MAJOR > 6) || \
	((_WRS_VXWORKS_MAJOR == 6) && (_WRS_VXWORKS_MINOR > 8)) || \
    ((_WRS_VXWORKS_MAJOR == 6) && (_WRS_VXWORKS_MINOR == 8) && (_WRS_VXWORKS_MAINT >= 2)))
	CAFE_OUTEV_INIT3(vxWorks_taskUnlockZero_VT_REF, &ev_handler_info_TASKUNLOCK_ZERO, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskUnlockNonZero_VT_REF, &ev_handler_info_TASKUNLOCK_NONZERO, 0),
#else
	CAFE_OUTEV_INIT3(vxWorks_taskUnlock_VT_REF, &ev_handler_info_TASKUNLOCK, 0),
#endif
	CAFE_OUTEV_INIT3(cafe_num_cores_VT_REF, &ev_handler_info_NUM_CPU, 0),
	CAFE_OUTEV_INIT3(cafe_timestamp_config_VT_REF, &ev_handler_info_TIMESTAMP_CONFIG, 0),
	CAFE_OUTEV_INIT3(vxWorks_dprintf_VT_REF, &ev_handler_info_DPRINTF_LOG, 0),
	CAFE_OUTEV_INIT3(vxWorks_edrErrorInject_VT_REF, &ev_handler_info_EDR_ERROR_INJECT, 0),
	CAFE_OUTEV_INIT3(vxWorks_eventReceive_VT_REF, &ev_handler_info_EVENTRECEIVE, 0),
	CAFE_OUTEV_INIT3(vxWorks_eventSend_VT_REF, &ev_handler_info_EVENTSEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_exception_VT_REF, &ev_handler_info_EXCEPTION, 0),
	CAFE_OUTEV_INIT3(vxWorks_intEnter_VT_REF, &ev_handler_info_INT_ENTER, 0),
#if defined(__VXWORKS__) && (_VX_CPU_FAMILY == _VX_PPC)
        CAFE_OUTEV_INIT3(vxWorks_intEnterPpcDec_VT_REF, &ev_handler_info_PPC_DECR_INT_ID, 0),
        CAFE_OUTEV_INIT3(vxWorks_intEnterPpcWd_VT_REF, &ev_handler_info_PPC_WD_INT_ID, 0),
#endif
	CAFE_OUTEV_INIT3(vxWorks_intExitK_VT_REF, &ev_handler_info_INT_EXIT_K, 0),
	CAFE_OUTEV_INIT3(vxWorks_intExit_VT_REF, &ev_handler_info_INT_EXIT, 0),
	CAFE_OUTEV_INIT3(vxWorks_ioClose_VT_REF, &ev_handler_info_IOCLOSE, 0),
	CAFE_OUTEV_INIT3(vxWorks_ioCreat_VT_REF, &ev_handler_info_IOCREAT, 0),
	CAFE_OUTEV_INIT3(vxWorks_ioDelete_VT_REF, &ev_handler_info_IODELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_ioIoctl_VT_REF, &ev_handler_info_IOIOCTL, 0),
	CAFE_OUTEV_INIT3(vxWorks_ioOpen_VT_REF, &ev_handler_info_IOOPEN, 0),
	CAFE_OUTEV_INIT3(vxWorks_ioRead_VT_REF, &ev_handler_info_IOREAD, 0),
	CAFE_OUTEV_INIT3(vxWorks_ioWrite_VT_REF, &ev_handler_info_IOWRITE, 0),
	CAFE_OUTEV_INIT3(vxWorks_isrCreate_VT_REF, &ev_handler_info_ISR_CREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_isrDestroy_VT_REF, &ev_handler_info_ISR_DESTROY, 0),
	CAFE_OUTEV_INIT3(vxWorks_isrDispatcherDone_VT_REF, &ev_handler_info_ISR_DISPATCHER_DONE, 0),
	CAFE_OUTEV_INIT3(vxWorks_isrDispatcherInvoke_VT_REF, &ev_handler_info_ISR_DISPATCHER_INVOKE, 0),
	CAFE_OUTEV_INIT3(vxWorks_isrInfo_VT_REF, &ev_handler_info_ISR_INFO, 0),
	CAFE_OUTEV_INIT3(vxWorks_kill_VT_REF, &ev_handler_info_KILL, 0),
	CAFE_OUTEV_INIT3(vxWorks_memAddToPool_VT_REF, &ev_handler_info_MEMADDTOPOOL, 0),
	CAFE_OUTEV_INIT3(vxWorks_memAlloc_VT_REF, &ev_handler_info_MEMALLOC, 0),
	CAFE_OUTEV_INIT3(vxWorks_memFree_VT_REF, &ev_handler_info_MEMFREE, 0),
	CAFE_OUTEV_INIT3(vxWorks_memPartCreate_VT_REF, &ev_handler_info_MEMPARTCREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_memRealloc_VT_REF, &ev_handler_info_MEMREALLOC, 0),
	CAFE_OUTEV_INIT3(vxWorks_msgQCreate_VT_REF, &ev_handler_info_MSGQCREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_msgQDelete_VT_REF, &ev_handler_info_MSGQDELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_msgQRecv_VT_REF, &ev_handler_info_MSGQRECEIVE, 0),
	CAFE_OUTEV_INIT3(vxWorks_msgQSend_VT_REF, &ev_handler_info_MSGQSEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_objEventReceive_VT_REF, &ev_handler_info_OBJ_EVENTRECEIVE, 0),
	CAFE_OUTEV_INIT3(vxWorks_objEventSend_VT_REF, &ev_handler_info_OBJ_EVENTSEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_objMsgDelete_VT_REF, &ev_handler_info_OBJ_MSGDELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_objMsgReceive_VT_REF, &ev_handler_info_OBJ_MSGRECEIVE, 0),
	CAFE_OUTEV_INIT3(vxWorks_objMsgSend_VT_REF, &ev_handler_info_OBJ_MSGSEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_objOwnerSet_VT_REF, &ev_handler_info_OWNERSET, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSemFlush_VT_REF, &ev_handler_info_OBJ_SEMFLUSH, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSemGive_VT_REF, &ev_handler_info_OBJ_SEMGIVE, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSemRTake_VT_REF, &ev_handler_info_OBJ_SEMRTAKE, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSemTake_VT_REF, &ev_handler_info_OBJ_SEMTAKE, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSemWTake_VT_REF, &ev_handler_info_OBJ_SEMWTAKE, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSigKill_VT_REF, &ev_handler_info_OBJ_SIGKILL, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSigPause_VT_REF, &ev_handler_info_OBJ_SIGPAUSE, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSigSuspend_VT_REF, &ev_handler_info_OBJ_SIGSUSPEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_objSigWait_VT_REF, &ev_handler_info_OBJ_SIGWAIT, 0),
	CAFE_OUTEV_INIT3(vxWorks_objWindRotate_VT_REF, &ev_handler_info_OBJ_WINDROTATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_pause_VT_REF, &ev_handler_info_PAUSE, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpChildWait_VT_REF, &ev_handler_info_RTPCHILDWAIT, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpDelete_VT_REF, &ev_handler_info_RTP_DELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpInitTask_VT_REF, &ev_handler_info_RTP_INIT_TASK, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpKill_VT_REF, &ev_handler_info_RTPKILL, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpLoad_VT_REF, &ev_handler_info_RTP_LOAD, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpName_VT_REF, &ev_handler_info_RTPNAME, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpPause_VT_REF, &ev_handler_info_RTPPAUSE, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSignal_VT_REF, &ev_handler_info_RTPSIGNAL, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSigPending_VT_REF, &ev_handler_info_RTPSIGPENDING, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSigProcMask_VT_REF, &ev_handler_info_RTPSIGPROCMASK, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSigQueue_VT_REF, &ev_handler_info_RTPSIGQUEUE, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSigTimedWait_VT_REF, &ev_handler_info_RTPSIGTIMEDWAIT, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSigWrapperDone_VT_REF, &ev_handler_info_RTPSIGWRAPPER_DONE, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSigWrapper_VT_REF, &ev_handler_info_RTPSIGWRAPPER, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSpawn_VT_REF, &ev_handler_info_RTP_SPAWN, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpSuspend_VT_REF, &ev_handler_info_RTPSUSPEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpTaskKill_VT_REF, &ev_handler_info_RTPTASKKILL, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpTaskSigQueue_VT_REF, &ev_handler_info_RTPTASKSIGQUEUE, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpUnload_VT_REF, &ev_handler_info_CAFE_RTP_UNLOAD, 0),
	CAFE_OUTEV_INIT3(vxWorks_sdCreate_VT_REF, &ev_handler_info_SD_CREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_sdDelete_VT_REF, &ev_handler_info_SD_DELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_sdMap_VT_REF, &ev_handler_info_SD_MAP, 0),
	CAFE_OUTEV_INIT3(vxWorks_sdUnmap_VT_REF, &ev_handler_info_SD_UNMAP, 0),
	CAFE_OUTEV_INIT3(vxWorks_semBCreate_VT_REF, &ev_handler_info_SEMBCREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semCCreate_VT_REF, &ev_handler_info_SEMCCREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semDelete_VT_REF, &ev_handler_info_SEMDELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semFlush_VT_REF, &ev_handler_info_SEMFLUSH, 0),
	CAFE_OUTEV_INIT3(vxWorks_semGive_VT_REF, &ev_handler_info_SEMGIVE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semMCreate_VT_REF, &ev_handler_info_SEMMCREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semMGiveForce_VT_REF, &ev_handler_info_SEMMGIVEFORCE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semRTake_VT_REF, &ev_handler_info_SEMRTAKE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semRWCreate_VT_REF, &ev_handler_info_SEMRWCREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semTake_VT_REF, &ev_handler_info_SEMTAKE, 0),
	CAFE_OUTEV_INIT3(vxWorks_semWTake_VT_REF, &ev_handler_info_SEMWTAKE, 0),
	CAFE_OUTEV_INIT3(vxWorks_signal_VT_REF, &ev_handler_info_SIGNAL, 0),
	CAFE_OUTEV_INIT3(vxWorks_sigSuspend_VT_REF, &ev_handler_info_SIGSUSPEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_sigWrapper_VT_REF, &ev_handler_info_SIGWRAPPER, 0),
	CAFE_OUTEV_INIT3(vxWorks_syscallBadId_VT_REF, &ev_handler_info_SYSCALL_BAD_ID, 0),
	CAFE_OUTEV_INIT3(vxWorks_syscallReturnError_VT_REF, &ev_handler_info_SYSCALL_RETURN_ERR, 0),
	CAFE_OUTEV_INIT3(vxWorks_syscallReturnOk_VT_REF, &ev_handler_info_SYSCALL_RETURN_OK, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskCont_VT_REF, &ev_handler_info_TASKCONT, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskCpuAffinitySet_VT_REF, &ev_handler_info_TASK_CPU_AFFINITY_SET, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskDelay_VT_REF, &ev_handler_info_TASKDELAY, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskDestroy_VT_REF, &ev_handler_info_TASKDESTROY, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskLock_VT_REF, &ev_handler_info_TASKLOCK, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskName_VT_REF, &ev_handler_info_TASKNAME, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskPrioritySet_VT_REF, &ev_handler_info_TASKPRIORITYSET, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskResume_VT_REF, &ev_handler_info_TASKRESUME, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskRotate_VT_REF, &ev_handler_info_TASKROTATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskSafe_VT_REF, &ev_handler_info_TASKSAFE, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskSpawn_VT_REF, &ev_handler_info_TASKSPAWN, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskStop_VT_REF, &ev_handler_info_TASKSTOP, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskSuspend_VT_REF, &ev_handler_info_TASKSUSPEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_taskUnsafe_VT_REF, &ev_handler_info_TASKUNSAFE, 0),
	CAFE_OUTEV_INIT3(vxWorks_tickAnnounce_VT_REF, &ev_handler_info_TICKANNOUNCE, 0),
	CAFE_OUTEV_INIT3(vxWorks_timerRollover_VT_REF, &ev_handler_info_TIMER_ROLLOVER, 0),
	CAFE_OUTEV_INIT3(vxWorks_wdCancel_VT_REF, &ev_handler_info_WDCANCEL, 0),
	CAFE_OUTEV_INIT3(vxWorks_wdCreate_VT_REF, &ev_handler_info_WDCREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_wdDelete_VT_REF, &ev_handler_info_WDDELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_wdStart_VT_REF, &ev_handler_info_WDSTART, 0),
	CAFE_OUTEV_INIT3(vxWorks_windCont_VT_REF, &ev_handler_info_WINDCONT, 0),
	CAFE_OUTEV_INIT3(vxWorks_windDelay_VT_REF, &ev_handler_info_WINDDELAY, 0),
	CAFE_OUTEV_INIT3(vxWorks_windDelete_VT_REF, &ev_handler_info_WINDDELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_windExitDispatchPI_VT_REF, &ev_handler_info_WIND_EXIT_DISPATCH_PI, 0),
	CAFE_OUTEV_INIT3(vxWorks_windExitDispatch_VT_REF, &ev_handler_info_WIND_EXIT_DISPATCH, 0),
	CAFE_OUTEV_INIT3(vxWorks_windExitIdle_VT_REF, &ev_handler_info_WIND_EXIT_IDLE, 0),
	CAFE_OUTEV_INIT3(vxWorks_windPendQFlush_VT_REF, &ev_handler_info_WINDPENDQFLUSH, 0),
	CAFE_OUTEV_INIT3(vxWorks_windPendQGet_VT_REF, &ev_handler_info_WINDPENDQGET, 0),
	CAFE_OUTEV_INIT3(vxWorks_windPendQPut_VT_REF, &ev_handler_info_WINDPENDQPUT, 0),
	CAFE_OUTEV_INIT3(vxWorks_windPendQTerminate_VT_REF, &ev_handler_info_WINDPENDQTERMINATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_windPrioritySetLower_VT_REF, &ev_handler_info_WINDPRIORITYSETLOWER, 0),
	CAFE_OUTEV_INIT3(vxWorks_windPrioritySetRaise_VT_REF, &ev_handler_info_WINDPRIORITYSETRAISE, 0),
	CAFE_OUTEV_INIT3(vxWorks_windResume_VT_REF, &ev_handler_info_WINDRESUME, 0),
	CAFE_OUTEV_INIT3(vxWorks_windSemDelete_VT_REF, &ev_handler_info_WINDSEMDELETE, 0),
	CAFE_OUTEV_INIT3(vxWorks_windSpawn_VT_REF, &ev_handler_info_WINDSPAWN, 0),
	CAFE_OUTEV_INIT3(vxWorks_windStop_VT_REF, &ev_handler_info_WINDSTOP, 0),
	CAFE_OUTEV_INIT3(vxWorks_windSuspend_VT_REF, &ev_handler_info_WINDSUSPEND, 0),
	CAFE_OUTEV_INIT3(vxWorks_windTickAnnounceTmrSlc_VT_REF, &ev_handler_info_WINDTICKANNOUNCETMRSLC, 0),
	CAFE_OUTEV_INIT3(vxWorks_windTickAnnounceTmrWd_VT_REF, &ev_handler_info_WINDTICKANNOUNCETMRWD, 0),
	CAFE_OUTEV_INIT3(vxWorks_windTickTimeout_VT_REF, &ev_handler_info_WINDTICKTIMEOUT, 0),
	CAFE_OUTEV_INIT3(vxWorks_windTickUndelay_VT_REF, &ev_handler_info_WINDTICKUNDELAY, 0),
	CAFE_OUTEV_INIT3(vxWorks_windUndelay_VT_REF, &ev_handler_info_WINDUNDELAY, 0),
	CAFE_OUTEV_INIT3(vxWorks_windWdCancel_VT_REF, &ev_handler_info_WINDWDCANCEL, 0),
	CAFE_OUTEV_INIT3(vxWorks_windWdStart_VT_REF, &ev_handler_info_WINDWDSTART, 0),

#if RTP_INSTRUMENTATION
	CAFE_OUTEV_INIT3(vxWorks_rtpMemPartCreate_VT_REF, &ev_handler_info_RTPMEMPARTCREATE, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpMemAlloc_VT_REF, &ev_handler_info_RTPMEMALLOC, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpMemFree_VT_REF, &ev_handler_info_RTPMEMFREE, 0),
	CAFE_OUTEV_INIT3(vxWorks_rtpMemRealloc_VT_REF, &ev_handler_info_RTPMEMREALLOC, 0),
#endif

#ifdef CAFE_VXWORKS_POWER_MANAGEMENT_EVENTS
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___POWERDOWN_VT_REF, &ev_handler_info_POWERDOWN, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___POWERUP_VT_REF, &ev_handler_info_POWERUP, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___POWERCSTATESET_VT_REF, &ev_handler_info_POWER_CSTATE_SET, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___POWERDEVENABLESET_VT_REF, &ev_handler_info_POWER_DEV_ENABLE_SET, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___POWERDEVTIMEOUTSET_VT_REF, &ev_handler_info_POWER_DEV_TIMEOUT_SET, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___POWERPSTATESET_VT_REF, &ev_handler_info_POWER_PSTATESET, 0),
	/* CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___PPM_PEV_VT_REF, &ev_handler_info_PPM_PEV, 0), */
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___PPM_POLICY_DISABLE_VT_REF, &ev_handler_info_PPM_POLICY_DISABLE, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___PPM_POLICY_ENABLE_VT_REF, &ev_handler_info_PPM_POLICY_ENABLE, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___PPM_POLICY_SET_CPU_VT_REF, &ev_handler_info_PPM_POLICY_SET_CPU, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___PPM_POLICY_SET_DEV_VT_REF, &ev_handler_info_PPM_POLICY_SET_DEV, 0),
	CAFE_OUTEV_INIT3(VXWORKS67___VXWORKS___PWRPOLICYMGRSYSSETPROCESS_VT_REF, &ev_handler_info_PPM_POLICY_SET_SYS, 0),
#endif
	CAFE_OUTEV_INIT3(vxWorks_moduleLoad_VT_REF, &ev_handler_info_CAFE_MODULE_LOAD, 0),
	CAFE_OUTEV_INIT3(vxWorks_moduleUnload_VT_REF, &ev_handler_info_CAFE_MODULE_UNLOAD, 0),
	CAFE_OUTEV_INIT2(user_event_signed_VT_REF,
			 &ev_handler_info_user_event_signed),
	CAFE_OUTEV_INIT2(user_event_unsigned_VT_REF,
			 &ev_handler_info_user_event_unsigned),
	CAFE_OUTEV_INIT2(user_event_float_VT_REF,
			 &ev_handler_info_user_event_float),
	CAFE_OUTEV_INIT2(user_event_string_VT_REF,
			 &ev_handler_info_user_event_string)
	//CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};

static const char *output_odata[] = {
	cafe_timestamp_VT_REF,
	cafe_core_id_VT_REF,
	cafe_skipcount_VT_REF,
#if defined(__VXWORKS__)
	vxWorks_context_VT_REF,
#endif
#if defined(__VXWORKS__) && defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
	cafe_stacktrace_VT_REF
#endif
};


#define EVENT_LIST_HANDLE_ENTRY(PTR)					\
	CAFE_LIST_ENTRY((PTR), struct cafe_user_event_handle, list)

#define user_event_to_vx_handle(h) \
	((struct vxworks_event_handle *)(((char *)(h)) - offsetof(struct vxworks_event_handle, event_type.user)))


static int update_handle_output_event_actions(struct vxworks_event_handle *h)
{
	int rc;
	unsigned int i;

	for (i = 0; i < CAFE_NELEMENTS(instance_list); ++i) {
		struct cafe_ep_inst *inst = instance_list[i];
		struct cafe_event_descr *event;
		struct cafe_event_action_list *sink_actions;
		cafe_eal_handle_t sink_actions_handle;
		struct vxworks_inst_data *node;

		if (inst == NULL)
			continue;

		node = (struct vxworks_inst_data *) inst->client_data;
		if (!node->started) {
			cafe_event_action_list_swap(&h->sink_actions[i],
						    CAFE_EAL_HANDLE_NULL);
			continue;
		}

		rc = cafe_type_list_parse(inst->session,
					  h->event_type.user.vt2str,
					  &event);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_event_action_list_alloc(
			inst, event, &sink_actions, &sink_actions_handle);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
		if (rc != CAFE_OK)
			return rc;

		cafe_event_action_list_swap(&h->sink_actions[i],
					    sink_actions_handle);
	}
	return CAFE_OK;
}

int cafe_user_event_add(const struct cafe_allocator *allocator,
			const char *vt2str,
			struct cafe_user_event_handle **handle)
{
	struct vxworks_event_handle *vx_h;
	struct cafe_list_head *ptr;
	unsigned int i;
	int rc;

	assert(allocator != NULL);
	assert(vt2str != NULL);
	assert(handle != NULL);
	CAFE_ASSERT_LOCKED();
	cafe_list_foreach(ptr, &user_defined_events) {
		struct cafe_user_event_handle *h;
		h = EVENT_LIST_HANDLE_ENTRY(ptr);
		if (strcmp(h->vt2str, vt2str) == 0) {
			h->refcnt++;
			*handle = h;
			return CAFE_OK;
		}
	}
	vx_h = CAFE_ALLOC(allocator, sizeof(*vx_h) + strlen(vt2str) + 1);
	if (vx_h == NULL)
		return CAFE_NOMEM;
	memset(vx_h, 0, sizeof(*vx_h));
	for (i = 0; i < CAFE_NELEMENTS(instance_list); ++i) {
		if (instance_list[i] == NULL)
			continue;
		vx_h->change_count[i] = instance_list[i]->session->output_event_action_change_count;
	}

	rc = update_handle_output_event_actions(vx_h);
	if (rc != CAFE_OK) {
		CAFE_FREE(allocator, vx_h);
		return rc;
	}

	vx_h->event_type.user.refcnt = 1;
	vx_h->event_type.user.vt2str = (char *)(vx_h + 1);
	strncpy((char *)vx_h->event_type.user.vt2str, vt2str, strlen(vt2str)+1);
	vx_h->event_type.user.allocator = allocator;
	cafe_list_add_last(&vx_h->event_type.user.list, &user_defined_events);

	*handle = &vx_h->event_type.user;
	return CAFE_OK;
}

void cafe_user_event_release(struct cafe_user_event_handle *h)
{
	struct vxworks_event_handle *vx_h = user_event_to_vx_handle(h);
	unsigned int i;

	assert(h != NULL);
	assert(h->refcnt > 0);
	CAFE_ASSERT_LOCKED();
	if (--h->refcnt > 0)
		return;
	for (i = 0; i < CAFE_NELEMENTS(vx_h->sink_actions); ++i)
		cafe_event_action_list_swap(&vx_h->sink_actions[i],
					    CAFE_EAL_HANDLE_NULL);

	cafe_list_remove(&h->list);
	CAFE_FREE(h->allocator, vx_h);
}

void cafe_user_event_generate(struct cafe_user_event_handle *h, void *data)
{
	struct vxworks_event_handle *vx_h = user_event_to_vx_handle(h);
	assert(h != NULL);
	cafe_cfg_vxworks_ev_process_handle(vx_h, data, NULL);
}

void cafe_user_event_signed(const char *label, long long value)
{
	struct user_event_signed data;

	data.label = (char *) label;
	data.value = value;
	cafe_user_event_generate(&ev_handler_info_user_event_signed.event_type.user, &data);
}

void cafe_user_event_unsigned(const char *label, unsigned long long value)
{
	struct user_event_unsigned data;

	data.label = (char *) label;
	data.value = value;
	cafe_user_event_generate(&ev_handler_info_user_event_unsigned.event_type.user, &data);
}

void cafe_user_event_float(const char *label, double value)
{
	struct user_event_float data;

	data.label = (char *) label;
	data.value = value;
	cafe_user_event_generate(&ev_handler_info_user_event_float.event_type.user, &data);
}

void cafe_user_event_string(const char *label, const char *value)
{
	struct user_event_string data;

	data.label = (char *) label;
	data.value = (char *) value;
	cafe_user_event_generate(&ev_handler_info_user_event_string.event_type.user, &data);
}


/* Macro only meant to be used from the ev_process_handle() routine.  Evaluates
   any filter orthogonal data entries, advancing the odata index past all
   consumed entries.  When a filter evaluates as false, move on to the next
   registered event. */
#define CHECK_FILTER()							\
	do {								\
		if (++j == num_types)					\
			goto done_odata;				\
									\
		while (event->types[j]->kind == CAFE_TYPE_FILTER) {	\
			bool result;					\
									\
			rc = check_filter(event, output_data, event->types[j]->ident->expr, &result); \
			if (rc != CAFE_OK) {				\
				cafe_dispatch_error(linst, rc);		\
				goto unlock;				\
			}						\
			if (!result)					\
				goto unlock;				\
			if (++j == event->num_types)			\
				goto done_odata;			\
		}							\
	} while(0)

static int check_filter(struct cafe_event_descr *event, void **event_data, struct cafe_vm_expr *expr, bool *expr_result)
{
	cafe_any_value result;
	struct cafe_expr_eval_context context;
	int rc;

	cafe_expr_eval_context_create(&context, NULL);
	context.event =	event;
	context.event_data = event_data;
	rc = cafe_expr_eval_value(&context, expr, &result);
	if (rc == CAFE_OK)
		*expr_result = result.ll != 0;
	return rc;
}
/*******************************************************************************
*
* cafe_cfg_vxworks_ev_process_handle - send event to instances
*
* If inst is non-null, then the event will be dispatched only to that instance:
* Otherwise, it will go to all instances listed in the action_list array for
* that event. This is so that at instance start, taskname etc. can be sent only
* to the instance that requires them. If dispatching to a particular instance,
* the state of that instance (running or stopped) is ignored
*
* RETURNS: N/A
*
* \NOMANUAL
*/

void cafe_cfg_vxworks_ev_process_handle(struct vxworks_event_handle *h, void *args,
					struct cafe_ep_inst *inst)
{
	int rc;
#if defined(__VXWORKS__)
	struct vxWorks_context ctx = {0};
	extern void *cafe_vxworks_allocator_partition_id(const struct cafe_allocator *);
#endif
#if defined(__VXWORKS__) && defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
#  ifdef CAFE_ALLOCA
	void **stack = NULL;
#  else
	void *stack[MAX_STACK_DEPTH];
#  endif
	struct cafe_vm_dynarray stack_arg;
#endif
	unsigned int i, j;
	void *output_data[NUM_ODATA_ITEMS + 1];
	output_data[0] = args;

	for (i = 0; i < CAFE_NELEMENTS(h->sink_actions); i++) {
		struct cafe_event_action_list *action_list = NULL;
		cafe_eal_handle_t unlock_handle;
		struct cafe_skipcount skipcount_odata;
		cafe_atomic_val curr_skipcount_limit;
		struct cafe_event_descr *event;
		int num_types;
		struct cafe_ep_inst *linst;

		linst = instance_list[i];
		if (linst == NULL)
			continue;

		if (inst != NULL && linst != inst)
			continue;
		rc = cafe_event_action_list_read_lock(&h->sink_actions[i],
						      &unlock_handle, &action_list);

		if (rc != CAFE_OK) {
			cafe_dispatch_error(linst, rc);
			continue;
		}

		if (action_list == NULL)
			goto unlock;

		if (linst->state != CAFE_EP_INST_STATE_STARTED && inst == NULL)
			goto unlock;

#if defined(__VXWORKS__)
		/* WRS_TODO: Remove once we stop using memPartAlloc so much. */
		if ((h == &ev_handler_info_MEMALLOC ||
		     h == &ev_handler_info_MEMREALLOC ||
		     h == &ev_handler_info_MEMFREE) &&
		    cafe_vx_ignore_allocations &&
		    *(void **)args == cafe_vxworks_allocator_partition_id(linst->allocator))
			goto unlock;
#endif

		curr_skipcount_limit = cafe_atomic_get(&h->skipcount_limit[i]);

		if (curr_skipcount_limit > 0) {
			cafe_atomic_val n;

			n = cafe_atomic_val_inc(&h->skipcount_value[i]) + 1;

			if (n < curr_skipcount_limit)
				goto unlock;

			if (!cafe_atomic_cas(&h->skipcount_value[i], n, 0))
				goto unlock;

			assert(n > 0);
			skipcount_odata.value = n;
		} else {
			skipcount_odata.value = 1;
		}

		event = action_list->event;
		num_types = event->num_types;
		j = 0;
		CHECK_FILTER();

		if (event->types[j] == linst->output_odata[ODATA_INDEX_TS]) {
			unsigned long long ts;
			ts = cafe_cfg_vxworks_orth_data_ts();
			output_data[j] = &ts;

			CHECK_FILTER();
		}

		if (event->types[j] == linst->output_odata[ODATA_INDEX_CORE]) {
			int coreid;
			coreid = cafe_cfg_vxworks_orth_data_coreid();
			output_data[j] = &coreid;

			CHECK_FILTER();
		}

		if (cafe_base_type(event->types[j]) == linst->output_odata[ODATA_INDEX_SKIPCOUNT]) {
			output_data[j] = &skipcount_odata;

			CHECK_FILTER();
		}

#if defined(__VXWORKS__)
		if (event->types[j] == linst->output_odata[ODATA_INDEX_CONTEXT]) {

			ctx.tid = (unsigned long)taskIdCurrent;
			ctx.pid = (unsigned long)RTP_ID_OF_TASK(ctx.tid);
			cafe_cfg_vxworks_cur_int_num_get (&ctx.intCount, &ctx.intNum);
			output_data[j] = &ctx;

			CHECK_FILTER();
		}
#endif

#if defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
		if (event->types[j] == linst->output_odata[ODATA_INDEX_STACK]) {
			struct workarea *w = (struct workarea *)
				instance_list[i]->workarea->u.data;
			REG_SET regs;
#  ifdef CAFE_ALLOCA
			if (stack == NULL)
				stack = (void **) CAFE_ALLOCA(w->max_stack_depth * sizeof(void *));
#  endif
			output_data[j] = &stack_arg;
			stack_arg.p = stack;

			ST_current_regs(&regs);
			stack_arg.l = cafe_vx_stacktrace(&regs, taskIdSelf(),
							 stack,
							 w->max_stack_depth,
							 FRAMES_TO_SKIP);
			CHECK_FILTER();
		}
#endif
	done_odata:
		cafe_dispatch_event_unlocked(action_list, output_data);
	unlock:
		cafe_event_action_list_read_unlock(unlock_handle);
	}
}

static void disable_events_for_inst(struct cafe_ep_inst *inst)
{
	unsigned int i;
	struct vxworks_inst_data *inst_data =
		(struct vxworks_inst_data *)inst->client_data;

	for (i = 0; i < inst->spec->n_output_events; i++) {
		unsigned int j;
		bool event_in_use = false;
		struct vxworks_event_handle *event_handle =
			(struct vxworks_event_handle *)outputs[i].impl_data;

		if (event_handle->change_count[inst_data->index] == inst_data->change_count)
			continue;

		cafe_event_action_list_swap(&event_handle->sink_actions[inst_data->index],
					    CAFE_EAL_HANDLE_NULL);
		for (j = 0; j < CAFE_NELEMENTS(instance_list); j++) {
			if (event_handle->sink_actions[j] != CAFE_EAL_HANDLE_NULL) {
				event_in_use = true;
				break;
			}
		}

		if (!event_in_use)
			cafe_ep_vxworks_kernel_disable_event(event_handle);
		event_handle->statedump_request[inst_data->index] = STATEDUMP_NONE;
	}
}

static int update_output_event_actions_for_inst(struct cafe_ep_inst *inst,
						bool from_start)
{
	struct cafe_list_head *ptr;
	unsigned int i, j;
	struct vxworks_inst_data *inst_data;

	inst_data = (struct vxworks_inst_data *)inst->client_data;

	for (i = 0; i < inst->enabled_output_events.ind; i++) {
		int rc;
		struct cafe_event_descr *event_descr = NULL;
		struct vxworks_event_handle *event_handle = NULL;
		struct cafe_event_action_info *event_action_info;
		struct cafe_event_action_list *action_list = NULL;
		cafe_eal_handle_t sink_actions_handle = CAFE_EAL_HANDLE_NULL;
		struct cafe_vm_expr *expr;

		event_descr = inst->enabled_output_events.list[i].event;
		rc = cafe_ep_inst_event_action_info_get(inst, event_descr,
							&event_action_info);
		if (rc != CAFE_OK || event_action_info == NULL)
			return rc;

		event_handle = (struct vxworks_event_handle *)
			inst->enabled_output_events.list[i].impl_data;
		if (event_handle == NULL)
			return CAFE_ERROR;

#if defined(ALLOW_MULTIPLE_STATE_EVENTS)
		if (event_handle->state_dump != NULL)
#else
		if ((event_handle->state_dump != NULL) &&
		    (event_handle->statedump_request[inst_data->index] == STATEDUMP_NONE))
#endif
			event_handle->statedump_request[inst_data->index] = STATEDUMP_REQUEST;


		/* Configure any requested skipcount. */
		expr = NULL;

		for (j = 0; j < event_action_info->event->num_types; j++) {
			const struct cafe_value_type *type = event_action_info->event->types[j];
			if (cafe_type_based_on(type, event_action_info->inst->output_odata[ODATA_INDEX_SKIPCOUNT])) {
				expr = cafe_param_value_get(type, "limit");
				break;
			}
		}

		if (expr != NULL) {
			if (expr->u.ull > UINT_MAX || expr->u.ull == 0)
				return CAFE_ERROR;
			if (cafe_atomic_get(&event_handle->skipcount_limit[inst_data->index]) == 0)
				(void)cafe_atomic_set(&event_handle->skipcount_value[inst_data->index], 0);
			(void)cafe_atomic_set(&event_handle->skipcount_limit[inst_data->index], 
			                (cafe_atomic_val)expr->u.ull);
		} else {
			(void)cafe_atomic_set(&event_handle->skipcount_limit[inst_data->index], 0);
		}

		if (!inst_data->started) {
			cafe_event_action_list_swap(&event_handle->sink_actions[inst_data->index],
						    sink_actions_handle);
			continue;
		}

		rc = cafe_event_action_list_alloc(inst, event_descr,
						  &action_list,
						  &sink_actions_handle);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_ep_inst_get_sink_event_actions(inst, action_list);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sink_actions_handle);
			return rc;
		}

		(void)cafe_ep_vxworks_kernel_enable_event(event_handle);

		cafe_event_action_list_swap(&event_handle->sink_actions[inst_data->index], sink_actions_handle);
		event_handle->change_count[inst_data->index] = inst_data->change_count;
	}
	disable_events_for_inst(inst);

	cafe_list_foreach(ptr, &user_defined_events) {
		struct vxworks_event_handle *vx_h =
			user_event_to_vx_handle(EVENT_LIST_HANDLE_ENTRY(ptr));

		(void)update_handle_output_event_actions(vx_h);
	}

	return dump_state_for_inst(inst, from_start);
}

static int vxworks_inst_create(struct cafe_ep_inst *inst)
{
	unsigned int i;
	struct vxworks_inst_data *inst_data;

#if defined(__VXWORKS__) && defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
	struct workarea *w =
		(struct workarea *)inst->workarea->u.data;

	if ((w->max_stack_depth > MAX_STACK_DEPTH) || (w->max_stack_depth == 0))
		w->max_stack_depth = MAX_STACK_DEPTH;
#endif
	for (i = 0; i < CAFE_NELEMENTS(instance_list); i++)
		if (instance_list[i] == NULL)
			break;

	if (i == CAFE_NELEMENTS(instance_list))
		return CAFE_NOMEM;

	if ((inst_data = CAFE_ALLOC(inst->allocator, sizeof *inst_data)) == NULL)
		return CAFE_NOMEM;

	inst_data->change_count = inst->session->output_event_action_change_count - 1;
	inst_data->started = false;
	inst_data->index = i;
	instance_list[i] = inst;
	inst->client_data = inst_data;

	return CAFE_OK;
}

static void vxworks_inst_delete(struct cafe_ep_inst *inst)
{
	struct vxworks_inst_data *inst_data =
		(struct vxworks_inst_data *)inst->client_data;

	instance_list[inst_data->index] = NULL;
	CAFE_FREE(inst->allocator, inst_data);
}

static int vxworks_inst_start(struct cafe_ep_inst *inst)
{
	struct vxworks_inst_data *inst_data =
		(struct vxworks_inst_data *)inst->client_data;

	inst_data->started = true;
	(void)update_output_event_actions_for_inst(inst, true);
	return cafe_cfg_vxworks_inst_start(inst);
}

static int vxworks_inst_stop(struct cafe_ep_inst *inst)
{
	struct vxworks_inst_data *inst_data =
		(struct vxworks_inst_data *)inst->client_data;

	inst_data->started = false;
	(void)update_output_event_actions_for_inst(inst, false);
	return cafe_cfg_vxworks_inst_stop(inst);
}

static int update_output_event_actions(struct cafe_ep_inst *inst)
{
	struct vxworks_inst_data *inst_data =
		(struct vxworks_inst_data *)inst->client_data;

	if (inst_data->change_count == inst->session->output_event_action_change_count)
		return CAFE_OK;

	inst_data->change_count = inst->session->output_event_action_change_count;
	return update_output_event_actions_for_inst(inst, false);
}

static struct cafe_ep_spec spec = {
	"vxworks_kernel",
	vxworks_inst_create,
	vxworks_inst_delete,
	vxworks_inst_start,
	vxworks_inst_stop,
	NULL,			/* get_input_event_actions */
	update_output_event_actions,
	NULL,			/* inject event */
	EVENT_TYPEDEFS,
	"t[vxWorks.workarea]",
	0,			/* number of input events */
	NULL,			/* input_events */
	CAFE_NELEMENTS(outputs),
	outputs,
	CAFE_NELEMENTS(output_odata),
	output_odata,
	NULL,					/* settings_commit() */
	NULL,					/* input_state_change */
	CAFE_NELEMENTS(od_sets),		/* n_output_odata_sets */
	od_sets					/* output_odata_sets */
};


/*******************************************************************************
*
* cafe_cfg_vxworks_kernel_init_cafe - initialize vxWorks Cafe configuration
*
* This function prepares the Cafe vxWorks configuration. It performs
* registration, and connects to the vxWorks event source
*
* RETURNS: CAFE_OK, or a Cafe error value
*
*/

int cafe_cfg_vxworks_kernel_init_cafe(void)
{
	int rc;
	unsigned int i;

	if (vxWorksInitialized)
		return CAFE_OK;

	for (i = 0; i < CAFE_NELEMENTS(outputs); i++) {
		int j;
		struct vxworks_event_handle *h = outputs[i].impl_data;

		memset(h->skipcount_value, 0, sizeof h->skipcount_value);
		memset(h->skipcount_limit, 0, sizeof h->skipcount_limit);

		for (j = 0; j < CAFE_NELEMENTS(h->sink_actions); ++j)
			h->sink_actions[j] = CAFE_EAL_HANDLE_NULL;

		if (h->event_id == 0)
			/* user events don't need connecting */
			continue;

		rc = cafe_cfg_vxworks_kernel_connect_event(outputs + i);
		if (rc != CAFE_OK) {
			(void)fprintf(stderr, "cafe_cfg_vxworks_kernel_connect_event() failed for %s: %s\n",
				outputs[i].type, cafe_strerror(rc));
			return rc;
		}
	}

	if (ev_handler_info_user_event_signed.event_type.user.list.next == NULL) {
		USER_EVENT_HANDLE_INIT(signed);
		USER_EVENT_HANDLE_INIT(unsigned);
		USER_EVENT_HANDLE_INIT(float);
		USER_EVENT_HANDLE_INIT(string);
	}

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		vxWorksInitialized = true;

	return rc;
}

void cafe_cfg_vxworks_kernel_delete(void)
{
	if (!vxWorksInitialized)
		return;

	vxWorksInitialized = false;
	(void)cafe_ep_spec_unregister(&spec);

	(void)cafe_cfg_vxworks_kernel_delete_internal();
}

static int dump_state_for_inst(struct cafe_ep_inst *inst, bool from_start)
{
	int rc = CAFE_OK;
	unsigned int i;
	int index;
	struct cafe_enabled_event *evt;
	struct vxworks_event_handle *event_handle = NULL;
	struct vxworks_inst_data *inst_data = inst->client_data;

	/* if not running, don't emit anything */
	if ((inst->state != CAFE_EP_INST_STATE_STARTED) &&
	    (inst->state != CAFE_EP_INST_STATE_STARTING))
		return CAFE_OK;

	index = inst_data->index;
	for (i = 0; i < inst->enabled_output_events.ind; i++) {
		evt = (struct cafe_enabled_event *)&inst->enabled_output_events.list[i];
		event_handle = (struct vxworks_event_handle *)evt->impl_data;

		if (event_handle == NULL || event_handle->state_dump == NULL)
			continue;

		if ((event_handle->statedump_request[index] == STATEDUMP_REQUEST) || from_start) {
			rc = event_handle->state_dump(event_handle, inst);
			if (rc != CAFE_OK)
				return rc;
			event_handle->statedump_request[index] = STATEDUMP_DONE;
		}
	}
	return rc;
}
