/* vxworks_kernel_host.c - host-only part of vxWorks Cafe event processor */

/* Copyright (c) 2010-2013 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
DESCRIPTION

This configuration routes events from the vxWorks OS into the Cafe
analysis system. In the Cafe system, it appears as a configuration which can be
instantiated multiple times. It has no input events, no settings, a minimal
private workarea, and output events.

*/

/* includes */

#include <cafe/cafe.h>
#include <cafe/ep/vxworks_kernel.h>
#include <cafe/autogenerated/vxworks.type.h>
#include "vxworks_events.h"
#include "vxworks_kernel_host.h"
#include <cafe/hashtbl.h>

struct vxworks_per_event_info {
	int event_id;
	int (*state_dump)(struct vxworks_event_handle *event_handle,
			  struct cafe_ep_inst *inst);
};


static struct cafe_hashtbl *event_hashtbl = NULL;


void cafe_cfg_vxworks_ev_process_id(int id, void *args, struct cafe_ep_inst *inst)
{
	struct cafe_ep_spec_output_evt *oe;
	struct vxworks_event_handle *h;
	oe = cafe_hashtbl_lookup(event_hashtbl, &id);
	h = oe->impl_data;
	cafe_cfg_vxworks_ev_process_handle(h, args, inst);
}


static int post_timestamp_config(struct vxworks_event_handle *event_handle,
				 struct cafe_ep_inst *inst)
{
	struct {
		unsigned long freq;
	} ts_config;

	ts_config.freq = 1234567;

 	cafe_cfg_vxworks_ev_process_id(EVENT_TIMESTAMP_CONFIG, &ts_config, inst);

	return CAFE_OK;
}


static int post_initial_task(struct vxworks_event_handle *event_handle,
			     struct cafe_ep_inst *inst)
{
	struct vxWorks_taskName tn;

	tn.state = 1;
	tn.priority = 2;
	tn.lockCnt = 3;
	tn.tid = 12345;
	tn.entryPt = 0x2000;
	tn.rtp = 0x3000;
	tn.affinity = 4;
	tn.name = (char *)"start-up-task";

 	cafe_cfg_vxworks_ev_process_id(EVENT_TASKNAME, &tn, inst);
	return CAFE_OK;
}


static int post_core_count(struct vxworks_event_handle *event_handle,
			   struct cafe_ep_inst *inst)
{
	struct cafe_num_cores nc;

	nc.num_cores = 2;
 	cafe_cfg_vxworks_ev_process_id(EVENT_NUM_CPU, &nc, inst);
	return CAFE_OK;
}

static int post_kernel_id (struct vxworks_event_handle *event_handle,
			   struct cafe_ep_inst *inst)
{
	struct vxWorks_rtpLoad kid;

	kid.dataSize = 1;
	kid.dataAddr = 2;
	kid.textSize = 3;
	kid.textAddr = 4;
	kid.entryPoint = 5;
	kid.rtpId = 6;
	kid.name = (char *)"vxWorks";

	cafe_cfg_vxworks_ev_process_id(EVENT_RTP_LOAD, &kid, inst);
	return CAFE_OK;
}

struct vxworks_per_event_info outputs [] = {
	{EVENT_TIMESTAMP_CONFIG, post_timestamp_config},
	{EVENT_TASKNAME, post_initial_task},
	{EVENT_NUM_CPU, post_core_count},
	{EVENT_RTPNAME, post_kernel_id}
};

/*******************************************************************************
*
* inst_start - callback for enabling the instance
*
* This function consults the output event list on the instance, and adds each
* one to the vxWorks event mapping
*
* RETURNS: CAFE_OK or a Cafe error code
*
* \NOMANUAL
*/

int cafe_cfg_vxworks_inst_start(struct cafe_ep_inst *inst)
{
	cafe_ep_inst_state_started(inst);
	return CAFE_OK;
}

/*******************************************************************************
*
* inst_stop - callback to disable event generation for this instance
*
* RETURNS: CAFE_OK or a Cafe error code
*
* \NOMANUAL
*/

int cafe_cfg_vxworks_inst_stop(struct cafe_ep_inst *inst)
{
	cafe_ep_inst_state_stopped(inst);
	return CAFE_OK;
}

/*******************************************************************************
*
* cafe_cfg_vxworks_orth_data_ts -
*
* RETURNS: N/A
*
* \NOMANUAL
*/

unsigned long long cafe_cfg_vxworks_orth_data_ts(void)
{
	return 987654321;
}

/*******************************************************************************
*
* cafe_cfg_vxworks_orth_data_coreid -
*
* RETURNS: N/A
*
* \NOMANUAL
*/

int cafe_cfg_vxworks_orth_data_coreid(void)
{
	return 1;
}


int cafe_cfg_vxworks_kernel_connect_event(struct cafe_ep_spec_output_evt *evt)
{
	int i;
	struct vxworks_event_handle *h = evt->impl_data;

	if (event_hashtbl == NULL) {
		event_hashtbl = cafe_hashtbl_new(64,
						 cafe_stdlib_allocator,
						 cafe_hashtbl_int_hash,
						 cafe_hashtbl_int_equals);
		if (event_hashtbl == NULL)
			return CAFE_NOMEM;
	}

	if (cafe_hashtbl_insert(event_hashtbl, &h->event_id, evt) != 0) {
		CAFE_FREE(cafe_stdlib_allocator, h);
				return CAFE_NOMEM;
	}

	/* connect state dump functions */
	for (i = 0; i < CAFE_NELEMENTS(outputs); i++) {
		if (outputs[i].event_id == h->event_id) {
			h->state_dump = outputs[i].state_dump;
			return CAFE_OK;
		}
	}
	return CAFE_OK;
}

int cafe_cfg_vxworks_kernel_delete_internal(void)
{
	cafe_hashtbl_delete(event_hashtbl, NULL, NULL, NULL, NULL);
	return CAFE_OK;
}

int cafe_ep_vxworks_kernel_enable_event(struct vxworks_event_handle *evt)
{
	return CAFE_OK;
}

void cafe_ep_vxworks_kernel_disable_event(struct vxworks_event_handle *evt)
{
}
