/* Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <errno.h>
#include <cafe/cafe.h>
#include <cafe/ptr_array.h>

#include "bailout_allocator.h"
#include "compiler.h"
#include "strbuf.h"

static const char *const progname = "cafe-type2vt";
static int join_vt_str;
static int compile_only;
static int quote_vt_str;
static int list_structs;
static int list_natures;
static int all_types_opt;

#ifdef __GNUC__
static int xfprintf(FILE *stream, const char *fmt, ...) __attribute__((format(printf, 2, 3)));
#endif

static int xfprintf(FILE *stream, const char *fmt, ...)
{
	int rc;
	va_list ap;

	va_start(ap, fmt);
	rc = vfprintf(stream, fmt, ap);
	va_end(ap);

	if (rc < 0) {
		fprintf(stderr, "write failed: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	return rc;
}

#define fprintf USE_xfprintf_instead

static void print_quoted_string(FILE *stream, const char *s, int terminator)
{
	const char *q = strchr(s, '"');

	if (q == NULL) {
		fputs(s, stream);
		fputc(terminator, stream);
	} else {
		const char *q = s;

		fputc('"', stream);

		while (*q != '\0') {
			switch (*q) {
			case '"':
				fputc('\\', stream);
			default:
				fputc(*q, stream);
				break;
			}
			q++;
		}
		fputc('"', stream);
		fputc(terminator, stream);
	}
}

static int qsort_cmp(const void *a, const void *b)
{
#ifdef _WIN32
	const struct pt_node **x = a;
	const struct pt_node **y = b;
#else
	struct pt_node *const *x = a;
	struct pt_node *const *y = b;
#endif
	assert((*x)->type == PT_STRUCT || (*x)->type == PT_NATURE);
	assert((*y)->type == PT_STRUCT || (*y)->type == PT_NATURE);
	return strcmp((*x)->left->istr->str, (*y)->left->istr->str);
}

struct get_types_ctx {
	const char *filename;
	struct cafe_ptr_array *types;
};

static int walk_top_level_types(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct get_types_ctx *ctx = (struct get_types_ctx *)data;

	if (visit != PT_WALK_PREORDER)
		return 1;

	if (n->type == PT_STRUCT || n->type == PT_NATURE)
		cafe_ptr_array_append(ctx->types, n);

	return 1;
}

static void get_types_for_filename(const char *filename, struct cafe_type_compiler *compiler, struct cafe_ptr_array *types)
{
	struct cafe_list_head *item;
	struct get_types_ctx ctx;

	ctx.filename = filename;
	ctx.types = types;

	cafe_list_foreach(item, &compiler->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		if (filename == NULL || all_types_opt || (p->root != NULL && strcmp(p->filename, filename)) == 0) {
			pt_walk(p->root, walk_top_level_types, &ctx);
		}
	}

	qsort(types->data, cafe_ptr_array_len(types), sizeof *types->data, qsort_cmp);
}

static void list_types(struct cafe_type_compiler *compiler, struct cafe_ptr_array *types)
{
	size_t i;

	for (i = 0; i < cafe_ptr_array_len(types); i++) {
		struct pt_node *n = cafe_ptr_array_val(types, i);

		if (n->type == PT_STRUCT && list_structs)
			xfprintf(stdout, "%s\n", n->left->istr->str);
		else if (n->type == PT_NATURE && list_natures)
			xfprintf(stdout, "%s\n", n->left->istr->str);
	}
}

static void print_types(struct cafe_type_compiler *compiler, struct cafe_ptr_array *types)
{
	size_t i;
	struct cafe_byte_buffer sb = STRBUF_INIT(bailout_allocator);

	for (i = 0; i < cafe_ptr_array_len(types); i++) {
		struct pt_node *n = cafe_ptr_array_val(types, i);

		if (compiler_value_type_str(n, &sb) == CAFE_OK) {
			if (!join_vt_str) {
				if (quote_vt_str)
					print_quoted_string(stdout, sb.data, '\n');
				else
					printf("%s\n", sb.data);

				strbuf_reset(&sb);
			}
		}
	}

	if (join_vt_str) {
		if (quote_vt_str)
			print_quoted_string(stdout, sb.data, '\n');
		else
			printf("%s\n", sb.data);
	}

	strbuf_release(&sb);
}

static void usage(void)
{
	xfprintf(stderr, "Usage: %s [-c] [-l] [-1] [-Q] [ <INPUT> ]\n", progname);
	xfprintf(stderr, "Converts Cafe Types (structs/natures) to Value Type strings\n");
	xfprintf(stderr, "  -c -- compile only\n");
	xfprintf(stderr, "  -s -- list struct types only; no conversion to VT\n");
	xfprintf(stderr, "  -n -- list nature types only; no conversion to VT\n");
	xfprintf(stderr, "  -1 -- print as 1 (long) string\n");
	xfprintf(stderr, "  -Q -- quote emitted types\n");
	exit(EXIT_FAILURE);
}

static int parse_opts(int argc, char *argv[])
{
	int opt;
	int ind = 1;

	while (ind < argc) {
		char *s = argv[ind];

		if (*s++ != '-')
			break;

		ind++;

		if (s[0] == '-' && s[1] == '\0')
			break;

		while ((opt = *s++) != '\0') {
			switch (opt) {
			case 'A':
				all_types_opt = 1;
				break;
			case 'Q':
				quote_vt_str = 1;
				break;
			case '1':
				join_vt_str = 1;
				break;
			case 'c':
				compile_only = 1;
				break;
			case 'n':
				list_natures = 1;
				break;
			case 's':
				list_structs = 1;
				break;
			case 'h':
			case '?':
			default:
				usage();
			}
		}
	}

	return ind;
}

int main(int argc, char *argv[])
{
	int rc;
	const char *filename = NULL;
	struct cafe_type_compiler compiler;
	int filename_index;
	struct cafe_ptr_array types;

	filename_index = parse_opts(argc, argv);

	if (filename_index < argc)
		filename = argv[filename_index];

	rc = compiler_init(&compiler, NULL, NULL, bailout_allocator);

	if (rc != CAFE_OK) {
		xfprintf(stderr, "%s: cannot initialize compiler\n", progname);
		return EXIT_FAILURE;
	}

	rc = compiler_parsefile(&compiler, filename);

	if (rc != CAFE_OK) {
		/* Errors already printed by the parser. */
		compiler_release(&compiler);
		return EXIT_FAILURE;
	}

	rc = compiler_validate(&compiler);

	if (rc != CAFE_OK || compile_only) {
		compiler_release(&compiler);
		return rc == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
	}

	assert(compiler.compilation_errors == 0);
	cafe_ptr_array_init(&types, bailout_allocator);
	get_types_for_filename(filename, &compiler, &types);

	if (list_structs || list_natures)
		list_types(&compiler, &types);
	else
		print_types(&compiler, &types);

	compiler_release(&compiler);
	cafe_ptr_array_release(&types);

	return EXIT_SUCCESS;
}
