/*
 * Copyright (c) 2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
24feb14,sdt  Disable unexpected NO_EFFECT warnings

DESCRIPTION

*/

#include <cafe/cafe.h>
#include <cafe/stdio.h>
#include "../src/strtoll.h"
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
#include "istr.h"
#include "token.h"
#include "grammar.h"
#include "lexer.h"

typedef enum {
	SYM_AND = 0,
	SYM_LE,
	SYM_EQ,
	SYM_GE,
	SYM_OR,
	SYM_NOT,
	SYM_MOD,
	SYM_TIMES,
	SYM_PLUS,
	SYM_MINUS,
	SYM_DIVIDE,
	SYM_QMARK,
	SYM_ASSIGNMENT,
	SYM_SEMI,
	SYM_COLON,
	SYM_COMMA,
	SYM_LT,
	SYM_GT,
	SYM_LPAREN,
	SYM_RPAREN,
	SYM_LBRACE,
	SYM_RBRACE,
	SYM_AT,
	SYM_HASH
} symbol_tag_t;

static struct {
	const char *str;
	unsigned int tkid;
	struct istr *istr;
} symbols[] = {
	{ "&&", TK_AND },
	{ "<=", TK_LE },
	{ "==", TK_EQ },
	{ ">=", TK_GE },
	{ "||", TK_OR },
	{ "!",	TK_NOT },
	{ "%",	TK_MOD },
	{ "*",	TK_TIMES },
	{ "+",	TK_PLUS },
	{ "-",	TK_MINUS },
	{ "/",	TK_DIVIDE },
	{ "?",	TK_QMARK },
	{ "=",	TK_ASSIGNMENT },
	{ ";",	TK_SEMI },
	{ ":",	TK_COLON },
	{ ",",	TK_COMMA },
	{ "<",	TK_LT },
	{ ">",	TK_GT },
	{ "(",	TK_LPAREN },
	{ ")",	TK_RPAREN },
	{ "{",	TK_LBRACE },
	{ "}",	TK_RBRACE },
	{ "@",	TK_AT },
	{ "#",	TK_HASH },
};

/* A note on the "include" keyword: by making "include" a keyword we
 * can leverage CMake's dependency analysis when we generate a header
 * file; otherwise it seems very difficult to hook into how cmake
 * handles dependencies for add_custom_target(). */

static struct {
	const char *str;
	unsigned int tkid;
	size_t strsz;
	struct istr *istr;
} keywords[] = {
	{ "include",	TK_IMPORT },
	{ "array",	TK_ARRAY },
	{ "blob",	TK_BLOB },
	{ "boolean",	TK_BOOLEAN },
	{ "char",	TK_CHAR },
	{ "const",	TK_CONST },
	{ "dict",	TK_DICT },
	{ "double",	TK_DOUBLE },
	{ "false",	TK_FALSE },
	{ "float",	TK_FLOAT },
	{ "import",	TK_IMPORT },
	{ "int",	TK_INT },
	{ "key",	TK_KEY },
	{ "long",	TK_LONG },
	{ "longlong",	TK_LONGLONG },
	{ "namespace",	TK_NAMESPACE },
	{ "nature",	TK_NATURE },
	{ "setting",	TK_SETTING },
	{ "short",	TK_SHORT },
	{ "string",	TK_STRING },
	{ "struct",	TK_STRUCT },
	{ "true",	TK_TRUE },
	{ "typedef",	TK_TYPEDEF },
	{ "uchar",	TK_UCHAR },
	{ "uint",	TK_UINT },
	{ "ulong",	TK_ULONG },
	{ "ulonglong",	TK_ULONGLONG },
	{ "ushort",	TK_USHORT }
};

void lexer_error(struct lexer *l, const struct src_loc *loc, const char *fmt, ...)
{
	va_list ap;
	struct cafe_byte_buffer sb = STRBUF_INIT(l->allocator);

	l->lexer_errors++;
	va_start(ap, fmt);
	strbuf_vfmt(&sb, fmt, ap);
	va_end(ap);
	assert(l->error_handler);
	l->error_handler(loc, sb.data);
	strbuf_release(&sb);
}

void lexer_default_error_handler(const struct src_loc *loc, const char *msg)
{
	if (loc != NULL) {
		fprintf(stderr, "%s%s%d:%d: ",
			loc->filename ? loc->filename : "",
			loc->filename ? ":" : "",
			loc->lineno, loc->column);
		fprintf(stderr, "error: ");
	} else {
		fprintf(stderr, "error: ");
	}

	fputs(msg, stderr);
	fputc('\n', stderr);
}

static const char *token_type_str(int id)
{
	switch (id) {
	case TK_AND:		return "AND";
	case TK_ARRAY:		return "ARRAY";
	case TK_ASSIGNMENT:	return "ASSIGNMENT";
	case TK_AT:		return "AT";
	case TK_BLOB:		return "BLOB";
	case TK_BOOLEAN:	return "BOOLEAN";
	case TK_CHAR:		return "CHAR";
	case TK_COLON:		return "COLON";
	case TK_COMMA:		return "COMMA";
	case TK_CONST:		return "CONST";
	case TK_DICT:		return "DICT";
	case TK_DIVIDE:		return "DIVIDE";
	case TK_DOCUMENTATION:	return "DOCUMENTATION";
	case TK_DOUBLE:		return "DOUBLE";
	case TK_EQ:		return "EQ";
	case TK_FALSE:		return "FALSE";
	case TK_FLOAT:		return "FLOAT";
	case TK_GE:		return "GE";
	case TK_GT:		return "GT";
	case TK_HASH:		return "HASH";
	case TK_IDENT:		return "IDENT";
	case TK_IMPORT:		return "IMPORT";
	case TK_INT:		return "INT";
	case TK_KEY:		return "KEY";
	case TK_LBRACE:		return "LBRACE";
	case TK_LE:		return "LE";
	case TK_LONG:		return "LONG";
	case TK_LONGLONG:	return "LONGLONG";
	case TK_LPAREN:		return "LPAREN";
	case TK_LT:		return "LT";
	case TK_MINUS:		return "MINUS";
	case TK_MOD:		return "MOD";
	case TK_NAMESPACE:	return "NAMESPACE";
	case TK_NATURE:		return "NATURE";
	case TK_NE:		return "NE";
	case TK_NOT:		return "NOT";
	case TK_NUMBER:		return "NUMBER";
	case TK_OR:		return "OR";
	case TK_PLUS:		return "PLUS";
	case TK_QMARK:		return "QMARK";
	case TK_RBRACE:		return "RBRACE";
	case TK_REAL:		return "REAL";
	case TK_RPAREN:		return "RPAREN";
	case TK_SEMI:		return "SEMI";
	case TK_SETTING:	return "SETTING";
	case TK_SHORT:		return "SHORT";
	case TK_STRING:		return "STRING";
	case TK_STRUCT:		return "STRUCT";
	case TK_STRVAL:		return "STRVAL";
	case TK_TIMES:		return "TIMES";
	case TK_TRUE:		return "TRUE";
	case TK_TYPEDEF:	return "TYPEDEF";
	case TK_UCHAR:		return "UCHAR";
	case TK_UINT:		return "UINT";
	case TK_ULONG:		return "ULONG";
	case TK_ULONGLONG:	return "ULONGLONG";
	case TK_USHORT:		return "USHORT";
	}

	return "<illegal>";
}

static inline bool is_ident_char(int c)
{
	return isalnum(c) || c == '_' || c == '.' || c == '@' || c == ':';
}

static inline int token_set(struct token *t, symbol_tag_t sym)
{
	t->id = symbols[sym].tkid;
	t->istr = symbols[sym].istr;
	t->name = token_type_str(t->id);

	return t->id;
}

static inline int consume(struct lexer *l)
{
	return (l->c = fgetc_unlocked(l->stream));
}

static inline int peeknext(struct lexer *l)
{
	int c = fgetc_unlocked(l->stream);
	ungetc(c, l->stream);
	return c;
}

static void consume_nl(struct lexer *l)
{
	if (l->c == '\r' || l->c == '\n') {
		if (l->c == '\r' && peeknext(l) == '\n') {
			consume(l);
			l->column++;
		}

		l->lineno++;
		l->column = 0;
	}
}

static void consume_nl_and_save(struct lexer *l, struct cafe_byte_buffer *sb)
{
	if (l->c == '\r' || l->c == '\n') {
		if (l->c == '\r' && peeknext(l) == '\n') {
			consume(l);
			strbuf_addc(sb, l->c);
			l->column++;
		}

		l->lineno++;
		l->column = 0;
	}
}

static struct istr *intern_stringbuffer(struct lexer *l)
{
	return istr_intern(l->strtbl, l->sb.data, l->sb.len, l->allocator);
}

static bool read_string(struct lexer *l, int delimiter)
{
	struct token *t = &l->token;

	assert(l->c == delimiter);
	t->loc.lineno = l->lineno;
	t->loc.column = l->column;
	consume(l);
	l->column++;
	t->name = token_type_str(TK_STRVAL);

	while (l->c != delimiter) {
		switch (l->c) {
		case EOF:
			lexer_error(l, &t->loc, "unterminated string");
			return 0;
		case '\r':
		case '\n': {
			lexer_error(l, &t->loc, "newline in string constant");
			return 0;
		}
		case '\\': {
			int c;
			consume(l);
			l->column++;
			switch (l->c) {
			case 'f': c = '\f'; break;
			case 'n': c = '\n'; break;
			case 'r': c = '\r'; break;
			case 't': c = '\t'; break;
			case 'v': c = '\v'; break;
			case '\n':
			case '\r':
				strbuf_addc(&l->sb, '\n'); /* reduced to just \n */
				consume_nl(l);
				continue;
			case EOF: /* pick up on this error at start of loop */
				continue;
			default:
				switch (l->c) {
				case '\\':
				case '\"':
				case '\'':
					strbuf_addc(&l->sb, l->c);
					consume(l);
					l->column++;
					continue;
				default: {
					struct src_loc loc;
					loc.filename = l->filename;
					loc.lineno = l->lineno;
					loc.column = l->column;
					lexer_error(l, &loc, "unexpected escape character");
					return 0;
				}
				}
			}
			strbuf_addc(&l->sb, c);
			consume(l);
			l->column++;
			continue;
		}
		default:
			strbuf_addc(&l->sb, l->c);
			consume(l);
			l->column++;
			continue;
		}
	}

	t->istr = intern_stringbuffer(l);
	t->id = TK_STRVAL;

	return t->id;
}

static bool read_comment(struct lexer *l)
{
	unsigned int nesting = 1;
	struct token *t = &l->token;

	assert(l->c == '/');
	t->id = 0;
	t->loc.lineno = l->lineno;
	t->loc.column = l->column;

	do {
		strbuf_addc(&l->sb, l->c);
		consume(l);
		l->column++;

		switch (l->c) {
		case EOF: {
			struct src_loc loc;
			loc.filename = l->filename;
			loc.lineno = l->lineno;
			loc.column = l->column;
			lexer_error(l, &loc, "end of file in comment");
			lexer_error(l, &t->loc, "comment started here");
			return false;
		}
		case '\r':
		case '\n':
			consume_nl_and_save(l, &l->sb);
			break;
		case '*':
			if (peeknext(l) == '/') {
				strbuf_addc(&l->sb, l->c);
				consume(l);
				nesting--;
			}
			break;
		case '/':
			if (peeknext(l) == '*') {
				strbuf_addc(&l->sb, l->c);
				consume(l);
				nesting++;
			}
			break;
		}
	} while (nesting != 0);

	strbuf_addc(&l->sb, l->c);

	if (l->sb.len > 3 && strncmp(l->sb.data, "/**", 3) == 0) {
		if (l->sb.data[3] != '/') {
			t->istr = intern_stringbuffer(l);
			t->id = TK_DOCUMENTATION;
			t->name = token_type_str(t->id);
		}
	}

	return true;
}

static int read_number(struct lexer *l)
{
	struct token *t = &l->token;
	bool is_floating_literal;
	bool is_negative;
	char *endptr;

	assert(isdigit(l->c) || l->c == '.' || l->c == '-');
	t->id = 0;
	t->loc.lineno = l->lineno;
	t->loc.column = l->column;
	is_negative = l->c == '-';

	if (is_negative)
		is_floating_literal = peeknext(l) == '.';
	else
		is_floating_literal = l->c == '.';

	strbuf_addc(&l->sb, l->c);

	/* WRS_TODO: add read_number_with_leading_0() for hex values. */

	while (isdigit(peeknext(l))) {
		consume(l);
		l->column++;
		strbuf_addc(&l->sb, l->c);
	}

	if (!is_floating_literal && peeknext(l) == '.') {
		is_floating_literal = true;
		do {
			consume(l);
			l->column++;
			strbuf_addc(&l->sb, l->c);
		} while (isdigit(peeknext(l)));
	}

	t->istr = intern_stringbuffer(l);

	if (is_floating_literal) {
		double d;

		errno = 0;
		d = strtod(l->sb.data, &endptr);
		d = *&d;	/* squash gcc's warn_unused_result warning. */

		if (errno != 0 || l->sb.data >= endptr) {
			if (errno == ERANGE)
				lexer_error(l, &t->loc, "%s", strerror(ERANGE));
			else
				lexer_error(l, &t->loc, "malformed number");
			return 0;
		}
	} else {
		if (is_negative) {
			long long ll;
			errno = 0;
			ll = strtoll(l->sb.data, &endptr, 10);
			ll = *&ll;   /* squash gcc's warn_unused_result warning. */
		} else {
			long long ull;
			errno = 0;
			ull = strtoull(l->sb.data, &endptr, 10);
			ull = *&ull; /* squash gcc's warn_unused_result warning. */
		}

		if (errno != 0 || l->sb.data >= endptr) {
			if (errno == ERANGE)
				lexer_error(l, &t->loc, "%s", strerror(ERANGE));
			else
				lexer_error(l, &t->loc, "malformed number");
			return 0;
		}
	}

	t->id = is_floating_literal ? TK_REAL : TK_NUMBER;
	t->name = token_type_str(t->id);

	return t->id;
}

void lexer_init(struct lexer *l, lexer_error_handler error_handler, const char *filename, FILE *stream, struct cafe_hashtbl *strtbl, const struct cafe_allocator *allocator)
{
	size_t i;

	memset(l, 0, sizeof *l);
	l->lineno = 1;
	l->column = 0;
	l->filename = filename;
	l->stream = stream;
	l->strtbl = strtbl;
	l->allocator = allocator;
	l->error_handler = error_handler;
	strbuf_init(&l->sb, l->allocator);

	for (i = 0; i < CAFE_NELEMENTS(symbols); i++)
		symbols[i].istr = istr_intern(l->strtbl, symbols[i].str,
					      strlen(symbols[i].str),
					      l->allocator);

	for (i = 0; i < CAFE_NELEMENTS(keywords); i++) {
		keywords[i].strsz = strlen(keywords[i].str);
		keywords[i].istr = istr_intern(l->strtbl, keywords[i].str,
					       keywords[i].strsz, l->allocator);
	}
}

int lexer_next_token(struct lexer *l)
{
	strbuf_reset(&l->sb);
	l->token.loc.filename = l->filename;
	l->token.name = NULL;

	for (;;) {
		consume(l);
		l->column++;
		l->token.loc.lineno = l->lineno;
		l->token.loc.column = l->column;
		l->token.id = 0;

		switch (l->c) {
		case EOF:
			return l->token.id = TK_EOF;
		case '\r':
		case '\n':
			consume_nl(l);
			continue;
		case '\t':
			l->column = (l->column + 7) & -8;
			continue;
		case ' ':
			continue;
		case '<': {
			if (peeknext(l) == '=') {
				consume(l);
				l->column++;
				return token_set(&l->token, SYM_LE);
			}
			return token_set(&l->token, SYM_LT);
		}
		case '=': {
			if (peeknext(l) == '=') {
				consume(l);
				l->column++;
				return token_set(&l->token, SYM_EQ);
			}
			return token_set(&l->token, SYM_ASSIGNMENT);
		}
		case '>': {
			if (peeknext(l) == '=') {
				consume(l);
				l->column++;
				return token_set(&l->token, SYM_GE);
			}
			return token_set(&l->token, SYM_GT);
		}
		case '!':
			return token_set(&l->token, SYM_NOT);
		case '%':
			return token_set(&l->token, SYM_MOD);
		case '*':
			return token_set(&l->token, SYM_TIMES);
		case '+':
			return token_set(&l->token, SYM_PLUS);
		case '/': {
			switch (peeknext(l)) {
			case '/':
				do {
					consume(l);
					l->column++;
				} while (l->c != '\r' && l->c != '\n');
				consume_nl(l);
				continue;
			case '*':
				if (!read_comment(l))
					return 0;
				if (l->token.id == TK_DOCUMENTATION)
					return TK_DOCUMENTATION;
				continue;
			default:
				return token_set(&l->token, SYM_DIVIDE);
			}
		}
		case '?':
			return token_set(&l->token, SYM_QMARK);
		case ';':
			return token_set(&l->token, SYM_SEMI);
		case ':':
			return token_set(&l->token, SYM_COLON);
		case ',':
			return token_set(&l->token, SYM_COMMA);
		case '(':
			return token_set(&l->token, SYM_LPAREN);
		case ')':
			return token_set(&l->token, SYM_RPAREN);
		case '{':
			return token_set(&l->token, SYM_LBRACE);
		case '}':
			return token_set(&l->token, SYM_RBRACE);
		case '@':
			return token_set(&l->token, SYM_AT);
		case '#':
			return token_set(&l->token, SYM_HASH);
		case '"':
			return read_string(l, '"');
		default:
			if (isspace(l->c))
				continue;

			if (isdigit(l->c)) {
				return read_number(l);
			} else if (l->c == '-' && isdigit(peeknext(l))) {
				return read_number(l);
			} else if (l->c == '.' && isdigit(peeknext(l))) {
				return read_number(l);
			} else if (l->c == '-') {
				return token_set(&l->token, SYM_MINUS);
			} else if (isalpha(l->c) || (l->c == '^' && isalpha(peeknext(l)))) {
				size_t i;
				bool escape_keyw = false;
				strbuf_reset(&l->sb);
				if (l->c == '^') {
					consume(l);
					l->column++;
					escape_keyw = true;
				}
				strbuf_addc(&l->sb, l->c);
				l->token.loc.lineno = l->lineno;
				l->token.loc.column = l->column;
				while ((l->c = peeknext(l)) != EOF && is_ident_char(l->c)) {
					strbuf_addc(&l->sb, l->c);
					consume(l);
					l->column++;
				}
				for (i = 0; !escape_keyw && i < CAFE_NELEMENTS(keywords); i++) {
					if (strcmp(l->sb.data, keywords[i].str) == 0) {
						l->token.id = keywords[i].tkid;
						l->token.name = token_type_str(l->token.id);
						l->token.istr = keywords[i].istr;
						return keywords[i].tkid;
					}
				}
				l->token.id = TK_IDENT;
				l->token.name = token_type_str(TK_IDENT);
				l->token.istr = intern_stringbuffer(l);
				return l->token.id;
			} else if (l->c == '|' && peeknext(l) == '|') {
				consume(l);
				l->column++;
				return token_set(&l->token, SYM_OR);
			} else if (l->c == '&' && peeknext(l) == '&') {
				consume(l);
				l->column++;
				return token_set(&l->token, SYM_AND);
			} else {
				strbuf_addc(&l->sb, l->c);
				l->token.istr = intern_stringbuffer(l);
				l->token.name = l->token.istr->str;
				return l->token.id = TK_ILLEGAL;
			}
			break;
		}
		assert(!feof(l->stream));
	}
	/* coverity[array_null] */
	assert(!"not reached");
}
