/*
 * Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

#include <stdio.h>
#include <cafe/cafe.h>
#include <assert.h>
#include "symtbl.h"
#include "istr.h"

#define PRINT_DEBUG 0

struct symbol *symbol_alloc(struct istr *istr, pt_type_t type, struct src_loc *loc, struct scope *scope, const struct cafe_allocator *allocator)
{
	struct symbol *symbol = CAFE_ALLOC(allocator, sizeof *symbol);

	if (symbol == NULL)
		return NULL;

	memset(symbol, 0, sizeof *symbol);
	assert(istr != NULL);
	cafe_list_init(&symbol->symbol_list_item);
	symbol->name = istr;
	symbol->type = type;
	symbol->symbol_type = NULL;
	symbol->loc = loc;
	symbol->node_def = NULL;
	symbol->node_def_parent = NULL;
	symbol->scope = scope;
	symbol->eval_type = PT_UNDEFINED;

	return symbol;
}

void symbol_free(struct symbol *symbol, const struct cafe_allocator *allocator)
{
	cafe_list_remove(&symbol->symbol_list_item);
	CAFE_FREE(allocator, symbol);
}

struct scope *scope_alloc(struct scope *enclosing_scope, const char *name, const struct cafe_allocator *allocator)
{
	struct scope *scope = CAFE_ALLOC(allocator, sizeof *scope);

	if (scope == NULL)
		return NULL;

	scope->symbols = cafe_hashtbl_new(5, allocator, istr_hashtbl_hash, istr_hashtbl_equals);

	if (scope->symbols == NULL) {
		CAFE_FREE(allocator, scope);
		return NULL;
	}

	scope->enclosing_scope = enclosing_scope;
	scope->name = name;
	cafe_list_init(&scope->symbol_list_head);
	cafe_list_init(&scope->scope_list_item);

	return scope;
}

void scope_free(struct scope *scope, const struct cafe_allocator *allocator)
{
	cafe_hashtbl_delete(scope->symbols, NULL, NULL, NULL, NULL);
	cafe_list_remove(&scope->scope_list_item);
	CAFE_FREE(allocator, scope);
}

int scope_add_symbol(struct scope *scope, struct symbol *symbol)
{
	int rc;

	assert(cafe_hashtbl_lookup(scope->symbols, symbol->name) == NULL);
	rc = cafe_hashtbl_insert(scope->symbols, (void *)symbol->name, symbol);

	if (rc != 0)
		return CAFE_NOMEM;

	cafe_list_add_last(&symbol->symbol_list_item, &scope->symbol_list_head);
#if PRINT_DEBUG
	fprintf(stderr, "define symbol '%s' in scope '%s' (load = %f)\n",
		symbol->name->str,
		scope->name,
		(double)cafe_hashtbl_count(scope->symbols) / cafe_hashtbl_capacity(scope->symbols));
#endif
	return rc;
}

struct symbol *scope_resolve(struct scope *scope, struct istr *istr)
{
	struct symbol *symbol;

	assert(scope);
	symbol = cafe_hashtbl_lookup(scope->symbols, istr);
#if PRINT_DEBUG
	fprintf(stderr, "resolving '%s' in scope '%s' (%s)\n", istr->str, scope->name, (symbol == NULL) ? "NOT FOUND" : "found");
#endif
	if (symbol != NULL)
		return symbol;

	if (scope->enclosing_scope != NULL)
		return scope_resolve(scope->enclosing_scope, istr);

	return NULL;
}

struct symbol *scope_resolve_field(struct scope *scope, struct istr *istr)
{
	struct symbol *symbol;

	symbol = cafe_hashtbl_lookup(scope->symbols, istr);
#if PRINT_DEBUG
	fprintf(stderr, "resolving '%s' in scope '%s' (%s)\n", istr->str, scope->name, (symbol == NULL) ? "NOT FOUND" : "found");
#endif
	return symbol;
}
