/* Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <cafe/cafe.h>
#include <assert.h>
#include "compiler.h"
#include "strbuf.h"
#include "symtbl.h"
#include "pt.h"

static struct vt_type {
	pt_type_t type;
	const char *tag;
} vt_types[] = {
	{ PT_ARRAY,	"a" },
	{ PT_BLOB,	"b" },
	{ PT_BOOLEAN,	"?" },
	{ PT_CHAR,	"c" },
	{ PT_DICT,	"D" },
	{ PT_DOUBLE,	"d" },
	{ PT_FLOAT,	"f" },
	{ PT_INT,	"i" },
	{ PT_LONG,	"l" },
	{ PT_LONGLONG,	"q" },
	{ PT_SHORT,	"h" },
	{ PT_STRING,	"s" },
	{ PT_UCHAR,	"C" },
	{ PT_UINT,	"I" },
	{ PT_ULONG,	"L" },
	{ PT_ULONGLONG, "Q" },
	{ PT_USHORT,	"H" },
};

static const struct vt_type *vt_type(pt_type_t type)
{
	size_t i;

	for (i = 0; i < CAFE_NELEMENTS(vt_types); i++) {
		if (vt_types[i].type == type)
			return &vt_types[i];
	}

	return NULL;
}

static const char *key_qualifier(const struct pt_node *n)
{
	size_t i;

	assert(n->type == PT_ATTRIBUTE);

	if (n->left == NULL)
		return "v";

	for (i = 0; i < n->left->children.count; i++) {
		if (n->left->children.list[i]->type == PT_KEY)
			return "k";
		if (n->left->children.list[i]->type == PT_SETTING)
			return "p";
	}

	return "v";
}

static int encode_type(struct pt_node *n, struct cafe_byte_buffer *sb)
{
	switch (n->type) {
	case PT_ARRAY:
		strbuf_fmt(sb, "%s[", vt_type(n->type)->tag);
		encode_type(n->left, sb);
		if (n->middle) {
			strbuf_fmt(sb, ","); /* dim */
			encode_type(n->middle, sb);
		}
		strbuf_fmt(sb, "]");
		break;
	case PT_DICT:
		strbuf_fmt(sb, "%s[", vt_type(n->type)->tag);
		encode_type(n->left, sb);
		strbuf_fmt(sb, "]");
		break;
	case PT_BLOB:
	case PT_BOOLEAN:
	case PT_CHAR:
	case PT_DOUBLE:
	case PT_FLOAT:
	case PT_INT:
	case PT_LONG:
	case PT_LONGLONG:
	case PT_REAL:
	case PT_SHORT:
	case PT_STRING:
	case PT_UCHAR:
	case PT_UINT:
	case PT_ULONG:
	case PT_ULONGLONG:
	case PT_USHORT:
		strbuf_fmt(sb, "%s", vt_type(n->type)->tag);
		break;
	case PT_IDENTIFIER:
		strbuf_fmt(sb, "t[%s]", n->istr->str);
		break;
	case PT_ATTRIBUTE:
	case PT_CONST:
	case PT_NUMBER:
		strbuf_fmt(sb, "%s", n->istr->str);
		break;
	case PT_AND:
	case PT_ASSIGNMENT:
	case PT_CONDITIONAL:
	case PT_DIVIDE:
	case PT_EQ:
	case PT_FALSE:
		strbuf_fmt(sb, "0");
		break;
	case PT_GE:
	case PT_GT:
	case PT_KEY:
	case PT_LE:
	case PT_LIST:
	case PT_LT:
	case PT_MINUS:
	case PT_MOD:
	case PT_NATURE:
	case PT_NATURE_REF:
	case PT_NE:
	case PT_NOT:
	case PT_NULL:
	case PT_OR:
	case PT_PLUS:
	case PT_ROOT:
	case PT_SETTING:
	case PT_STRUCT:
	case PT_STRVAL:
	case PT_TIMES:
	case PT_TRUE:
		strbuf_fmt(sb, "0");
		break;
	case PT_TYPEDEF:
		break;
	case PT_UNDEFINED:
		assert(0);
		break;
	}

	return CAFE_OK;
}

static void write_expr(struct cafe_byte_buffer *sb, struct pt_node *n)
{
	if (n == NULL)
		return;

	write_expr(sb, n->left);

	if (n->type == PT_STRVAL)
		strbuf_fmt(sb, "\"%s\"", n->istr->str);
	else if (n->type == PT_TRUE)
		strbuf_fmt(sb, "1");
	else if (n->type == PT_FALSE)
		strbuf_fmt(sb, "0");
	else
		strbuf_fmt(sb, "%s", n->istr->str);

	write_expr(sb, n->right);
}

static int walk_natures(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *ctx, void *data)
{
	struct cafe_byte_buffer *sb = (struct cafe_byte_buffer *)data;

	if (visit != PT_WALK_PREORDER)
		return 1;

	if (n->type == PT_NATURE_REF) {
		strbuf_fmt(sb, "n[%s", n->left->istr->str);
		if (n->middle != NULL) {
			size_t i;

			for (i = 0; i < n->middle->children.count; i++) {
				struct pt_node *c = n->middle->children.list[i];
				strbuf_fmt(sb, ",");
				write_expr(sb, c);
			}
		}
		strbuf_fmt(sb, "]");
	}

	return 1;
}

static int walk_attributes(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *ctx, void *data)
{
	struct cafe_byte_buffer *sb = (struct cafe_byte_buffer *)data;

	if (n->type == PT_ATTRIBUTE) {
		switch (visit) {
		case PT_WALK_PREORDER:
			strbuf_fmt(sb, "%s[%s,", key_qualifier(n), n->right->istr->str);
			encode_type(n->middle, sb);
			if (n->arg5 != NULL) {
				strbuf_fmt(sb, ",");
				pt_walk(n->arg5, walk_natures, sb);
			}
			if (n->arg4 != NULL) {
				if (n->arg5 == NULL)
					strbuf_fmt(sb, ",");
				if (n->arg4->left != NULL) {
					strbuf_fmt(sb, ",");
					write_expr(sb, n->arg4->left);
				}
			}
			break;
		case PT_WALK_POSTORDER:
			strbuf_fmt(sb, "]");
			break;
		}
	}

	return 1;
}

static int walk_types(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *ctx, void *data)
{
	struct cafe_byte_buffer *sb = (struct cafe_byte_buffer *)data;

	if (n->type == PT_STRUCT || n->type == PT_NATURE) {
		switch (visit) {
		case PT_WALK_PREORDER:
			strbuf_fmt(sb, "%c[%s",
				   n->type == PT_STRUCT ? 'S' : 'N',
				   n->left->istr->str);
			if (n->right != NULL) {
				strbuf_fmt(sb, ",");
				pt_walk(n->right, walk_attributes, sb);
			}
			if (n->middle) {
				if (n->right == NULL)
					strbuf_fmt(sb, ",");
				strbuf_fmt(sb, ",");
				pt_walk(n->middle, walk_natures, sb);
			}
			break;
		case PT_WALK_POSTORDER:
			strbuf_fmt(sb, "]");
			break;
		}
	} else if (n->type == PT_TYPEDEF) {
		switch (visit) {
		case PT_WALK_PREORDER:
			strbuf_fmt(sb, "T[%s,", n->middle->istr->str);
			encode_type(n->left, sb);
			break;
		case PT_WALK_POSTORDER:
			strbuf_fmt(sb, "]");
			break;
		}
	}

	return level <= 1;
}

int compiler_value_type_str(struct pt_node *n, struct cafe_byte_buffer *sb)
{
	assert(n);
	assert(n->type == PT_STRUCT ||
	       n->type == PT_NATURE ||
	       n->type == PT_TYPEDEF);
	pt_walk(n, walk_types, sb);
	return CAFE_OK;
}
