/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
02sep14,sdt  Fixed build warnings

DESCRIPTION
  Implementations of Cafe OS primitives for VxWorks.
*/


#if CAFE_OS_VXWORKS

#include "cafe_os.h"
#include <cafe/cafe.h>
#include <cafe/ccdiff.h>
#include <vxWorks.h>
#include <semLib.h>
#include <edrLib.h>
#include <intLib.h>
#include <taskLib.h>
#include <private/vmLibP.h>
#include <stdio.h>
#include "cafe_os_vxworks.h"
#include "cafe_assert.h"
#if defined(TCF_GUARDS)
#include <framework/events.h>
#endif

static SEM_ID cafe_mutex;
static SEM_ID cafe_cond;

#if (_WRS_VXWORKS_MAJOR < 6)
typedef int TASK_ID;
#endif


size_t getpagesize(void)
{
	size_t pagesize = VM_PAGE_SIZE_GET();
	if (pagesize == (size_t)ERROR)
		return 4096;
	return pagesize;
}


int cafe_sched_yield(void)
{
	(void)taskDelay(0);
	return 0;
}

int cafe_lock_init(void)
{
	cafe_mutex = semMCreate(SEM_Q_PRIORITY);
	cafe_cond = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
	return (cafe_mutex == NULL || cafe_cond == NULL) ? CAFE_ERROR : CAFE_OK;
}

bool cafe_islocked(void)
{
	return (TASK_ID) SEM_OWNER(cafe_mutex) == (TASK_ID)taskIdSelf();
}

void __cafe_lock(const char *file, int line)
{
#if defined(TCF_GUARDS)
	__cafe_assert(!is_dispatch_thread(), file, line, NULL);
#endif
	__cafe_assert(!cafe_islocked(), file, line, NULL);
	CAFE_ASSERT_OK(semTake(cafe_mutex, WAIT_FOREVER));
}

void __cafe_unlock(const char *file, int line)
{
#if defined(TCF_GUARDS)
	__cafe_assert(!is_dispatch_thread(), file, line, NULL);
#endif
	__cafe_assert(cafe_islocked(), file, line, NULL);
	CAFE_ASSERT_OK(semGive(cafe_mutex));
}

int cafe_notify_signal(void)
{
	CAFE_ASSERT_LOCKED();
	CAFE_ASSERT_OK(semFlush(cafe_cond));
	return CAFE_OK;
}

int cafe_notify_wait(void)
{
	CAFE_ASSERT_LOCKED();
	CAFE_ASSERT_OK(semExchange(cafe_mutex, cafe_cond, WAIT_FOREVER));
	CAFE_ASSERT_OK(semTake(cafe_mutex, WAIT_FOREVER));
	return CAFE_OK;
}

void __cafe_assert_rtn(const char *expr, const char *file, int line, 
                       const char *msg)
{
	const char *str = msg ? msg : expr;

	if (INT_CONTEXT() == TRUE) {
#ifdef _WRS_CONFIG_SMP
		/*
		 * EDR log is spinlock and intCpuLock restricted so
		 * for SMP we call printf directly in a hope to get
		 * something useful out of the UART to aid debugging.
		 */
		fdprintf(2, "%s:%d Assertion `%s' failed\n", file, line, str);
#else
		(void)EDR_INJECT(EDR_FACILITY_INTERRUPT | EDR_SEVERITY_FATAL,
		                 file, line, NULL, NULL, NULL, msg);
#endif
	} else {
#ifndef _WRS_CONFIG_SMP
		(void)EDR_INJECT(EDR_FACILITY_KERNEL | EDR_SEVERITY_WARNING,
		                 file, line, NULL, NULL, NULL, msg);
#endif
		(void)fdprintf(2, "%s:%d Assertion `%s' failed\n", file, line, 
		               str);
		(void)taskSuspend(0);
	}
}

#endif	/* CAFE_OS_VXWORKS */
