/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
20Aug14,sdt  Prevent crash when type->keymap is NULL - WB4-4006
24feb14,sdt  Disable unexpected NO_EFFECT warnings 
21feb14,sdt  Fix compiler warnings

DESCRIPTION

*/

#include "cafe_assert.h"
#include <stddef.h>
#include <string.h>
#include <cafe/cafe.h>
#include <cafe/dict.h>
#if defined(CAFE_DEBUG)
#include <stdio.h>
#endif

#if defined(_MSC_VER)
#pragma warning(disable:4244)	/* conversion error: possible loss of data */
#endif

typedef cafe_hashval_t hashval_t;

static bool invalidHashval = false;

#if defined(CAFE_DEBUG)
static inline bool is_type_valid(const struct cafe_dict_struct *d,
				 const char * func)
#else
static inline bool is_type_valid(const struct cafe_dict_struct *d)
#endif
{
    invalidHashval = false;
    if ((d->type == NULL) ||
        ((d->type->elements > 0) && (d->type->keymap==0))) {
#if defined(CAFE_DEBUG)
        fprintf(stderr, "%s : invalid type: dict=0x%p type=0x%p, " \
                "keymap=0x%p, type_str=%s\n",
                func, (void*)d, (void*)d->type,
                (void*) d->type->keymap, d->type->type_str);
#endif
        invalidHashval = true;
    }
    return !invalidHashval;
}

#if defined(CAFE_DEBUG)
#  define IS_TYPE_VALID(d) is_type_valid(d, __FUNCTION__)
#else
#  define IS_TYPE_VALID(d) is_type_valid(d)
#endif

static hashval_t hash_buf(hashval_t hashval, const void *p, size_t l)
{
	const unsigned char *cp = p;
	while (l-- > 0)
		hashval = hashval * 17 + *cp++;
	return hashval;
}

static hashval_t cafe_float_hash(float v)
{
	return hash_buf(0, &v, sizeof v);
}

static hashval_t cafe_double_hash(double v)
{
	return hash_buf(0, &v, sizeof v);
}

static bool cafe_float_same_value(const float *v1, const float *v2)
{
	return memcmp(v1, v2, sizeof *v1) == 0;
}

static bool cafe_double_same_value(const double *v1, const double *v2)
{
	return memcmp(v1, v2, sizeof *v1) == 0;
}

static bool cafe_float_same_key(const float *v1, const cafe_any_value *key)
{
	float v2 = (float)key->d;
	return cafe_float_same_value(v1, &v2);
}

static bool cafe_double_same_key(const double *v1, const cafe_any_value *key)
{
	double v2 = key->d;
	return cafe_double_same_value(v1, &v2);
}

static hashval_t hashValue(hashval_t hashval,
			   const struct cafe_value_type *type,
			   const void *base)
{
	unsigned int i;
	hashval_t v = 0;
	struct cafe_vm_attr *attr;

	type = cafe_base_type(type);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
		v = *(char *)base;
		break;

	case CAFE_TYPE_UCHAR:
		v = *(unsigned char *)base;
		break;

	case CAFE_TYPE_SHORT:
		v = *(short *)base;
		break;

	case CAFE_TYPE_USHORT:
		v = *(unsigned short *)base;
		break;

	case CAFE_TYPE_INT:
		v = *(int *)base;
		break;

	case CAFE_TYPE_UINT:
		v = *(unsigned int *)base;
		break;

	case CAFE_TYPE_LONG:
		v = *(long *)base;
		break;

	case CAFE_TYPE_ULONG:
		v = *(unsigned long *)base;
		break;

	case CAFE_TYPE_LONGLONG:
		v = (hashval_t) *(long long *)base;
		break;

	case CAFE_TYPE_ULONGLONG:
		v = (hashval_t) *(unsigned long long *)base;
		break;

	case CAFE_TYPE_BOOL:
		v = *(bool *)base;
		break;

	case CAFE_TYPE_FLOAT:
		v = cafe_float_hash(*(float *)base);
		break;

	case CAFE_TYPE_DOUBLE:
		v = cafe_double_hash(*(double *)base);
		break;

	case CAFE_TYPE_STRUCT:
		cafe_type_attr_foreach(attr, type) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			hashval = hashValue(hashval, cafe_attr_type(attr),
					    ((char *)base) + attr->offset);
		}
		return hashval;

	case CAFE_TYPE_ARRAY:
		for (i = 0; i < type->elements; i++) {
			hashval = hashValue(hashval, cafe_element_type(type), base);
			base = ((char *)base) + cafe_element_type(type)->size;
		}
		return hashval;

	case CAFE_TYPE_DYNARRAY: {
		const struct cafe_vm_dynarray *a = base;
		base = a->p;
		for (i = 0; i < a->l; i++) {
			hashval = hashValue(hashval, cafe_element_type(type),
					    base);
			base = ((char *)base) + cafe_element_type(type)->size;
		}
		return hashval;
	}

	case CAFE_TYPE_STRING: {
		char *s = *(char **)base;
		if (s == NULL)
			return hash_buf(hashval, "", 0);
		return hash_buf(hashval, s, strlen(s));
	}

	case CAFE_TYPE_BLOB:
		return hash_buf(hashval,
				((cafe_any_value *) base)->blob.p,
				((cafe_any_value *) base)->blob.l);

	default:
		/* coverity[array_null] */
		assert(!"unexpected cafe_value_type in hashValue()");
	}
	return hashval * 17 + v;
}

static hashval_t hashKeyValues(const struct cafe_dict_struct *dict,
			       const void *base)
{
	unsigned int i;
	hashval_t hashval = 0;
	struct cafe_vm_attr *attr;

	assert(dict != NULL);
	if (!IS_TYPE_VALID(dict))
	        return 0;

	for (i = 0; i < dict->type->elements; i++) {
		attr = dict->type->keymap[i];
		hashval = hashValue(hashval, cafe_attr_type(attr),
				    ((char *)base) + attr->offset);
	}
	return hashval;
}

static hashval_t hashKey(hashval_t hashval, const struct cafe_value_type *type,
			 const cafe_any_value *key)
{
	hashval_t v = 0;

	type = cafe_base_type(type);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
		v = (char)key->ll;
		break;

	case CAFE_TYPE_UCHAR:
		v = (unsigned char)key->ull;
		break;

	case CAFE_TYPE_SHORT:
		v = (short)key->ll;
		break;

	case CAFE_TYPE_USHORT:
		v = (unsigned short)key->ull;
		break;

	case CAFE_TYPE_INT:
		v = (int)key->ll;
		break;

	case CAFE_TYPE_UINT:
		v = (unsigned int)key->ull;
		break;

	case CAFE_TYPE_LONG:
		v = (long)key->ll;
		break;

	case CAFE_TYPE_ULONG:
		v = (unsigned long)key->ull;
		break;

	case CAFE_TYPE_LONGLONG:
		v = (hashval_t)key->ll;
		break;

	case CAFE_TYPE_ULONGLONG:
		v = (hashval_t)key->ull;
		break;

	case CAFE_TYPE_BOOL:
		v = (bool)key->ll;
		break;

	case CAFE_TYPE_FLOAT:
		v = cafe_float_hash((float)key->d);
		break;

	case CAFE_TYPE_DOUBLE:
		v = cafe_double_hash(key->d);
		break;

	case CAFE_TYPE_STRUCT:
	case CAFE_TYPE_ARRAY:
	case CAFE_TYPE_DYNARRAY:
		return hashValue(hashval, type, key->p);

	case CAFE_TYPE_STRING: {
		char *s = key->str;
		if (s == NULL)
			return hash_buf(hashval, "", 0);
		return hash_buf(hashval, s, strlen(s));
	}

	case CAFE_TYPE_BLOB:
		return hash_buf(hashval, key->blob.p, key->blob.l);

	default:
		/* coverity[array_null] */
		assert(!"unexpected cafe_value_type in hashKey()");
	}
	return hashval * 17 + v;
}

static hashval_t hashKeys(const struct cafe_dict_struct *dict,
			  const cafe_any_value *keys)
{
	unsigned int i;
	hashval_t hashval = 0;
	struct cafe_vm_attr *attr;

	assert(dict != NULL);
        if (!IS_TYPE_VALID(dict))
                return 0;

	for (i = 0; i < dict->type->elements; i++) {
		attr = dict->type->keymap[i];
		hashval = hashKey(hashval, cafe_attr_type(attr), keys + i);
	}
	return hashval;
}

bool cafe_same_value(const struct cafe_value_type *type,
		     const void *base1, const void *base2)
{
	unsigned int i;
	struct cafe_vm_attr *attr;

	type = cafe_base_type(type);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
		return *(char *)base1 == *(char *)base2;

	case CAFE_TYPE_UCHAR:
		return *(unsigned char *)base1 == *(unsigned char *)base2;

	case CAFE_TYPE_SHORT:
		return *(short *)base1 == *(short *)base2;

	case CAFE_TYPE_USHORT:
		return *(unsigned short *)base1 == *(unsigned short *)base2;

	case CAFE_TYPE_INT:
		return *(int *)base1 == *(int *)base2;

	case CAFE_TYPE_UINT:
		return *(unsigned int *)base1 == *(unsigned int *)base2;

	case CAFE_TYPE_LONG:
		return *(long *)base1 == *(long *)base2;

	case CAFE_TYPE_ULONG:
		return *(unsigned long *)base1 == *(unsigned long *)base2;

	case CAFE_TYPE_LONGLONG:
		return *(long long *)base1 == *(long long *)base2;

	case CAFE_TYPE_ULONGLONG:
		return *(unsigned long long *)base1 ==
		    *(unsigned long long *)base2;

	case CAFE_TYPE_BOOL:
		return *(bool *)base1 == *(bool *)base2;

	case CAFE_TYPE_FLOAT:
		return cafe_float_same_value(base1, base2);

	case CAFE_TYPE_DOUBLE:
		return cafe_double_same_value(base1, base2);

	case CAFE_TYPE_STRUCT:
		cafe_type_attr_foreach(attr, type) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			if (!cafe_same_value(cafe_attr_type(attr),
					     ((char *)base1) + attr->offset,
					     ((char *)base2) + attr->offset))
				return false;
		}
		return true;

	case CAFE_TYPE_ARRAY:
		for (i = 0; i < type->elements; i++) {
			if (!cafe_same_value(cafe_element_type(type), base1,
					     base2))
				return false;
			base1 = ((char *)base1) + cafe_element_type(type)->size;
			base2 = ((char *)base2) + cafe_element_type(type)->size;
		}
		return true;

	case CAFE_TYPE_DYNARRAY:{
			const struct cafe_vm_dynarray *a1 = base1;
			const struct cafe_vm_dynarray *a2 = base2;
			if (a1->l != a2->l)
				return 0;
			base1 = a1->p;
			base2 = a2->p;
			for (i = 0; i < a1->l; i++) {
				if (!cafe_same_value(cafe_element_type(type),
						     base1, base2))
					return false;
				base1 = ((char *)base1) + cafe_element_type(type)->size;
				base2 = ((char *)base2) + cafe_element_type(type)->size;
			}
			return true;
		}

	case CAFE_TYPE_STRING: {
		const char *a = *(char **)base1 == NULL ? "" : *(char **)base1;
		const char *b = *(char **)base2 == NULL ? "" : *(char **)base2;
		return strcmp(a, b) == 0;
	}

	case CAFE_TYPE_BLOB:{
		size_t l = ((cafe_any_value *) base1)->blob.l;
		if (l != ((cafe_any_value *) base2)->blob.l)
			return false;
		return memcmp(((cafe_any_value *) base1)->blob.p,
			      ((cafe_any_value *) base2)->blob.p, l) == 0;
	}

	default:
		/* coverity[array_null] */
		assert(!"unexpected cafe_value_type in cafe_same_value()");
		return true;
	}
}

static bool sameKeyValues(const struct cafe_dict_struct *dict,
			  const void *base1, const void *base2)
{
	unsigned int i;
	struct cafe_vm_attr *attr;

	assert(dict != NULL);
        if (!IS_TYPE_VALID(dict))
                return false;

	for (i = 0; i < dict->type->elements; i++) {
		attr = dict->type->keymap[i];
		if (!cafe_same_value(cafe_attr_type(attr),
				     ((char *)base1) + attr->offset,
				     ((char *)base2) + attr->offset))
			return false;
	}
	return true;
}

static bool sameKey(const struct cafe_value_type *type, const void *base,
		    const cafe_any_value *key)
{
	type = cafe_base_type(type);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
		return *(char *)base == (char)key->ll;

	case CAFE_TYPE_UCHAR:
		return *(unsigned char *)base == (unsigned char)key->ull;

	case CAFE_TYPE_SHORT:
		return *(short *)base == (short)key->ll;

	case CAFE_TYPE_USHORT:
		return *(unsigned short *)base == (unsigned short)key->ull;

	case CAFE_TYPE_INT:
		return *(int *)base == (int)key->ll;

	case CAFE_TYPE_UINT:
		return *(unsigned int *)base == (unsigned int)key->ull;

	case CAFE_TYPE_LONG:
		return *(long *)base == (long)key->ll;

	case CAFE_TYPE_ULONG:
		return *(unsigned long *)base == (unsigned long)key->ull;

	case CAFE_TYPE_LONGLONG:
		return *(long long *)base == (long long)key->ll;

	case CAFE_TYPE_ULONGLONG:
		return *(unsigned long long *)base ==
		    (unsigned long long)key->ull;

	case CAFE_TYPE_BOOL:
		return *(bool *)base == (bool)key->ll;

	case CAFE_TYPE_FLOAT:
		return cafe_float_same_key(base, key);

	case CAFE_TYPE_DOUBLE:
		return cafe_double_same_key(base, key);

	case CAFE_TYPE_STRUCT:
	case CAFE_TYPE_ARRAY:
	case CAFE_TYPE_DYNARRAY:
		return cafe_same_value(type, base, key->p);

	case CAFE_TYPE_STRING:
		if (*(char **)base == key->str)
			return true;
		if (*(char **)base == NULL)
			return false;
		return strcmp(*(char **)base,
			      key->str == NULL ? "" : key->str) == 0;

	case CAFE_TYPE_BLOB:{
		size_t l = ((cafe_any_value *) base)->blob.l;
		if (l != key->blob.l)
			return false;
		return memcmp(((cafe_any_value *) base)->blob.p,
			      key->blob.p, l) == 0;
	}

	default:
		/* coverity[array_null] */
		assert(!"unexpected cafe_value_type in sameKey()");
		return true;
	}
}

static bool sameKeys(const struct cafe_dict_struct *dict,
		     const struct cafe_value_data *p,
		     const cafe_any_value *keys)
{
	unsigned int i;
	struct cafe_vm_attr *attr;

	assert(dict != NULL);
        if (!IS_TYPE_VALID(dict))
                return false;

	for (i = 0; i < dict->type->elements; i++) {
		attr = dict->type->keymap[i];
		if (!sameKey(cafe_attr_type(attr), p->u.data + attr->offset,
			     keys + i)) {
			return false;
		}
	}
	return true;
}

bool cafe_value_copy(const struct cafe_value_type *type,
		     void *base1, const void *base2,
		     const struct cafe_allocator *allocator)
{
	unsigned int i;
	struct cafe_vm_attr *attr;

	type = cafe_base_type(type);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
		*(char *)base1 = *(char *)base2;
		break;

	case CAFE_TYPE_UCHAR:
		*(unsigned char *)base1 = *(unsigned char *)base2;
		break;

	case CAFE_TYPE_SHORT:
		*(short *)base1 = *(short *)base2;
		break;

	case CAFE_TYPE_USHORT:
		*(unsigned short *)base1 = *(unsigned short *)base2;
		break;

	case CAFE_TYPE_INT:
		*(int *)base1 = *(int *)base2;
		break;

	case CAFE_TYPE_UINT:
		*(unsigned int *)base1 = *(unsigned int *)base2;
		break;

	case CAFE_TYPE_LONG:
		*(long *)base1 = *(long *)base2;
		break;

	case CAFE_TYPE_ULONG:
		*(unsigned long *)base1 = *(unsigned long *)base2;
		break;

	case CAFE_TYPE_LONGLONG:
		*(long long *)base1 = *(long long *)base2;
		break;

	case CAFE_TYPE_ULONGLONG:
		*(unsigned long long *)base1 = *(unsigned long long *)base2;
		break;

	case CAFE_TYPE_BOOL:
		*(bool *)base1 = *(bool *)base2;
		break;

	case CAFE_TYPE_FLOAT:
		*(float *)base1 = *(float *)base2;
		break;

	case CAFE_TYPE_DOUBLE:
		*(double *)base1 = *(double *)base2;
		break;

	case CAFE_TYPE_STRUCT:
		cafe_type_attr_foreach(attr, type) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			if (!cafe_value_copy(cafe_attr_type(attr),
					     ((char *)base1) + attr->offset,
					     ((char *)base2) + attr->offset,
					     allocator))
				return false;
		}
		break;

	case CAFE_TYPE_ARRAY:
		for (i = 0; i < type->elements; i++) {
			if (!cafe_value_copy(cafe_element_type(type), base1,
					     base2, allocator))
				return false;
			base1 = ((char *)base1) + cafe_element_type(type)->size;
			base2 = ((char *)base2) + cafe_element_type(type)->size;
		}
		break;

	case CAFE_TYPE_DYNARRAY: {
		struct cafe_vm_dynarray *a1 = base1;
		const struct cafe_vm_dynarray *a2 = base2;
		assert(a1->c == 0);
		assert(a1->l == 0);
		assert(a1->p == NULL);
		if (a2->l == 0)
			break;
		a1->p = CAFE_ALLOC(allocator, a2->l * cafe_element_type(type)->size);
		if (a1->p == NULL)
			return false;
		a1->c = a2->l;
		base1 = a1->p;
		base2 = a2->p;
		for (i = 0; i < a2->l; i++) {
			if (!cafe_value_copy(cafe_element_type(type), base1,
					     base2, allocator))
			{
				a1->l = i;	/* Cleanup up to this point */
				return false;
			}
			base1 = ((char *)base1) + cafe_element_type(type)->size;
			base2 = ((char *)base2) + cafe_element_type(type)->size;
		}
		a1->l = i;
		break;
	}

	case CAFE_TYPE_STRING: {
		char *s = *(char **)base2;
		size_t l = s ? strlen(s) : 0;
		void *p = CAFE_ALLOC(allocator, l + 1);
		if (p == NULL)
			return false;
		*(char **)base1 = p;
		memcpy(p, s ? s : "", l + 1);
		break;
	}

	case CAFE_TYPE_BLOB:{
		size_t l = ((struct cafe_vm_blob *) base2)->l;
		void *p = CAFE_ALLOC(allocator, l);
		if (p == NULL)
			return false;
		((struct cafe_vm_blob *) base1)->l = l;
		((struct cafe_vm_blob *) base1)->p = p;
		memcpy(p, ((struct cafe_vm_blob *) base2)->p, l);
		break;
	}

	case CAFE_TYPE_DICT:
	{
		struct cafe_dict_struct *dict1 =
			(struct cafe_dict_struct *)base1;
		struct cafe_dict_struct *dict2 =
			(struct cafe_dict_struct *)base2;
		struct cafe_value_data *data1;
		struct cafe_value_data *data2;
		struct cafe_dict_iterator iter;

		data2 = cafe_dict_first(dict2, &iter);
		while (data2 != NULL) {
			int rc = cafe_value_create(data2->type, &data1);
			if (rc != CAFE_OK)
				return false;
			if (!cafe_value_copy(data2->type, data1->u.data,
					     data2->u.data, allocator))
				return false;
			if (cafe_dict_replace(dict1, data1) == NULL)
				return false;
			data2 = cafe_dict_next(dict2, &iter);
		}
		break;
	}

	default:
		/* coverity[array_null] */
		assert(!"unexpected cafe_value_type in cafe_value_copy()");
		break;
	}
	return true;
}

static bool copyKey(const struct cafe_value_type *type, void *base,
		    const cafe_any_value *key,
		    const struct cafe_allocator *allocator)
{
	type = cafe_base_type(type);
	assert(type != NULL);
	switch (type->kind) {
	case CAFE_TYPE_CHAR:
		*(char *)base = (char)key->ll;
		break;

	case CAFE_TYPE_UCHAR:
		*(unsigned char *)base = (unsigned char)key->ull;
		break;

	case CAFE_TYPE_SHORT:
		*(short *)base = (short)key->ll;
		break;

	case CAFE_TYPE_USHORT:
		*(unsigned short *)base = (unsigned short)key->ull;
		break;

	case CAFE_TYPE_INT:
		*(int *)base = (int)key->ll;
		break;

	case CAFE_TYPE_UINT:
		*(unsigned int *)base = (unsigned int)key->ull;
		break;

	case CAFE_TYPE_LONG:
		*(long *)base = (long)key->ll;
		break;

	case CAFE_TYPE_ULONG:
		*(unsigned long *)base = (unsigned long)key->ull;
		break;

	case CAFE_TYPE_LONGLONG:
		*(long long *)base = (long long)key->ll;
		break;

	case CAFE_TYPE_ULONGLONG:
		*(unsigned long long *)base = (unsigned long long)key->ull;
		break;

	case CAFE_TYPE_BOOL:
		*(bool *)base = (bool)key->ll;
		break;

	case CAFE_TYPE_FLOAT:
		*(float *)base = (float)key->d;
		break;

	case CAFE_TYPE_DOUBLE:
		*(double *)base = key->d;
		break;

	case CAFE_TYPE_STRUCT:
	case CAFE_TYPE_ARRAY:
	case CAFE_TYPE_DYNARRAY:
		return cafe_value_copy(type, base, key->p, allocator);

	case CAFE_TYPE_STRING: {
		size_t l = 0;
		void *p;
		if (key->str != NULL)
			l = strlen(key->str);
		p = CAFE_ALLOC(allocator, l + 1);
		if (p == NULL)
			return false;
		*(char **)base = p;
		if (key->str == NULL)
			*(char *)p = '\0';
		else
			memcpy(p, key->str, l + 1);
		break;
	}

	case CAFE_TYPE_BLOB:{
		size_t l = key->blob.l;
		void *p = CAFE_ALLOC(allocator, l);
		if (p == NULL)
			return 1;
		((cafe_any_value *) base)->blob.l = l;
		((cafe_any_value *) base)->blob.p = p;
		memcpy(p, key->blob.p, l);
		break;
	}

	default:
		/* coverity[array_null] */
		assert(!"unexpected cafe_value_type in copyKey()");
		break;
	}
	return true;
}

static bool copyKeys(const struct cafe_dict_struct *dict,
		     struct cafe_value_data *p,
		     const cafe_any_value *keys,
		     const struct cafe_allocator *allocator)
{
	unsigned int i;
	struct cafe_vm_attr *attr;

	assert(dict != NULL);
        if (!IS_TYPE_VALID(dict))
                return false;

	for (i = 0; i < dict->type->elements; i++) {
		attr = dict->type->keymap[i];
		if (!copyKey(cafe_attr_type(attr), p->u.data + attr->offset,
			     keys + i, allocator)) {
			return false;
		}
	}
	return true;
}

/*
 * cafe_dict_create - Create new dictionary and return handle
 *
 * <type>: type information for this dictionary.
 */
void cafe_dict_create(struct cafe_dict_struct *dict,
		      const struct cafe_value_type *type)
{
	type = cafe_base_type(type);
	assert(dict != NULL);
	assert(dict->nentries == 0);
	assert(dict->nbuckets == 0);
	assert(dict->type == NULL);
	assert(dict->buckets == NULL);
	assert(type != NULL);
	assert(type->kind == CAFE_TYPE_DICT);
	assert(cafe_element_type(type) != NULL);
	assert(cafe_base_type(cafe_element_type(type)) != NULL);
	assert(cafe_base_type(cafe_element_type(type))->kind == CAFE_TYPE_STRUCT);
	assert(type->elements > 0);
	assert(type->keymap != NULL);

	dict->type = type;
	assert(dict->type->ident);
	assert(dict->type->ident->ns);
	assert(dict->type->ident->ns->allocator);
}

/*
 * cafe_dict_delete - Delete dictionary
 *
 * <dict>: handle to dictionary to delete
 */
void cafe_dict_delete(struct cafe_dict_struct *dict)
{
	cafe_dict_reset(dict);
	dict->type = NULL;
}

/*
 * cafe_dict_reset - Reset dictionary
 *
 * <dict>: handle to dictionary to reset
 */
void cafe_dict_reset(struct cafe_dict_struct *dict)
{
	size_t i;
	size_t entries;
	struct cafe_dict_entry *p;
	struct cafe_dict_entry *next;
	const struct cafe_allocator *allocator;

	assert(dict != NULL);
	assert(dict->type != NULL);
	allocator = dict->type->ident->ns->allocator;
	if (dict->nbuckets != 0) {
		assert(dict->buckets != NULL);
		if ((entries = dict->nentries) != 0) {
			for (i = 0; i < dict->nbuckets; i++) {
				next = dict->buckets[i];
				while ((p = next) != NULL) {
					next = p->next;
					cafe_value_delete(p->ref);
					CAFE_FREE(allocator, p);
					entries--;
				}
			}
			assert(entries == 0);
			dict->nentries = 0;
		}
		CAFE_FREE(allocator, dict->buckets);
		dict->buckets = NULL;
		dict->nbuckets = 0;
	} else {
		assert(dict->nentries == 0);
		assert(dict->buckets == NULL);
	}
}

/*
 * cafe_dict_lookup - Lookup entry in dictionary, returns NULL if not found,
 * otherwise returns pointer to data
 *
 * <dict>: handle to dictionary to search
 *
 * <keys>: array of key values.  The number of keys must match the
 * dictionary type declaration passed to cafe_dict_create.
 */
struct cafe_value_data *cafe_dict_lookup(const struct cafe_dict_struct *dict,
					 const cafe_any_value *keys)
{
	unsigned long hashval;
	struct cafe_dict_entry *p;

	assert(dict != NULL);
	assert(keys != NULL);
	if (dict->nbuckets == 0)
		return NULL;
	hashval = hashKeys(dict, keys);
	if (invalidHashval)
	    return NULL;

	p = dict->buckets[hashval % dict->nbuckets];
	while (p != NULL) {
		if (sameKeys(dict, p->ref, keys))
			return p->ref;
		p = p->next;
	}
	return NULL;
}

/*
 * cafe_dict_resize - Resize dictionary (double the number of buckets)
 *
 * <dict>: handle to dictionary to resize
 */
static void cafe_dict_resize(struct cafe_dict_struct *dict)
{
	struct cafe_dict_struct new_dict;
	const struct cafe_allocator *allocator;
	struct cafe_dict_entry *p;
	unsigned int i;

	allocator = dict->type->ident->ns->allocator;

	memset(&new_dict, 0, sizeof(new_dict));
	cafe_dict_create(&new_dict, dict->type);
	new_dict.nbuckets = dict->nbuckets * 2;
	while (dict->nentries > new_dict.nbuckets * 2) {
		new_dict.nbuckets *= 2;
	}
	new_dict.buckets = CAFE_ALLOC(allocator, new_dict.nbuckets * sizeof *new_dict.buckets);
	if (new_dict.buckets == NULL) {
		return;
	}
	memset(new_dict.buckets, 0, new_dict.nbuckets * sizeof *new_dict.buckets);

	/* Re-insert cafe_dict_entries into new buckets */
	for (i = 0; i < dict->nbuckets; ++i) {
		unsigned long hashval;
		struct cafe_dict_entry **pp;

		p = dict->buckets[i];
		if (p == NULL)
			continue;
		do {
			hashval = hashKeyValues(dict, p->ref->u.data);
			if (invalidHashval)
		                return;

			pp = &new_dict.buckets[hashval % new_dict.nbuckets];
			while (*pp != NULL) {
				assert (!sameKeyValues(dict, (*pp)->ref->u.data, p->ref->u.data));
				pp = &(*pp)->next;
			}
			*pp = p;
			p = p->next;
			(*pp)->next = NULL;
			pp = &(*pp)->next;
		} while (p != NULL);
	}
	/* Free old buckets */
	CAFE_FREE(allocator, dict->buckets);
	dict->nbuckets = new_dict.nbuckets;
	dict->buckets = new_dict.buckets;
}

/*
 * cafe_dict_get - Lookup entry in dictionary, allocates new entry if not
 * found, returns pointer to data
 *
 * <dict>: handle to dictionary to search
 *
 * <keys>: array of key values.  The number of keys must match the
 * dictionary type declaration passed to cafe_dict_create.
 */
struct cafe_value_data *cafe_dict_get(struct cafe_dict_struct *dict,
				      const cafe_any_value *keys)
{
	unsigned long hashval;
	struct cafe_dict_entry **pp;
	struct cafe_dict_entry *p;
	struct cafe_value_data *d;
	const struct cafe_allocator *allocator;

	assert(dict != NULL);
	assert(keys != NULL);
	allocator = dict->type->ident->ns->allocator;
	hashval = hashKeys(dict, keys);
        if (invalidHashval)
                return NULL;

	if (dict->nbuckets != 0) {
		pp = &dict->buckets[hashval % dict->nbuckets];
		p = *pp;
		while (p != NULL) {
			if (sameKeys(dict, p->ref, keys))
				return p->ref;
			p = p->next;
		}
	} else {
		dict->nbuckets = 64;
		dict->buckets = CAFE_ALLOC(allocator, dict->nbuckets * sizeof *dict->buckets);
		if (dict->buckets == NULL) {
			dict->nbuckets = 0;
			return NULL;
		}
		memset(dict->buckets, 0,
		       dict->nbuckets * sizeof *dict->buckets);
		pp = &dict->buckets[hashval % dict->nbuckets];
	}
	if ((p = CAFE_ALLOC(allocator, sizeof *p)) == NULL)
		return NULL;
	if (cafe_value_create(cafe_element_type(dict->type), &d) != CAFE_OK) {
		CAFE_FREE(allocator, p);
		return NULL;
	}
	p->ref = d;
	if (!copyKeys(dict, p->ref, keys, dict->type->ident->ns->allocator))
		return NULL;

	p->next = *pp;
	dict->nentries++;
	*pp = p;
	if (dict->nentries > dict->nbuckets * 2) {
		cafe_dict_resize(dict);
	}
	return d;
}

/*
 * cafe_dict_replace - Replace or add an entry in dictionary
 *
 * <dict>: handle to dictionary to search
 *
 * <d>: data for dict entry
 */
struct cafe_value_data *cafe_dict_replace(struct cafe_dict_struct *dict,
					  struct cafe_value_data *d)
{
	unsigned long hashval;
	struct cafe_dict_entry **pp;
	struct cafe_dict_entry *p;
	const struct cafe_allocator *allocator;

	assert(dict != NULL);
	assert(d != NULL);
	if (cafe_element_type(dict->type) != d->type) {
		cafe_value_delete(d);
		return NULL;
	}
	allocator = dict->type->ident->ns->allocator;
	hashval = hashKeyValues(dict, d->u.data);
	if (invalidHashval)
                return NULL;

	if (dict->nbuckets != 0) {
		pp = &dict->buckets[hashval % dict->nbuckets];
		p = *pp;
		while (p != NULL) {
			if (sameKeyValues(dict, p->ref->u.data, d->u.data)) {
				cafe_value_delete(p->ref);
				p->ref = d;
				return d;
			}
			p = p->next;
		}
	} else {
		dict->nbuckets = 64;
		dict->buckets = CAFE_ALLOC(allocator, dict->nbuckets * sizeof *dict->buckets);
		if (dict->buckets == NULL) {
			dict->nbuckets = 0;
			cafe_value_delete(d);
			return NULL;
		}
		memset(dict->buckets, 0,
		       dict->nbuckets * sizeof *dict->buckets);
		pp = &dict->buckets[hashval % dict->nbuckets];
	}
	if ((p = CAFE_ALLOC(allocator, sizeof *p)) == NULL) {
		cafe_value_delete(d);
		return NULL;
	}
	p->ref = d;
	p->next = *pp;
	dict->nentries++;
	*pp = p;
	if (dict->nentries > dict->nbuckets * 2) {
		cafe_dict_resize(dict);
	}
	return d;
}

/*
 * cafe_dict_clear - Clear entry from dictionary
 *
 * <dict>: handle to dictionary to search
 *
 * <keys>: array of key values.  The number of keys must match the
 * dictionary type declaration passed to cafe_dict_create.
 */
void cafe_dict_clear(struct cafe_dict_struct *dict, const cafe_any_value *keys)
{
	unsigned long hashval;
	struct cafe_dict_entry **pp;
	struct cafe_dict_entry *p;
	const struct cafe_allocator *allocator;

	assert(dict != NULL);
	assert(keys != NULL);
	if (dict->nbuckets == 0) {
		return;
	}
	allocator = dict->type->ident->ns->allocator;
	hashval = hashKeys(dict, keys);
	if (invalidHashval)
                return;

	pp = &dict->buckets[hashval % dict->nbuckets];
	while ((p = *pp) != NULL) {
		if (sameKeys(dict, p->ref, keys))
			break;
		pp = &p->next;
	}
	if (p == NULL)
		return;
	*pp = p->next;
	cafe_value_delete(p->ref);
	CAFE_FREE(allocator, p);
	dict->nentries--;
}

/*
 * cafe_dict_remove - Remove an existing entry in dictionary
 *
 * <dict>: handle to dictionary to search
 *
 * <d>: data of dict entry to remove
 */
void cafe_dict_remove(struct cafe_dict_struct *dict,
		      struct cafe_value_data *d)
{
	unsigned long hashval;
	struct cafe_dict_entry **pp;
	struct cafe_dict_entry *p;
	const struct cafe_allocator *allocator;

	assert(dict != NULL);
	assert(d != NULL);
	assert(dict->nbuckets != 0);
	assert(cafe_element_type(dict->type) == d->type);

	allocator = dict->type->ident->ns->allocator;
	hashval = hashKeyValues(dict, d->u.data);
        if (invalidHashval)
                return;

	pp = &dict->buckets[hashval % dict->nbuckets];
	while ((p = *pp) != NULL) {
		if (sameKeyValues(dict, p->ref->u.data, d->u.data))
			break;
		pp = &p->next;
	}
	assert(p != NULL);
	*pp = p->next;
	cafe_value_delete(p->ref);
	CAFE_FREE(allocator, p);
	dict->nentries--;
}

/*
 * cafe_dict_first - lookup and return first entry in dictionary and
 * initialize iterator, returns NULL if no more entries are available.
 *
 * <dict>: handle to dictionary to iterate over
 *
 * <iter>: opaque data structure used by cafe_dict_first and cafe_dict_next iterate
 */
struct cafe_value_data *cafe_dict_first(const struct cafe_dict_struct *dict,
					struct cafe_dict_iterator *iter)
{
	size_t i;
	struct cafe_dict_entry *p = NULL;

	assert(dict != NULL);
	assert(iter != NULL);
	for (i = 0; i < dict->nbuckets; i++) {
		if ((p = dict->buckets[i]) != NULL)
			break;
	}
	if (p == NULL)
		return NULL;
	iter->nbuckets = dict->nbuckets;
	iter->bucket = i;
	iter->next = p->next;
	return p->ref;
}

/*
 * cafe_dict_next - lookup and return next entry in dictionary, returns NULL
 * if no more entries are available.
 *
 * <dict>: handle to dictionary to iterate over
 *
 * <iter>: opaque data structure used by cafe_dict_first and cafe_dict_next iterate
 */
struct cafe_value_data *cafe_dict_next(const struct cafe_dict_struct *dict,
				       struct cafe_dict_iterator *iter)
{
	size_t i;
	struct cafe_dict_entry *p;

	assert(dict != NULL);
	assert(iter != NULL);
	assert(dict->nbuckets == iter->nbuckets);
	if ((p = iter->next) == NULL) {
		for (i = iter->bucket + 1; i < dict->nbuckets; i++) {
			if ((p = dict->buckets[i]) != NULL)
				break;
		}
		if (p == NULL)
			return NULL;
		iter->bucket = i;
	}
	iter->next = p->next;
	return p->ref;
}

cafe_hashval_t cafe_dict_hash(
	struct cafe_dict_struct *dict,
	struct cafe_value_data *d)
{
	return hashKeyValues(dict, d->u.data);
}

cafe_hashval_t cafe_hash_value(
	const struct cafe_value_type *type,
	const void *value)
{
	return hashValue(0, type, value);
}
