/* log_buffer_diagnostics.c -  */

/* Copyright (c) 2009-2010, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

#include <stdio.h>
#include <ctype.h>  /* for isprint() */
#include <cafe/inttypes.h>
#include <cafe/log_buffer.h>

#include "log_buffer_diagnostics.h"


/* Don't check return code of printf and fputs (fix Prevent warning) */

#define CAFE_PRINTF     (void)printf

/**
 *  Emit the most basic statistics of the given log buffer to stdout.
 */
void cafe_dump_log_header(const struct cafe_log_buf *log)
{
	CAFE_PRINTF("log @ %p:\n", log);
	CAFE_PRINTF("\tsize: %lu\n", (unsigned long)log->size);
	CAFE_PRINTF("\tdata: %p\n", log->data);
	CAFE_PRINTF("\tread_head: %p\n", log->read_head);
	CAFE_PRINTF("\twrite_head: %p\n", log->write_head);
	CAFE_PRINTF("\ttrailing_top: %p\n", log->trailing_top);
}


/**
 *  Emit the byte-by-byte contents of the given log buffer to stdout.
 */
void cafe_dump_log_contents(const struct cafe_log_buf *log)
{
	size_t i;

	for (i = 0; i < log->size; ++i) {
		CAFE_PRINTF("\t\t%p: %hhx", log->data + i,
		       *(int8_t *) (log->data + i));
		if (isprint(log->data[i]))
			CAFE_PRINTF(" \"%c\"", log->data[i]);
		if (log->read_head == log->data + i)
			CAFE_PRINTF("\t<---- read");

		if (log->write_head == log->data + i)
			CAFE_PRINTF("\t<---- write");
		CAFE_PRINTF("\n");
	}
}


/**
 *  Emit everything we know about the given log buffer to stdout.
 */
void cafe_dump_log(const struct cafe_log_buf *log)
{
	cafe_dump_log_header(log);
	cafe_dump_log_contents(log);
}


/**
 *  Run basic sanity checks, returning false iff we can see for a fact the
 * given log is invalid.
 */
bool cafe_log_validate(const struct cafe_log_buf *const log)
{
	const char *const log_max = log->data + log->size;

	if ((log->read_head < log->data) || (log->read_head >= log_max)) {
		CAFE_PRINTF("Read head out of bounds (%p) (%p, %p)\n",
		       log->read_head, log->data, log_max);
		return false;
	}

	if ((log->write_head < log->data) || (log->write_head >= log_max)) {
		CAFE_PRINTF("Write head out of bounds (%p) (%p, %p)\n",
		       log->write_head, log->data, log_max);
		return false;
	}

	if ((log->trailing_top != 0) &&
	    ((log->trailing_top < log->data)
	     || (log->trailing_top > log_max))) {
		CAFE_PRINTF("Trailing top out of bounds (%p) (%p, %p)\n",
		       log->trailing_top, log->data, log_max);
		return false;
	}

	/* now make sure the read and write heads don't exceed the trailing
	   top */
	if ((log->trailing_top != 0) && (log->read_head >= log->trailing_top))
	{
		CAFE_PRINTF("Read head past trailing top (%p) (%p, %p)\n",
		       log->read_head, log->data, log->trailing_top);
		return false;
	}

	if ((log->trailing_top != 0) && (log->write_head >= log->trailing_top))
	{
		CAFE_PRINTF("Write head past trailing top (%p) (%p, %p)\n",
		       log->write_head, log->data, log->trailing_top);
		return false;
	}

	return true;
}
