/* test_streams.c - test log_stream routines */

/* Copyright (c) 2010-2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <assert.h>
#include <limits.h>
#include <string.h>  /* for memcmp() */
#include "../cafe_os.h"  /* for getpagesize() */
#include <cafe/log_stream.h>
#include <test/cafe_CUnitTest.h>
#include "../log_buffer_diagnostics.h"
#include "../leb128.h"
#include "../time_utils.h"


static
int test_dynarray_marshall(void)
{
	char buf[200];
	unsigned long data[20];
	struct cafe_vm_dynarray array;
	struct cafe_vm_dynarray *verify_array;
	void *arg[1];
	struct cafe_event_descr event;
	struct cafe_namespace *ns;
	size_t size_static;
	size_t size_variable;
	unsigned int i;

	CUT_ASSERT_CAFE_OK(cafe_namespace_create(cafe_stdlib_allocator, &ns));
	event.num_types = 1;
	CUT_ASSERT_CAFE_OK(cafe_type_parse(ns, "S[test,v[dynarray,a[L]]]",
					   &event.types[0]));

	/* initialize our data */
	array.l = 10;
	array.p = data;
	for (i = 0; i < 10; ++i)
		data[i] = 0xdeadbeef + i;
	arg[0] = &array;

	/* write the array to the buffer */
	size_static = calc_entry_size_static(&event);
	CUT_ASSERT(size_static > sizeof(struct cafe_vm_dynarray));
	size_variable = calc_entry_size_variable(&event, arg);
	CUT_ASSERT_EQUAL(size_variable, array.l * sizeof(long));
	cafe_write_event(buf, 42, 0xdeadbeef, &event, arg,
			 size_static, size_static + size_variable);

	/* validate written data */
	verify_array = (struct cafe_vm_dynarray *)
		&buf[sizeof(long) + sizeof(long) + sizeof(size_t)];
	CUT_ASSERT_EQUAL(array.l, verify_array->l);
	CUT_ASSERT_EQUAL(0, memcmp(array.p, verify_array->p,
				   array.l * sizeof(long)));

	cafe_namespace_delete(ns);
	return 0;
}

#if 0
/**
 * Make sure logstream create routines succeed and fail as expected.
 */
static
int test_logstream_create(void)
{
	struct cafe_output_stream *ostream;

	CUT_ASSERT_EQUAL(CAFE_ERROR,
			 logstream_open_output(2, cafe_stdlib_allocator,
					       &ostream));
	CUT_ASSERT_EQUAL(CAFE_NOMEM,
			 logstream_open_output(SIZE_MAX, cafe_stdlib_allocator,
					       &ostream));
	CUT_ASSERT_CAFE_OK(logstream_open_output(64, cafe_stdlib_allocator,
						 &ostream));

	CAFE_FREE(cafe_stdlib_allocator, ostream);
	return 0;
}
#endif


struct writer_thread_args {
	struct cafe_output_stream *ostream;
	long long max;
	unsigned long long umax;
	bool abort;
};


static
void *writer_thread(void *arg)
{
	struct writer_thread_args *wta = arg;
	unsigned long long ui;
	long long i;

	for (ui = 0; ui < wta->umax; ++ui) {
		if (wta->abort)
			return NULL;
		cafe_leb128_swrite_ull(wta->ostream, ui);
	}

	for (i = -wta->max; i < wta->max; ++i) {
		if (wta->abort)
			return NULL;
		cafe_leb128_swrite_ll(wta->ostream, i);
	}

	return NULL;
}


static
unsigned long long read_uval(struct cafe_log_buf *log)
{
	unsigned long long val;
	char buf[64];
	char *msg;
	size_t size;
	char *trav;
	size_t i = 0;

	memset(buf, 0, sizeof(buf));
	do {
		msg = cafe_log_read_reserve(log, &size);
		while (size == 0) {
			struct timespec delay;
			timespec_from_ticks(5000, NSECS_PER_SEC, &delay);
			nanosleep(&delay, 0);
			msg = cafe_log_read_reserve(log, &size);
		}

		if (size > sizeof(buf) - i)
			size = sizeof(buf) - i;

		memcpy(&buf[i], msg, size);
		trav = cafe_leb128_mread_ull(buf, &val);
		if ((trav - buf) <= (int)(i + size)) {
			cafe_log_read_commit(log, trav - &buf[i]);
			return val;
		}
		i += size;
		cafe_log_read_commit(log, size);
	} while(1);
}


static
long long read_val(struct cafe_log_buf *log)
{
	long long val;
	char buf[64];
	char *msg;
	size_t size;
	char *trav;
	size_t i = 0;

	memset(buf, 0, sizeof(buf));
	do {
		/* read as many bytes as possible into a local buffer, then
		   attempt to decode the buffered bytes.  If the decode routine
		   uses more bytes than we've buffered, then read more bytes
		   from the log_buffer until we're able to completely decode
		   the value */
		msg = cafe_log_read_reserve(log, &size);
		while (size == 0) {
			struct timespec delay;
			timespec_from_ticks(5000, NSECS_PER_SEC, &delay);
			nanosleep(&delay, 0);
			msg = cafe_log_read_reserve(log, &size);
		}

		if (size > sizeof(buf) - i)
			size = sizeof(buf) - i;

		memcpy(&buf[i], msg, size);
		trav = cafe_leb128_mread_ll(buf, &val);
		if ((trav - buf) <= (int)(i + size)) {
			cafe_log_read_commit(log, trav - &buf[i]);
			return val;
		}
		i += size;
		cafe_log_read_commit(log, size);
	} while(1);
}


/**
 *  Make sure read and writes behave as expected, and validating the basic
 * leb128 encoding/decoding of integrals.
 */
static
int test_logstream_concurrency(void)
{
	struct cafe_log_buf *buf;
	struct writer_thread_args thread_args;
	char *msg;
	size_t size, size2;
	pthread_t writer;
	long long i;
	unsigned long long ui;
	struct timespec delay;

	/* seed the number generator */
	if (clock_gettime(CLOCK_REALTIME, &delay))
		perror("error seeding with time");
	else
		srand((unsigned int)delay.tv_nsec);

	do {
		/* limit ourselves to using ~ a page */
		size = rand() % getpagesize();
	} while (size <= 1);

	buf = cafe_log_create(size, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(buf);
	CUT_ASSERT_CAFE_OK(cafe_logstream_open_output(buf, cafe_stdlib_allocator,
						      &thread_args.ostream,
						      NULL, NULL));

	/* Make sure that the buffer will get full at least once, but
	 * limit the number of value to keep the test fast. */
	do {
		size *= 2;
	} while(size < 3000);

	thread_args.max = size;
	thread_args.umax = size;
	thread_args.abort = false;

	/* spawn the writer thread */
	if (pthread_create(&writer, 0, writer_thread, &thread_args)) {
		perror("failed to create writer thread");
		CUT_FAIL("failed to create writer thread");
	}

	/* wait until writer blocks. */
	//td: would be better to do this deterministically
	timespec_from_ticks(1, 1, &delay);
	nanosleep(&delay, 0);
	cafe_log_read_reserve(buf, &size);
	msg = cafe_log_read_reserve(buf, &size2);
	CUT_ASSERT_NOT_NULL(msg);
	CUT_ASSERT(size2 != 0);
	CUT_ASSERT(size == size2);

	/* start validating data */
	for (ui = 0; ui < thread_args.umax; ++ui) {
		unsigned long long val = read_uval(buf);
		if (val != ui) {
			thread_args.abort = true;
			if (pthread_join(writer, NULL))
				perror("Error waiting for the writer thread");
			cafe_dump_log(buf);
			printf("%llu != %llu\n", val, ui);
		}
		CUT_ASSERT_EQUAL(val, ui);
	}

	for (i = -thread_args.max; i < thread_args.max; ++i) {
		long long val = read_val(buf);
		if (val != i) {
			thread_args.abort = true;
			if (pthread_join(writer, NULL))
				perror("Error waiting for the writer thread");
			cafe_dump_log(buf);
			printf("%lld != %lld\n", val, i);
		}
		CUT_ASSERT_EQUAL(val, i);
	}

	if (pthread_join(writer, NULL))
		perror("Error waiting for the writer thread");

	CAFE_FREE(cafe_stdlib_allocator, buf);
	CAFE_FREE(cafe_stdlib_allocator, thread_args.ostream);
	return 0;
}


CUT_BEGIN_TEST_HARNESS(stream_suite)
	CUT_RUN_TEST(test_dynarray_marshall);
	CUT_RUN_TEST(test_logstream_concurrency);
CUT_END_TEST_HARNESS
