# Test macros taken from the chasm project
# The following macros are defined:
#   phoenix_build_test
#   phoenix_make_test
# Their syntax is:
#   phoenix_build_test(testname libraries file1 [file2 ...])
#     The first argument is the name of the test executable to build and the
#     second is the name of a variable containing the libraries that it needs
#     to be linked against. The remaining arguments are the files that are
#     needed to build the test.
#   phoenix_make_test(testname instance [arg1 ...])
#     The first argument is the name of the test and the second is the name of
#     the instance of the test. It creates a target named
#     `test-${testname}-${instance}' which runs the test by itself as well as
#     valgrind and callgrind targets that run the test throught the respective
#     tool. Tests also have group targets based on `testname'.

set(PHOENIX_TESTING OFF CACHE BOOL "Build tests")
set(PHOENIX_VALGRIND_GENERATE_SUPPRESSIONS OFF CACHE BOOL "Output suppression rules for valgrind leak detections")
set(PHOENIX_VALGRIND_VERBOSE OFF CACHE BOOL "Make valgrind verbose")
set(PHOENIX_VALGRIND_USE_SUPPRESSIONS ON CACHE BOOL "Suppress known leaks in valgrind")

set(phoenix_valgrind_arguments)

if(PHOENIX_VALGRIND_GENERATE_SUPPRESSIONS)
  set(phoenix_valgrind_arguments
    ${phoenix_valgrind_arguments}
    "--gen-suppressions=all")
endif()

if(PHOENIX_VALGRIND_VERBOSE)
  set(phoenix_valgrind_arguments
    ${phoenix_valgrind_arguments}
    "--verbose")
endif()

if(PHOENIX_VALGRIND_USE_SUPPRESSIONS)
  set(phoenix_valgrind_arguments
    ${phoenix_valgrind_arguments})
endif()

enable_testing()

if(UNIX)
  add_custom_target(valgrind)
  add_custom_target(callgrind)
endif()

macro(phoenix_build_test testname libraries)
  add_executable(test-${testname} ${ARGN})
  target_link_libraries(test-${testname} ${${libraries}})
  if(UNIX)
    add_custom_target(tests-${testname})
    add_custom_target(valgrind-${testname})
    add_custom_target(callgrind-${testname})
    add_dependencies(valgrind valgrind-${testname})
    add_dependencies(callgrind callgrind-${testname})
  endif()
endmacro()

macro(phoenix_add_test testname instance)
  add_test(test-${testname}-${instance}
    ${CMAKE_RUNTIME_OUTPUT_DIRECTORY}/test-${testname}
    ${ARGN})
  if(UNIX)
    set_tests_properties(test-${testname}-${instance}
      PROPERTIES
      FAIL_REGULAR_EXPRESSION "^Error")
    set(tests_${testname}
      ${tests_${testname}}
      test-${testname}-${instance})
    add_custom_target(test-${testname}-${instance})
    add_custom_target(valgrind-${testname}-${instance})
    add_custom_target(callgrind-${testname}-${instance})
    add_custom_command(
      TARGET  valgrind-${testname}-${instance}
      COMMAND valgrind
      --quiet
      --leak-check=full
      --show-reachable=yes
      --log-file=${CMAKE_RUNTIME_OUTPUT_DIRECTORY}/valgrind.log.${testname}.${instance}
      ${phoenix_valgrind_arguments}
      ${CMAKE_RUNTIME_OUTPUT_DIRECTORY}/test-${testname}
      ${ARGN}
      WORKING_DIRECTORY
      ${CMAKE_BINARY_DIR}
      COMMENT "Running valgrind on test \"${testname}\" instance \"${instance}\"")
    add_dependencies(valgrind-${testname}
      valgrind-${testname}-${instance})
    add_custom_command(
      TARGET  callgrind-${testname}-${instance}
      COMMAND valgrind
      --tool=callgrind
      --dump-instr=yes
      --log-file=${CMAKE_RUNTIME_OUTPUT_DIRECTORY}/callgrind.log.${testname}.${instance}
      ${CMAKE_RUNTIME_OUTPUT_DIRECTORY}/test-${testname}
      ${ARGN}
      WORKING_DIRECTORY
      ${CMAKE_BINARY_DIR}
      COMMENT
      "Running callgrind on test \"${testname}\" instance \"${instance}\"")
    add_dependencies(callgrind-${testname}
      callgrind-${testname}-${instance})
  endif()
endmacro()
