/* Copyright (c) 2010 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
16sep14,sdt  Fixed sizeof-mismatch warnings (US45196)
*/

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include <hpc/hpc_ep.h>
#include <hpc/hpc_drv/hpcHwDrvApi.h>
#include <hpc/drv_evt_cfg/hpcDrvEvtCfg.h>
#include <hpc/drv_evt_cfg/i86/i86ArchPmEvtCfg.h>
#include <hpc/drv_evt_cfg/i86/hpcI86vxw.h>
#include <hpc/drv_evt_cfg/i86/i86ArchPm1DrvSet.h>
#include <hpc/drv_evt_cfg/i86/i86ArchPm2DrvSet.h>
#include <cafe/autogenerated/hpc_common.type.h>
#include <cafe/autogenerated/hpc_x86archpm.type.h>

#include "i86ArchPmEvtCfgP.h"
#include "../../hpc_utils.h"
#include "hpcI86PerfIsr.h"

#define NUM_ARCH_V2_EVENTS	3

static void * (*pMalloc)(size_t);
static void (*pFree)(void *);
static unsigned int cpuidPerfMonEax;
static unsigned int cpuidPerfMonEbx;
static unsigned int cpuidPerfMonEcx;
static unsigned int cpuidPerfMonEdx;

/*****************************************************************************/

/* Architecture Performance Monitoring */

/*
 * The following macros decode the result of cpuid instruction
 * where initial EAX value is 0x0A.
 */

#define PMON_VERSION_ID(_eax_)		(_eax_ & 0xff)
#define NUM_COUNTERS(_eax_)		((_eax_ & (0xff << 8)) >> 8)
#define COUNTER_BIT_SIZE(_eax_)		((_eax_ & (0xff << 16)) >> 16)
#define NUM_ARCH_EVENTS(_eax_)		((_eax_ & (0xff << 24)) >> 24)

#define ARCH_EVENTS_BITS(_ebx_)		(_ebx_ & 0x3f)
#define ARCH_CORE_CYCLES(_ebx_)		((_ebx_ & (1 << 0)) == 0)
#define ARCH_INST_RETIRED(_ebx_)	((_ebx_ & (1 << 1)) == 0)
#define ARCH_REF_CYCLES(_ebx_)		((_ebx_ & (1 << 2)) == 0)
#define ARCH_LLC_REFERENCE(_ebx_)	((_ebx_ & (1 << 3)) == 0)
#define ARCH_LLC_MISSES(_ebx_)		((_ebx_ & (1 << 4)) == 0)
#define ARCH_BRANCH_RETIRED(_ebx_)	((_ebx_ & (1 << 5)) == 0)
#define ARCH_BRANCH_MISPREDICT_RETIRED(_ebx_) ((_ebx_ & (1 << 6)) == 0)


/* Note, these two are not valid in Arch Perf Mon Version 1 */
#define NUM_FIXED_COUNTERS(_edx_)	(_edx_ & 0x1f)
#define FIXED_COUNTER_BIT_SIZE(_edx_)	((_edx_ & (0xff << 5)) >> 5)

/* Note this array must be kept in sync with archEventsDesc array. It is
 * also organized in pairs. Each pair entry contains the arch v1 named
 * event followed by a generic named event, which could be empty */
static struct i86Arch1EventDescription evtDescV1[2 * NUM_ARCH_V1_EVENTS] = {
    /* CPU_CLK_UNHALTED-CORE: Unhalted core cycles */
    {hpc_x86archpm1_CPU_CLK_UNHALTED_CORE_VT_DEF,
        {500000},
        {0x3c, 0, 0}},
    {NULL, {0}, {0, 0, 0}},    /* no common name for this event */

    /* INSTRUCTION_RETIRED: Instruction retired */
    {hpc_x86archpm1_INSTRUCTION_RETIRED_VT_DEF,
        {500000},
        {0xC0, 0, 0}},
    {NULL, {0}, {0, 0, 0}},    /* no common name for this event */

    /* CPU_CLK_UNHALTED-REF: Unhalted reference cycles Measures bus cycle */
    {hpc_x86archpm1_CPU_CLK_UNHALTED_REF_VT_DEF,
        {500000},
        {0x3c, 0x1, 0}},
    {NULL, {0}, {0, 0, 0}},    /* no common name for this event */

    /* LLC_REFERENCE: LL cache references */
    {hpc_x86archpm1_LLC_REFERENCE_VT_DEF,
        {1000},
        {0x2E, 0x4F, 0}},
    {NULL, {0}, {0, 0, 0}},    /* no common name for this event */

    /* LLC_MISSES: LL cache misses */
    {hpc_x86archpm1_LLC_MISSES_VT_DEF,
        {500},
        {0x2E, 0x41, 0}},
    {NULL, {0}, {0, 0, 0}},    /* no common name for this event */

    /* BRANCH_INSTRUCTION_RETIRED: Branch instruction retired */
    {hpc_x86archpm1_BRANCH_INSTRUCTION_RETIRED_VT_DEF,
        {1000},
        {0xC4, 0x0, 0}},
    {NULL, {0}, {0, 0, 0}},    /* no common name for this event */

    /* BRANCH_MISSES_RETIRED: Mispredicted Branch Instruction retired */
    {hpc_x86archpm1_BRANCH_MISSES_RETIRED_VT_DEF,
        {1000},
        {0xC5, 0x0, 0}},
    {NULL, {0}, {0, 0, 0}}};    /* no common name for this event */

/* Note this array must be kept in sync with archFixedEventsDesc array
 * It is also organized in pairs. Each pair entry contains the arch v2 named
 * event followed by a generic named event, which could be empty */
static struct i86Arch2EventDescription evtDescV2[2 * NUM_ARCH_V2_EVENTS] = {
    /* CPU_CLK_UNHALTED-CORE: Unhalted core cycles */
    {hpc_x86archpm2_CPU_CLK_UNHALTED_CORE_VT_DEF,
        {500000},
        {0}},
    {NULL, {0}, {0}},	/* no common name for this event */

    /* INSTRUCTION_RETIRED: Instruction retired */
    {hpc_x86archpm2_INSTRUCTION_RETIRED_VT_DEF,
        {500000},
        {1}},
    /* total_instructions */
    {hpc_total_instructions_VT_DEF,
        {500000},
        {1}},

    /* CPU_CLK_UNHALTED-REF: Unhalted reference cycles Measures bus cycle */
    {hpc_x86archpm2_CPU_CLK_UNHALTED_REF_VT_DEF,
        {500000},
        {2}},
    /* total_cycles */
    {hpc_total_cycles_VT_DEF,
        {500000},
        {2}}};

static
struct i86Arch1EventDescription v1EvtEnabledDesc[2 * NUM_ARCH_V1_EVENTS];
static
struct i86Arch2EventDescription v2EvtEnabledDesc[2 * NUM_ARCH_V2_EVENTS];

static struct driverInstArchV1 hpcHwDrvI86ArchV1;
static struct driverInstArchV2 hpcHwDrvI86ArchV2;

static struct hpcDriverConfig i86ArchV1HpcDriverConfig;
static struct hpcDriverConfig i86ArchV2HpcDriverConfig;

static int (*pArchV1Handler)(struct driverInstArchV1 *, void *, int, long long int);
static int (*pArchV2Handler)(struct driverInstArchV2 *, void *, int, long long int);

static int defaultArchV1Handler(struct driverInstArchV1 * p, void * pRegs,
				int cpu, long long int status)
    {
    return 0;
    }

static int defaultArchV2Handler(struct driverInstArchV2 * p, void * pRegs,
				int cpu, long long int status)
    {
    return 0;
    }

static int i86ArchPmHandler
    (
    void * pRegs,
    int cpu
    )
    {
    int p = 0; /* processed */
    long long int pmGlobalStatus;

    pentiumMsrGet(IA32_PERF_GLOBAL_STATUS, (long long int *)&pmGlobalStatus);

    while (pmGlobalStatus)
        {
	/* Clear the overflow */
        pentiumMsrSet(IA32_PERF_GLOBAL_OVF_CTRL,
                      (long long int *)&pmGlobalStatus);

        if (pmGlobalStatus << 32)
	    p += (*pArchV1Handler)(&hpcHwDrvI86ArchV1, pRegs, cpu,
							    pmGlobalStatus);

        if (pmGlobalStatus >> 32)
	    p += (*pArchV2Handler)(&hpcHwDrvI86ArchV2, pRegs, cpu,
							    pmGlobalStatus);
        /* check the status again */
        pentiumMsrGet(IA32_PERF_GLOBAL_STATUS,
                      (long long int *) &pmGlobalStatus);
        }

    return p;
    }

static STATUS enableArchPerf1Isr
    (
    int (*pHandler)(struct driverInstArchV1 *, void *, int, long long int)
    )
    {
    hpcI86PerfIsrHandler(0, i86ArchPmHandler);
    pArchV1Handler = pHandler;
    hpcVxwI86PerfIsrConnectVec(hpcI86PerfMonVector, hpcI86PerfMonHandler);
    hpcVxwI86PerfMonIsrEnable();
    return OK;
    }

static STATUS disableArchPerf1Isr(int force)
    {
    hpcVxwI86PerfMonIsrDisable();
    hpcVxwI86PerfIsrDisconnectVec(hpcI86PerfMonHandler, force);
    hpcI86PerfIsrHandler(0, NULL);
    pArchV1Handler = defaultArchV1Handler;
    return OK;
    }

static STATUS enableArchPerf2Isr
    (
    int (*pHandler)(struct driverInstArchV2 *, void *, int, long long int)
    )
    {
    hpcI86PerfIsrHandler(0, i86ArchPmHandler);
    pArchV2Handler = pHandler;
    hpcVxwI86PerfIsrConnectVec(hpcI86PerfMonVector, hpcI86PerfMonHandler);
    hpcVxwI86PerfMonIsrEnable();
    return OK;
    }

static STATUS disableArchPerf2Isr(int force)
    {
    hpcVxwI86PerfMonIsrDisable();
    hpcVxwI86PerfIsrDisconnectVec(hpcI86PerfMonHandler, force);
    hpcI86PerfIsrHandler(0, NULL);
    pArchV2Handler = defaultArchV2Handler;
    return OK;
    }

STATUS i86ArchPmV1Free
    (
    struct hpcDriverConfig * pConfig
    )
    {
    struct driverInstArchV1 * pInst;

    if (pConfig == NULL)
        return ERROR;

    pInst = pConfig->pDrvInst;

    if (pInst->ppDrvHwCtrData == NULL)
        return ERROR;

    (*pFree)(pInst->ppDrvHwCtrData);

    return OK;
    }

static void addArchV1Events
    (
    int *                             pNumEvents,
    int *                             pNumGenericEvents,
    struct i86Arch1EventDescription * pOutEventDesc
    )
    {
    int i;
    *pNumEvents = 0;
    *pNumGenericEvents = 0;

    for (i = 0; i < NUM_ARCH_V1_EVENTS; i++)
        {
	if ((cpuidPerfMonEbx & (1 << i)) == 0)
            {
            memcpy(pOutEventDesc, &evtDescV1[i * 2],
                    sizeof(struct i86Arch1EventDescription));
            pOutEventDesc++;
            (*pNumEvents)++;

            if (evtDescV1[(i * 2) + 1].eventVtStr != NULL)
                {
                /* copy the generic named event as well */
                memcpy(pOutEventDesc, &evtDescV1[(i * 2) + 1],
                        sizeof(struct i86Arch1EventDescription));
                (*pNumGenericEvents)++;
                pOutEventDesc++;
                }
            }
        }
    }

struct hpcDriverConfig * i86ArchPmV1Configure
    (
    int				      numCores,
    unsigned int		      numNonArchEvents,
    struct i86Arch1EventDescription * pEventDesc
    )
    {
    int i, numEvts, numGenEvts;
    size_t size;
    struct i86Arch1EventDescription * pOutEventDesc;

    if (PMON_VERSION_ID(cpuidPerfMonEax) < 1)
	    return NULL;

    if (numNonArchEvents == 0)
	pOutEventDesc = v1EvtEnabledDesc;
    else
	pOutEventDesc = pEventDesc;

    for (i = 0; i < (2 * NUM_ARCH_V1_EVENTS); i++)
        memset(&v1EvtEnabledDesc[i], 0, sizeof(struct i86Arch1EventDescription));

    addArchV1Events(&numEvts, &numGenEvts, pOutEventDesc + numNonArchEvents);

    if (NUM_COUNTERS(cpuidPerfMonEax) == 0 ||
        COUNTER_BIT_SIZE(cpuidPerfMonEax) == 0 ||
        numCores == 0)
        {
	return NULL;
        }

    numEvts += numGenEvts + numNonArchEvents;

    memset(&hpcHwDrvI86ArchV1, 0, sizeof(hpcHwDrvI86ArchV1));
    hpcHwDrvI86ArchV1.pDriverName = "I86 Arch Perf Monitor V1";
    hpcHwDrvI86ArchV1.numCounters = NUM_COUNTERS(cpuidPerfMonEax);
    hpcHwDrvI86ArchV1.numCounterBits = COUNTER_BIT_SIZE(cpuidPerfMonEax);
    hpcHwDrvI86ArchV1.pHandler = NULL;
    hpcHwDrvI86ArchV1.enablePerfIsrPerCore = enableArchPerf1Isr;
    hpcHwDrvI86ArchV1.disablePerfIsrPerCore = disableArchPerf1Isr;
    hpcHwDrvI86ArchV1.enablePerfIsr = NULL;
    hpcHwDrvI86ArchV1.disablePerfIsr = NULL;
    if (PMON_VERSION_ID(cpuidPerfMonEax) == 3)
        hpcHwDrvI86ArchV1.archPerfMonVer = I86_ARCH_PM_DRV1_V3;
    else
        hpcHwDrvI86ArchV1.archPerfMonVer = I86_ARCH_PM_DRV1_V1;
    hpcHwDrvI86ArchV1.msrPmc = IA32_PMC0;
    hpcHwDrvI86ArchV1.msrPerfEvtSel = IA32_PERFEVTSEL0;

    size = sizeof(struct hpcHwCtrData *) * hpcHwDrvI86ArchV1.numCounters *
								    numCores;
    hpcHwDrvI86ArchV1.ppDrvHwCtrData = (*pMalloc)(size);

    if (hpcHwDrvI86ArchV1.ppDrvHwCtrData == NULL)
	return NULL;

    memset(hpcHwDrvI86ArchV1.ppDrvHwCtrData, 0, size);

    i86ArchV1HpcDriverConfig.numEvents = numEvts;
    i86ArchV1HpcDriverConfig.numCpuCores = numCores;
    i86ArchV1HpcDriverConfig.pDrvInst = (void *)&hpcHwDrvI86ArchV1;
    i86ArchV1HpcDriverConfig.pFuncs = &hpcHwDrvI86Arch1Rtns;
    i86ArchV1HpcDriverConfig.pEventSettingsGet = i86ArchPm1EventSettingsGet;
    i86ArchV1HpcDriverConfig.pEventDesc = pOutEventDesc;
    i86ArchV1HpcDriverConfig.eventSettingSize =
				    sizeof(struct i86Arch1EventDescription);

    return &i86ArchV1HpcDriverConfig;
    }

static void addArchV2Events
    (
    int * pNumEvents,
    int * pNumGenericEvents,
    int numCounters
    )
    {
    int i;
    struct i86Arch2EventDescription *pEvtDesc = v2EvtEnabledDesc;

    *pNumEvents = 0;
    *pNumGenericEvents = 0;

    for (i = 0; i < numCounters; i++)
        {
        memcpy(pEvtDesc, &evtDescV2[i * 2],
                                      sizeof(struct i86Arch2EventDescription));
        pEvtDesc++;
        (*pNumEvents)++;

        if (evtDescV2[(i * 2) + 1].eventVtStr != NULL)
            {
            memcpy(pEvtDesc, &evtDescV2[(i * 2) + 1],
                                      sizeof(struct i86Arch2EventDescription));
            (*pNumGenericEvents)++;
            pEvtDesc++;
            }
        }
    }

STATUS i86ArchPmV2Free
    (
    struct hpcDriverConfig * pConfig
    )
    {
    struct driverInstArchV2 * pInst;

    if (pConfig == NULL)
        return ERROR;

    pInst = pConfig->pDrvInst;

    if (pInst->ppDrvHwCtrData == NULL)
        return ERROR;

    (*pFree)(pInst->ppDrvHwCtrData);

    return OK;
    }

struct hpcDriverConfig * i86ArchPmV2Configure
    (
    int	numCores
    )
    {
    size_t size;
    int numEvts, numGenEvts;

    if (PMON_VERSION_ID(cpuidPerfMonEax) < 2)
        return NULL;

    memset(&hpcHwDrvI86ArchV2, 0, sizeof(hpcHwDrvI86ArchV2));

    hpcHwDrvI86ArchV2.numCounters = NUM_FIXED_COUNTERS(cpuidPerfMonEdx);

    if (hpcHwDrvI86ArchV2.numCounters > NUM_ARCH_V2_EVENTS)
        return NULL;

    addArchV2Events(&numEvts, &numGenEvts, hpcHwDrvI86ArchV2.numCounters);

    if (numEvts != hpcHwDrvI86ArchV2.numCounters)
        return NULL;

    size = sizeof(struct hpcHwCtrData *) * hpcHwDrvI86ArchV2.numCounters *
								    numCores;

    hpcHwDrvI86ArchV2.ppDrvHwCtrData = (*pMalloc)(size);

    if (hpcHwDrvI86ArchV2.ppDrvHwCtrData == NULL)
	return NULL;

    memset(hpcHwDrvI86ArchV2.ppDrvHwCtrData, 0, size);

    hpcHwDrvI86ArchV2.pDriverName = "I86 Arch Perf Monitor V2";
    hpcHwDrvI86ArchV2.numCounterBits = FIXED_COUNTER_BIT_SIZE(cpuidPerfMonEdx);
    hpcHwDrvI86ArchV2.pHandler = NULL;
    hpcHwDrvI86ArchV2.enablePerfIsrPerCore = enableArchPerf2Isr;
    hpcHwDrvI86ArchV2.disablePerfIsrPerCore = disableArchPerf2Isr;
    hpcHwDrvI86ArchV2.enablePerfIsr = NULL;
    hpcHwDrvI86ArchV2.disablePerfIsr = NULL;
    if (PMON_VERSION_ID(cpuidPerfMonEax) < 3)
        hpcHwDrvI86ArchV2.archPerfMonVer = I86_ARCH_PM_DRV2_V2;
    else
        hpcHwDrvI86ArchV2.archPerfMonVer = I86_ARCH_PM_DRV2_V3;
    hpcHwDrvI86ArchV2.msrFixedCtr = IA32_FIXED_CTR0;
    hpcHwDrvI86ArchV2.msrFixedCtrCtrl = IA32_FIXED_CTR_CTRL;
    hpcHwDrvI86ArchV2.msrGlobalStatus = IA32_PERF_GLOBAL_STATUS;
    hpcHwDrvI86ArchV2.msrGlobalCtrl = IA32_PERF_GLOBAL_CTRL;
    hpcHwDrvI86ArchV2.msrGlobalOvfCtrl = IA32_PERF_GLOBAL_OVF_CTRL;

    i86ArchV2HpcDriverConfig.numEvents = numEvts + numGenEvts;
    i86ArchV2HpcDriverConfig.numCpuCores = numCores;
    i86ArchV2HpcDriverConfig.pDrvInst = (void *)&hpcHwDrvI86ArchV2;
    i86ArchV2HpcDriverConfig.pFuncs = &hpcHwDrvI86Arch2Rtns;
    i86ArchV2HpcDriverConfig.pEventSettingsGet = i86ArchPm2EventSettingsGet;
    i86ArchV2HpcDriverConfig.pEventDesc = v2EvtEnabledDesc;
    i86ArchV2HpcDriverConfig.eventSettingSize =
				 sizeof(struct i86Arch2EventDescription);

    return &i86ArchV2HpcDriverConfig;
    }

STATUS hpcI86ArchPerfMonInit
    (
    void * 	 (*pMallocFunc)(size_t),
    void 	 (*pFreeFunc)(void *),
    int 	 errata,
    unsigned int edxErrata
    )
    {
    memset(&i86ArchV1HpcDriverConfig, 0, sizeof(i86ArchV1HpcDriverConfig));
    memset(&i86ArchV2HpcDriverConfig, 0, sizeof(i86ArchV2HpcDriverConfig));
    memset(&hpcHwDrvI86ArchV1, 0, sizeof(hpcHwDrvI86ArchV1));
    memset(&hpcHwDrvI86ArchV2, 0, sizeof(hpcHwDrvI86ArchV2));

    if (pMallocFunc == NULL || pFreeFunc == NULL)
	return ERROR;

    pMalloc = pMallocFunc;
    pFree = pFreeFunc;

    cpuidPerfMonEax = 0xa;
    i86cpuid (&cpuidPerfMonEax, &cpuidPerfMonEbx, &cpuidPerfMonEcx,
		    &cpuidPerfMonEdx);

    if (errata)
        cpuidPerfMonEdx = edxErrata;

    pArchV1Handler = defaultArchV1Handler;
    pArchV2Handler = defaultArchV2Handler;

    return OK;
    }

static void setCpuidValue
    (
    unsigned int eax,
    unsigned int ebx,
    unsigned int ecx,
    unsigned int edx
    )
    {
    cpuidPerfMonEax = eax;
    cpuidPerfMonEbx = ebx;
    cpuidPerfMonEcx = ecx;
    cpuidPerfMonEdx = edx;
    }

#define MSR_DEBUGCTL 0x01d9
#define FREEZE_WHILE_SMM (1ULL << 14)

void hpcI86ArchPerfMonFreezeWhileSmmSet(void)
    {
    long long int value;
    pentiumMsrGet(MSR_DEBUGCTL, (long long int *)&value);
    value |= FREEZE_WHILE_SMM;
    pentiumMsrSet(MSR_DEBUGCTL, (long long int *)&value);
    }

void hpcI86ArchPerfMonFreezeWhileSmmClear(void)
    {
    long long int value;
    pentiumMsrGet(MSR_DEBUGCTL, (long long int *)&value);
    value &= ~FREEZE_WHILE_SMM;
    pentiumMsrSet(MSR_DEBUGCTL, (long long int *)&value);
    }

void hpcI86ArchPerfMonShow(void)
    {
#define MSR_IA32_MISC_ENABLE 0x1A0
#define MSR_IA32_PERF_CAPABILITIES 0x345
    unsigned int cpuid_function, eax, ebx, ecx, edx;

    cpuid_function = eax = 1;
    i86cpuid (&eax, &ebx, &ecx, &edx);
    printf("cpuid f=0x%x eax=0x%x ebx=0x%x ecx=0x%x edx=0x%x\n",
		    cpuid_function, eax, ebx, ecx, edx);
    /* For CPUID=1, check PDCM (Perfmon and Debug Capability) bit of ecx
     * to see if the processor supports the Performance Capabilities */
    if (ecx & (1 << 15))
        {
	long long int value;
        pentiumMsrGet(MSR_IA32_PERF_CAPABILITIES, (long long int *)&value);
	printf("IA32_PERF_CAPABILITIES = 0x%llx\n", value);
	printf("LBR format\t\t\t\t\t0x%llx\n", value & 0x1f);
	printf("PEBS Trap\t\t\t\t\t%s\n", (value & (1 << 6)) ? "True" : "False");
	printf("PEBS Save Arch Regs\t\t\t\t%s\n", (value & (1 << 7)) ? "True" : "False");
	printf("PEBS Record Format\t\t\t\t0x%llx\n", (value & (0xf << 8)) >> 8);
	printf("Freeze while SMM is supported\t\t\t%s\n", (value & (1 << 12)) ? "True" : "False");
	printf("Full width of counter writable via IA32_A_PMCx\t%s\n", (value & (1 << 13)) ? "True" : "False");

        pentiumMsrGet(MSR_IA32_MISC_ENABLE, (long long int *)&value);
        printf("IA32_MISC_ENABLE = 0x%llx\n", value);
	printf("Perf Mon Available\t\t\t\t%s\n", (value & (1 << 7)) ? "True" : "False");
	printf("BTS Unavailable\t\t\t\t\t%s\n", (value & (1 << 11)) ? "True" : "False");
	printf("PEBS Unavailable\t\t\t\t%s\n", (value & (1 << 11)) ? "True" : "False");

        pentiumMsrGet(MSR_DEBUGCTL, (long long int *)&value);
        printf("IA32_DEBUGCTLMSR = 0x%llx\n", value);
        }
    else
        {
	printf("IA32_PERF_CAPABILITIES not supported\n");
	return;
	}

    eax = cpuid_function = 0xa;
    i86cpuid (&eax, &ebx, &ecx, &edx);
    printf("cpuid f=0x%x eax=0x%x ebx=0x%x ecx=0x%x edx=0x%x\n",
		cpuid_function, eax, ebx, ecx, edx);
    printf("Version\t\t\t\t%d\n", PMON_VERSION_ID(eax));
    printf("Number of counters\t\t%d\n", NUM_COUNTERS(eax));
    printf("Counter bits\t\t\t%d\n", COUNTER_BIT_SIZE(eax));
    printf("Number of architecture events\t%u\n", NUM_ARCH_EVENTS(eax));
    printf("Arch core cycles event\t\t\t\t%s\n",
           ARCH_CORE_CYCLES(ebx) ? "YES" : "NO");
    printf("Arch instructions retired event\t\t\t%s\n",
           ARCH_INST_RETIRED(ebx) ? "YES" : "NO");
    printf("Arch reference cycles event\t\t\t%s\n",
           ARCH_REF_CYCLES(ebx) ? "YES" : "NO");
    printf("Arch last level cache reference event\t\t%s\n",
           ARCH_LLC_REFERENCE(ebx) ? "YES" : "NO");
    printf("Arch last level cache misses event\t\t%s\n",
           ARCH_LLC_MISSES(ebx) ? "YES" : "NO");
    printf("Arch branches retired event\t\t\t%s\n",
           ARCH_BRANCH_RETIRED(ebx) ? "YES" : "NO");
    printf("Arch branches mispredicted retired event\t%s\n",
           ARCH_BRANCH_MISPREDICT_RETIRED(ebx) ? "YES" : "NO");
    printf("Number of fixed counters\t%d\n", NUM_FIXED_COUNTERS(edx));
    printf("Fixed counter bits\t\t%d\n", FIXED_COUNTER_BIT_SIZE(edx));
    }

/******************************************************************************
*/

struct i86ArchPmTestFuncs i86ArchPmTest =
    {
    setCpuidValue,
    &i86ArchV1HpcDriverConfig,
    &i86ArchV2HpcDriverConfig,
    hpcI86PerfMonHandler
    };
