/* Copyright (c) 2011, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
16sep14,sdt  Fixed checked-return and sizeof-mismatch warnings (US45196)
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include <cafe/types.h>
#include <cafe/autogenerated/hpc_common.type.h>
#include <cafe/autogenerated/hpc_nehalem_unc.type.h>
#include <hpc/hpc_drv/hpcHwDrvApi.h>
#include <hpc/drv_evt_cfg/hpcDrvEvtCfg.h>
#include <hpc/drv_evt_cfg/i86/i86ArchPmEvtCfg.h>
#include <hpc/drv_evt_cfg/i86/i86NehalemUncEvtCfg.h>
#include <hpc/drv_evt_cfg/i86/i86ArchPm1DrvSet.h>
#include <hpc/drv_evt_cfg/i86/i86ArchPm2DrvSet.h>
#include <hpc/drv_evt_cfg/i86/hpcI86vxw.h>
#include "../../hpc_utils.h"
#include "hpcI86PerfIsr.h"

/* This behavior was observed on an 8core Nehalem CPUID=0106Ex.
 * The UNCORE register MSRs can be read/written on cores 0 to 3. If you write
 * the value on core0, the same value is read back on core2. This implies that
 * the UNCORE MSRs are shared across cores. If you try and read them on
 * cores 4,5,6,7 (for an 8 core system) the value read is 0. Trying to write
 * to UNCORE MSRs on cores 4 to 7 also resulted in an exception. It is not
 * known if this is expected or not.
 * The Uncore PMI can only be enabled on one core. Setting multiple bits
 * in MSR_UNCORE_PERF_GLOBAL_CTRL_MSR:EN_PMI_COREn results in a target
 * exception.
 */

static int pmiIntCore;
static int (*pArchUncoreHandler)(struct driverInstArchV1 *, void *, int, long long int);
static int (*pArchUncoreFixedHandler)(struct driverInstArchV2 *, void *, int, long long int);
static struct driverInstArchV1 hpcHwDrvI86Uncore;
static struct driverInstArchV2 hpcHwDrvI86UncoreFixed;

static int defaultUncoreHandler(struct driverInstArchV1 * p, void * pRegs,
				int cpu, long long int status)
    {
    return 0;
    }

static int defaultUncoreFixedHandler(struct driverInstArchV2 * p, void * pRegs,
				int cpu, long long int status)
    {
    return 0;
    }

static int i86UncorePmHandler
    (
    void * pRegs,
    int cpu
    )
    {
    int p = 0; /* processed */
    long long int pmGlobalStatus;

    if (cpu != pmiIntCore)
        return 0;
    pentiumMsrGet(MSR_UNCORE_PERF_GLOBAL_STATUS,
                  (long long int *)&pmGlobalStatus);

    while (pmGlobalStatus)
        {
        if (pmGlobalStatus << 32)
	    p += (*pArchUncoreHandler)(&hpcHwDrvI86Uncore, pRegs, cpu,
							    pmGlobalStatus);
        if (pmGlobalStatus >> 32)
	    p += (*pArchUncoreFixedHandler)(&hpcHwDrvI86UncoreFixed, pRegs, cpu,
							    pmGlobalStatus);
	/* Clear the overflow */
        pentiumMsrSet(MSR_UNCORE_PERF_GLOBAL_OVF_CTRL,
                      (long long int *)&pmGlobalStatus);

        /* check the status again */
        pentiumMsrGet(MSR_UNCORE_PERF_GLOBAL_STATUS,
                      (long long int *)&pmGlobalStatus);
        }

    return p;
    }

/* uncore events use the i86 arch perf mon V1 driver */

#define HPC_NUM_I86_COREI7_UNC_EVENTS (175 + 0)


/* Note this array must be kept in sync with archEventsDesc array. It Also
 * includes space for arch pmon version 1 events */
struct i86Arch1EventDescription
    x86NehalemUncEvtDesc[HPC_NUM_I86_COREI7_UNC_EVENTS + (2 * NUM_ARCH_V1_EVENTS)] = {
    /* common named events */

    /* hw specific named events. Note the eventUmask is zero, since the
     * event VT description contains it. */
    {hpc_nehalem_UNC_GQ_CYCLES_FULL_READ_TRACKER_VT_DEF,
	{1000},		/* reloadMin */
	{0x00, 0, 0}},	/* eventNum, eventUmask, restrictedCtrMask */
    {hpc_nehalem_UNC_GQ_CYCLES_FULL_WRITE_TRACKER_VT_DEF,
	{1000},
	{0x00, 0, 0}},
    {hpc_nehalem_UNC_GQ_CYCLES_FULL_PEER_PROBE_TRACKER_VT_DEF,
	{1000},
	{0x00, 0, 0}},
    {hpc_nehalem_UNC_GQ_CYCLES_NOT_EMPTY_READ_TRACKER_VT_DEF,
	{1000},
	{0x01, 0, 0}},
    {hpc_nehalem_UNC_GQ_CYCLES_NOT_EMPTY_WRITE_TRACKER_VT_DEF,
	{1000},
	{0x01, 0, 0}},
    {hpc_nehalem_UNC_GQ_CYCLES_NOT_EMPTY_PEER_PROBE_TRACKER_VT_DEF,
	{1000},
	{0x01, 0, 0}},
    {hpc_nehalem_UNC_GQ_ALLOC_READ_TRACKER_VT_DEF,
	{1000},
	{0x03, 0, 0}},
    {hpc_nehalem_UNC_GQ_ALLOC_RT_L3_MISS_VT_DEF,
	{1000},
	{0x03, 0, 0}},
    {hpc_nehalem_UNC_GQ_ALLOC_RT_TO_L3_RESP_VT_DEF,
	{1000},
	{0x03, 0, 0}},
    {hpc_nehalem_UNC_GQ_ALLOC_RT_TO_RTID_ACQUIRED_VT_DEF,
	{1000},
	{0x03, 0, 0}},
    {hpc_nehalem_UNC_GQ_ALLOC_WT_TO_RTID_ACQUIRED_VT_DEF,
	{1000},
	{0x03, 0, 0}},
    {hpc_nehalem_UNC_GQ_ALLOC_WRITE_TRACKER_VT_DEF,
	{1000},
	{0x03, 0, 0}},
    {hpc_nehalem_UNC_GQ_ALLOC_PEER_PROBE_TRACKER_VT_DEF,
	{1000},
	{0x03, 0, 0}},
    {hpc_nehalem_UNC_GQ_DATA_FROM_QPI_VT_DEF,
	{1000},
	{0x04, 0, 0}},
    {hpc_nehalem_UNC_GQ_DATA_FROM_QMC_VT_DEF,
	{1000},
	{0x04, 0, 0}},
    {hpc_nehalem_UNC_GQ_DATA_FROM_L3_VT_DEF,
	{1000},
	{0x04, 0, 0}},
    {hpc_nehalem_UNC_GQ_DATA_FROM_CORES_02_VT_DEF,
	{1000},
	{0x04, 0, 0}},
    {hpc_nehalem_UNC_GQ_DATA_FROM_CORES_13_VT_DEF,
	{1000},
	{0x04, 0, 0}},
    {hpc_nehalem_UNC_GQ_DATA_TO_QPI_QMC_VT_DEF,
	{1000},
	{0x05, 0, 0}},
    {hpc_nehalem_UNC_GQ_DATA_TO_L3_VT_DEF,
	{1000},
	{0x05, 0, 0}},
    {hpc_nehalem_UNC_GQ_DATA_TO_CORES_VT_DEF,
	{1000},
	{0x05, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_LOCAL_HOME_I_STATE_VT_DEF,
	{1000},
	{0x06, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_LOCAL_HOME_S_STATE_VT_DEF,
	{1000},
	{0x06, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_LOCAL_HOME_FWD_S_STATE_VT_DEF,
	{1000},
	{0x06, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_LOCAL_HOME_FWD_I_STATE_VT_DEF,
	{1000},
	{0x06, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_LOCAL_HOME_CONFLICT_VT_DEF,
	{1000},
	{0x06, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_LOCAL_HOME_WB_VT_DEF,
	{1000},
	{0x06, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_REMOTE_HOME_I_STATE_VT_DEF,
	{1000},
	{0x07, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_REMOTE_HOME_S_STATE_VT_DEF,
	{1000},
	{0x07, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_REMOTE_HOME_FWD_S_STATE_VT_DEF,
	{1000},
	{0x07, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_REMOTE_HOME_FWD_I_STATE_VT_DEF,
	{1000},
	{0x07, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_REMOTE_HOME_CONFLICT_VT_DEF,
	{1000},
	{0x07, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_REMOTE_HOME_WB_VT_DEF,
	{1000},
	{0x07, 0, 0}},
    {hpc_nehalem_UNC_SNP_RESP_TO_REMOTE_HOME_HITM_VT_DEF,
	{1000},
	{0x07, 0, 0}},
    {hpc_nehalem_UNC_L3_HITS_READ_VT_DEF,
	{1000},
	{0x08, 0, 0}},
    {hpc_nehalem_UNC_L3_HITS_WRITE_VT_DEF,
	{1000},
	{0x08, 0, 0}},
    {hpc_nehalem_UNC_L3_HITS_PROBE_VT_DEF,
	{1000},
	{0x08, 0, 0}},
    {hpc_nehalem_UNC_L3_HITS_ANY_VT_DEF,
	{1000},
	{0x08, 0, 0}},
    {hpc_nehalem_UNC_L3_MISS_READ_VT_DEF,
	{1000},
	{0x09, 0, 0}},
    {hpc_nehalem_UNC_L3_MISS_WRITE_VT_DEF,
	{1000},
	{0x09, 0, 0}},
    {hpc_nehalem_UNC_L3_MISS_PROBE_VT_DEF,
	{1000},
	{0x09, 0, 0}},
    {hpc_nehalem_UNC_L3_MISS_ANY_VT_DEF,
	{1000},
	{0x09, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_IN_M_STATE_VT_DEF,
	{1000},
	{0x0a, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_IN_E_STATE_VT_DEF,
	{1000},
	{0x0a, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_IN_S_STATE_VT_DEF,
	{1000},
	{0x0a, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_IN_F_STATE_VT_DEF,
	{1000},
	{0x0a, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_IN_ANY_VT_DEF,
	{1000},
	{0x0a, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_OUT_M_STATE_VT_DEF,
	{1000},
	{0x0b, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_OUT_E_STATE_VT_DEF,
	{1000},
	{0x0b, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_OUT_S_STATE_VT_DEF,
	{1000},
	{0x0b, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_OUT_I_STATE_VT_DEF,
	{1000},
	{0x0b, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_OUT_F_STATE_VT_DEF,
	{1000},
	{0x0b, 0, 0}},
    {hpc_nehalem_UNC_L3_LINES_OUT_ANY_VT_DEF,
	{1000},
	{0x0b, 0, 0}},
    {hpc_nehalem_UNC_QHL_REQUESTS_IOH_READS_VT_DEF,
	{1000},
	{0x20, 0, 0}},
    {hpc_nehalem_UNC_QHL_REQUESTS_IOH_WRITES_VT_DEF,
	{1000},
	{0x20, 0, 0}},
    {hpc_nehalem_UNC_QHL_REQUESTS_REMOTE_READS_VT_DEF,
	{1000},
	{0x20, 0, 0}},
    {hpc_nehalem_UNC_QHL_REQUESTS_REMOTE_WRITES_VT_DEF,
	{1000},
	{0x20, 0, 0}},
    {hpc_nehalem_UNC_QHL_REQUESTS_LOCAL_READS_VT_DEF,
	{1000},
	{0x20, 0, 0}},
    {hpc_nehalem_UNC_QHL_REQUESTS_LOCAL_WRITES_VT_DEF,
	{1000},
	{0x20, 0, 0}},
    {hpc_nehalem_UNC_QHL_CYCLES_FULL_IOH_VT_DEF,
	{1000},
	{0x21, 0, 0}},
    {hpc_nehalem_UNC_QHL_CYCLES_FULL_REMOTE_VT_DEF,
	{1000},
	{0x21, 0, 0}},
    {hpc_nehalem_UNC_QHL_CYCLES_FULL_LOCAL_VT_DEF,
	{1000},
	{0x21, 0, 0}},
    {hpc_nehalem_UNC_QHL_CYCLES_NOT_EMPTY_IOH_VT_DEF,
	{1000},
	{0x22, 0, 0}},
    {hpc_nehalem_UNC_QHL_CYCLES_NOT_EMPTY_REMOTE_VT_DEF,
	{1000},
	{0x22, 0, 0}},
    {hpc_nehalem_UNC_QHL_CYCLES_NOT_EMPTY_LOCAL_VT_DEF,
	{1000},
	{0x22, 0, 0}},
    {hpc_nehalem_UNC_QHL_OCCUPANCY_IOH_VT_DEF,
	{1000},
	{0x23, 0, 0}},
    {hpc_nehalem_UNC_QHL_OCCUPANCY_REMOTE_VT_DEF,
	{1000},
	{0x23, 0, 0}},
    {hpc_nehalem_UNC_QHL_OCCUPANCY_LOCAL_VT_DEF,
	{1000},
	{0x23, 0, 0}},
    {hpc_nehalem_UNC_QHL_ADDRESS_CONFLICTS_2WAY_VT_DEF,
	{1000},
	{0x24, 0, 0}},
    {hpc_nehalem_UNC_QHL_ADDRESS_CONFLICTS_3WAY_VT_DEF,
	{1000},
	{0x24, 0, 0}},
    {hpc_nehalem_UNC_QHL_CONFLICT_CYCLES_IOH_VT_DEF,
	{1000},
	{0x25, 0, 0}},
    {hpc_nehalem_UNC_QHL_CONFLICT_CYCLES_REMOTE_VT_DEF,
	{1000},
	{0x25, 0, 0}},
    {hpc_nehalem_UNC_QHL_CONFLICT_CYCLES_LOCAL_VT_DEF,
	{1000},
	{0x25, 0, 0}},
    {hpc_nehalem_UNC_QHL_TO_QMC_BYPASS_VT_DEF,
	{1000},
	{0x26, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_FULL_READ_CH0_VT_DEF,
	{1000},
	{0x27, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_FULL_READ_CH1_VT_DEF,
	{1000},
	{0x27, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_FULL_READ_CH2_VT_DEF,
	{1000},
	{0x27, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_FULL_WRITE_CH0_VT_DEF,
	{1000},
	{0x27, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_FULL_WRITE_CH1_VT_DEF,
	{1000},
	{0x27, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_FULL_WRITE_CH2_VT_DEF,
	{1000},
	{0x27, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISOC_FULL_READ_CH0_VT_DEF,
	{1000},
	{0x28, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISOC_FULL_READ_CH1_VT_DEF,
	{1000},
	{0x28, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISOC_FULL_READ_CH2_VT_DEF,
	{1000},
	{0x28, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISOC_FULL_WRITE_CH0_VT_DEF,
	{1000},
	{0x28, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISOC_FULL_WRITE_CH1_VT_DEF,
	{1000},
	{0x28, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISOC_FULL_WRITE_CH2_VT_DEF,
	{1000},
	{0x28, 0, 0}},
    {hpc_nehalem_UNC_QMC_BUSY_READ_CH0_VT_DEF,
	{1000},
	{0x29, 0, 0}},
    {hpc_nehalem_UNC_QMC_BUSY_READ_CH1_VT_DEF,
	{1000},
	{0x29, 0, 0}},
    {hpc_nehalem_UNC_QMC_BUSY_READ_CH2_VT_DEF,
	{1000},
	{0x29, 0, 0}},
    {hpc_nehalem_UNC_QMC_BUSY_WRITE_CH0_VT_DEF,
	{1000},
	{0x29, 0, 0}},
    {hpc_nehalem_UNC_QMC_BUSY_WRITE_CH1_VT_DEF,
	{1000},
	{0x29, 0, 0}},
    {hpc_nehalem_UNC_QMC_BUSY_WRITE_CH2_VT_DEF,
	{1000},
	{0x29, 0, 0}},
    {hpc_nehalem_UNC_QMC_OCCUPANCY_CH0_VT_DEF,
	{1000},
	{0x2a, 0, 0}},
    {hpc_nehalem_UNC_QMC_OCCUPANCY_CH1_VT_DEF,
	{1000},
	{0x2a, 0, 0}},
    {hpc_nehalem_UNC_QMC_OCCUPANCY_CH2_VT_DEF,
	{1000},
	{0x2a, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISSOC_OCCUPANCY_CH0_VT_DEF,
	{1000},
	{0x2b, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISSOC_OCCUPANCY_CH1_VT_DEF,
	{1000},
	{0x2b, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISSOC_OCCUPANCY_CH2_VT_DEF,
	{1000},
	{0x2b, 0, 0}},
    {hpc_nehalem_UNC_QMC_ISSOC_READS_ANY_VT_DEF,
	{1000},
	{0x2b, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_READS_CH0_VT_DEF,
	{1000},
	{0x2c, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_READS_CH1_VT_DEF,
	{1000},
	{0x2c, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_READS_CH2_VT_DEF,
	{1000},
	{0x2c, 0, 0}},
    {hpc_nehalem_UNC_QMC_NORMAL_READS_ANY_VT_DEF,
	{1000},
	{0x2c, 0, 0}},
    {hpc_nehalem_UNC_QMC_HIGH_PRIORITY_READS_CH0_VT_DEF,
	{1000},
	{0x2d, 0, 0}},
    {hpc_nehalem_UNC_QMC_HIGH_PRIORITY_READS_CH1_VT_DEF,
	{1000},
	{0x2d, 0, 0}},
    {hpc_nehalem_UNC_QMC_HIGH_PRIORITY_READS_CH2_VT_DEF,
	{1000},
	{0x2d, 0, 0}},
    {hpc_nehalem_UNC_QMC_HIGH_PRIORITY_READS_ANY_VT_DEF,
	{1000},
	{0x2d, 0, 0}},
    {hpc_nehalem_UNC_QMC_CRITICAL_PRIORITY_READS_CH0_VT_DEF,
	{1000},
	{0x2e, 0, 0}},
    {hpc_nehalem_UNC_QMC_CRITICAL_PRIORITY_READS_CH1_VT_DEF,
	{1000},
	{0x2e, 0, 0}},
    {hpc_nehalem_UNC_QMC_CRITICAL_PRIORITY_READS_CH2_VT_DEF,
	{1000},
	{0x2e, 0, 0}},
    {hpc_nehalem_UNC_QMC_CRITICAL_PRIORITY_READS_ANY_VT_DEF,
	{1000},
	{0x2e, 0, 0}},
    {hpc_nehalem_UNC_QMC_WRITES_FULL_CH0_VT_DEF,
	{1000},
	{0x2f, 0, 0}},
    {hpc_nehalem_UNC_QMC_WRITES_FULL_CH1_VT_DEF,
	{1000},
	{0x2f, 0, 0}},
    {hpc_nehalem_UNC_QMC_WRITES_FULL_CH2_VT_DEF,
	{1000},
	{0x2f, 0, 0}},
    {hpc_nehalem_UNC_QMC_WRITES_FULL_ANY_VT_DEF,
	{1000},
	{0x2f, 0, 0}},
    {hpc_nehalem_UNC_QMC_WRITES_PARTIAL_CH0_VT_DEF,
	{1000},
	{0x2f, 0, 0}},
    {hpc_nehalem_UNC_QMC_WRITES_PARTIAL_CH1_VT_DEF,
	{1000},
	{0x2f, 0, 0}},
    {hpc_nehalem_UNC_QMC_WRITES_PARTIAL_CH2_VT_DEF,
	{1000},
	{0x2f, 0, 0}},
    {hpc_nehalem_UNC_QMC_WRITES_PARTIAL_ANY_VT_DEF,
	{1000},
	{0x2f, 0, 0}},
    {hpc_nehalem_UNC_QMC_CANCEL_CH0_VT_DEF,
	{1000},
	{0x30, 0, 0}},
    {hpc_nehalem_UNC_QMC_CANCEL_CH1_VT_DEF,
	{1000},
	{0x30, 0, 0}},
    {hpc_nehalem_UNC_QMC_CANCEL_CH2_VT_DEF,
	{1000},
	{0x30, 0, 0}},
    {hpc_nehalem_UNC_QMC_CANCEL_ANY_VT_DEF,
	{1000},
	{0x30, 0, 0}},
    {hpc_nehalem_UNC_QMC_PRIORITY_UPDATES_CH0_VT_DEF,
	{1000},
	{0x31, 0, 0}},
    {hpc_nehalem_UNC_QMC_PRIORITY_UPDATES_CH1_VT_DEF,
	{1000},
	{0x31, 0, 0}},
    {hpc_nehalem_UNC_QMC_PRIORITY_UPDATES_CH2_VT_DEF,
	{1000},
	{0x31, 0, 0}},
    {hpc_nehalem_UNC_QMC_PRIORITY_UPDATES_ANY_VT_DEF,
	{1000},
	{0x31, 0, 0}},
    {hpc_nehalem_UNC_QHL_FRC_ACK_CNFLTS_LOCAL_VT_DEF,
	{1000},
	{0x33, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_SINGLE_FLIT_HOME_LINK_0_VT_DEF,
	{1000},
	{0x40, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_SINGLE_FLIT_SNOOP_LINK_0_VT_DEF,
	{1000},
	{0x40, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_SINGLE_FLIT_NDR_LINK_0_VT_DEF,
	{1000},
	{0x40, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_SINGLE_FLIT_HOME_LINK_1_VT_DEF,
	{1000},
	{0x40, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_SINGLE_FLIT_SNOOP_LINK_1_VT_DEF,
	{1000},
	{0x40, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_SINGLE_FLIT_NDR_LINK_1_VT_DEF,
	{1000},
	{0x40, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_SINGLE_FLIT_LINK_0_VT_DEF,
	{1000},
	{0x40, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_SINGLE_FLIT_LINK_1_VT_DEF,
	{1000},
	{0x40, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_MULTI_FLIT_DRS_LINK_0_VT_DEF,
	{1000},
	{0x41, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_MULTI_FLIT_NCB_LINK_0_VT_DEF,
	{1000},
	{0x41, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_MULTI_FLIT_NCS_LINK_0_VT_DEF,
	{1000},
	{0x41, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_MULTI_FLIT_DRS_LINK_1_VT_DEF,
	{1000},
	{0x41, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_MULTI_FLIT_NCB_LINK_1_VT_DEF,
	{1000},
	{0x41, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_MULTI_FLIT_NCS_LINK_1_VT_DEF,
	{1000},
	{0x41, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_MULTI_FLIT_LINK_0_VT_DEF,
	{1000},
	{0x41, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_STALLED_MULTI_FLIT_LINK_1_VT_DEF,
	{1000},
	{0x41, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_HEADER_BUSY_LINK_0_VT_DEF,
	{1000},
	{0x42, 0, 0}},
    {hpc_nehalem_UNC_QPI_TX_HEADER_BUSY_LINK_1_VT_DEF,
	{1000},
	{0x42, 0, 0}},
    {hpc_nehalem_UNC_QPI_RX_NO_PPT_CREDIT_STALLS_LINK_0_VT_DEF,
	{1000},
	{0x43, 0, 0}},
    {hpc_nehalem_UNC_QPI_RX_NO_PPT_CREDIT_STALLS_LINK_1_VT_DEF,
	{1000},
	{0x43, 0, 0}},
    {hpc_nehalem_UNC_DRAM_OPEN_CH0_VT_DEF,
	{1000},
	{0x60, 0, 0}},
    {hpc_nehalem_UNC_DRAM_OPEN_CH1_VT_DEF,
	{1000},
	{0x60, 0, 0}},
    {hpc_nehalem_UNC_DRAM_OPEN_CH2_VT_DEF,
	{1000},
	{0x60, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PAGE_CLOSE_CH0_VT_DEF,
	{1000},
	{0x61, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PAGE_CLOSE_CH1_VT_DEF,
	{1000},
	{0x61, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PAGE_CLOSE_CH2_VT_DEF,
	{1000},
	{0x61, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PAGE_MISS_CH0_VT_DEF,
	{1000},
	{0x62, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PAGE_MISS_CH1_VT_DEF,
	{1000},
	{0x62, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PAGE_MISS_CH2_VT_DEF,
	{1000},
	{0x62, 0, 0}},
    {hpc_nehalem_UNC_DRAM_READ_CAS_CH0_VT_DEF,
	{1000},
	{0x63, 0, 0}},
    {hpc_nehalem_UNC_DRAM_READ_CAS_AUTOPRE_CH0_VT_DEF,
	{1000},
	{0x63, 0, 0}},
    {hpc_nehalem_UNC_DRAM_READ_CAS_CH1_VT_DEF,
	{1000},
	{0x63, 0, 0}},
    {hpc_nehalem_UNC_DRAM_READ_CAS_AUTOPRE_CH1_VT_DEF,
	{1000},
	{0x63, 0, 0}},
    {hpc_nehalem_UNC_DRAM_READ_CAS_CH2_VT_DEF,
	{1000},
	{0x63, 0, 0}},
    {hpc_nehalem_UNC_DRAM_READ_CAS_AUTOPRE_CH2_VT_DEF,
	{1000},
	{0x63, 0, 0}},
    {hpc_nehalem_UNC_DRAM_WRITE_CAS_CH0_VT_DEF,
	{1000},
	{0x64, 0, 0}},
    {hpc_nehalem_UNC_DRAM_WRITE_CAS_AUTOPRE_CH0_VT_DEF,
	{1000},
	{0x64, 0, 0}},
    {hpc_nehalem_UNC_DRAM_WRITE_CAS_CH1_VT_DEF,
	{1000},
	{0x64, 0, 0}},
    {hpc_nehalem_UNC_DRAM_WRITE_CAS_AUTOPRE_CH1_VT_DEF,
	{1000},
	{0x64, 0, 0}},
    {hpc_nehalem_UNC_DRAM_WRITE_CAS_CH2_VT_DEF,
	{1000},
	{0x64, 0, 0}},
    {hpc_nehalem_UNC_DRAM_WRITE_CAS_AUTOPRE_CH2_VT_DEF,
	{1000},
	{0x64, 0, 0}},
    {hpc_nehalem_UNC_DRAM_REFRESH_CH0_VT_DEF,
	{1000},
	{0x65, 0, 0}},
    {hpc_nehalem_UNC_DRAM_REFRESH_CH1_VT_DEF,
	{1000},
	{0x65, 0, 0}},
    {hpc_nehalem_UNC_DRAM_REFRESH_CH2_VT_DEF,
	{1000},
	{0x65, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PRE_ALL_CH0_VT_DEF,
	{1000},
	{0x66, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PRE_ALL_CH1_VT_DEF,
	{1000},
	{0x66, 0, 0}},
    {hpc_nehalem_UNC_DRAM_PRE_ALL_CH2_VT_DEF,
	{1000},
	{0x66, 0, 0}}};

#define NUM_PMC_COUNTERS 8
static struct hpcHwCtrData * hwCtrData[NUM_PMC_COUNTERS];

static STATUS enableUncoreIsrCoreIter(int cpu, void * data)
    {
    hpcVxwI86PerfIsrConnectVec(hpcI86PerfMonVector, hpcI86PerfMonHandler);
    hpcVxwI86PerfMonIsrEnable();
    return OK;
    }

static STATUS enableUncoreIsr
    (
    int (*pHandler)(struct driverInstArchV1 *, void *, int, long long int)
    )
    {
    hpcI86PerfIsrHandler(1, i86UncorePmHandler);
    pArchUncoreHandler = pHandler;
    return hpcExecuteOnSingleCpu(pmiIntCore, enableUncoreIsrCoreIter, NULL);
    }

static STATUS disableUncoreIsrCoreIter(int cpu, void * data)
    {
    int *pForce = data;
    hpcVxwI86PerfMonIsrDisable();
    hpcVxwI86PerfIsrDisconnectVec(hpcI86PerfMonHandler, *pForce);
    return OK;
    }

static STATUS disableUncoreIsr(int force)
    {
    STATUS rc = hpcExecuteOnSingleCpu(pmiIntCore, disableUncoreIsrCoreIter, &force);
    hpcI86PerfIsrHandler(1, NULL);
    pArchUncoreHandler = defaultUncoreHandler;
    return rc;
    }

static struct driverInstArchV1 hpcHwDrvI86Uncore =
    {
    "Nehalem uncore pmc events",
    NUM_PMC_COUNTERS,
    48, /* counter bits */
    hwCtrData,
    NULL,
    NULL,
    NULL,
    I86_ARCH_PM_DRV1_UNCORE,
    MSR_UNCORE_PMC0,
    MSR_UNCORE_PERFEVTSEL0,
    enableUncoreIsr,
    disableUncoreIsr,
    0, 0
    };

static struct hpcDriverConfig i86NehalenUncDriverConfig = {
    HPC_NUM_I86_COREI7_UNC_EVENTS,
    1, /* Uncore events are per system, not per core */
    (void *)&hpcHwDrvI86Uncore,
    &hpcHwDrvI86Arch1Rtns,
    i86ArchPm1EventSettingsGet,
    sizeof(struct i86Arch1EventDescription),
    x86NehalemUncEvtDesc
    };

/* uncore fixed events use the i86 arch perf mon V2 driver */

#define HPC_NUM_I86_COREI7_UNC_FIXED_EVENTS (1)

static struct i86Arch2EventDescription
	x86NehalemUncFixedEvtDesc[HPC_NUM_I86_COREI7_UNC_FIXED_EVENTS] = {
    /* U-Clock */
    {hpc_nehalem_UNC_U_CLOCK_CYCLES_VT_DEF,
        {500000},
        {0}}};

static STATUS enableUncoreFixedIsr
    (
    int (*pHandler)(struct driverInstArchV2 *, void *, int, long long int)
    )
    {
    hpcI86PerfIsrHandler(1, i86UncorePmHandler);
    pArchUncoreFixedHandler = pHandler;
    return hpcExecuteOnSingleCpu(pmiIntCore, enableUncoreIsrCoreIter, NULL);
    }

static STATUS disableUncoreFixedIsr(int force)
    {
    STATUS rc = hpcExecuteOnSingleCpu(pmiIntCore, disableUncoreIsrCoreIter, &force);
    hpcI86PerfIsrHandler(1, NULL);
    pArchUncoreFixedHandler = defaultUncoreFixedHandler;
    return rc;
    }

#define NUM_FIXED_COUNTERS 1
static struct hpcHwCtrData * hwCtrDataFixed[NUM_FIXED_COUNTERS];

static struct driverInstArchV2 hpcHwDrvI86UncoreFixed =
    {
    "Nehalem uncore fixed event",
    NUM_FIXED_COUNTERS,
    48, /* counter bits */
    hwCtrDataFixed,
    NULL,
    NULL,
    NULL,
    I86_ARCH_PM_DRV2_UNCORE,
    MSR_UNCORE_FIXED_CTR0,
    MSR_UNCORE_FIXED_CTR_CTRL,
    MSR_UNCORE_PERF_GLOBAL_STATUS,
    MSR_UNCORE_PERF_GLOBAL_CTRL,
    MSR_UNCORE_PERF_GLOBAL_OVF_CTRL,
    enableUncoreFixedIsr,
    disableUncoreFixedIsr,
    0, 0
    };

static struct hpcDriverConfig i86NehalenUncFixedDriverConfig = {
    HPC_NUM_I86_COREI7_UNC_FIXED_EVENTS,
    1, /* Uncore events are per system, not per core */
    (void *)&hpcHwDrvI86UncoreFixed,
    &hpcHwDrvI86Arch2Rtns,
    i86ArchPm2EventSettingsGet,
    sizeof(struct i86Arch2EventDescription),
    x86NehalemUncFixedEvtDesc
    };

static STATUS enableUncorePmi(int cpu, void * data)
    {
    long long int msrDebugControl;
    long long int pmGlobalControl =
        0x1ULL << (48 + pmiIntCore) | /* enable PMI interrupt on core intCore */
        1ULL << 32 |	/* enable 1 fixed counter */
        0xffULL;	/* enable 8 general purpose counters */

    pentiumMsrSet(MSR_UNCORE_PERF_GLOBAL_CTRL,
                  (long long int *)&pmGlobalControl);

    pentiumMsrGet(MSR_IA32_DEBUG_CTL, (long long int *)&msrDebugControl);
    msrDebugControl |= 1ULL<<13; /* ENABLE_UNCORE_PMI bit */
    pentiumMsrSet(MSR_IA32_DEBUG_CTL, (long long int *)&msrDebugControl);

    return OK;
    }

static STATUS disableUncorePmi(int cpu, void * data)
    {
    long long int msrDebugControl;
    long long int pmGlobalControl = 0ULL;

    /* could only write to this register on cores 0 to four on device Nehalem
     * 8-core CPUID=0106Ex */
    if (cpu < 4)
        pentiumMsrSet(MSR_UNCORE_PERF_GLOBAL_CTRL,
                      (long long int *)&pmGlobalControl);

    pentiumMsrGet(MSR_IA32_DEBUG_CTL, (long long int *)&msrDebugControl);
    msrDebugControl &= ~(1ULL<<13); /* ENABLE_UNCORE_PMI bit */
    pentiumMsrSet(MSR_IA32_DEBUG_CTL, (long long int *)&msrDebugControl);

    return OK;
    }

STATUS hpcI86NehalemUncFinit (void)
    {
    (void)hpcExecuteOnSingleCpu(pmiIntCore, disableUncorePmi, NULL);
    return OK;
    }

STATUS hpcI86NehalemUncInit
    (
    unsigned int intCore
    )
    {
    pmiIntCore = intCore;
    (void)hpcIterateCpus(pmiIntCore, disableUncorePmi, NULL);
    (void)hpcExecuteOnSingleCpu(pmiIntCore, enableUncorePmi, NULL);
    return OK;
    }

struct hpcDriverConfig * hpcI86NehalemUncDrvCfgGet(void)
    {
    return &i86NehalenUncDriverConfig;
    }

struct hpcDriverConfig * hpcI86NehalemUncFixedDrvCfgGet(void)
    {
    return &i86NehalenUncFixedDriverConfig;
    }
