/* Copyright (c) 2010 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include <hpc/hpc_drv/hpcHwDrvApi.h>
#include <hpc/hpc_drv/i86/i86ArchPm2Drv.h>

/* Array index for a 2 dimensional array */
#define A2D(ROW,COLUMN,NCOLUMNS) (ROW * NCOLUMNS + COLUMN)

struct hpcHwCtrData
    {
    void *		pClient;
    HPC_COUNTER		countValue;
    int			counter;
    HPC_COUNTER		reloadValue;
    HPC_COUNTER		counterWrap;
    int			fixedCounterAddrs;
    unsigned int	ctrlEnableMask;
    unsigned long long	ctrlMaskNot;
    unsigned int	ctrCtrlValue;
    };

/* This macro is used to cast HPC_COUNTER type for use with pentiumMsrGet(). */
#define HPC_CAST(V) ((long long int *)&V)

/* Architecture Performance Monitoring Version 2 registers. */

/* bits definitions for IA32_FIXED_CTR_CTRL register */
#define FIXED_ENABLE_OS		1
#define FIXED_ENABLE_US		(1<<1)
#define FIXED_PMI		(1<<3)
#define FIXED_ANY		(1<<2)

#define FIXED_PMI_UNCORE	(1<<2)
#define FIXED_ENABLE_UNCORE	(1<<0)

static STATUS hpcHwDriverShow
    (
    void * pDrvInst
    )
    {
    struct driverInstArchV2 * pInst = pDrvInst;
    long long int value;
    int n;

    /* Display fixed counters  */
    for (n = 0; n < pInst->numCounters; n++)
        {
        pentiumMsrGet(pInst->msrFixedCtr + n, &value);
        printf("Fixed CTR%d\t\t0x%llx\n", n, value);
        }

    pentiumMsrGet(pInst->msrFixedCtrCtrl, &value);
    printf("Fixed Control\t\t0x%llx\n", value);

    pentiumMsrGet(pInst->msrGlobalStatus, &value);
    printf("Global status\t\t0x%llx\n", value);

    pentiumMsrGet(pInst->msrGlobalCtrl, &value);
    printf("Global control\t\t0x%llx\n", value);

    return OK;
    }

/******************************************************************************
*/

static int perfIntHandler
    (
    struct driverInstArchV2 *	pInst,
    void * 			pRegs,
    int				cpu,
    long long int		perfGlobalStatus
    )
    {
    int eventsProcessed = 0;
    unsigned int counter;
    HPC_COUNTER pmcValue;
    unsigned int status = (unsigned int)(perfGlobalStatus >> 32);
    int ctrPerCore;

    if (pInst->archPerfMonVer == I86_ARCH_PM_DRV2_UNCORE)
        ctrPerCore = 0;
    else
        ctrPerCore = cpu;

    if ((status & pInst->ctrCounterMask) == 0)
        return 0;

    for (counter = 0; counter < pInst->numCounters; counter++)
        {
	if ((1 << counter) & status)
	    {
	    int idx2d = A2D(ctrPerCore, counter, pInst->numCounters);
	    struct hpcHwCtrData * pCtr;
            long long int control;

	    pCtr = pInst->ppDrvHwCtrData[idx2d];
	    if (pCtr == NULL)
		continue;

	    /* get counter status */
            pentiumMsrGet (pInst->msrFixedCtrCtrl, &control);

            /* check if it is still running */
            if ((control & pCtr->ctrlEnableMask) == 0)
                continue;

	    /* stop the counter */
            control &= pCtr->ctrlMaskNot;
            pentiumMsrSet(pInst->msrFixedCtrCtrl, &control);

	    /* read counter */
	    pentiumMsrGet (pCtr->fixedCounterAddrs, HPC_CAST(pmcValue));

	    pmcValue += pInst->ctrRollover - pCtr->reloadValue;
	    (*pInst->pHandler) (pCtr->pClient, pRegs, pmcValue, cpu);

	    /* reload counter */
	    pentiumMsrSet(pCtr->fixedCounterAddrs, HPC_CAST(pCtr->reloadValue));

	    /* restart the counter */
            control |= pCtr->ctrCtrlValue;
            pentiumMsrSet(pInst->msrFixedCtrCtrl, &control);

	    eventsProcessed = 1;
	    }
        }

    return eventsProcessed;
    }

/******************************************************************************
*/

static STATUS hpcHwCtrStart
    (
    void *		  pDrvInst,
    struct hpcHwCtrData * pCtr
    )
    {
    struct driverInstArchV2 * pInst = pDrvInst;
    long long int control;
    size_t idx2d;
    int key;

    key = intCpuLock();
    pentiumMsrGet (pInst->msrFixedCtrCtrl, &control);
    if (control & pCtr->ctrlEnableMask)
        {
        intCpuUnlock(key);
	printf ("counter %u is already running, counter 0x%llx %s\n",
	           pCtr->counter, control, pInst->pDriverName);
	return ERROR;
        }

    idx2d = A2D(vxCpuIndexGet(), pCtr->counter, pInst->numCounters);
    pInst->ppDrvHwCtrData[idx2d] = pCtr;

    /* set the counters current value */
    pentiumMsrSet (pCtr->fixedCounterAddrs, HPC_CAST(pCtr->countValue));

    control &= pCtr->ctrlMaskNot;
    control |= pCtr->ctrCtrlValue;
    pentiumMsrSet(pInst->msrFixedCtrCtrl, &control);
    intCpuUnlock(key);

    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwCtrStop
    (
    void *		  pDrvInst,
    struct hpcHwCtrData * pCtr,
    HPC_COUNTER *	  pCount
    )
    {
    struct driverInstArchV2 * pInst = pDrvInst;
    long long int control;
    size_t idx2d;
    int key;

    key = intCpuLock();
    pentiumMsrGet (pInst->msrFixedCtrCtrl, &control);

    if ((control & pCtr->ctrlEnableMask) == 0)
        {
        intCpuUnlock(key);
	printf ("counter %u is not running, control 0x%llx %s\n",
	           pCtr->counter, control, pInst->pDriverName);
	return ERROR;
        }

    /* stop the counter */
    control &= pCtr->ctrlMaskNot;
    pentiumMsrSet(pInst->msrFixedCtrCtrl, &control);
    intCpuUnlock(key);

    /* save the counters value */
    pentiumMsrGet (pCtr->fixedCounterAddrs, HPC_CAST(pCtr->countValue));

    idx2d = A2D(vxCpuIndexGet(), pCtr->counter, pInst->numCounters);
    pInst->ppDrvHwCtrData[idx2d] = NULL;

    if (pCount != NULL)
        {
	if (pCtr->countValue >= pCtr->reloadValue)
	    *pCount = pCtr->countValue - pCtr->reloadValue;
	else
	    *pCount = pCtr->countValue + (pInst->ctrRollover - pCtr->reloadValue);
        }

    return OK;
    }

/******************************************************************************
* This API can be called from ISR context
*/

static STATUS hpcHwCtrRead
    (
    void *		  pDrvInst,
    struct hpcHwCtrData * pCtr,
    HPC_COUNTER	*	  pValue
    )
    {
    struct driverInstArchV2 * pInst = pDrvInst;
    long long int control;
    long long int pmcValue;
    int key;

    /* stop the counter in order to read it */
    key = intCpuLock();
    pentiumMsrGet (pInst->msrFixedCtrCtrl, &control);
    control &= pCtr->ctrlMaskNot;
    pentiumMsrSet(pInst->msrFixedCtrCtrl, &control);
    /* read counter value */
    pentiumMsrGet (pCtr->fixedCounterAddrs, &pmcValue);
    /* restart the counter */
    control |= pCtr->ctrCtrlValue;
    pentiumMsrSet(pInst->msrFixedCtrCtrl, &control);
    intCpuUnlock(key);

    /* Check if the counter has rolled over and the ISR to reload it has not
     * run yet */
    if (pmcValue >= pCtr->reloadValue)
        *pValue = pmcValue - pCtr->reloadValue;
    else
	*pValue = pmcValue + (pInst->ctrRollover - pCtr->reloadValue);

    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwDeleteCtrForEvent
    (
    void *		  pDrvInst,
    struct hpcHwCtrData * pCtr
    )
    {
    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwUpdateCtrReload
    (
    void * 		  pDrvInst,
    struct hpcHwCtrData * pCtr,
    HPC_COUNTER *	  pReload
    )
    {
    struct driverInstArchV2 * pInst = pDrvInst;

    if (*pReload > pInst->ctrRollover || *pReload == 0)
        {
        /*
         * set a value that will cause maximum number of perf counter increments
         * before perf rollover interrupt occurs.
         */
	pCtr->reloadValue = 0;
        *pReload = pInst->ctrRollover;
        return OK;
        }
    pCtr->reloadValue = pInst->ctrRollover - *pReload;
    pCtr->countValue = pCtr->reloadValue;

    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwCreateCtrForEvent
    (
    void *		  pDrvInst,
    void *		  pEvtSettings,
    HPC_COUNTER *	  pReload,
    struct hpcHwCtrData * pCtr,
    void *		  pClient,
    HPC_EVENT_ITER_FUNC	  pEventIterFunc,
    void *		  pIterArg
    )
    {
    STATUS status;
    struct driverInstArchV2 * pInst = pDrvInst;
    struct i86Arch2Settings * pSettings = pEvtSettings;
    struct hpcHwCtrData * pCtrIter;
    int counterBitShift;

    if (pDrvInst == NULL || pEvtSettings == NULL)
	return ERROR;

    /* determine if the counter is already in use */
    while ((pCtrIter = (*pEventIterFunc)(pIterArg)) != NULL)
	if (pCtrIter->counter == pSettings->counter)
		return ERROR;

    pCtr->counter = pSettings->counter;
    pCtr->pClient = pClient;

    if ((status = hpcHwUpdateCtrReload(pInst, pCtr, pReload)) != OK)
        {
        printf("unable to get hw reload value\n");
        return status;
        }

    counterBitShift = pCtr->counter << 2;
    pCtr->fixedCounterAddrs = pInst->msrFixedCtr + pCtr->counter;
    pCtr->ctrlEnableMask = 0x3 << counterBitShift;

    if (pInst->archPerfMonVer == I86_ARCH_PM_DRV2_UNCORE)
        pCtr->ctrCtrlValue = FIXED_PMI_UNCORE | FIXED_ENABLE_UNCORE;
    else
        pCtr->ctrCtrlValue = FIXED_PMI | FIXED_ENABLE_OS | FIXED_ENABLE_US;

    if (pInst->archPerfMonVer == I86_ARCH_PM_DRV2_V3)
        {
        pCtr->ctrCtrlValue |= FIXED_ANY;
        pCtr->ctrlMaskNot = 0xfULL << counterBitShift;
        }
    else if (pInst->archPerfMonVer == I86_ARCH_PM_DRV2_V2)
        pCtr->ctrlMaskNot = 0xbULL << counterBitShift;
    else
        pCtr->ctrlMaskNot = 0x5ULL << counterBitShift; /* UNCORE bits */

    pCtr->ctrlMaskNot = NOT(pCtr->ctrlMaskNot);
    pCtr->ctrCtrlValue <<= counterBitShift;

    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwDriverInit
    (
    void *			 pDrvInst,
    struct hpcHwDriverInitData * pDrvInstInitData,
    size_t *			 sizeOfHwPrivData
    )
    {
    STATUS rc = OK;
    struct driverInstArchV2 * pInst = pDrvInst;
    int bit;

    pInst->pHandler = pDrvInstInitData->pPerfIntCallback;
    *sizeOfHwPrivData = sizeof(struct hpcHwCtrData);

    for (bit = 0; bit < pInst->numCounterBits; bit++)
	pInst->ctrRollover |= (1ULL << bit);

    for (bit = 0; bit < pInst->numCounters; bit++)
	pInst->ctrCounterMask |= 1 << bit;

    if (pInst->enablePerfIsr != NULL)
        if ((rc = pInst->enablePerfIsr(perfIntHandler)) != OK)
            printf("failed to enable i86 v2 perf isr %u\n", rc);

    return rc;
    }

/******************************************************************************
*/

static STATUS hpcHwDriverFinit
    (
    void * pDrvInst
    )
    {
    STATUS rc = OK;
    struct driverInstArchV2 * pInst = pDrvInst;
    int force = 1; /* WRS_TODO add this as a parameter... */

    if (pInst->disablePerfIsr != NULL)
        if ((rc = pInst->disablePerfIsr(force)) != OK)
            printf("failed to disable i86 v2 perf isr %d, %s\n", rc,
                      pInst->pDriverName);
    return rc;
    }

/******************************************************************************
*/

static void resetRegisters
    (
    struct driverInstArchV2 * pInst
    )
    {
    int n;
    long long int value = 0ULL;

    for (n = 0; n < pInst->numCounters; n++)
        pentiumMsrSet(pInst->msrFixedCtr + n, &value);

    pentiumMsrSet(pInst->msrFixedCtrCtrl, &value);
    pentiumMsrGet(pInst->msrGlobalStatus, &value);
    pentiumMsrSet(pInst->msrGlobalOvfCtrl, &value);
    }

/******************************************************************************
*/

static STATUS hpcHwPerCoreInit
    (
    void * pDrvInst
    )
    {
    struct driverInstArchV2 * pInst = pDrvInst;
    STATUS rc = OK;

    resetRegisters(pInst);

    if (pInst->enablePerfIsrPerCore != NULL)
        if ((rc = pInst->enablePerfIsrPerCore(perfIntHandler)) != OK)
            printf("failed to enable i86 v2 perf isr per core %u\n", rc);

    return rc;
    }

/******************************************************************************
*/

static STATUS hpcHwPerCoreFinit
    (
    void * pDrvInst,
    int    force
    )
    {
    struct driverInstArchV2 * pInst = pDrvInst;
    STATUS rc = OK;

    resetRegisters(pInst);

    if (pInst->disablePerfIsrPerCore != NULL)
        if ((rc = pInst->disablePerfIsrPerCore(force)) != OK)
            printf("failed to disable i86 v2 perf isr per core %d, %s\n", rc,
                      pInst->pDriverName);

    return rc;
    }

/******************************************************************************
*/

STATUS hpcHwI86ArchPm2DriverInit(void)
    {
    return OK;
    }

/******************************************************************************
*/

STATUS hpcHwI86ArchPm2DriverFinit(void)
    {
    return OK;
    }

/******************************************************************************
*/

struct hpcHwDriverFuncs hpcHwDrvI86Arch2Rtns =
    {
    hpcHwCtrStart,
    hpcHwCtrStop,
    hpcHwCtrRead,
    hpcHwUpdateCtrReload,
    hpcHwCreateCtrForEvent,
    hpcHwDeleteCtrForEvent,
    hpcHwDriverInit,
    hpcHwDriverFinit,
    hpcHwPerCoreInit,
    hpcHwPerCoreFinit,
    hpcHwDriverShow
    };
