/* perf_lib.c - linux perf-related library routines */

/* Copyright (c) 2011-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
28oct14,tle  Insert dummy gap fields for padding bytes (WB4-4481)
20Aug14,sdt  Add support of space in array name field - WB4-4020

DESCRIPTION

*/

#include <string.h>  /* for memcpy() */
#include <ctype.h>  /* for isblank() */
#include <errno.h>
#include <assert.h>
#include <sys/vfs.h>
#include <linux/magic.h>  /* for DEBUGFS_MAGIC */
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include "../include/perf_lib.h"


#undef DEBUG_EVENTS
#if defined(DEBUG_EVENTS)
#  define dprintf(fmt, ...)			\
	fprintf(stderr, fmt, ## __VA_ARGS__)
#else
#  define dprintf(fmt, ...)
#endif


/* prefix to add to numeric-only subsystem names */
#define DIGIT_SUBSYSTEM_PREFIX  "core"

const char *perf_hw_category_defs[] = {
	PERF_HW_CATEGORY_DEF,
	NULL,
};

const char *perf_sw_category_defs[] = {
	PERF_SW_CATEGORY_DEF,
	NULL,
};

/* perf hardware event definitions */
const char *PERF_HW_VT[] = {
	"S[perf.cycles,,n[cafe.LABEL,\"Cycles\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.instructions,,n[cafe.LABEL,\"Instructions\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.cache_references,,n[cafe.LABEL,\"Cache Accesses\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.cache_misses,,n[cafe.LABEL,\"Cache Misses\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.branches,,n[cafe.LABEL,\"Branches\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.branch_misses,,n[cafe.LABEL,\"Branch Misses\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.bus_cycles,,n[cafe.LABEL,\"Bus Cycles\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.frontend_stalled_cycles,,n[cafe.LABEL,\"Frontend Stalled Cycles\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.backend_stalled_cycles,,n[cafe.LABEL,\"Backend Stalled Cycles\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	"S[perf.ref_cycles,,n[cafe.LABEL,\"Ref Cycles\"]" PERF_HW_CATEGORY_REF "n[cafe.COUNTER_METRIC_CLASS,1]" "]",
	NULL
};

/* perf software event definitions */
const char *PERF_SW_VT[] = {
	"S[perf.cpu_clock_msecs]",
	"S[perf.task_clock_msecs]",
	PERF_SW_PAGE_FAULTS_VT_DEF,
	"S[perf.context_switches,,n[cafe.LABEL,\"Context Switches\"]" PERF_SW_CATEGORY_REF cafe_COUNTER_METRIC_CLASS_VT_NAT "]",
	"S[perf.CPU_migrations,,n[cafe.LABEL,\"CPU Migrations\"]" PERF_SW_CATEGORY_REF cafe_COUNTER_METRIC_CLASS_VT_NAT "]",
	"S[perf.minor_faults,,n[cafe.LABEL,\"Minor Faults\"]" PERF_SW_CATEGORY_REF cafe_COUNTER_METRIC_CLASS_VT_NAT "]",
	"S[perf.major_faults,,n[cafe.LABEL,\"Major Faults\"]" PERF_SW_CATEGORY_REF cafe_COUNTER_METRIC_CLASS_VT_NAT "]",
	"S[perf.alignment_faults,,n[cafe.LABEL,\"Alignment Faults\"]" PERF_SW_CATEGORY_REF cafe_COUNTER_METRIC_CLASS_VT_NAT "]",
	"S[perf.emulation_faults,,n[cafe.LABEL,\"Emulation Faults\"]" PERF_SW_CATEGORY_REF cafe_COUNTER_METRIC_CLASS_VT_NAT "]",
	NULL
};

/**
 *  Return the valuetype string corresponding to the given perf event ID.
 * Returns null on error.
 */
const char *perf_eid_to_vt(uint32_t perf_eid)
{
	switch(perf_eid) {
	case PERF_RECORD_SAMPLE:
		return PERF_SAMPLE_DEF;
	case PERF_RECORD_MMAP:
		return PERF_MMAP_DEF;
	case PERF_RECORD_COMM:
		return PERF_COMM_DEF;
	case PERF_RECORD_FORK:
		return PERF_FORK_DEF;
	case PERF_RECORD_EXIT:
		return PERF_EXIT_DEF;
	case PERF_RECORD_LOST:
		return PERF_RECORD_LOST_DEF;
	case PERF_RECORD_READ:
		return PERF_READ_DEF;
	case PERF_RECORD_THROTTLE:
		return PERF_THROTTLE_DEF;
	case PERF_RECORD_UNTHROTTLE:
		return PERF_UNTHROTTLE_DEF;
	}
	return NULL;
}


/**
 *  Return the valuetype string corresponding to the given perf hardware cache
 * event configuration.  Returns null on error.
 */
const char *perf_hw_cache_valuetype(enum perf_hw_cache_id type,
				    enum perf_hw_cache_op_id op,
				    enum perf_hw_cache_op_result_id result)
{
	static const char *hw_cache[] = {
		"L1_dcache",
		"L1_icache",
		"LLC",
		"dTLB",
		"iTLB",
		"branch",
		"node",
		NULL
	};

	static const char *hw_cache_op[] = {
		"load",
		"store",
		"prefetch",
		NULL
	};

	static const char *hw_cache_result[] = {
		"accesses",
		"misses",
		NULL
	};
	char buf[256];

	if (type >= sizeof(hw_cache) || op >= sizeof(hw_cache_op)
	    || result >= sizeof(hw_cache_result))
		return NULL;

	if (hw_cache[type] == NULL || hw_cache_op[op] == NULL
	    || hw_cache_result[result] == NULL)
		return NULL;

	snprintf(buf, sizeof buf,
		 "S[perf.%s_%s_%s,,"
		 "n[cafe.COUNTER_METRIC_CLASS,1]"
		 "n[cafe.LABEL,\"%s-%s%s%s%s\"]" PERF_HW_CATEGORY_REF "]",
		 hw_cache[type], hw_cache_op[op], hw_cache_result[result],
		 hw_cache[type],
		 hw_cache_op[op],
		 result == 1 ? "" : "s",
		 result == 1 ? "-" : "",
		 result == 1 ? hw_cache_result[result] : "");

	return strdup(buf);
}


/**
 *  Returns the "config" value needed to express the given hardware cache event
 * to the perf_event attribute.
 */
uint64_t perf_hw_cache_config(enum perf_hw_cache_id type,
			      enum perf_hw_cache_op_id op,
			      enum perf_hw_cache_op_result_id result)
{
	uint64_t config;

	config = type;
	config |= op << 8;
	config |= result << 16;

	return config;
}


/**
 * A convenience routine; perf event data has strings embedded, but our
 * valuetype routines expect all strings to be indirect.  The "out" buffer must
 * be large enough to accomodate the written data.
 */
//td: support buffer size bounding checks
bool relocate_strings(const struct cafe_value_type *type, const char *in,
		      char *out)
{
	struct cafe_vm_attr *attr;

	cafe_type_attr_foreach(attr, type) {
		const struct cafe_value_type *subtype = cafe_attr_type(attr);
		switch(subtype->kind) {
		case CAFE_TYPE_STRING:
		{
			size_t len = strlen(in) + 1;
			char **addr = (char **) (out + attr->offset);

			*addr = (char *)in;
			in += len;
			break;
		}
		default:
		{
			size_t size = cafe_attr_type(attr)->size;
			cafe_value_copy(subtype, out + attr->offset, in,
					cafe_stdlib_allocator);
			in += size;
		}
		}
	}
	return true;
}


/**
 *  Return the single-character representation of a given cafe_value_type_tag.
 */
static
char vt_type_id(cafe_value_type_tag tag)
{
	/* needs to stay in sync with value_type.h */
	char tag_map[] = {
		'c',  /* CAFE_TYPE_CHAR    0 */
		'C',  /* CAFE_TYPE_UCHAR */
		'h',  /* CAFE_TYPE_SHORT */
		'H',  /* CAFE_TYPE_USHORT */
		'i',  /* CAFE_TYPE_INT */
		'I',  /* CAFE_TYPE_UINT    5 */
		'l',  /* CAFE_TYPE_LONG */
		'L',  /* CAFE_TYPE_ULONG */
		'q',  /* CAFE_TYPE_LONGLONG */
		'Q',  /* CAFE_TYPE_ULONGLONG, */
		's',  /* CAFE_TYPE_STRING  10 */
#if 0
	CAFE_TYPE_BLOB,
	CAFE_TYPE_DICT,
	CAFE_TYPE_BOOL,
	CAFE_TYPE_TYPE,
	CAFE_TYPE_FLOAT,		/* 15 */
	CAFE_TYPE_DOUBLE,
	CAFE_TYPE_NATURE_DEF,
	CAFE_TYPE_STRUCT,
	CAFE_TYPE_ARRAY,
	CAFE_TYPE_DYNARRAY,		/* 20 */
	CAFE_TYPE_AGGR_REF,
	CAFE_TYPE_TYPEDEF,
	CAFE_TYPE_INSTANCE,
#endif
	};

	if (tag > CAFE_TYPE_STRING)
		return 0;

	return tag_map[tag];
}


/**
 *  Read a null-terminated string from the given file stream and return a
 * malloc()ed buffer with a copy.
 */
char *read_string(FILE *in)
{
	char buf[1024];
	unsigned int i;

	for (i = 0; i < sizeof(buf); ++i) {
		if (fread_unlocked(&buf[i], 1, 1, in) != 1)
			return NULL;

		if ((buf[i] == '\0') || (buf[i] == '\n')) {
			char *str = malloc(i + 1);
			if (str == NULL)
				return NULL;
			memcpy(str, buf, i+1);
			if (buf[i] == '\n')
				str[i] = '\0';
			return str;
		}
	}

	return NULL;
}


/**
 *  Read a string from the given file stream and compare it to the given
 * "expected" string.  If they match, return true, otherwise rewind the stream
 * to the same offset seen before calling this routine and return false.
 */
bool expect_string(const char *str, FILE *in, bool null_terminated)
{
	char buf[32];
	const long off = ftell(in);
	size_t len = strlen(str);

	if (null_terminated)
		++len;

	if (fread_unlocked(buf, len, 1, in) != 1)
		goto fail;
	if (memcmp(buf, str, len) != 0)
		goto fail;
	return true;
fail:
	printf("Failed to find expected string \"%s\" at 0x%lx\n", str, off);
	return false;
}


/**
 *  Return the number of digits in a base 10 representation of the given
 * integer.
 */
static
unsigned int count_digits(unsigned long n)
{
	unsigned int i = 0;

	do {
		n /= 10;
		i++;
	} while (n != 0);

	return i;
}


/**
 *  Maps cafe_value_type_tag values to a human-readable string.
 */
static const char * vt_descr[] = {
	"char",			/* 0 */
	"unsigned char",
	"short",
	"unsigned short",
	"int",
	"unsigned int",		/* 5 */
	"long",
	"unsigned long",
	"long long",
	"unsigned long long",
	"string",		/* 10 */
	"blob",
	"dict",
	"bool",
	"type",
	"float",		/* 15 */
	"double",
	"nature",
	"struct",
	"array",
	"dynarry",		/* 20 */
	"aggregation",
	"typedef",
	"instance",
	"function"
};


static inline
void dump_event(const struct linux_event *e)
{
	unsigned int i;
	printf("%u: %s (%u)\n", e->id, e->name, e->n_parameters);
	for (i = 0; i < e->n_parameters; ++i)
		printf("\t%s %s (%lu, %lu)\n", vt_descr[e->params[i].type],
		       e->params[i].name, e->params[i].size,
		       e->params[i].offset);
}


/**
 *  Given a linux trace event record, construct its value type representation,
 * and return it as an allocated string (which callers should free).
 */
int trace_event_valuetype(const struct linux_event *e, char **vt_string,
			  const struct cafe_ptr_array *natures,
			  const struct cafe_ptr_array *fields)
{
	unsigned int i;
	unsigned int gap_index = 0;
	size_t len = 3;		/* "S[]" */
	char *type;
	int idx;

	len += strlen(e->name);
	if (natures != NULL) {
		if (cafe_ptr_array_len(natures) > 0)
			len++;	/* for ',' */
		for (i = 0; i < cafe_ptr_array_len(natures); i++)
			len += strlen((char *)cafe_ptr_array_val(natures, i));
	}
	if (e->n_parameters > 0)
		++len;  /* for ',' */
	for (i = 0; i < e->n_parameters; ++i) {
		const char *is_array = strchr(e->params[i].name, '[');

		len += 5;  /* v[,x] */
		len += strlen(e->params[i].name);
 		if (is_array != NULL)
			len -= strlen(is_array);
		if (e->params[i].array_size > 0) {
			len += 4;  /* a[x,] */
			len += count_digits(e->params[i].array_size);
		}
	}

	if (fields != NULL) {
		for (i = 0; i < e->n_parameters; i++) {
			size_t j;

			for (j = 0; j < cafe_ptr_array_len(fields); j++) {
				struct field_type_annotation *fa;
				fa = cafe_ptr_array_val(fields, j);
				if (strcmp(e->params[i].name, fa->name) == 0) {
					break;
				}
			}

			if (j != cafe_ptr_array_len(fields)) {
				struct field_type_annotation *fa;

				fa = cafe_ptr_array_val(fields, j);
				if (cafe_ptr_array_len(&fa->natures) > 0)
					len += 1; /* for ',' */
				for (j = 0; j < cafe_ptr_array_len(&fa->natures); j++)
					len += strlen((char *)cafe_ptr_array_val(&fa->natures, j));
			}
		}
	}

	/* Determine length of gap fields */
	for (i = 1; i < e->n_parameters; i++) {
		unsigned long gap = e->params[i].offset - e->params[i-1].offset - e->params[i-1].size;
		while (gap) {
			/* For each set bit in the gap we need a field */
			if (gap & 1)
				len += 9; /* v[gapN,x] */
			gap >>= 1;
		}
	}

	type = malloc(len + 1);	/* + '\0' */
	if (type == NULL)
		return CAFE_NOMEM;
	idx = sprintf(type, "S[%s", e->name);
	if (e->n_parameters > 0)
		type[idx++] = ',';
	for (i = 0; i < e->n_parameters; ++i) {
		char type_id = vt_type_id(e->params[i].type);
		const char *is_array = strchr(e->params[i].name, '[');

		if (type_id == 0) {
			printf("Unknown type %u\n", e->params[i].type);
			free(type);
			return CAFE_ERROR;
		}

		if (i > 0) {
			/* Insert gap field(s) if necessary */
			unsigned long gap = e->params[i].offset - e->params[i-1].offset - e->params[i-1].size;
			unsigned int field_size = 1;
			while (gap) {
				if (gap & 1) {
					type[idx++] = 'v';
					type[idx++] = '[';
					type[idx++] = 'g';
					type[idx++] = 'a';
					type[idx++] = 'p';
					if (gap_index < 10)
						type[idx++] = (char) gap_index + '0';
					else
						type[idx++] = (char) gap_index - 10 + 'A';
					++gap_index;
					type[idx++] = ',';
					if (field_size == sizeof(char))
						type[idx++] = vt_type_id(CAFE_TYPE_CHAR);
					else if (field_size == sizeof(short))
						type[idx++] = vt_type_id(CAFE_TYPE_SHORT);
					else if (field_size == sizeof(int))
						type[idx++] = vt_type_id(CAFE_TYPE_INT);
					else if (field_size == sizeof(long))
						type[idx++] = vt_type_id(CAFE_TYPE_LONG);
					else if (field_size == sizeof(long long))
						type[idx++] = vt_type_id(CAFE_TYPE_LONGLONG);
					else {
						printf("Gap too large: %u\n", field_size);
						free(type);
						return CAFE_ERROR;
					}
					type[idx++] = ']';
				}
				gap >>= 1;
				field_size <<= 1;
			}
		}
		type[idx++] = 'v';
		type[idx++] = '[';
		if (is_array == NULL) {
			strcpy(&type[idx], e->params[i].name);
			idx += strlen(e->params[i].name);
		} else {
			unsigned long l = is_array - e->params[i].name;
			memcpy(&type[idx], e->params[i].name, l);
			idx += l;
		}
		type[idx++] = ',';
		if (e->params[i].array_size > 0) {
			type[idx++] = 'a';
			type[idx++] = '[';
		}
		type[idx++] = type_id;
		if (e->params[i].array_size > 0)
			idx += sprintf(&type[idx], ",%lu]",
				       e->params[i].array_size);

		if (fields != NULL) {
			size_t j;

			for (j = 0; j < cafe_ptr_array_len(fields); j++) {
				struct field_type_annotation *fa;
				fa = cafe_ptr_array_val(fields, j);
				if (strcmp(e->params[i].name, fa->name) == 0)
					break;
			}

			if (j != cafe_ptr_array_len(fields)) {
				struct field_type_annotation *fa;

				fa = cafe_ptr_array_val(fields, j);
				if (cafe_ptr_array_len(&fa->natures) > 0)
					type[idx++] = ',';
				for (j = 0; j < cafe_ptr_array_len(&fa->natures); j++) {
					const char *s;
					s = (const char *)cafe_ptr_array_val(&fa->natures, j);
					strcpy(&type[idx], s);
					idx += strlen(s);
				}
			}
		}

		type[idx++] = ']';
	}

	if (natures != NULL) {
		if (cafe_ptr_array_len(natures) > 0)
			type[idx++] = ',';
		for (i = 0; i < cafe_ptr_array_len(natures); i++) {
			const char *s = cafe_ptr_array_val(natures, i);
			strcpy(&type[idx], s);
			idx += strlen(s);
		}
	}

	type[idx++] = ']';
	type[idx++] = '\0';
	*vt_string = type;
	assert(strlen(*vt_string) == len);

	return CAFE_OK;
}


static inline
void consume_whitespace(const char **str)
{
	while(isblank(**str))
		++*str;
}


static inline
bool consume_token(const char *val, const char **str)
{
	size_t len = strlen(val);
	if (strncmp(*str, val, len) != 0)
		return false;

	*str += len;
	return true;
}


static inline
bool consume_ident(char **val, const char **str)
{
	const char *trav = *str;
	bool array_def = false;

	assert(str != NULL);
	while (*trav != ';') {
		if (*trav == '[')
			array_def = true;
		if (*trav == ']')
			array_def = false;
		if (!array_def && (*trav == ' '))
			break;
		if (*trav == '\n')
			return false;
		++trav;
	}

	if (val != NULL) {
		size_t len = trav - *str;

		*val = malloc(len + 1);
		memcpy(*val, *str, len);
		(*val)[len] = 0;
	}
	*str = trav;
	return true;
}


static inline
void consume_ulong(unsigned long *val, const char **str)
{
	*val = strtoul(*str, (char **) str, 0);
}


/**
 *  Given a field line from a trace event format file, of the form:
 *     field:<type> <name>;   offset:<int>;       size:<int>; signed:<int>;
 * parse it, and populate the given linux_event_param structure.
 */
static
bool parse_field(const char *line, struct linux_event_param *param)
{
#if defined(DEBUG_EVENTS)
	const char *original_line = line;
#endif
	const char *param_name = NULL;
	bool is_integral_type = false;

	consume_whitespace(&line);
	if (!consume_token("field:", &line) != 0) {
		dprintf("Error parsing field token from %s\n", line);
		return false;
	}

	is_integral_type = true;
	do {
		if (consume_token("* ", &line)) {
			param->type = CAFE_TYPE_ULONG;
			is_integral_type = false;
			continue;
		}
		if (consume_token("char ", &line)) {
			param->type = CAFE_TYPE_CHAR;
			is_integral_type = false;
			continue;
		}

		param_name = line;
		if (!consume_ident(NULL, &line)) {
			dprintf("Unable to parse type: %s", original_line);
			return false;
		}
		if (consume_token(" ", &line))
			continue;

	} while(!consume_token(";", &line));

	if (!consume_ident(&param->name, &param_name)) {
		dprintf("Error parsing field identifier from \"%s\" (%s)\n",
			param_name, original_line);
		return false;
	}

	/* check for array variants */
	param->array_size = 0;
	char *array_size = strchr(param->name, '[');
	if (array_size != NULL) {
		*array_size = 0;
		++array_size;
		if ((!is_integral_type) && (param->type == CAFE_TYPE_CHAR))
			param->type = CAFE_TYPE_STRING;
		else if (isdigit(*array_size)) {
			unsigned long size = strtoul(array_size, NULL, 10);
			param->array_size = size;
		} else {
			/* the size is specified as a string constant,
			   may need a way to handle these cases */
		}
	}

	consume_whitespace(&line);

	if (!consume_token("offset:", &line) != 0) {
		dprintf("Error parsing offset token from %s\n", line);
		goto fail_and_free_name;
	}
	consume_ulong(&param->offset, &line);
	if (!consume_token(";", &line)) {
		dprintf("Error parsing offset delimiter from %s\n", line);
		goto fail_and_free_name;
	}
	consume_whitespace(&line);

	if (!consume_token("size:", &line) != 0) {
		dprintf("Error parsing offset token from %s\n", line);
		goto fail_and_free_name;
	}
	consume_ulong(&param->size, &line);

	if (is_integral_type) {
		/* use size and sign to determine the most appropriate
		   value-type */
		unsigned long is_signed;
		unsigned int type_size;

		if (!consume_token(";", &line)) {
			dprintf("Error parsing size delimiter from %s\n", line);
			goto fail_and_free_name;
		}
		consume_whitespace(&line);

		if (!consume_token("signed:", &line) != 0) {
			dprintf("Error parsing signed token from %s\n", line);
			goto fail_and_free_name;
		}
		consume_ulong(&is_signed, &line);

		type_size = (param->array_size == 0) ?
			param->size : param->size / param->array_size;
			
		if (type_size == sizeof(char)) {
			if (is_signed)
				param->type = CAFE_TYPE_CHAR;
			else
				param->type = CAFE_TYPE_UCHAR;
		} else if (type_size == sizeof(short)) {
			if (is_signed)
				param->type = CAFE_TYPE_SHORT;
			else
				param->type = CAFE_TYPE_USHORT;
		} else if (type_size == sizeof(int)) {
			if (is_signed)
				param->type = CAFE_TYPE_INT;
			else
				param->type = CAFE_TYPE_UINT;
		} else if (type_size == sizeof(long)) {
			if (is_signed)
				param->type = CAFE_TYPE_LONG;
			else
				param->type = CAFE_TYPE_ULONG;
		} else if (type_size == sizeof(long long)) {
			if (is_signed)
				param->type = CAFE_TYPE_LONGLONG;
			else
				param->type = CAFE_TYPE_ULONGLONG;
		} else {
			dprintf("Unknown assumed integral type\n\t%s\n",
				original_line);
			goto fail_and_free_name;
		}
	}

	return true;

fail_and_free_name:
	free(param->name);
	return false;
}


/**
 *  Given a file stream pointing to the beginning of a trace event format
 * specification, parse it return a pointer to an allocated linux_event
 * structure (which should be released with release_linux_event() when no
 * longer in use.
 */
struct linux_event *parse_event_format(const char *subsystem, FILE *in)
{
	char *line = NULL;
	char *name;
	size_t line_size;
	long param_offset;
	unsigned int i;
	unsigned int n_params;
	unsigned int id;
	struct linux_event *event;
	bool in_common_section;

	assert(subsystem != NULL);
	assert(in != NULL);
	if (!expect_string("name: ", in, false))
		return NULL;
	name = read_string(in);

	if (!expect_string("ID: ", in, false))
		goto free_name_and_fail;

	if (fscanf(in, "%u\n", &id) != 1) {
		dprintf("Unable to parse ID\n");
		goto free_name_and_fail;
	}

	if (getline(&line, &line_size, in) == -1) {
		dprintf("Error reading (%u)\n", __LINE__);
		goto free_name_and_fail;
	}

	if (strcmp("format:\n", line) != 0) {
		dprintf("Unexpected input (\"%s\") while looking for "
			"\"format:\"\n", line);
		goto free_name_and_fail;
	}

	/* count parameters */
	param_offset = ftell(in);
	in_common_section = true;
	n_params = 0;
	while (getline(&line, &line_size, in) != -1) {
		if (strcmp("\n", line) == 0)
			break;
		if (strstr(line, "common_type") != NULL)
			continue;
		++n_params;
	}
	while (getline(&line, &line_size, in) != -1) {
		if (strcmp("\n", line) == 0)
			break;
		++n_params;
	}

	event = malloc(sizeof(*event)
		       + n_params * sizeof(struct linux_event_param));
	if (event == NULL)
		goto free_name_and_fail;

	event->n_parameters = 0;
	event->name = NULL;
	in_common_section = true;
	fseek(in, param_offset, SEEK_SET);
	for (i = 0; i < n_params; ++i) {
		if (getline(&line, &line_size, in) == -1)
			break;
		if (strcmp("\n", line) == 0) {
			if (in_common_section) {
				in_common_section = false;
				if (getline(&line, &line_size, in) == -1)
					break;
			} else
				break;
		}

		if (strstr(line, "common_type") != NULL) {
			--i;
			continue;
		}

		if (!parse_field(line, &event->params[i])) {
			dprintf("error parsing event %s/%s\n", subsystem, name);
			goto free_event_and_fail;
		}
		++event->n_parameters;
	}
	free(line);
	line = NULL;
	if (i != n_params)
		goto free_event_and_fail;

	if (isdigit(subsystem[0])) {
		event->name = malloc(strlen(name) + strlen(subsystem)
				     + strlen(DIGIT_SUBSYSTEM_PREFIX) + 2);
		if (event->name == NULL)
			goto free_event_and_fail;

		sprintf(event->name, DIGIT_SUBSYSTEM_PREFIX "%s.%s",
			subsystem, name);
	} else {
	    event->name = malloc(strlen(name) + strlen(subsystem) + 2);
	    if (event->name == NULL)
		    goto free_event_and_fail;

	    sprintf(event->name, "%s.%s", subsystem, name);
	}

	free(name);
	for (name = event->name; *name != 0; ++name) {
		if (*name == '-')
			*name = '_';
	}

	event->id = id;
	event->n_parameters = n_params;

	return event;

free_event_and_fail:
	release_linux_event(event);
	free(line);
free_name_and_fail:
	free(name);
	return NULL;
}


/**
 *  Release any memory claimed by a linux_event structure created by a call to
 * parse_event_format().
 */
void release_linux_event(struct linux_event *e)
{
	unsigned int i;

	for (i = 0; i < e->n_parameters; ++i)
		free(e->params[i].name);
	free(e->name);
	free(e);
}


/**
 *  Mount debugfs if necessary.  Returns true iff debugfs is mounted as
 * /sys/kernel/debug.
 */
bool mount_debugfs(void)
{
	struct statfs statbuf;

	if (statfs("/sys/kernel/debug", &statbuf) != 0)
		return false;

	if (statbuf.f_type == DEBUGFS_MAGIC)
		return true;

	if (system("mount -t debugfs none /sys/kernel/debug") == -1)
		return false;

	return true;
}
