/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
08oct14,tle  Fix multiple symbols for same location (WB4-4180)
26aug14,tle  Register system_info EP for wrlinux targets (US30550)
02sep14,tle  Always register symresolve_actions (WB4-4108)
*/

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#if CAFE_OS_UNIX
#include <signal.h>
#endif
#if defined(HAVE_CAFE_CONFIG_H)
#include <cafe/config.h>
#endif
#if defined(HAVE_PR_SET_NAME)
#include <sys/prctl.h>
#endif
#include <cafe/cafe.h>
#include <cafe/cafe_print.h>
#include <cafe/errors.h>
#include <cafe/script.h>
#include <cafe/ep/print.h>
#include <cafe/ep/text_file_provider.h>
#include <cafe/ep/file_stream_reader.h>
#include <cafe/ep/file_stream_writer.h>
#include <cafe/ep/timer_ep.h>
#include <cafe/ep/timestamp_normalizer.h>
#include <cafe/ep/realtime_ep.h>
#include <cafe/ep/aggregator.h>
#include <cafe/ep/config_service.h>
#if defined(HAVE_LINUX_PERF)
#include <cafe/ep/linux_trace.h>
#endif
#if defined(__linux__)
#include <cafe/ep/ostimer.h>
#include <cafe/ep/procfs_cfg.h>
#endif
#if defined(HAVE_UTSNAME_H) && defined(HAVE_UNAME) && !defined(VALUE_ADD)
#include <cafe/ep/sysinfo.h>
#endif
#if defined(HAVE_TCF)
#include <tcf/framework/events.h>
#include <tcf/framework/context.h>
#include <tcf/framework/context-dispatcher.h>
#include <tcf/framework/trace.h>
#include <tcf/framework/channel.h>
#include <tcf/services/pathmap.h>
#include <tcf/services/memorymap.h>
#include "../ep/tcf_utils.h"
#endif
#include "../src/cafe_trace.h"
#include "../src/time_utils.h"
#include "../src/cafe_os.h"


#if defined(__GNUC__)
static void die_with_cafe_error(int rc, const char *msg, ...) __attribute__((format(printf, 2, 3)));
#endif

static int cliopt_t_flag;	   /* cafe trace flag */
static int cliopt_p_flag;	   /* parse only flag */
static int cliopt_timeout_value;   /* in milliseconds */
static int cliopt_delay_timeout_value; /* in milliseconds */

/* This three-tuple is used to track and be notified when the script instrument
 has finished. */
static pthread_mutex_t mutex;
static pthread_cond_t cond;
static bool instrument_is_running = true;


static struct cafe_list_head pathmap_list = CAFE_LIST_INIT(pathmap_list);


struct pathmap {
	struct cafe_list_head node;
	const char *dest;
	char source[0];
};

static struct cafe_instrument *session;

#if 0
#if !defined(SLURP_BUFSZ)
#define SLURP_BUFSZ	(128)
#endif

static int slurp_stream(FILE *stream, char **result)
{
	char buf[SLURP_BUFSZ];
	char *str = NULL;
	size_t remaining = 0, capacity = 0;

	while (!feof(stream) && fgets(buf, SLURP_BUFSZ, stream) != NULL) {
		size_t bufsz = strlen(buf) + 1;

		if (bufsz > remaining) {
			char *new_str;

			capacity += SLURP_BUFSZ;
			new_str = realloc(str, capacity);
			if (new_str == NULL) {
				free(str);
				str = NULL;
				*result = NULL;
				return -1;
			}
			str = new_str;
			remaining = SLURP_BUFSZ;
			if (capacity == SLURP_BUFSZ)
				*str = '\0';
		}
		strcat(str, buf);
		remaining -= bufsz;
	}
	*result = str;
	return 0;
}
#endif

static void die_with_cafe_error(int rc, const char *msg, ...)
{
	va_list ap;
	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	fprintf(stderr, ": %s\n", cafe_strerror(rc));
	exit(EXIT_FAILURE);
}

static void register_file_stream_ep(void)
{
	int rc;

	/* Register file stream reader/writer configurations. */

	if ((rc = text_file_provider_register()) != CAFE_OK)
		die_with_cafe_error(rc, "cannot register text file provider event processor");

	if ((rc = cafe_fstream_reader_cfg_register()) != CAFE_OK)
		die_with_cafe_error(rc, "cannot register fstream_reader event processor");

	if ((rc = cafe_fstream_writer_cfg_register()) != CAFE_OK)
		die_with_cafe_error(rc, "cannot register fstream_writer event processor");
}

static void register_event_processors(void)
{
	register_file_stream_ep();
	cafe_realtime_ep_register();
	cafe_print_configuration_register();
	cafe_timer_ep_register();
	cafe_timestamp_normalizer_ep_register();
#if defined(__linux__)
	cafe_ostimer_cfg_register();
	procfs_cfg_register();
#endif
	symresolve_aggregator_ini();
	cafe_aggregator_ep_register();
#if defined(HAVE_LINUX_PERF) && defined(HAVE_FTW)
	linux_trace_register();
#endif
#if defined(HAVE_UTSNAME_H) && defined(HAVE_UNAME) && !defined(VALUE_ADD)
	sysinfo_cfg_register();
#endif
}

static void unregister_event_processors(void)
{
	cafe_aggregator_ep_unregister();
	cafe_realtime_ep_unregister();
	cafe_timer_ep_unregister();
	cafe_timestamp_normalizer_ep_unregister();
#if defined(__linux__)
	cafe_ostimer_cfg_unregister();
	procfs_cfg_unregister();
#endif
#if defined(HAVE_LINUX_PERF) && defined(HAVE_FTW)
	linux_trace_unregister();
#endif
#if defined(HAVE_UTSNAME_H) && defined(HAVE_UNAME) && !defined(VALUE_ADD)
	sysinfo_cfg_unregister();
#endif
}

static void print_usage_and_exit(void)
{
	/* Note: Don't promote the -D (delay on exit) as it is an
	 * internal debug aid. */

	fprintf(stderr, "usage: cafe-script "
#ifdef HAVE_TCF
		"[-r <milliseconds>] [-p] [-t] [-L <tcf-log-filename>] [-l <tcf-log-mode>] <filename>\n"
		"  -l -- TCF log mode\n"
		"  -L -- TCF log file; specify \"-L-\" for stdout\n"
		"  -m <source>;<destination> -- add object path mapping\n"
#else
		"[-r <milliseconds>] [-p] [-t] <filename>\n"
#endif
		"  -r -- run for <time>\n"
		"  -p -- parse only\n"
		"  -t -- enable cafe trace\n");
	exit(EXIT_FAILURE);
}

static void error_handler(struct cafe_script *script,
			  const void *client_data,
			  const char *filename,
			  int linenum,
			  const char *fmt, va_list ap)
{
	if (linenum >= 1 && filename != NULL && *filename != '\0') {
		fprintf(stderr, "%s:%d: ", filename, linenum);
	} else if (filename != NULL && *filename != '\0') {
		fprintf(stderr, "%s: ", filename);
	}
	vfprintf(stderr, fmt, ap);
	if (cafe_error_cur.rc != CAFE_OK) {
		fputs(": ", stderr);
		cafe_error_print(NULL, stderr, ": ");
		cafe_error_clear(NULL);
	}
	fputc('\n', stderr);
}

static void stop_notifier(struct cafe_ep_inst *inst)
{
	pthread_mutex_lock(&mutex);
	instrument_is_running = false;
	pthread_cond_broadcast(&cond);
	pthread_mutex_unlock(&mutex);
}


#if CAFE_OS_UNIX
static volatile int got_signal;

static void signal_handler(int signum)
{
	got_signal = 1;
}
static void *signal_processor(void *arg)
{
	struct sigaction sa;
	sigset_t signal_set;

#if defined(HAVE_PR_SET_NAME)
	if (prctl(PR_SET_NAME, "cafe-script-sigproc") != 0)
		perror("Error setting thread name");
#endif

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;

	sa.sa_handler = signal_handler;
	if (sigaction(SIGINT, &sa, 0) != 0) {
		perror("sigaction");
		exit(EXIT_FAILURE);
	}

	sa.sa_handler = signal_handler;
	if (sigaction(SIGTERM, &sa, 0) != 0) {
		perror("sigaction");
		exit(EXIT_FAILURE);
	}

	sigfillset(&signal_set);

	if (pthread_sigmask(SIG_UNBLOCK, &signal_set, NULL) != 0) {
		perror("pthread_sigmask");
		exit(EXIT_FAILURE);
	}

	for (;;) {
		pause();

		if (got_signal) {
			pthread_mutex_lock(&mutex);
			instrument_is_running = false;
			pthread_cond_broadcast(&cond);
			pthread_mutex_unlock(&mutex);
			break;
		}
	}

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = SIG_DFL;

	if (sigaction(SIGINT, &sa, 0) != 0) {
		perror("sigaction");
		exit(EXIT_FAILURE);
	}

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = SIG_DFL;

	if (sigaction(SIGTERM, &sa, 0) != 0) {
		perror("sigaction");
		exit(EXIT_FAILURE);
	}

	return arg;
}
#endif

#if defined(HAVE_TCF)

static void *tcf_event_thread(void *arg)
{
	Protocol *protocol;
	TCFBroadcastGroup *bcg;

#if defined(HAVE_PR_SET_NAME)
	if (prctl(PR_SET_NAME, "cafe-script-tcfevent") != 0)
		perror("Error setting thread name");
#endif
	ini_events_queue();
	protocol = protocol_alloc();
	bcg = broadcast_group_alloc();
	ini_path_map_service(protocol, bcg);
#if defined(SERVICE_MemoryMap) && SERVICE_MemoryMap
	ini_memory_map_service(protocol, bcg);
#endif
#if defined(ENABLE_ContextMux) && ENABLE_ContextMux
	ini_contexts();
	ini_context_dispatcher();
#endif

	/* Signal initialization of event queue */
	pthread_mutex_lock(&mutex);
	pthread_cond_signal(&cond);
	pthread_mutex_unlock(&mutex);

	run_event_loop();

	broadcast_group_free(bcg);
	protocol_release(protocol);

	return NULL;
}

static void tcf_stop(void *arg)
{
	cancel_event_loop();
}
#endif

static void error_notifier(struct cafe_ep_inst *inst, int cafe_errno)
{
	fprintf(stderr, "cafe session error: session=%s, ep=%s: %s\n",
		inst->session->name, inst->spec->name, cafe_strerror(cafe_errno));
}

static bool parse_args(unsigned int argc, const char *argv[],
		       const char **filename,
		       unsigned int *script_argc, const char ***script_argv)
{
	unsigned int optind;

	for (optind = 1; optind < argc; ++optind) {
		const char *opt = argv[optind];
		const char *optarg;
		if (*opt != '-')
			break;

		++opt;
		optarg = optind < argc ? argv[optind + 1] : NULL;

		switch (*opt) {
#if defined(HAVE_TCF)
		case 'L':
			open_log_file(optarg);
			++optind;
			break;
		case 'l':
			parse_trace_mode(optarg, &log_mode);
			++optind;
			break;
		case 'm':
		{
			struct pathmap *p;
			const char *trav;
			size_t len = 0;
			size_t colon = -1;

			for (trav = optarg; *trav != 0; ++trav, ++len)
				if (*trav == ';')
					colon = len;

			if (colon == -1) {
				fprintf(stderr, "Badly formatted pathmap "
					"\"%s\"\n", optarg);
				return false;
			}

			p = malloc(sizeof(*p) + len + 1);
			if (p == NULL)
				exit(-1);

			memcpy(p->source, optarg, len + 1);
			p->dest = p->source + colon + 1;
			p->source[colon] = 0;
			cafe_list_add_last(&p->node, &pathmap_list);
			++optind;
			break;
		}
#endif  /* if defined(HAVE_TCF) */
		case 'r':
			cliopt_timeout_value = atoi(optarg);
			if (cliopt_timeout_value < 0)
				cliopt_timeout_value = 0;
			++optind;
			break;
		case 'D':
			cliopt_delay_timeout_value = atoi(optarg);
			if (cliopt_delay_timeout_value < 0)
				cliopt_delay_timeout_value = 0;
			++optind;
			break;
		case 'p':
			cliopt_p_flag = 1;
			break;
		case 't':
			cliopt_t_flag = 1;
			break;
		default:
			fprintf(stderr, "Unknown switch '%s'\n", argv[optind]);
			return false;
		}
	}

	if (optind == argc) {
		fputs("No input script specified\n", stderr);
		return false;
	}

	*filename = argv[optind];
	++optind;

	*script_argc = argc - optind;
	*script_argv = &argv[optind];
	return true;
}


int main(int argc, const char *argv[])
{
	int rc;
	const char *filename = "";
	FILE *fp = stdin;
	struct timespec delay;
	unsigned int script_argc;
	const char **script_argv;
#if defined(HAVE_TCF)
	pthread_t tcf_event_thread_id;
	struct cafe_list_head *trav, *tmp;
#endif
#if CAFE_OS_UNIX
	pthread_t signal_thread_id;
	sigset_t signal_set;
#endif

	if (!parse_args(argc, argv, &filename, &script_argc, &script_argv))
		print_usage_and_exit();

	fp = fopen(filename, "r");
	if (fp == NULL) {
		fprintf(stderr, "cannot open '%s': %s\n",
			filename, strerror(errno));
		return EXIT_FAILURE;
	}

	rc = pthread_mutex_init(&mutex, NULL);
	if (rc != 0) {
		perror("pthread_mutex_init");
		return EXIT_FAILURE;
	}

	rc = pthread_cond_init(&cond, NULL);
	if (rc != 0) {
		perror("pthread_cond_init");
		return EXIT_FAILURE;
	}

#if CAFE_OS_UNIX
	sigfillset(&signal_set);

	if (pthread_sigmask(SIG_BLOCK, &signal_set, NULL) != 0) {
		perror("pthread_sigmask");
		exit(EXIT_FAILURE);
	}

	rc = pthread_create(&signal_thread_id, NULL, signal_processor, NULL);
	if (rc != 0) {
		perror("pthread_create");
		return EXIT_FAILURE;
	}
#endif
	if ((rc = cafe_lib_init()) != CAFE_OK)
		die_with_cafe_error(rc, "cafe_lib_init() failed");

#if defined(HAVE_TCF)
	cafe_config_channel_up = NULL;
	pthread_mutex_lock(&mutex);
	rc = pthread_create(&tcf_event_thread_id, NULL, tcf_event_thread, NULL);
	if (rc != 0) {
		perror("pthread_create");
		return EXIT_FAILURE;
	}

	/* Need to wait until TCF event queue and path map service
	 * have been initialized before setting path maps */
	rc = pthread_cond_wait(&cond, &mutex);
	pthread_mutex_unlock(&mutex);
	cafe_list_foreach_safe(trav, &pathmap_list, tmp) {
		struct pathmap *p = (struct pathmap *) trav;
		tcf_add_pathmap(p->source, p->dest);
		free(p);
	}
#endif

	CAFE_LOCK();
	register_event_processors();

	rc = cafe_script_parse_from_file(cafe_stdlib_allocator,
					 fp, filename,
					 script_argc, script_argv,
					 error_handler,
					 (void *)filename,
					 &session);
	fclose(fp);
	if (rc != CAFE_OK)
		return EXIT_FAILURE;
	if (cliopt_p_flag)
		return EXIT_SUCCESS;
	if (cliopt_t_flag)	/* must come after cafe_lib_init() */
		cafe_trace = 1;
	session->error_notifier = error_notifier;
	session->stop_notifier = stop_notifier;
	rc = cafe_instrument_start(session);
	if (rc != CAFE_OK)
		die_with_cafe_error(rc, "session start failed");
	CAFE_UNLOCK();
	rc = 0;
	pthread_mutex_lock(&mutex);
	if (cliopt_timeout_value == 0) {
		while (instrument_is_running && (rc == 0))
			rc = pthread_cond_wait(&cond, &mutex);
	} else {
		delay.tv_sec = delay.tv_nsec = 0;
		clock_gettime(CLOCK_REALTIME, &delay);
		timespec_ticks_add(cliopt_timeout_value, MILLIS_PER_SEC, &delay);
		while (instrument_is_running && (rc == 0))
			rc = pthread_cond_timedwait(&cond, &mutex, &delay);
	}
	pthread_mutex_unlock(&mutex);
	if (rc != 0 && rc != ETIMEDOUT)
		fprintf(stderr, "Error during delay: %s", strerror(rc));

#if CAFE_OS_UNIX
	if (!got_signal)
		pthread_kill(signal_thread_id, SIGTERM);
	pthread_join(signal_thread_id, NULL);
	pthread_detach(signal_thread_id);
	sigfillset(&signal_set);
	if (pthread_sigmask(SIG_UNBLOCK, &signal_set, NULL) != 0) {
		perror("pthread_sigmask");
		exit(EXIT_FAILURE);
	}
#endif
	CAFE_LOCK();
	rc = cafe_instrument_delete(session);
	if (rc != CAFE_OK)
		die_with_cafe_error(rc, "cannot delete session");
	unregister_event_processors();
	CAFE_UNLOCK();
#if defined(HAVE_TCF)
	/* This sleep allows the debugging of various TCF events to
	 * take place before the tcf events loops are shutdown (e.g.,
	 * ELF cache cleanup). */
#if CAFE_OS_UNIX
	if (cliopt_delay_timeout_value != 0) {
		delay.tv_sec = delay.tv_nsec = 0;
		clock_gettime(CLOCK_REALTIME, &delay);
		timespec_ticks_add(cliopt_delay_timeout_value, MILLIS_PER_SEC, &delay);
		clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &delay, NULL);
	}
#endif
	post_event(tcf_stop, NULL);
	pthread_join(tcf_event_thread_id, NULL);
#endif
	return rc == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
}
