/* leb-tool.c - quick calculator for leb128 conversions */

/* Copyright (c) 2010-2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>  /* for LINE_MAX */
#include <errno.h>
#include <cafe/stdbool.h>
#include "../cafe/src/leb128.h"
#include "../cafe/src/strtoll.h"

#ifndef LINE_MAX
#define LINE_MAX (4096)
#endif

#if !defined(LLONG_MAX)
#  define LLONG_MAX ((~(long long)0) >> 1)
#endif

#if !defined(ULLONG_MAX)
#  define ULLONG_MAX (~(unsigned long long)0)
#endif

#if !defined(LLONG_MIN)
#  define LLONG_MIN (~((~(long long)0) >> 1))
#endif


static bool opt_signed = false;
static bool opt_encode = false;


static
void usage(const char *progname)
{
	printf("Usage: %s [-s] [-e]\n", progname);
}


static
void parse_cmdline(unsigned int argc, char *argv[])
{
	unsigned int ind;
	int c;

	/* note: we don't use getopt here because that would upset win32 */

	for (ind = 1; ind < argc; ind++) {
		const char *s = argv[ind];
		if (*s != '-') {
			break;
		}
		++s;
		for (c = *s++; c != '\0'; c = *s++) {
			switch (c) {
			case 's':
				opt_signed = true;
				break;

			case 'e':
				opt_encode = true;
				break;

			case '?':
			case 'h':
				usage(argv[0]);
				exit(EXIT_SUCCESS);

			default:
				printf("illegal command line switch '%c'\n", c);
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
		}
	}
}


static
void parse_signed(void)
{
	char line[LINE_MAX];
	long byte;
	long long accum = 0;
	unsigned int shift = 0;

	printf("Parsing input as a signed leb128 stream\n"
	       "Enter one byte per line:\n");
	while(1) {
		if (fgets(line, sizeof(line), stdin) == NULL)
			break;
		
		byte = strtoul(line, 0, 16);

		if ((byte & 128) == 0) {
			if (byte & 64)
				/* Sign-extend */
				byte |= -64;
			accum |= ((long long) byte) << shift;
			printf("Complete: %llx (%lld)\n", accum, accum);
			accum = 0;
			shift = 0;
			continue;
		}

		accum |= (((long long) byte & 127) << shift);
		shift += 7;
		printf("(%llx)\n", accum);
	}

}


static
void parse_unsigned(void)
{
	char line[LINE_MAX];
	unsigned long byte;
	unsigned long long accum = 0;
	unsigned int shift = 0;

	printf("Parsing input as an unsigned leb128 stream\n"
	       "Enter one byte per line:\n");
	while(1) {
		if (fgets(line, sizeof(line), stdin) == NULL)
			break;
		
		byte = strtoul(line, 0, 16);

		accum |= (((unsigned long long)byte & 127) << shift);
		shift += 7;

		if (byte & 128)
			printf("(%llx)\n", accum);
		else {
			printf("Complete: %llx (%llu)\n", accum, accum);
			accum = 0;
			shift = 0;
		}
	}

}


static
void encode_unsigned(void)
{
	char line[LINE_MAX];
	int8_t buf[64];
	unsigned long long val;
	int size;
	int i;

	printf("Encoding input to unsigned leb128 stream\n"
	       "Enter one number per line:\n");
	while(1) {
		if (fgets(line, sizeof(line), stdin) == NULL)
			break;

		val = strtoull(line, 0, 0);
		if ((val == ULLONG_MAX) && (errno == ERANGE)) {
			printf("Value out of range; too big or too small\n");
			continue;
		}

		size = cafe_leb128_mwrite_ull(buf, val);
		for (i = 0; i < size; ++i)
			printf("%hhx  ", buf[i]);
		printf(" (%d bytes)\n", size);
	}
}


static
void encode_signed(void)
{
	char line[LINE_MAX];
	int8_t buf[64];
	long long val;
	int size;
	int i;

	printf("Encoding input to signed leb128 stream\n"
	       "Enter one number per line:\n");
	while(1) {
		if (fgets(line, sizeof(line), stdin) == NULL)
			break;

		val = strtoll(line, 0, 0);
		if (((val == LLONG_MIN) || (val == LLONG_MAX))
		    && (errno == ERANGE))
		{
			printf("Value out of range; too big or too small\n");
			continue;
		}

		size = cafe_leb128_mwrite_ll(buf, val);
		for (i = 0; i < size; ++i)
			printf("%hhx  ", buf[i]);
		printf(" (%d bytes)\n", size);
	}
}


int main(int argc, char *argv[])
{
	parse_cmdline(argc, argv);
	if (opt_encode)
		if (opt_signed)
			encode_signed();
		else
			encode_unsigned();
	else
		if (opt_signed)
			parse_signed();
		else
			parse_unsigned();
	return EXIT_SUCCESS;
}
