/* tcf_utils.c - TCF utilities */

/* Copyright (c) 2009-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
 * TCF Utilities.
 *
 * Contains utility routines which can be generically useful to TCF peers.  At
 * some point, most of these routines should be properly refactored into
 * libraries, which may end up being part of the TCF agent sources proper.
 */

#include <stdio.h>
#include <stdarg.h>
#include <cafe/inttypes.h>
#include <tcf/framework/trace.h>
#include <tcf/framework/exceptions.h>
#include <tcf/framework/json.h>

#include <cafe/stdbool.h>

#include "tcf_utils.pkg.h"


/**
 *
 */
void send_command(Channel *channel, const char *service, const char *command,
		  ReplyHandlerCB handler, const char *fmt, ...)
{
	va_list ap;
	protocol_send_command(channel, service, command, handler, channel);
	va_start(ap, fmt);
	for (; *fmt != 0; ++fmt)
		write_stringz(&channel->out, va_arg(ap, char *));
	va_end(ap);

	write_stream(&channel->out, MARKER_EOM);
}


/**
 *
 */
void display_tcf_reply(Channel *c, void *client_data, int error)
{
	int i;

	if (error) {
		trace(LOG_ALWAYS, "Reply error %d: %s\n",
		      error, errno_to_str(error));
		return;
	}
	putchar('\"');
	for (;;) {
		i = read_stream(&c->inp);
		if (i == MARKER_EOM) break;
		if (i == 0) i = ' ';
		putchar(i);
	}
	putchar('\"');
	putchar('\n');

	/* We flush the stream to be able to connect to the client with pipes
	 * and receive the message when it's displayed */
	fflush(0);
}


/**
 *
 */
void error_struct_callback(InputStream *in, const char *name, void *value)
{
	char buf[1024];
	int size;

	size = snprintf(buf, sizeof(buf), "\t%s: ", name);
	if (strcmp(name, "Time") == 0) {
		snprintf(buf, sizeof(buf) - size,
			 "%"PRId64"\n", json_read_int64(in));
	} else if (strcmp(name, "Code") == 0) {
		snprintf(buf, sizeof(buf) - size,
			 "%"PRId64"\n", json_read_int64(in));
	} else if (strcmp(name, "Format") == 0) {
		json_read_string(in, &buf[size], sizeof(buf) - size);
	} else {
		printf("UNKNOWN\n");
	}
	trace(LOG_ALWAYS, buf);
}


/**
 *
 */
void virtualstream_read(InputStream *in, char **msg, size_t *size,
			int64_t *lost, bool *end)
{
	*msg = json_read_alloc_binary(in, size);
	json_test_char(in, 0);
	if (peek_stream(in) != 0) {
		json_read_struct(in, error_struct_callback, 0);
	}
	json_test_char(in, 0);
	*lost = json_read_int64(in);
	json_test_char(in, 0);
        *end = json_read_boolean(in);
	json_test_char(in, 0);
	json_test_char(in, MARKER_EOM);

	/* We flush the stream to be able to connect to the client with pipes
	 * and receive the message when it's displayed */
	fflush(0);
}
