/* vs_logger.c -  TCF VirtualStream Logger client. */

/* Copyright (c) 2009-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
 * The VirtualStream Logger is a simple TCF agent which connects to the
 * requested virtual stream on the given peer, then listens for an event
 * definition event.  It uses this event to define subsequent events sent on
 * the stream, which it then listens for and emits to stdout.
 */

#define ENABLE_Trace (1)
#if defined(HAVE_CAFE_CONFIG_H)
#  include <cafe/config.h>
#endif

#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <signal.h>
#include <tcf/framework/mdep.h>
#include <tcf/framework/asyncreq.h>
#include <tcf/framework/events.h>
#include <tcf/framework/trace.h>
#include <tcf/framework/json.h>
#include <tcf/framework/protocol.h>
#include <tcf/framework/errors.h>
#include <tcf/framework/exceptions.h>
#include <cafe/stdio.h>
#include <cafe/log_stream.h>
#include <cafe/event_map.h>
#include <cafe/cafe.h>
#include <cafe/value_print.h>
#include "tcf_utils.pkg.h"

static const char *progname;

/* application-specific log messages */
#define LOG_APP         (0x1000)
#define LOG_APP_DEBUG   (0x2000)


/* number of bytes to pull with each read request.  This must be more than the
   target's VirtualStream buffer size (currently 0x1000), otherwise we would
   have to handle re-assembling truncated reads on our end */
static const char *const STREAM_READ_CHUNK = "8192";

/* options gleaned from the command line */
static const char *dest_url;
static const char *stream_id;
static FILE *output_file = 0;
static bool CAFE_VALUE_SHOW_PRIVATE = false;

static Channel *channel;
static Protocol *protocol;
static struct cafe_event_map *event_map;
static struct cafe_value_type *metadata_event;


/* Store for any bytes read from the input stream which can't be decoded
   without the rest of the message. */
static struct cafe_byte_buffer msg_fragment;


static bool header_read = false;

static void display_read_reply(Channel *c, void *client_data, int error);


/**
 *
 */
static
void display_connect_reply(Channel *c, void *client_data, int error)
{
	if (error) {
		trace(LOG_ALWAYS, "Reply error %d: %s\n",
		      error, errno_to_str(error));
		return;
	}
	if (peek_stream(&c->inp) != 0) {
		trace(LOG_ALWAYS, "ERROR connecting to stream:\n");
		json_read_struct(&c->inp, error_struct_callback, 0);
		exit(EXIT_FAILURE);
	}
	json_test_char(&c->inp, 0);
	json_test_char(&c->inp, MARKER_EOM);

	send_command(c, "Streams", "read", display_read_reply,
		     "ss", stream_id, STREAM_READ_CHUNK);

	/* We flush the stream to be able to connect to the client with pipes
	 * and receive the message when it's displayed */
	fflush(0);
}


/**
 *
 */
static
void handle_disconnect_reply(Channel *c, void *client_data, int error)
{
	if (error) {
		trace(LOG_ALWAYS, "DISCONNECT Reply error %d: %s\n",
		      error, errno_to_str(error));
		return;
	}
	if (peek_stream(&c->inp) != 0) {
		trace(LOG_ALWAYS, "ERROR disconnecting to stream:\n");
		json_read_struct(&c->inp, error_struct_callback, 0);
		exit(EXIT_FAILURE);
	}
	json_test_char(&c->inp, 0);
	json_test_char(&c->inp, MARKER_EOM);

	cancel_event_loop();
}


#if 0
static
void dump_words(const void *address, unsigned int len)
{
	const unsigned long *words = address;
	unsigned int i;

	for (i = 0; i < len; ++i) {
		printf("%p:\t%lx\n", &words[i], words[i]);
	}
}
#endif
#if 0
#include <ctype.h>
static
void dump_bytes(const void *address, unsigned int len)
{
	const char *words = address;
	unsigned int i;

	for (i = 0; i < len; ++i) {
		printf("%p:\t%hhx", &words[i], words[i]);
		if (isascii(words[i]))
			printf("\t%c", words[i]);
		printf("\n");
	}
}
#endif


/**
 * Decodes a single event from the given byte stream (bounded by "max_bytes").
 * Returns true iff the event can be decoded successfully, and sets "size"
 * equal to the number of bytes decoded.
 */
static
bool decode_event(char *msg, size_t max_bytes, size_t *size)
{
	unsigned long long eid;
	unsigned long long msg_size;
	int8_t *read_head;
	struct cafe_remote_event *remote;
	void **output_event_data;
	unsigned int i;
	static struct cafe_byte_buffer buf = {NULL};

	if (buf.data == NULL)
		cafe_bbuf_init(&buf, cafe_stdlib_allocator, 1024);

	read_head = cafe_mdecode_event_header((int8_t *)msg, max_bytes,
					      &msg_size, &eid);
	if (read_head == NULL)
		/* we don't have the full header in the given stream */
		return false;

	*size = (size_t)(((char *)read_head - msg) + msg_size);
	if (*size > max_bytes)
		return false;

	if (eid == 0) {
		/* metadata event */
		struct cafe_metadata md;
		struct cafe_remote_event *tmp;
		int rc;

		/* argument values */
		cafe_mdecode_val(metadata_event, &md, cafe_stdlib_allocator,
				 event_map, read_head);
		trace(LOG_APP, "%llx (0x%llx, %u),", md.eid, msg_size,
		      md.event_data.l);
		for (i = 0; i < md.event_data.l; ++i)
			trace(LOG_APP, " \"%s\"",
			      ((char **)md.event_data.p)[i]);
		CAFE_LOCK();
		rc = cafe_event_map_add_event(event_map, &md, NULL, &tmp);
		if (rc != CAFE_OK)
		  trace(LOG_ALWAYS, "Failed to add event description "
			"\"%s\"", ((char **)md.event_data.p)[0]);
		CAFE_UNLOCK();
		cafe_release_values(metadata_event, &md,
				    cafe_stdlib_allocator);

		return true;
	}

	remote = cafe_event_map_lookup(event_map, eid);
	if (remote == NULL) {
		trace(LOG_ALWAYS, "Received unknown event id %llx", eid);
		return true;
	}

	trace(LOG_APP, "E: (%s)", remote->event->types[0]->ident->name);
	/* read event data orthogonal data payload */
	if (cafe_mread_event_data(read_head, remote->event,
				  cafe_stdlib_allocator, event_map,
				  &buf) != CAFE_OK)
	{
		trace(LOG_ALWAYS, "unable to allocate %"PRIsize_t" bytes, "
		      "skipping event", remote->event->types[0]->size);
		return true;
	}

	output_event_data = (void **)buf.data;

	for (i = 0; i < remote->event->num_types; ++i) {
		if (log_mode & LOG_APP) {
			struct cafe_value_print_options options =
				CAFE_VALUE_PRINT_OPTIONS_INIT;
			options.field_names = true;
			fprintf(log_file, "\t");
			cafe_value_print(log_file, NULL,
					 remote->event->types[i],
					 output_event_data[i],
					 &options);
			fprintf(log_file, "\n");
		}
		cafe_release_values(remote->event->types[i],
				    output_event_data[i],
				    cafe_stdlib_allocator);
	}
	return true;
}


/**
 *  Concatenate the given bytes onto the end of any existing message fragment.
 */
static
void cat_message_fragment(const char *msg, size_t size)
{
	int rc = cafe_bbuf_append(&msg_fragment, msg, size);
	if (rc != CAFE_OK) {
		printf("error concatenating message fragments: %s",
		       cafe_strerror(rc));
		exit(EXIT_FAILURE);
	}
}


/**
 *
 */
static
void display_read_reply(Channel *c, void *client_data, int error)
{
	size_t size;
	char *msg;
	char *msg_base;
	bool end;
	int64_t lost;
	size_t processed;
	size_t total_processed;

	if (error) {
		trace(LOG_ALWAYS, "Reply error %d: %s",
		      error, errno_to_str(error));
		return;
	}

	virtualstream_read(&c->inp, &msg, &size, &lost, &end);
	msg_base = msg;

	/* dump output to file, if requested */
	if ((output_file != 0) && (size != 0)) {
		if (fwrite_unlocked(msg, size, 1, output_file) != 1) {
			trace(LOG_ALWAYS, "Unable to write %d bytes: %s",
			      size, strerror(errno));
		}
		fflush_unlocked(output_file);
	}
	trace(LOG_APP_DEBUG, "Read %d bytes", size);

	/* concatenate with any leftover fragments, if necessary */
	if (msg_fragment.len != 0) {
		cat_message_fragment(msg, size);
		free(msg_base);
		msg = msg_base = msg_fragment.data;
		size = msg_fragment.len;
	}

	if (!header_read) {
		int8_t *read_head =
			cafe_mdecode_stream_header((int8_t *)msg, size, NULL);
		if (read_head == NULL)
			fprintf(stderr, "Error reading stream header\n");
		else {
			size -= ((char *)read_head - msg);
			msg = (char *)read_head;
		}
		header_read = true;
	}

	for (total_processed = 0; total_processed < size;
	     total_processed += processed)
	{
		if (!decode_event(msg + total_processed,
				  size - total_processed, &processed))
			break;
	}

	if (total_processed == size) {
		if (msg_fragment.len == 0)
			free(msg_base);
		else
			msg_fragment.len = 0;
	} else {
		const size_t unprocessed = size - total_processed;
		if (msg_fragment.len == 0)
			/* store undecoded bytes as a message fragment */
			cat_message_fragment(msg + total_processed,
					     unprocessed);
		else {
			memmove(msg_fragment.data,
				msg_fragment.data + total_processed,
				unprocessed);
			msg_fragment.len = unprocessed;
		}
	}
	trace(LOG_APP_DEBUG, "Processed %d bytes", total_processed);

	if (end) {
		trace(LOG_APP_DEBUG, "%s closed, exiting.\n", stream_id);
		cafe_bbuf_release(&msg_fragment);
		send_command(channel, "Streams", "disconnect",
			     handle_disconnect_reply, "s", stream_id);
		return;
	}

	send_command(channel, "Streams", "read", display_read_reply, "ss",
		     stream_id, STREAM_READ_CHUNK);
}


static
void channel_connecting(Channel *c)
{
	send_hello_message(c);
}


static
void channel_connected(Channel *c)
{
	/* connect to virtual stream */
	send_command(c, "Streams", "connect", display_connect_reply,
		     "s", stream_id);
}


static
void channel_receive(Channel *c)
{
	handle_protocol_message(c);
}


static
void channel_disconnected(Channel *c)
{
	fprintf(log_file, "Channel closed, exiting..\n");
	if (channel == c)
		channel = NULL;
	protocol_release(protocol);
	exit(EXIT_SUCCESS);
}


static
void connect_done(void *args, int error, Channel *c)
{
	PeerServer *ps = (PeerServer *) args;

	if (error) {
		trace(LOG_ALWAYS, "cannot connect to peer: %s", dest_url);
		exit(EXIT_FAILURE);
	}

	protocol = protocol_alloc();
	protocol_reference(protocol);

	c->connecting = channel_connecting;
	c->connected = channel_connected;
	c->receive = channel_receive;
	c->disconnected = channel_disconnected;
	c->protocol = protocol;
	channel_start(c);
	channel = c;

	peer_server_free(ps);

	trace(LOG_APP, "Connected to \"%s\", reading %s", dest_url, stream_id);
}


static
void usage(void)
{
	printf("Usage: %s [-o output file] [-l logging level] [-p] <url> "
	       "[stream id]\n",
	       progname);
}


static
void parse_cmdline(int argc, char *argv[])
{
	int ind;
	int c;

	/* set some quick defaults */
	log_mode = LOG_TCFLOG | LOG_APP;  // | LOG_PROTOCOL | LOG_EVENTCORE;
	log_file = stdout;

	for (ind = 1; ind < argc; ind++) {
		const char * s = argv[ind];
		if (*s != '-') {
			break;
		}
		s++;
		while ((c = *s++) != '\0') {
			switch (c) {
			case 'o':
			case 'l':
			case 'p':
				if (*s == '\0') {
					if (++ind >= argc) {
						printf("no argument given to option '%c'\n", c);
						exit(EXIT_FAILURE);
					}
					s = argv[ind];
				}
				switch (c) {
				case 'o':
					output_file = fopen(s, "w");
					if (output_file == 0) {
						printf("unable to open output file \"%s\""
						       ": %s\n", s, strerror(errno));
						exit(EXIT_FAILURE);
					}
					printf("Streaming output to file: %s\n", s);
					break;

				case 'l':
					log_mode = strtoul(s, 0, 0);
					printf("Logging at level 0x%x\n", log_mode);
					break;

				case 'p':
					CAFE_VALUE_SHOW_PRIVATE = true;
					break;

				default:
					printf("illegal option '%c'\n", c);
					exit(EXIT_FAILURE);
				}
				s = "";
				break;

			default:
				printf("illegal option '%c'\n", c);
				exit(EXIT_FAILURE);
			}
		}
	}
	if (ind >= argc) {
		usage();
		exit(EXIT_FAILURE);
	}
	dest_url = argv[ind++];

	/* stream id is optional */
	if (ind == argc) {
		stream_id = "\"VS0\"";
		return;
	}

	stream_id = malloc(strlen(argv[ind]) + 3);
	if (stream_id == 0) {
		perror("Constructing stream identifier");
		exit(EXIT_FAILURE);
	}
	sprintf((char *)stream_id, "\"%s\"", argv[ind]);
	++ind;

	if (ind != argc) {
		usage();
		exit(EXIT_FAILURE);
	}
}


int main(int argc, char *argv[])
{
	struct cafe_instrument *session;
	PeerServer *ps;
	int rc;

	progname = argv[0];
	cafe_bbuf_init(&msg_fragment, cafe_stdlib_allocator, 0);
	rc = cafe_lib_init();
	if (rc != CAFE_OK) {
		printf("Error initializing Cafe lib: %s\n",
		       cafe_strerror(rc));
		return 0;
	}
	CAFE_LOCK();
	rc = cafe_instrument_create("vs_logger", &session);
	CAFE_UNLOCK();
	event_map = cafe_event_map_create(session);
	rc = cafe_type_parse(session->ns,
			     cafe_metadata_VT_DEF, &metadata_event);

	parse_cmdline(argc, argv);

#if !defined(_WIN32)
	signal(SIGPIPE, SIG_IGN);
#endif
	ini_mdep();
	ini_trace();
	ini_asyncreq();
	ini_events_queue();

	ps = channel_peer_from_url(dest_url);
	if (ps == NULL) {
		printf("invalid server URL (-s option value): %s\n", dest_url);
		return EXIT_FAILURE;
	}
	channel_connect(ps, connect_done, ps);

	run_event_loop();

	//channel_close(channel);
	cafe_event_map_destroy(event_map);
	CAFE_LOCK();
	rc = cafe_instrument_delete(session);
	CAFE_UNLOCK();

	return EXIT_SUCCESS;
}
