/*******************************************************************************
 * Copyright (c) 2007, 2010, 2013-2014 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * You may elect to redistribute this code under either of these licenses.
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*
 * This module handles process/thread OS contexts and their state machine.
 */

#ifndef D_context_vxworks
#define D_context_vxworks

#include <tcf/config.h>
#include <tcf/framework/context.h>
#include <tcf/services/tcf_elf.h>
#include <tcf/framework/events.h>

/* Defines fake IDs for the main containers of the hierarchy */

#define VX_SYSTEM_FOLDER_ID     1
#define VX_KERNEL_FOLDER_ID     2
#define VX_RTPS_FOLDER_ID       3

/* Defines for IO redirection file descriptors */

#define PIPE_INOUT_MASTER_FD(x)	x->pipe_fds[0][0]
#define PIPE_INOUT_SLAVE_FD(x)	x->pipe_fds[0][1]
#define PIPE_ERR_MASTER_FD(x)	x->pipe_fds[1][0]
#define PIPE_ERR_SLAVE_FD(x)	x->pipe_fds[1][1]

/*
 * VXDBG_POST_EVENT - Macro to invoke post_event() via VxDBG event task.
 *
 * post_event() must not be invoked directly from other tasks than TCF
 * tasks as post_event() (which uses pthread_self()) converts calling
 * task into pthread. In such cases, the following macro must be used.
 *
 * Arguments:
 *      call_back_rtn : post_event call-back routine,
 *      call_back_arg : post_event call-back argument.
 *      static_args   : static POST_EVENT_ARGS provided by caller, or
 *                      NULL when POST_EVENT_ARGS must be allocated/freed.
 */
#define VXDBG_POST_EVENT(call_back_rtn,call_back_arg,static_args) {            \
    if (static_args != NULL) {                                                 \
        ((POST_EVENT_ARGS *)static_args)->call_back = call_back_rtn;           \
        ((POST_EVENT_ARGS *)static_args)->arg = call_back_arg;                 \
        vxdbgEventInject(NULL, vxdbg_post_event, static_args);                 \
    } else {                                                                   \
        POST_EVENT_ARGS * event_args = calloc (1, sizeof (POST_EVENT_ARGS));   \
        event_args->call_back = call_back_rtn;                                 \
        event_args->arg = call_back_arg;                                       \
        vxdbgEventInject(NULL, vxdbg_post_and_free_event, event_args);         \
    }                                                                          \
}

typedef enum
    {
    VX_OBJ_SYS_FOLDER,   /* Obj is System folder */
    VX_OBJ_KERNEL,       /* Obj is VxWorks Kernel */
    VX_OBJ_RTPS_FOLDER,  /* Obj is RTPs folder */
    VX_OBJ_RTP,          /* Obj is an RTP */
    VX_OBJ_TASK,         /* Obj is a TASK */
    VX_OBJ_UNKNOWN       /* Unknown object type */
    } VX_OBJ_TYPE;

typedef struct {
    pid_t          pid;                 /* VxWorks object id */
    pid_t          ppid;                /* VxWorks object parent id */
    VX_OBJ_TYPE    obj_type;            /* VxWorks object type */
    char *	   obj_name;		/* VxWorks object type */
    BOOL           can_attach;		/* Can be attached by debugger? */
    BOOL           attached;            /* Is attached? */
    BOOL           system_container;	/* Is a system container? */
					/* VxWorks, Kernel or RTPs */
    BOOL           task_creation;	/* Support task creation */
    int            max_task_args;       /* Maximum # of task arguments */
    BOOL           rtp_creation;	/* Support RTP creation */

    /* Task specific information */
    int            task_priority;       /* VxWorks task priority */
    char           task_status[16];     /* VxWorks task status */
} VX_OBJ_INFO;

typedef struct {
    Channel *	channel;		/* Channel for sending reply */
    char *	token;			/* Command token */
    const char * exe;			/* RTP file name */
    const char ** args;			/* RTP args */
    const char ** envp;			/* RTP env vars */
    int		priority;		/* RTP init task priority */
    size_t	stack_size;		/* RTP init task stack size */
    int		rtp_options;		/* RTP Spawn options */
    int		init_task_options;	/* RTP init task options */
    void *	prs;			/* ptr to ChildProcess struct */
    int		redirect_io;		/* Redirect STD I/O */
    int		attach;			/* Attach RTP ? */
    int		attach_mode;		/* Attach mode: StopAtMain/Entry */
    pid_t 	pid;			/* RTP id */
    char *	pipe_name;		/* Device name used for IOredirection */
    int 	pipe_fds[2][2];		/* STD IO file descriptors */
    int		err;			/* Spawn errno */
} RTP_SPAWN_REQ;

typedef struct TerminateDoneArgs {
    Channel * c;
    char token[256];
    int error;
    char * msg;
} TerminateDoneArgs;

typedef struct post_event_args {
    EventCallBack *     call_back;      /* post_event call-back routine */
    void *              arg;            /* post_event call-back argument */
} POST_EVENT_ARGS;

/*
 * Get VxWorks ID from given context.
 */
extern pid_t get_context_task_id(Context * ctx);

/*
 * Get children of given <id>. If <attached_only> is true, then only return
 * contexts that are attached.
 */
extern void vx_get_children(Channel * c, char * id, int attached_only);

/*
 * Get TCF id from given VxWorks Id.
 */
extern char * vx_get_id(pid_t pid);

/*
 * Get TCF id from given VxWorks object information
 */
extern char * vx_get_id_from_obj_info(VX_OBJ_INFO * obj_info);

/*
 * Get VxWorks object information from given TCF id.
 * Return a VX_OBJ_INFO pointer which must be freed using vx_free_obj_info().
 */
extern VX_OBJ_INFO * vx_get_obj_info (char * id);

/*
 * Free a VxWorks object information structure returned by vx_get_obj_info().
 */
extern void vx_free_obj_info (VX_OBJ_INFO * vx_obj_info);

/*
 * create a VxWorks task.
 */
extern pid_t vx_task_create (char * name, int priority, int task_options, ssize_t stack_size, ContextAddress start_address, char * args[], int attach, int attach_mode, int * p_err, ContextAttachCallBack * attach_cb, void * attach_cb_data);

/*
 * create a VxWorks RTP.
 */
extern int vx_rtp_create (char * token, Channel * chnl, void * prs, const char * rtp_file_name, const char * argv[], const char * envp[], int priority, ssize_t stack_size, int rtp_options, int init_task_options, int redirect_io, int attach, int attach_mode, int * p_err);

/*
 * Terminate a VxWorks task or rtp.
 */
extern int vx_context_terminate (pid_t pid, void * terminate_done, void * data);

/*
 * Send reply to Processes start command.
 */
extern void start_process_send_reply (RTP_SPAWN_REQ * pReq);

/*
 * Send exit event for processes started with Processes start command.
 */
extern void start_process_send_exited_event (pid_t pid, int exit_code);

/*
 * Check if the execution context stack reaches its limit.
 */
extern int vx_trace_stack_bottom_check(StackFrame * frame);

/* 
 * Utility to read a ELF file program header
 * This routine is used to read the program header of a ELF file. It is up to 
 * the caller to free the memory allocated to store resulting program header.
 * Returns 0 on success and -1 on error.
 */
extern int vx_elf_read_phdr(const char * file_name, unsigned * pheader_cnt, ELF_PHeader ** ppheaders);

/*
 * VxDBG event inject call-back used by VXDBG_POST_EVENT() macro.
 * This call-back does not free given args (POST_EVENT_ARGS).
 */
extern void vxdbg_post_event(void * args);

/*
 * VxDBG event inject call-back used by VXDBG_POST_EVENT() macro.
 * This call-back does free given args (POST_EVENT_ARGS) after posting event.
 */
extern void vxdbg_post_and_free_event(void * args);
#endif /* D_context_vxworks */
