/*
 * Copyright (c) 2013-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

#include <tcf/config.h>

#if SERVICE_WR_HostFs
#include "WR_HostFs.h"
#include <tcf/framework/channel.h>
#include <tcf/framework/json.h>
#include <tcf/framework/errors.h>
#include <tcf/framework/myalloc.h>
#include <tcf/framework/events.h>
#include <tcf/services/WR_FileSystem_proxy.h>

/* VxWorks headers */
#include <taskLib.h>
#include <errnoLib.h>
#include <hostFsLib.h>
#include <ctype.h>
#include <assert.h>

#define AUTOMOUNT_ENABLED_BY_DEFAULT
#ifdef  WR_HOSTFS_DEBUG
#include <tcf/framework/trace.h>
#define WR_HOSTFS_DEBUG_LOG(msg,args...)     \
        {                                  \
        trace (LOG_ALWAYS, msg, ## args);  \
        }
#else
#define WR_HOSTFS_DEBUG_LOG(msg,args...)
#endif  /* DEBUG */

typedef struct {
    char * device_name;
    char * client_data;
    BOOL   auto_mount;
    BOOL   auto_mount_rw;
    BOOL   auto_pathmap;
} DevCreateCmdArgs;

typedef struct {
    char * mountpoint;
    char * root_path;
    char * client_data;
    BOOL rdonly;
} MountCmdArgs;

static const char *   WR_HOSTFS = "WR_HostFs";

void send_wr_hostfs_evt(void * args) {
	 
    Channel * chan = NULL;
    HostfsEvtArgs * evt_args = (HostfsEvtArgs*)args;

    assert (is_dispatch_thread());
    assert (evt_args != NULL);
    assert (evt_args->channel != NULL);
    assert (evt_args->name != NULL);
    assert (evt_args->id != NULL);

    chan = (Channel *)evt_args->channel;
    write_stringz(&chan->out, "E");
    write_stringz(&chan->out, WR_HOSTFS);
    write_stringz(&chan->out, evt_args->name);
    json_write_string(&chan->out, evt_args->id);
    write_stream(&chan->out, MARKER_EOM);
    loc_free(evt_args->id);
    loc_free(evt_args);
    channel_unlock(chan);
} 


static void read_devcreate_params(InputStream * inp, const char * name, void * args) {
    DevCreateCmdArgs * devargs = (DevCreateCmdArgs*)args;

    if (strcmp(name, "DeviceName") == 0) {
        devargs->device_name = json_read_alloc_string(inp);
    }
    else if (strcmp(name, "ClientData") == 0) {
        devargs->client_data = json_read_alloc_string(inp);
    }
    else if (strcmp(name, "AutoMountReadOnly") == 0) {
        devargs->auto_mount_rw = !json_read_boolean(inp);
    }
    else if (strcmp(name, "AutoMount") == 0) {
        devargs->auto_mount = json_read_boolean(inp);
    }
    else if (strcmp(name, "AutoPathMap") == 0) {
        devargs->auto_pathmap = json_read_boolean(inp);
    }
    else {
        json_skip_object(inp);
    }
}

static void free_devcreate_args (DevCreateCmdArgs * args) {
    if (args->device_name != NULL) 
	loc_free(args->device_name);
    
    if (args->client_data != NULL) 
	loc_free(args->client_data);

    loc_free(args);
}

static void command_dev_create(char * token, Channel * chnl) {
    int error = 0;
    DevCreateCmdArgs * args = NULL;
    STATUS status;
    WR_HOSTFS_DEBUG_LOG("WR_HostFs devCreate.");
    char * id;
      
    /* parse arguments */

    args = (DevCreateCmdArgs*)loc_alloc_zero(sizeof(DevCreateCmdArgs));
#ifdef AUTOMOUNT_ENABLED_BY_DEFAULT
    args->auto_mount = TRUE;
#endif

    json_read_struct(&chnl->inp, read_devcreate_params, args);
    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);

    WR_HOSTFS_DEBUG_LOG("WR_HostFs devCreate Params: args->device_name = %s.", args->device_name);
    status = hostFsDevInit(chnl, args->device_name, args->auto_mount, args->auto_mount_rw, args->auto_pathmap, args->client_data, &id);

    if (status != OK) {
	/* vxWorks hostFs driver reported an error */
	
	if (errno == EBUSY)
	    error = set_errno(ERR_OTHER, "A hostFs device already exists on target.");
	else if (errno == EINVAL)
	    error = set_errno(ERR_OTHER, "At least one parameter is incorrect");
	else 
            error = set_errno(ERR_OTHER, "Hostfs device creation failed.");
    }
        
    if ((error != 0) || (!args->auto_mount)) {
	/* 
  	 * no auto mount of host device: answer TCF
	 * request now.
	 */
    	write_stringz(&chnl->out, "R");
    	write_stringz(&chnl->out, token);
    	write_errno(&chnl->out, error);

    	if (error == 0) {
    	    json_write_string(&chnl->out, id);
    	} else {
    	    write_stringz(&chnl->out, "null");
    	}

    	write_stream(&chnl->out, MARKER_EOM);
        free_devcreate_args(args);
	return;
    }
	
    free_devcreate_args(args);

    /* 
     * auto mount host drives is enabled and device creation succeeded.
     * Let's query the host drives list from the value-add and create 
     * a hostfs mount point for each.
     */
	
    wr_fsproxy_roots (chnl, token, id, TRUE);
    
}

static void command_dev_delete(char * token, Channel * chnl) {
    int error = 0;
    char * id = NULL;

    WR_HOSTFS_DEBUG_LOG("WR_HostFs devDelete.");

    id = json_read_alloc_string(&chnl->inp);

    WR_HOSTFS_DEBUG_LOG("WR_HostFs devDelete Param: device ID = %s", id);

    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);

    /* check that the device exists */

    if ((id == NULL) || (hostFsPropertyGet(id, NULL) != OK)) {
	error = set_errno(ERR_OTHER, "Unknown hotfs device.");
    } else if (hostFsDevDelete(chnl) != OK) {
	int err = errnoGet ();
	if (err == ENODEV) error = set_errno(ERR_OTHER, "Permission denied.");
	else error = set_errno(ERR_OTHER, "Hostfs device deletion failed.");
    }

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);
    write_errno(&chnl->out, error);
    write_stream(&chnl->out, MARKER_EOM);
    if (id!= NULL) loc_free(id);
}

static void free_mount_args(MountCmdArgs* args) {
    
    if (args->mountpoint != NULL) 
	loc_free(args->mountpoint);
    
    if (args->root_path != NULL) 
	loc_free(args->root_path);

    if (args->client_data != NULL) 
	loc_free(args->client_data);

    loc_free(args);
}

static void read_mount_params(InputStream * inp, const char * name, void * args) {
    MountCmdArgs * mntargs = (MountCmdArgs*)args;

    if (strcmp(name, "MountPointName") == 0) {
        mntargs->mountpoint = json_read_alloc_string(inp);
    }
    else if (strcmp(name, "RootPath") == 0) {
        mntargs->root_path = json_read_alloc_string(inp);
    }
    else if (strcmp(name, "ClientData") == 0) {
        mntargs->client_data = json_read_alloc_string(inp);
    }
    else if (strcmp(name, "IsReadOnly") == 0) {
        mntargs->rdonly = json_read_boolean(inp);
    }
    else {
        json_skip_object(inp);
    }
}

static void command_mount(char * token, Channel * chnl) {
    int error = 0;
    MountCmdArgs * args = NULL;
    char * device_id = NULL;
    char * id = NULL;
    HOSTFS_PROP_INFO devInfo;
    
    WR_HOSTFS_DEBUG_LOG("WR_HostFs mount.");

    args = (MountCmdArgs*)loc_alloc_zero(sizeof(MountCmdArgs));
    device_id = json_read_alloc_string(&chnl->inp);
    json_test_char(&chnl->inp, MARKER_EOA);
    json_read_struct(&chnl->inp, read_mount_params, args);
    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);
    WR_HOSTFS_DEBUG_LOG("WR_HostFs mount Params: deviceId = %s, mountpoint=%s, root_path=%s",
	device_id, args->mountpoint, args->root_path);

    
    if ((device_id == NULL) || (hostFsPropertyGet(device_id, &devInfo) != OK)) {
        error = set_errno(ERR_OTHER, "Unknown hostfs device.");
    } else if (hostFsMountPointAdd(chnl, device_id, args->mountpoint, args->root_path, args->rdonly, args->client_data, &id, FALSE, devInfo.prop.dev.autoPathMap) != OK) {
	int err = errno;
	
	if (err == EINVAL)
	    error = set_errno(ERR_OTHER, "At least one parameter is incorrect.");

        else if (err == ENOSPC)
	    error = set_errno(ERR_OTHER, "Maximum number of mount point reached.");
    	else if ((err == EPERM)  || (err == ENODEV))
	    error = set_errno(ERR_OTHER, "Permission denied.");
        else
	    error = set_errno(ERR_OTHER, "hostFs mount point creation failed.");
    }
	
    free_mount_args(args);
    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);
    write_errno(&chnl->out, error);
    if (device_id != NULL) loc_free (device_id);
    if (error == 0) {
    	json_write_string(&chnl->out, id);
    } else {
    	write_stringz(&chnl->out, "null");
    }
    write_stream(&chnl->out, MARKER_EOM);
}

static void command_umount(char * token, Channel * chnl) {
    int error = 0;
    char * id = NULL;

    WR_HOSTFS_DEBUG_LOG("WR_HostFs umount.");

    id = json_read_alloc_string(&chnl->inp);

    WR_HOSTFS_DEBUG_LOG("WR_HostFs umount Param: mount ID = %s", id);

    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);

    /* check that mount point is valid */

    if ((id == NULL) || (hostFsPropertyGet(id, NULL) != OK)) {
	 error = set_errno(ERR_OTHER, "Unknown mount point.");
    } else if (hostFsMountPointDelete(chnl, id) != OK) {
	int err = errnoGet();
	if (err == ENODEV) error = set_errno(ERR_OTHER, "Permission denied.");
        else error = set_errno(ERR_OTHER, "Mount point deletion failed.");
    }

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);
    write_errno(&chnl->out, error);
    write_stream(&chnl->out, MARKER_EOM);
    if (id != NULL) loc_free(id);
}

static void command_get_children(char * token, Channel * chnl) {
    int error = 0;
    char * id = NULL;

    WR_HOSTFS_DEBUG_LOG("WR_HostFs getChildren.");

    id = json_read_alloc_string(&chnl->inp);

    WR_HOSTFS_DEBUG_LOG("WR_HostFs getChildren Param: ID = %s", id);

    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);

    if ((id == NULL) || strcmp(id, "null") == 0) {

	/* id = NULL : provide device ID if it exists on target */
	write_errno(&chnl->out, error);
	write_stream(&chnl->out, '[');
	if (hostFsPropertyGet(DEV_ID, NULL) == OK) {
	    json_write_string(&chnl->out, DEV_ID);
	}
	write_stream(&chnl->out, ']');
    } else {
	/* id != NULL, provide mount points list */
	char * id_list = NULL;
  	char * cur_id = NULL;
	int  nb_mnt = 0;
	int nb_mnt_written = 0;
	
	/* check that device correspoding to ID exists */

	if (hostFsPropertyGet(id, NULL) != OK) {
	    error = set_errno(ERR_OTHER, "Invalid context ID.");
	} else if (hostFsMntListGet(id, &id_list, &nb_mnt) != OK) {
	    error = set_errno(ERR_OTHER, "Cannot determine children list.");
	}
	
	write_errno(&chnl->out, error);
	write_stream(&chnl->out, '[');
 	cur_id = id_list;

	while (nb_mnt_written < nb_mnt)
	    {
	    json_write_string(&chnl->out, cur_id);
	    cur_id += strlen (cur_id) + 1;
	    nb_mnt_written++;
   	    if (nb_mnt_written < nb_mnt) 
		write_stream(&chnl->out, ',');
	    }
	write_stream(&chnl->out, ']');

	if (id_list != NULL) loc_free(id_list);
    } 

    write_stream(&chnl->out, MARKER_EOA);
    write_stream(&chnl->out, MARKER_EOM);
    if (id != NULL) loc_free(id);
}

static void write_properties(OutputStream * out, char * id, HOSTFS_PROP_INFO * prop) {
    write_stream(out, '{');
    json_write_string(out, "ID");
    write_stream(out, ':');
    json_write_string(out, id);
    write_stream(out, ',');

    json_write_string(out, "Name");
    write_stream(out, ':');
    json_write_string(out, prop->name);
    write_stream(out, ',');

    json_write_string(out, "IsDevice");
    write_stream(out, ':');
    json_write_boolean(out, prop->isDev);
    write_stream(out, ',');

    if (!prop->isDev) {
   	/* mount point properties */

        json_write_string(out, "ParentID");
        write_stream(out, ':');
        json_write_string(out, prop->prop.mnt.devId);
        write_stream(out, ',');

        json_write_string(out, "RootPath");
        write_stream(out, ':');
        json_write_string(out, prop->prop.mnt.rootPath);
        write_stream(out, ',');

	json_write_string(out, "TargetPath");
        write_stream(out, ':');
        json_write_string(out, prop->prop.mnt.targetPath);
        write_stream(out, ',');

	json_write_string(out, "IsReadOnly");
        write_stream(out, ':');
        json_write_boolean(out, prop->prop.mnt.readOnly);
	
    } else {
	json_write_string(out, "AutoMount");
        write_stream(out, ':');
        json_write_boolean(out, prop->prop.dev.autoMount);

	if (prop->prop.dev.autoMount) {
            write_stream(out, ',');
	    json_write_string(out, "AutoMountReadOnly");
            write_stream(out, ':');
            json_write_boolean(out, prop->prop.dev.autoReadOnly);
	}

        write_stream(out, ',');
     	json_write_string(out, "AutoPathMap");
        write_stream(out, ':');
        json_write_boolean(out, prop->prop.dev.autoPathMap);
    }

    if (prop->clientData != NULL) {
        write_stream(out, ',');
        json_write_string(out, "ClientData");
        write_stream(out, ':');
        json_write_string(out, prop->clientData);    
    }
    write_stream(out, '}');
    write_stream(out, MARKER_EOA);
}

static void command_get_properties(char * token, Channel * chnl) {
    int error = 0;
    char * id = NULL;
    
    HOSTFS_PROP_INFO prop;
    memset (&prop, 0, sizeof (HOSTFS_PROP_INFO));

    WR_HOSTFS_DEBUG_LOG("WR_HostFs getProperties.");

    id = json_read_alloc_string(&chnl->inp);
    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);

    WR_HOSTFS_DEBUG_LOG("WR_HostFs getProperties Param: ID = %s", id);

    if ((id == NULL) || (hostFsPropertyGet (id, &prop) != OK)) {
	error = set_errno(ERR_OTHER, "Invalid context ID.");
    }  

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);
    write_errno(&chnl->out, error);

    if (error != 0) {
        write_stringz(&chnl->out, "null");
    } else {
	write_properties (&chnl->out, id, &prop);
    }
    write_stream(&chnl->out, MARKER_EOM);
    if (id != NULL) loc_free(id);
}

static void command_get_capabilities(char * token, Channel * chnl) {
    BOOL autoMountIsDefault;
    char * id = NULL;

#ifdef AUTOMOUNT_ENABLED_BY_DEFAULT
    autoMountIsDefault = TRUE;
#else
    autoMountIsDefault = FALSE;
#endif
    
    WR_HOSTFS_DEBUG_LOG("WR_HostFs getCapabilities");

    id = json_read_alloc_string(&chnl->inp);
    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);

    write_errno(&chnl->out, 0);
    write_stream(&chnl->out, '{');

    json_write_string(&chnl->out, "MultipleDevice");
    write_stream(&chnl->out, ':');
    json_write_boolean(&chnl->out, 0);
    write_stream(&chnl->out, ',');

    json_write_string(&chnl->out, "MultipleMountPoint");
    write_stream(&chnl->out, ':');
    json_write_boolean(&chnl->out, 1);
    write_stream(&chnl->out, ',');

    json_write_string(&chnl->out, "MaxMountPointPerDevice");
    write_stream(&chnl->out, ':');
    json_write_ulong(&chnl->out, (unsigned long)MAX_MNT_PER_DEV);
    write_stream(&chnl->out, ',');

    json_write_string(&chnl->out, "AutoMount");
    write_stream(&chnl->out, ':');
    json_write_boolean(&chnl->out, 1);
    write_stream(&chnl->out, ',');

    json_write_string(&chnl->out, "AutoPathMap");
    write_stream(&chnl->out, ':');
    json_write_boolean(&chnl->out, 1);

    write_stream(&chnl->out, '}');
        
    write_stream(&chnl->out, MARKER_EOA);
    write_stream(&chnl->out, MARKER_EOM);
    if (id != NULL) loc_free (id);
}

static void command_automount_refresh (char * token, Channel * chnl) {
    int error = 0;
    char * id = NULL;
    BOOL devValid = FALSE;

    WR_HOSTFS_DEBUG_LOG("WR_HostFs getChildren.");
    
    id = json_read_alloc_string(&chnl->inp);

    WR_HOSTFS_DEBUG_LOG("WR_HostFs autoMountRefresh Param: ID = %s", id);

    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);

    devValid = (hostFsPropertyGet(DEV_ID, NULL) == OK);

    if (!devValid)
        error = set_errno(ERR_OTHER, "No hostfs device created on target.");
    else if ((id == NULL) || (strncmp(DEV_ID, id, MAX_DEVID_SIZE) != 0))
        error = set_errno(ERR_OTHER, "Invalid context ID.");
    
    if (id != NULL) loc_free(id);
    if (error != 0){
	write_stringz(&chnl->out, "R");
    	write_stringz(&chnl->out, token);
    	write_errno(&chnl->out, error);
    	write_stream(&chnl->out, MARKER_EOM);
        return;
    } 
    /* 
     * Let's query the host drives list from the value-add and update 
     * the hostfs mount point list accordingly.
     */
	
    wr_fsproxy_roots (chnl, token, NULL, FALSE);
}

static void channel_close_listener(Channel * chnl) {
    STATUS stat =  hostFsDevDelete(chnl);

    assert (stat == 0);
}

/*
 * ini_wr_hostfs_service - initialize the service WR_HostFs.
 */
void ini_wr_hostfs_service (Protocol * proto) {
    static int ini_wr_hostfs = 0;
    if (ini_wr_hostfs == 0) {
        add_channel_close_listener(channel_close_listener);
        ini_wr_hostfs = 1;
    }    
    add_command_handler(proto, WR_HOSTFS, "devCreate", command_dev_create);
    add_command_handler(proto, WR_HOSTFS, "devDelete", command_dev_delete);
    add_command_handler(proto, WR_HOSTFS, "mount", command_mount);
    add_command_handler(proto, WR_HOSTFS, "umount", command_umount);
    add_command_handler(proto, WR_HOSTFS, "getChildren", command_get_children);
    add_command_handler(proto, WR_HOSTFS, "getProperties", command_get_properties);
    add_command_handler(proto, WR_HOSTFS, "getCapabilities", command_get_capabilities);
    add_command_handler(proto, WR_HOSTFS, "autoMountRefresh", command_automount_refresh);
}
#endif /* SERVICE_WR_HostFs */
