/*
 * Copyright (c) 2013-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use of this
 * software may be licensed only pursuant to the terms of an applicable Wind
 * River license agreement.
 */


#include <tcf/config.h>

#if SERVICE_WR_Module

#include <assert.h>
#include <sys/stat.h>
#include "WR_Module.h"
#include <errnoLib.h>
#include <ctype.h>
#include <tcf/framework/protocol.h>
#include <tcf/framework/cpudefs.h>
#include <tcf/framework/cache.h>
#include <tcf/framework/events.h>
#include <tcf/framework/exceptions.h>
#include <tcf/framework/json.h>
#include <tcf/framework/link.h>
#include <tcf/framework/myalloc.h>
#include <tcf/framework/trace.h>
#include <tcf/context-vxworks.h>
#include <tcf/worker-module.h>
#include <system/VxWorks/tcf/context-vxworks.h>

#ifndef TRUE
#  define TRUE 1
#endif
#ifndef FALSE
#  define FALSE 0
#endif
#ifndef BOOL
#   define BOOL int
#endif

#define KEY_ID "ID"
#define KEY_ParentID "ParentID"
#define KEY_ModuleID  "ModuleId"
#define KEY_ModuleFile "ModuleFile"
#define KEY_Signatue "Signature"
#define KEY_CanUnload "CanUnload"
#define KEY_Undefined "Undefined"
#define KEY_Commons "Commons"
#define KEY_SectionList "SectionList"
#define KEY_SegmentList "SegmentList"
#define KEY_Name "Name"
#define KEY_Size "Size"
#define KEY_Addr "Addr"
#define KEY_Flags "Flags"

#define KEY_CMD_LOAD                "load"
#define KEY_CMD_UNLOAD              "unload"
#define KEY_CMD_GET_CHILDREN        "getChildren"

#define KEY_LoadFlags               "LoadFlags"
#define KEY_UnloadFlags             "UnloadFlags"

#define KEY_Filter                  "Filter"
#define KEY_ModuleType              "ModuleType"
#define KEY_KernelImage             "KernelImage"
#define KEY_KernelModule            "KernelModule"
#define VAL_KernelImage             KEY_KernelImage
#define VAL_KernelModule            KEY_KernelModule
#define KEY_SharedLibrary           "SharedLibrary"
#define KEY_ProcessImage            "ProcessImage"

#define KEY_Type                    "Type"
#define VAL_TypeModule              "Module"
#define VAL_TypeUndefSymbolList     "UndefSymbolList"
#define VAL_TypeCommonSymbolList    "CommonSymbolList"
#define VAL_TypeSection             "SectionList"
#define VAL_TypeSegment             "SegmentList"

#define VX_KERNEL_MODULE_NAME "vxWorks"

#define NODE_UNKNOW_ID          0
#define NODE_MODULE_ID          1
#define NODE_SYM_LST_UNDEF_ID   2
#define NODE_SYM_LST_COMMON_ID  3
#define NODE_SECTION_LST_ID     4
#define NODE_SEGMENT_LST_ID     5
#define NODE_KERNEL_MODULE_ID   6

static const char *   WR_MODULE                    = "WR_Module";
static const char *   WR_MODULE_LOAD_EVENT_ADDED   = "moduleAdded";
static const char *   WR_MODULE_LOAD_EVENT_REMOVED = "moduleRemoved";

static const unsigned WR_MODULE_LOAD_MAGIC = 0x35715964;
static int openedChannelCount = 0;

static TCFBroadcastGroup * broadcast_group = NULL;

/* ModuleDesc contains information to fully describe a module */
typedef struct ModuleDesc {
    unsigned    magic;      /* MAGIC number to check module integrity */
    char *      ctxId;      /* the ctx id owner of this file */
    MODULE_ID   vxModId;    /* the vxWorks module ID */
    long unsigned nodeType;   /* It identify the node type one of NODE_xxx */
} ModuleDesc;

/*
 * Parameter used by the ouModuleId function
 */
typedef struct ParamOutModuleId {
    void *      streamOut;  /* the output channel stream */
    char *      ctxId;      /* the owner ctx id  */
    int         nbModule;  /* number of the module */
} ParamOutModuleId;

/* Get the kernel TCF ID
 */

static char *kernelTCFIdGet () {
    /* in dispatcher thread, no need to copy because we are the only one
     * using the string returned by vx_get_id, which is a static buffer */
    assert(is_dispatch_thread());
    return vx_get_id((pid_t) kernelIdGet());
}

static BOOL is_tcfid_valid (char *tcfid) {
    VX_OBJ_INFO *ctxObjInfo = NULL;
    BOOL ret = FALSE;
    ctxObjInfo = vx_get_obj_info (tcfid);
    if (ctxObjInfo == NULL)
        return FALSE;
    if (ctxObjInfo->obj_type == VX_OBJ_KERNEL) {
        ret = TRUE;
    } else {
        ret = FALSE;
    }
    vx_free_obj_info (ctxObjInfo);
    return ret;
}

/*
 * read_hex - translate string to unint64
 * This function convert a string hexa value to an unsigned int 64 bits.
 *
 * param IN hexStr : the string number coded hexa
 * return the result of the conversion.
 *
 */
static unsigned long read_hex(const char ** hexStr) {
    unsigned long res = 0;
    const char * pHexStr = *hexStr;

    while (TRUE) {
        if (*pHexStr >= '0' && *pHexStr <= '9') {
            res = (res << 4) | (*pHexStr - '0');
        } else if (*pHexStr >= 'A' && *pHexStr <= 'F') {
            res = (res << 4) | (*pHexStr - 'A' + 10);
        } else {
            break;
        }

        pHexStr++;
    }

    *hexStr = pHexStr;
    return res;
}

/*
 * id2module - translate a moduleID to a module descriptor
 *
 * This function retrieve the module description that correspond to the given ID.
 * module id format is on of :
 *    vxMod.<vxWorks module id>.<sub node type>.<context owner ID>
 *
 * This function allocate a module descriptor, so the caller is in charge to free it.
 *
 * \param IN  id the tcfID that identify the module or a sub node
 *
 * \return module descriptor if the conversion succeed, else NULL
 */
static ModuleDesc * id2module(const char * id) {
    ModuleDesc * mod = NULL;
    char *p = NULL;
    long unsigned modId = 0;
    long unsigned nodeType = NODE_UNKNOW_ID;

    /* retrieve the module ID first */

    if (id != NULL &&
            id[0] == 'v' && id[1] == 'x' &&
            id[2] == 'M' && id[3] == 'o' && id[4] == 'd' && id[5] == '.') {

        p = (char *)id + 6;

        /* read the vxWorks module ID */
        modId = read_hex((const char **)&p);

        /* A sub node is describe by
         * the caracter '.' followed by hexa value folowed by a '.'
         */
        if (*p == '.') {
            p++; /* jump the . */
            /* Just get it for now. */
            nodeType = read_hex((const char **)&p);
        } else {
            /* This is not a valid ID, exit */
            printf ("[%s:%d] Not a valid id p!='.' (%s)\n", __FILE__, __LINE__, p);
            return NULL;
        }

        /* The last field is the context owner */
        if (*p == '.') {
            p++;
        } else {
            /* This is not a valid ID, exit */
            printf ("[%s:%d] Not a valid id p!='.' (%s)\n", __FILE__, __LINE__, p);
            return NULL;
        }

        /* p point to the context ID that own this module */
        if (is_tcfid_valid(p)) {

            /* Check the module validity */
            if ((nodeType == NODE_KERNEL_MODULE_ID) ||
                ((modId == 0) && (nodeType != NODE_KERNEL_MODULE_ID)) ||
                (moduleIdFigure ((MODULE_ID) modId) != NULL)) {

                /* save the context ID */
                mod = (ModuleDesc *) loc_alloc_zero(sizeof(ModuleDesc));
                if (mod == NULL) {
                    /* Memory problem */
                    printf ("[%s:%d] out of memory\n", __FILE__, __LINE__);
                    return NULL;
                }
                mod->magic    = WR_MODULE_LOAD_MAGIC;
                mod->ctxId    = loc_strdup(p);
                mod->vxModId  = (MODULE_ID) modId;
                mod->nodeType = nodeType;
            }
        } else {
            printf ("[%s:%d] not a valid contextID\n", __FILE__, __LINE__);
        }
    }
    return mod;
}

/*
 * module2id - create an ID base on module descriptor
 *
 */
static char * module2id(ModuleDesc desc){
    char moduleIdPart [50]; /* this enough to store first part of id */
    char * moduleId = NULL;

    assert(desc.magic == WR_MODULE_LOAD_MAGIC);

    /* module id is vxMod.<vxWorks module id>.<Node type>.<context owner ID>*/
    snprintf(moduleIdPart, 50, "vxMod.%lX.%X",
            (long unsigned) desc.vxModId, desc.nodeType);

    moduleId = loc_alloc_zero (strlen(moduleIdPart) + strlen(desc.ctxId) + 2);
    if (moduleId != NULL) {
        sprintf(moduleId, "%s.%s", moduleIdPart, desc.ctxId);
    }

    return moduleId;
}

static char * vxModId2id (MODULE_ID modId) {
    ModuleDesc modDesc;
    memset(&modDesc, 0, sizeof(modDesc));

    modDesc.magic = WR_MODULE_LOAD_MAGIC;
    modDesc.ctxId = kernelTCFIdGet();
    modDesc.vxModId = modId;
    modDesc.nodeType = NODE_MODULE_ID;

    char * modIdStr = module2id(modDesc);
    return modIdStr;
}

/*
 * moduleDescFree - Free the module descriptor and its field
 */
static void moduleDescFree (ModuleDesc * mod) {

    if (mod != NULL) {
        if (mod->ctxId != NULL) {
            loc_free(mod->ctxId);
        }
        loc_free(mod);
    }
}

/*
 * wrModuleSendLoadResponse - Send the response to the Load command.
 * Either everything is OK or we had a failure. Provide details for failure.
 */

void wrModuleSendLoadResponse (WR_MODULE_LOAD_LOAD_REQ *pReq) {
    write_stringz(&pReq->channel->out, "R");
    write_stringz(&pReq->channel->out, pReq->token);
    write_errno(&pReq->channel->out, pReq->loader_errno);
    
    /* If it's OK, send the Module Id in TCF format */
    if (pReq->modId != NULL) {
        ModuleDesc modDesc;
        char * modIdStr;
        memset(&modDesc, 0, sizeof(modDesc));

        modDesc.magic = WR_MODULE_LOAD_MAGIC;
        modDesc.ctxId = pReq->processId;
        modDesc.vxModId = pReq->modId;
        modDesc.nodeType = NODE_MODULE_ID;

        modIdStr = module2id(modDesc);

        assert (modIdStr == NULL); /* Stop here, there is a fatal error */

        if (modIdStr != NULL) {
            json_write_string (&pReq->channel->out, modIdStr);
            loc_free(modIdStr);
        }
    }
    write_stream(&pReq->channel->out, MARKER_EOA);
    write_stream(&pReq->channel->out, MARKER_EOM);
}

static BOOL outEachCommonSymbol (SYMBOL_ID symbolId, MODULE_ID moduleId, _Vx_usr_arg_t arg) {
    /* Get the stream out */
    ParamOutModuleId * param = (ParamOutModuleId *) arg;
    OutputStream * out = param->streamOut;

    if (param->nbModule > 0) {
        write_stream (out, ',');
    }
    param->nbModule++;
    write_stream(out, '{');
    json_write_string(out, KEY_Name);
    write_stream(out, ':');
    json_write_string(out, symbolId->name);
    write_stream(out, ',');
    json_write_string(out, KEY_Addr);
    write_stream(out, ':');
    json_write_long(out, (long int) symbolId->value);
    write_stream(out, '}');
    return TRUE; /* List all common*/
}

static BOOL outEachUndefSymbol (char * symbolName, MODULE_ID moduleId,_Vx_usr_arg_t arg) {
    /* Get the stream out */
    ParamOutModuleId * param = (ParamOutModuleId *) arg;
    OutputStream * out = param->streamOut;

    if (param->nbModule > 0) {
        write_stream (out, ',');
    }
    param->nbModule++;
    write_stream(out, '{');
    json_write_string(out, KEY_Name);
    write_stream(out, ':');
    json_write_string(out, symbolName);
    write_stream(out, '}');
    return TRUE; /* List all common*/
}

static BOOL outEachSegment (SEGMENT_ID segmentId, MODULE_ID moduleId, _Vx_usr_arg_t arg) {
    ParamOutModuleId * param = (ParamOutModuleId *) arg;
    OutputStream * out = param->streamOut;
    char * flags = NULL;

    if (segmentId->type == SEGMENT_TEXT) {
        flags = "R-X";
    } else if ((segmentId->type == SEGMENT_DATA) || (segmentId->type == SEGMENT_BSS)) {
        flags = "RW-";
    } else {
        /* Should not happen */
        flags = "---";
    }

    if (param->nbModule > 0) {
        write_stream (out, ',');
    }
    param->nbModule++;

    write_stream(out, '{');

    /*
     *  Segment are not named
    json_write_string(out, KEY_Name);
    write_stream(out, ':');
    json_write_string(out, segmentId->name);
    write_stream(out, ',');
    */
    json_write_string(out, "SegType");
    write_stream(out, ':');
    switch (segmentId->type) {
        case SEGMENT_TEXT:
            json_write_string(out, "TEXT");
            break;
        case SEGMENT_DATA:
            json_write_string(out, "DATA");
            break;
        case SEGMENT_BSS:
            json_write_string(out, "BSS");
            break;
    }
    write_stream(out, ',');

    json_write_string(out, KEY_Addr);
    write_stream(out, ':');
    json_write_ulong(out, (unsigned long) segmentId->address);

    write_stream(out, ',');

    json_write_string(out, KEY_Size);
    write_stream(out, ':');
    json_write_ulong(out, (unsigned long) segmentId->size);

    write_stream(out, ',');
    json_write_string(out, KEY_Flags);
    write_stream(out, ':');
    json_write_string(out, flags);

    /*
    write_stream(out, ',');
    json_write_string(out, "Flags_2");
    write_stream(out, ':');
    json_write_uint64(out, segmentId->flags);
    */

    write_stream(out, '}');

    return TRUE;
}

/*
 * outEachModuleId - output the translated module ID
 *
 * This function is used by getChildrenOfProcess() only
 */
static BOOL outEachModuleId (MODULE_ID mod, _Vx_usr_arg_t arg) {
    ModuleDesc modDesc;
    char * modIdStr = NULL;
    ParamOutModuleId * param = (ParamOutModuleId *) arg;
    OutputStream * out = param->streamOut;

    memset(&modDesc, 0, sizeof(modDesc));

    /* skip hidden module like vxWorks.sym */
    if ((mod->flags & HIDDEN_MODULE) != HIDDEN_MODULE) {
        if (param->nbModule > 0) {
            param->nbModule++;
            write_stream (out, ',');
        }

        modDesc.magic = WR_MODULE_LOAD_MAGIC;
        modDesc.ctxId = param->ctxId;
        modDesc.vxModId = mod;
        modDesc.nodeType = NODE_MODULE_ID;

        modIdStr = module2id(modDesc);

        assert (modIdStr == NULL); /* Stop here, there is a fatal error */

        if (modIdStr != NULL) {
            json_write_string (out, modIdStr);
            loc_free(modIdStr);
        }
    }

    return TRUE; /* List all module */
}

static void wrModuleSendNotification (MODULE_ID modId, const char *event_name) {
    OutputStream * out;
    ParamOutModuleId param;

    memset(&param, 0, sizeof(param));

    if (((modId->flags & HIDDEN_MODULE) != HIDDEN_MODULE) &&
            (broadcast_group != NULL)) {
        out = &broadcast_group->out;

        write_stringz(out, "E");
        write_stringz(out, WR_MODULE);

        write_stringz(out, event_name);

        param.nbModule = 0;
        param.streamOut = out;
        param.ctxId = kernelTCFIdGet();
        json_write_string (out, param.ctxId);
        write_stream(out, MARKER_EOA);
        write_stream(out, '[');
        outEachModuleId (modId, (_Vx_usr_arg_t) &param);
        write_stream(out, ']');
        write_stream(out, MARKER_EOA);
        write_stream(out, MARKER_EOM);
    }
}

/*
 * wrModuleAdded_cb - hook called when a module is loaded
 * Broadcast an event.
 */
static void wrModuleAdded_cb_dispatcher (void *data) {
    MODULE_ID modId = (MODULE_ID) data;

    if (modId != NULL) {
        wrModuleSendNotification(modId, WR_MODULE_LOAD_EVENT_ADDED);
    }
}

static void wrModuleAdded_cb (MODULE_ID modId, UINT32 flags) {
    /* We can get called from the worker-thread !
     * The post_event is then defered to the dispatcher thread for sending
     * the asynchronous event !
     */
    VXDBG_POST_EVENT(wrModuleAdded_cb_dispatcher, modId, NULL);
}

static void wrModuleRemoved_cb_dispatcher (void *data) {
    MODULE_ID modId = (MODULE_ID) data;

    if (modId != NULL) {
        wrModuleSendNotification(modId, WR_MODULE_LOAD_EVENT_REMOVED);
    }
}

static void wrModuleRemoved_cb (MODULE_ID modId, UINT32 flags) {
    /* Can be called from anythread */
    VXDBG_POST_EVENT(wrModuleRemoved_cb_dispatcher, modId, NULL);
}

/*
 * event_channel_opened - A channel has been opened, add hook
 */
static void  event_channel_opened(Channel * c) {
    /*
     * This hook will broadcast the message, so all
     * connected client will receive message.
     */
    if ((++openedChannelCount) == 1) {
        moduleLoadHookAdd(wrModuleAdded_cb);
        moduleUnloadHookAdd(wrModuleRemoved_cb);
    }
}

/*
 * event_channel_closed - the channel has been close, cleanup the module list
 */
static void event_channel_closed(Channel * c) {
    if ((--openedChannelCount) == 0) {
        moduleLoadHookDelete(wrModuleAdded_cb);
        moduleUnloadHookDelete(wrModuleRemoved_cb);
    }
}

void add_private_modules_info (MODULE_ID modId, MemoryRegion *r) {
    MemoryRegionAttribute *attr1 = NULL;

    if ((modId == NULL) || (r == NULL))
        return;

    /* Adding the moduleId to the memoryMap */
    attr1 =  (MemoryRegionAttribute *) loc_alloc_zero(sizeof(MemoryRegionAttribute));
    if (attr1 == NULL)
        return;

    attr1->name = loc_strdup (KEY_ModuleID);
    if (attr1->name == NULL) {
        loc_free(attr1);
           return;
    }

    attr1->value = loc_alloc_zero (50);
    if (attr1->value == NULL) {
        loc_free(attr1->name);
        loc_free(attr1);
           return;
    }

    sprintf(attr1->value,"\"0x%lx\"", modId);
    if (r->attrs == NULL) {
        r->attrs = attr1;
    } else {
        attr1->next = r->attrs;
        r->attrs = attr1;
    }
}

/*
 * getChildrenOfProcess
 */
static int getChildrenOfProcess (void * out, char * parentCtxId) {
    int error = 0;
    ParamOutModuleId param;
    ModuleDesc modDesc;
    char * kernelModule = NULL;

    memset(&param, 0, sizeof(param));
    memset(&modDesc, 0, sizeof(modDesc));

    param.streamOut = out;
    param.ctxId = parentCtxId;

    modDesc.magic = WR_MODULE_LOAD_MAGIC;
    modDesc.ctxId = parentCtxId;
    modDesc.nodeType = NODE_KERNEL_MODULE_ID;

    kernelModule = module2id(modDesc);
    if (kernelModule != NULL) {
        json_write_string (out, kernelModule);
        param.nbModule = 1;
        loc_free (kernelModule);
    }

    /* Return the module list of this context */
    moduleEach (outEachModuleId, (_Vx_usr_arg_t)(&param));
    return error;
}

/*
 * getChildrenOfModule - get the children list of a module
 * The filter parameter remove not matching node
 *
 */
static void getChildrenOfModule(void * out, ModuleDesc *node, int filter) {
    char * subNode;
    int childrenCount = 0;

    if (node->vxModId != 0) {
        MODULE_ID mod = node->vxModId;

        if (((filter == NODE_UNKNOW_ID) || (filter == NODE_SYM_LST_UNDEF_ID)) &&
                (mod->undefSymCount != 0)) {
            node->nodeType = NODE_SYM_LST_UNDEF_ID;
            subNode = module2id(*node);
            if (subNode != NULL) {
                if (childrenCount > 0) {
                    write_stream (out, ',');
                }
                json_write_string (out, subNode);
                loc_free (subNode);
                childrenCount++;
            }
        }

        if (((filter == NODE_UNKNOW_ID) || (filter == NODE_SYM_LST_COMMON_ID)) &&
                (mod->commTotalSize != 0)) {
            node->nodeType = NODE_SYM_LST_COMMON_ID;
            subNode = module2id(*node);
            if (subNode != NULL) {
                if (childrenCount > 0) {
                    write_stream (out, ',');
                }
                json_write_string (out, subNode);
                loc_free (subNode);
                childrenCount++;
            }
        }
    }

    /* Hack if kernel = > get segment, else DKM so, get section */

    if (node->nodeType != NODE_KERNEL_MODULE_ID) {

        if ((filter == NODE_UNKNOW_ID) || (filter == NODE_SECTION_LST_ID)) {
            node->nodeType = NODE_SECTION_LST_ID;
            subNode = module2id(*node);
            if (subNode != NULL) {
                if (childrenCount > 0) {
                    write_stream(out, ',');
                }
                json_write_string(out, subNode);
                loc_free(subNode);
                childrenCount++;
            }
        }
    }
}

static void getPropertiesOfSymCommon(void *out, char * me, ModuleDesc *node) {
    ParamOutModuleId param;

    memset(&param, 0, sizeof(param));
    param.streamOut = out;
    param.ctxId = NULL;

    json_write_string(out, KEY_ID);
    write_stream(out, ':');
    json_write_string(out, me);

    write_stream(out, ',');
    json_write_string(out, KEY_ParentID);
    write_stream(out, ':');
    json_write_string(out, node->ctxId);

    write_stream(out, ',');
    json_write_string(out, KEY_Type);
    write_stream(out, ':');
    json_write_string(out, VAL_TypeCommonSymbolList);

    write_stream(out, ',');
    json_write_string(out, KEY_Commons);
    write_stream(out, ':');
    write_stream(out, '[');
    /* add symbol list of common here [{"Name":"","Addr":address},...]*/
    moduleCommonSymEach(node->vxModId, outEachCommonSymbol,(_Vx_usr_arg_t)(&param),
                        NULL);
    write_stream(out, ']');

}

static void getPropertiesOfSymUndef(void *out, char * me, ModuleDesc *node) {
    ParamOutModuleId param;

    memset(&param, 0, sizeof(param));
    param.streamOut = out;
    param.ctxId = NULL;

    json_write_string(out, KEY_ID);
    write_stream(out, ':');
    json_write_string(out, me);

    write_stream(out, ',');
    json_write_string(out, KEY_ParentID);
    write_stream(out, ':');
    json_write_string(out, node->ctxId);

    write_stream(out, ',');
    json_write_string(out, KEY_Type);
    write_stream(out, ':');
    json_write_string(out, VAL_TypeUndefSymbolList);

    write_stream(out, ',');
    json_write_string(out, KEY_Undefined);
    write_stream(out, ':');
    write_stream(out, '[');
    /* add symbol list of undef here [{"Name":""},...]*/
    moduleUndefSymEach(node->vxModId, outEachUndefSymbol,(_Vx_usr_arg_t)(&param),
                        NULL);
    write_stream(out, ']');
}

static void getPropertiesOfSection(OutputStream *out, char * me, ModuleDesc *node) {
    MODULE_INFO info;
    memset(&info, 0, sizeof(info));

    json_write_string(out, KEY_ID);
    write_stream(out, ':');
    json_write_string(out, me);

    write_stream(out, ',');
    json_write_string(out, KEY_ParentID);
    write_stream(out, ':');
    json_write_string(out, node->ctxId);

    write_stream(out, ',');
    json_write_string(out, KEY_Type);
    write_stream(out, ':');
    json_write_string(out, VAL_TypeSection);

    write_stream(out, ',');
    json_write_string(out, KEY_SectionList);
    write_stream(out, ':');
    write_stream(out, '[');

    if (moduleInfoGet(node->vxModId, &info) == OK) {
        SECTION_ID sectionId = moduleFirstSectionGet(node->vxModId);
        unsigned int flags = 0;

        while (sectionId != NULL) {
            char * flags = NULL;
            write_stream(out, '{');

            json_write_string(out, KEY_Name);
            write_stream(out, ':');
            json_write_string(out, sectionId->name);

            write_stream(out, ',');

            json_write_string(out, KEY_Addr);
            write_stream(out, ':');
            json_write_ulong(out, (unsigned long) sectionId->address);

            write_stream(out, ',');

            json_write_string(out, KEY_Size);
            write_stream(out, ':');
            json_write_ulong(out, (unsigned long) sectionId->size);

            write_stream(out, ',');

            if (sectionId->type == SECTION_RODATA) {
                flags = "R--";
            } else if (sectionId->type == SECTION_TEXT) {
                flags = "R-X";
            } else if ((sectionId->type == SECTION_DATA) || (sectionId->type == SECTION_BSS)) {
                flags = "RW-";
            } else {
                /* Should not append */
                flags = "---";
            }
            json_write_string(out, KEY_Flags);
            write_stream(out, ':');
            json_write_string(out, flags);

            write_stream(out, '}');

            sectionId = moduleNextSectionGet(sectionId);
            if (sectionId != NULL) {
                write_stream(out, ',');
            }
        }
    }

    write_stream(out, ']');
}

static void getPropertiesOfSegment(void *out, char * me, ModuleDesc *node) {
    MODULE_ID modId = node->vxModId;
    ParamOutModuleId param;
    memset(&param, 0, sizeof(param));

    json_write_string(out, KEY_ID);
    write_stream(out, ':');
    json_write_string(out, me);

    write_stream(out, ',');
    json_write_string(out, KEY_ParentID);
    write_stream(out, ':');
    json_write_string(out, node->ctxId);

    write_stream(out, ',');
    json_write_string(out, KEY_Type);
    write_stream(out, ':');
    json_write_string(out, VAL_TypeSegment);

    write_stream(out, ',');
    json_write_string(out, KEY_SegmentList);
    write_stream(out, ':');
    write_stream(out, '[');

    param.streamOut = out;

    /* For the kernel, the API moduleSegEach does not work */
    if (node->vxModId != 0) {
        moduleSegEach(node->vxModId, outEachSegment, (_Vx_usr_arg_t) &param);
    }

    write_stream(out, ']');
}

/*
 * get the properties of the module
 */
static void getPropertiesOfModule(void *out, char * me, ModuleDesc *node) {

    /* Return the module information */
    json_write_string(out, KEY_ID);
    write_stream(out, ':');
    json_write_string(out, me);

    write_stream(out, ',');
    json_write_string(out, KEY_ParentID);
    write_stream(out, ':');
    json_write_string(out, node->ctxId);

    write_stream(out, ',');
    json_write_string(out, KEY_Type);
    write_stream(out, ':');
    json_write_string(out, VAL_TypeModule);

    if (node->nodeType != NODE_KERNEL_MODULE_ID) {
        char vxModuleId [50]; /* this enough to store the id */
        write_stream(out, ',');
        json_write_string(out, KEY_ModuleID);
        write_stream(out, ':');
        snprintf(vxModuleId, 50, "0x%lx", (long unsigned) node->vxModId);
        json_write_string(out, vxModuleId);
    }

    write_stream(out, ',');
    json_write_string(out, KEY_ModuleFile);
    write_stream(out, ':');
    if (node->nodeType == NODE_KERNEL_MODULE_ID) {
        char * kernel = NULL;
        BOOT_PARAMS bootParams;    /* VxWorks boot parameters */
        char * pModule;

        /* Add VxWorks module (From bootline) */
        bootStringToStruct (sysBootLine, &bootParams);
        kernel = bootParams.bootFile;
        if (kernel == NULL) {
            kernel = VX_KERNEL_MODULE_NAME;
        }

        json_write_string(out, kernel);

    } else {
        if (node->vxModId->nameWithPath != NULL) {
            json_write_string(out, node->vxModId->nameWithPath);
        } else {
            json_write_string(out, "No file path available");
        }
    }

    if (node->nodeType != NODE_KERNEL_MODULE_ID) {
        write_stream(out, ',');
        json_write_string(out, KEY_Signatue);
        write_stream(out, ':');
        write_stream(out, '{');
        /*
        json_write_string(out, KEY_Signatue_CRC);
        write_stream(out, ':');
        json_write_ulong(out, 0);
        */
        write_stream(out, '}');
    }

    write_stream(out, ',');
    json_write_string(out, KEY_ModuleType);
    write_stream(out, ':');
    if (node->nodeType == NODE_KERNEL_MODULE_ID) {
        json_write_string(out, VAL_KernelImage);
    } else {
        json_write_string(out, VAL_KernelModule);
    }

    write_stream(out, ',');
    json_write_string(out, KEY_CanUnload);
    write_stream(out, ':');
    if (node->nodeType == NODE_KERNEL_MODULE_ID) {
        json_write_boolean(out, 0);
    } else {
        json_write_boolean(out, 1);
    }
}

/*
 * cb_getFilter - extract one filter value of cmd getChildren
 * Example format {"Filter":{"Type":"CommonSymbolList"}
 */
static void cb_getFilter (InputStream * inp, const char * name, void * args) {
    int * filter = (int *) args;
    /* Today only the KEY_Type is supported */
    if ((*filter == NODE_UNKNOW_ID) && (strcmp (name, KEY_Type) == 0)) {
        char * filterStr = json_read_alloc_string(inp);
        if (strcmp (filterStr, VAL_TypeModule) == 0) {
            *filter = NODE_MODULE_ID;
        } else if (strcmp (filterStr, VAL_TypeCommonSymbolList) == 0) {
            *filter = NODE_SYM_LST_COMMON_ID;
        } else if (strcmp (filterStr, VAL_TypeUndefSymbolList) == 0) {
            *filter = NODE_SYM_LST_UNDEF_ID;
        } else if (strcmp (filterStr, VAL_TypeSection) == 0) {
            *filter = NODE_SECTION_LST_ID;
        } else if (strcmp (filterStr, VAL_TypeSegment) == 0) {
            *filter = NODE_SEGMENT_LST_ID;
        } else {
            *filter = NODE_UNKNOW_ID;
        }
    } else {
        /* Unsupported entry, just skip it for now. */
        json_skip_object(inp);
    }
}

/*
 * cb_getFilter - extract one filter value of cmd getChildren
 * Example format {"Filter":{"Type":"CommonSymbolList"}
 */
static void cb_getFilterObject (InputStream * inp, const char * name, void * args) {
    int * filter = (int *) args;
    if ((*filter == NODE_UNKNOW_ID) && (strcmp (name, KEY_Filter) == 0)) {
        int filterValue = 0;
        json_read_struct(inp, cb_getFilter, &filterValue);
        *filter = filterValue;
    } else {
        /* Unsupported entry, just skip it for now. */
        json_skip_object(inp);
    }
}

/*
 * cb_loadFlags - extract one option value of cmd load
 * Example format {"LoadFlags":8}
 */
static void cb_loadFlags (InputStream * inp, const char * name, void * args) {
    int * loadFlags = (int *) args;
    if (strcmp (name, KEY_LoadFlags) == 0) {
        *loadFlags = (int)json_read_ulong(inp);
    } else {
        /* Unsupported entry, just skip it for now. */
        char *objStr = json_read_object(inp);
        trace(LOG_ALWAYS,  "%s.load : Unsupported option \"%s\":%s\n", WR_MODULE, name, objStr);
        loc_free(objStr);
    }
}

/*
 * cb_unloadFlags - extract one option value of cmd unload
 * Example format {"UnloadFlags":8}
 */
static void cb_unloadFlags (InputStream * inp, const char * name, void * args) {
    int * loadFlags = (int *) args;
    if (strcmp (name, KEY_UnloadFlags) == 0) {
        *loadFlags = (int)json_read_ulong(inp);
    } else {
        /* Unsupported entry, just skip it for now. */
        char *objStr = json_read_object(inp);
        trace(LOG_ALWAYS,  "%s.unload : Unsupported option \"%s\":%s\n", WR_MODULE, name, objStr);
        loc_free(objStr);
    }
}

/*
 * getChildren - returns the module ID list of the given execution context
 *
 * C . <token> . WrModuleLoad . getChildren . <string: context ID> . <object: filter>
 *
 * This command retrieves the module list of the given context ID. It's used to
 * discover all modules of a context.
 *
 * The parameter <object: filter> is not mandatory.
 *
 * Reply:
 *     R . <token> . <error report> . <array of IDs>
 *         <array of  IDs>
 *                  null
 *                  [ <ID list> ]
 *
 *             <ID list>
 *                      <string:  ID>
 *                      <ID list> , <string: ID>
 *
 */
static void command_getChildren(char * token, Channel * chnl) {
    int error = 0;
    char * parentCtxId = NULL;
    BOOL validCtx = FALSE;
    ModuleDesc * node = NULL;
    int ch = 0;
    int filter = NODE_UNKNOW_ID;

    parentCtxId = json_read_alloc_string(&chnl->inp);
    json_test_char(&chnl->inp, MARKER_EOA);

    /* Skip white space if any */
    while (1) {
        ch = json_peek(&chnl->inp);
        if (isspace(ch)) {
            read_stream(&chnl->inp);
        } else {
            break;
        }
    }

    if (ch == '{') {
        json_read_struct(&chnl->inp, cb_getFilterObject, &filter);
        json_test_char(&chnl->inp, MARKER_EOA);
    } else if (ch != MARKER_EOM) {
        /* drop this object, we do not support it. */
        json_skip_object(&chnl->inp);
        json_test_char(&chnl->inp, MARKER_EOA);
    }

    json_test_char(&chnl->inp, MARKER_EOM);

    if (parentCtxId != NULL) {

        node = id2module(parentCtxId);
        if (node != NULL) {
            /* We got a module, that's fine */
        } else {
            /*
             * This is not a module ID context, so
             * check the Process context validity
             */
            validCtx = is_tcfid_valid (parentCtxId);
            if (!validCtx) {
                error = ERR_INV_CONTEXT;
            }
        }
    } else {
        error = ERR_INV_CONTEXT;
    }

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);
    write_errno(&chnl->out, error);

    if (error != 0) {

        /*
         * Output an error message
         */

        /* No children available so NULL */
        json_write_string(&chnl->out, NULL);

    } else if (validCtx) {
        write_stream(&chnl->out, '[');

        getChildrenOfProcess(&chnl->out, parentCtxId);
        write_stream(&chnl->out, ']');

    } else if (node != NULL) {
        write_stream(&chnl->out, '[');

        switch (node->nodeType) {
            case NODE_UNKNOW_ID:
                printf ("[%s:%d] Not a valid context of getChildren (%s)\n", __FILE__, __LINE__, parentCtxId);
                break;

            case NODE_MODULE_ID: {

                getChildrenOfModule(&chnl->out, node, filter);
                break;
                }

            case NODE_KERNEL_MODULE_ID:
                getChildrenOfModule(&chnl->out, node, filter);
                break;
        }
        write_stream(&chnl->out, ']');
    }
    
    write_stream(&chnl->out, MARKER_EOA); /* end of string */
    write_stream(&chnl->out, MARKER_EOM);

    if (parentCtxId != NULL) {
        loc_free(parentCtxId);
    }
}

/*
 * command_getCapabilities - get the service capabilities
 *
 *  C . <token> . WrModuleLoad . getCapabilities . <string context ID> .
 *
 * The command returns the capabilities of the services.
 *
 * Reply:
 * R . <token> . <error report> . <object: capabilities>
 *
 */
static void command_getCapabilities(char * token, Channel * chnl) {
    int error = 0;
    char * ctxId = NULL;
    VX_OBJ_INFO *ctxObjInfo = NULL;

    ctxId = json_read_alloc_string(&chnl->inp);
    json_test_char(&chnl->inp, MARKER_EOA);

    json_test_char(&chnl->inp, MARKER_EOM);

    /* check the context validity*/
    ctxObjInfo = vx_get_obj_info (ctxId);
    if (ctxObjInfo == NULL) {
        error = ERR_INV_CONTEXT;
    }

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);
    write_errno(&chnl->out, error);

    if (error == 0) {
        write_stream(&chnl->out, '{');
        if (ctxObjInfo->obj_type == VX_OBJ_KERNEL) {

            json_write_string(&chnl->out, KEY_CMD_LOAD);
            write_stream(&chnl->out, ':');
            write_stream(&chnl->out, '{');
                json_write_string(&chnl->out, KEY_KernelModule);
                write_stream(&chnl->out, ':');
                json_write_boolean(&chnl->out, 1);
            write_stream(&chnl->out, '}');

            write_stream(&chnl->out, ',');

            json_write_string(&chnl->out, KEY_CMD_UNLOAD);
            write_stream(&chnl->out, ':');
            write_stream(&chnl->out, '{');
                json_write_string(&chnl->out, KEY_KernelModule);
                write_stream(&chnl->out, ':');
                json_write_boolean(&chnl->out, 1);
            write_stream(&chnl->out, '}');
        }

        write_stream(&chnl->out, '}');

    } else {

        /* no list available */
        json_write_string(&chnl->out, NULL);
    }

    write_stream(&chnl->out, MARKER_EOA);
    write_stream(&chnl->out, MARKER_EOM);

    if (ctxId != NULL) {
        loc_free(ctxId);
    }

    if (ctxObjInfo != NULL) {
        vx_free_obj_info (ctxObjInfo);
    }
}

/*
 * command_getProperties - The command retrieves context info for given context IDs.
 *
 * Command allows to query multiple contexts at once. The root node context ID must be a ProcessID that is a memory context.
 *
 * Reply:
 * R . <token> . <array of context data> . <error report> .
 *
 * <array of context data>
 *     => null
 *     => [ ]
 *     => [ <context data list> ]
 * <context data list>
 *     => <context data>
 *     => <context data list> , <context data>
 *
 * <context data>
 *     => null
 *     => <object>
 *
 */
static void command_getProperties(char * token, Channel * chnl) {
    int error = 0;
    ModuleDesc * node = NULL;
    int idx;
    int idsCnt = 0;
    char **ids;

    ids = json_read_alloc_string_array(&chnl->inp, &idsCnt);

    json_test_char(&chnl->inp, MARKER_EOA);
    json_test_char(&chnl->inp, MARKER_EOM);

    for (idx = 0; idx < idsCnt; idx++) {
        node = NULL;

        node = id2module(ids[idx]);
        if (node == NULL) {
            error = ERR_INV_CONTEXT;
            break;
        }
    }

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);
    write_errno(&chnl->out, error);


    if (error != 0) {
         /* no list to output */
        write_stream(&chnl->out, MARKER_EOA);

    } else {
        write_stream(&chnl->out, '[');

        for (idx = 0; idx < idsCnt; idx++) {
            node = NULL;

            node = id2module(ids[idx]);
            assert (node != NULL);

            if (idx > 0) {
                write_stream(&chnl->out, ',');
            }

            if (node != NULL) {
                write_stream(&chnl->out, '{');

                switch (node->nodeType) {
                    case NODE_UNKNOW_ID: {
                        printf ("[%s:%d] Not a valid context of getContext (%s)\n", __FILE__, __LINE__, ids[idx]);
                        break;
                    }

                    case NODE_MODULE_ID: {
                        getPropertiesOfModule(&chnl->out, ids[idx], node);
                        break;
                    }

                    case NODE_KERNEL_MODULE_ID: {
                        getPropertiesOfModule(&chnl->out, ids[idx], node);
                        break;
                    }

                        /* Symbol list */
                    case NODE_SYM_LST_UNDEF_ID: {
                        getPropertiesOfSymUndef(&chnl->out, ids[idx], node);
                        break;
                    }
                    case NODE_SYM_LST_COMMON_ID: {
                        getPropertiesOfSymCommon(&chnl->out,ids[idx], node);
                        break;
                    }

                    /* Section & Segments */
                    case NODE_SECTION_LST_ID: {
                        getPropertiesOfSection(&chnl->out, ids[idx], node);
                        break;
                    }
                    case NODE_SEGMENT_LST_ID: {
                        getPropertiesOfSegment(&chnl->out, ids[idx], node);
                        break;
                    }
                }
                write_stream(&chnl->out, '}');
            }
        }

        write_stream(&chnl->out, ']');

        write_stream(&chnl->out, MARKER_EOA);
    }


    write_stream(&chnl->out, MARKER_EOM);

    loc_free(ids);
}

/*
 * command_load - load the module
 *
 * C . <token> . WrModuleLoad . load . <string: context ID> . <string: fileName> . <object : loadFlags>
 *
 * This command is a wrapper to the vxWorks API moduleLoadAt().
 *
 * Reply:
 * R . <token> . <error report> .
 *
 */

static void command_load(char * token, Channel * chnl) {

    int error = 0;
    char * fileName = NULL;
    char * ctxId = NULL;
    int loadFlags = 0;
    BOOL validCtx = FALSE;

    ctxId = json_read_alloc_string(&chnl->inp);
    json_test_char(&chnl->inp, MARKER_EOA);

    fileName = json_read_alloc_string(&chnl->inp);
    json_test_char(&chnl->inp, MARKER_EOA);

    json_read_struct(&chnl->inp, cb_loadFlags, (void *)&loadFlags);
    json_test_char(&chnl->inp, MARKER_EOA);

    json_test_char(&chnl->inp, MARKER_EOM);

    /* check the context validity*/
    if (error == 0) {
        validCtx = is_tcfid_valid (ctxId);
        if (!validCtx) {
            error = ERR_INV_CONTEXT;
        }
    }

    if ((error == 0) && (fileName == NULL)) {
        /* Invalid file name */
        error = ERR_UNSUPPORTED;
    }

    /* Only reply in case of error because wrong ID is given */
    if (error) {
        write_stringz(&chnl->out, "R");
        write_stringz(&chnl->out, token);
        write_errno(&chnl->out, error);
        write_stream(&chnl->out, MARKER_EOA); /* No moduleID */
        write_stream(&chnl->out, MARKER_EOM);
        if (ctxId != NULL) {
            loc_free(ctxId);
        }
        if (fileName != NULL) {
            loc_free(fileName);
        }
    } else {
        /* Delegate this work to a worker Thread */
        Wr_ModuleLoad_Worker (token, chnl, ctxId, fileName, loadFlags);
    }
}

/*
 * command_unload - unload the specified module
 *
 * C . <token> . WrModuleLoad . unload . <string: module ID> .
 *
 * This is a wrapper of vxWorks API unldByModuleId.
 * See the vxWorks documentation for more details about unldByModuleId
 *
 * Reply:
 * R . <token> . <error report> .
 *
 */
static void command_unload(char * token, Channel * chnl) {
    int error = 0;
    char * modStrId = NULL;
    char * ctxStrId = NULL;
    ModuleDesc * modId = NULL;
    int unloadFlags = 0;
    BOOL validCtx = FALSE;

    modStrId = json_read_alloc_string(&chnl->inp);
    json_test_char(&chnl->inp, MARKER_EOA);

    json_read_struct(&chnl->inp, cb_unloadFlags, (void *)&unloadFlags);
    json_test_char(&chnl->inp, MARKER_EOA);

    json_test_char(&chnl->inp, MARKER_EOM);

    /* search the modID */
    if ((error == 0) && (modStrId != NULL)) {
        modId = id2module(modStrId);
        if (modId == NULL) {
            /*
            error = ERR_OTHER;
            error = set_errno(ERR_OTHER, "No such module");
            */
            error = EINVAL;
        } else {
            ctxStrId = modId->ctxId;
        }
    }

    /* check the context validity*/

    if ((error == 0) && (ctxStrId != NULL)) {
        validCtx = is_tcfid_valid (ctxStrId);
        if (!validCtx) {
            error = ERR_INV_CONTEXT;
        }
    }

    if (error == 0) {
        /* Option and context are valid */
        if (unldByModuleId(modId->vxModId, unloadFlags)) {
            error = set_errno(ERR_OTHER, strerror(errnoGet()));
        }
    }

    write_stringz(&chnl->out, "R");
    write_stringz(&chnl->out, token);
    write_errno(&chnl->out, error);

    write_stream(&chnl->out, MARKER_EOM);

    if (modStrId != NULL) {
        loc_free(modStrId);
    }

    moduleDescFree(modId);
}

/*
 * ini_wr_module_service - initialize the service WR_ModuleLoad.
 *
 */
void ini_wr_module_service(Protocol * proto, TCFBroadcastGroup * bcg) {
    broadcast_group = bcg;

    add_channel_open_listener(event_channel_opened);
    add_channel_close_listener(event_channel_closed);
    add_command_handler(proto, WR_MODULE, "getCapabilities", command_getCapabilities);
    add_command_handler(proto, WR_MODULE, "getChildren", command_getChildren);
    add_command_handler(proto, WR_MODULE, "getProperties", command_getProperties);

    add_command_handler(proto, WR_MODULE, "load", command_load);
    add_command_handler(proto, WR_MODULE, "unload", command_unload);
}

#endif /* SERVICE_WrModule */

