/* linux_trace.c - EP exposing linux trace events */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
05sep14,tle  Fix dispatching events in stopped state (WB4-4006)
*/

#include <unistd.h>
#include <string.h>  /* for memset() and friends */
#include <fcntl.h>
#include <ftw.h>
#include <ctype.h>
#include <error.h>
#include <errno.h>
#include <pthread.h>
#include <poll.h>
#include <dirent.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/param.h>  /* for MAX() */
#if defined(HAVE_CAFE_CONFIG_H)
#  include <cafe/config.h>
#endif
#if defined(HAVE_PR_SET_NAME)
#  include <sys/prctl.h>
#endif
#include <cafe/cafe.h>
#include <cafe/autogenerated/perf_mem_events.type.h>
#include <cafe/stdio.h>
#include <cafe/atomic.h>
#include <cafe/ep/linux_trace.h>
#include <cafe/byte_buffer.h>
#include <cafe/hashtbl.h>
#include <cafe/array.h>
#include <cafe/utils.h>
#include "symresolve_actions.h"
#include "../../platforms/linux/include/perf_lib.h"
#include "../../platforms/linux/include/memory_map.h"
#include "../../platforms/linux/include/kernel_map_parser.h"
#include "../src/tokenizer.h"
#include "../src/time_utils.h"  /* NSECS_PER_SEC */
//#include "/usr/src/linux/include/linux/perf_event.h"
#include <include/linux/perf_event.h>
#include <asm/unistd.h>  /* for __NR_perf_event_open */

#define PERF_ANNOTATIONS_DEFAULT_FILENAME "/etc/cafe/perf.annotations"

#if !defined(__NR_perf_event_open)
//#define __NR_perf_event_open  (319)  /* ppc */
#define __NR_perf_event_open  (336)  /* x86 */
#endif

#define alignof(TYPE) offsetof(struct { char c; TYPE val; }, val)

/* strictly used for debugging */
#if defined(RAW_DUMP_DEBUG)
static FILE *out = NULL;
#endif


#undef TRACE_MEMTBL_DEBUG
#if defined(TRACE_MEMTBL_DEBUG)
#  define MTD_LOG(fmt, ...)			\
	fprintf(stderr, fmt, ## __VA_ARGS__)
#else
#  define MTD_LOG(fmt, ...)
#endif


/* default MULTI_BUFFER to enabled */
static const char *MULTI_BUFFER = "";


/* event types defined specifically by this EP */
enum EP_defined_event_types {
	INTRINSIC_PERF_EVENT = PERF_TYPE_MAX,  /* fork, comm, mmap, etc. */
	SYNTHETIC_EVENT
};


/* maps a linux trace event spec to the per-cpu identifier and controls, as
   well as the per-EP-instance data, like action lists */
struct event_action_map {
	enum perf_type_id event_type;
	union {
		enum perf_hw_id hardware_event;
		enum perf_sw_ids software_event;
		uint64_t hw_cache_event;
		struct linux_event *trace_event;
	};
	int (*state_dump)(struct cafe_event_action_info *h,
			  struct cafe_ep_inst *inst);
};


/*  The per-cpu event bookkeeping info */
struct event_cpu_handle {
	bool hashed;
	uint64_t stream_id;  /* id used in the kernel's buffer to
				represent this event */
	int fd;
	/* previous count value, to calculate actual skipcount as
	   delta */
	uint64_t last_event_count;
	struct instance_data *event_instance;
};


/* the per-EP-instance bookkeeping data */
struct instance_data {
	struct cafe_event_action_info *action_info;
	unsigned long change_count;
	struct event_action_map *event_map;
	uint64_t buffer_marker;

	/* support requesting different odata at runtime */
	uint64_t sample_type;

	uint64_t skipcount_limit;
	uint64_t skipcount_freq;

	struct event_cpu_handle cpu[0];
};


struct linux_trace_workarea {
	unsigned long buffer_size;
};


static void syscall_brk_memtable_handler(struct cafe_event_action_info *info,
					 void *event_data[]);
static void syscall_mmap_memtable_handler(struct cafe_event_action_info *info,
					  void *event_data[]);
static void syscall_mmap_pgoff_memtable_handler(struct cafe_event_action_info *info,
						void *event_data[]);
static void syscall_munmap_memtable_handler(struct cafe_event_action_info *info,
					    void *event_data[]);
struct memtable_input_spec {
	const char *name;
	const CAFE_ACTION_HANDLER handler;
	char *vt;
	struct event_action_map map;
} memtable_input_specs[] = {
	{"syscalls.sys_exit_brk", syscall_brk_memtable_handler},
	{"syscalls.sys_enter_mmap", syscall_mmap_memtable_handler},
	{"syscalls.sys_enter_mmap_pgoff", syscall_mmap_pgoff_memtable_handler},
	{"syscalls.sys_enter_munmap", syscall_munmap_memtable_handler}
};


struct process_memory_record {
	uint32_t pid;
	uint64_t heap_base;
	uint64_t heap_top;
	uint64_t mapped_memory;
#if defined(TRACE_MEMTBL_DEBUG)
	char pname[1024];
	uint32_t ppid;
#endif
};


struct memtable_inst_data {
	struct cafe_event_descr *event;
	struct cafe_event_action_info *info;
};


struct event_buffer {
	/* file descriptor controlling this buffer */
	int fd;
	struct perf_event_mmap_page *buffer;
	const char *data;  /* pointer to the first byte of the
			      ring-buffer */
	unsigned long last_read;
	struct cafe_byte_buffer wrap_buf;
};


struct linux_trace_inst {
	bool timestamp_config_event_dispatch_pending;
	struct cafe_event_action_info *timestamp_config_action_info;
	struct cafe_value_type *timestamp_config_type;
	struct cafe_value_type *timestamp_type;

	unsigned long change_count;

	/* lock protecting access to emap members. */
	pthread_mutex_t actions_lock;

	/* Thread responsible for consuming the kernel's event buffer */
	pthread_t reader_thread;
	bool reader_should_exit;

	/* tracks the file descriptors we need to wake up on after calling
	   poll(). */
	struct pollfd *event_poll;

	/* Used by EP instances to signal the reader thread that the event_poll
	   array has been updated */
	int reader_wake_pipe[2];

	struct instance_data perf_event_idata[PERF_RECORD_SAMPLE];

	/* map of stream-id to event action map objects */
	struct cafe_hashtbl *emap;

	/* number of pages worth of buffer to map.  Must be 1 + 2^n  */
	unsigned int buf_pages;

	/* file descriptor which all events are redirected to */
	int *group_fd;

	/* a hash of: (PID) -> (process_memory_record) */
	struct cafe_hashtbl *process_memory_map;
	struct cafe_value_type *perf_mmap_type;
	struct cafe_value_type *perf_comm_type;
	struct cafe_value_type *perf_fork_type;
	struct cafe_value_type *perf_exit_type;

	struct cafe_value_type *memtbl_heap_type;
	struct cafe_event_action_info *memtbl_heap_info;
	struct cafe_value_type *memtbl_mmap_type;
	struct cafe_event_action_info *memtbl_mmap_info;

	struct cafe_ep_inst *inst;
	struct memtable_inst_data memtbl_input_events[CAFE_NELEMENTS(memtable_input_specs)];

	struct event_buffer buffers[0];
};


static int trace_create(struct cafe_ep_inst *inst);
static void trace_destroy(struct cafe_ep_inst *inst);
static int trace_start(struct cafe_ep_inst *inst);
static int trace_stop(struct cafe_ep_inst *inst);
static int trace_update_output_event_actions(struct cafe_ep_inst *inst);
static unsigned int update_poll_fds(struct linux_trace_inst *ti);
static void update_poll_fds_signal(int signal_fd);
static int timestamp_config_state_dump_handler(struct cafe_event_action_info *h,
					       struct cafe_ep_inst *inst);
static void disable_events_for_inst(const struct cafe_ep_inst *inst);
static void *parse_buffer_thread(void *arg);


static const struct {
	const char *name;
	int (*state_dump)(struct cafe_event_action_info *h,
			  struct cafe_ep_inst *inst);
} synthesized_events[] = {
	{cafe_timestamp_config_VT_REF, timestamp_config_state_dump_handler},
	{perf_mem_heap_VT_REF},
	{perf_mem_mmap_VT_REF}
};


/* page size of the current process */
static unsigned long page_size;


/* number of cpus on the current machine */
static unsigned int num_cpus;


/*  */
static const unsigned int INITIAL_HASH_SIZE = 64;


static const unsigned long DEFAULT_BUFFER_SIZE = 1024 * 1024;


/* In order to provide monotonically increasing timestamps for events which
   perf doesn't timestamp itself, we track the most recently seen real timestamp
   to tack on to timestamp-less events. */
static struct cafe_timestamp current_time = {
	.timestamp = 0
};


/* array of cafe_ep_spec_output_evt objects.  Each of these objects' impl_data
   handle points to the associated event_action_map object */
static struct cafe_array linux_event_specs;


/* Maps perf event type to specific action lists. */
struct event_action_map *perf_event_action_map[PERF_RECORD_SAMPLE];


/* The type of data requested for every perf "sample" event (which includes
   all software tracepoints.  This data is then conditionally relayed with
   individual events based on which orthogonal data pieces are selected on
   a given instance. These are the odata pieces which are required by the perf
   EP to adequately parse the sample.  There are additional pieces which can
   be optionally generated. */
static const uint64_t SAMPLE_TYPE = PERF_SAMPLE_IP | PERF_SAMPLE_TID
	| PERF_SAMPLE_TIME | PERF_SAMPLE_ADDR | PERF_SAMPLE_CPU
	| PERF_SAMPLE_ID | PERF_SAMPLE_STREAM_ID
	| PERF_SAMPLE_READ;

//td: make period, read, conditionally switched

/* symbolic names for indices in the sample_type_offsets array */
enum perf_sample_types_offset {
	PERF_SAMPLE_IP_OFFSET,
	PERF_SAMPLE_CONTEXT_OFFSET,
	PERF_SAMPLE_TIME_OFFSET,
	PERF_SAMPLE_ADDR_OFFSET,
	PERF_SAMPLE_ID_OFFSET,
	PERF_SAMPLE_STREAM_ID_OFFSET,
	PERF_SAMPLE_CPU_OFFSET,
	PERF_SAMPLE_PERIOD_OFFSET,
	PERF_SAMPLE_READ_OFFSET,
	PERF_SAMPLE_CALLCHAIN_OFFSET,
	PERF_SAMPLE_RAW_OFFSET,
	NUM_PERF_SAMPLE_TYPES
};


/* Size of data chunk appended when PERF_SAMPLE_READ is asserted.  The size of
   this data is variable based on the value of the .read_format attribute (set
   when the event fd is opened) but currently we hardcode this to:
   sizeof(value) + sizeof(PERF_FORMAT_ID) */
#define SIZEOF_READ_FORMAT  sizeof(uint64_t) + sizeof(uint64_t)

static const unsigned int ODATA_SIZES[NUM_PERF_SAMPLE_TYPES] = {
	/* PERF_SAMPLE_IP */ sizeof(uint64_t),
	/* PERF_SAMPLE_TID */ sizeof(struct perf_context),
	/* PERF_SAMPLE_TIME */ sizeof(uint64_t),
	/* PERF_SAMPLE_ADDR */ sizeof(uint64_t),
	/* PERF_SAMPLE_ID */ sizeof(uint64_t),
	/* PERF_SAMPLE_STREAM_ID */ sizeof(uint64_t),
	/* PERF_SAMPLE_CPU */ sizeof(uint64_t),
	/* PERF_SAMPLE_PERIOD */ sizeof(uint64_t),
	/* PERF_SAMPLE_READ */ SIZEOF_READ_FORMAT
};


/* array of offsets indicating where in a sample we can find a particular
   orthogonal sample data type. */
static unsigned int sample_type_offsets[NUM_PERF_SAMPLE_TYPES];

#define EP_TYPES							\
	PERF_MEM_EVENTS_TYPE_ALL_VT_DEFS				\
	cafe_core_id_VT_DEF						\
	cafe_CURRENT_CORE_ID_VT_DEF					\
	cafe_COUNTER_METRIC_CLASS_VT_DEF				\
	cafe_SUM_METRIC_CLASS_VT_DEF					\
	cafe_METRIC_CATEGORY_VT_DEF					\
	cafe_METRIC_CLASS_VT_DEF					\
	cafe_output_inst_thread_VT_DEF					\
	cafe_sample_freq_VT_DEF						\
	cafe_setting_VT_DEF						\
	cafe_skipcount_VT_DEF						\
	cafe_timestamp_config_VT_DEF					\
	cafe_timestamp_VT_DEF						\
	LX_CURRENT_THREAD_ID_DEF					\
	LX_CURRENT_PROCESS_ID_DEF					\
	PERF_ADDR_VT_DEF						\
	PERF_COMM_DEF							\
	PERF_CONTEXT_VT_DEF						\
	PERF_EXIT_DEF							\
	PERF_FORK_DEF							\
	PERF_ID_VT_DEF							\
	PERF_IP_VT_DEF							\
	PERF_MMAP_DEF							\
	PERF_PERIOD_VT_DEF						\
	PERF_READ_DEF							\
	PERF_RECORD_LOST_DEF						\
	PERF_STACK_TRACE_VT_DEF						\
	PERF_STREAM_ID_VT_DEF						\
	PERF_THROTTLE_DEF						\
	PERF_UNTHROTTLE_DEF						\
	"S[trace.workarea,v[buffer_size,L," cafe_setting_VT_NAT "],"	\
	cafe_output_inst_thread_VT_NAT "]"

/* indices into orthogonal data items */
enum {
	ODATA_INDEX_CONTEXT,
	ODATA_INDEX_TIMESTAMP,
	ODATA_INDEX_IP,
	ODATA_INDEX_ADDR,
	ODATA_INDEX_ID,
	ODATA_INDEX_STREAM_ID,
	ODATA_INDEX_CPU,
	ODATA_INDEX_PERIOD,
	ODATA_INDEX_STACK,
	ODATA_INDEX_SKIPCOUNT,
	ODATA_INDEX_SAMPLE_FREQ,
	NUM_ODATA_ITEMS
};


/* Orthogonal data types available for perf sample events.  Other perf event
   types may need to synthesize this information */
static const char *output_odata[] = {
	PERF_CONTEXT_VT_REF,
	cafe_timestamp_VT_REF,
	PERF_IP_VT_REF,
	PERF_ADDR_VT_REF,
	PERF_ID_VT_REF,
	PERF_STREAM_ID_VT_REF,
	cafe_core_id_VT_REF,
	PERF_PERIOD_VT_REF,
	PERF_STACK_TRACE_VT_REF,
	cafe_skipcount_VT_REF,
	cafe_sample_freq_VT_REF
};


static int od_set_all[] = {
	ODATA_INDEX_CONTEXT,
	ODATA_INDEX_TIMESTAMP,
	ODATA_INDEX_IP,
	ODATA_INDEX_ADDR,
	ODATA_INDEX_ID,
	ODATA_INDEX_STREAM_ID,
	ODATA_INDEX_CPU,
	ODATA_INDEX_PERIOD,
	ODATA_INDEX_STACK,
	ODATA_INDEX_SKIPCOUNT,
	ODATA_INDEX_SAMPLE_FREQ
};

static int od_set_no_stacktrace[] = {
	ODATA_INDEX_CONTEXT,
	ODATA_INDEX_TIMESTAMP,
	ODATA_INDEX_IP,
	ODATA_INDEX_ADDR,
	ODATA_INDEX_ID,
	ODATA_INDEX_STREAM_ID,
	ODATA_INDEX_CPU,
	ODATA_INDEX_PERIOD,
	ODATA_INDEX_SKIPCOUNT,
	ODATA_INDEX_SAMPLE_FREQ
};


static int od_set_ts[] = {
	ODATA_INDEX_TIMESTAMP
};


static int od_set_memtbl[] = {
	ODATA_INDEX_CONTEXT
};


/* indices into od_sets array */
enum {
	OD_SET_ALL = 0,
	OD_SET_TS,
	OD_SET_MEMTBL,
	OD_SET_NO_STACKTRACE,
	OD_SET_NONE
};


static struct cafe_ep_spec_output_od_set od_sets[] = {
	{CAFE_NELEMENTS(od_set_all), od_set_all},
	{CAFE_NELEMENTS(od_set_ts), od_set_ts},
	{CAFE_NELEMENTS(od_set_memtbl), od_set_memtbl},
	{CAFE_NELEMENTS(od_set_no_stacktrace), od_set_no_stacktrace},
	{0, NULL}
};


static struct cafe_ep_spec spec = {
	"linux_trace",
	trace_create,			/* create_hook */
	trace_destroy,			/* delete_hook */
	trace_start,
	trace_stop,
	NULL,                           /* get_input_event_actions */
	trace_update_output_event_actions,
	NULL,
	NULL,		       /* types: populated by registration. */
	"t[trace.workarea]",
	0,
	NULL,		/* input events */
	0,
	NULL,		/* output events, populated as part of registering the
			   EP spec. */
	CAFE_NELEMENTS(output_odata),
	output_odata,
	NULL,           /* settings commit */
	NULL,           /* input_state_change */
	CAFE_NELEMENTS(od_sets),
	od_sets
};


/**
 *  Set the NONBLOCK flag on the given file descriptor.
 */
static
bool set_fd_nonblocking(int fd)
{
	long flags = fcntl(fd, F_GETFL);

	if (flags < 0) {
		perror("Error getting file descriptor flags");
		return false;
	}

	flags |= O_NONBLOCK;
	if (fcntl(fd, F_SETFL, flags) < 0) {
		perror("Error setting file descriptor flags");
		return false;
	}

	return true;
}


static
unsigned long hashtbl_int64_hash(const void *k)
{
	//td: this could be improved.
	const int64_t *i64 = k;
	return 0xffffffff & *i64;
//	return int32hash(*((const unsigned int *)k));
}


static
int hashtbl_int64_equals(const void *a, const void *b)
{
	return *((const int64_t *)a) == *((const int64_t *)b);
}


/**
 *  Delete memtable entry row.  Meant to be used as a hashtbl_delete() callback
 */
static
void memtable_row_delete(void *v, void *arg)
{
	const struct cafe_allocator *a = arg;
	CAFE_FREE(a, v);
}


/**
 *  Builds and initializes a perf buffer for use by a single linux trace EP
 * instance.  It accomplishes this by activating a single non-intrusive event,
 * flagged to generate perf's infrastructure events (comm, fork, exit, mmap,
 * etc.).
 */
static
int perf_buffer_create(struct event_buffer *buf, unsigned long size,
		       unsigned int cpu, unsigned int master_fd)
{
	const pid_t pid = -1;
	const unsigned long flags = 0;
	struct perf_event_attr attr;
	int fd;

	/* this combination of attributes sets up the event to be a clock tick
	   that never fires (0 Hz) */
	memset(&attr, 0, sizeof(attr));
	attr.type = PERF_TYPE_SOFTWARE;
	attr.sample_type = 0;
	attr.freq = 1;
	attr.sample_freq = 0;
	attr.config = PERF_COUNT_SW_CPU_CLOCK;
	attr.size = sizeof(attr);
	attr.mmap = 1;
	//attr.munmap = 1;
	//attr.mmap_data = 1;
	attr.comm = 1;
	attr.task = 1;
	attr.read_format = PERF_FORMAT_ID;
	attr.inherit = 1;
	//attr.disabled = 1;

	fd = syscall(__NR_perf_event_open, &attr, pid, cpu, -1, flags);
	if (fd < 0) {
		fprintf(stderr,
			"Error opening event file descriptor %llu: %s (%d)\n",
			attr.config, strerror(errno), errno);
		return CAFE_ERROR;
	}

	if ((MULTI_BUFFER == NULL) && (cpu != 0)) {
		if (ioctl(fd, PERF_EVENT_IOC_SET_OUTPUT, master_fd) == -1) {
			perror("failed to redirect output");
			close(fd);
			return CAFE_ERROR;
		}
	} else {
		buf->buffer = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
		if (buf->buffer == MAP_FAILED) {
			perror("failed to mmap");
			close(fd);
			return CAFE_ERROR;
		}

		buf->data = (char *) buf->buffer + page_size;
		buf->last_read = 0;
	}

	buf->fd = fd;
	return CAFE_OK;
}


/**
 *  Close a trace EP's perf buffer, releasing all related resources.
 */
static
void perf_buffer_destroy(struct event_buffer *buf, unsigned long size)
{
	if (close(buf->fd))
		perror("Error closing perf group fd");
	if (munmap(buf->buffer, size))
		perror("Error unmapping perf buffer");
	buf->buffer = MAP_FAILED;
	buf->data = 0;
	buf->fd = -1;
}


static
int trace_create(struct cafe_ep_inst *inst)
{
	struct linux_trace_workarea *w =
		(struct linux_trace_workarea *) inst->workarea->u.data;
	struct linux_trace_inst *ti;
	int rc;
	unsigned int i;

	ti = CAFE_ALLOC(inst->allocator, sizeof(*ti) + num_cpus * sizeof(struct event_buffer));
	if (ti == NULL)
		return CAFE_NOMEM;

	ti->timestamp_config_event_dispatch_pending = false;
	rc = cafe_type_parse(inst->session->ns, cafe_timestamp_config_VT_DEF,
			     &ti->timestamp_config_type);

	if (rc != CAFE_OK) {
		CAFE_FREE(inst->allocator, ti);
		return rc;
	}

	rc = cafe_type_parse(inst->session->ns, cafe_timestamp_VT_DEF,
			     &ti->timestamp_type);

	if (rc != CAFE_OK) {
		CAFE_FREE(inst->allocator, ti);
		return rc;
	}

	if (w->buffer_size == 0)
		w->buffer_size = DEFAULT_BUFFER_SIZE;

	ti->buf_pages = 1;
	while (ti->buf_pages * page_size < w->buffer_size)
		ti->buf_pages <<= 1;
	w->buffer_size = ti->buf_pages * page_size;
	++ti->buf_pages;

	ti->event_poll = CAFE_ALLOC(inst->allocator,
				    (cafe_array_len(&linux_event_specs) + 1)
				    * sizeof(struct pollfd) * num_cpus);
	if (ti->event_poll == NULL) {
		CAFE_FREE(inst->allocator, ti);
		return CAFE_NOMEM;
	}

	ti->emap = cafe_hashtbl_new(INITIAL_HASH_SIZE * num_cpus,
				    inst->allocator,
				    hashtbl_int64_hash, hashtbl_int64_equals);
	if (ti->emap == NULL) {
		CAFE_FREE(inst->allocator, ti->event_poll);
		CAFE_FREE(inst->allocator, ti);
		return CAFE_NOMEM;
	}

	if (pipe(ti->reader_wake_pipe) == -1) {
		perror("Error creating reader thread signalling pipe");
		cafe_hashtbl_delete(ti->emap, NULL, NULL, NULL, NULL);
		CAFE_FREE(inst->allocator, ti->event_poll);
		CAFE_FREE(inst->allocator, ti);
		return CAFE_ERROR;
	}
	set_fd_nonblocking(ti->reader_wake_pipe[0]);

	ti->event_poll[0].fd = ti->reader_wake_pipe[0];
	ti->event_poll[0].events = POLLIN;

	rc = pthread_mutex_init(&ti->actions_lock, NULL);
	if (rc != 0) {
		fprintf(stderr, "Error creating action list mutex: %s\n",
			strerror(rc));
		close(ti->reader_wake_pipe[1]);
		close(ti->reader_wake_pipe[0]);
		cafe_hashtbl_delete(ti->emap, NULL, NULL, NULL, NULL);
		CAFE_FREE(inst->allocator, ti->event_poll);
		CAFE_FREE(inst->allocator, ti);
		return CAFE_ERROR;
	}

	ti->change_count = inst->session->output_event_action_change_count - 1;
	inst->client_data = ti;
	ti->reader_should_exit = false;

	for (i = 0; i < PERF_RECORD_SAMPLE; ++i) {
		ti->perf_event_idata[i].event_map = perf_event_action_map[i];
		ti->perf_event_idata[i].action_info = NULL;
	}

	ti->inst = inst;

	for (i = 0; i < num_cpus; ++i) {
		rc = perf_buffer_create(&ti->buffers[i],
					ti->buf_pages * page_size, i,
					ti->buffers[0].fd);
		if (rc != CAFE_OK) {
			fprintf(stderr,
				"Failed to create perf buffer: %s\n",
				cafe_strerror(rc));
			close(ti->reader_wake_pipe[1]);
			close(ti->reader_wake_pipe[0]);
			cafe_hashtbl_delete(ti->emap, NULL, NULL, NULL, NULL);
			CAFE_FREE(inst->allocator, ti->event_poll);
			CAFE_FREE(inst->allocator, ti);
			return rc;
		}
	}

	rc = pthread_create(&ti->reader_thread, NULL, parse_buffer_thread,
			    ti);
	if (rc != 0) {
		fprintf(stderr,
			"Error creating reader_thread: %s\n", strerror(rc));
		pthread_mutex_destroy(&ti->actions_lock);
		close(ti->reader_wake_pipe[1]);
		close(ti->reader_wake_pipe[0]);
		cafe_hashtbl_delete(ti->emap, NULL, NULL, NULL, NULL);
		CAFE_FREE(inst->allocator, ti->event_poll);
		CAFE_FREE(inst->allocator, ti);
		return CAFE_ERROR;
	}

	ti->process_memory_map = cafe_hashtbl_new(INITIAL_HASH_SIZE,
						  inst->allocator,
						  cafe_hashtbl_int_hash,
						  cafe_hashtbl_int_equals);
	if (ti->process_memory_map == NULL) {
		fprintf(stderr, "Error creating process_memory_map\n");
		return CAFE_NOMEM;
	}

	// memtable input types
	for (i = 0; i < CAFE_NELEMENTS(ti->memtbl_input_events); ++i) {
		struct cafe_value_type *types[2];

		if (memtable_input_specs[i].vt == NULL) {
			/* occurs when this target doesn't support a certain
			   input event */
			ti->memtbl_input_events[i].event = NULL;
			ti->memtbl_input_events[i].info = NULL;
			continue;
		}

		rc = cafe_type_parse(inst->session->ns,
				     memtable_input_specs[i].vt, &types[0]);
		if (rc != CAFE_OK) {
			fprintf(stderr, "Error parsing memtable input type "
				"(%s: %s): %s\n",
				memtable_input_specs[i].name,
				memtable_input_specs[i].vt, cafe_strerror(rc));
			return rc;
		}
		types[1] = inst->output_odata[ODATA_INDEX_CONTEXT];

		rc = cafe_cfg_get_event(inst->session, 2, types,
					&ti->memtbl_input_events[i].event);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_ep_inst_event_action_info_get(inst,
							ti->memtbl_input_events[i].event,
							&ti->memtbl_input_events[i].info);
		if (rc != CAFE_OK)
			return rc;
	}

	rc = cafe_type_parse(inst->session->ns, PERF_MMAP_REF,
			     &ti->perf_mmap_type);
	if (rc != CAFE_OK) {
		fprintf(stderr, "Error parsing mmap type: %s\n",
			cafe_strerror(rc));
		return rc;
	}
	rc = cafe_type_parse(inst->session->ns, PERF_COMM_REF,
			     &ti->perf_comm_type);
	if (rc != CAFE_OK) {
		fprintf(stderr, "Error parsing comm type: %s\n",
			cafe_strerror(rc));
		return rc;
	}
	rc = cafe_type_parse(inst->session->ns, PERF_EXIT_REF,
			     &ti->perf_exit_type);
	if (rc != CAFE_OK) {
		fprintf(stderr, "Error parsing exit type: %s\n",
			cafe_strerror(rc));
		return rc;
	}
	rc = cafe_type_parse(inst->session->ns, PERF_FORK_REF,
			     &ti->perf_fork_type);
	if (rc != CAFE_OK) {
		fprintf(stderr, "Error parsing fork type: %s\n",
			cafe_strerror(rc));
		return rc;
	}

	// memtable output event types
	rc = cafe_type_parse(inst->session->ns, perf_mem_heap_VT_REF,
			     &ti->memtbl_heap_type);
	if (rc != CAFE_OK) {
		fprintf(stderr, "Error parsing memtable heap type: %s\n",
			cafe_strerror(rc));
		return rc;
	}
	rc = cafe_type_parse(inst->session->ns, perf_mem_mmap_VT_REF,
			     &ti->memtbl_mmap_type);
	if (rc != CAFE_OK) {
		fprintf(stderr, "Error parsing memtable mmap type: %s\n",
			cafe_strerror(rc));
		return rc;
	}
	ti->memtbl_heap_info = NULL;
	ti->memtbl_mmap_info = NULL;
	ti->inst = inst;

	return rc;
}


static
void trace_destroy(struct cafe_ep_inst *inst)
{
	struct linux_trace_inst *client_data = inst->client_data;
	int rc;
	void *retval;
	unsigned int i;

	++inst->session->output_event_action_change_count;
	disable_events_for_inst(inst);

	/* first stop the reader thread */
	client_data->reader_should_exit = true;
	update_poll_fds_signal(client_data->reader_wake_pipe[1]);
	rc = pthread_join(client_data->reader_thread, &retval);
	if (rc != 0)
		fprintf(stderr, "Error joining buffer thread: %s\n",
			strerror(rc));
	if (MULTI_BUFFER == NULL)
		perf_buffer_destroy(&client_data->buffers[0],
				    client_data->buf_pages * page_size);
	else
		for (i = 0; i < num_cpus; ++i)
			perf_buffer_destroy(&client_data->buffers[i],
					    client_data->buf_pages * page_size);
	close(client_data->reader_wake_pipe[1]);
	close(client_data->reader_wake_pipe[0]);

	cafe_hashtbl_delete(client_data->emap, NULL, NULL, NULL, NULL);
	cafe_hashtbl_delete(client_data->process_memory_map, NULL, NULL,
			    memtable_row_delete, (void *)inst->allocator);

	rc = pthread_mutex_destroy(&client_data->actions_lock);
	if (rc != 0)
		fprintf(stderr, "Error destroying action list mutex: %s\n",
			strerror(rc));

	CAFE_FREE(inst->allocator, client_data->event_poll);
	CAFE_FREE(inst->allocator, inst->client_data);
}


/**
 * Invoke the given synthetic event's state-dumping callback routine.
 */
static
int dump_state_for_event(struct cafe_ep_inst *inst,
			 struct cafe_enabled_event *event)
{
	struct cafe_event_action_info *event_action_info;
	struct event_action_map *action_map;

	assert(inst != NULL);
	assert(event != NULL);

	action_map = event->impl_data;
	if (action_map->state_dump == NULL)
		return CAFE_OK;

	event_action_info =
		cafe_ep_inst_event_action_info_find(inst, event->event);

	if (event_action_info == NULL)
		return CAFE_OK;

	return action_map->state_dump(event_action_info, inst);
}


/**
 *  Populate the given timestamp odata structure with values representing the
 * current time, only if the given event requests a timestamp.  If no timestamp
 * was requested, return false.
 */
static
bool add_timestamp_odata(const struct cafe_event_action_info *actions,
			 struct cafe_timestamp *ts, bool lock)
{
	if (actions->event->num_types == 1)
		return false;

	if (actions->event->num_types > 2) {
		assert(!"Event does not support orthogonal data types other "
		       "than timestamp");
		return false;
	}

	if (actions->event->types[1] == actions->inst->output_odata[ODATA_INDEX_TIMESTAMP]) {
		struct linux_trace_inst *client_data =
			actions->inst->client_data;
		if (current_time.timestamp != 0) {
			/* locking here assures atomic read of the timestamp
			   value */
			if (lock)
				pthread_mutex_lock(&client_data->actions_lock);
			ts->timestamp = current_time.timestamp;
			if (lock)
				pthread_mutex_unlock(&client_data->actions_lock);
			return true;
		} else {
#if 0
			struct timespec t;
			unsigned long long temp;

			if (clock_gettime(CLOCK_MONOTONIC, &t) != 0)
				perror("Error getting current time");
			temp = t.tv_sec;
			temp *= NSECS_PER_SEC;
			temp += t.tv_nsec;
			ts->timestamp = temp;
#else
			/* this is a hack.  At the moment the perf EP will
			   enable events before emitting events it synthesizes
			   itself, e.g., timestamp_config and initial process
			   descriptors.  This results in clients receiving the
			   synthesized events before buffered samples (as
			   desired) but the later samples will have timestamps
			   indicating they came before the synthesized events.
			   When we change the EP to enable buffering of perf
			   events only after having been started, we can remove
			   this hack. */
			ts->timestamp = 0;
#endif
		}
		return true;
	} else {
		/* unhandled orthogonal data requested */
		assert(!"Event does not support orthogonal data types other "
		       "than timestamp");
		return false;
	}
}


/**
 * This function acts as an intercept for all calls to
 * cafe_dispatch_event() in this EP.  It checks to see if there's an
 * enablement for:
 *
 *   t[cafe.timestamp_config] t[cafe.timestamp]
 *
 * and if the event to be dispatched has t[cafe.timestamp] as
 * orthogonal data then it copies that value before dispatching the
 * config event.
 *
 * We do this to ensure that we get monotonically increasing
 * timestamps as this EP dispatches some events which are effectively
 * synthetic and those that come from the kernel (i.e., Perf).
 *
 * If we were to source the timestamp for the above enablement from
 * userspace using clock_gettime() then that time value can and will
 * be in the future compared to the events that Perf has already
 * generated and subsequently buffered.  The net result is that
 * downstream EPs receive events where time appears to go backwards.
 */
static
void dispatch_event(struct cafe_event_action_info *action_info,
		    void **event_data)
{
	struct linux_trace_inst *ti = action_info->inst->client_data;

	if (ti->timestamp_config_action_info != NULL &&
	    ti->timestamp_config_event_dispatch_pending) {
		if (ti->timestamp_config_action_info->event->num_types == 1) {
			struct cafe_timestamp_config timestamp_config;
			void *tmp_data[1];
			tmp_data[0] = &timestamp_config;
			timestamp_config.frequency = NSECS_PER_SEC;
			cafe_dispatch_event(&ti->timestamp_config_action_info->sink_actions, tmp_data);
			ti->timestamp_config_event_dispatch_pending = false;
		} else if (action_info->event->num_types > 1) {
			size_t i;
			struct cafe_value_type *timestamp_type = NULL;

			for (i = 1; i < action_info->event->num_types; i++) {
				if (action_info->event->types[i] == ti->timestamp_type) {
					timestamp_type = action_info->event->types[i];
					break;
				}
			}

			if (timestamp_type != NULL) {
				struct cafe_timestamp_config timestamp_config;
				void *tmp_data[2];

				timestamp_config.frequency = NSECS_PER_SEC;
				tmp_data[0] = &timestamp_config;
				tmp_data[1] = event_data[i];
				cafe_dispatch_event(&ti->timestamp_config_action_info->sink_actions, tmp_data);
				ti->timestamp_config_event_dispatch_pending = false;
			}
		}
	}

	cafe_dispatch_event(&action_info->sink_actions, event_data);
}


/**
 *
 */
static
void dispatch_memtable_update(struct cafe_event_action_info *action_info,
			      void *event_data, uint32_t pid)
{
	void *data[NUM_ODATA_ITEMS + 1];
	unsigned int i;
	struct perf_context context = {0, 0};

	if (action_info == NULL)
		return;

	data[0] = event_data;
	for (i = 1; i < action_info->event->num_types; ++i)
		if (action_info->event->types[i] == action_info->inst->output_odata[ODATA_INDEX_CONTEXT]) {
			context.pid = pid;
			context.tid = pid;
			data[i] = &context;
		}
	cafe_dispatch_event(&action_info->sink_actions, data);
}


/**
 *
 */
static
void perf_mmap_memtable_handler(struct cafe_ep_inst *inst, void *event_data[])
{
	struct linux_trace_inst *ti = inst->client_data;
	struct perf_mmap_event *m = event_data[0];
	struct perf_mem_mmap mmap;
	struct process_memory_record *proc_data;

	if (m->pid == 0)
		return;

	proc_data = cafe_hashtbl_lookup(ti->process_memory_map, &m->pid);
	if (proc_data == NULL) {
		MTD_LOG("%s: unable to find pid %d\n", __FUNCTION__, m->pid);
		return;
	}

	if (strcmp(m->filename, "[heap]") == 0) {
		struct perf_mem_heap heap;

		proc_data->heap_base = m->addr;
		proc_data->heap_top = m->addr + m->len;
		heap.size = m->len;
		MTD_LOG("%s: %d, heap->%llu\n",
			__FUNCTION__, m->pid, heap.size);
		dispatch_memtable_update(ti->memtbl_heap_info, &heap, m->pid);
		return;
	}

	proc_data->mapped_memory += m->len;
	mmap.size = proc_data->mapped_memory;
	dispatch_memtable_update(ti->memtbl_mmap_info, &mmap, m->pid);
}


/**
 *  Returns true iff the routine was able to determine the given process'
 * heap boundaries.
 */
static
bool sample_heap_region(unsigned int pid, uint64_t *base, uint64_t *top)
{
	struct memory_region region;
	char buf[PATH_MAX];
	FILE *f;

	snprintf(buf, sizeof(buf), "/proc/%u/maps", pid);
	f = fopen(buf, "r");
	if (f == NULL)
		return false;

	while (memory_region_get_next(f, &region))
		if (strcmp("[heap]", region.name) == 0) {
			*base = region.start;
			*top = region.end;
			fclose(f);
			return true;
		}

	fclose(f);
	return false;
}


/**
 *
 */
static
void perf_comm_memtable_handler(struct cafe_ep_inst *inst, void *event_data[])
{
	struct linux_trace_inst *ti = inst->client_data;
	struct perf_comm_event *c = event_data[0];
	struct perf_mem_heap heap;
	struct perf_mem_mmap mmap;
	struct process_memory_record *proc_data;

	// filter out events for kernel regions and threads
	if ((c->pid == 0) || (c->pid != c->tid))
		return;

	proc_data = cafe_hashtbl_lookup(ti->process_memory_map, &c->pid);
	if (proc_data == NULL) {
		proc_data = CAFE_ALLOC(inst->allocator, sizeof(*proc_data));
		if (proc_data == NULL)
			return;

		proc_data->pid = c->pid;
		if (cafe_hashtbl_insert(ti->process_memory_map,
					&proc_data->pid, proc_data))
		{
			MTD_LOG("Unable to hash new process mem record\n");
			return;
		}
	}

	if (!sample_heap_region(c->pid,
				&proc_data->heap_base, &proc_data->heap_top))
	{
		proc_data->heap_base = 0;
		proc_data->heap_top = 0;
	}
	proc_data->mapped_memory = 0;
	heap.size = proc_data->heap_top - proc_data->heap_base;
	mmap.size = 0;
#if defined(TRACE_MEMTBL_DEBUG)
	strncpy(proc_data->pname, c->comm, sizeof(proc_data->pname));
#endif
	MTD_LOG("%s: (%d, %s) 0x%"PRIx64" - 0x%"PRIx64"\theap->%llu\n",
		__FUNCTION__, c->pid, proc_data->pname, proc_data->heap_base, proc_data->heap_top, heap.size);
	dispatch_memtable_update(ti->memtbl_heap_info, &heap, c->pid);
	dispatch_memtable_update(ti->memtbl_mmap_info, &mmap, c->pid);
}


/**
 *
 */
static
void perf_exit_memtable_handler(struct cafe_ep_inst *inst, void *event_data[])
{
	struct linux_trace_inst *ti = inst->client_data;
	struct perf_exit_event *e = event_data[0];
	struct perf_mem_heap heap;
	struct perf_mem_mmap mmap;
	struct process_memory_record *proc_data;

	// filter out events for threads
	if (e->pid != e->tid)
		return;

	proc_data = cafe_hashtbl_remove(ti->process_memory_map, &e->pid,
					NULL, NULL);
	if (proc_data == NULL) {
		MTD_LOG("%s: unable to find pid %d (%d) (%u, %u)\n",
			__FUNCTION__,
			e->pid, e->tid, e->ppid, e->ptid);
		return;
	}

	CAFE_FREE(inst->allocator, proc_data);
	heap.size = 0;
	mmap.size = 0;
	MTD_LOG("%s: %d, heap->%llu\n", __FUNCTION__, e->pid, heap.size);
	dispatch_memtable_update(ti->memtbl_heap_info, &heap, e->pid);
	dispatch_memtable_update(ti->memtbl_mmap_info, &mmap, e->pid);
}


/**
 *
 */
static
void perf_fork_memtable_handler(struct cafe_ep_inst *inst, void *event_data[])
{
	struct linux_trace_inst *ti = inst->client_data;
	struct perf_fork_event *f = event_data[0];
	struct perf_mem_heap heap;
	struct perf_mem_mmap mmap;
	struct process_memory_record *proc_data;
	struct process_memory_record *pproc_data;

	// filter out events for threads
	if (f->pid != f->tid)
		return;

	pproc_data = cafe_hashtbl_lookup(ti->process_memory_map, &f->ppid);
	if (pproc_data == NULL) {
		MTD_LOG("%s: unable to find ppid %d\n", __FUNCTION__, f->ppid);
		return;
	}

	proc_data = CAFE_ALLOC(inst->allocator, sizeof(*proc_data));
	if (proc_data == NULL)
		return;

	proc_data->pid = f->pid;
	proc_data->heap_base = pproc_data->heap_base;
	proc_data->heap_top = pproc_data->heap_top;
	proc_data->mapped_memory = pproc_data->mapped_memory;
	if (cafe_hashtbl_insert(ti->process_memory_map,
				&proc_data->pid, proc_data))
	{
		MTD_LOG("Unable to hash forked process\n");
		CAFE_FREE(inst->allocator, proc_data);
		return;
	}

	heap.size = proc_data->heap_top - proc_data->heap_base;
	mmap.size = proc_data->mapped_memory;
#if defined(TRACE_MEMTBL_DEBUG)
	strncpy(proc_data->pname, pproc_data->pname, sizeof(proc_data->pname));
	proc_data->ppid = f->ppid;
#endif
	MTD_LOG("%s: %d, heap->%llu\n", __FUNCTION__, f->pid, heap.size);
	dispatch_memtable_update(ti->memtbl_heap_info, &heap, f->pid);
	dispatch_memtable_update(ti->memtbl_mmap_info, &mmap, f->pid);
}


/**
 *
 */
static
void syscall_brk_memtable_handler(struct cafe_event_action_info *info,
				  void *event_data[])
{
	struct linux_trace_inst *ti = info->inst->client_data;
	struct perf_mem_heap heap;
	struct {
		unsigned char common_flags;
		unsigned char common_preempt_count;
		int common_pid;
		int common_lock_depth;
		int nr;
		long ret;
	} *b = event_data[0];
	struct perf_context *c = event_data[1];
	struct process_memory_record *proc_data;

	proc_data = cafe_hashtbl_lookup(ti->process_memory_map, &c->pid);
	if (proc_data == NULL) {
		MTD_LOG("%s: unable to find pid %d (%d)\n", __FUNCTION__,
			c->pid, c->tid);
		return;
	}

	if (proc_data->heap_base == 0)
		proc_data->heap_base = b->ret;

	proc_data->heap_top = b->ret;
	heap.size = proc_data->heap_top - proc_data->heap_base;
	MTD_LOG("%s: %d, %s, heap->%llu\n", __FUNCTION__, c->pid,
		proc_data->pname, heap.size);
	dispatch_memtable_update(ti->memtbl_heap_info, &heap, c->pid);
}


/**
 *
 */
static
void syscall_mmap_memtable_handler(struct cafe_event_action_info *info,
				   void *event_data[])
{
	struct linux_trace_inst *ti = info->inst->client_data;
	struct perf_mem_mmap mmap;
	struct {
		unsigned char common_flags;
		unsigned char common_preempt_count;
		int common_pid;
		int common_lock_depth;
		int nr;
		unsigned long addr;
		unsigned long len;
		unsigned long prot;
		unsigned long flags;
		unsigned long fd;
		unsigned long off;
	} *m = event_data[0];
	struct perf_context *c = event_data[1];
	struct process_memory_record *proc_data;

	proc_data = cafe_hashtbl_lookup(ti->process_memory_map, &c->pid);
	if (proc_data == NULL) {
		MTD_LOG("%s: unable to find pid %d\n", __FUNCTION__, c->pid);
		return;
	}

	proc_data->mapped_memory += m->len;
	mmap.size = proc_data->mapped_memory;
	dispatch_memtable_update(ti->memtbl_mmap_info, &mmap, c->pid);
}


/**
 *
 */
static
void syscall_mmap_pgoff_memtable_handler(struct cafe_event_action_info *info,
					 void *event_data[])
{
	struct linux_trace_inst *ti = info->inst->client_data;
	struct perf_mem_mmap mmap;
	struct {
		unsigned char common_flags;
		unsigned char common_preempt_count;
		int common_pid;
		int common_lock_depth;
		int nr;
		unsigned long addr;
		unsigned long len;
		unsigned long prot;
		unsigned long flags;
		unsigned long fd;
		unsigned long pgoff;
	} *m = event_data[0];
	struct perf_context *c = event_data[1];
	struct process_memory_record *proc_data;

	proc_data = cafe_hashtbl_lookup(ti->process_memory_map, &c->pid);
	if (proc_data == NULL) {
		MTD_LOG("%s: unable to find pid %d\n", __FUNCTION__, c->pid);
		return;
	}

	proc_data->mapped_memory += m->len;
	mmap.size = proc_data->mapped_memory;
	dispatch_memtable_update(ti->memtbl_mmap_info, &mmap, c->pid);
}


/**
 *
 */
static
void syscall_munmap_memtable_handler(struct cafe_event_action_info *info,
				     void *event_data[])
{
	struct linux_trace_inst *ti = info->inst->client_data;
	struct perf_mem_mmap mmap;
	struct {
		unsigned char common_flags;
		unsigned char common_preempt_count;
		int common_pid;
		int common_lock_depth;
		int nr;
		unsigned long addr;
		size_t len;
	} *m = event_data[0];
	struct perf_context *c = event_data[1];
	struct process_memory_record *proc_data;

	proc_data = cafe_hashtbl_lookup(ti->process_memory_map, &c->pid);
	if (proc_data == NULL) {
		MTD_LOG("%s: unable to find pid %d\n", __FUNCTION__, c->pid);
		return;
	}

	proc_data->mapped_memory -= m->len;
	mmap.size = proc_data->mapped_memory;
	dispatch_memtable_update(ti->memtbl_mmap_info, &mmap, c->pid);
}


/**
 *  Emit mmap events describing the executable regions of the current system's
 * kernel.
 */
static
void emit_kernel_mmap_event(struct cafe_ep_inst *inst, struct cafe_event_action_info *actions)
{
	static const char *const KERNEL_OBJECT_NAME = "vmlinux";
	ksym_entry_t *ksyms;
	int n_ksyms = kernel_symbols_read(&ksyms);
	struct perf_mmap_event e;
	void *data[NUM_ODATA_ITEMS + 1] = {&e};
	unsigned int i;

	if (n_ksyms == -1) {
		fprintf(stderr, "Unable to read kernel symbols\n");
		return;
	}

	/* locate the first symbol belonging to a kernel module */
	i = n_ksyms - 1;
	while (ksyms[i].module[0] != 0)
		--i;

	e.pid = 0;
	e.tid = 0;
	e.addr = 0;
	e.pgoff = SYMRESOLVE_OFFSET_UNKNOWN;
	e.filename = (char *)KERNEL_OBJECT_NAME;
	if (i == n_ksyms - 1) {
		e.len = ksyms[i].address + 0x1000;
		dispatch_event(actions, data);
		goto done;
	}

	e.len = ksyms[i + 1].address;
	dispatch_event(actions, data);

	++i;
	e.filename = ksyms[i].module;
	e.addr = ksyms[i].address;
	for (++i; i < n_ksyms - 1; ++i) {
		if (strcmp(e.filename, ksyms[i].module) != 0) {
			e.len = ksyms[i].address - e.addr;
			dispatch_event(actions, data);
			e.filename = ksyms[i].module;
			e.addr = ksyms[i].address;
		}
	}
	e.len = ksyms[i - 1].address - e.addr + 0x1000;
	dispatch_event(actions, data);
done:
	free(ksyms);
}


/**
 *  Emit mmap events describing the executable regions of the given process.
 */
static
void emit_mmap_event(struct cafe_ep_inst *inst,
		     struct cafe_event_action_info *actions, uint32_t pid)
{
	struct memory_region region;
	char buf[PATH_MAX];
	struct perf_mmap_event e;
	struct cafe_timestamp ts;
	void *data[NUM_ODATA_ITEMS + 1] = {&e};
	FILE *f;

	snprintf(buf, sizeof(buf), "/proc/%u/maps", pid);
	f = fopen(buf, "r");
	if (f == NULL)
		return;

	if ((actions != NULL) && (add_timestamp_odata(actions, &ts, true)))
		data[1] = &ts;

	e.pid = pid;
	e.tid = pid;
	while (memory_region_get_next(f, &region)) {
		e.addr = region.start;
		e.len = region.end - region.start;
		e.pgoff = region.file_offset;
		e.filename = region.name;
		if (actions != NULL)
			dispatch_event(actions, data);
		perf_mmap_memtable_handler(inst, data);
	}
	fclose(f);
}


/**
 *  A utility routine which scans the directory given in the path parameter and
 * invokes the given callback routine for each subdirectory which has a purely
 * numeric name.
 */
static
int for_each_numeric_subdir(const char *path,
			    void (*callback)(unsigned long id, void *arg),
			    void *arg)
{
	DIR *proc_dir;
	struct dirent *result;
	struct dirent *dir_entry;
	int err;

	assert(path != NULL);
	dir_entry = malloc(offsetof(struct dirent, d_name)
			   + pathconf(path, _PC_NAME_MAX) + 1);
	if (dir_entry == NULL)
		return CAFE_NOMEM;

	proc_dir = opendir(path);
	if (proc_dir == NULL) {
		free(dir_entry);
		printf("Error scanning %s: %s\n", path, strerror(errno));
		return CAFE_ERROR;
	}

	err = readdir_r(proc_dir, dir_entry, &result);
	while ((err == 0) && (result != NULL)) {
		char *endptr;
		unsigned long id;

		errno = 0;
		id = strtoul(dir_entry->d_name, &endptr, 10);

		if ((errno == 0) && (*endptr == '\0'))
			callback(id, arg);

		err = readdir_r(proc_dir, dir_entry, &result);
	}

	if (err != 0) {
		closedir(proc_dir);
		free(dir_entry);
		printf("Error opening %s subdirectory entry: %s\n",
		       path, strerror(err));
		return CAFE_ERROR;
	}

	closedir(proc_dir);
	free(dir_entry);
	return CAFE_OK;
}


struct emit_thread_event_arg {
	struct cafe_ep_inst *inst;
	struct perf_comm_event comm_event;
	struct cafe_event_action_info *actions;
};


/**
 *  Emits a "comm"and event representing the thread given by the tid and
 * emit_thread_event_arg parameters.  This routine is meant to be a callback
 * passed to the for_each_numeric_subdir() function.
 */
static
void emit_thread_event(unsigned long tid, void *arg)
{
	void *data[NUM_ODATA_ITEMS + 1];
	struct emit_thread_event_arg *a = arg;
	struct cafe_timestamp ts;

	assert(arg != NULL);
	data[0] = &a->comm_event;
	a->comm_event.tid = tid;

	if (a->actions != NULL) {
		if (add_timestamp_odata(a->actions, &ts, true))
			data[1] = &ts;

		dispatch_event(a->actions, data);
	}
	perf_comm_memtable_handler(a->inst, data);
}


struct emit_process_event_arg {
	struct cafe_ep_inst *inst;
	struct cafe_event_action_info *action_info;
	struct cafe_event_action_info *mmap_actions;
};


/**
 * Emit perf "comm"and events representing all threads of the given pid.  This
 * routine is meant to be a callback passed to the for_each_numeric_subdir()
 * function, with a pointer to an emit_process_event_arg structure.  If this
 * argument sets a value for mmap_actions, mmap events will be emitted
 * describing this process's executable memory map.
 */
static
void emit_process_event(unsigned long pid, void *arg)
{
	char buf[MAX(PATH_MAX, BUFSIZ)];
	char command[16];
	struct emit_thread_event_arg a;
	struct emit_process_event_arg *proc_args = arg;
	FILE *f;

	assert(arg != NULL);
	a.inst = proc_args->inst;
	a.actions = proc_args->action_info;

	snprintf(buf, sizeof(buf), "/proc/%lu/status", pid);
	f = fopen(buf, "r");
	if (f == NULL)
		return;

	do {
		if (fgets_unlocked(buf, sizeof(buf), f) == NULL) {
			fclose(f);
			return;
		}

		if (sscanf(buf, "Name: %s", command) == 1) {
			a.comm_event.comm = command;
			break;
		}
	} while(1);
	fclose(f);

	snprintf(buf, sizeof(buf), "/proc/%lu/exe", pid);
	if (readlink(buf, buf, sizeof(buf)) == -1) {
		/* this is actually a kernel thread, handle it as such */
		a.comm_event.pid = 0;
		emit_thread_event(pid, &a);
		return;
	}
	a.comm_event.pid = pid;

	snprintf(buf, sizeof(buf), "/proc/%lu/task", pid);
	for_each_numeric_subdir(buf, emit_thread_event, &a);

	emit_mmap_event(proc_args->inst, proc_args->mmap_actions, pid);
}


static
int trace_start(struct cafe_ep_inst *inst)
{
	struct linux_trace_inst *client_data = inst->client_data;
	struct emit_process_event_arg proc_args;
	unsigned int i;
	int rc = CAFE_OK;

	assert(inst != NULL);

	/* Change state before dispatching events (WB4-4006) */
	pthread_mutex_lock(&client_data->actions_lock);
	cafe_ep_inst_state_started(inst);

	memset(&proc_args, 0, sizeof(proc_args));
	proc_args.inst = inst;

	for (i = 0; i < inst->enabled_output_events.ind; ++i) {
		struct cafe_enabled_event *evt;
		struct event_action_map *action_map;

		evt = &inst->enabled_output_events.list[i];
		action_map = evt->impl_data;
		if (action_map == perf_event_action_map[PERF_RECORD_COMM]) {
			proc_args.action_info = cafe_ep_inst_event_action_info_find(inst, evt->event);
			continue;
		}

		if (action_map == perf_event_action_map[PERF_RECORD_MMAP]) {
			proc_args.mmap_actions = cafe_ep_inst_event_action_info_find(inst, evt->event);
			continue;
		}

		if (action_map->event_type != (enum perf_type_id) SYNTHETIC_EVENT)
			continue;

		rc = dump_state_for_event(inst, evt);
		if (rc != CAFE_OK)
			goto out;
	}

	if (proc_args.mmap_actions != NULL)
		emit_kernel_mmap_event(inst, proc_args.mmap_actions);
	/* emit events describing existing processes */
	for_each_numeric_subdir("/proc", emit_process_event, &proc_args);

out:
	/* Let the buffer reader thread continue */
	pthread_mutex_unlock(&client_data->actions_lock);
	return rc;
}


/**
 *  Ensures the "stopped" state is asserted only after we're sure no further
 * events will be dispatched via the given EP instance.
 */
static
int trace_stop(struct cafe_ep_inst *inst)
{
	struct linux_trace_inst *client_data = inst->client_data;
	pthread_mutex_lock(&client_data->actions_lock);
	cafe_ep_inst_state_stopped(inst);
	pthread_mutex_unlock(&client_data->actions_lock);
	return CAFE_OK;
}


/**
 *  Dispatch the cafe.timestamp_config event with the timestamp frequency to
 * any registered sinks.
 */
static
int timestamp_config_state_dump_handler(struct cafe_event_action_info *h,
					struct cafe_ep_inst *inst)
{
	struct linux_trace_inst *ti = inst->client_data;

	/*
	 * This will actually get dispatched when a valid timestamp
	 * becomes available.  See dispatch_event().
	 */
	if (ti->timestamp_config_action_info != NULL)
		ti->timestamp_config_event_dispatch_pending = true;

	return CAFE_OK;
}


/**
 *  Garner the offset of the last byte of readable data from perf's ring buffer
 * as an offset from the buffer's base address.
 */
static
unsigned long perfbuf_head_get(const struct perf_event_mmap_page *buffer)
{
	unsigned long head = buffer->data_head;
#if ( defined(__i386__) || defined(__i586__) || defined(__i686__) || defined(__x86_64__) ) && defined(__SSE2__)
	asm volatile("mfence":::"memory");
#else
#warning No memory barrier defined for this architecture.
#endif
	return head;
}


/**
 *  Set the offset of the last byte of consumed data from perf's ring buffer
 * as an offset from the buffer's base address.
 */
static
void perfbuf_tail_write(struct perf_event_mmap_page *buffer, unsigned long tail)
{
#if ( defined(__i386__) || defined(__i586__) || defined(__i686__) || defined(__x86_64__) ) && defined(__SSE2__)
	asm volatile("mfence":::"memory");
#else
#warning No memory barrier defined for this architecture.
#endif
	buffer->data_tail = tail;
}


/**
 *  Given a byte-stream of raw perf sample data, generate pointers to
 * individual pieces of orthogonal data as specified by the given action_info.
 * The callstack is the only piece of data that we can't reference in-place;
 * it needs to be expanded out into a cafe_vm_dynarray structure provided by
 * the caller.
 */
static
void populate_odata(void *data[NUM_ODATA_ITEMS + 1],
		    struct cafe_vm_dynarray *stack,
		    struct cafe_skipcount *skipcount_odata,
		    struct instance_data *idata,
		    char *raw_data_base)
{
	const struct cafe_event_descr *event = idata->action_info->event;
	struct cafe_value_type **odata = idata->action_info->inst->output_odata;
	unsigned int i;

//	unsigned int offset = sample_type_offsets[PERF_SAMPLE_READ_OFFSET]
//		+ ODATA_SIZES[PERF_SAMPLE_READ_OFFSET];

	for (i = 1; i < event->num_types; ++i) {
		if (event->types[i] == odata[ODATA_INDEX_IP])
			data[i] = raw_data_base + sample_type_offsets[PERF_SAMPLE_IP_OFFSET];
		else if (event->types[i] == odata[ODATA_INDEX_CONTEXT])
			data[i] = raw_data_base + sample_type_offsets[PERF_SAMPLE_CONTEXT_OFFSET];
		else if (event->types[i] == odata[ODATA_INDEX_TIMESTAMP]) {
			uint64_t *ts = (uint64_t *) (raw_data_base + sample_type_offsets[PERF_SAMPLE_TIME_OFFSET]);
			data[i] = ts;
			current_time.timestamp = *ts;
		} else if (event->types[i] == odata[ODATA_INDEX_ADDR])
			data[i] = raw_data_base + sample_type_offsets[PERF_SAMPLE_ADDR_OFFSET];
		else if (event->types[i] == odata[ODATA_INDEX_ID])
			data[i] = raw_data_base + sample_type_offsets[PERF_SAMPLE_ID_OFFSET];
		else if (event->types[i] == odata[ODATA_INDEX_STREAM_ID])
			data[i] = raw_data_base + sample_type_offsets[PERF_SAMPLE_STREAM_ID_OFFSET];
		else if (event->types[i] == odata[ODATA_INDEX_CPU])
			data[i] = raw_data_base + sample_type_offsets[PERF_SAMPLE_CPU_OFFSET];
		else if (event->types[i] == odata[ODATA_INDEX_PERIOD])
			data[i] = raw_data_base + sample_type_offsets[PERF_SAMPLE_PERIOD_OFFSET];
		else if (event->types[i] == odata[ODATA_INDEX_STACK]) {
			uint64_t *raw_stack = (uint64_t *) (raw_data_base + sample_type_offsets[PERF_SAMPLE_CALLCHAIN_OFFSET]);
			assert(idata->sample_type & PERF_SAMPLE_CALLCHAIN);
			data[i] = stack;
			stack->l = stack->c = *raw_stack;
			raw_stack++;
			stack->p = raw_stack;
		} else if ((cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SKIPCOUNT]) ||
			   (cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SAMPLE_FREQ])) {
			uint64_t *cpu;
			cpu = (uint64_t *) (raw_data_base + sample_type_offsets[PERF_SAMPLE_CPU_OFFSET]);
			const uint64_t *count = (uint64_t *)(raw_data_base + sample_type_offsets[PERF_SAMPLE_READ_OFFSET]);
			uint64_t delta;
			if ((*cpu >= 0) && (*cpu < num_cpus)) {
				delta = *count - idata->cpu[*cpu].last_event_count;
				idata->cpu[*cpu].last_event_count = *count;
				skipcount_odata->value = delta;
			} else {
				fprintf(stderr, "%s: Bad CPU number in %"PRIu64" %u\n", __FUNCTION__, *cpu, num_cpus);
				skipcount_odata->value = 0;
			}
			data[i] = skipcount_odata;
		}
	}
}


/**
 *  Given a pointer to the beginning of a perf sample event, (i.e., a pointer to
 * its perf_event_header) return the event_action_map corresponding to the
 * specific event.  Returns NULL on error.
 */
static
struct instance_data *parse_sample_event(const char **trav,
					 struct cafe_hashtbl *emap)
{
	uint64_t stream_event_id;
	struct event_cpu_handle *cpu;
	struct instance_data *id;

	stream_event_id = *(uint64_t *) (*trav + sample_type_offsets[PERF_SAMPLE_STREAM_ID_OFFSET]);
	cpu = cafe_hashtbl_lookup(emap, &stream_event_id);
	if (cpu == NULL) {
		fprintf(stderr, "skipping unregistered sample event %"PRIu64"\n",
			stream_event_id);
		return NULL;
	}
	id = cpu->event_instance;

	/* parse past the orthogonal sample data types */
	if (id->sample_type & PERF_SAMPLE_CALLCHAIN) {
		uint64_t stack_depth;
		*trav += sample_type_offsets[PERF_SAMPLE_CALLCHAIN_OFFSET];
		stack_depth = *(uint64_t *) *trav;
		*trav += sizeof(uint64_t);
		*trav += stack_depth * sizeof(uint64_t);
	} else
		*trav += sample_type_offsets[PERF_SAMPLE_READ_OFFSET]
			+ ODATA_SIZES[PERF_SAMPLE_READ_OFFSET];

	/* raw_data size */
	*trav += sizeof(uint32_t);

	/* advance past the event id parameter */
	*trav += sizeof(uint16_t);

	return id;
}


/**
 *  Given the perf_event_header of a non-sample perf event, return the
 * event_action_map corresponding to the specific event.  Returns NULL on error.
 */
static
struct instance_data *parse_perf_event(const char **trav,
				       struct instance_data perf_events[])
{
	const struct perf_event_header *event_header =
		(struct perf_event_header *) *trav;

	if ((event_header->type == 0) ||
	    (event_header->type >= PERF_RECORD_SAMPLE) ||
		event_header->size < alignof(struct perf_event_header)) {
		fprintf(stderr, "unknown PERF event 0x%x\n", event_header->type);
		return NULL;
	}

	//printf("PERF event %u @ %p\n", event_header->type, *trav);
	*trav += sizeof(struct perf_event_header);
	return &perf_events[event_header->type];
}


/**
 *  Returns true if an event_action was found for this event, and the event
 * data was successfully formatted into the given buffer.  Must be called with
 * the actions_lock held.
 */
static
bool format_event_data(const struct instance_data *idata,
		       const char *raw_data, char *event_data)
{
	if (idata->action_info == NULL)
		return false;

	relocate_strings(idata->action_info->event->types[0], raw_data,
			 event_data);
	return true;
}


/**
 *  Given a pointer to a buffered perf event record, decode and dispatch it.
 */
static
bool parse_event(const char *trav,
		 void *output_event_data[NUM_ODATA_ITEMS + 1],
		 char event_data[512], struct linux_trace_inst *client_data)
{
	struct perf_event_header *event_header =
		(struct perf_event_header *) trav;
	struct instance_data *idata = NULL;

	pthread_mutex_lock(&client_data->actions_lock);
	if (client_data->inst->state != CAFE_EP_INST_STATE_STARTED)
		goto out;

	if (event_header->type == PERF_RECORD_SAMPLE)
		idata = parse_sample_event(&trav, client_data->emap);
	else {
		idata = parse_perf_event(&trav, client_data->perf_event_idata);
		if ((event_header->type == PERF_RECORD_EXIT) &&
		    (relocate_strings(client_data->perf_exit_type,
				      trav, event_data)))
		{
			perf_exit_memtable_handler(client_data->inst,
						   output_event_data);
		}
	}

	if ((idata != NULL) && (format_event_data(idata, trav, event_data))) {
		struct cafe_timestamp ts;
		struct cafe_skipcount skipcount_odata;
		struct cafe_vm_dynarray stack;

		if (event_header->type == PERF_RECORD_SAMPLE) {
			populate_odata(output_event_data, &stack, &skipcount_odata,
				       idata, (char *)event_header);
		} else {
			if (add_timestamp_odata(idata->action_info, &ts, false))
				output_event_data[1] = &ts;
		}
		dispatch_event(idata->action_info, output_event_data);
	}

	// dispatch internal input events for building memtable rows
	if (event_header->type == PERF_RECORD_COMM) {
		relocate_strings(client_data->perf_comm_type, trav, event_data);
		perf_comm_memtable_handler(client_data->inst,
					   output_event_data);
	} else if (event_header->type == PERF_RECORD_FORK) {
		relocate_strings(client_data->perf_fork_type,
				 trav, event_data);
		perf_fork_memtable_handler(client_data->inst,
					   output_event_data);
	} else if (event_header->type == PERF_RECORD_MMAP) {
		relocate_strings(client_data->perf_mmap_type, trav, event_data);
		perf_mmap_memtable_handler(client_data->inst,
					   output_event_data);
	}
out:
	pthread_mutex_unlock(&client_data->actions_lock);

	return idata != NULL;
}


/**
 *  Read as many perf events as possible from a single continguous bytestream
 * and dispatch them as cafe events.
 */
static
void buffer_read_contiguous(unsigned long size, unsigned long mask,
			    struct event_buffer *buffer,
			    struct linux_trace_inst *client_data)
{
	struct perf_event_header *event_header;
	void *output_event_data[NUM_ODATA_ITEMS + 1];
	char event_data[512];
	const char *data = buffer->data + (buffer->last_read & mask);
	const char *trav = data;

#if defined(RAW_DUMP_DEBUG)
	if (fwrite_unlocked(data, size, 1, out) != 1)
		perror("failed to dump data");
	fflush_unlocked(out);
#endif

	output_event_data[0] = event_data;

	/* consume any wrapped data */
	if (buffer->wrap_buf.len != 0) {
		unsigned long chunk_size;
		bool parsed_event;

		if (buffer->wrap_buf.len < sizeof(struct perf_event_header)) {
			chunk_size = sizeof(struct perf_event_header) - buffer->wrap_buf.len;
			cafe_bbuf_append(&buffer->wrap_buf, trav, chunk_size);
			trav += chunk_size;
		}

		event_header = (struct perf_event_header *)buffer->wrap_buf.data;
		chunk_size = event_header->size - buffer->wrap_buf.len;
		cafe_bbuf_append(&buffer->wrap_buf, trav, chunk_size);
		trav += chunk_size;

		parsed_event = parse_event(buffer->wrap_buf.data,
					   output_event_data,
					   event_data, client_data);
		buffer->wrap_buf.len = 0;
		if (!parsed_event)
			return;
	}

	/* now read exclusively from the incoming buffer */
	while(trav < data + size) {
		const char *next_event;

		if (trav + sizeof(struct perf_event_header) > data + size) {
			cafe_bbuf_append(&buffer->wrap_buf, trav,
					 data + size - trav);
			break;
		}

		event_header = (struct perf_event_header *)trav;
		next_event = trav + event_header->size;

		if (next_event > data + size) {
			cafe_bbuf_append(&buffer->wrap_buf, trav,
					 data + size - trav);
			break;
		}

		if (!parse_event(trav, output_event_data, event_data, client_data))
			break;
		trav = next_event;
	}
}


/**
 *  Drain all available data from a given buffer.
 */
static
void buffer_read(struct event_buffer *buffer, unsigned long mask,
		 struct linux_trace_inst *client_data)
 {
	 unsigned long data_head = perfbuf_head_get(buffer->buffer);
	 unsigned long size = data_head - buffer->last_read;

	 //printf("%s: %u (0x%lx 0x%lx 0x%lx) (%lx)\n", __FUNCTION__, __LINE__,
	 //buffer->last_read, data_head, size, mask);
	 if ((buffer->last_read & mask) + size > (mask + 1)) {
		 /* if the write head has wrapped, consume the leftover
		    consecutive bytes before wrapping the read head */
		 size = mask + 1 - (buffer->last_read & mask);
		 buffer_read_contiguous(size, mask, buffer, client_data);
		 buffer->last_read += size;
		 //perfbuf_tail_write(buffer->last_read);
		 size = data_head - buffer->last_read;
		 //printf("%s: %u (0x%lx 0x%lx 0x%lx)\n", __FUNCTION__, __LINE__,
		 //buffer->last_read, data_head, size);
	 }

	 assert(size <= ((client_data->buf_pages - 1) * page_size));
	 buffer_read_contiguous(size, mask, buffer, client_data);
	 buffer->last_read += size;
	 perfbuf_tail_write(buffer->buffer, buffer->last_read);
 }


/**
 *  The entry point for a thread meant to read the kernel's event ring buffer
 * and distribute cafe events to client EPs.
 */
static
void *parse_buffer_thread(void *arg)
{
	int rc;
	unsigned int i;
	unsigned int num_poll_events = 1;
	struct linux_trace_inst *client_data = arg;
	const unsigned int mask = ((client_data->buf_pages - 1) * page_size) - 1;

#if defined(HAVE_PR_SET_NAME)
	if (prctl(PR_SET_NAME, "perfbuf_read") != 0)
		perror("Error setting thread name");
#endif

	if (MULTI_BUFFER == NULL) {
		rc = cafe_bbuf_init(&client_data->buffers[0].wrap_buf,
				    cafe_stdlib_allocator, 1);
		if (rc != CAFE_OK) {
			fprintf(stderr,
				"Error initializing wrap buffer: %s\n",
				cafe_strerror(rc));
			return NULL;
		}
	} else
		for (i = 0; i < num_cpus; ++i) {
			rc = cafe_bbuf_init(&client_data->buffers[i].wrap_buf,
					    cafe_stdlib_allocator, 1);
			if (rc != CAFE_OK) {
				fprintf(stderr,
					"Error initializing wrap buffer %u: %s\n",
					i, cafe_strerror(rc));
				return NULL;
			}
		}

#if defined(RAW_DUMP_DEBUG)
	out = fopen("trace_out.raw", "w");
#endif

	while (!client_data->reader_should_exit) {
		if (poll(client_data->event_poll, num_poll_events, 100) == -1) {
			fprintf(stderr, "Error poll()ing for %u events: %s",
				num_poll_events, strerror(errno));
			break;
		}

		/* drain any wakeup signals */
		if (client_data->event_poll[0].revents & POLLIN) {
			char c;
			while (read(client_data->reader_wake_pipe[0], &c,
				    sizeof(c)) > 0)
				;  //printf("read 1 signal byte\n");
			num_poll_events = update_poll_fds(client_data);
		}

		if (client_data->buffers[0].data == MAP_FAILED)
			continue;

		if (MULTI_BUFFER == NULL)
			buffer_read(&client_data->buffers[0], mask,
				    client_data);
		else
			for (i = 0; i < num_cpus; ++i)
				buffer_read(&client_data->buffers[i], mask,
					    client_data);
	}

	if (MULTI_BUFFER == NULL)
		cafe_bbuf_release(&client_data->buffers[0].wrap_buf);
	else
		for (i = 0; i < num_cpus; ++i)
			cafe_bbuf_release(&client_data->buffers[i].wrap_buf);
#if defined(RAW_DUMP_DEBUG)
	fclose(out);
#endif
	return NULL;
}


/**
 *  Given a trace event id, e.g., .../debug/tracing/events/<class>/<event>/id,
 * activate kernel output of this event and tie it into the existing mmap()ed
 * ring buffer.
 */
static
int activate_trace_event(struct instance_data *id,
			 struct linux_trace_inst *ti)
{
	unsigned int cpu;
	const pid_t pid = -1;
	const unsigned long flags = 0;
	struct perf_event_attr attr;
	int fd;
	struct {
		uint64_t count;
		uint64_t id;
	} fd_data;

	if (id->event_map->event_type >= PERF_TYPE_MAX)
		/* no need to activate non-sample events */
		return CAFE_OK;

	memset(&attr, 0, sizeof(attr));
	attr.type = id->event_map->event_type;
	attr.sample_type = SAMPLE_TYPE | id->sample_type;
	if (id->skipcount_freq == 0)
		attr.sample_period = id->skipcount_limit;
	else {
		attr.freq = 1;
		attr.sample_freq = id->skipcount_freq;
	}
	if (attr.type == PERF_TYPE_TRACEPOINT) {
		attr.config = id->event_map->trace_event->id;
		attr.sample_type |= PERF_SAMPLE_RAW;
	} else if (attr.type == PERF_TYPE_HW_CACHE)
		attr.config = id->event_map->hw_cache_event;
	else if (attr.type == PERF_TYPE_SOFTWARE)
		attr.config = id->event_map->software_event;
	else {
		assert(attr.type == PERF_TYPE_HARDWARE);
		attr.config = id->event_map->hardware_event;
	}

	/* Until we have a good abstraction for expressing desired frequency,
	   hardcode the cpu_clock event to fire at 100 Hz. */
	if ((attr.type == PERF_TYPE_SOFTWARE)
	    && (attr.config == PERF_COUNT_SW_CPU_CLOCK)) {
		attr.sample_period = 10000000;
	}
	if ((attr.type == PERF_TYPE_SOFTWARE)
	    && (attr.config == PERF_COUNT_SW_TASK_CLOCK)) {
		attr.exclude_idle = 1;
		attr.sample_period = 10000000;
	}

	attr.size = sizeof(attr);
	attr.read_format = PERF_FORMAT_ID;
	attr.inherit = 1;
	//attr.disabled = 1;

	//printf("** %u %llu %llu\n", attr.type, attr.config, attr.sample_type);
	for (cpu = 0; cpu < num_cpus; ++cpu) {
		assert(id->cpu[cpu].hashed == false);

		fd = syscall(__NR_perf_event_open, &attr, pid, cpu, -1, flags);
		if (fd < 0) {
			printf("Error opening event file descriptor %llu: %s (%d)\n",
			       attr.config, strerror(errno), errno);
			id->cpu[cpu].fd = fd;
			return CAFE_ERROR;
		}

		if (read(fd, &fd_data, sizeof(fd_data)) == -1) {
			perror("Error reading event file descriptor");
			close(fd);
			return CAFE_ERROR;
		}
		id->cpu[cpu].stream_id = fd_data.id;
		//printf("%u : %"PRIu64" %"PRIu64"\n", cpu, fd_data.id, fd_data.count);
		if (MULTI_BUFFER == NULL) {
			if (ioctl(fd, PERF_EVENT_IOC_SET_OUTPUT, ti->buffers[0].fd) == -1)
				goto fail_ioctl;
		} else if (ioctl(fd, PERF_EVENT_IOC_SET_OUTPUT, ti->buffers[cpu].fd) == -1) {
		fail_ioctl:
			perror("failed to redirect output");
			close(fd);
			return CAFE_ERROR;
		}
		id->cpu[cpu].fd = fd;
		id->cpu[cpu].event_instance = id;

		if (cafe_hashtbl_insert(ti->emap, &id->cpu[cpu].stream_id,
					&id->cpu[cpu]))
		{
			close(fd);
			fprintf(stderr,
				"Unable to hash newly activated event\n");
			return CAFE_NOMEM;
		}
		id->cpu[cpu].hashed = true;
	}
	return CAFE_OK;
}


/**
 *  Deactivate the underlying perf trace event corresponding to the given
 * event_action_map structure.
 */
static
void deactivate_trace_event(struct instance_data *id)
{
	unsigned int i;

	id->buffer_marker = 0;
	for (i = 0; i < num_cpus; ++i) {
		if (id->cpu[i].fd <= 0)
			continue;

		if (close(id->cpu[i].fd) != 0)
			perror("Error closing tracepoint file descriptor");

		id->cpu[i].fd = -1;
	}
}


/**
 *  Deactivate the underlying perf trace event corresponding to the given
 * event_action_map structure.
 */
static
void delete_trace_event(const struct cafe_ep_inst *inst,
			struct instance_data *id,
			struct cafe_hashtbl *emap)
{
	unsigned int i;

	deactivate_trace_event(id);
	for (i = 0; i < num_cpus; ++i) {
		if (!id->cpu[i].hashed)
			continue;
		if (cafe_hashtbl_remove(emap, &id->cpu[i].stream_id,
					NULL, NULL) == NULL)
			printf("Unable to remove mapping for stream id"
			       ": %" PRIu64 "\n", id->cpu[i].stream_id);
	}
	CAFE_FREE(inst->allocator, id);
}


/**
 *  Returns true iff the given event contains the given orthogonal data type.
 */
static
bool wants_odata(const struct cafe_event_descr *event,
		 const struct cafe_value_type *odata_type)
{
	unsigned int i;

	for (i = 1; i < event->num_types; ++i) {
		const struct cafe_value_type *type = event->types[i];
		if (cafe_type_based_on(type, odata_type))
			return true;
	}

	return false;
}


/**
 *  Disable all event generation for all events no longer enabled on this EP
 * instance.  An event for a given EP instance is no longer enabled when its
 * per-instance change count does not equal the session's change count.  This
 * routine iterates over all possible events and removes pointers to action
 * lists of those disabled events.  If an event is found to have no more active
 * clients, its descriptor is closed, disabling the underlying data generation.
 */
static
void disable_events_for_inst(const struct cafe_ep_inst *inst)
{
	struct linux_trace_inst *client_data = inst->client_data;
	struct cafe_hashtbl_iter iter;

	pthread_mutex_lock(&client_data->actions_lock);
	cafe_hashtbl_iter_init(client_data->emap, &iter);

	while (cafe_hashtbl_iter_next(client_data->emap, &iter) != 0) {
		struct event_cpu_handle *cpu = iter.val;
		struct instance_data *idata = cpu->event_instance;

		if (idata->change_count != inst->session->output_event_action_change_count) {
			delete_trace_event(inst, idata, client_data->emap);
			cafe_hashtbl_iter_init(client_data->emap, &iter);
		}
	}

	pthread_mutex_unlock(&client_data->actions_lock);
	update_poll_fds_signal(client_data->reader_wake_pipe[1]);
}


/**
 *  Ensure the given event_action_info object is contained in our global
 * event->action list mappings.  If the action_info object is a new one,
 * add it to our map and activate the associated linux trace event if
 * necessary.
 */
static
int update_ep_action_map(struct event_action_map *map,
			 struct cafe_event_action_info *event_action_info,
			 uint64_t requested_skipcount,
			 uint64_t requested_freq,
			 uint64_t requested_odata,
			 unsigned long change_generation)
{
	int rc;
	struct instance_data *id;
	struct linux_trace_inst *client_data =
		event_action_info->inst->client_data;
	struct cafe_hashtbl_iter iter;

	assert(map);
	assert(event_action_info);

	if (map->event_type == (enum perf_type_id) SYNTHETIC_EVENT)
		return CAFE_OK;

	pthread_mutex_lock(&client_data->actions_lock);

	if (map->event_type == (enum perf_type_id) INTRINSIC_PERF_EVENT) {
		unsigned int i;
		for (i = 1; i < PERF_RECORD_SAMPLE; ++i) {
			if (map == client_data->perf_event_idata[i].event_map) {
				id = &client_data->perf_event_idata[i];
				id->action_info = event_action_info;
				id->change_count = change_generation;
				pthread_mutex_unlock(&client_data->actions_lock);
				return CAFE_OK;
			}
		}
	}

	cafe_hashtbl_iter_init(client_data->emap, &iter);
	while (cafe_hashtbl_iter_next(client_data->emap, &iter) != 0) {
		struct event_cpu_handle *cpu = iter.val;
		id = cpu->event_instance;
		if (id->event_map == map) {
			if (id->action_info == event_action_info) {
				/* Unmodified event */
				id->change_count = change_generation;
				pthread_mutex_unlock(&client_data->actions_lock);
				return CAFE_OK;
			}
			/* WRS_TODO: This will cause loss of pending
			 * events associated with this id. Instead of
			 * deleting the trace event immediately we
			 * should deactivate it here and delete it
			 * when we are sure the buffer does not
			 * contain any pending events of this type. */
			delete_trace_event(event_action_info->inst, id,
					   client_data->emap);
			break;
		}
	}
	id = CAFE_ALLOC(event_action_info->inst->allocator,
			sizeof(*id) + num_cpus * sizeof(id->cpu[0]));
	if (id == NULL) {
		pthread_mutex_unlock(&client_data->actions_lock);
		return CAFE_NOMEM;
	}

	id->action_info = event_action_info;
	id->change_count = change_generation;
	id->event_map = map;
	id->buffer_marker = 0;
	id->sample_type = requested_odata;
	id->skipcount_limit = requested_skipcount;
	id->skipcount_freq = requested_freq;
	memset(&id->cpu, 0, num_cpus * sizeof(id->cpu[0]));

	rc = activate_trace_event(id, event_action_info->inst->client_data);
	if (rc != CAFE_OK) {
		delete_trace_event(event_action_info->inst, id,
				   client_data->emap);
		pthread_mutex_unlock(&client_data->actions_lock);
		return rc;
	}
	pthread_mutex_unlock(&client_data->actions_lock);
	return rc;
}


/**
 *  Signals the buffer-parsing thread to wake up and update the event
 * descriptors it poll()s on.  Should be called any time after a batch of
 * events are enabled or disabled.
 */
static
void update_poll_fds_signal(int signal_pipe)
{
	unsigned int i = 0;

	if (-1 == write(signal_pipe, &i, 1)) {
		perror("Error waking reader thread, not updating "
		       "file descriptor list");
		return;
	}
	sched_yield();
}


/**
 *  Rewrites the event_poll list with all the current file descriptors we
 * need to watch for incoming events.  Should only be called from the thread
 * which actually poll()s on the descriptor list (i.e., the buffer reading
 * thread).  The event_poll array must be pre-sized large enough to accomodate
 * the maximum possible event descriptors.
 */
static
unsigned int update_poll_fds(struct linux_trace_inst *ti)
{
	unsigned int num_poll_events = 1;
	struct cafe_hashtbl_iter i;

	cafe_hashtbl_iter_init(ti->emap, &i);
	pthread_mutex_lock(&ti->actions_lock);
	while (cafe_hashtbl_iter_next(ti->emap, &i) != 0) {
		const struct event_cpu_handle *cpu = i.val;

		assert(cpu->fd > 0);

		ti->event_poll[num_poll_events].fd = cpu->fd;
		ti->event_poll[num_poll_events].events = POLLIN;
		++num_poll_events;
	}
	pthread_mutex_unlock(&ti->actions_lock);

	return num_poll_events;
}


/**
 *  Get the limit value from the skipcount parameter on the given event
 * description.  If the event does not have a skipcount parameter orthogonal
 * event, it returns false.  Otherwise it returns true and the limit value.
 */
static
bool skipcount_get_limit(const struct cafe_event_descr *event,
			 const struct cafe_value_type *skipcount_type,
			 unsigned long long *limit)
{
	unsigned int i;

	for (i = 0; i < event->num_types; ++i) {
		const struct cafe_value_type *type = event->types[i];
		if (cafe_type_based_on(type, skipcount_type)) {
			const struct cafe_vm_expr *expr =
				cafe_param_value_get(type, "limit");
			*limit = expr->u.ull;
			return true;
		}
	}

	return false;
}


/**
 * Get the sample_freq from the events given event description. If the event
 * does not have a sample_freq orthogonal event, it returns false.  Otherwise
 * it returns true and the requested value.
 */
static
bool sample_freq_value_get(const struct cafe_event_descr *event,
			   const struct cafe_value_type *sample_freq_type,
			   unsigned long long *freq)
{
	unsigned int i;

	for (i = 0; i < event->num_types; ++i) {
		const struct cafe_value_type *type = event->types[i];
		if (cafe_type_based_on(type, sample_freq_type)) {
			const struct cafe_vm_expr *expr =
				cafe_param_value_get(type, "rate");
			*freq = expr->u.ull;
			return true;
		}
	}
	return false;
}


/**
 *
 */
static
int memtbl_input_event_activate(struct cafe_ep_inst *inst,
				struct memtable_inst_data *input_event,
				struct memtable_input_spec *input_spec)
{
	struct cafe_event_action_list *input_actions;
	cafe_eal_handle_t input_actions_handle;
	int rc;
	rc = cafe_event_action_list_alloc(inst, input_event->event,
					  &input_actions,
					  &input_actions_handle);
	if (rc != CAFE_OK)
		return rc;
	rc = cafe_add_event_action(inst, input_actions, input_spec->handler,
				   input_event->info);
	if (rc != CAFE_OK)
		return rc;
	cafe_event_action_list_swap(&input_event->info->sink_actions,
				    input_actions_handle);
	rc = update_ep_action_map(&input_spec->map, input_event->info, 1, 0, 0,
				  inst->session->output_event_action_change_count);
	return rc;
}


/**
 *
 */
static
int trace_update_output_event_actions(struct cafe_ep_inst *inst)
{
	unsigned int i;
	int rc;
	struct cafe_event_action_info *event_action_info;
	struct cafe_event_action_list *sink_actions;
	struct linux_trace_inst *ti = inst->client_data;
	unsigned long long limit, freq;

	/* Check if anything changed since last time we updated. */
	if (ti->change_count == inst->session->output_event_action_change_count)
		return CAFE_OK;
	ti->change_count = inst->session->output_event_action_change_count;
	ti->timestamp_config_action_info = NULL;
	ti->timestamp_config_event_dispatch_pending = false;

	for (i = 0; i < inst->enabled_output_events.ind; ++i) {
		struct cafe_event_descr *event;
		cafe_eal_handle_t sink_actions_handle;
		struct event_action_map *mapping;
		uint64_t requested_odata;

		event = inst->enabled_output_events.list[i].event;
		rc = cafe_ep_inst_event_action_info_get(inst, event,
							&event_action_info);
		if (rc != CAFE_OK)
			return rc;
		if (event_action_info == NULL)
			continue;
		if (event->types[0] == ti->timestamp_config_type) {
			ti->timestamp_config_action_info = event_action_info;
			ti->timestamp_config_event_dispatch_pending = true;
		}

		mapping = inst->enabled_output_events.list[i].impl_data;
		if (mapping == NULL)
			return CAFE_ERROR;

		if (skipcount_get_limit(event, inst->output_odata[ODATA_INDEX_SKIPCOUNT], &limit)) {
			if (limit == 0)
				return CAFE_ERROR;
			freq = 0;
		} else if (sample_freq_value_get(event, inst->output_odata[ODATA_INDEX_SAMPLE_FREQ], &freq)) {
			if (freq == 0)
				return CAFE_ERROR;
			limit = 0;
		} else if (mapping->event_type == PERF_TYPE_HARDWARE ||
			   mapping->event_type == PERF_TYPE_HW_CACHE) {
			/* Sample hardware events by default */
			limit = 0;
			freq = 100;
		} else {
			limit = 1;
			freq = 0;
		}

		requested_odata = 0;
		if (wants_odata(event, inst->output_odata[ODATA_INDEX_STACK]))
			requested_odata |= PERF_SAMPLE_CALLCHAIN;

		if (wants_odata(event, inst->output_odata[ODATA_INDEX_CPU]))
			requested_odata |= PERF_SAMPLE_CPU;

		/* We now know what action to take for the selected event, add
		   it to our list attached to the event */

		rc = cafe_event_action_list_alloc(inst, event, &sink_actions,
						  &sink_actions_handle);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sink_actions_handle);
			return rc;
		}

		if (event->types[0] == ti->memtbl_heap_type) {
			rc = memtbl_input_event_activate(inst, &ti->memtbl_input_events[0], &memtable_input_specs[0]);
			if (rc != CAFE_OK)
				return rc;

			ti->memtbl_heap_info = event_action_info;

		} else if (event->types[0] == ti->memtbl_mmap_type) {
#if 0
			unsigned int j;
			for (j = 1; j < CAFE_NELEMENTS(memtable_input_specs);
			     ++j)
			{
				rc = memtbl_input_event_activate(inst, &ti->memtbl_input_events[j], &memtable_input_specs[j]);
				if (rc != CAFE_OK)
					return rc;
			}
#endif
			ti->memtbl_mmap_info = event_action_info;
		}
		cafe_event_action_list_swap(&event_action_info->sink_actions,
					    sink_actions_handle);

		rc = update_ep_action_map(mapping, event_action_info,
					  (uint64_t)limit, (uint64_t)freq, requested_odata,
					  inst->session->output_event_action_change_count);

		/* For events that fail to enable, we just ignore it and
		 *  carry on. */
		if (rc != CAFE_OK) {
			printf("Failed to enable event %s\n", event->types[0]->ident->name);
		}
	}
	disable_events_for_inst(inst);
	return CAFE_OK;
}


static
void dump_mapping(struct event_action_map *map)
{
#if 0
	printf("map @ %p\n", map);
	printf("\tevent @ %p\n", map->event);
	printf("\t\tname: %s\n", map->event->name);
	printf("\tfs: %d\n", map->fd);
	/* array of ep_actions objects */
	//struct cafe_array actions;
#endif
}


static
struct event_action_map *alloc_map(uint64_t event_type)
{
	struct event_action_map *map = malloc(sizeof(*map));
	if (map == NULL)
		return NULL;
	map->event_type = event_type;
	map->state_dump = NULL;
	return map;
}


/**
 *
 */
static
int add_event(char *format_path, int base)
{
	/* assumes there are exactly two levels to the event name */
	const char *scope;
	char *trav = &format_path[base];
	struct linux_event *event;
	int rc;
	FILE *in = fopen(format_path, "r");
	struct cafe_ep_spec_output_evt e_spec;
	struct event_action_map *map;
	unsigned int i;

	if (in == NULL) {
		printf("error reading %s: %s\n", format_path, strerror(errno));
		return CAFE_ERROR;
	}

	/* find beginning of the event name */
	trav -= 2;
	while (*trav != '/')
		--trav;
	*trav = '\0';

	/* find beginning of the event scope */
	--trav;
	while (*trav != '/')
		--trav;
	scope = trav + 1;
	event = parse_event_format(scope, in);
	fclose(in);

	if (event == NULL)
		return CAFE_ERROR;

	map = alloc_map(PERF_TYPE_TRACEPOINT);
	if (map == NULL)
		return CAFE_NOMEM;

	map->trace_event = event;
	e_spec.type = NULL;	/* filled in later */
	e_spec.impl_data = map;
	e_spec.od_set_idx = OD_SET_ALL;

	for (i = 0; i < CAFE_NELEMENTS(memtable_input_specs); ++i) {
		if ((memtable_input_specs[i].vt == NULL) &&
		    (strcmp(memtable_input_specs[i].name, event->name) == 0))
		{
			rc = trace_event_valuetype(event, &memtable_input_specs[i].vt, NULL, NULL);
			if (rc != CAFE_OK)
				printf("Error building vt string for %s: %s\n",
				       event->name, cafe_strerror(rc));
			memcpy(&memtable_input_specs[i].vt[2], "_my", 3);
			memcpy(&memtable_input_specs[i].map, map,
			       sizeof(*map));
		}
	}
	rc = cafe_array_append(&linux_event_specs, &e_spec);
	dump_mapping(map);
	return rc;
}


/**
 *
 */
static
int nftw_find_format_files(const char *path, const struct stat *sb, int type,
			   struct FTW *ftwbuf)
{
	char *path_copy;

	if (type != FTW_F)
		return FTW_CONTINUE;

	if (strcmp("format", &path[ftwbuf->base]) != 0)
		return FTW_CONTINUE;

	path_copy = strdup(path);
	if (path_copy != NULL) {
		add_event(path_copy, ftwbuf->base);
		free(path_copy);
	}
	return FTW_SKIP_SIBLINGS;
}


/**
 * Simple pattern matcher (like glob).  Only '*' is special.
 */
static
int match(const char *text, const char *pattern)
{
	for (;;) {
		switch (*pattern++) {
		case '\0':
			return (*text != '\0') ? -1 : 0;
		case '*': {
			const char *mark = text;

			while (*text++ != '\0')
				continue;

			do {
				text--;
				if (match(text, pattern) == 0)
					return 0;
			} while (text > mark);

			break;
		}
		default:
			if (*text++ != pattern[-1])
				return 1;
			break;
		}
	}
}


/*
 * I'm deliberately grouping all the type annotation types and
 * functions together (to avoid future merge conflicts).
 *
 * The type annotation machinery has its own allocator.  Why?  I
 * wanted to reduce the nesting of handling allocation failures to
 * just:
 *
 *    if (rc != CAFE_OK)
 *            return rc;
 *
 * as opposed to:
 *
 *    if (rc != CAFE_OK) {
 *            free(lots);
 *            free(of);
 *            free(other);
 *            free(nested);
 *            free(allocations);
 *            return rc;
 *    }
 *
 * The type_annotator_allocator will free up whatever hasn't been
 * free'd when the allocator is destroyed.  We can do this because
 * once the annotations have been loaded and applied, we don't need
 * any of the memory that was allocated as part of that machination.
 *
 * WRS_TODO: We should introduce the notion of a hierarchical allocator.
 */

#define ALLOCATOR_LIST_NODE(PTR)					\
	CAFE_LIST_ENTRY((PTR), struct type_annotator_allocator_node, list_item)

static void *type_annotator_alloc(const struct cafe_allocator *a, size_t size);
static void *type_annotator_realloc(const struct cafe_allocator *a, void *ptr, size_t size);
static void type_annotator_free(const struct cafe_allocator *a, void *ptr);


struct type_annotator_allocator_node {
	struct cafe_list_head list_item;
};


struct type_annotator_allocator {
	struct cafe_allocator base;
	struct cafe_list_head list_head; /* list of all allocations */
};


static struct type_annotator_allocator type_annotator_allocator = {
	{ type_annotator_alloc, type_annotator_realloc, type_annotator_free }
};


struct type_annotator {
	const struct cafe_allocator *allocator;

	/* type_map: a set of annotations to be applied to the value
	 * type representation of Perf's events.  The mapping is:
	 * map<char *type_name, struct type_annotation *>. */
	struct cafe_hashtbl *type_map;
	struct cafe_hashtbl *glob_map;

	/* The list of all dynamically discovered VT definitions. */
	struct cafe_ptr_array types;

	/* An ordered list of glob types from the glob_map.
	 *
	 * WRS_TODO: create a cafe_hashtbl that has a linked list of all
	 * the entries running through it.  That would prevent this
	 * duplication. */
	struct cafe_ptr_array glob_types;

	/* Flag to print the applied annotations. */
	long debug;
};

struct type_annotation_entry {
	const struct cafe_allocator *allocator;
	char *event_name;
	unsigned int lineno;
	struct cafe_ptr_array type_natures; /* char */
	struct cafe_ptr_array fields;       /* field_type_annotation */
};

typedef int(*TYPE_ANNOTATOR_NATURE_ADD)(struct type_annotator *ta,
					struct type_annotation_entry *tae);

static const char *parse_l(const char *fmt, long *result)
{
	char *str;

	errno = 0;
	*result = strtol(fmt, &str, 0);
	if (errno != 0 || fmt >= str)
		return NULL;
	return str;
}


static
void *type_annotator_alloc(const struct cafe_allocator *a, size_t size)
{
	struct type_annotator_allocator *ta;
	struct type_annotator_allocator_node *node;

	assert(a == &type_annotator_allocator.base);

	if ((node = malloc(size + sizeof *node)) == NULL)
		return NULL;

	ta = (struct type_annotator_allocator *)a;
	assert(ta == &type_annotator_allocator);
	cafe_list_init(&node->list_item);
	cafe_list_add_last(&node->list_item, &ta->list_head);

	return (char *)node + sizeof *node;
}


static
void *type_annotator_realloc(const struct cafe_allocator *a,
			     void *ptr,
			     size_t size)
{
	char *old_ptr;
	char *new_ptr;
	struct type_annotator_allocator_node *new_node;
	struct type_annotator_allocator_node *old_node;
	struct type_annotator_allocator_node curr_node;
	struct type_annotator_allocator *ta;

	assert(a == &type_annotator_allocator.base);

	if (ptr == NULL)
		return type_annotator_alloc(a, size);

	old_ptr = (char *)ptr - sizeof *old_node;
	curr_node = *(struct type_annotator_allocator_node *)old_ptr;

	if ((new_ptr = realloc(old_ptr, size + sizeof *new_node)) == NULL)
		return NULL;

	if (new_ptr == old_ptr)
		return ptr;

	ta = (struct type_annotator_allocator *)a;
	new_node = (struct type_annotator_allocator_node *)new_ptr;
	cafe_list_remove(&curr_node.list_item);
	cafe_list_add_last(&new_node->list_item, &ta->list_head);

	return (char *)new_ptr + sizeof *new_node;
}


static
void type_annotator_free(const struct cafe_allocator *a, void *ptr)
{
#if !defined(NDEBUG)
	struct type_annotator_allocator *ta =
		(struct type_annotator_allocator *)a;
#endif
	struct type_annotator_allocator_node *node;

	assert(a == &type_annotator_allocator.base);

	if (ptr != NULL) {
		assert(!cafe_list_is_empty(&ta->list_head));
		ptr = (char *)ptr - sizeof *node;
		node = (struct type_annotator_allocator_node *)ptr;
		cafe_list_remove(&node->list_item);
		free(ptr);
	}
}


static
struct cafe_allocator *type_annotator_allocator_create(void)
{
	cafe_list_init(&type_annotator_allocator.list_head);
	return &type_annotator_allocator.base;
}


static
void type_annotator_allocator_delete(const struct cafe_allocator *allocator)
{
	struct type_annotator_allocator *ta;
	struct cafe_list_head *list_item;
	struct cafe_list_head *tmp;

	assert(allocator == &type_annotator_allocator.base);
	ta = (struct type_annotator_allocator *)allocator;

	cafe_list_foreach_safe(list_item, &ta->list_head, tmp) {
		struct type_annotator_allocator_node *node;
		node = ALLOCATOR_LIST_NODE(list_item);
		cafe_list_remove(&node->list_item);
		free(node);
	}

	assert(cafe_list_is_empty(&ta->list_head));
}


static
void type_annotator_annotation_free(struct type_annotation_entry *ta)
{
	size_t i;

	for (i = 0; i < cafe_ptr_array_len(&ta->type_natures); i++)
		CAFE_FREE(ta->allocator, cafe_ptr_array_val(&ta->type_natures, i));

	for (i = 0; i < cafe_ptr_array_len(&ta->fields); i++) {
		size_t j;
		struct field_type_annotation *fa;
		fa = cafe_ptr_array_val(&ta->fields, i);
		if (fa != NULL)
			CAFE_FREE(ta->allocator, fa->name);
		for (j = 0; j < cafe_ptr_array_len(&fa->natures); j++)
			CAFE_FREE(ta->allocator, cafe_ptr_array_val(&fa->natures, j));
		cafe_ptr_array_release(&fa->natures);
		CAFE_FREE(ta->allocator, fa);
	}

	CAFE_FREE(ta->allocator, ta->event_name);
	cafe_ptr_array_release(&ta->type_natures);
	cafe_ptr_array_release(&ta->fields);
	CAFE_FREE(ta->allocator, ta);
}


static
int type_annotator_annotation_alloc(struct type_annotator *a,
				    const char *type_name,
				    unsigned int lineno,
				    struct type_annotation_entry **result)
{
	struct type_annotation_entry *val;
	char *key = CAFE_STRDUP(a->allocator, type_name);

	if (key == NULL)
		return CAFE_NOMEM;

	val = CAFE_ALLOC(a->allocator, sizeof *val);
	if (val == NULL) {
		CAFE_FREE(a->allocator, key);
		return CAFE_NOMEM;
	}

	val->allocator = a->allocator;
	val->event_name = key;
	val->lineno = lineno;
	cafe_ptr_array_init(&val->type_natures, a->allocator);
	cafe_ptr_array_init(&val->fields, a->allocator);

	if (result != NULL)
		*result = val;

	return CAFE_OK;
}


static
const char *nature_ref_name_get(const char *s, const char **startp, size_t *sz)
{
	size_t len = strlen(s);

	if (len > 2 && s[0] == 'n' && s[1] == '[') {
		size_t i;

		for (i = 2; i < len; i++) {
			if (s[i] == ']' || s[i] == ',')
				break;
		}
		*startp = &s[2];
		*sz = &s[i] - &s[2];

		return &s[2];
	}

	return NULL;
}


static
const char *nature_def_name_get(const char *s, const char **startp, size_t *sz)
{
	size_t len = strlen(s);

	if (len > 2 && s[0] == 'N' && s[1] == '[') {
		size_t i;

		for (i = 2; i < len; i++) {
			if (s[i] == ']' || s[i] == ',')
				break;
		}
		*startp = &s[2];
		*sz = &s[i] - &s[2];

		return &s[2];
	}

	return NULL;
}


#if 0
static
const char *event_name_label(const char *perf_event_name)
{
	const char *p = strchr(perf_event_name, '.');

	if (p != NULL) {
		const char *prefix = perf_event_name;
		size_t prefix_len = p - perf_event_name;
	}
}
#endif


static
bool type_annotator_nature_ref_exists_p(const struct type_annotator *a,
					const char *nature_ref,
					const struct cafe_ptr_array *natures)
{
	size_t i;
	const char *nature_name;
	size_t nature_name_sz;

	if (nature_ref_name_get(nature_ref, &nature_name, &nature_name_sz) == NULL)
		return false;

	for (i = 0; i < cafe_ptr_array_len(natures); i++) {
		const char *element;
		const char *tmp_name;
		size_t tmp_sz;

		element = cafe_ptr_array_val(natures, i);

		if (nature_ref_name_get(element, &tmp_name, &tmp_sz) == NULL)
			continue;

		if (strncmp(nature_name, tmp_name, nature_name_sz) == 0)
			return true;
	}

	return false;
}


static
bool type_annotator_list_has_nature_ref_p(const struct cafe_ptr_array *list,
					  const char *nature_name)
{
	size_t i;

	/* NOT expecting n[<type_ref>]. */
	assert(strchr(nature_name, '[') == NULL);
	assert(strchr(nature_name, ']') == NULL);

	for (i = 0; i < cafe_ptr_array_len(list); i++) {
		const char *element = cafe_ptr_array_val(list, i);
		const char *s;
		size_t sz;

		if (nature_ref_name_get(element, &s, &sz) == NULL)
			continue;

		if (strncmp(s, nature_name, sz) == 0)
			return true;
	}

	return false;
}


static
bool type_annotator_list_has_nature_def_p(const struct cafe_ptr_array *list,
					  const char *nature_name)
{
	size_t i;

	/* NOT expecting N[<type_def>]. */
	assert(strchr(nature_name, '[') == NULL);
	assert(strchr(nature_name, ']') == NULL);

	for (i = 0; i < cafe_ptr_array_len(list); i++) {
		const char *element = cafe_ptr_array_val(list, i);
		const char *s;
		size_t sz;

		if (nature_def_name_get(element, &s, &sz) == NULL)
			continue;

		if (strncmp(s, nature_name, sz) == 0 && !nature_name[sz])
			return true;
	}

	return false;
}


static
bool type_annotator_have_nature_type_ref_p(const struct type_annotator *ta,
					   const char *nature_name)
{
	return type_annotator_list_has_nature_ref_p(&ta->types, nature_name);
}


static
bool type_annotator_have_nature_type_def_p(const struct type_annotator *ta,
					   const char *nature_name)
{
	return type_annotator_list_has_nature_def_p(&ta->types, nature_name);
}


#if 0
static
bool type_annotator_have_nature_type_def_p_old(const struct type_annotator *ta,
					       const char *nature_def)
{
	size_t i;

	for (i = 0; i < cafe_ptr_array_len(&ta->types); i++) {
		const char *element = cafe_ptr_array_val(&ta->types, i);

		if (strcmp(element, nature_def) == 0)
			return true;
	}

	return false;
}
#endif

#if 0
static
bool type_annotator_have_nature_type_def_p_old(const struct type_annotator *ta,
					       const char *nature_def)
{
	size_t i;

	for (i = 0; i < cafe_ptr_array_len(&ta->types); i++) {
		const char *element = cafe_ptr_array_val(&ta->types, i);

		if (strcmp(element, nature_def) == 0)
			return true;
	}

	return false;
}
#endif


static
int type_annotator_annotate_cmd(struct type_annotator *a,
				unsigned int lineno,
				struct cafe_tokenizer *t)
{
	char **wordp;
	int rc;
	size_t i;
	struct type_annotation_entry *tae;
	struct cafe_ptr_array *natures;
	struct cafe_hashtbl *map;

	if (!(strcmp(t->wordv[0], "annotate") == 0 && t->wordc > 2))
		return CAFE_OK;

	if (strchr(t->wordv[1], '*') != NULL)
		map = a->glob_map;
	else
		map = a->type_map;

	tae = cafe_hashtbl_lookup(map, t->wordv[1]);

	if (tae == NULL) {
		rc = type_annotator_annotation_alloc(a, t->wordv[1], lineno,
						     &tae);
		if (rc != CAFE_OK)
			return rc;

		rc = cafe_hashtbl_insert(map, tae->event_name, tae);
		if (rc != CAFE_OK)
			return rc;

		if (map == a->glob_map) {
			rc = cafe_ptr_array_append(&a->glob_types, tae);
			if (rc != CAFE_OK)
				return rc;
		}
	}

	if (strcmp("-f", t->wordv[2]) == 0) {
		if (t->wordv[3] == NULL) /* not enough arguments */
			return CAFE_OK;

		for (i = 0; i < cafe_ptr_array_len(&tae->fields); i++) {
			struct field_type_annotation *fa;

			fa = cafe_ptr_array_val(&tae->fields, i);

			if (strcmp(t->wordv[3], fa->name) == 0) {
				natures = &fa->natures;
				break;
			}
		}

		if (i == cafe_ptr_array_len(&tae->fields)) {
			struct field_type_annotation *fa;

			fa = CAFE_ALLOC(a->allocator, sizeof *fa);
			if (fa == NULL)
				return CAFE_NOMEM;

			cafe_ptr_array_init(&fa->natures, a->allocator);
			fa->name = CAFE_STRDUP(a->allocator, t->wordv[3]);

			if (fa->name == NULL) {
				CAFE_FREE(a->allocator, fa);
				return CAFE_NOMEM;
			}

			rc = cafe_ptr_array_append(&tae->fields, fa);
			if (rc != CAFE_OK)
				return rc;

			natures = &fa->natures;
		}
		wordp = &t->wordv[4];
	} else {
		wordp = &t->wordv[2];
		natures = &tae->type_natures;
	}

	assert(natures != NULL);

	for (; *wordp != NULL; wordp++) {
		char *word;

		if (type_annotator_nature_ref_exists_p(a, *wordp, natures))
			continue;

		word = CAFE_STRDUP(a->allocator, *wordp);
		if (word == NULL)
			return CAFE_NOMEM;

		rc = cafe_ptr_array_append(natures, word);
		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}


static
int type_annotator_type_cmd(struct type_annotator *a,
			    unsigned int lineno,
			    struct cafe_tokenizer *t)
{
	char **wordp;
	int rc;

	if (!(strcmp(t->wordv[0], "type") == 0 && t->wordc > 1))
		return CAFE_OK;

	for (wordp = &t->wordv[1]; *wordp != NULL; wordp++) {
		char *word = CAFE_STRDUP(a->allocator, *wordp);

		if (word == NULL)
			return CAFE_NOMEM;

		rc = cafe_ptr_array_append(&a->types, word);
		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}


static
int type_annotator_debug_cmd(struct type_annotator *a,
			     unsigned int lineno,
			     struct cafe_tokenizer *t)
{
	if (strcmp(t->wordv[0], "debug") == 0 && t->wordc > 1)
		parse_l(t->wordv[1], &a->debug);

	return CAFE_OK;
}


static
void type_annotator_free_hashtbl_val(void *v, void *client_data)
{
	type_annotator_annotation_free(v);
}


static
int type_annotator_init(struct type_annotator *a)
{
	struct cafe_hashtbl *h1, *h2;

	a->allocator = type_annotator_allocator_create();
	if (a->allocator == NULL)
		return CAFE_NOMEM;

	cafe_ptr_array_init(&a->types, a->allocator);
	cafe_ptr_array_init(&a->glob_types, a->allocator);

	h1 = cafe_hashtbl_new(137, a->allocator,
			      cafe_hashtbl_string_hash,
			      cafe_hashtbl_string_equals);
	if (h1 == NULL)
		return CAFE_NOMEM;

	h2 = cafe_hashtbl_new(13, a->allocator,
			      cafe_hashtbl_string_hash,
			      cafe_hashtbl_string_equals);

	if (h2 == NULL) {
		cafe_hashtbl_delete(h1, NULL, NULL, NULL, NULL);
		return CAFE_NOMEM;
	}

	a->type_map = h1;
	a->glob_map = h2;
	a->debug = 0;

	return CAFE_OK;
}


static
void type_annotator_release(struct type_annotator *ta)
{
	size_t i;

	cafe_hashtbl_delete(ta->type_map, NULL, NULL,
			    type_annotator_free_hashtbl_val, ta);

	cafe_hashtbl_delete(ta->glob_map, NULL, NULL,
			    type_annotator_free_hashtbl_val, ta);

	for (i = 0; i < cafe_ptr_array_len(&ta->types); i++)
		CAFE_FREE(ta->allocator, cafe_ptr_array_val(&ta->types, i));

	cafe_ptr_array_release(&ta->types);
	cafe_ptr_array_release(&ta->glob_types);
	type_annotator_allocator_delete(ta->allocator);
}


static
int type_annotator_load(struct type_annotator *a, const char *filename)
{
	FILE *fp;
	char *line = NULL;
	size_t line_sz = 0;
	size_t nbytes;
	struct cafe_tokenizer t;
	unsigned int lineno = 0;

	cafe_tokenizer_init(&t);
	fp = fopen(filename, "r");

	if (fp == NULL)
		return CAFE_OK;	/* no annotations is not an error! */

	while ((nbytes = getline(&line, &line_sz, fp)) != -1) {
		int rc;

		lineno++;
		cafe_tokenizer_reset(&t);
		rc = cafe_tokenize(&t, line, a->allocator);

		switch (rc) {
		case CAFE_TOKENIZER_ENOMEM:
			cafe_tokenizer_release(&t, a->allocator);
			return CAFE_NOMEM;
		case CAFE_TOKENIZER_QUOTED_WORD_NOT_TERMINATED:
			fprintf(stderr, "%s:%d: quoted word not terminated\n",
				filename, lineno);
			cafe_tokenizer_release(&t, a->allocator);
			return CAFE_ERROR;
		case CAFE_TOKENIZER_ESCAPED_CHARACTER_IS_ILLEGAL:
			fprintf(stderr, "%s:%d: escaped character is illegal\n",
				filename, lineno);
			cafe_tokenizer_release(&t, a->allocator);
			return CAFE_ERROR;
		default:
			assert(rc == CAFE_TOKENIZER_SUCCESS);
		}

		if (t.wordc < 1)
			continue;

		if (strcmp(t.wordv[0], "annotate") == 0)
			rc = type_annotator_annotate_cmd(a, lineno, &t);
		else if (strcmp(t.wordv[0], "type") == 0)
			rc = type_annotator_type_cmd(a, lineno, &t);
		else if (strcmp(t.wordv[0], "debug") == 0)
			rc = type_annotator_debug_cmd(a, lineno, &t);

		if (rc != CAFE_OK) {
			cafe_tokenizer_release(&t, a->allocator);
			return rc;
		}
	}

	cafe_tokenizer_release(&t, a->allocator);
	free(line);
	fclose(fp);

	return CAFE_OK;
}


static
int type_annotator_glob_merge(const struct type_annotator *ta,
			      struct cafe_ptr_array *natures,
			      const struct type_annotation_entry *src)
{
	int rc;
	size_t i;

	for (i = 0; i < cafe_ptr_array_len(&src->type_natures); i++) {
		char *nature = cafe_ptr_array_val(&src->type_natures, i);

		if (type_annotator_nature_ref_exists_p(ta, nature, natures))
			continue;

		nature = CAFE_STRDUP(ta->allocator, nature);
		if (nature == NULL)
			return CAFE_NOMEM;

		rc = cafe_ptr_array_append(natures, nature);
		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}


static
void type_annotation_dump(const char *event_name,
			  struct cafe_ptr_array *natures,
			  struct cafe_ptr_array *fields)
{
#if 0
	size_t i, j;

	printf("%s: ", event_name);

	for (i = 0; i < cafe_ptr_array_len(natures); i++) {
		printf("%s ", (char *)cafe_ptr_array_val(natures, i));
	}

	printf("(fields: ");

	for (i = 0; fields != NULL && i < cafe_ptr_array_len(fields); i++) {
		struct field_type_annotation *fa;
		fa = (struct field_type_annotation *)cafe_ptr_array_val(fields, i);
		printf("[%s ", fa->name);
		for (j = 0; j < cafe_ptr_array_len(&fa->natures); j++) {
			printf("%s ", (char *)cafe_ptr_array_val(&fa->natures, j));
		}
		printf("])");
	}

	printf(")\n");
#endif
}


static
int type_annotator_annotation_label_add(struct type_annotator *ta,
					struct type_annotation_entry *tae)
{
	char *subsytem_delim = strchr(tae->event_name, '.');
	char *label = tae->event_name;

	if (subsytem_delim != NULL && *(subsytem_delim + 1) != '\0')
		label = subsytem_delim + 1;

	label = cafe_mprintf(ta->allocator, "n[cafe.LABEL,\"%s\"]", label);
	if (label == NULL)
		return CAFE_NOMEM;

	return cafe_ptr_array_append(&tae->type_natures, label);
}


static
int type_annotator_annotation_counter_add(struct type_annotator *ta,
					  struct type_annotation_entry *tae)
{
	char *ident = tae->event_name;

	ident = cafe_mprintf(ta->allocator, "n[cafe.COUNTER_METRIC_CLASS]");

	if (ident == NULL)
		return CAFE_NOMEM;

	return cafe_ptr_array_append(&tae->type_natures, ident);
}


static
int type_annotator_annotation_category_add(struct type_annotator *ta,
					   struct type_annotation_entry *tae)
{
	char *subsytem_delim = strchr(tae->event_name, '.');
	char *nat_ident;
	int len;

	if (subsytem_delim == NULL)
		return CAFE_OK;

	len = (int)(subsytem_delim - tae->event_name);
	nat_ident = cafe_mprintf(ta->allocator, "category.perf_%.*s", len, tae->event_name);

	if (nat_ident == NULL)
		return CAFE_NOMEM;

	if (!type_annotator_have_nature_type_ref_p(ta, nat_ident)) {
		int rc;
		char *ref;

		ref = cafe_mprintf(ta->allocator, "n[%s]", nat_ident);
		if (ref == NULL) {
			CAFE_FREE(ta->allocator, nat_ident);
			return CAFE_NOMEM;
		}

		rc = cafe_ptr_array_append(&tae->type_natures, ref);
		if (rc != CAFE_OK) {
			CAFE_FREE(ta->allocator, nat_ident);
			return rc;
		}
	}

	if (!type_annotator_have_nature_type_def_p(ta, nat_ident)) {
		int rc;
		char *def;

		def = cafe_mprintf(ta->allocator,
				   "N[%s,,"
				   "n[cafe.METRIC_CATEGORY,\"perf_%.*s\",\"%.*s (Perf)\"]"
				   "]",
				   nat_ident,
				   len, tae->event_name,
				   len, tae->event_name);

		if (def == NULL) {
			CAFE_FREE(ta->allocator, nat_ident);
			return CAFE_NOMEM;
		}

		rc = cafe_ptr_array_append(&ta->types, def);
		if (rc != CAFE_OK) {
			CAFE_FREE(ta->allocator, nat_ident);
			return rc;
		}
	}

	CAFE_FREE(ta->allocator, nat_ident);

	return CAFE_OK;
}


static
bool type_annotator_entry_has_nature_ref_p(struct type_annotator *ta,
					   struct type_annotation_entry *tae,
					   const char *nature_ref)
{
	return type_annotator_list_has_nature_ref_p(&tae->type_natures, nature_ref);
}


static
int type_annotator_cafe_annotations_add(struct type_annotator *ta,
					struct type_annotation_entry *tae)
{
	static struct {
		const char *nature;
		TYPE_ANNOTATOR_NATURE_ADD handler;
	} natures[] = {
		{ "cafe.LABEL", type_annotator_annotation_label_add },
		{ "cafe.COUNTER_METRIC_CLASS", type_annotator_annotation_counter_add }
	};
	size_t i;

	for (i = 0; i < CAFE_NELEMENTS(natures); i++) {
		int rc;

		if (type_annotator_entry_has_nature_ref_p(ta, tae, natures[i].nature))
			continue;

		rc = natures[i].handler(ta, tae);

		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}


static
int type_annotator_dynamic_annotations_add(struct type_annotator *ta,
					   struct type_annotation_entry *tae)
{
	static struct {
		TYPE_ANNOTATOR_NATURE_ADD handler;
	} annotators[] = {
		{ type_annotator_annotation_category_add }
	};
	size_t i;

	for (i = 0; i < CAFE_NELEMENTS(annotators); i++) {
		int rc;

		rc = annotators[i].handler(ta, tae);

		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}


static
int type_annotator_event_annotate(struct type_annotator *ta,
				  struct cafe_ep_spec_output_evt *e_spec)
{
	int rc;
	size_t i;
	struct linux_event *event;
	char *type;
	struct type_annotation_entry *tae;

	event = ((struct event_action_map *)e_spec->impl_data)->trace_event;
	tae = cafe_hashtbl_lookup(ta->type_map, event->name);

	if (tae == NULL) {
		rc = type_annotator_annotation_alloc(ta, event->name, 0, &tae);

		if (rc != CAFE_OK)
			return rc;

		rc = cafe_hashtbl_insert(ta->type_map, tae->event_name, tae);

		if (rc != CAFE_OK)
			return rc;
	}

	/* Append all matching glob type natures. */

	for (i = 0; i < cafe_ptr_array_len(&ta->glob_types); i++) {
		const struct type_annotation_entry *glob_tae;

		glob_tae = cafe_ptr_array_val(&ta->glob_types, i);

		if (match(tae->event_name, glob_tae->event_name) == 0) {
			rc = type_annotator_glob_merge(ta, &tae->type_natures, glob_tae);

			if (rc != CAFE_OK)
				return rc;
		}
	}

	/* Add any missing Cafe annotations. */
	rc = type_annotator_cafe_annotations_add(ta, tae);
	if (rc != CAFE_OK)
		return rc;

	/* Anything else? */
	rc = type_annotator_dynamic_annotations_add(ta, tae);
	if (rc != CAFE_OK)
		return rc;

	type_annotation_dump(event->name, &tae->type_natures, &tae->fields);

	rc = trace_event_valuetype(event, &type, &tae->type_natures, &tae->fields);

	if (rc == CAFE_OK)
		e_spec->type = type;

	if (ta->debug)
		printf("%s\n", type);

	return rc;
}


static
int build_output_event_list(struct type_annotator *ta)
{
	struct cafe_ep_spec_output_evt e_spec;
	struct event_action_map *map;
	int rc;
	unsigned int i;
	const char *filename;
	char *all_types;
	const char **strp;

	filename = getenv("LINUX_TRACE_PERF_ANNOTATIONS");

	if (filename == NULL)
		filename = PERF_ANNOTATIONS_DEFAULT_FILENAME;

	if (filename != NULL) {
		rc = type_annotator_load(ta, filename);
		if (rc != CAFE_OK)
			return rc;
	}

	/* add trace sample event types */
	cafe_array_init(&linux_event_specs, cafe_stdlib_allocator,
			sizeof(struct cafe_ep_spec_output_evt));
	rc = cafe_array_capacity_set(&linux_event_specs,
				     PERF_COUNT_HW_MAX + PERF_COUNT_SW_MAX);
	if (rc != CAFE_OK)
		return rc;

	if (nftw("/sys/kernel/debug/tracing/events", nftw_find_format_files,
		 10, FTW_ACTIONRETVAL) == -1)
	{
		perror("Error locating event descriptions");
		return CAFE_ERROR;
	}

	for (i = 0; i < cafe_array_len(&linux_event_specs); i++) {
		struct cafe_ep_spec_output_evt *spec;

		spec = &cafe_array_val(&linux_event_specs,
				       struct cafe_ep_spec_output_evt, i);
		rc = type_annotator_event_annotate(ta, spec);
		if (rc != CAFE_OK)
			return rc;
	}

	for (i = 0; i < PERF_COUNT_HW_MAX; ++i) {
		if (PERF_HW_VT[i] == NULL) {
			fprintf(stderr, "linux trace EP:%u Warning: perf HW "
				"events [%d .. %d] not supported\n",
				__LINE__, i, PERF_COUNT_HW_MAX - 1);
			break;
		}

		map = alloc_map(PERF_TYPE_HARDWARE);
		if (map == NULL)
			return CAFE_NOMEM;
		map->hardware_event = i;

		e_spec.type = PERF_HW_VT[i];
		e_spec.impl_data = map;
		e_spec.od_set_idx = OD_SET_ALL;
		rc = cafe_array_append(&linux_event_specs, &e_spec);
		if (rc != CAFE_OK)
			return rc;
	}

	for (i = 0; i < PERF_COUNT_SW_MAX; ++i) {
		if (PERF_SW_VT[i] == NULL) {
			fprintf(stderr, "linux trace EP:%u Warning: perf SW "
				"events [%d .. %d] not supported\n",
				__LINE__, i, PERF_COUNT_SW_MAX-1);
			break;
		}

		map = alloc_map(PERF_TYPE_SOFTWARE);
		if (map == NULL)
			return CAFE_NOMEM;
		map->software_event = i;

		e_spec.type = PERF_SW_VT[i];
		e_spec.impl_data = map;

		if (strcmp(e_spec.type, PERF_SW_PAGE_FAULTS_VT_DEF) == 0)
			e_spec.od_set_idx = OD_SET_NO_STACKTRACE;
		else
			e_spec.od_set_idx = OD_SET_ALL;

		rc = cafe_array_append(&linux_event_specs, &e_spec);
		if (rc != CAFE_OK)
			return rc;
	}

	for (i = 0; i < PERF_COUNT_HW_CACHE_MAX; ++i) {
		unsigned int j, k;
		for (j = 0; j < PERF_COUNT_HW_CACHE_OP_MAX; ++j) {
			for (k = 0; k < PERF_COUNT_HW_CACHE_RESULT_MAX; k++) {
				const char* type_desc =
					perf_hw_cache_valuetype(i, j, k);
				if (type_desc == NULL) {
					fprintf(stderr, "linux trace EP:%u Warning: perf cache events >= (%u,%u,%u) not supported\n", __LINE__,i,j,k);
					break;
				}

				map = alloc_map(PERF_TYPE_HW_CACHE);
				if (map == NULL)
					return CAFE_NOMEM;
				map->hw_cache_event = perf_hw_cache_config(i, j, k);

				e_spec.type = type_desc;
				e_spec.impl_data = map;
				e_spec.od_set_idx = OD_SET_ALL;
				rc = cafe_array_append(&linux_event_specs, &e_spec);
				if (rc != CAFE_OK)
					return rc;
			}
		}
	}

	/* add the non-sample event descriptions */
	perf_event_action_map[0] = NULL;
	for (i = 1; i < PERF_RECORD_SAMPLE; ++i) {
		map = alloc_map(INTRINSIC_PERF_EVENT);
		if (map == NULL)
			return CAFE_NOMEM;

		perf_event_action_map[i] = map;
		e_spec.type = perf_eid_to_vt(i);
		if (e_spec.type == NULL)
			return CAFE_ERROR;
		e_spec.impl_data = map;
		e_spec.od_set_idx = OD_SET_NONE;
		rc = cafe_array_append(&linux_event_specs, &e_spec);
		if (rc != CAFE_OK)
			return rc;
	}

	/* add descriptions of synthesized events */
	for (i = 0; i < CAFE_NELEMENTS(synthesized_events); ++i) {
		map = alloc_map(SYNTHETIC_EVENT);
		if (map == NULL)
			return CAFE_NOMEM;

		map->state_dump = synthesized_events[i].state_dump;
		e_spec.type = synthesized_events[i].name;
		e_spec.impl_data = map;
		// only the first synthetic event is a non-memtable event
		if (i == 0)
			e_spec.od_set_idx = OD_SET_TS;
		else
			e_spec.od_set_idx = OD_SET_MEMTBL;

		if ((i == 1) && (memtable_input_specs[0].vt == NULL))
			continue;

		if ((i == 2) &&
		    ((memtable_input_specs[1].vt == NULL) ||
		     (memtable_input_specs[2].vt == NULL) ||
		     (memtable_input_specs[3].vt == NULL)))
			continue;

		rc = cafe_array_append(&linux_event_specs, &e_spec);
		if (rc != CAFE_OK)
			return rc;
	}

	for (strp = perf_hw_category_defs; *strp != NULL; strp++) {
		char *s = CAFE_STRDUP(ta->allocator, *strp);
		rc = cafe_ptr_array_append(&ta->types, s);
		if (rc != CAFE_OK)
			return rc;
	}

	for (strp = perf_sw_category_defs; *strp != NULL; strp++) {
		char *s = CAFE_STRDUP(ta->allocator, *strp);
		rc = cafe_ptr_array_append(&ta->types, s);
		if (rc != CAFE_OK)
			return rc;
	}

	all_types = cafe_mprintf(cafe_stdlib_allocator, "%s", EP_TYPES);

	for (i = 0; i < cafe_ptr_array_len(&ta->types); i++) {
		char *s = cafe_ptr_array_val(&ta->types, i);
		char *q;

		q = cafe_mprintf(cafe_stdlib_allocator, "%s%s", all_types, s);

		if (q == NULL) {
			CAFE_FREE(cafe_stdlib_allocator, all_types);
			all_types = NULL;
			return CAFE_NOMEM;
		}
		CAFE_FREE(cafe_stdlib_allocator, all_types);
		all_types = q;
	}

	spec.types = all_types;
	spec.output_events = linux_event_specs.data;
	spec.n_output_events = cafe_array_len(&linux_event_specs);

	return CAFE_OK;
}


/**
 *  Calculates the individual offsets for each type of orthogonal data offered
 * by perf sample events.  These offsets are in bytes from the beginning of the
 * event record (the first byte of the perf_event_header).
 */
static
void calc_sample_odata_offsets(void)
{
	unsigned int offset = sizeof(struct perf_event_header);

	if (SAMPLE_TYPE & PERF_SAMPLE_IP) {
		sample_type_offsets[PERF_SAMPLE_IP_OFFSET] = offset;
		offset += ODATA_SIZES[PERF_SAMPLE_IP_OFFSET];
	}

	if (SAMPLE_TYPE & PERF_SAMPLE_TID) {
		sample_type_offsets[PERF_SAMPLE_CONTEXT_OFFSET] = offset;
		/* one 32-bit value for each the tid and the pid */
		offset += ODATA_SIZES[PERF_SAMPLE_CONTEXT_OFFSET];
	}

	if (SAMPLE_TYPE & PERF_SAMPLE_TIME) {
		sample_type_offsets[PERF_SAMPLE_TIME_OFFSET] = offset;
		offset += ODATA_SIZES[PERF_SAMPLE_TIME_OFFSET];
	}

	if (SAMPLE_TYPE & PERF_SAMPLE_ADDR) {
		sample_type_offsets[PERF_SAMPLE_ADDR_OFFSET] = offset;
		offset += ODATA_SIZES[PERF_SAMPLE_ADDR_OFFSET];
	}

	if (SAMPLE_TYPE & PERF_SAMPLE_ID) {
		sample_type_offsets[PERF_SAMPLE_ID_OFFSET] = offset;
		offset += ODATA_SIZES[PERF_SAMPLE_ID_OFFSET];
	}

	if (SAMPLE_TYPE & PERF_SAMPLE_STREAM_ID) {
		sample_type_offsets[PERF_SAMPLE_STREAM_ID_OFFSET] = offset;
		offset += ODATA_SIZES[PERF_SAMPLE_STREAM_ID_OFFSET];
	}

	if (SAMPLE_TYPE & PERF_SAMPLE_CPU) {
		sample_type_offsets[PERF_SAMPLE_CPU_OFFSET] = offset;
		/* the cpu value is a 32-bit quantity, but perf allocates
		   64 bits for it */
		offset += ODATA_SIZES[PERF_SAMPLE_CPU_OFFSET];
	}

	if (SAMPLE_TYPE & PERF_SAMPLE_PERIOD) {
		sample_type_offsets[PERF_SAMPLE_PERIOD_OFFSET] = offset;
		offset += ODATA_SIZES[PERF_SAMPLE_PERIOD_OFFSET];
	}

	if (SAMPLE_TYPE & PERF_SAMPLE_READ) {
		sample_type_offsets[PERF_SAMPLE_READ_OFFSET] = offset;
		offset += ODATA_SIZES[PERF_SAMPLE_READ_OFFSET];
	}

	/* the callchain will always be the last piece of data, saves
	   us from having to update the offset value afterwards. */
	sample_type_offsets[PERF_SAMPLE_CALLCHAIN_OFFSET] = offset;
}


static bool spec_registered = false;

int linux_trace_register(void)
{
	int rc;
	struct type_annotator type_annotator;

	assert(CAFE_STACK_MARKER_KERNEL == PERF_CONTEXT_KERNEL);
	assert(CAFE_STACK_MARKER_USER == PERF_CONTEXT_USER);
	if (spec_registered)
		return CAFE_OK;

	/* Must be root */
	if (getuid() != 0) {
		MTD_LOG("%s: error, must be root!\n", __func__);
		return CAFE_ERROR;
	}

	num_cpus = sysconf(_SC_NPROCESSORS_ONLN);
	if (num_cpus == 0)
		num_cpus = 1;

	page_size = sysconf(_SC_PAGE_SIZE);
	if (page_size == -1) {
		perror("Error determining page size");
		return CAFE_ERROR;
	}

	if (!mount_debugfs()) {
		perror("Error mounting debugfs");
		return CAFE_ERROR;
	}

	calc_sample_odata_offsets();

	rc = type_annotator_init(&type_annotator);

	if (rc != CAFE_OK)
		return rc;

	rc = build_output_event_list(&type_annotator);

	if (rc != CAFE_OK) {
		type_annotator_release(&type_annotator);
		return rc;
	}

	type_annotator_release(&type_annotator);

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		spec_registered = true;

	const char * cafe_multibuf = getenv("CAFE_MULTIBUF");
    
    /* if shell var undefined or not set to "0" use MULTI_BUFFER on */
	if (!cafe_multibuf || strncmp(cafe_multibuf, "0", 1)) {
		puts("Enabling per-cpu perf event buffering");
	}
    else {
		MULTI_BUFFER = NULL;
		puts("Disabling per-cpu perf event buffering");
	}
	return rc;
}


void linux_trace_unregister(void)
{
	unsigned int i;

	if (!spec_registered)
		return;

	for (i = 0; i < cafe_array_len(&linux_event_specs); ++i) {
		struct cafe_ep_spec_output_evt *e_spec =
			&cafe_array_val(&linux_event_specs,
					struct cafe_ep_spec_output_evt, i);
		struct event_action_map *map = e_spec->impl_data;

		if (map->event_type == PERF_TYPE_TRACEPOINT) {
			release_linux_event(map->trace_event);
			free((char *)e_spec->type);
		} else if (map->event_type == PERF_TYPE_HW_CACHE)
			free((char *)e_spec->type);
		free(e_spec->impl_data);
	}
	cafe_array_release(&linux_event_specs);

	for (i = 0; i < CAFE_NELEMENTS(memtable_input_specs); ++i) {
		free(memtable_input_specs[i].vt);
		memtable_input_specs[i].vt = NULL;
	}

	cafe_ep_spec_unregister(&spec);
	spec_registered = false;
	CAFE_FREE(cafe_stdlib_allocator, (char *)spec.types);
	spec.types = NULL;
	cafe_ep_spec_unregister(&spec);
	spec_registered = false;
}
