/* tcf_utils.c - convenience routines for TCF-enabled clients */

/*
 * Copyright (c) 2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
24feb14,sdt  Disable SLEEP warnings

DESCRIPTION
This file holds convenience routines for TCF-enabled clients 
*/

#include <assert.h>
#include <pthread.h>
#include <stdio.h>
#include <cafe/lock.h>
#include <cafe/stdbool.h>

/* In VxWorks config, we need to take care of both cafe and agent NDEBUG flags
 * (IOW both _WRS_CONFIG_RUNTIME_ANALYSIS_DEBUG and
 *  _WRS_CONFIG_DEBUG_AGENT_VERBOSE settings)
 * Because state of NDEBUG flag impacts check_error definition see
 * tcf/framework/errors.h for more details
 */

#define cafe_check_error	check_error
#if defined CAFE_OS_VXWORKS
#   if defined DEBUG_AGENT_VERBOSE
#   	undef NDEBUG
#   elif !defined NDEBUG
#   	define NDEBUG
#   endif
#   include <tcf/framework/errors.h>
#endif /* CAFE_OS_VXWORKS */

#include <tcf/services/pathmap.h>
#include <tcf/framework/exceptions.h>
#include <tcf/framework/json.h>
#include <tcf/framework/myalloc.h>
#include <tcf/framework/streams.h>
#include "tcf_utils.h"


#if defined(_WIN32) && !defined(PATH_MAX)
#  define PATH_MAX  (1024)
#endif

typedef struct {
	EventCallBack *handler;
	void *arg;
	pthread_mutex_t wait_lock;
	pthread_cond_t wait_cond;
	bool done;
} sync_handler_arg;


static
void sync_handler(void *arg)
{
	sync_handler_arg *a = arg;

	a->handler(a->arg);

	cafe_check_error(pthread_mutex_lock(&a->wait_lock));
	a->done = true;
	/* lock removed by a different thread */
	/* coverity[sleep] */
	cafe_check_error(pthread_cond_signal(&a->wait_cond));
	cafe_check_error(pthread_mutex_unlock(&a->wait_lock));
}


/**
 * Has the same effect as a call to post_event(), except this routine only
 * returns after the given callback has completed.
 */
void post_event_sync(EventCallBack *handler, void *arg)
{
	sync_handler_arg a;

	assert(!cafe_islocked());
	assert(handler != NULL);

	cafe_check_error(pthread_mutex_init(&a.wait_lock, NULL));
	cafe_check_error(pthread_cond_init(&a.wait_cond, NULL));
	a.handler = handler;
	a.arg = arg;
	a.done = false;

	/* lock removed by a different thread */
	/* coverity[sleep] */
	cafe_check_error(pthread_mutex_lock(&a.wait_lock));
	post_event(sync_handler, &a);

	/* lock removed by a different thread */
	/* coverity[sleep] */
	/* coverity[loop_condition] */
	do cafe_check_error(pthread_cond_wait(&a.wait_cond, &a.wait_lock));
	while (!a.done);
	
	cafe_check_error(pthread_mutex_unlock(&a.wait_lock));
	
	/* cleanup */
        cafe_check_error(pthread_mutex_destroy(&a.wait_lock));
        cafe_check_error(pthread_cond_destroy(&a.wait_cond));
}


void cafe_post_event_sync(EventCallBack *handler, void *arg)
{
	CAFE_UNLOCK();
	post_event_sync(handler, arg);
	CAFE_LOCK();
}


struct add_pathmap_arg {
	const char *source;
	const char *dest;
	struct Channel *c;
};


static
char *write_json_string(const char *str)
{
	ByteArrayOutputStream buf;
	OutputStream *out;
	char *json_str;

	out = create_byte_array_output_stream(&buf);
	json_write_string(out, str);
	write_stream(out, 0);
	get_byte_array_output_stream_data(&buf, &json_str, NULL);
	return json_str;
}


/**
 *  Callback used by tcf_add_pathmap().
 */
static
void map_path(void *arg)
{
	struct add_pathmap_arg *a = arg;
	Trap trap;

	if (set_trap(&trap)) {
		PathMapRuleAttribute * source_attr = (PathMapRuleAttribute *) loc_alloc_zero(
				sizeof(PathMapRuleAttribute));
		PathMapRuleAttribute * dest_attr = (PathMapRuleAttribute *) loc_alloc_zero(
				sizeof(PathMapRuleAttribute));
		source_attr->name = loc_strdup(PATH_MAP_SOURCE);
		source_attr->value = write_json_string(a->source);
		dest_attr->name = loc_strdup(PATH_MAP_DESTINATION);
		dest_attr->value = write_json_string(a->dest);
		source_attr->next = dest_attr;
		dest_attr->next = NULL;
		create_path_mapping(source_attr);
		clear_trap(&trap);
	} else {
		cafe_check_error(trap.error);
	}
}


void tcf_add_pathmap(const char *source, const char *destination)
{
	struct add_pathmap_arg a;

	a.source = source;
	a.dest = destination;

	if (cafe_islocked())
		cafe_post_event_sync(map_path, &a);
	else
		post_event_sync(map_path, &a);

}
