/* cpu_hog_fork.c - process-based cpu exerciser */

/* Copyright (c) 2011-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/wait.h>
#include <pthread.h>
#include <assert.h>
#include <unistd.h>
#include "../../src/time_utils.h"

#define CLOCK_ID  CLOCK_THREAD_CPUTIME_ID

static void busyloop(unsigned long duration)
{
	int rc;
	struct timespec end_time, current_time;

	rc = clock_gettime(CLOCK_ID, &current_time);
	if (rc != 0)
		fprintf(stderr, "Error getting current time: %s\n",
			strerror(errno));
	assert(rc == 0);
	end_time = current_time;
	timespec_ticks_add(duration, NSECS_PER_SEC, &end_time);
	rc = clock_gettime(CLOCK_ID, &current_time);
	assert(rc == 0);
	while (timespec_cmp(&current_time, &end_time) < 0) {
		rc = clock_gettime(CLOCK_ID, &current_time);
		assert(rc == 0);
	}
}

static void idle(unsigned long duration)
{
	struct timespec ts;

	timespec_from_ticks(duration, NSECS_PER_SEC, &ts);
	while (nanosleep(&ts, &ts))
		perror("Error nanosleep()ing");
}

static void cpu_ramp(unsigned int num_steps)
{
#define GRANULARITY 5

	do {
		unsigned int i;
		for (i = 0; i <= num_steps; ++i) {
			float load = i / (float)num_steps;
			busyloop((load * NSECS_PER_SEC) / GRANULARITY);
			idle((1.0 - load) * NSECS_PER_SEC / GRANULARITY);
		}
	} while (1);
}

static
void hog_cpu_til_signal(void)
{
	cpu_ramp(10);
	while (1);
}


static
void process_func(int cpuid)
{
	struct sigaction act;

	/* reset signal handler to overwrite parent process's handler */
	act.sa_handler = SIG_DFL;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGTERM);

	if (sigaction(SIGTERM, &act, NULL)) {
		perror("Failed to install SIGTERM handler");
		exit(EXIT_FAILURE);
	}

	//cpu_hogging_set_affinity(cpuid);
	hog_cpu_til_signal();
}


static
void cpu_hogging_stop(unsigned int num, pid_t *children)
{
	unsigned int i;
	int status;

	/* terminate forked processes */
	//v2printf("Stop cpu hogging!\n");
	for (i = 0; i < num; ++i)
		kill(children[i], SIGTERM);

	for (i = 0; i < num; ++i) {
		waitpid(children[i], &status, 0);
		if (WIFEXITED(status)) {
			//v3printf("process %d exited normally\n", process[i]);
		}
	}
	free(children);
}


/**
 * cpu_hogging_start - start a process on each cpu
 *
 * exit program when failed.
 */
static
pid_t *cpu_hogging_start(unsigned int n_processes)
{
	unsigned int i;
	//v2printf("Start cpu hogging!\n");
	pid_t *children = malloc(n_processes * sizeof(pid_t));

	if (children == NULL) {
		perror("Unable to allocate process list");
		exit(EXIT_FAILURE);
	}

	for (i = 0; i < n_processes; ++i) {
		children[i] = fork();
		if (children[i] == 0){
			process_func(i);
		} else if (children[i] == -1) {
			perror("Failed to fork cpu-hogging child");
			cpu_hogging_stop(--i, children);
			exit(EXIT_FAILURE);
		}
	}

	return children;
}


int main(int argc, char *argv[])
{
	unsigned int nr;
	pid_t *children;

	if (argc == 2) {
		nr = atoi(argv[1]);
	} else {
		nr = sysconf(_SC_NPROCESSORS_ONLN);
		if (nr == -1) {
			perror("Couldn't determine number of CPUs");
			return EXIT_FAILURE;
		}
	}

	children = cpu_hogging_start(nr);

	do {
		pause();
	} while (1);

	cpu_hogging_stop(nr, children);

	return EXIT_SUCCESS;
}
