/* timestamp_normalizer - normalize timestamp values w.r.t clock rollovers */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

/*
 * This configuration normalizes timestamp values so that downstream
 * event processors need not deal with with timestamp values where the
 * time source has rolled over or where time goes backwards.
 */

#include <string.h>
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/value_print.h>
#include <cafe/ep/timestamp_normalizer.h>
#include <assert.h>

static void timestamp_freq_with_ts_act_handler(struct cafe_event_action_info *, void **);
static void timestamp_freq_act_handler(struct cafe_event_action_info *, void **);
static void wildcard_with_ts_act_handler(struct cafe_event_action_info *, void **);

struct dispatch_data {
	unsigned int count;
	unsigned int capacity;
	void **data;
};

struct timestamp_event {
	unsigned long long tick;
};

struct timestamp_freq_event {
	unsigned long long frequency;
};

struct ts_normalizer_inst {
	unsigned long long cumulative_tick;
	unsigned long long clock_frequency;
	long change_count;
	struct dispatch_data dispatch_data;
};

static const char *wildcard1_TYPE[] = {
	cafe_wildcard_VT_REF
};

static const char *wildcard2_TYPE[] = {
	cafe_wildcard_VT_REF,
	cafe_timestamp_VT_REF
};

static const char *timestamp_freq1_TYPE[] = {
	cafe_timestamp_config_VT_REF
};

static const char *timestamp_freq2_TYPE[] = {
	cafe_timestamp_config_VT_REF,
	cafe_timestamp_VT_REF
};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT2(timestamp_freq2, timestamp_freq_with_ts_act_handler),
	CAFE_INPEV_INIT2(timestamp_freq1, timestamp_freq_act_handler),
	CAFE_INPEV_INIT2(wildcard2, wildcard_with_ts_act_handler),
	CAFE_INPEV_INIT(wildcard1)
};

static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};

static const char *orthogonal_data[] = {
	cafe_timestamp_VT_REF
};

static int grow_dispatch_data(struct ts_normalizer_inst *i,
			      unsigned int new_capacity,
			      const struct cafe_allocator *allocator)
{
	if (new_capacity <= i->dispatch_data.capacity)
		return CAFE_OK;

	i->dispatch_data.capacity = new_capacity;
	i->dispatch_data.data = CAFE_REALLOC
		(allocator,
		 i->dispatch_data.data,
		 i->dispatch_data.capacity * sizeof *i->dispatch_data.data);

	return (i->dispatch_data.data == NULL) ? CAFE_NOMEM : CAFE_OK;
}

/*
 * Normalize incoming timestamp values to ensure time always increases
 */
static void normalize_timestamp(struct ts_normalizer_inst *i,
				const struct timestamp_event *timestamp_event)
{
	if (timestamp_event->tick > i->cumulative_tick)
		i->cumulative_tick = timestamp_event->tick;
}

static void wildcard_with_ts_act_handler(struct cafe_event_action_info *info,
					 void **data)
{
	int rc;
	unsigned int j;
	struct ts_normalizer_inst *i;

	assert(info->event->num_types > 1);
	i = (struct ts_normalizer_inst *)info->inst->client_data;

	if (i->clock_frequency == 0) {
		cafe_dispatch_event(&info->sink_actions, data);
		return;
	}

	rc = grow_dispatch_data(i, info->event->num_types, info->inst->allocator);

	if (rc != CAFE_OK) {
		cafe_dispatch_error(info->inst, rc);
		return;
	}

	normalize_timestamp(i, data[info->odmap[1]]);

	for (j = 0; j < info->event->num_types; j++) {
		i->dispatch_data.data[j] = data[j];
	}
	i->dispatch_data.data[info->odmap[1]] = &i->cumulative_tick;

	cafe_dispatch_event(&info->sink_actions, i->dispatch_data.data);
}

static void timestamp_freq_act_handler(struct cafe_event_action_info *info,
				       void **data)
{
	int rc;
	unsigned int j;
	struct ts_normalizer_inst *i;
	const struct timestamp_freq_event *e;
	struct timestamp_freq_event freq_evt;

	e = (struct timestamp_freq_event *)data[0];
	i = (struct ts_normalizer_inst *)info->inst->client_data;

	i->clock_frequency = e->frequency;

	if (e->frequency == 0) {
		cafe_dispatch_event(&info->sink_actions, data);
		return;
	}

	rc = grow_dispatch_data(i, info->event->num_types, info->inst->allocator);

	if (rc != CAFE_OK) {
		cafe_dispatch_error(info->inst, rc);
		return;
	}

	freq_evt.frequency = e->frequency;

	i->dispatch_data.data[0] = &freq_evt;
	for (j = 1; j < info->event->num_types; j++) {
		i->dispatch_data.data[j] = data[j];
	}

	cafe_dispatch_event(&info->sink_actions, i->dispatch_data.data);
}

static void timestamp_freq_with_ts_act_handler(struct cafe_event_action_info *info,
					       void **data)
{
	int rc;
	unsigned int j;
	struct ts_normalizer_inst *i;
	const struct timestamp_freq_event *e;
	struct timestamp_freq_event freq_evt;

	e = (struct timestamp_freq_event *)data[0];
	i = (struct ts_normalizer_inst *)info->inst->client_data;

	i->clock_frequency = e->frequency;

	if (e->frequency == 0) {
		cafe_dispatch_event(&info->sink_actions, data);
		return;
	}

	rc = grow_dispatch_data(i, info->event->num_types, info->inst->allocator);

	if (rc != CAFE_OK) {
		cafe_dispatch_error(info->inst, rc);
		return;
	}

	normalize_timestamp(i, data[info->odmap[1]]);

	freq_evt.frequency = e->frequency;

	i->dispatch_data.data[0] = &freq_evt;
	for (j = 1; j < info->event->num_types; j++) {
		i->dispatch_data.data[j] = data[j];
	}
	i->dispatch_data.data[info->odmap[1]] = &i->cumulative_tick;

	cafe_dispatch_event(&info->sink_actions, i->dispatch_data.data);
}

static int timestamp_normalizer_create(struct cafe_ep_inst *inst)
{
	struct ts_normalizer_inst *i;

	if ((i = CAFE_ALLOC(inst->allocator, sizeof *i)) == NULL)
		return CAFE_NOMEM;

	memset(i, 0, sizeof *i);
	i->change_count = inst->session->output_event_action_change_count - 1;
	inst->client_data = i;

	return CAFE_OK;
}

static void timestamp_normalizer_delete(struct cafe_ep_inst *inst)
{
	struct ts_normalizer_inst *i = (struct ts_normalizer_inst *)inst->client_data;
	CAFE_FREE(inst->allocator, i->dispatch_data.data);
	CAFE_FREE(inst->allocator, inst->client_data);
}

static int get_input_event_actions(struct cafe_ep_inst *inst,
				   struct cafe_event_descr *event,
				   struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *event_info;
	struct cafe_event_action_list *sink_actions;
	cafe_eal_handle_t sink_actions_handle;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);
	if ((rc != CAFE_OK) || (event_info == NULL))
		return rc;

	if (event_info->impl_data == NULL)
		/* forward dispatches to interested sinks */
		return cafe_ep_inst_get_sink_event_actions(inst, actions);

	rc = cafe_event_action_list_alloc(inst, event,
					  &sink_actions,
					  &sink_actions_handle);

	if (rc != CAFE_OK)
		return rc;

	/* Ask the sink if it is interested in this event type. */

	rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);

	if (rc != CAFE_OK) {
		cafe_event_action_list_release(&sink_actions_handle);
		return rc;
	}

	cafe_event_action_list_swap(&event_info->sink_actions,
				    sink_actions_handle);

	return cafe_add_event_action(inst, actions,
				     (CAFE_ACTION_HANDLER)event_info->impl_data,
				     event_info);
}

#define ALL_TYPES				\
	cafe_timestamp_config_VT_DEF		\
	cafe_timestamp_VT_DEF			\
	cafe_wildcard_VT_DEF

static struct cafe_ep_spec spec = {
	"cafe_timestamp_normalizer",
	timestamp_normalizer_create,
	timestamp_normalizer_delete,
	NULL,			/* start */
	NULL,			/* stop */
	get_input_event_actions,
	NULL,			/* update_output_event_actions */
	NULL,			/* inject event */
	ALL_TYPES,
	NULL,			/* workarea */
	CAFE_NELEMENTS(input_events),
	input_events,
	CAFE_NELEMENTS(output_events),
	output_events,
	CAFE_NELEMENTS(orthogonal_data),
	orthogonal_data
};

static bool cfg_registered = false;

void cafe_timestamp_normalizer_ep_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}

int cafe_timestamp_normalizer_ep_register(void)
{
	int rc = CAFE_OK;

	if (!cfg_registered) {
		rc = cafe_ep_spec_register(&spec);
		if (rc == CAFE_OK)
			cfg_registered = true;
	}

	return rc;
}
