/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
06jun14,tle  Keep event order when disabling events (WB4-3685)

DESCRIPTION

*/

#include <string.h>		/* memset() */
#include <stdarg.h>
#include <cafe/cafe.h>
#include <cafe/errors.h>
#include <cafe/expr.h>
#include <cafe/inst.h>
#include <cafe/value_type.h>
#include <cafe/autogenerated/cafe.type.h>
#include "cafe_assert.h"
#include "cafe_trace.h"

static int cafe_ep_inst_disconnect_edge(struct cafe_ep_inst_edge *edge);

static int inst_create(struct cafe_instrument *session,
		       const struct cafe_ep_spec *spec,
		       unsigned int n_settings,
		       const struct cafe_cfg_setting *settings,
		       struct cafe_ep_inst **result)
{
	int rc;
	unsigned int i;
	struct cafe_ep_inst *inst;

	if ((inst = CAFE_ALLOC(session->allocator, sizeof(*inst))) == NULL)
		return CAFE_NOMEM;
	memset(inst, 0, sizeof(*inst));

	inst->state = CAFE_EP_INST_STATE_INIT;
	inst->allocator = session->allocator;
	inst->session = session;
	inst->spec = spec;

	if (spec->types != NULL) {
		rc = cafe_type_parse(session->ns, spec->types, NULL);
		if (rc != CAFE_OK)
			goto error;
	}

	if (spec->workarea != NULL) {
		rc = cafe_type_parse(session->ns, spec->workarea, &inst->workarea_type);
		if (rc != CAFE_OK)
			goto error;
		if (inst->workarea_type == NULL) {
			rc = CAFE_ERROR;
			goto error;
		}
		if (cafe_find_nature_str(inst->workarea_type->ident,
					 cafe_input_any_ctx_VT_DEF))
			inst->flags |= CAFE_EP_INST_FLAG_INPUT_ANY_CTX;
		if (cafe_find_nature_str(inst->workarea_type->ident,
					 cafe_output_inst_thread_VT_DEF))
			inst->flags |= CAFE_EP_INST_FLAG_OUTPUT_INST_THREAD;
		if (cafe_find_nature_str(inst->workarea_type->ident,
					 cafe_output_session_thread_VT_DEF))
			inst->flags |= CAFE_EP_INST_FLAG_OUTPUT_SESSION_THREAD;
	}

	if (spec->n_input_events > 0) {
		inst->input_events = CAFE_ALLOC(
			inst->allocator,
			spec->n_input_events * sizeof *inst->input_events);
		if (inst->input_events == NULL) {
			rc = CAFE_NOMEM;
			goto error;
		}
		for (i = 0; i < spec->n_input_events; i++) {
			struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES] = {0};
			unsigned int j;

			assert(spec->input_events[i].num_types > 0);
			assert(spec->input_events[i].num_types < CAFE_MAX_EVENT_TYPES);
			for (j = 0; j < spec->input_events[i].num_types; j++) {
				rc = cafe_type_parse(session->ns,
						     spec->input_events[i].type[j],
						     &types[j]);
				if (rc != CAFE_OK)
					goto error;
			}
			rc = cafe_cfg_get_event(
				session,
				spec->input_events[i].num_types,
				types,
				&inst->input_events[i]);
			if (rc != CAFE_OK)
				goto error;
		}
	}

	if (spec->n_output_events > 0) {
		inst->output_events = CAFE_ALLOC(
			inst->allocator,
			spec->n_output_events * sizeof *inst->output_events);
		if (inst->output_events == NULL) {
			rc = CAFE_NOMEM;
			goto error;
		}
		for (i = 0; i < spec->n_output_events; i++) {
			rc = cafe_type_parse(session->ns,
					     spec->output_events[i].type,
					     &inst->output_events[i]);
			if (rc != CAFE_OK)
				goto error;
		}
	}

	if (spec->n_output_odata > 0) {
		inst->output_odata = CAFE_ALLOC(
			inst->allocator,
			spec->n_output_odata * sizeof *inst->output_odata);
		if (inst->output_odata == NULL) {
			rc = CAFE_NOMEM;
			goto error;
		}
		for (i = 0; i < spec->n_output_odata; i++) {
			rc = cafe_type_parse(session->ns,
					     spec->output_odata[i],
					     &inst->output_odata[i]);
			if (rc != CAFE_OK)
				goto error;
		}
	}

	if (inst->workarea_type != NULL) {
		rc = cafe_value_create(inst->workarea_type, &inst->workarea);
		if (rc != CAFE_OK)
			goto error;
	}

	if (n_settings != 0) {
		struct cafe_value_data *workarea = inst->workarea;
		if (workarea == NULL) {
			rc = CAFE_ERROR;
			goto error;
		}
		for (i = 0; i < n_settings; i++) {
			rc = cafe_value_store(inst->workarea_type,
					      workarea->u.data,
					      settings[i].name,
					      &settings[i].value,
					      inst->allocator);
			if (rc != CAFE_OK)
				goto error;
		}
	}

	cafe_list_init(&inst->source_list_head);
	cafe_list_init(&inst->sink_list_head);

	if (cafe_list_is_empty(&session->inst_list_head)) {
		inst->order = 0;
	} else {
		struct cafe_ep_inst *last_inst =
			SESSION_INST_LIST_ENTRY(session->inst_list_head.prev);
		inst->order = last_inst->order + 1;
	}
	cafe_list_add_last(&inst->session_inst_list, &session->inst_list_head);

	*result = inst;

	return CAFE_OK;

error:
	if (inst->workarea != NULL)
		cafe_value_delete(inst->workarea);
	CAFE_FREE(inst->allocator, inst->output_odata);
	CAFE_FREE(inst->allocator, inst->output_events);
	CAFE_FREE(inst->allocator, inst->input_events);
	CAFE_FREE(inst->allocator, inst);
	return rc;
}

static int inst_free(struct cafe_ep_inst *inst)
{
	unsigned int i;
	struct cafe_event_action_info *event_info;

	assert(inst);

	for (i = 0; i < inst->event_action_data_list.ind; i++) {
		event_info = inst->event_action_data_list.list[i];
		cafe_event_action_list_release(&event_info->sink_actions);
		CAFE_FREE(inst->allocator, event_info->odmap);
		CAFE_FREE(inst->allocator, event_info);
	}
	CAFE_FREE(inst->allocator, inst->event_action_data_list.list);
	CAFE_FREE(inst->allocator, inst->enabled_output_events.list);

	if (inst->workarea != NULL)
		cafe_value_delete(inst->workarea);

	cafe_list_remove(&inst->session_inst_list);

	CAFE_FREE(inst->allocator, inst->output_odata);
	CAFE_FREE(inst->allocator, inst->output_events);
	CAFE_FREE(inst->allocator, inst->input_events);
	CAFE_FREE(inst->allocator, inst);

	return CAFE_OK;
}

int cafe_ep_inst_create(struct cafe_instrument *session,
			const struct cafe_ep_spec *spec,
			unsigned int n_settings,
			const struct cafe_cfg_setting *settings,
			struct cafe_ep_inst **result)
{
	int rc;
	struct cafe_ep_inst *inst = NULL;

	CAFE_ASSERT_LOCKED();

	rc = inst_create(session, spec, n_settings, settings, &inst);
	if (rc != CAFE_OK)
		return rc;

	if (spec->create_inst != NULL) {
		rc = spec->create_inst(inst);
		if (rc != CAFE_OK) {
			(void)inst_free(inst);
			return rc;
		}
	}

	inst->state = CAFE_EP_INST_STATE_STOPPED;

	*result = inst;

	return CAFE_OK;
}

int cafe_ep_inst_delete(struct cafe_ep_inst *inst)
{
	CAFE_ASSERT_LOCKED();
	assert(inst);

	if (inst->state != CAFE_EP_INST_STATE_STOPPED)
		return CAFE_INVALID_STATE;

	if (!cafe_list_is_empty(&inst->source_list_head) ||
	    !cafe_list_is_empty(&inst->sink_list_head))
		return CAFE_ERROR;

	inst->state = CAFE_EP_INST_STATE_DELETING;
	if (inst->spec->delete_inst != NULL)
		inst->spec->delete_inst(inst);

	return inst_free(inst);
}

static struct cafe_ep_inst *join_exec_ctx(
	struct cafe_ep_inst *exec_ctx1,
	struct cafe_ep_inst *exec_ctx2)
{
	if (exec_ctx1 == NULL)
		return exec_ctx2;
	if (exec_ctx2 == NULL)
		return exec_ctx1;
	if (exec_ctx1 == CAFE_EXEC_CTX_ANY)
		return exec_ctx1;
	if (exec_ctx2 == CAFE_EXEC_CTX_ANY)
		return exec_ctx2;
	if (exec_ctx1 == exec_ctx2)
		return exec_ctx1;
	return CAFE_EXEC_CTX_ANY;
}

static int check_resolved_sources(struct cafe_ep_inst *sink)
{
	struct cafe_list_head *list_item;
	struct cafe_ep_inst *exec_ctx = NULL;

	CAFE_ASSERT_LOCKED();
	assert(sink);

	cafe_list_foreach(list_item, &sink->source_list_head) {
		struct cafe_ep_inst_edge *edge =
			SOURCE_LIST_NODE_ENTRY(list_item);
		if (edge->source->order < 0)
			return CAFE_ERROR;
		exec_ctx = join_exec_ctx(exec_ctx, edge->source->output_ctx);
	}
	if (exec_ctx == CAFE_EXEC_CTX_ANY &&
	    (sink->flags & CAFE_EP_INST_FLAG_INPUT_ANY_CTX) == 0)
		return CAFE_ERROR;
	sink->input_ctx = exec_ctx;
	if (sink->flags & CAFE_EP_INST_FLAG_OUTPUT_INST_THREAD) {
		exec_ctx = sink;
	} else if (sink->flags & CAFE_EP_INST_FLAG_OUTPUT_SESSION_THREAD) {
		exec_ctx = CAFE_EXEC_CTX_SESSION;
	} else if (sink->spec->n_input_events == 0) {
		exec_ctx = CAFE_EXEC_CTX_ANY;
	}
	sink->output_ctx = exec_ctx;
	return CAFE_OK;
}

/*
 * This function sorts instances in topological order such that all
 * sources come before their connected sinks.  Instances without
 * direct or indirect connection have undefined ordering.
 *
 * As part of the sorting, the connections are checked for circles and
 * to make sure that the threadding model between a source and its
 * sink is matching.
 */
static int update_inst_order(struct cafe_instrument *session)
{
	struct cafe_list_head *list_item;
	struct cafe_list_head *tmp;
	int status;
	const int ORDER_RESOLVED = 1;
	const int ORDER_UNRESOLVED = 2;
	int order = 0;

	CAFE_ASSERT_LOCKED();

	cafe_list_foreach_safe(list_item, &session->inst_list_head, tmp) {
		struct cafe_ep_inst *inst =
			SESSION_INST_LIST_ENTRY(list_item);
		inst->order = -1;
	}
	do {
		status = 0;
		cafe_list_foreach_safe(list_item, &session->inst_list_head, tmp) {
			struct cafe_ep_inst *inst =
				SESSION_INST_LIST_ENTRY(list_item);
			if (inst->order >= 0)
				/* Reached resolved instance so the
				 * rest of the list is already
				 * ordered */
				break;
			if (check_resolved_sources(inst) == CAFE_OK) {
				/* All sources are resolved, move
				 * instance to temporary list, after
				 * any instance it depends on  */
				cafe_list_remove(&inst->session_inst_list);
				cafe_list_add_last(&inst->session_inst_list, &session->inst_list_head);
				inst->order = order++;
				status |= ORDER_RESOLVED;
			} else {
				status |= ORDER_UNRESOLVED;
			}
		}
	} while (status == (ORDER_RESOLVED | ORDER_UNRESOLVED));

	if (status & ORDER_UNRESOLVED)
		return CAFE_INVALID_CONNECTION;
	return CAFE_OK;
}

int cafe_ep_inst_connect(struct cafe_ep_inst *source,
			 struct cafe_ep_inst *sink)
{
	struct cafe_ep_inst_edge *edge;
	int rc;

	CAFE_ASSERT_LOCKED();
	assert(source);
	assert(sink);

	if (source->session != sink->session)
		return CAFE_ERROR;

	edge = CAFE_ALLOC(source->allocator, sizeof *edge);
	if (edge == NULL)
		return CAFE_NOMEM;
	edge->source = source;
	edge->sink = sink;
	cafe_list_add_last(&edge->sink_list, &source->sink_list_head);
	cafe_list_add_last(&edge->source_list, &sink->source_list_head);
	rc = update_inst_order(source->session);
	if (rc != CAFE_OK) {
		(void)cafe_ep_inst_disconnect_edge(edge);
		(void)update_inst_order(source->session);
		return rc;
	}
	source->session->output_event_action_change_count++;
	return CAFE_OK;
}

static int cafe_ep_inst_disconnect_edge(struct cafe_ep_inst_edge *edge)
{
	CAFE_ASSERT_LOCKED();

	if (edge->source == NULL || edge->sink == NULL)
		return CAFE_ERROR;

	cafe_list_remove(&edge->source_list);
	cafe_list_remove(&edge->sink_list);
	edge->source->session->output_event_action_change_count++;
	CAFE_FREE(edge->source->allocator, edge);

	return CAFE_OK;
}

int cafe_ep_inst_disconnect(struct cafe_ep_inst *source,
			    struct cafe_ep_inst *sink)
{
	struct cafe_list_head *list_item;

	CAFE_ASSERT_LOCKED();
	assert(source);
	assert(sink);

	cafe_list_foreach(list_item, &sink->source_list_head) {
		struct cafe_ep_inst_edge *edge =
			SOURCE_LIST_NODE_ENTRY(list_item);
		assert(edge->sink == sink);
		if (edge->source != source)
			continue;
		(void)cafe_ep_inst_disconnect_edge(edge);
		return CAFE_OK;
	}
	return CAFE_ERROR;
}

void cafe_ep_inst_disconnect_all(struct cafe_ep_inst *inst)
{
	struct cafe_list_head *list_item, *tmp;

	CAFE_ASSERT_LOCKED();

	cafe_list_foreach_safe(list_item, &inst->source_list_head, tmp) {
		struct cafe_ep_inst_edge *edge =
			SOURCE_LIST_NODE_ENTRY(list_item);
		int rc;
		rc = cafe_ep_inst_disconnect_edge(edge);
		assert(rc == CAFE_OK);
		if (rc != CAFE_OK)
			TRACE(("Error disconnecting edge (%s)\n",
			       cafe_strerror(rc)));
	}

	cafe_list_foreach_safe(list_item, &inst->sink_list_head, tmp) {
		struct cafe_ep_inst_edge *edge =
			SINK_LIST_NODE_ENTRY(list_item);
		int rc;
		rc = cafe_ep_inst_disconnect_edge(edge);
		assert(rc == CAFE_OK);
		if (rc != CAFE_OK)
			TRACE(("Error disconnecting edge (%s)\n",
			       cafe_strerror(rc)));
	}
}

int cafe_ep_inst_start(struct cafe_ep_inst *inst)
{
	int rc = CAFE_OK;

	CAFE_ASSERT_LOCKED();
	assert(inst);

	if (inst->state == CAFE_EP_INST_STATE_STARTED)
		return CAFE_OK;
	if (inst->state != CAFE_EP_INST_STATE_STOPPED)
		return CAFE_INVALID_STATE;

	inst->state = CAFE_EP_INST_STATE_STARTING;
	if (inst->spec->start)
		rc = inst->spec->start(inst);
	else
		cafe_ep_inst_state_started(inst);

	return rc;
}

int cafe_ep_inst_stop(struct cafe_ep_inst *inst)
{
	CAFE_ASSERT_LOCKED();
	assert(inst);

	if (inst->state == CAFE_EP_INST_STATE_STOPPED)
		return CAFE_OK;
	if (inst->state != CAFE_EP_INST_STATE_STARTED)
		return CAFE_INVALID_STATE;

	inst->state = CAFE_EP_INST_STATE_STOPPING;
	if (inst->spec->stop != NULL)
		return inst->spec->stop(inst);

	cafe_ep_inst_state_stopped(inst);
	return CAFE_OK;
}

static void cafe_input_state_change(struct cafe_ep_inst *inst)
{
	struct cafe_instrument *session = inst->session;
	struct cafe_list_head *list_item;

	if (!(inst->flags & CAFE_EP_INST_FLAG_OUTPUT_INST_THREAD))
		return;
	cafe_list_foreach(list_item, &session->inst_list_head) {
		struct cafe_ep_inst *inst2 =
			SESSION_INST_LIST_ENTRY(list_item);
		if (inst2->spec->input_state_change &&
		    inst2->input_ctx == inst)
			inst2->spec->input_state_change(inst2);
	}
}

void cafe_ep_inst_state_started(struct cafe_ep_inst *inst)
{
	CAFE_ASSERT_LOCKED();
	assert(inst->state == CAFE_EP_INST_STATE_STARTING);
	inst->state = CAFE_EP_INST_STATE_STARTED;

	cafe_input_state_change(inst);

	TRACE(("started cfg %p(%s)\n", inst, inst->spec->name));

	if (inst->session->start_notifier)
		inst->session->start_notifier(inst);
}

void cafe_ep_inst_state_stopped(struct cafe_ep_inst *inst)
{
	CAFE_ASSERT_LOCKED();

	if (inst->state == CAFE_EP_INST_STATE_STOPPED)
		return;
	assert(inst->state == CAFE_EP_INST_STATE_STOPPING ||
	       inst->state == CAFE_EP_INST_STATE_STARTING ||
	       inst->state == CAFE_EP_INST_STATE_STARTED);
	inst->state = CAFE_EP_INST_STATE_STOPPED;

	cafe_input_state_change(inst);

	TRACE(("stopped cfg %p(%s)\n", inst, inst->spec->name));

	if (inst->session->stop_notifier)
		inst->session->stop_notifier(inst);
}

void cafe_dispatch_event_unlocked(const struct cafe_event_action_list *actions,
				  void **event_data)
{
	struct cafe_event_action *action;
	unsigned int i;

	assert(event_data);
	if (actions == NULL)
		return;
	for (i = 0; i < actions->ind; i++) {
		action = actions->list + i;
		action->handler(action->info, event_data);
	}
}

void cafe_dispatch_event(cafe_eal_handle_t *handle,
			 void **event_data)
{
	int rc;
	struct cafe_event_action_list *actions;
	cafe_eal_handle_t unlock_handle;

	rc = cafe_event_action_list_read_lock(handle, &unlock_handle, &actions);
	if (rc != CAFE_OK)
		return;
	if (actions != NULL)
		cafe_dispatch_event_unlocked(actions, event_data);
	cafe_event_action_list_read_unlock(unlock_handle);
}

void cafe_dispatch_error(struct cafe_ep_inst *inst, int error)
{
	if (inst->session->error_notifier)
		inst->session->error_notifier(inst, error);
}

void cafe_dispatch_status(struct cafe_ep_inst *inst, enum cafe_log_level l,
			  const char *msg)
{
	if (inst->session->status_notifier) {
		inst->session->status_notifier(inst, l, msg);
		return;
	}

	if (l == CAFE_LOG_ERROR) {
		(void)fprintf(stderr, "ERROR: %s\n", msg);
		return;
	}

	if (l == CAFE_LOG_WARNING)
		(void)fputs("WARNING: ", stdout);
	else if (l == CAFE_LOG_INFO) {
		if (cafe_trace <= 0)
			return;
		(void)fputs("INFO: ", stdout);
	}
	(void)fputs(msg, stdout);
	(void)fputc('\n', stdout);
}

struct cafe_event_action_info *cafe_ep_inst_event_action_info_find(
	const struct cafe_ep_inst *inst,
	const struct cafe_event_descr *event)
{
	unsigned int i;

	for (i = 0; i < inst->event_action_data_list.ind; i++) {
		struct cafe_event_action_info *event_info =
			inst->event_action_data_list.list[i];
		if (event_info->event == event)
			return event_info;
	}
	return NULL;
}

static int cafe_ep_inst_event_action_info_add(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	unsigned int *odmap,
	void *impl_data,
	struct cafe_event_action_info **result)
{
	struct cafe_event_action_info_list *event_info_list =
		&inst->event_action_data_list;
	struct cafe_event_action_info *event_info;

	if (event_info_list->ind == event_info_list->max) {
		struct cafe_event_action_info **new_list;
		if (event_info_list->ind == 0) {
			event_info_list->max = 1;
		} else {
			event_info_list->max *= 2;
		}
		new_list = CAFE_REALLOC(
			inst->allocator, event_info_list->list,
			event_info_list->max * sizeof *new_list);
		if (new_list == NULL) {
			CAFE_FREE(inst->allocator, odmap);
			return CAFE_NOMEM;
		}
		event_info_list->list = new_list;
	}
	event_info = CAFE_ALLOC(inst->allocator, sizeof *event_info);
	if (event_info == NULL) {
		CAFE_FREE(inst->allocator, odmap);
		return CAFE_NOMEM;
	}
	event_info->inst = inst;
	event_info->event = event;
	event_info->odmap = odmap;
	event_info->impl_data = impl_data;
	event_info->sink_actions = CAFE_EAL_HANDLE_NULL;
	event_info_list->list[event_info_list->ind++] = event_info;
	*result = event_info;
	return CAFE_OK;
}

int cafe_ep_inst_event_action_info_get(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_info **result)
{
	unsigned int j;
	unsigned int *odmap;
	void *impl_data;
	struct cafe_event_action_info *event_info;

	CAFE_ASSERT_LOCKED();
	assert(inst);
	assert(event);
	assert(event->num_types > 0);
	assert(result);

	/* Reuse existing if there is one */
	event_info = cafe_ep_inst_event_action_info_find(inst, event);
	if (event_info != NULL) {
		*result = event_info;
		return CAFE_OK;
	}

	/* WRS_TODO: we should not be called for instances without any
	 * input events, but for now we workaround it until vxworks
	 * and other specs are fixed. */
	if (inst->spec->n_input_events == 0)
		return cafe_ep_inst_event_action_info_add(
			inst, event, NULL, NULL, result);

	/* Find matching input event */

	/* Use event->num_types as a proxy for the input event
	 * num_types since we know it is greater or equal. */
	odmap = CAFE_ALLOC(inst->allocator, event->num_types * sizeof *odmap);
	if (odmap == NULL)
		return CAFE_NOMEM;
	odmap[0] = 0;
	for (j = 0; j < inst->spec->n_input_events; j++) {
		unsigned int k;
		struct cafe_value_type *type;
		struct cafe_event_descr *input_event = inst->input_events[j];

		if (event->num_types < input_event->num_types)
			continue;
		type = input_event->types[0];
		if (!cafe_type_based_on(event->types[0], type) &&
		    type != inst->session->wildcard_type)
			continue;
		for (k = 1; k < input_event->num_types; k++) {
			unsigned int l;

			type = input_event->types[k];
			for (l = 1; l < event->num_types; l++) {
				if (cafe_type_based_on(event->types[l], type) ||
				    type == inst->session->wildcard_type)
					break;
			}
			if (l >= event->num_types)
				break;
			odmap[k] = l;
		}
		if (k >= input_event->num_types)
			break;
	}
	if (j >= inst->spec->n_input_events) {
		CAFE_FREE(inst->allocator, odmap);
		*result = NULL;
		return CAFE_OK;
	}
	impl_data = inst->spec->input_events[j].impl_data;

	return cafe_ep_inst_event_action_info_add(
		inst, event, odmap, impl_data, result);
}

int cafe_ep_inst_event_action_info_get2(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_descr *input_event,
	void *impl_data,
	struct cafe_event_action_info **result)
{
	unsigned int k;
	unsigned int *odmap;
	struct cafe_value_type *type;
	struct cafe_event_action_info *event_info;

	CAFE_ASSERT_LOCKED();
	assert(inst);
	assert(event);
	assert(event->num_types > 0);
	assert(input_event);
	assert(input_event->num_types > 0);
	assert(result);

	/* Reuse existing if there is one */
	event_info = cafe_ep_inst_event_action_info_find(inst, event);
	if (event_info != NULL) {
		*result = event_info;
		return CAFE_OK;
	}

	type = input_event->types[0];
	if (!cafe_type_based_on(event->types[0], type))
		return CAFE_ERROR;

	odmap = CAFE_ALLOC(inst->allocator,
			   (input_event->num_types + 1) * sizeof *odmap);
	if (odmap == NULL)
		return CAFE_NOMEM;
	odmap[0] = 0;

	for (k = 1; k < input_event->num_types; k++) {
		unsigned int l;

		type = input_event->types[k];
		for (l = 1; l < event->num_types; l++) {
			if (cafe_type_based_on(event->types[l], type))
				break;
		}
		if (l < event->num_types)
			odmap[k] = l;
		else
			odmap[k] = CAFE_ODMAP_NONE;
	}
	if (k < input_event->num_types) {
		CAFE_FREE(inst->allocator, odmap);
		*result = NULL;
		return CAFE_OK;
	}
	odmap[k] = CAFE_ODMAP_LAST;

	return cafe_ep_inst_event_action_info_add(
		inst, event, odmap, impl_data, result);
}

int cafe_add_event_action(
	struct cafe_ep_inst *inst,
	struct cafe_event_action_list *actions,
	CAFE_ACTION_HANDLER handler,
	struct cafe_event_action_info *info)
{
	struct cafe_event_action *p;

	CAFE_ASSERT_LOCKED();
	assert(inst);
	assert(actions);
	assert(handler);
	assert(info);

	if (actions->ind == actions->max) {
		if (actions->max == 0) {
			actions->max = 1;
		} else {
			actions->max *= 2;
		}
		actions->list = CAFE_REALLOC(
			inst->allocator, actions->list,
			actions->max * sizeof *actions->list);
		if (actions->list == NULL)
			return CAFE_NOMEM;
	}
	p = &actions->list[actions->ind++];
	p->handler = handler;
	p->info = info;
	return CAFE_OK;
}

int cafe_ep_inst_get_sink_event_actions(
	struct cafe_ep_inst *source,
	struct cafe_event_action_list *actions)
{
	struct cafe_list_head *list_item;

	CAFE_ASSERT_LOCKED();
	assert(source);
	assert(actions);

	cafe_list_foreach(list_item, &source->sink_list_head) {
		struct cafe_ep_inst *sink =
			SINK_LIST_NODE_ENTRY(list_item)->sink;
		if (sink->spec->get_input_event_actions != NULL) {
			int rc = sink->spec->get_input_event_actions(
				sink, actions->event, actions);
			if (rc != CAFE_OK)
				return rc;
		}
	}
	return CAFE_OK;
}


static
struct cafe_vm_expr *resolve_expr_arg(
	struct cafe_value_context *context,
	struct cafe_vm_expr *e)
{
	struct cafe_event_descr *event = context->event;

	assert(e->kind == CAFE_EXPR_ARG);
	if (event != NULL && e->u.ull < event->num_types) {
		if (context->odmap) {
			e->u.ull = context->odmap[e->u.ull];
			assert(e->u.ull  < event->num_types);
		}
		e->type = event->types[e->u.ull];
	}
	return e;
}


/**
 *  Resolve orthogonal data entries for an enabled event instance.  This
 * routine returns the first odata entry index which cannot be satisfied with
 * the current available type provided by the EP spec, and isn't a filter type.
 * Each filter entry is resolved as encountered in the context of the current
 * event, and CAFE_OK is returned iff all types up to the returned index are
 * successfully resolved.
 */
static
int resolve_event_odata(const struct cafe_value_type *available_odata,
			const struct cafe_event_descr *event,
			unsigned int *odata_idx,
			struct cafe_value_context *context)
{
	unsigned int i;

	for (i = *odata_idx; i < event->num_types; ++i) {
		struct cafe_vm_expr *e;
		if (cafe_type_based_on(event->types[i], available_odata))
			continue;

		if (event->types[i]->kind != CAFE_TYPE_FILTER)
			goto done;
		e = cafe_expr_resolve(context, event->types[i]->ident->expr);
		assert(e != NULL);
		if (e->type == NULL) {
			cafe_error_set(NULL, CAFE_ERROR, "error resolving filter expression entry: %s", event->types[i]->type_str);
			return CAFE_ERROR;
		}
	}

done:
	*odata_idx = i;
	return CAFE_OK;
}


/**
 *  Flag the given instance to enable the desired event.  Returns CAFE_OK if
 * the event and all its orthogonal data can be enabled.
 */
int cafe_ep_inst_enable_output_event(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event)
{
	unsigned int i, j;
	struct cafe_value_context *context;
	void *impl_data = NULL;
	struct cafe_ep_spec_output_evt *output_event = NULL;
	int rc;

	CAFE_ASSERT_LOCKED();
	assert(inst != NULL);
	assert(event != NULL);

	/* first make sure the given instance supports the requested output
	   event type */
	for (i = 0; i < inst->spec->n_output_events; ++i)
		if (cafe_type_based_on(event->types[0], inst->output_events[i])) {
			output_event = &inst->spec->output_events[i];
			impl_data = inst->spec->output_events[i].impl_data;
			break;
		}
	if (i == inst->spec->n_output_events)
		return CAFE_ERROR;

	/* now make sure the instance supports creation of any
	   required orthogonal data and that they appear in the same
	   order as the spec */
	context = cafe_value_context_create(inst->allocator);
	assert(context != NULL);
	context->event = event;
	context->ns = inst->session->ns;
	context->resolve_expr_arg = resolve_expr_arg;
	context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
	rc = CAFE_OK;
	j = 1;
	if (inst->spec->n_output_odata_sets == 0 ||
	    inst->spec->output_odata_sets == NULL) {
		for (i = 0; i < inst->spec->n_output_odata; ++i) {
			if ((rc != CAFE_OK) || (j >= event->num_types))
				break;
			rc = resolve_event_odata(inst->output_odata[i], event,
						 &j, context);
		}
	} else {
		struct cafe_ep_spec_output_od_set *od_set;
		od_set = &inst->spec->output_odata_sets[output_event->od_set_idx];
		for (i = 0; i < (unsigned int)od_set->length; ++i) {
			if ((rc != CAFE_OK) || (j >= event->num_types))
				break;
			rc = resolve_event_odata(inst->output_odata[od_set->od_items[i]],
						 event, &j, context);
		}
	}
	cafe_value_context_delete(context);
	
	if (rc != CAFE_OK)
		return rc;
	if (j < event->num_types)
		return CAFE_ERROR;

	for (i = 0; i < inst->enabled_output_events.ind; i++) {
		if (event == inst->enabled_output_events.list[i].event) {
			assert(inst->enabled_output_events.list[i].impl_data == impl_data);
			return CAFE_OK;
		}
	}
	if (i == inst->enabled_output_events.max) {
		if (i == 0) {
			inst->enabled_output_events.max = 1;
		} else {
			inst->enabled_output_events.max *= 2;
		}
		inst->enabled_output_events.list = CAFE_REALLOC(
			inst->allocator, inst->enabled_output_events.list,
			inst->enabled_output_events.max *
			sizeof *inst->enabled_output_events.list);
		if (inst->enabled_output_events.list == NULL)
			return CAFE_NOMEM;
	}
	inst->enabled_output_events.list[i].event = event;
	inst->enabled_output_events.list[i].impl_data = impl_data;
	inst->enabled_output_events.ind = i + 1;
	inst->session->output_event_action_change_count++;
	return CAFE_OK;
}

int cafe_ep_inst_disable_output_event(
	struct cafe_ep_inst *inst,
	const struct cafe_event_descr *event)
{
	unsigned int i, ind;
	struct cafe_enabled_event *list;

	CAFE_ASSERT_LOCKED();

	ind = inst->enabled_output_events.ind;
	list = inst->enabled_output_events.list;

	for (i = 0; i < ind; i++) {
		if (event == list[i].event) {
			while (i < (ind - 1)) {
				list[i] = list[i + 1];
				++i;
			}
			inst->enabled_output_events.ind = ind - 1;
			inst->session->output_event_action_change_count++;
			break;
		}
	}
	return CAFE_OK;
}


/**
 *  Dispatch the given event to the given configuration instance.  Returns
 * CAFE_ERROR if the instance does not allow event injection, otherwise,
 * returns the result of the event dispatch action.
 */
int cafe_ep_inst_inject_event(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	void *data[])
{
	if (inst->spec->inject_event == NULL)
		return CAFE_ERROR;

	return inst->spec->inject_event(inst, event, data);
}

struct cafe_ep_inst *cafe_ep_inst_iter_first(
	struct cafe_instrument *session,
	struct cafe_ep_inst_iter *iter)
{
	struct cafe_list_head *list_head, *list_entry;

	CAFE_ASSERT_LOCKED();
	assert(iter);

	iter->list_head = list_head = &session->inst_list_head;
	iter->list_entry = list_entry = list_head->next;

	if (list_head == list_entry)
		return NULL;

	return SESSION_INST_LIST_ENTRY(list_entry);
}

struct cafe_ep_inst *cafe_ep_inst_iter_next(
	struct cafe_ep_inst_iter *iter)
{
	struct cafe_list_head *list_head, *list_entry;

	CAFE_ASSERT_LOCKED();
	assert(iter);

	list_head = iter->list_head;
	list_entry = iter->list_entry;
	iter->list_entry = list_entry = list_entry->next;

	if (list_head == list_entry)
		return NULL;

	return SESSION_INST_LIST_ENTRY(list_entry);
}
