/* ep_worker.c - Worker thread library for event processors that needs it. */

/* Copyright (c) 2011-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
06jun14,tle  Fix handling of worker abort (WB4-3650)
24feb14,sdt  Disable unexpected NO_EFFECT warnings

DESCRIPTION

*/

#include <stdio.h>
#include <errno.h>
#if defined(HAVE_CAFE_CONFIG_H)
#  include <cafe/config.h>
#endif
#if defined(HAVE_PR_SET_NAME)
#  include <sys/prctl.h>
#endif
#include <cafe/cafe.h>
#include <cafe/atomic.h>
#include <cafe/ep_worker.h>
#include "cafe_os.h"
#include "cafe_assert.h"


enum ep_worker_state {
	state_stopped,
	state_starting,
	state_started,
	state_stopping,
	state_exiting,
	state_exited
};

struct cafe_ep_worker {
	struct cafe_ep_inst *inst;
	void (*client_func)(void *client_data);
	void *client_data;
	pthread_t tid;
	cafe_atomic_val state;
	bool aborted;
#if defined(HAVE_PR_SET_NAME)
	char thread_name[0];
#endif
};

static void *ep_worker_handler(void *arg)
{
	struct cafe_ep_worker *worker = arg;
	struct cafe_ep_inst *inst = worker->inst;

#if defined(HAVE_PR_SET_NAME)
	if (prctl(PR_SET_NAME, worker->thread_name) != 0)
		perror("Error setting thread name");
#endif
	for(;;) {
		if (cafe_atomic_get(&worker->state) != state_started) {
			CAFE_LOCK();
			switch(cafe_atomic_get(&worker->state)) {
			case state_stopped:
				if (cafe_notify_wait() != CAFE_OK) {
					break;
				}
				CAFE_UNLOCK();
				continue;

			case state_starting:
				if (worker->aborted)
					goto stop_and_exit;
				cafe_atomic_set(&worker->state, state_started);
				cafe_notify_signal();
				cafe_ep_inst_state_started(inst);
				CAFE_UNLOCK();
				continue;

			case state_stopping:
				if (worker->aborted)
					goto stop_and_exit;
				cafe_atomic_set(&worker->state, state_stopped);
				cafe_notify_signal();
				cafe_ep_inst_state_stopped(inst);
				CAFE_UNLOCK();
				continue;

			stop_and_exit:
				cafe_ep_inst_state_stopped(inst);
			case state_exiting:
				cafe_atomic_set(&worker->state, state_exited);
				cafe_notify_signal();
				CAFE_UNLOCK();
				return 0;

			case state_started:
			case state_exited:
			default:
				/* coverity[array_null] */
				assert(!"bad state in ep_worker_handler");
				break;
			}
			worker->aborted = true;
			CAFE_UNLOCK();
			continue;
		}
		worker->client_func(worker->client_data);
	}
}

int cafe_ep_worker_create(
	const char *thread_name,
	struct cafe_ep_inst *inst,
	void (*client_func)(void *client_data),
	void *client_data,
	struct cafe_ep_worker **result)
{
	struct cafe_ep_worker *worker;
	size_t len;
	int rc;

	assert(inst != NULL);
	assert(client_func != NULL);
	assert(result != NULL);
#if defined(HAVE_PR_SET_NAME)
	char tmp_buf[256];
	if (thread_name == NULL) {
		snprintf(tmp_buf, sizeof(tmp_buf), "%s-worker",
			 inst->spec->name);
		thread_name = tmp_buf;
	}
	len = strlen(thread_name) + 1;
#else
	len = 0;
#endif
	worker = CAFE_ALLOC(inst->allocator, sizeof(*worker) + len);
	if (worker == NULL)
		return CAFE_NOMEM;

	worker->inst = inst;
	worker->client_func = client_func;
	worker->client_data = client_data;
	worker->aborted = false;
#if defined(HAVE_PR_SET_NAME)
	memcpy(&worker->thread_name, thread_name, len);
#endif
	cafe_atomic_set(&worker->state, state_stopped);
	rc = pthread_create(&worker->tid, NULL, ep_worker_handler, worker);
	if (rc != 0) {
		CAFE_FREE(inst->allocator, worker);
		return CAFE_ERROR;
	}
	*result = worker;
	return CAFE_OK;
}

void cafe_ep_worker_delete(
	struct cafe_ep_worker *worker)
{
	int err;
	enum ep_worker_state state = cafe_atomic_get(&worker->state);

	CAFE_ASSERT_LOCKED();

	if (state != state_exited) {
		/*
		 * Join thread only if not yet exited,
		 * otherwise pthread_join() fails on windows.
		 */
		cafe_ep_worker_abort(worker);

		CAFE_UNLOCK();
		err = pthread_join(worker->tid, NULL);
		if (err != 0)
			fprintf(stderr, "Error joining worker thread: %s\n",
				strerror(err));
		CAFE_LOCK();
	}

	CAFE_FREE(worker->inst->allocator, worker);
}

int cafe_ep_worker_start(
	struct cafe_ep_worker *worker)
{
	enum ep_worker_state state = cafe_atomic_get(&worker->state);
	CAFE_ASSERT_LOCKED();

	if (state == state_started || state == state_starting)
		return CAFE_OK;

	if (state != state_stopped || worker->aborted)
		return CAFE_ERROR;

	cafe_atomic_set(&worker->state, state_starting);
	cafe_notify_signal();
	return CAFE_OK;
}

int cafe_ep_worker_stop(
	struct cafe_ep_worker *worker)
{
	enum ep_worker_state state = cafe_atomic_get(&worker->state);
	CAFE_ASSERT_LOCKED();

	if (state == state_stopped || state == state_stopping)
		return CAFE_OK;

	if (state != state_started)
		return CAFE_ERROR;

	cafe_atomic_set(&worker->state, state_stopping);
	cafe_notify_signal();
	return CAFE_OK;
}

void cafe_ep_worker_abort(
	struct cafe_ep_worker *worker)
{
	enum ep_worker_state state = cafe_atomic_get(&worker->state);
	CAFE_ASSERT_LOCKED();

	if (state == state_stopped)
		cafe_atomic_set(&worker->state, state_exiting);
	else if (state == state_started)
		cafe_atomic_set(&worker->state, state_stopping);

	worker->aborted = true;
	cafe_notify_signal();
}

int cafe_ep_worker_wait(
	struct cafe_ep_worker *worker)
{
	enum ep_worker_state state = cafe_atomic_get(&worker->state);

	CAFE_ASSERT_LOCKED();

	while (state == state_starting ||
	       state == state_stopping ||
	       state == state_exiting)
	{
		int rc = cafe_notify_wait();
		if (rc != CAFE_OK)
			return rc;
		state = cafe_atomic_get(&worker->state);
	}
	return CAFE_OK;
}

bool cafe_ep_worker_isrunning(
	struct cafe_ep_worker *worker)
{
	return cafe_atomic_get(&worker->state) == state_started;
}
