/*
 * Copyright (c) 2007-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/


#include <cafe/stdio.h>
#include "cafe_assert.h"
#include <errno.h>
#include "leb128.h"

/**************	Local data, types, fns and macros ***************/

#define SAFE_SIZE(BYTES)	(((BYTES)*8)/7)
#define BORDER_SIZE(BYTES)	(((BYTES)*8+6)/7)
#define UBORDER_ERR(BYTES, LAST) ((c >> (((BYTES)*8)%7)) != 0)
#define SBORDER_ERR(BYTES, LAST) ((c >> (((BYTES)*8)%7)) != 0 && \
				  (c >> (((BYTES)*8)%7)) != -1)
#define	UOVERFLOW(SIZE, VAL, LAST) ((SIZE) > SAFE_SIZE(sizeof (VAL)) && \
				    (SAFE_SIZE(sizeof (VAL)) == BORDER_SIZE(sizeof (VAL)) || \
				     ((SIZE) == BORDER_SIZE(sizeof (VAL)) && \
				      UBORDER_ERR(sizeof (VAL), c))))
#define	SOVERFLOW(SIZE, VAL, LAST) ((SIZE) > SAFE_SIZE(sizeof (VAL)) && \
				    (SAFE_SIZE(sizeof (VAL)) == BORDER_SIZE(sizeof (VAL)) || \
				     ((SIZE) == BORDER_SIZE(sizeof (VAL)) && \
				      SBORDER_ERR(sizeof (VAL), c))))

/**************	Implementation of exported functions ************/


void *cafe_leb128_mread_ul(const void *buf, unsigned long *v)
{
	unsigned char *p = (unsigned char *)buf;
	int c;
	int sz = 0;
	int shift = 0;
	unsigned long n = 0;

	do {
		c = *p++;
		sz++;
		n |= (((unsigned long)c & 127) << shift);
		shift += 7;
	} while(c & 128);
	if(UOVERFLOW(sz, n, c)) return NULL;
	*v = n;
	return p;
}


void *cafe_leb128_mread_ull(const void *buf, unsigned long long *v)
{
	unsigned char *p = (unsigned char *)buf;
	int c;
	int sz = 0;
	int shift = 0;
	unsigned long long n = 0;

	do {
		c = *p++;
		sz++;
		n |= (((unsigned long long)c & 127) << shift);
		shift += 7;
	} while(c & 128);
	if(UOVERFLOW(sz, n, c)) return NULL;
	*v = n;
	return p;
}


/**
 * Similar to cafe_leb128_mread_ull(), but guards against reading past the end
 * of valid memory.
 */
void *cafe_leb128_mread_ull_guarded(const void *buf, size_t max_size,
				    unsigned long long *v)
{
	unsigned char *p = (unsigned char *)buf;
	int c;
	int sz = 0;
	int shift = 0;
	unsigned long long n = 0;

	do {
		if (sz == max_size)
			return NULL;
		c = *p++;
		sz++;
		n |= (((unsigned long long)c & 127) << shift);
		shift += 7;
	} while(c & 128);
	if(UOVERFLOW(sz, n, c)) return NULL;
	*v = n;
	return p;
}


void *cafe_leb128_mread_l(const void *buf, long *v)
{
	unsigned char *p = (unsigned char *)buf;
	int c;
	int sz = 0;
	int shift = 0;
	long n = 0;

	for(;;) {
		c = *p++;
		sz++;
		if((c & 128) == 0) break;
		n |= (((long)c & 127) << shift);
		shift += 7;
	}
	if(c & 64) {
		/* Sign-extend */
		c |= -64;
	}
	n |= (((long)c) << shift);
	if(SOVERFLOW(sz, n, c)) return NULL;
	*v = n;
	return p;
}


void *cafe_leb128_mread_ll(const void *buf, long long *v)
{
	unsigned char *p = (unsigned char *)buf;
	int c;
	int sz = 0;
	int shift = 0;
	long long n = 0;

	for(;;) {
		c = *p++;
		sz++;
		if((c & 128) == 0) break;
		n |= (((long long)c & 127) << shift);
		shift += 7;
	}
	if(c & 64) {
		/* Sign-extend */
		c |= -64;
	}
	n |= (((long long)c) << shift);
	if(SOVERFLOW(sz, n, c)) return NULL;
	*v = n;
	return p;
}


int cafe_leb128_mwrite_ll(int8_t *out, long long n)
{
	unsigned int sz = 1;

	while(n < -64 || n >= 64) {
		out[sz - 1] = (int8_t) ((n & 127) + 128);
		n >>= 7;
		sz++;
	}
	out[sz - 1] = (int8_t) (n & 0x7f);
	return sz;
}


int cafe_leb128_mwrite_ull(int8_t *out, unsigned long long n)
{
	unsigned int sz = 1;

	while(n >= 128) {
		out[sz - 1] = (int8_t) ((n & 127) + 128);
		n >>= 7;
		sz++;
	}
	out[sz - 1] = (int8_t) (n & 0x7f);
	return sz;
}


int cafe_leb128_fread_ul(FILE *f, unsigned long *v)
{
	int c;
	int sz = 0;
	int shift = 0;
	unsigned long n = 0;

	do {
		if((c = fgetc_unlocked(f)) == EOF) {
			return sz == 0 ? LEB128ERR_NONUM : LEB128ERR_INVAL;
		}
		sz++;
		n |= (((unsigned long)c & 127) << shift);
		shift += 7;
	} while(c & 128);
	if(UOVERFLOW(sz, n, c)) return LEB128ERR_OVERFLOW;
	*v = n;
	return sz;
}


int cafe_leb128_fread_ull(FILE *f, unsigned long long *v)
{
	int c;
	int sz = 0;
	int shift = 0;
	unsigned long long n = 0;

	do {
		if((c = fgetc_unlocked(f)) == EOF) {
			return sz == 0 ? LEB128ERR_NONUM : LEB128ERR_INVAL;
		}
		sz++;
		n |= (((unsigned long long)c & 127) << shift);
		shift += 7;
	} while(c & 128);
	if(UOVERFLOW(sz, n, c)) return LEB128ERR_OVERFLOW;
	*v = n;
	return sz;
}


int cafe_leb128_fread_l(FILE *f, long *v)
{
	int c;
	int sz = 0;
	int shift = 0;
	long n = 0;

	for(;;) {
		if((c = fgetc_unlocked(f)) == EOF) {
			return sz == 0 ? LEB128ERR_NONUM : LEB128ERR_INVAL;
		}
		sz++;
		if((c & 128) == 0) break;
		n |= (((long)c & 127) << shift);
		shift += 7;
	}
	if(c & 64) {
		/* Sign-extend */
		c |= -64;
	}
	n |= (((long)c) << shift);
	if(SOVERFLOW(sz, n, c)) return LEB128ERR_OVERFLOW;
	*v = n;
	return sz;
}


int cafe_leb128_fread_ll(FILE *f, long long *v)
{
	int c;
	int sz = 0;
	int shift = 0;
	long long n = 0;

	for(;;) {
		if((c = fgetc_unlocked(f)) == EOF) {
			return sz == 0 ? LEB128ERR_NONUM : LEB128ERR_INVAL;
		}
		sz++;
		if((c & 128) == 0) break;
		n |= (((long long)c & 127) << shift);
		shift += 7;
	}
	if(c & 64) {
		/* Sign-extend */
		c |= -64;
	}
	n |= (((long long)c) << shift);
	if(SOVERFLOW(sz, n, c)) return LEB128ERR_OVERFLOW;
	*v = n;
	return sz;
}


int cafe_leb128_fwrite_ull(FILE *f, unsigned long long n)
{
	int sz = 1;

	if (f != NULL) {
		while(n >= 128) {
			(void)fputc_unlocked(((int)n & 127) + 128, f);
			n >>= 7;
			sz++;
		}
		(void)fputc_unlocked((int)n, f);
	} else {
		while(n >= 128) {
			n >>= 7;
			sz++;
		}
	}
	return sz;
}


int cafe_leb128_fwrite_ll(FILE *f, long long n)
{
	int sz = 1;

	if(f != NULL) {
		while(n < -64 || n >= 64) {
			(void)fputc_unlocked(((int)n & 127) + 128, f);
			n >>= 7;
			sz++;
		}
		(void)fputc_unlocked((int)n & 0x7f, f);
	} else {
		while(n < -64 || n >= 64) {
			n >>= 7;
			sz++;
		}
	}
	return sz;
}


int cafe_leb128_swrite_ull(struct cafe_output_stream *out, unsigned long long n)
{
	int sz = 1;

	if (out != NULL) {
		while (n >= 128) {
			out->write(out, (int) ((n & 127) + 128));
			n >>= 7;
			++sz;
		}
		out->write(out, (int) n);
	} else {
		while (n >= 128) {
			n >>= 7;
			++sz;
		}
	}
	return sz;
}


int cafe_leb128_swrite_ll(struct cafe_output_stream *out, long long n)
{
	int sz = 1;

	if (out != NULL) {
		while (n < -64 || n >= 64) {
			out->write(out, (int) ((n & 127) + 128));
			n >>= 7;
			++sz;
		}
		out->write(out, (int) (n & 0x7f));
	} else {
		while (n < -64 || n >= 64) {
			n >>= 7;
			++sz;
		}
	}
	return sz;
}
