/* log_stream.c - Collects utilities around serializing event data. */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

#include "cafe_assert.h"
#include <limits.h>             /* for SSIZE_MAX */
#include <string.h>             /* for memcpy() */
#include <ctype.h>              /* for isdigit() */
#include <cafe/cafe.h>
#include <cafe/log_stream.h>
#include <cafe/dict.h>
#include <cafe/stdio.h>
#include <time.h>
#include "cafe_os.h"  /* for portable nanosleep() implementations */
#include "leb128.h"
#include "time_utils.h"


#if (CAFE_OS_VXWORKS)
    #if (_BYTE_ORDER == _LITTLE_ENDIAN)
        #define CAFE_LITTLE_ENDIAN
    #else
        #define CAFE_BIG_ENDIAN
    #endif
#else
    #if defined(__LITTLE_ENDIAN__)
        #define CAFE_LITTLE_ENDIAN
    #elif defined(__BIG_ENDIAN__)
        #define CAFE_BIG_ENDIAN
    #else
        #error "Host endianness not defined"
    #endif
#endif


#ifndef SSIZE_MAX
# define SSIZE_MAX ((ssize_t) (SIZE_MAX / 2))
#endif


/* Bytestream header */
static const char HEADER[] = {0x7f, 'C', 'a', 'f', 'e', ' '};


/* Version of the cafe event stream format used for new output streams */
const unsigned int cafe_slog_version = 0;


#define CAFE_LOG_STREAM_MAP_DICT					\
	"S[cafe_log_event_map,k[key,L]v[event_id,Q]]"			\
	"D[t[cafe_log_event_map]]"

struct cafe_log_event_map {
	unsigned long key;
	unsigned long long event_id;
};


struct cafe_slog_encoded_inst {
	struct cafe_instrument *session;
	struct cafe_output_stream *out;
	unsigned long event_id;
	struct cafe_dict_struct dict;
	struct cafe_event_descr *metadata_event;
};


typedef union {
	struct {
#if defined(CAFE_LITTLE_ENDIAN)
		uint32_t m1:23;
		uint32_t exp:8;
		uint32_t sign:1;
#else
		uint32_t sign:1;
		uint32_t exp:8;
		uint32_t m1:23;
#endif
	} b;
	float f;
} ieee_bits_f;

typedef union {
	struct {
#if defined(CAFE_LITTLE_ENDIAN)
		uint32_t m2:32;
		uint32_t m1:20;
		uint32_t exp:11;
		uint32_t sign:1;
#else
		uint32_t sign:1;
		uint32_t exp:11;
		uint32_t m1:20;
		uint32_t m2:32;
#endif
	} b;
	double d;
} ieee_bits_d;


static int fread_be(FILE *in, void *p, size_t l)
{
	int c;
	unsigned char *cp = p;

	while(l-- > 0) {
		c = fgetc_unlocked(in);
		if (c == EOF)
			return -1;
#if defined(CAFE_LITTLE_ENDIAN)
		cp[l] = (unsigned char) c;
#else
		*cp++ = (unsigned char) c;
#endif
	}
	return 0;
}


static const char *mread_be(const char *in, void *p, size_t l)
{
	unsigned char *cp = p;

	while(l-- > 0) {
#if defined(CAFE_LITTLE_ENDIAN)
		cp[l] = *in++;
#else
		*cp++ = *in++;
#endif
	}
	return in;
}


static ssize_t fread_float(FILE *in, float *arg)
{
	ieee_bits_f v;

	if (0 != fread_be(in, &v, sizeof v))
		return -1;
	*arg = v.f;
	return sizeof v;
}


static ssize_t fread_double(FILE *in, double *arg)
{
	ieee_bits_d v;

	if (0 != fread_be(in, &v, sizeof v))
		return -1;
	*arg = v.d;
	return sizeof v;
}


static const char *mread_float(const char *in, float *arg)
{
	ieee_bits_f v;

	in = mread_be(in, &v, sizeof v);
	*arg = v.f;
	return in;
}


static const char *mread_double(const char *in, double *arg)
{
	ieee_bits_d v;

	in = mread_be(in, &v, sizeof v);
	*arg = v.d;
	return in;
}


static size_t swrite_be(struct cafe_output_stream *out, const void *p, size_t l)
{
	char c;
	const unsigned char *cp = p;

	if (out == NULL)
		return l;

	while (l-- > 0) {
#if defined(__LITTLE_ENDIAN__) || _BYTE_ORDER==_LITTLE_ENDIAN
		c = cp[l];
#else
		c = *cp++;
#endif
		out->write(out, c);
	}
	return l;
}


static ssize_t swrite_float(struct cafe_output_stream *out, const float *arg)
{
	ieee_bits_f v;

	v.f = *arg;
	return swrite_be(out, &v, sizeof v);
}


static ssize_t swrite_double(struct cafe_output_stream *out, const double *arg)
{
	ieee_bits_d v;

	v.d = *arg;
	return swrite_be(out, &v, sizeof v);
}


/**
 *  Calculate the number of bytes needed to encode the static portion of this
 * event.  This includes the event bookkeeping and its parameter values.  It
 * does NOT include any variable length data, such as strings or other
 * variable-size arrays.
 */
size_t calc_entry_size_static(const struct cafe_event_descr *event)
{
	unsigned int i;

	/* account for the event id, sequence id, and record size */
	size_t size = sizeof(long) + sizeof(long) + sizeof(size_t);
	assert(event);

	/* add in size of parameter payload, if there is one */
	for (i = 0; i < event->num_types; ++i)
		size += event->types[i]->size;

	return size;
}


/**
 *  Calculate the number of bytes needed to encode any variable-length data
 * carried by the given event.
 */
size_t calc_entry_size_variable(const struct cafe_event_descr *event,
				void *data[])
{
	unsigned int i;
	size_t size = 0;

	for (i = 0; i < event->num_types; ++i) {
		struct cafe_vm_attr *attr;
		const char *ptr = (char *) data[i];

		cafe_type_attr_foreach(attr, event->types[i]) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			if (CAFE_TYPE_STRING == cafe_attr_type(attr)->kind) {
				const char *string_data =
					*(char **)(ptr + attr->offset);
				if (string_data != NULL)
					/* space for NULL when unpacking */
					size += strlen(string_data) + 1;
			} else if (CAFE_TYPE_BLOB == cafe_attr_type(attr)->kind) {
				size += ((struct cafe_vm_blob*)
					 (ptr + attr->offset))->l;
			} else if (CAFE_TYPE_DYNARRAY == cafe_attr_type(attr)->kind) {
				//td: doesn't handle complex subtypes
				const struct cafe_value_type *element_type =
					cafe_element_type(cafe_attr_type(attr));
				const size_t element_size = element_type->size;
				size += ((struct cafe_vm_dynarray *)
					 (ptr + attr->offset))->l * element_size;
			}
		}
	}
	return size;
}


/**
 *  Write one event's worth of data to the given address.  This happens in two
 * pieces, the statically sized data and then any variable-sized data.  This
 * scheme allows clients to keep all static data consecutive, with
 * pointers to any subsequent variable-sized data.
 */
static
char *write_event_data(char *msg, char *var_ptr,
		       const struct cafe_value_type *type, const void *data)
{
	struct cafe_vm_attr *attr;

	type = cafe_base_type(type);
	assert(type != NULL);

	/* statically sized data first */
	memcpy(msg, data, type->size);

	/* followed by any variable-sized data */
	cafe_type_attr_foreach(attr, type) {
		struct cafe_value_type *attrtype =
			cafe_base_type(cafe_attr_type(attr));
		if (attr->kind != CAFE_ATTR_VAR &&
		    attr->kind != CAFE_ATTR_KEY)
			continue;
		assert(attrtype != NULL);
		switch(attrtype->kind) {
		case CAFE_TYPE_STRING:
		{
			size_t len;
			const char *arg =
				*(char **) ((char *)data + attr->offset);

			/* if arg is the offset of a member in the original,
			   buff_arg is the offset in the copy */

			char **buff_arg = (char **)(msg + attr->offset);

			/* special case: If ptr == NULL, don't copy string data
			   to log */
			if (arg == NULL)
				continue;

			len = strlen(arg) + 1;
			/* copy the string into the log buffer */
			memcpy(var_ptr, arg, len);

			/* update the pointer in the logged arguments to point
			   to the buffered copy of the string */
			*buff_arg = var_ptr;
			var_ptr += len;
			break;
		} /* case CAFE_TYPE_STRING */

		case CAFE_TYPE_BLOB:
                {
			struct cafe_vm_blob *event_arg;
			struct cafe_vm_blob *dst_arg;

                        event_arg = (struct cafe_vm_blob*)
				((char *)data + attr->offset);

			/* if event_arg is the offset of a member in the
			   original, dst_arg is the offset in the copy */

			dst_arg = (struct cafe_vm_blob*) (msg + attr->offset);

			/* copy the blob data into the log buffer */
			memcpy(var_ptr, event_arg->p, event_arg->l);

			/* update the pointer in the logged arguments to point
			   to the buffered copy of the string */
			dst_arg->p = var_ptr;
			var_ptr += event_arg->l;
                        break;
		} /* case CAFE_TYPE_BLOB */

		case CAFE_TYPE_DYNARRAY:
		{
			//td: does not handle complex subtypes; anything that needs pointers further relocated
			const struct cafe_value_type *element_type =
				cafe_element_type(attrtype);
			const size_t element_size = element_type->size;
			struct cafe_vm_dynarray *event_arg;
			struct cafe_vm_dynarray *dst_arg;

                        event_arg = (struct cafe_vm_dynarray *)
				((char *)data + attr->offset);

			/* if event_arg is the offset of a member in the
			   original, dst_arg is the offset in the copy */

			dst_arg = (struct cafe_vm_dynarray *)
				(msg + attr->offset);

			/* copy the array data into the log buffer */
			memcpy(var_ptr, event_arg->p,
			       event_arg->l * element_size);

			/* update the pointer in the logged arguments to point
			   to the buffered copy of the string */
			dst_arg->p = var_ptr;
			var_ptr += event_arg->l * element_size;
                        break;
		} /* case CAFE_TYPE_DYNARRAY */

		case CAFE_TYPE_STRUCT:
		case CAFE_TYPE_DICT:
			(void)printf("Can't handle type %d!\n", attrtype->kind);
		default:
			break;
		}
	} /* for 0 .. vt->elements */
	return var_ptr;
}


/**
 *  Writes the given event to memory starting at the given "msg" address.  The
 * message format is:
 *    message sequence number  <---  msg
 *    event ID
 *    message size
 *    event parameters
 *    event parameter variable data (strings, dynamic arrays, etc.)
 *    beginning of next message <--- return value
 *
 * Where:
 *   - size_total includes all bytes in this record, so (msg + size) will be
 *     the next message in the buffer
 *   - event parameters is a strict (unpacked) copy of the given event's
 *     parameter structure.  This means a pointer to this address can be cast
 *     to (and used as) the event-specific argument struct.
 *   - the "variable data" area contains copies of any data referenced by
 *     pointers from the event parameter.  For example, any strings,
 *     structures, or dynamic arrays will be copied here.  The pointers in the
 *     logged event parameter will be updated to point at these copies.
 *
 *  Returns a pointer to the next unwritten byte.
 */
char *cafe_write_event(char *msg, unsigned long sequence,
		       unsigned long event_handle,
		       const struct cafe_event_descr *event,
		       void *data[],
		       size_t size_static, size_t size_total)
{
	unsigned int i;
	char *var_ptr = msg + size_static;
	char *msg_start = msg;

	assert(msg != 0);
	assert(event != 0);

	/* write sequence */
	*(long *)msg = sequence;
	msg += sizeof(long);

	/* write event ID */
	*(unsigned long *) msg = event_handle;
	msg += sizeof(unsigned long);

	/* write event record size */
	*(size_t *) msg = size_total;
	msg += sizeof(size_t);

	/* write data payload */
	for (i = 0; i < event->num_types; ++i) {
		var_ptr = write_event_data(msg, var_ptr,
					   event->types[i], data[i]);
		msg += event->types[i]->size;
	}
	return msg_start + size_total;
}


/**
 *  Given a pointer to the beginning of an event record (in a message log),
 * populate "sequence" and "event" with the event sequence number and event id
 * respectively. Returns a pointer to the first address after the given event.
 */
const char *cafe_parse_event(const char *msg, unsigned long *sequence,
			     void **event_handle, const void **args)
{
	size_t size;
	const char *orig = msg;
	assert(msg != 0);
	assert(sequence != 0);
	assert(event_handle != 0);
	assert(args != 0);

	/* sequence id is first */
	*sequence = *(unsigned long *)msg;
	msg += sizeof(unsigned long);

	/* event descriptor is next */
	*event_handle = *(void **)msg;
	msg += sizeof(void *);

	/* event record size is next */
	size = *(size_t *) msg;
	msg += sizeof(size_t);

	*args = msg;

	return orig + size;
}


/**
 *
 */
void cafe_release_values(const struct cafe_value_type *vt, void *arg,
			 const struct cafe_allocator *allocator)
{
	if (!vt)
		return;

	assert(allocator);
	vt = cafe_base_type(vt);
	assert(vt != NULL);
	switch (vt->kind) {
	case CAFE_TYPE_STRING:
		CAFE_FREE(allocator, *((char **)arg));
		break;
	case CAFE_TYPE_BLOB:
        {
		struct cafe_vm_blob *blob = (struct cafe_vm_blob *) arg;
		CAFE_FREE(allocator, blob->p);
		break;
        }

	case CAFE_TYPE_DICT:
	{
		struct cafe_dict_struct *dict =
			(struct cafe_dict_struct *) arg;
		cafe_dict_delete(dict);
		break;
	}

	case CAFE_TYPE_STRUCT:
	{
		struct cafe_vm_attr *attr;

		cafe_type_attr_foreach(attr, vt) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			cafe_release_values(cafe_attr_type(attr),
					(char *) arg + attr->offset,
					allocator);
		}
		break;
	}

	case CAFE_TYPE_DYNARRAY:
	{
		struct cafe_vm_dynarray *array =
			(struct cafe_vm_dynarray *) arg;
		const struct cafe_value_type *subtype = cafe_element_type(vt);
		char *trav = (char *) array->p;
		unsigned int i;
		for (i = 0; i < array->l; ++i) {
			cafe_release_values(subtype, trav, allocator);
			trav += subtype->size;
		}
		CAFE_FREE(allocator, array->p);
	}

	default:
		;
	}
}


/***********************************************************************
 * File encoding routines
 */


/**
 *  Consume the stream header.  On success, sets 'version' to the file's format
 * version and returns CAFE_OK.
 */
int cafe_fdecode_stream_header(FILE *f, unsigned long *version)
{
	int tmp;
	unsigned long v;
	unsigned int i;
	bool version_found;

	assert(f != NULL);
	for (i = 0; i < sizeof(HEADER); ++i) {
		tmp = fgetc_unlocked(f);
		if (tmp != HEADER[i])
			return CAFE_ERROR;
	}

	version_found = false;
	v = 0;
	do {
		tmp = fgetc_unlocked(f);
		if (tmp == 0) {
			if ((!version_found) || (v != 0))
				return CAFE_ERROR;
			if (version != NULL)
				*version = v;
			return CAFE_OK;
		}
		if (!isdigit(tmp & 0xFF))
			return CAFE_ERROR;

		v *= 10;
		v += tmp - '0';
		version_found = true;
	} while(true);
}


/**
 *  Given an encoded byte stream, decode the event size and event id, returning
 * the number of bytes consumed from the stream, or a negative LEB128 status on
 * error.
 */
int cafe_fdecode_event_header(FILE *msg, unsigned long long *size,
			      unsigned long long *event)
{
	int size_bytes, eid_bytes;

	/* event id */
	eid_bytes = cafe_leb128_fread_ull(msg, event);
	if (eid_bytes < 0)
		return eid_bytes;

	/* event record size */
	size_bytes = cafe_leb128_fread_ull(msg, size);
	if (size_bytes < 0)
		return size_bytes;

	return size_bytes + eid_bytes;
}


static
size_t fread_str(FILE *in, const struct cafe_allocator *allocator,
		 const char **s)
{
	int sz;
	unsigned long n;
	char *str;

	sz = cafe_leb128_fread_ul(in, &n);
	if (sz < 0)
		return sz;

	if (n == 0) {
		*s = 0;
		return sz;
	}

	//td: avoid use of malloc here
	str = CAFE_ALLOC(allocator, n + 1);
	if (str != NULL) {
		if (1 != fread_unlocked(str, n, 1, in))
			return (size_t) -1;
		str[n] = 0;
	}
	*s = str;
	return sz+n;
}


/**
 *  Unpack the bytestream from "in" according to the given value type, and
 * populate "arg" with the results.  Returns the number of bytes consumed
 * from the input stream, or -1 on failure.
 */
ssize_t cafe_fdecode_val(struct cafe_value_type *vt_orig, void *arg,
			 const struct cafe_allocator *allocator, 
			 struct cafe_event_map *event_map, FILE *in)
{
#if defined(_MSC_VER)
#pragma warning(push)
#pragma warning(disable:4244)	/* conversion error: possible loss of data */
#endif
	ssize_t size;
	struct cafe_value_type *vt;

	if (!vt_orig)
		return 0;

	vt = cafe_base_type(vt_orig);
	assert(vt != NULL);
	switch (vt->kind) {
	case CAFE_TYPE_CHAR:
	{
		long x;
		size = cafe_leb128_fread_l(in, &x);
		*(char *) arg = (char) x;
		return size;
	}
	case CAFE_TYPE_UCHAR:
	{
		unsigned long x;
		size = cafe_leb128_fread_ul(in, &x);
		*(unsigned char *) arg = (unsigned char) x;
		return size;
	}
	case CAFE_TYPE_SHORT:
	{
		long x;
		size = cafe_leb128_fread_l(in, &x);
		*(short *) arg = (short) x;
		return size;
	}
	case CAFE_TYPE_USHORT:
	{
		unsigned long x;
		size = cafe_leb128_fread_ul(in, &x);
		*(unsigned short *) arg = (unsigned short) x;
		return size;
	}
	case CAFE_TYPE_INT:
	{
		long x;
		size = cafe_leb128_fread_l(in, &x);
		*(int *) arg = (int) x;
		return size;
	}
	case CAFE_TYPE_UINT:
	{
		unsigned long x;
		size = cafe_leb128_fread_ul(in, &x);
		*(unsigned int *) arg = (unsigned int) x;
		return size;
	}
	case CAFE_TYPE_LONG:
		return cafe_leb128_fread_l(in,  (long *) arg);
	case CAFE_TYPE_ULONG:
		return cafe_leb128_fread_ul(in,  (unsigned long *) arg);
	case CAFE_TYPE_LONGLONG:
		return cafe_leb128_fread_ll(in,  (long long *) arg);
	case CAFE_TYPE_ULONGLONG:
		return cafe_leb128_fread_ull(in,  (unsigned long long *) arg);
	case CAFE_TYPE_BOOL:
	{
		unsigned long x;
		size = cafe_leb128_fread_ul(in, &x);
		*(bool *) arg = (bool) x;
		return size;
	}
	case CAFE_TYPE_FLOAT:
		return fread_float(in,  (float *) arg);
	case CAFE_TYPE_DOUBLE:
		return fread_double(in,  (double *) arg);

	case CAFE_TYPE_STRING:
		return fread_str(in, allocator, (const char **) arg);

	case CAFE_TYPE_BLOB:
        {
		struct cafe_vm_blob *blob = (struct cafe_vm_blob *) arg;
		unsigned long len;

		size = cafe_leb128_fread_ul(in, &len);
		blob->l = len;
		if (blob->l == 0) {
			blob->p = 0;
			return size;
		}

		//td: avoid use of malloc here by referencing mapped stream address
		blob->p = CAFE_ALLOC(allocator, blob->l);
		if (blob->p == 0) {
			(void)printf("%s: Error allocating\n", __func__);
			abort();
		}
		if (1 != fread_unlocked(blob->p, blob->l, 1, in))
			return -1;
		return size + blob->l;
        }

	case CAFE_TYPE_ARRAY:
	{
		unsigned int i;
		struct cafe_value_type *element_type = cafe_element_type(vt);
		size_t element_size = element_type->size;
		size = 0;
		for (i = 0; i < vt->elements; ++i) {
			void *element = (char *) arg + (element_size * i);
			size += cafe_fdecode_val(element_type, element,
						 allocator, event_map, in);
		}
		return size;
	}

	case CAFE_TYPE_DYNARRAY:
	{
		unsigned int i;
		unsigned long elements;
		struct cafe_vm_dynarray *array =
			(struct cafe_vm_dynarray *) arg;
		struct cafe_value_type *element_type = cafe_element_type(vt);
		size_t element_size = element_type->size;

		/* count array elements */
		size = cafe_leb128_fread_ul(in, &elements);

		array->l = elements;
		array->c = elements;
		array->p = CAFE_ALLOC(allocator, element_size * elements);
		if (!array->p)
			return -1;

		for (i = 0; i < elements; ++i) {
			void *element = (char *)array->p
				+ (element_size * i);
			size += cafe_fdecode_val(element_type, element,
						 allocator, event_map, in);
		}

		return size;
	}

	case CAFE_TYPE_DICT:
	{
		struct cafe_dict_struct *dict = (struct cafe_dict_struct *) arg;
		struct cafe_value_type *element_type = cafe_element_type(vt);
		unsigned int i;
		unsigned long rows = 0;

		/* initialize dictionary */
		memset(dict, 0, sizeof(*dict));
		cafe_dict_create(dict, vt_orig);

		/* count dict rows */
		size = cafe_leb128_fread_ul(in, &rows);

		/* read each row in turn */
		for (i = 0; i < rows; ++i) {
			struct cafe_value_data *row =
				CAFE_ALLOC(allocator,
					   sizeof(*row) + element_type->size);
			if (row == NULL)
				return -1;
			row->type = element_type;
			size += cafe_fdecode_val(element_type, row->u.data,
						 allocator, event_map, in);
			if (cafe_dict_replace(dict, row) == NULL)
				return -1;
		}
		return size;
	}

	case CAFE_TYPE_STRUCT:
	{
		struct cafe_vm_attr *attr;
		size = 0;

		cafe_type_attr_foreach(attr, vt) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			size += cafe_fdecode_val(cafe_attr_type(attr),
						 (char *) arg + attr->offset,
						 allocator, event_map, in);
		}

		return size;
	}

	case CAFE_TYPE_FILTER:
		return 0;

	default:
		(void)fprintf(stderr, "%s: UNKNOWN TYPE: %d\n", __func__, vt->kind);
		return -1;
	}
#if defined(_MSC_VER)
#pragma warning(pop)
#endif
}


/**
 *  Given the parameter value type, and an array of any orthogonal data value
 * types, read all data into the given data area (*data).  The data area will
 * be resized if it's too small for the requested values.
 *
 * Clients should free() the data area after they will no longer reuse it, as
 * well as calling release_values() on the data area itself beforehand.
 *
 * WRS_TODO: make sure each struct in data is aligned. It might be better
 * to return an array of pointers to the data.  This is what the file
 * reader expects.
 */
int cafe_fread_event_data(FILE *input,
			  const struct cafe_event_descr *event,
			  const struct cafe_allocator *allocator,
			  struct cafe_event_map *event_map,
			  struct cafe_byte_buffer *data_buf)
{
	size_t offset;
	unsigned int i;
	char *buf;
	void **event_data;
	int rc;

	/* make sure we have enough space to read in the
	   non-variable-length data */
	offset = event->num_types * sizeof(void *);
	for (i = 0; i < event->num_types; ++i) {
	        if (event->types[i]->size == 0)
	            continue;
		offset = (offset + event->types[i]->align - 1) &
				(0 - event->types[i]->align);
		offset += event->types[i]->size;
	}

	rc = cafe_bbuf_capacity_set(data_buf, offset);
	if (rc != CAFE_OK)
		return rc;

	buf = data_buf->data;
	event_data = (void **)buf;
	offset = event->num_types * sizeof(void *);
	for (i = 0; i < event->num_types; ++i) {
		offset = (offset + event->types[i]->align - 1) &
				(0 - event->types[i]->align);
		if (event->types[i]->size != 0)
			event_data[i] = buf + offset;
		else
			event_data[i] = NULL;

		cafe_fdecode_val(event->types[i], buf + offset,
				 allocator, event_map, input);
		offset += event->types[i]->size;
	}

	return CAFE_OK;
}


/***********************************************************************
 * Stream encoding routines
 */

static ssize_t sencode_val(struct cafe_output_stream *out,
			   const struct cafe_value_type *vt, const void *arg);

static ssize_t slog_encoded_write_raw(
	struct cafe_output_stream *out, unsigned long long event_id,
	const struct cafe_event_descr *event, void **event_data);


/**
 *  Create a handle used for encoding event data to a buffered memory stream.
 * cafe_slog_encoded_delete() should be called to free this handle when
 * finished.
 */
int cafe_slog_encoded_create(
	struct cafe_instrument *session,
	struct cafe_output_stream *ostream,
	struct cafe_slog_encoded_inst **result)
{
	struct cafe_slog_encoded_inst *slog;
	struct cafe_value_type *type;
	int rc;

	assert(ostream != NULL);
	assert(result != NULL);
	slog = CAFE_ALLOC(session->allocator, sizeof(*slog));
	if (slog == NULL)
		return CAFE_NOMEM;

	slog->session = session;
	slog->out = ostream;
	slog->event_id = 1;
	memset(&slog->dict, 0, sizeof(struct cafe_dict_struct));

	rc = cafe_type_parse(slog->session->ns,
			     cafe_metadata_VT_DEF, &type);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_cfg_get_event(slog->session, 1, &type, &slog->metadata_event);
	if (rc != CAFE_OK)
		return rc;

	rc = cafe_type_parse(slog->session->ns,
			     CAFE_LOG_STREAM_MAP_DICT, &type);
	if (rc != CAFE_OK || type == NULL) {
		CAFE_FREE(session->allocator, slog);
		return CAFE_ERROR;
	}
	cafe_dict_create(&slog->dict, type);

	ostream->write_block(ostream, HEADER, sizeof(HEADER));
	assert(cafe_slog_version < 10);
	ostream->write(ostream, '0' + cafe_slog_version);
	ostream->write(ostream, 0);

	*result = slog;
	return CAFE_OK;
}


void cafe_slog_encoded_delete(struct cafe_slog_encoded_inst *slog)
{
	cafe_dict_delete(&slog->dict);
	CAFE_FREE(slog->session->allocator, slog);
}


static
ssize_t slog_encoded_write_raw(
	struct cafe_output_stream *out, unsigned long long event_id,
	const struct cafe_event_descr *event, void **event_data)
{
	unsigned int i;
	size_t msg_size;
	ssize_t bytes;

	/* calculate size of event data */
	msg_size = 0;
	for (i = 0; i < event->num_types; ++i)
		msg_size += sencode_val(0, event->types[i], event_data[i]);

	/* event id */
	bytes = cafe_leb128_swrite_ull(out, 0 + event_id);

	/* event record size */
	bytes += cafe_leb128_swrite_ull(out, (long long) msg_size);

	for (i = 0; i < event->num_types; ++i)
		bytes += sencode_val(out, event->types[i], event_data[i]);

	return bytes;
}


static
ssize_t cafe_slog_event_descr(
	struct cafe_slog_encoded_inst *slog,
	unsigned long long event_id,
	const struct cafe_event_descr *event,
	bool genref)
{
	unsigned int i;
	ssize_t bytes;
	struct cafe_metadata metadata;
	char **type_strings;
	void *output_event_data[1];

	type_strings = CAFE_ALLOC(
		slog->session->allocator,
		event->num_types * sizeof *type_strings);
	if (type_strings == NULL)
		return 0;
	metadata.eid = event_id;
	metadata.event_data.c = 0;
	metadata.event_data.l = event->num_types;
	metadata.event_data.p = type_strings;
	for (i = 0; i < event->num_types; ++i)
		if (!genref) {
			type_strings[i] = event->types[i]->type_str;
		} else {
			assert(cafe_base_type(event->types[i]) != NULL);
			type_strings[i] = cafe_type_ref_str(
					    cafe_base_type(event->types[i]));
		}

	output_event_data[0] = &metadata;
	bytes = slog_encoded_write_raw(
		slog->out, 0, slog->metadata_event, output_event_data);
	if (genref)
		for (i = 0; i < event->num_types; ++i)
			CAFE_FREE(slog->session->allocator, type_strings[i]);
	CAFE_FREE(slog->session->allocator, type_strings);
	return bytes;
}


static
bool cafe_slog_writer_pre_log_type(struct cafe_value_type *vt, void *arg)
{
	struct cafe_slog_encoded_inst *slog = arg;
	cafe_any_value keys[1];
	struct cafe_value_data *entry;
	struct cafe_log_event_map *map;

	switch(vt->kind) {
	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_TYPE:
	case CAFE_TYPE_STRING:
	case CAFE_TYPE_BLOB:
	case CAFE_TYPE_FILTER:
		return true;
	}
 	keys[0].ull = (unsigned long)vt;
	entry = cafe_dict_get(&slog->dict, keys);
	assert(entry);
	map = (struct cafe_log_event_map *)entry->u.data;
	if (map->event_id != 0)
		return true;
	map->event_id = slog->event_id++;
	return false;
}


static void cafe_slog_writer_post_log_type(struct cafe_value_type *vt,
					   void *arg)
{
	struct cafe_slog_encoded_inst *slog = arg;
	cafe_any_value keys[1];
	struct cafe_value_data *entry;
	struct cafe_log_event_map *map;
	struct cafe_event_descr event;

	switch(vt->kind) {
	case CAFE_TYPE_DICT:
	case CAFE_TYPE_ARRAY:
	case CAFE_TYPE_DYNARRAY:
	case CAFE_TYPE_INSTANCE:
	case CAFE_TYPE_FILTER:
		return;
	}

 	keys[0].ull = (unsigned long)vt;
	entry = cafe_dict_lookup(&slog->dict, keys);
	assert(entry);
	map = (struct cafe_log_event_map *)entry->u.data;
	assert(map->event_id != 0);

	event.num_types = 1;
	event.types[0] = vt;
	(void)cafe_slog_event_descr(slog, map->event_id, &event, false);
}


/**
 *  If the given event has not been expressed in the event stream yet, emit its
 * description and add it to the types table for future lookups.  Returns the
 * stream event id, or 0 on error.
 */
static
unsigned long long slog_get_stream_eid(struct cafe_slog_encoded_inst *slog,
				  const struct cafe_event_descr *event)
{
	cafe_any_value keys[1];
	struct cafe_value_data *data;
	struct cafe_log_event_map *map;
	unsigned int i;

	if (event->num_types == 1) {
		keys[0].ull = (unsigned long) event->types[0];
	} else {
		keys[0].ull = (unsigned long) event;
	}

	data = cafe_dict_lookup(&slog->dict, keys);
	if (data != NULL) {
		map = (struct cafe_log_event_map *)data->u.data;
		return map->event_id;
	}

	for (i = 0; i < event->num_types; ++i) {
		cafe_type_iter_dep_types(event->types[i],
					 cafe_slog_writer_pre_log_type,
					 cafe_slog_writer_post_log_type,
					 slog);
	}

	data = cafe_dict_get(&slog->dict, keys);
	if (data == NULL)
		return 0;
	map = (struct cafe_log_event_map *)data->u.data;
	if (map->event_id == 0) {
		map->event_id = slog->event_id++;
		(void)cafe_slog_event_descr(slog, map->event_id, event, true);
	}
	return map->event_id;
}


/**
 *  Write an event to a memory address.  This is an atomic operation; the event
 * will be fully written at the return of this function.  Both the event and
 * its data payload will be written.
 *
 * Returns the number of bytes written to the stream, or an leb128 error code
 * on failure.
 */
ssize_t cafe_slog_encoded_write(struct cafe_slog_encoded_inst *slog,
	const struct cafe_event_descr *event,
	void *event_data[])
{
	unsigned long long stream_id;

	stream_id = slog_get_stream_eid(slog, event);
	return slog_encoded_write_raw(slog->out, stream_id, event, event_data);
}


/**
 *  Given the parameter value type, and an array of any orthogonal data value
 * types, read all data into the given data area (*data).  The data area will
 * be resized if it's too small for the requested values.
 *
 * Clients should free() the data area after they will no longer reuse it, as
 * well as calling release_values() on the data area itself beforehand.
 *
 * WRS_TODO: make sure each struct in data is aligned. It might be better
 * to return an array of pointers to the data.  This is what the file
 * reader expects.
 */
int cafe_mread_event_data(int8_t *input,
			  const struct cafe_event_descr *event,
			  const struct cafe_allocator *allocator,
			  struct cafe_event_map *event_map,
			  struct cafe_byte_buffer *data_buf)
{
	size_t offset;
	unsigned int i;
	char *buf;
	void **event_data;
	int rc;

	/* make sure we have enough space to read in the
	   non-variable-length data */
	offset = event->num_types * sizeof(void *);
	for (i = 0; i < event->num_types; ++i) {
		offset = (offset + event->types[i]->align - 1) &
				(0 - event->types[i]->align);
		offset += event->types[i]->size;
	}

	rc = cafe_bbuf_capacity_set(data_buf, offset);
	if (rc != CAFE_OK)
		return rc;
	data_buf->len = offset;

	buf = data_buf->data;
	event_data = (void **)buf;
	offset = event->num_types * sizeof(void *);
	for (i = 0; i < event->num_types; ++i) {
		offset = (offset + event->types[i]->align - 1) &
				(0 - event->types[i]->align);
		if (event->types[i]->size != 0)
			event_data[i] = buf + offset;
		else
			event_data[i] = NULL;

		assert(input != NULL);
		input = cafe_mdecode_val(event->types[i], buf + offset,
					 allocator, event_map, input);
		offset += event->types[i]->size;
	}

	return CAFE_OK;
}


/**
 *  Consume the stream header.  On success, sets 'version' to the stream's
 * format version and returns CAFE_OK.
 */
int8_t *cafe_mdecode_stream_header(int8_t *msg, size_t msg_size,
				   unsigned long *version)
{
	unsigned long v;
	bool version_found;

	assert(msg != NULL);
	if ((msg_size < sizeof(HEADER))
	    || (memcmp(msg, HEADER, sizeof(HEADER)) != 0))
		return NULL;

	msg += sizeof(HEADER);
	version_found = false;
	v = 0;
	while (*msg != 0) {
		if (!isdigit(*msg))
			return NULL;

		v *= 10;
		v += *msg - '0';
		version_found = true;
		++msg;
	};

	if ((!version_found) || (v != 0))
		return NULL;
	if (version != NULL)
		*version = v;
	return ++msg;
}


/**
 *  Given an encoded byte stream, decode the event size and event id, returning
 * a pointer to the next unread byte in the stream, or NULL on error.
 */
int8_t *cafe_mdecode_event_header(int8_t *msg, size_t msg_size,
				  unsigned long long *size,
				  unsigned long long *event)
{
	int8_t *original_msg = msg;

	/* event id */
	msg = cafe_leb128_mread_ull_guarded(msg, msg_size, event);
	if (!msg)
		return NULL;
	msg_size -= msg - original_msg;

	/* event record size */
	msg = cafe_leb128_mread_ull_guarded(msg, msg_size, size);
	if (!msg)
		return NULL;

	return msg;
}


static
char *mread_str(char *in, const struct cafe_allocator *allocator,
		const char **s)
{
	unsigned long n;
	char *str;

	in = cafe_leb128_mread_ul(in, &n);
	if (n == 0) {
		*s = 0;
		return in;
	}
	assert(in != NULL);

	//td: avoid use of malloc here by referencing mapped stream address
	str = CAFE_ALLOC(allocator, n + 1);
	if (str == 0) {
		(void)printf("%s: Error allocating\n", __func__);
		abort();
	}
	if(str != NULL)
		memcpy(str, in, n);
	str[n] = 0;
	*s = str;
	return in + n;
}


static
size_t swrite_bytes(struct cafe_output_stream *out, const char *s, size_t len)
{
	size_t sz;

	if (s == 0)
		return cafe_leb128_swrite_ull(out, 0);

	sz = cafe_leb128_swrite_ull(out, len);
	if (out != NULL)
		out->write_block(out, s, len);
	return sz + len;
}


/**
 *  Encode the data starting at "arg" into memory starting at "out", according
 * to the types given by "vt".
 * Returns the number of bytes written, or an leb128 error code.
 */
static
ssize_t sencode_val(struct cafe_output_stream *out,
		    const struct cafe_value_type *vt, const void *arg)
{
	vt = cafe_base_type(vt);
	assert(vt != NULL);
	switch (vt->kind) {
	case CAFE_TYPE_CHAR:
		return cafe_leb128_swrite_ll(out,  *(char *) arg);
	case CAFE_TYPE_UCHAR:
		return cafe_leb128_swrite_ull(out, *(unsigned char *) arg);
	case CAFE_TYPE_SHORT:
		return cafe_leb128_swrite_ll(out,  *(short *) arg);
	case CAFE_TYPE_USHORT:
		return cafe_leb128_swrite_ull(out, *(unsigned short *) arg);
	case CAFE_TYPE_INT:
		return cafe_leb128_swrite_ll(out,  *(int *) arg);
	case CAFE_TYPE_UINT:
		return cafe_leb128_swrite_ull(out, *(unsigned int *) arg);
	case CAFE_TYPE_LONG:
		return cafe_leb128_swrite_ll(out,  *(long *) arg);
	case CAFE_TYPE_ULONG:
		return cafe_leb128_swrite_ull(out, *(unsigned long *) arg);
	case CAFE_TYPE_LONGLONG:
		return cafe_leb128_swrite_ll(out,  *(long long *) arg);
	case CAFE_TYPE_ULONGLONG:
		return cafe_leb128_swrite_ull(out, *(unsigned long long *) arg);
	case CAFE_TYPE_BOOL:
		if (*(bool *) arg)
			return cafe_leb128_swrite_ull(out,  1);
		else
			return cafe_leb128_swrite_ull(out,  0);
	case CAFE_TYPE_FLOAT:
		return swrite_float(out, (float *) arg);
	case CAFE_TYPE_DOUBLE:
		return swrite_double(out, (double *) arg);

	case CAFE_TYPE_STRING:
	{
		const char *str = *(char **) arg;

		if (str == 0)
			return cafe_leb128_swrite_ull(out, 0);

		return swrite_bytes(out, str, strlen(str));
	}

	case CAFE_TYPE_BLOB:
        {
		const struct cafe_vm_blob *blob = (struct cafe_vm_blob *) arg;
		return swrite_bytes(out, blob->p, blob->l);
        }

	case CAFE_TYPE_ARRAY:
	{
		unsigned int i;
		size_t size = 0;

		for (i = 0; i < vt->elements; ++i) {
			const void *element =
				(char *)arg + (cafe_element_type(vt)->size * i);
			size += sencode_val(out, cafe_element_type(vt),
					    element);
		}
		return size;
	}

	case CAFE_TYPE_DYNARRAY:
	{
		unsigned int i;
		struct cafe_vm_dynarray *array =
			(struct cafe_vm_dynarray *) arg;

		/*  write element count */
		size_t size = cafe_leb128_swrite_ull(out, array->l);
		for (i = 0; i < array->l; ++i) {
			void *element = (char *) array->p
				+ (cafe_element_type(vt)->size * i);
			size += sencode_val(out, cafe_element_type(vt),
					    element);
		}

		return size;
	}

	case CAFE_TYPE_STRUCT:
	{
		struct cafe_vm_attr *attr;
		size_t bytes = 0;

		cafe_type_attr_foreach(attr, vt) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			bytes += sencode_val(out, cafe_attr_type(attr),
					     (char *)arg + attr->offset);
		}

		return bytes;
	}

	case CAFE_TYPE_DICT:
	{
		struct cafe_dict_struct *dict = (struct cafe_dict_struct *) arg;
		size_t size = 0;
		struct cafe_dict_iterator iter;
		struct cafe_value_data *val;

		size = cafe_leb128_swrite_ull(out, cafe_dict_count_rows(dict));

		/* emit each row in turn */
		val = cafe_dict_first(dict, &iter);
		for (;val != 0; val = cafe_dict_next(dict, &iter)) {
			size += sencode_val(out, cafe_element_type(vt),
					    val->u.data);
		}
		return size;
	}

	case CAFE_TYPE_FILTER:
		return 0;

	default:
		(void)fprintf(stderr, "%s: UNKNOWN TYPE: %d\n", __func__, vt->kind);
		return -1;
	}
}


int8_t *cafe_mdecode_val(struct cafe_value_type *vt_orig, void *arg,
			 const struct cafe_allocator *allocator,
			 struct cafe_event_map *event_map, int8_t *in)
{
#if defined(_MSC_VER)
#pragma warning(push)
#pragma warning(disable:4244)	/* conversion error: possible loss of data */
#endif
	int8_t *pos;
	struct cafe_value_type *vt;

	vt = cafe_base_type(vt_orig);
	assert(vt != NULL);
	switch (vt->kind) {
	case CAFE_TYPE_CHAR:
	{
		long x;
		pos = cafe_leb128_mread_l(in, &x);
		*(char *) arg = (char) x;
		return pos;
	}
	case CAFE_TYPE_UCHAR:
	{
		unsigned long x;
		pos = cafe_leb128_mread_ul(in, &x);
		*(unsigned char *) arg = (unsigned char) x;
		return pos;
	}
	case CAFE_TYPE_SHORT:
	{
		long x;
		pos = cafe_leb128_mread_l(in, &x);
		*(short *) arg = (short) x;
		return pos;
	}
	case CAFE_TYPE_USHORT:
	{
		unsigned long x;
		pos = cafe_leb128_mread_ul(in, &x);
		*(unsigned short *) arg = (unsigned short) x;
		return pos;
	}
	case CAFE_TYPE_INT:
	{
		long x;
		pos = cafe_leb128_mread_l(in, &x);
		*(int *) arg = (int) x;
		return pos;
	}
	case CAFE_TYPE_UINT:
	{
		unsigned long x;
		pos = cafe_leb128_mread_ul(in, &x);
		*(unsigned int *) arg = (unsigned int) x;
		return pos;
	}
	case CAFE_TYPE_LONG:
		return cafe_leb128_mread_l(in,  (long *) arg);
	case CAFE_TYPE_ULONG:
		return cafe_leb128_mread_ul(in,  (unsigned long *) arg);
	case CAFE_TYPE_LONGLONG:
		return cafe_leb128_mread_ll(in,  (long long *) arg);
	case CAFE_TYPE_ULONGLONG:
		return cafe_leb128_mread_ull(in,  (unsigned long long *) arg);
	case CAFE_TYPE_BOOL:
	{
		unsigned long x;
		pos = cafe_leb128_mread_ul(in, &x);
		*(bool *)arg = (bool) x;
		return pos;
	}
	case CAFE_TYPE_FLOAT:
		return (int8_t *)mread_float((char *)in,  (float *) arg);
	case CAFE_TYPE_DOUBLE:
		return (int8_t *)mread_double((char *)in,  (double *) arg);

	case CAFE_TYPE_STRING:
		return (int8_t *)mread_str((char *)in, allocator, (const char **) arg);

	case CAFE_TYPE_BLOB:
        {
		struct cafe_vm_blob *blob = (struct cafe_vm_blob *) arg;
		unsigned long len;

		in = cafe_leb128_mread_ul(in, &len);
		blob->l = len;
		if (blob->l == 0) {
			blob->p = 0;
			return in;
		}

		//td: avoid use of malloc here by referencing mapped stream address
		blob->p = CAFE_ALLOC(allocator, blob->l);
		if (blob->p == 0) {
			(void)printf("%s: Error allocating\n", __func__);
			abort();
		}
		assert(in != NULL);
		memcpy(blob->p, in, blob->l);
		return in + blob->l;
        }

	case CAFE_TYPE_ARRAY:
	{
		unsigned int i;
		for (i = 0; i < vt->elements; ++i) {
			void *element = (char *) arg
				+ (cafe_element_type(vt)->size * i);
			assert(in != NULL);
			in = cafe_mdecode_val(cafe_element_type(vt), element,
					      allocator, event_map, in);
		}
		return in;
	}

	case CAFE_TYPE_DYNARRAY:
	{
		unsigned int i;
		unsigned long elements;
		struct cafe_vm_dynarray *array =
			(struct cafe_vm_dynarray *) arg;

		/* count array elements */
		in = cafe_leb128_mread_ul(in, &elements);

		array->l = elements;
		array->c = elements;
		array->p = CAFE_ALLOC(allocator,
				      cafe_element_type(vt)->size * elements);
		if (!array->p)
			return 0;

		for (i = 0; i < elements; ++i) {
			void *element = (char *) array->p
				+ (cafe_element_type(vt)->size * i);
			assert(in != NULL);
			in = cafe_mdecode_val(cafe_element_type(vt), element,
					      allocator, event_map, in);
		}

		return in;
	}

	case CAFE_TYPE_DICT:
	{
		struct cafe_dict_struct *dict = (struct cafe_dict_struct *) arg;
		struct cafe_value_type *element_type = cafe_element_type(vt);
		unsigned int i;
		unsigned long rows = 0;

		/* initialize dictionary */
		memset(dict, 0, sizeof(*dict));
		cafe_dict_create(dict, vt_orig);

		/* count dict rows */
		in = cafe_leb128_mread_ul(in, &rows);

		/* read each row in turn */
		for (i = 0; i < rows; ++i) {
			struct cafe_value_data *row =
				CAFE_ALLOC(allocator,
					   sizeof(*row) + element_type->size);
			if (row == NULL)
				return 0;
			row->type = element_type;
			in = cafe_mdecode_val(element_type, row->u.data,
					      allocator, event_map, in);
			if (cafe_dict_replace(dict, row) == NULL)
				return 0;
		}
		return in;
	}

	case CAFE_TYPE_STRUCT:
	{
		struct cafe_vm_attr *attr;

		cafe_type_attr_foreach(attr, vt) {
			if (attr->kind != CAFE_ATTR_VAR &&
			    attr->kind != CAFE_ATTR_KEY)
				continue;
			assert(in != NULL);
			in = cafe_mdecode_val(cafe_attr_type(attr),
					      (char *) arg + attr->offset,
					      allocator, event_map, in);
		}

		return in;
	}

	case CAFE_TYPE_FILTER:
		return in;

	default:
		(void)fprintf(stderr, "%s: UNKNOWN TYPE: %d\n", __func__, vt->kind);
		return 0;
	}
#if defined(_MSC_VER)
#pragma warning(pop)
#endif
}


struct logstream {
	struct cafe_output_stream ostream;
	struct cafe_log_buf *buf;
	void (*full_callback)(void *client_data);
	void *client_data;
};


/**
 *  Sleep for the given delay period, then adjust the delay period upwards for
 * subsequent delays.
 */
static inline
void adaptive_delay(struct timespec *t)
{
	(void)nanosleep(t, 0);
	t->tv_nsec *= 2;
	if (t->tv_nsec >= NSECS_PER_SEC)
		/* cap adaptive delay at 1 second */
		t->tv_nsec = NSECS_PER_SEC - 1;
}


/**
 * Write a single byte to the given log stream; blocking when the log buffer
 * is full.
 */
static
void logstream_write(struct cafe_output_stream *stream, int byte)
{
	struct logstream *lstream = (struct logstream *) stream;
	struct timespec delay;
	char *msg = cafe_log_write_reserve(lstream->buf, 1);

	timespec_from_ticks(500000, NSECS_PER_SEC, &delay);
	while (msg == NULL) {
		if (lstream->full_callback != NULL)
			lstream->full_callback(lstream->client_data);
		adaptive_delay(&delay);
		msg = cafe_log_write_reserve(lstream->buf, 1);
	}
	*msg = (char) byte;
	cafe_log_write_commit(lstream->buf, 1);
}


/**
 * Write a block of bytes to the given log stream; blocking when the log buffer
 * is full.
 */
static
void logstream_write_block(struct cafe_output_stream *stream, const char *block,
			   size_t size)
{
	struct logstream *lstream = (struct logstream *) stream;
	size_t avail;
	size_t written = 0;
	char *msg;
	struct timespec delay;

	timespec_from_ticks(500000, NSECS_PER_SEC, &delay);
	while (written < size) {
		msg = cafe_log_write_reserve_min(lstream->buf, 1, &avail);

		while (msg == NULL) {
			if (lstream->full_callback != NULL)
				lstream->full_callback(lstream->client_data);
			adaptive_delay(&delay);
			msg = cafe_log_write_reserve_min(lstream->buf, 1, &avail);
		}
		if (avail > (size - written))
			avail = size - written;
		memcpy(msg, block + written, avail);
		cafe_log_write_commit(lstream->buf, avail);
		written += avail;
	}
}


/**
 *  Create a new memory-buffered output stream, utilizing the given log buffer.
 * The result should be freed by the client when not in use.
 */
int cafe_logstream_open_output(struct cafe_log_buf *buf,
			       const struct cafe_allocator *allocator,
			       struct cafe_output_stream **result,
			       void (*full_callback)(void *client_data),
			       void *client_data)
{
	struct logstream *lstream;

	assert(result != NULL);
	assert(allocator != NULL);

	lstream = CAFE_ALLOC(allocator, sizeof(*lstream));
	if (lstream == NULL)
		return CAFE_NOMEM;

	lstream->buf = buf;
	lstream->ostream.write = logstream_write;
	lstream->ostream.write_block = logstream_write_block;
	lstream->full_callback = full_callback;
	lstream->client_data = client_data;

	*result = &lstream->ostream;
	return CAFE_OK;
}


struct filestream {
	struct cafe_output_stream ostream;
	FILE *file;
};


/**
 * Write a single byte to the given log stream.
 */
static
void filestream_write(struct cafe_output_stream *stream, int byte)
{
	struct filestream *fstream = (struct filestream *) stream;
	(void)fputc_unlocked(byte, fstream->file);
}


/**
 * Write a block of bytes to the given log stream.
 */
static
void filestream_write_block(struct cafe_output_stream *stream,
			    const char *block, size_t size)
{
	struct filestream *fstream = (struct filestream *) stream;

	if (size == 0)
		return;

	if (fwrite_unlocked(block, size, 1, fstream->file) != 1)
		perror("Error writing file stream");
}


/**
 *  Create a new file output stream.  The result should be freed by the client
 * when not in use.
 */
int cafe_filestream_open_output(FILE *out,
				const struct cafe_allocator *allocator,
				struct cafe_output_stream **result)
{
	struct filestream *fstream;

	assert(result != NULL);
	assert(allocator != NULL);

	fstream = CAFE_ALLOC(allocator, sizeof(*fstream));
	if (fstream == NULL)
		return CAFE_NOMEM;

	fstream->file = out;
	fstream->ostream.write = filestream_write;
	fstream->ostream.write_block = filestream_write_block;

	*result = &fstream->ostream;
	return CAFE_OK;
}
