/* utils.c - general routines */

/* Copyright (c) 2010-2013 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include "cafe_os.h"
#include <assert.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>		/* strcmp */
#include <stddef.h>		/* size_t */
#include <cafe/utils.h>

/*
 * Print into memory obtained from malloc().  sprintf fmt/args into a
 * buffer that grows to accomodate the result. Returns NULL if there
 * was insufficient memory, or a pointer to the formatted message on
 * success.
 */
char *cafe_mprintf(const struct cafe_allocator *allocator, const char *fmt, ...)
{
	/* Guess we need no more than 32 bytes. */
	int n, size = 32;
	char *p, *np;
	va_list ap;

	if ((p = CAFE_ALLOC(allocator, size)) == NULL)
		return NULL;

	while (1) {
		/* Try to print in the allocated space. */

		va_start(ap, fmt);
		n = vsnprintf(p, size, fmt, ap);
		va_end(ap);

		/* If that worked, return the string. */

		if (n > -1 && n < size)
			return p;

		/* Else try again with more space. */

		if (n > -1)	    /* glibc 2.1 */
			size = n+1; /* precisely what is needed */
		else		    /* glibc 2.0 */
			size *= 2;  /* twice the old size */

		if ((np = CAFE_REALLOC(allocator, p, size)) == NULL) {
			CAFE_FREE(allocator, p);
			return NULL;
		} else {
			p = np;
		}
	}
}

/*
 *  Enable all output events listed in the spec.
 */
int cafe_inst_enable_all_output_events(struct cafe_ep_inst *inst)
{
	unsigned int i;
	int rc;

	for (i = 0; i < inst->spec->n_output_events; i++) {
		struct cafe_event_descr *event;
		struct cafe_value_type *type;

		rc = cafe_type_parse(inst->session->ns,
				     inst->spec->output_events[i].type,
				     &type);
		if (rc != CAFE_OK || type == NULL)
			continue;
		rc = cafe_cfg_get_event(inst->session, 1, &type, &event);
		if (rc != CAFE_OK)
			continue;
		cafe_ep_inst_enable_output_event(inst, event);
	}
	return CAFE_OK;
}

static
int cafe_type_list_parse2(
	struct cafe_instrument *session,
	const char **sp,
	struct cafe_event_descr **result)
{
	int rc;
	int typeno;
	const char *s = *sp;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];

	typeno = 0;
	for (;;) {
		if (typeno >= CAFE_NELEMENTS(types))
			return CAFE_INVALID_VALUE_TYPE_STR;
		rc = cafe_type_parse2(session->ns, &s, ",", &types[typeno]);
		if (rc != CAFE_OK)
			return rc;
		assert(s != NULL && types[typeno] != NULL);
		typeno++;
		if (*s != ',')
			break;
		s++;
	}
	rc = cafe_cfg_get_event(session, typeno, types, result);
	if (rc != CAFE_OK)
		return rc;
	*sp = s;
	return CAFE_OK;
}

int cafe_type_list_parse(
	struct cafe_instrument *session,
	const char *s,
	struct cafe_event_descr **result)
{
	int rc;

	rc = cafe_type_list_parse2(session, &s, result);
	if (rc != CAFE_OK)
		return rc;
	if (*s != '\0')
		return CAFE_INVALID_VALUE_TYPE_STR;
	return CAFE_OK;
}
